﻿using System;
using System.Collections.Generic;

using Nabu.Diagnostics;

namespace DeviceManagerExample
{
	class Program
	{
		static void Main(string[] args)
		{
			Console.WindowWidth = Console.LargestWindowWidth;
			Console.BufferHeight = 16384;

			DeviceManager dm = new DeviceManager();

			foreach (DeviceClass dc in dm.Classes)
			{
				Console.ForegroundColor = ConsoleColor.Yellow;
				Console.WriteLine();
				Console.WriteLine("[{0}, {1}]", dc.Name, dc.Description);

				using (DeviceInstanceCollection dic = dc.GetInstances())
				{
					foreach (DeviceInstance di in dic)
					{
						Console.WriteLine();

						#region Properties

						Console.ForegroundColor = ConsoleColor.Green;
						Console.WriteLine("  [Properties]");

						foreach (DevicePropertyID devicePropertyID in Enum.GetValues(typeof(DevicePropertyID)))
						{
							string propertyName = (devicePropertyID + ":").PadRight(32);
							object propertyValue;

							if (di.Properties.TryGetValue(devicePropertyID, out propertyValue))
							{
								Type propertyType = (propertyValue == null) ? null : propertyValue.GetType();

								if (propertyType == typeof(int))
								{
									int value = (int)propertyValue;

									if (value != -1)
									{
										Console.ForegroundColor = ConsoleColor.White;

										switch (devicePropertyID)
										{
											case DevicePropertyID.ConfigFlags:
												Console.WriteLine(
													"    {0}{1}",
													propertyName,
													((DeviceConfigFlags)value));
												break;
											case DevicePropertyID.Capabilities:
												Console.WriteLine(
													"    {0}{1}",
													propertyName,
													((DeviceCapabilityFlags)value));
												break;
											case DevicePropertyID.LegacyBusType:
												Console.WriteLine(
													"    {0}{1}",
													propertyName,
													((DeviceInterfaceType)value));
												break;
											case DevicePropertyID.DeviceType:
												Console.WriteLine(
													"    {0}{1}",
													propertyName,
													((FileDeviceType)value));
												break;
											case DevicePropertyID.RemovalPolicy:
											case DevicePropertyID.RemovalPolicyHWDefault:
											case DevicePropertyID.RemovalPolicyOverride:
												Console.WriteLine(
													"    {0}{1}",
													propertyName,
													((DeviceRemovalPolicy)value));
												break;
											case DevicePropertyID.InstallState:
												Console.WriteLine(
													"    {0}{1}",
													propertyName,
													((DeviceInstallState)value));
												break;
											default:
												Console.WriteLine(
													"    {0}{1}",
													propertyName,
													value);
												break;
										}

									}
								}
								else if (propertyType == typeof(string))
								{
									string value = (string)propertyValue;

									if (value != null)
									{
										Console.ForegroundColor = ConsoleColor.Gray;
										Console.WriteLine(
											"    {0}{1}",
											propertyName,
											value);
									}
								}
								else if (propertyType == typeof(string[]))
								{
									string[] value = (string[])propertyValue;

									if (value.Length > 0)
									{
										Console.ForegroundColor = ConsoleColor.DarkGreen;
										Console.WriteLine("    {0}", propertyName);

										foreach (string item in value)
										{
											Console.WriteLine("      {0}", item);
										}
									}
								}
							}
						}

						#endregion

						#region Status

						if ((di.Flags & DeviceInstanceFlags.HAS_PROBLEM) != 0)
						{
							Console.ForegroundColor = ConsoleColor.Red;
						}
						else
						{
							Console.ForegroundColor = ConsoleColor.Gray;
						}

						Console.WriteLine("  [Status]");

						Console.WriteLine("    {0}", di.Flags);
						Console.WriteLine("    {0}", di.Problem);

						#endregion

						#region Resources

						ICollection<DeviceMemoryRange> memoryRanges = di.GetMemoryResources();
						ICollection<DevicePortRange> portRanges = di.GetPortResources();
						ICollection<DeviceDmaRange> dmaRanges = di.GetDmaResources();
						ICollection<DeviceIrqRange> irqRanges = di.GetIrqResources();

						if (
							(memoryRanges.Count > 0) ||
							(portRanges.Count > 0) ||
							(dmaRanges.Count > 0) ||
							(irqRanges.Count > 0))
						{
							Console.ForegroundColor = ConsoleColor.Green;
							Console.WriteLine("  [Resources]");
							Console.ForegroundColor = ConsoleColor.Cyan;
						}

						if (memoryRanges.Count > 0)
						{
							Console.WriteLine("    Memory:");

							foreach (DeviceMemoryRange range in memoryRanges)
							{
								if (range.Is32Bit)
								{
									Console.WriteLine("      0x{0:X8}-0x{1:X8}; {2}", range.Min, range.Max, range.Flags);
								}
								else
								{
									Console.WriteLine("      0x{0:X16}-0x{1:X16}; {2}", range.Min, range.Max, range.Flags);
								}
							}
						}

						if (portRanges.Count > 0)
						{
							Console.WriteLine("    Ports:");

							foreach (DevicePortRange range in portRanges)
							{
								Console.WriteLine("      0x{0:X4}-0x{1:X4}; {2}", range.Min, range.Max, range.Flags);
							}
						}

						if (dmaRanges.Count > 0)
						{
							Console.WriteLine("    DMA:");

							foreach (DeviceDmaRange range in dmaRanges)
							{
								Console.WriteLine("      Channel {0}; 0x{1:X4}-0x{2:X4}; {3}", range.Channel, range.Min, range.Max, range.Flags);
							}
						}

						if (irqRanges.Count > 0)
						{
							Console.WriteLine("    IRQ:");

							foreach (DeviceIrqRange range in irqRanges)
							{
								Console.WriteLine("      Number {0}; 0x{1:X4}-0x{2:X4}; {3}", range.Number, range.Min, range.Max, range.Flags);
							}
						}

						#endregion

						#region Driver

						ICollection<DeviceDriver> drivers = di.GetDrivers();

						if (drivers.Count > 0)
						{
							Console.ForegroundColor = ConsoleColor.Green;
							Console.WriteLine("  [Drivers]");
							Console.ForegroundColor = ConsoleColor.Magenta;

							foreach (DeviceDriver driver in drivers)
							{
								Console.WriteLine(
									"    Manufacturer: {0}",
									driver.Manufacturer);
								Console.WriteLine(
									"    Provider:     {0}",
									driver.Provider);
								Console.WriteLine(
									"    Description:  {0}",
									driver.Description);
								Console.WriteLine(
									"    Release Date: {0}",
									driver.ReleaseDate);
								Console.WriteLine(
									"    Version:      {0}",
									driver.Version);

								if (driver.Files.Count > 0)
								{
									Console.WriteLine("    Files:");

									foreach (string file in driver.Files)
									{
										Console.WriteLine("      {0}", file);
									}
								}
							}
						}

						#endregion
					}
				}
			}

			Console.ForegroundColor = ConsoleColor.Gray;
		}
	}
}
