﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Threading;
using System.Windows.Forms;

using Nabu;
using Nabu.Forms;
using Nabu.Mobile;

namespace MobileDeviceDumpExample
{
	public partial class MainForm : Form
	{
		#region Types

		private enum WorkerCommand
		{
			None,
			DumpFileSystem,
			DumpRegistry,
			Quit
		}

		private sealed class MobileDeviceDescriptor
		{
			#region Fields

			private Guid _id;
			private string _name;

			#endregion

			#region Methods

			public MobileDeviceDescriptor(Guid id, string name)
			{
				this._id = id;
				this._name = name;
			}

			#endregion

			#region Properties

			public Guid ID
			{
				get { return this._id; }
			}

			public string Name
			{
				get { return this._name; }
			}

			#endregion
		}

		#endregion

		#region Fields

		private Thread _workerThread;
		private ManualResetEvent _workerEvent;
		private WorkerCommand _workerCommand;
		private Guid _workerDeviceID;
		private List<MobileDeviceDescriptor> _deviceDescriptors;
		private bool _isUpdatingUI;

		#endregion

		#region Methods

		#region Constructors

		public MainForm()
		{
			InitializeComponent();

			this._workerThread = new Thread(ThreadProc);
			this._workerEvent = new ManualResetEvent(false);
			this._workerCommand = WorkerCommand.None;
			this._workerThread.SetApartmentState(ApartmentState.MTA);
			this._workerThread.Start();
		}

		#endregion

		#region Dump FileSystem

		private void EnumFileSystem(
			MobileSession mobileSession,
			string sourcePath,
			string destinationPath)
		{
			if (sourcePath == "\\Storage Card\\")
			{
				return;
			}

			foreach (MobileFileFindInfo fileInfo in mobileSession.GetFiles(sourcePath + "*.*"))
			{
				if ((fileInfo.Attributes & FileAttributes.Directory) != 0)
				{
					string newSourcePath = sourcePath + fileInfo.Name + "\\";
					string newDestinationPath = destinationPath + fileInfo.Name + "\\";

					Directory.CreateDirectory(newDestinationPath);

					CrossThreadControlAccessor.SetText(
						this._labelStatus,
						newSourcePath);

					EnumFileSystem(
						mobileSession,
						newSourcePath,
						newDestinationPath);

					CrossThreadControlAccessor.SetText(
						this._labelStatus,
						sourcePath);
				}
				else
				{
					try
					{
						using (Stream mobileStream =
							mobileSession.CreateFileStream(
								sourcePath + fileInfo.Name,
								FileMode.Open,
								FileAccess.Read,
								FileShare.ReadWrite))
						{
							using (Stream hostStream =
								new FileStream(
									destinationPath + fileInfo.Name,
									FileMode.Create,
									FileAccess.Write,
									FileShare.ReadWrite))
							{
								byte[] buffer = new byte[65536];
								int length = (int)mobileStream.Length;
								int position = 0;

								for (; position < length; )
								{
									int count =
										mobileStream.Read(
											buffer,
											0,
											Math.Min(length - position, buffer.Length));

									hostStream.Write(buffer, 0, count);

									position += count;
								}
							}
						}
					}
					catch
					{
					}
				}
			}
		}

		#endregion

		#region Dump Registry

		private void EnumRegistryKey(
			TextWriter textWriter,
			MobileRegistryKey registryKey,
			string registryPath)
		{
			if (
				(registryPath == @"HKEY_CURRENT_USER\Performance") ||
				(registryPath == @"HKEY_CURRENT_USER\System\State\Shell") ||
				(registryPath == @"HKEY_LOCAL_MACHINE\Comm") ||
				(registryPath == @"HKEY_LOCAL_MACHINE\Drivers") ||
				(registryPath == @"HKEY_LOCAL_MACHINE\SECURITY\SIM") ||
				(registryPath == @"HKEY_LOCAL_MACHINE\Software\Microsoft\ActiveSync\Address") ||
				(registryPath == @"HKEY_LOCAL_MACHINE\Software\Microsoft\Shell\Rai") ||
				(registryPath == @"HKEY_LOCAL_MACHINE\System\State\DateTime") ||
				(registryPath == @"HKEY_LOCAL_MACHINE\System\State\Phone") ||
				(registryPath == @"HKEY_LOCAL_MACHINE\System\Uptime"))
			{
				return;
			}

			textWriter.WriteLine();
			textWriter.WriteLine(
				"[{0}]",
				registryPath);

			CrossThreadControlAccessor.SetText(
				this._labelStatus,
				registryPath);

			foreach (KeyValuePair<string, object> value in registryKey.Values)
			{
				if (value.Value == null)
				{
					textWriter.WriteLine(
						"{0} = NULL",
						value.Key);
				}
				else
				{
					Type valueType = value.Value.GetType();

					if (valueType == typeof(int))
					{
						textWriter.WriteLine(
							"{0} = {1}:{2} (0x{2:x8})",
							value.Key,
							valueType.Name,
							value.Value);
					}
					else if (valueType == typeof(long))
					{
						textWriter.WriteLine(
							"{0} = {1}:{2} (0x{2:x16})",
							value.Key,
							valueType.Name,
							value.Value);
					}
					else if (valueType == typeof(byte[]))
					{
						byte[] typeValue = (byte[])value.Value;

						textWriter.Write(
							"{0} = {1}:0x",
							value.Key,
							valueType.Name);

						foreach (byte item in typeValue)
						{
							textWriter.Write("{0:x2}", item);
						}

						textWriter.WriteLine();
					}
					else if (valueType == typeof(string[]))
					{
						string[] typeValue = (string[])value.Value;

						textWriter.Write(
							"{0} = {1}:",
							value.Key,
							valueType.Name);

						if (typeValue.Length > 0)
						{
							textWriter.Write("\"{0}\"", typeValue[0]);

							for (int index = 1; index < typeValue.Length; index++)
							{
								textWriter.Write(", \"{0}\"", typeValue[index]);
							}
						}

						textWriter.WriteLine();
					}
					else
					{
						textWriter.WriteLine(
							"{0} = {1}:\"{2}\"",
							value.Key,
							valueType.Name,
							value.Value);
					}
				}
			}

			foreach (MobileRegistryKey key in registryKey.Keys)
			{
				EnumRegistryKey(
					textWriter,
					key,
					string.Format("{0}\\{1}", registryPath, key.Name));
			}

			registryKey.Dispose();
		}

		#endregion

		#region Mobile Host

		private void MobileHost_DeviceConnected(object sender, EventArgs<MobileDevice> e)
		{
			this._deviceDescriptors.Add(
				new MobileDeviceDescriptor(
					e.Value.DeviceID,
					GetDeviceName(e.Value)));

			base.Invoke(new MethodInvoker(UpdateUI));
		}

		private void MobileHost_DeviceDisconnected(object sender, EventArgs<MobileDevice> e)
		{
			foreach (MobileDeviceDescriptor deviceDescriptor in this._deviceDescriptors)
			{
				if (deviceDescriptor.ID == e.Value.DeviceID)
				{
					this._deviceDescriptors.Remove(deviceDescriptor);
					break;
				}
			}

			base.Invoke(new MethodInvoker(UpdateUI));
		}

		private string GetDeviceName(MobileDevice mobileDevice)
		{
			using (MobileSession session = mobileDevice.CreateSession())
			{
				try
				{
					using (MobileRegistryKey key =
					session.OpenRegistryKey(
						MobileRegistryKeys.HKEY_LOCAL_MACHINE,
						"Software\\Microsoft\\ActiveSync",
						MobileRegistryRights.Read))
					{
						object value = key.Values["DeviceModel"];

						if (value != null)
						{
							return value.ToString().Trim(' ', '\0', '\r', '\n', '\t');
						}
					}
				}
				catch
				{
				}
			}

			return mobileDevice.DeviceName;
		}

		#endregion

		#region Thread

		/// <summary>
		/// We should do all work in separate thread, to keep UI alive.
		/// Cross thread COM calls are impossible due to need of making GUI thread run in STA apartment.
		/// </summary>
		private void ThreadProc()
		{
			using (MobileHost mobileHost = new MobileHost())
			{
				this._deviceDescriptors = new List<MobileDeviceDescriptor>();

				foreach (MobileDevice mobileDevice in mobileHost.Devices)
				{
					this._deviceDescriptors.Add(
						new MobileDeviceDescriptor(
							mobileDevice.DeviceID,
							GetDeviceName(mobileDevice)));
				}

				mobileHost.DeviceConnected += new EventHandler<EventArgs<MobileDevice>>(MobileHost_DeviceConnected);
				mobileHost.DeviceDisconnected += new EventHandler<EventArgs<MobileDevice>>(MobileHost_DeviceDisconnected);

				base.Invoke(new MethodInvoker(UpdateUI));

				bool isQuitPending = false;

				while ((!isQuitPending) && this._workerEvent.WaitOne())
				{
					this._workerEvent.Reset();

					switch (this._workerCommand)
					{
						case WorkerCommand.DumpFileSystem:
							CrossThreadControlAccessor.SetEnabled(this._comboDevice, false);
							CrossThreadControlAccessor.SetEnabled(this._buttonDumpFileSystem, false);
							CrossThreadControlAccessor.SetEnabled(this._buttonDumpRegistry, false);
							CrossThreadControlAccessor.SetEnabled(this._buttonExit, false);

							foreach (MobileDevice mobileDevice in mobileHost.Devices)
							{
								if (mobileDevice.DeviceID == this._workerDeviceID)
								{
									using (MobileSession mobileSession = mobileDevice.CreateSession())
									{
										EnumFileSystem(
											mobileSession,
											"\\",
											this._folderBrowserDialog.SelectedPath + "\\");
									}
								}
							}

							CrossThreadControlAccessor.SetText(this._labelStatus, string.Empty);
							CrossThreadControlAccessor.SetEnabled(this._buttonExit, true);
							CrossThreadControlAccessor.SetEnabled(this._buttonDumpRegistry, true);
							CrossThreadControlAccessor.SetEnabled(this._buttonDumpFileSystem, true);
							CrossThreadControlAccessor.SetEnabled(this._comboDevice, true);
							break;
						case WorkerCommand.DumpRegistry:
							CrossThreadControlAccessor.SetEnabled(this._comboDevice, false);
							CrossThreadControlAccessor.SetEnabled(this._buttonDumpFileSystem, false);
							CrossThreadControlAccessor.SetEnabled(this._buttonDumpRegistry, false);
							CrossThreadControlAccessor.SetEnabled(this._buttonExit, false);

							foreach (MobileDevice mobileDevice in mobileHost.Devices)
							{
								if (mobileDevice.DeviceID == this._workerDeviceID)
								{
									using (TextWriter textWriter = File.CreateText(this._saveFileDialog.FileName))
									{
										using (MobileSession mobileSession = mobileDevice.CreateSession())
										{
											EnumRegistryKey(
												textWriter,
												mobileSession.OpenRegistryKey(
													MobileRegistryKeys.HKEY_CLASSES_ROOT,
													MobileRegistryRights.Read),
													"HKEY_CLASSES_ROOT");
											EnumRegistryKey(
												textWriter,
												mobileSession.OpenRegistryKey(
													MobileRegistryKeys.HKEY_LOCAL_MACHINE,
													MobileRegistryRights.Read),
													"HKEY_LOCAL_MACHINE");
											EnumRegistryKey(
												textWriter,
												mobileSession.OpenRegistryKey(
													MobileRegistryKeys.HKEY_CURRENT_USER,
													MobileRegistryRights.Read),
													"HKEY_CURRENT_USER");
										}
									}
								}
							}

							CrossThreadControlAccessor.SetText(this._labelStatus, string.Empty);
							CrossThreadControlAccessor.SetEnabled(this._buttonExit, true);
							CrossThreadControlAccessor.SetEnabled(this._buttonDumpRegistry, true);
							CrossThreadControlAccessor.SetEnabled(this._buttonDumpFileSystem, true);
							CrossThreadControlAccessor.SetEnabled(this._comboDevice, true);
							break;
						case WorkerCommand.Quit:
							isQuitPending = true;
							break;
					}
				}
			}
		}

		#endregion

		#region Utility

		private void UpdateUI()
		{
			if (!this._isUpdatingUI)
			{
				this._isUpdatingUI = true;

				this._comboDevice.Items.Clear();

				foreach (MobileDeviceDescriptor deviceDescriptor in this._deviceDescriptors)
				{
					this._comboDevice.Items.Add(deviceDescriptor);
				}

				if (this._comboDevice.Items.Count == 0)
				{
					this._comboDevice.Enabled = false;
					this._buttonDumpFileSystem.Enabled = false;
					this._buttonDumpRegistry.Enabled = false;
				}
				else
				{
					this._comboDevice.Enabled = true;

					if (this._comboDevice.Items.Count == 1)
					{
						this._comboDevice.SelectedIndex = 0;
					}

					if (this._comboDevice.SelectedIndex == -1)
					{
						this._buttonDumpFileSystem.Enabled = false;
						this._buttonDumpRegistry.Enabled = false;
					}
					else
					{
						this._buttonDumpFileSystem.Enabled = true;
						this._buttonDumpRegistry.Enabled = true;
					}
				}

				this._isUpdatingUI = false;
			}
		}

		#endregion

		#region Form

		private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			if (this._buttonExit.Enabled)
			{
				this._workerCommand = WorkerCommand.Quit;
				this._workerEvent.Set();
				this._workerThread.Join();
			}
			else
			{
				e.Cancel = true;
			}
		}

		#endregion

		#region Controls

		private void ComboDevice_SelectedIndexChanged(object sender, EventArgs e)
		{
			UpdateUI();
		}

		private void ButtonDumpFileSystem_Click(object sender, EventArgs e)
		{
			if (this._folderBrowserDialog.ShowDialog(this) == DialogResult.OK)
			{
				this._workerCommand = WorkerCommand.DumpFileSystem;
				this._workerDeviceID = ((MobileDeviceDescriptor)this._comboDevice.SelectedItem).ID;
				this._workerEvent.Set();
			}
		}

		private void ButtonDumpRegistry_Click(object sender, EventArgs e)
		{
			if (this._saveFileDialog.ShowDialog(this) == DialogResult.OK)
			{
				this._workerCommand = WorkerCommand.DumpRegistry;
				this._workerDeviceID = ((MobileDeviceDescriptor)this._comboDevice.SelectedItem).ID;
				this._workerEvent.Set();
			}
		}

		private void ButtonExit_Click(object sender, EventArgs e)
		{
			base.Close();
		}

		#endregion

		#endregion
	}
}