﻿using System;
using System.Collections.Generic;
using System.Net.Security;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;

using Nabu.Net;

namespace Pop3ClientExample
{
	internal sealed class Pop3ClientCallback : IPop3ClientCallback<TcpClientSocket>
	{
		#region Pop3ClientCallback

		private string _hostname;
		private Pop3ClientRequestType _request = Pop3ClientRequestType.Unknown;
		private int _messageNumber = 0;
		private int _messageIndex = 0;

		public Pop3ClientCallback(string hostname)
		{
			this._hostname = hostname;
		}

		#endregion

		#region IPop3ClientCallback

		public void OnError(
			Pop3ClientSession<TcpClientSocket> session,
			Pop3ClientState state,
			string text)
		{
			Console.WriteLine("OnError: {0}, '{1}'", state, text);
		}

		public void OnCredentials(
			Pop3ClientSession<TcpClientSocket> session,
			out string hostname,
			out string username,
			out string password)
		{
			Console.WriteLine("OnCredentials");
			hostname = this._hostname;
			Console.Write("\tUsername: ");
			username = Console.ReadLine();
			Console.Write("\tPassword: ");
			password = Console.ReadLine();
		}

		public void OnRequest(
			Pop3ClientSession<TcpClientSocket> session,
			out Pop3ClientRequest request)
		{
			switch (this._request)
			{
				case Pop3ClientRequestType.Unknown:
					this._request = Pop3ClientRequestType.Stat;
					break;
				case Pop3ClientRequestType.Stat:
					this._request = Pop3ClientRequestType.List;
					break;
				case Pop3ClientRequestType.List:
					this._messageIndex++;

					if (this._messageIndex >= this._messageNumber)
					{
						this._messageIndex = 0;
						this._request = Pop3ClientRequestType.Uidl;
					}
					break;
				case Pop3ClientRequestType.Uidl:
					this._messageIndex++;

					if (this._messageIndex >= this._messageNumber)
					{
						this._messageIndex = 0;
						this._request = Pop3ClientRequestType.Retrive;
					}
					break;
				case Pop3ClientRequestType.Retrive:
					this._messageIndex++;

					if (this._messageIndex >= this._messageNumber)
					{
						this._messageIndex = 0;
						this._request = Pop3ClientRequestType.Reset;
					}
					break;
				case Pop3ClientRequestType.Reset:
					if (session.GetOptionSupported(Pop3ClientOptions.CommandTop))
					{
						this._request = Pop3ClientRequestType.Top;
					}
					else
					{
						this._request = Pop3ClientRequestType.Quit;
					}
					break;
				case Pop3ClientRequestType.Top:
					this._messageIndex++;

					if (this._messageIndex >= this._messageNumber)
					{
						this._messageIndex = 0;
						this._request = Pop3ClientRequestType.Quit;
					}
					break;
				case Pop3ClientRequestType.Quit:
					break;
			}

			Console.WriteLine("OnRequest");
			Console.WriteLine("\t{0} {1}", this._request, this._messageIndex);

			request = new Pop3ClientRequest(this._request, this._messageIndex, 0);
		}

		public void OnStat(
			Pop3ClientSession<TcpClientSocket> session,
			int messageTotalCount,
			int messageTotalSize)
		{
			Console.WriteLine("OnStat");
			Console.WriteLine("\t{0} messages, {1} bytes", messageTotalCount, messageTotalSize);

			this._messageNumber = messageTotalCount;
		}

		public void OnList(
			Pop3ClientSession<TcpClientSocket> session,
			IDictionary<int, int> messageSizes)
		{
			Console.WriteLine("OnList");
			Console.WriteLine("\t{0} messages", messageSizes.Count);

			foreach (KeyValuePair<int, int> messageSize in messageSizes)
			{
				Console.WriteLine("\t{0}; {1}",
					messageSize.Key,
					messageSize.Value);
			}
		}

		public void OnUidl(
			Pop3ClientSession<TcpClientSocket> session,
			IDictionary<int, string> messageUIDs)
		{
			Console.WriteLine("OnUidl");
			Console.WriteLine("\t{0} messages", messageUIDs.Count);

			foreach (KeyValuePair<int, string> messageSize in messageUIDs)
			{
				Console.WriteLine("\t{0}; {1}",
					messageSize.Key,
					messageSize.Value);
			}
		}

		public void OnTop(
			Pop3ClientSession<TcpClientSocket> session,
			ArraySegment<byte> segment)
		{
			Console.WriteLine("OnTop: {0} bytes", segment.Count);
		}

		public void OnRetrive(
			Pop3ClientSession<TcpClientSocket> session,
			ArraySegment<byte> segment)
		{
			Console.WriteLine("OnRetrive: {0} bytes", segment.Count);
		}

		#endregion

		#region ITcpNodeCallback

		public void OnOpenSession(Pop3ClientSession<TcpClientSocket> session)
		{
			Console.WriteLine("CONNECTED.");
		}

		public void OnCloseSession(Pop3ClientSession<TcpClientSocket> session)
		{
			Console.WriteLine("DISCONNECTED.");
		}

		public void OnSslAuthenticate(Pop3ClientSession<TcpClientSocket> session, out X509CertificateCollection certificates, out SslProtocols enabledSslProtocols, out bool mutualAuthenticationRequired, out bool checkCertificateRevocation)
		{
			certificates = null;
			mutualAuthenticationRequired = false;
			enabledSslProtocols = SslProtocols.Default;
			checkCertificateRevocation = false;
		}

		public bool OnSslValidateCertificate(
			Pop3ClientSession<TcpClientSocket> session,
			X509Certificate certificate,
			X509Chain chain,
			SslPolicyErrors sslPolicyErrors)
		{
			return true;
		}

		public X509Certificate OnSslSelectCertificate(
			Pop3ClientSession<TcpClientSocket> session,
			string targetHost,
			X509CertificateCollection localCertificates,
			X509Certificate remoteCertificate,
			string[] acceptableIssuers)
		{
			return null;
		}

		#endregion
	}

	public sealed class Program
	{
		public static void Main(string[] args)
		{
			Console.WriteLine("POP3 Client Example");
			Console.WriteLine();
			Console.WriteLine("Connection Parameters:");
			Console.Write("\tHost: ");

			string host = Console.ReadLine();

			Console.Write("\tPort: ");

			string port = Console.ReadLine();

			using (Pop3Client pop3Client =
				new Pop3Client(
					new Pop3ClientCallback(
						host)))
			{
				pop3Client.OpenSession(
					host,
					ushort.Parse(port),
					TcpConnectionWrapper.None,
					null);
				pop3Client.WaitForSessions();
			}
		}
	}
}
