﻿using System;
using System.Diagnostics;
using System.IO;
using System.Windows.Forms;

using Nabu.Reflection;
using Nabu.Text;

namespace ScriptEngineExample.FF
{
	public partial class MainForm : Form
	{
		#region Fields

		private IScriptExecutionHost _executionHost = new ManagedScriptUnsecureExecutionHost();
		private IScriptExecutionEngine _executionEngine;
		private IGeneratedSourceCodeUnit _sourceCode;
		private string _workerScopeName = string.Format("NS_{0:N}", Guid.NewGuid());
		private string _workerTypeName = "CustomWorker";

		#endregion

		#region Methods

		private void Execute()
		{
			using (StringWriter output = new StringWriter())
			{
				Stopwatch watch = new Stopwatch();

				watch.Start();

				IntPtr cookie = IntPtr.Zero;

				try
				{
					cookie = this._executionEngine.CreateInstance(this._workerScopeName, this._workerTypeName);

					this._executionEngine.SetProperty(cookie, "Output", output);

					bool isAllowed;

					try
					{
						isAllowed = (bool)this._executionEngine.InvokeMethod(cookie, "PreProcess");
					}
					catch (MissingMethodException)
					{
						isAllowed = true;
					}

					if (isAllowed)
					{
						this._executionEngine.InvokeMethod(cookie, "InProcess");

						try
						{
							this._executionEngine.InvokeMethod(cookie, "PostProcess");
						}
						catch (MissingMethodException)
						{
						}
					}
				}
				finally
				{
					this._executionEngine.DestroyInstance(cookie);
				}

				watch.Stop();

				output.WriteLine();
				output.WriteLine("{0}; Processed in {1}.", DateTime.Now, watch.Elapsed);

				this._textOutput.Text = output.ToString();
			}
		}

		public MainForm()
		{
			InitializeComponent();
		}

		private void ButtonProcess_Click(object sender, EventArgs e)
		{
			try
			{
				TemplateProcessor templateProcessor =
					new TemplateProcessor(
						new CustomTemplateAccessor(this._textSource.Lines));

				templateProcessor.Parse("default.template");
				templateProcessor.Generate(this._workerScopeName, this._workerTypeName);

				this._sourceCode = templateProcessor.SourceCode;

				this._textCode.Text = this._sourceCode.Value;

				this._executionEngine =
					this._executionHost.CreateEngine(
						new ISourceCodeUnit[]
						{
							this._sourceCode
						});

				Execute();
			}
			catch (TemplateParserException exception)
			{
				this._textOutput.Clear();

				this._textOutput.AppendText(
						string.Format(
							"{0} {1}\r\n",
							exception.ErrorType,
							exception.Position));
			}
			catch (ScriptEngineCompilerException exception)
			{
				this._textOutput.Clear();

				foreach (ScriptEngineCompilerMessage message in exception.Messages)
				{
					this._textOutput.AppendText(
						string.Format(
							"{0} @ {1}/{2}; {3} {4}\r\n",
							message.IsError ? "ERROR" : "WARNING",
							message.CodePosition,
							message.TextPosition,
							message.MessageID,
							message.MessageText));
				}
			}
			catch (Exception exception)
			{
				this._textOutput.Text =
					string.Format(
						"ERROR:\r\n{0}\r\n{1}",
						exception.Message,
						exception.StackTrace);
			}
		}

		private void ButtonExecute_Click(object sender, EventArgs e)
		{
			try
			{
				if (this._executionEngine != null)
				{
					Execute();
				}
			}
			catch (ScriptEngineCompilerException exception)
			{
				this._textOutput.Clear();

				foreach (ScriptEngineCompilerMessage message in exception.Messages)
				{
					this._textOutput.AppendText(
						string.Format(
							"{0} @ {1}/{2}; {3} {4}\r\n",
							message.IsError ? "ERROR" : "WARNING",
							message.CodePosition,
							message.TextPosition,
							message.MessageID,
							message.MessageText));
				}
			}
			catch (Exception exception)
			{
				this._textOutput.Text =
					string.Format(
						"ERROR:\r\n{0}\r\n{1}",
						exception.Message,
						exception.StackTrace);
			}
		}

		private void TextCode_MouseDoubleClick(object sender, MouseEventArgs e)
		{
			if (this._sourceCode != null)
			{
				int index = this._textCode.GetCharIndexFromPosition(e.Location);
				int line = this._textCode.GetLineFromCharIndex(index);
				int column = index - this._textCode.GetFirstCharIndexFromLine(line);
				long position = ((long)line << 32) + column;

				foreach (TextPositionMap map in this._sourceCode.PositionMap)
				{
					long start = ((long)map.DestinationStart.Line << 32) + map.DestinationStart.Column;
					long end = ((long)map.DestinationEnd.Line << 32) + map.DestinationEnd.Column;

					if ((position >= start) && (position < end))
					{
						int sourceStart =
							this._textSource.GetFirstCharIndexFromLine(map.SourceStart.Line) +
							map.SourceStart.Column;
						int sourceEnd =
							this._textSource.GetFirstCharIndexFromLine(map.SourceEnd.Line) +
							map.SourceEnd.Column;
						int destinationStart =
							this._textCode.GetFirstCharIndexFromLine(map.DestinationStart.Line) +
							map.DestinationStart.Column;
						int destinationEnd =
							this._textCode.GetFirstCharIndexFromLine(map.DestinationEnd.Line) +
							map.DestinationEnd.Column;

						this._textSource.Select(sourceStart, sourceEnd - sourceStart);
						this._textCode.Select(destinationStart, destinationEnd - destinationStart);
					}
				}
			}
		}

		#endregion
	}
}
