﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Collections Library                      */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;

namespace Nabu.Collections
{
	[Serializable]
	public abstract class BindingListExBase<T> :
		IBindingListEx<T>,
		IBindingList,
		IObjectObserver,
		ICancelAddNew,
		IRaiseItemChangedEvents
	{
		#region BindingListExBase

		#region Constants

		private const string IndexingExceptionMessage = "Indexing is not supported.";
		private const string SortingExceptionMessage = "Sorting is not supported.";
		private const string SearchExceptionMessage = "Search is not supported.";

		#endregion

		#region Types

		private sealed class ObjectObserverContext
		{
			#region ObjectObserverContext

			#region Fields

			private BindingListExBase<T> _list;
			private int _index;

			#endregion

			#region Methods

			[DebuggerStepThrough]
			public ObjectObserverContext(BindingListExBase<T> list, int index)
			{
				this._list = list;
				this._index = index;
			}

			#endregion

			#region Properties

			public int Index
			{
				[DebuggerStepThrough]
				get { return this._index; }
			}

			#endregion

			#endregion

			#region Object

			[DebuggerStepThrough]
			public override bool Equals(object obj)
			{
				ObjectObserverContext context = obj as ObjectObserverContext;

				if (obj == null)
				{
					return false;
				}
				else
				{
					return
						(this._list == context._list) &&
						(this._index == context._index);
				}
			}

			[DebuggerStepThrough]
			public override int GetHashCode()
			{
				return this._list.GetHashCode() ^ this._index;
			}

			#endregion
		}

		#endregion

		#region Fields

		private IListEx<T> _list;
		private int _newItemIndex = -1;

		#endregion

		#region Methods

		#region Utility

		[DebuggerStepThrough]
		private bool RegisterObserver(T item, int index)
		{
			IObservableObject observableObject = item as IObservableObject;

			if (observableObject == null)
			{
				return false;
			}
			else
			{
				observableObject.RegisterObserver(
					new ObjectObserverContext(
						this,
						index),
					this);

				return true;
			}
		}

		[DebuggerStepThrough]
		private bool UnregisterObserver(T item, int index)
		{
			IObservableObject observableObject = item as IObservableObject;

			if (observableObject == null)
			{
				return false;
			}
			else
			{
				observableObject.UnregisterObserver(
					new ObjectObserverContext(
						this,
						index));

				return true;
			}
		}

		[DebuggerStepThrough]
		private void SubscribeToNotifications(T item, int index)
		{
			if (!RegisterObserver(item, index))
			{
				INotifyPropertyChanging notifyChanging = item as INotifyPropertyChanging;

				if (notifyChanging != null)
				{
					notifyChanging.PropertyChanging += new PropertyChangingEventHandler(Item_PropertyChanging);
				}

				INotifyPropertyChanged notifyChanged = item as INotifyPropertyChanged;

				if (notifyChanged != null)
				{
					notifyChanged.PropertyChanged += new PropertyChangedEventHandler(Item_PropertyChanged);
				}
			}
		}

		[DebuggerStepThrough]
		private void UnsubscribeFromNotifications(T item, int index)
		{
			if (!UnregisterObserver(item, index))
			{
				INotifyPropertyChanging notifyChanging = item as INotifyPropertyChanging;

				if (notifyChanging != null)
				{
					notifyChanging.PropertyChanging -= new PropertyChangingEventHandler(Item_PropertyChanging);
				}

				INotifyPropertyChanged notifyChanged = item as INotifyPropertyChanged;

				if (notifyChanged != null)
				{
					notifyChanged.PropertyChanged -= new PropertyChangedEventHandler(Item_PropertyChanged);
				}
			}
		}

		#endregion

		#region Event Handlers

		private void Item_PropertyChanging(object sender, PropertyChangingEventArgs e)
		{
			T target = (T)sender;
			int index = this._list.IndexOf(target);

			if (!RaiseChanging(index, target, target))
			{
				throw new InvalidOperationException(e.PropertyName);
			}
		}

		private void Item_PropertyChanged(object sender, PropertyChangedEventArgs e)
		{
			T target = (T)sender;
			int index = this._list.IndexOf(target);

			RaiseListChange(index);
			RaiseChanged(index, target, target);
		}

		#endregion

		#region Event Raisers

		#region IBindingList Events

		[DebuggerStepThrough]
		private void RaiseListChanged(ListChangedType type, int indexNew, int indexOld)
		{
			ListChangedEventArgs eventArgs = new ListChangedEventArgs(type, indexNew, indexOld);

			if (this._listChangedEventHandler != null)
			{
				this._listChangedEventHandler(this, eventArgs);
			}
		}

		[DebuggerStepThrough]
		private void RaiseListReset()
		{
			RaiseListChanged(ListChangedType.Reset, 0, 0);
		}

		[DebuggerStepThrough]
		private void RaiseListAdd(int index)
		{
			RaiseListChanged(ListChangedType.ItemAdded, index, 0);
		}

		[DebuggerStepThrough]
		private void RaiseListDelete(int index)
		{
			RaiseListChanged(ListChangedType.ItemDeleted, index, 0);
		}

		[DebuggerStepThrough]
		private void RaiseListChange(int index)
		{
			RaiseListChanged(ListChangedType.ItemChanged, index, 0);
		}

		[DebuggerStepThrough]
		private void RaiseListMove(int indexNew, int indexOld)
		{
			RaiseListChanged(ListChangedType.ItemMoved, indexNew, indexOld);
		}

		#endregion

		#region IBindingListEx<T> Events

		[DebuggerStepThrough]
		private bool RaiseReseting()
		{
			bool result;

			if (this._resetingHandler != null)
			{
				CancelEventArgs eventArgs = new CancelEventArgs();

				this._resetingHandler(this, eventArgs);

				result = !eventArgs.Cancel;
			}
			else
			{
				result = true;
			}

			if (result)
			{
				for (int i = 0; i < this._list.Count; i++)
				{
					UnsubscribeFromNotifications(this._list[i], i);
				}
			}

			return result;
		}

		[DebuggerStepThrough]
		private void RaiseReseted()
		{
			for (int i = 0; i < this._list.Count; i++)
			{
				SubscribeToNotifications(this._list[i], i);
			}

			if (this._resetedHandler != null)
			{
				this._resetedHandler(this, EventArgs.Empty);
			}

			EndNew(this._newItemIndex);
		}

		[DebuggerStepThrough]
		private bool RaiseAdding(int index, T value)
		{
			bool result;

			if (this._addingHandler != null)
			{
				BindingListExAddingRemovingEventArgs<T> eventArgs =
					new BindingListExAddingRemovingEventArgs<T>(
						index,
						value);

				this._addingHandler(this, eventArgs);

				result = !eventArgs.Cancel;
			}
			else
			{
				result = true;
			}

			if (result)
			{
				for (int i = index; i < this._list.Count; i++)
				{
					UnregisterObserver(this._list[i], i);
				}
			}

			return result;
		}

		[DebuggerStepThrough]
		private void RaiseAdded(int index, T value)
		{
			SubscribeToNotifications(value, index);

			for (int i = index + 1; i < this._list.Count; i++)
			{
				RegisterObserver(this._list[i], i);
			}

			if (this._addedHandler != null)
			{
				this._addedHandler(
					this,
					new BindingListExAddedRemovedEventArgs<T>(
						index,
						value));
			}

			this._newItemIndex = index;
		}

		[DebuggerStepThrough]
		private bool RaiseRemoving(int index, T value)
		{
			bool result;

			if (this._removingHandler != null)
			{
				BindingListExAddingRemovingEventArgs<T> eventArgs =
					new BindingListExAddingRemovingEventArgs<T>(
						index,
						value);

				this._removingHandler(this, eventArgs);

				result = !eventArgs.Cancel;
			}
			else
			{
				result = true;
			}

			if (result)
			{
				for (int i = index + 1; i < this._list.Count; i++)
				{
					UnregisterObserver(this._list[i], i);
				}
			}

			return result;
		}

		[DebuggerStepThrough]
		private void RaiseRemoved(int index, T value)
		{
			UnsubscribeFromNotifications(value, index);

			for (int i = index; i < this._list.Count; i++)
			{
				RegisterObserver(this._list[i], i);
			}

			if (this._removedHandler != null)
			{
				this._removedHandler(
					this,
					new BindingListExAddedRemovedEventArgs<T>(
						index,
						value));
			}

			EndNew(this._newItemIndex);
		}

		[DebuggerStepThrough]
		private bool RaiseChanging(int index, T oldValue, T newValue)
		{
			bool result;

			if (this._changingHandler != null)
			{
				BindingListExChangingEventArgs<T> eventArgs =
					new BindingListExChangingEventArgs<T>(
						index,
						oldValue,
						newValue);

				this._changingHandler(this, eventArgs);

				result = !eventArgs.Cancel;
			}
			else
			{
				result = true;
			}

			return result;
		}

		[DebuggerStepThrough]
		private void RaiseChanged(int index, T oldValue, T newValue)
		{
			if (this._changedHandler != null)
			{
				this._changedHandler(
					this,
					new BindingListExChangedEventArgs<T>(
						index,
						oldValue,
						newValue));
			}

			EndNew(this._newItemIndex);
		}

		[DebuggerStepThrough]
		private bool RaiseMoving(int index1, int index2)
		{
			bool result;

			if (this._movingHandler != null)
			{
				BindingListExMovingEventArgs<T> eventArgs =
					new BindingListExMovingEventArgs<T>(
						index1,
						index2);

				this._movingHandler(this, eventArgs);

				result = !eventArgs.Cancel;
			}
			else
			{
				result = true;
			}

			if (result)
			{
				UnregisterObserver(this._list[index1], index1);
				UnregisterObserver(this._list[index2], index2);
			}

			return result;
		}

		[DebuggerStepThrough]
		private void RaiseMoved(int index1, int index2)
		{
			RegisterObserver(this._list[index1], index1);
			RegisterObserver(this._list[index2], index2);

			if (this._movedHandler != null)
			{
				this._movedHandler(
					this,
					new BindingListExMovedEventArgs<T>(
						index1,
						index2));
			}

			EndNew(this._newItemIndex);
		}

		#endregion

		#endregion

		#region Contructors

		[DebuggerStepThrough]
		protected BindingListExBase()
		{
			this._list = new ListEx<T>();
		}

		[DebuggerStepThrough]
		protected BindingListExBase(int capacity)
		{
			this._list = new ListEx<T>(capacity);
		}

		[DebuggerStepThrough]
		protected BindingListExBase(IEnumerable<T> collection)
		{
			this._list = new ListEx<T>(collection);
		}

		[DebuggerStepThrough]
		protected BindingListExBase(IEnumerable<T> collection, int capacity)
		{
			this._list = new ListEx<T>(collection, capacity);
		}

		[DebuggerStepThrough]
		protected BindingListExBase(T[] array)
		{
			this._list = new ListEx<T>(array);
		}

		[DebuggerStepThrough]
		protected BindingListExBase(T[] array, bool attach)
		{
			this._list = new ListEx<T>(array, attach);
		}

		#endregion

		#region Abstract

		protected abstract T CreateNew();

		#endregion

		#endregion

		#endregion

		#region IBindingListEx<T>

		#region Methods

		[DebuggerStepThrough]
		public void Swap(int index1, int index2)
		{
			if (RaiseMoving(index1, index2))
			{
				T item = this._list[index1];

				this._list[index1] = this._list[index2];
				this._list[index2] = item;

				RaiseMoved(index1, index2);
				RaiseListMove(index2, index1);
			}
		}

		#endregion

		#region Propeties

		public IListEx<T> InnerList
		{
			[DebuggerStepThrough]
			get { return this._list; }
		}

		#endregion

		#region Events

		#region Reseting

		[NonSerialized]
		private CancelEventHandler _resetingHandler;

		public event CancelEventHandler Reseting
		{
			add { this._resetingHandler += value; }
			remove { this._resetingHandler -= value; }
		}

		#endregion

		#region Reseted

		[NonSerialized]
		private EventHandler _resetedHandler;

		public event EventHandler Reseted
		{
			add { this._resetedHandler += value; }
			remove { this._resetedHandler -= value; }
		}

		#endregion

		#region Adding

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExAddingRemovingEventArgs<T>> _addingHandler;

		public event EventHandler<BindingListExAddingRemovingEventArgs<T>> Adding
		{
			add { this._addingHandler += value; }
			remove { this._addingHandler -= value; }
		}

		#endregion

		#region Added

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExAddedRemovedEventArgs<T>> _addedHandler;

		public event EventHandler<BindingListExAddedRemovedEventArgs<T>> Added
		{
			add { this._addedHandler += value; }
			remove { this._addedHandler -= value; }
		}

		#endregion

		#region Removing

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExAddingRemovingEventArgs<T>> _removingHandler;

		public event EventHandler<BindingListExAddingRemovingEventArgs<T>> Removing
		{
			add { this._removingHandler += value; }
			remove { this._removingHandler -= value; }
		}

		#endregion

		#region Removed

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExAddedRemovedEventArgs<T>> _removedHandler;

		public event EventHandler<BindingListExAddedRemovedEventArgs<T>> Removed
		{
			add { this._removedHandler += value; }
			remove { this._removedHandler -= value; }
		}

		#endregion

		#region Changing

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExChangingEventArgs<T>> _changingHandler;

		public event EventHandler<BindingListExChangingEventArgs<T>> Changing
		{
			add { this._changingHandler += value; }
			remove { this._changingHandler -= value; }
		}

		#endregion

		#region Changed

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExChangedEventArgs<T>> _changedHandler;

		public event EventHandler<BindingListExChangedEventArgs<T>> Changed
		{
			add { this._changedHandler += value; }
			remove { this._changedHandler -= value; }
		}

		#endregion

		#region Moving

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExMovingEventArgs<T>> _movingHandler;

		public event EventHandler<BindingListExMovingEventArgs<T>> Moving
		{
			add { this._movingHandler += value; }
			remove { this._movingHandler -= value; }
		}

		#endregion

		#region Moved

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExMovedEventArgs<T>> _movedHandler;

		public event EventHandler<BindingListExMovedEventArgs<T>> Moved
		{
			add { this._movedHandler += value; }
			remove { this._movedHandler -= value; }
		}

		#endregion

		#endregion

		#endregion

		#region IListEx<T>

		#region Methods

		#region Capacity

		[DebuggerStepThrough]
		public void EnsureCapacity(int capacity)
		{
			this._list.EnsureCapacity(capacity);
		}

		[DebuggerStepThrough]
		public void TrimExcess(float ratio)
		{
			this._list.TrimExcess(ratio);
		}

		[DebuggerStepThrough]
		public void TrimExcess()
		{
			this._list.TrimExcess();
		}

		#endregion

		#region Range

		[DebuggerStepThrough]
		public IListEx<T> GetRange(int startIndex, int count)
		{
			return this._list.GetRange(startIndex, count);
		}

		[DebuggerStepThrough]
		public void AddRange(IEnumerable<T> collection)
		{
			if (RaiseReseting())
			{
				this._list.AddRange(collection);
				RaiseReseted();
				RaiseListReset();
				EndNew(this._newItemIndex);
			}
		}

		[DebuggerStepThrough]
		public void AddRange(IListEx<T> list, int offset, int count)
		{
			if (RaiseReseting())
			{
				this._list.AddRange(list, offset, count);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void AddRange(IListEx<T> list)
		{
			if (RaiseReseting())
			{
				this._list.AddRange(list);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void AddRange(T[] array, int offset, int count)
		{
			if (RaiseReseting())
			{
				this._list.AddRange(array, offset, count);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void AddRange(T[] array)
		{
			if (RaiseReseting())
			{
				this._list.AddRange(array);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void InsertRange(int startIndex, IEnumerable<T> collection)
		{
			if (RaiseReseting())
			{
				this._list.InsertRange(startIndex, collection);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void InsertRange(int startIndex, IListEx<T> list, int offset, int count)
		{
			if (RaiseReseting())
			{
				this._list.InsertRange(startIndex, list, offset, count);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void InsertRange(int startIndex, IListEx<T> list)
		{
			if (RaiseReseting())
			{
				this._list.InsertRange(startIndex, list);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void InsertRange(int startIndex, T[] array, int offset, int count)
		{
			if (RaiseReseting())
			{
				this._list.InsertRange(startIndex, array, offset, count);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void InsertRange(int startIndex, T[] array)
		{
			if (RaiseReseting())
			{
				this._list.InsertRange(startIndex, array);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void RemoveRange(int startIndex, int count)
		{
			if (RaiseReseting())
			{
				this._list.RemoveRange(startIndex, count);
				RaiseReseted();
				RaiseListReset();
			}
		}

		#endregion

		#region Copy and Convert

		public void FromArray(T[] array, bool attach)
		{
			if (RaiseReseting())
			{
				this._list.FromArray(array, attach);
				RaiseReseted();
				RaiseListReset();
			}
		}

		public void FromArray(T[] array)
		{
			if (RaiseReseting())
			{
				this._list.FromArray(array);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public T[] ToArray()
		{
			return this._list.ToArray();
		}

		[DebuggerStepThrough]
		public T[] DangerousGetArray()
		{
			return this._list.DangerousGetArray();
		}

		[DebuggerStepThrough]
		public ArraySegment<T> ToArraySegment()
		{
			return this._list.ToArraySegment();
		}

		[DebuggerStepThrough]
		public void CopyTo(T[] array)
		{
			this._list.CopyTo(array);
		}

		[DebuggerStepThrough]
		public void CopyTo(int index, T[] array, int arrayIndex, int count)
		{
			this._list.CopyTo(index, array, arrayIndex, count);
		}

		[DebuggerStepThrough]
		public IListEx<TOutput> ConvertAll<TOutput>(Converter<T, TOutput> converter)
		{
			return this._list.ConvertAll<TOutput>(converter);
		}

		#endregion

		#region Reverse

		[DebuggerStepThrough]
		public void Reverse(int index, int count)
		{
			if (RaiseReseting())
			{
				this._list.Reverse(index, count);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void Reverse()
		{
			if (RaiseReseting())
			{
				this._list.Reverse();
				RaiseReseted();
				RaiseListReset();
			}
		}

		#endregion

		#region Sort

		[DebuggerStepThrough]
		public void Sort(int startIndex, int count)
		{
			if (RaiseReseting())
			{
				this._list.Sort(startIndex, count);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void Sort()
		{
			if (RaiseReseting())
			{
				this._list.Sort();
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void Sort(IComparer<T> comparer, int startIndex, int count)
		{
			if (RaiseReseting())
			{
				this._list.Sort(comparer, startIndex, count);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void Sort(IComparer<T> comparer)
		{
			if (RaiseReseting())
			{
				this._list.Sort(comparer);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void Sort(Comparison<T> comparer, int startIndex, int count)
		{
			if (RaiseReseting())
			{
				this._list.Sort(comparer, startIndex, count);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void Sort(Comparison<T> comparer)
		{
			if (RaiseReseting())
			{
				this._list.Sort(comparer);
				RaiseReseted();
				RaiseListReset();
			}
		}

		#endregion

		#region Search

		[DebuggerStepThrough]
		public int IndexOf(T item, int startIndex, int count)
		{
			return this._list.IndexOf(item, startIndex, count);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, int startIndex)
		{
			return this._list.IndexOf(item, startIndex);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, IComparer<T> comparer, int startIndex, int count)
		{
			return this._list.IndexOf(item, comparer, startIndex, count);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, IComparer<T> comparer, int startIndex)
		{
			return this._list.IndexOf(item, comparer, startIndex);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, IComparer<T> comparer)
		{
			return this._list.IndexOf(item, comparer);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, Comparison<T> comparer, int startIndex, int count)
		{
			return this._list.IndexOf(item, comparer, startIndex, count);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, Comparison<T> comparer, int startIndex)
		{
			return this._list.IndexOf(item, comparer, startIndex);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, Comparison<T> comparer)
		{
			return this._list.IndexOf(item, comparer);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, int startIndex, int count)
		{
			return this._list.LastIndexOf(item, startIndex, count);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, int startIndex)
		{
			return this._list.LastIndexOf(item, startIndex);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item)
		{
			return this._list.LastIndexOf(item);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, IComparer<T> comparer, int startIndex, int count)
		{
			return this._list.LastIndexOf(item, comparer, startIndex, count);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, IComparer<T> comparer, int startIndex)
		{
			return this._list.LastIndexOf(item, comparer, startIndex);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, IComparer<T> comparer)
		{
			return this._list.LastIndexOf(item, comparer);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, Comparison<T> comparer, int startIndex, int count)
		{
			return this._list.LastIndexOf(item, comparer, startIndex, count);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, Comparison<T> comparer, int startIndex)
		{
			return this._list.LastIndexOf(item, comparer, startIndex);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, Comparison<T> comparer)
		{
			return this._list.LastIndexOf(item, comparer);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, int startIndex, int count)
		{
			return this._list.BinarySearch(item, startIndex, count);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, int startIndex)
		{
			return this._list.BinarySearch(item, startIndex);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item)
		{
			return this._list.BinarySearch(item);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, IComparer<T> comparer, int startIndex, int count)
		{
			return this._list.BinarySearch(item, comparer, startIndex, count);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, IComparer<T> comparer, int startIndex)
		{
			return this._list.BinarySearch(item, comparer, startIndex);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, IComparer<T> comparer)
		{
			return this._list.BinarySearch(item, comparer);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, Comparison<T> comparer, int startIndex, int count)
		{
			return this._list.BinarySearch(item, comparer, startIndex, count);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, Comparison<T> comparer, int startIndex)
		{
			return this._list.BinarySearch(item, comparer, startIndex);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, Comparison<T> comparer)
		{
			return this._list.BinarySearch(item, comparer);
		}

		#endregion

		#region Predicates

		#region Exists

		[DebuggerStepThrough]
		public bool Exists(Predicate<T> match)
		{
			return this._list.Exists(match);
		}

		#endregion

		#region FindAll

		[DebuggerStepThrough]
		public IListEx<T> FindAll(Predicate<T> match, int startIndex, int count)
		{
			return this._list.FindAll(match, startIndex, count);
		}

		[DebuggerStepThrough]
		public IListEx<T> FindAll(Predicate<T> match, int startIndex)
		{
			return this._list.FindAll(match, startIndex);
		}

		[DebuggerStepThrough]
		public IListEx<T> FindAll(Predicate<T> match)
		{
			return this._list.FindAll(match);
		}

		[DebuggerStepThrough]
		public IListEx<int> FindAllIndex(Predicate<T> match, int startIndex, int count)
		{
			return this._list.FindAllIndex(match, startIndex, count);
		}

		[DebuggerStepThrough]
		public IListEx<int> FindAllIndex(Predicate<T> match, int startIndex)
		{
			return this._list.FindAllIndex(match, startIndex);
		}

		[DebuggerStepThrough]
		public IListEx<int> FindAllIndex(Predicate<T> match)
		{
			return this._list.FindAllIndex(match);
		}

		#endregion

		#region FindFirst

		[DebuggerStepThrough]
		public int FindFirstIndex(Predicate<T> match, int startIndex, int count)
		{
			return this._list.FindFirstIndex(match, startIndex, count);
		}

		[DebuggerStepThrough]
		public int FindFirstIndex(Predicate<T> match, int startIndex)
		{
			return this._list.FindFirstIndex(match, startIndex);
		}

		[DebuggerStepThrough]
		public int FindFirstIndex(Predicate<T> match)
		{
			return this._list.FindFirstIndex(match);
		}

		#endregion

		#region FindLast

		[DebuggerStepThrough]
		public int FindLastIndex(Predicate<T> match, int startIndex, int count)
		{
			return this._list.FindLastIndex(match, startIndex, count);
		}

		[DebuggerStepThrough]
		public int FindLastIndex(Predicate<T> match, int startIndex)
		{
			return this._list.FindLastIndex(match, startIndex);
		}

		[DebuggerStepThrough]
		public int FindLastIndex(Predicate<T> match)
		{
			return this._list.FindLastIndex(match);
		}

		#endregion

		#region RemoveAll

		[DebuggerStepThrough]
		public int RemoveAll(Predicate<T> match, int startIndex, int count)
		{
			return this._list.RemoveAll(match, startIndex, count);
		}

		[DebuggerStepThrough]
		public int RemoveAll(Predicate<T> match, int startIndex)
		{
			return this._list.RemoveAll(match, startIndex);
		}

		[DebuggerStepThrough]
		public int RemoveAll(Predicate<T> match)
		{
			return this._list.RemoveAll(match);
		}

		#endregion

		#region TrueForAll

		[DebuggerStepThrough]
		public bool TrueForAll(Predicate<T> match, int startIndex, int count)
		{
			return this._list.TrueForAll(match, startIndex, count);
		}

		[DebuggerStepThrough]
		public bool TrueForAll(Predicate<T> match, int startIndex)
		{
			return this._list.TrueForAll(match, startIndex);
		}

		[DebuggerStepThrough]
		public bool TrueForAll(Predicate<T> match)
		{
			return this._list.TrueForAll(match);
		}

		#endregion

		#region FalseForAll

		[DebuggerStepThrough]
		public bool FalseForAll(Predicate<T> match, int startIndex, int count)
		{
			return this._list.FalseForAll(match, startIndex, count);
		}

		[DebuggerStepThrough]
		public bool FalseForAll(Predicate<T> match, int startIndex)
		{
			return this._list.FalseForAll(match, startIndex);
		}

		[DebuggerStepThrough]
		public bool FalseForAll(Predicate<T> match)
		{
			return this._list.FalseForAll(match);
		}

		#endregion

		#endregion

		#region Actions

		[DebuggerStepThrough]
		public void ForEach(Action<T> action, int startIndex, int count)
		{
			this._list.ForEach(action, startIndex, count);
		}

		[DebuggerStepThrough]
		public void ForEach(Action<T> action, int startIndex)
		{
			this._list.ForEach(action, startIndex);
		}

		[DebuggerStepThrough]
		public void ForEach(Action<T> action)
		{
			this._list.ForEach(action);
		}

		[DebuggerStepThrough]
		public void ForEach(Action<T> action, Predicate<T> match, int startIndex, int count)
		{
			this._list.ForEach(action, match, startIndex, count);
		}

		[DebuggerStepThrough]
		public void ForEach(Action<T> action, Predicate<T> match, int startIndex)
		{
			this._list.ForEach(action, match, startIndex);
		}

		[DebuggerStepThrough]
		public void ForEach(Action<T> action, Predicate<T> match)
		{
			this._list.ForEach(action, match);
		}

		#endregion

		#endregion

		#region Properties

		public int Capacity
		{
			[DebuggerStepThrough]
			get { return this._list.Capacity; }
			[DebuggerStepThrough]
			set { this._list.Capacity = value; }
		}

		public T First
		{
			[DebuggerStepThrough]
			get { return this._list[0]; }
		}

		public T Last
		{
			[DebuggerStepThrough]
			get { return this._list[this._list.Count - 1]; }
		}

		public bool IsAddAllowed
		{
			[DebuggerStepThrough]
			get { return this._list.IsAddAllowed; }
			[DebuggerStepThrough]
			set { this._list.IsAddAllowed = value; }
		}

		public bool IsRemoveAllowed
		{
			[DebuggerStepThrough]
			get { return this._list.IsRemoveAllowed; }
			[DebuggerStepThrough]
			set { this._list.IsRemoveAllowed = value; }
		}

		public bool IsChangeAllowed
		{
			[DebuggerStepThrough]
			get { return this._list.IsChangeAllowed; }
			[DebuggerStepThrough]
			set { this._list.IsChangeAllowed = value; }
		}

		#endregion

		#endregion

		#region IList<T>

		#region Methods

		[DebuggerStepThrough]
		public int IndexOf(T item)
		{
			return this._list.IndexOf(item);
		}

		[DebuggerStepThrough]
		public void Insert(int index, T item)
		{
			if (RaiseAdding(index, item))
			{
				this._list.Insert(index, item);
				RaiseAdded(index, item);
				RaiseListAdd(index);
			}
		}

		#endregion

		#region Properties

		public T this[int index]
		{
			[DebuggerStepThrough]
			get
			{
				return this._list[index];
			}

			[DebuggerStepThrough]
			set
			{
				T item = this._list[index];

				if (RaiseChanging(index, item, value))
				{
					UnsubscribeFromNotifications(item, index);
					this._list[index] = value;
					SubscribeToNotifications(value, index);
					RaiseChanged(index, item, value);
					RaiseListChange(index);
				}
			}
		}

		#endregion

		#endregion

		#region ICollection<T>

		[DebuggerStepThrough]
		public void Add(T item)
		{
			if (RaiseAdding(this._list.Count, item))
			{
				this._list.Add(item);
				RaiseAdded(this._list.Count - 1, item);
				RaiseListAdd(this._list.Count - 1);
			}
		}

		[DebuggerStepThrough]
		public bool Contains(T item)
		{
			return this._list.Contains(item);
		}

		[DebuggerStepThrough]
		public void CopyTo(T[] array, int arrayIndex)
		{
			this._list.CopyTo(array, arrayIndex);
		}

		[DebuggerStepThrough]
		public bool Remove(T item)
		{
			int index = this._list.IndexOf(item);

			if (index == -1)
			{
				return false;
			}

			if (RaiseRemoving(index, item))
			{
				this._list.RemoveAt(index);
				RaiseRemoved(index, item);
				RaiseListDelete(index);

				return true;
			}
			else
			{
				return false;
			}
		}

		#endregion

		#region IEnumerable<T>

		[DebuggerStepThrough]
		public IEnumerator<T> GetEnumerator()
		{
			return this._list.GetEnumerator();
		}

		#endregion

		#region IBindingList

		#region Methods

		[DebuggerStepThrough]
		public void AddIndex(PropertyDescriptor property)
		{
			throw new NotSupportedException(IndexingExceptionMessage);
		}

		[DebuggerStepThrough]
		public object AddNew()
		{
			T item = CreateNew();

			Add(item);

			return item;
		}

		[DebuggerStepThrough]
		public void ApplySort(PropertyDescriptor property, ListSortDirection direction)
		{
			throw new NotSupportedException(SortingExceptionMessage);
		}

		[DebuggerStepThrough]
		public int Find(PropertyDescriptor property, object key)
		{
			throw new NotSupportedException(SearchExceptionMessage);
		}

		[DebuggerStepThrough]
		public void RemoveIndex(PropertyDescriptor property)
		{
			throw new NotSupportedException(IndexingExceptionMessage);
		}

		[DebuggerStepThrough]
		public void RemoveSort()
		{
			throw new NotSupportedException(SortingExceptionMessage);
		}

		#endregion

		#region Properties

		public bool AllowNew
		{
			[DebuggerStepThrough]
			get { return this._list.IsAddAllowed; }
		}

		public bool AllowRemove
		{
			[DebuggerStepThrough]
			get { return this._list.IsRemoveAllowed; }
		}

		public bool AllowEdit
		{
			[DebuggerStepThrough]
			get { return this._list.IsChangeAllowed; }
		}

		public bool IsSorted
		{
			[DebuggerStepThrough]
			get { return false; }
		}

		public ListSortDirection SortDirection
		{
			[DebuggerStepThrough]
			get { throw new NotSupportedException(SortingExceptionMessage); }
		}

		public PropertyDescriptor SortProperty
		{
			[DebuggerStepThrough]
			get { throw new NotSupportedException(SortingExceptionMessage); }
		}

		public bool SupportsChangeNotification
		{
			[DebuggerStepThrough]
			get { return true; }
		}

		public bool SupportsSearching
		{
			[DebuggerStepThrough]
			get { return false; }
		}

		public bool SupportsSorting
		{
			[DebuggerStepThrough]
			get { return false; }
		}

		#endregion

		#region Events

		[NonSerialized]
		private ListChangedEventHandler _listChangedEventHandler;

		public event ListChangedEventHandler ListChanged
		{
			[DebuggerStepThrough]
			add { this._listChangedEventHandler += value; }
			[DebuggerStepThrough]
			remove { this._listChangedEventHandler -= value; }
		}

		#endregion

		#endregion

		#region IList

		#region Methods

		[DebuggerStepThrough]
		public int Add(object value)
		{
			T item = (T)value;

			if (RaiseAdding((this._list as System.Collections.IList).Count, item))
			{
				int index = (this._list as System.Collections.IList).Add(item);

				RaiseAdded(index, item);
				RaiseListAdd(index);

				return index;
			}
			else
			{
				return -1;
			}
		}

		[DebuggerStepThrough]
		public void Remove(object value)
		{
			T item = (T)value;

			int index = (this._list as System.Collections.IList).IndexOf(item);

			if (index != -1)
			{
				if (RaiseRemoving(index, item))
				{
					(this._list as System.Collections.IList).RemoveAt(index);
					RaiseRemoved(index, item);
					RaiseListDelete(index);
				}
			}
		}

		[DebuggerStepThrough]
		public void RemoveAt(int index)
		{
			T item = (T)(this._list as System.Collections.IList)[index];

			if (RaiseRemoving(index, item))
			{
				(this._list as System.Collections.IList).RemoveAt(index);
				RaiseRemoved(index, item);
				RaiseListDelete(index);
			}
		}

		[DebuggerStepThrough]
		public void Clear()
		{
			if (RaiseReseting())
			{
				(this._list as System.Collections.IList).Clear();
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public bool Contains(object value)
		{
			return this._list.Contains((T)value);
		}

		[DebuggerStepThrough]
		public int IndexOf(object value)
		{
			return this._list.IndexOf((T)value);
		}

		[DebuggerStepThrough]
		public void Insert(int index, object value)
		{
			T item = (T)value;

			if (RaiseAdding(index, item))
			{
				(this._list as System.Collections.IList).Insert(index, item);
				RaiseAdded(index, item);
				RaiseListAdd(index);
			}
		}

		#endregion

		#region Properties

		public bool IsFixedSize
		{
			[DebuggerStepThrough]
			get { return ((System.Collections.IList)this._list).IsFixedSize; }
		}

		public bool IsReadOnly
		{
			[DebuggerStepThrough]
			get { return this._list.IsReadOnly; }
		}

		object System.Collections.IList.this[int index]
		{
			[DebuggerStepThrough]
			get
			{
				return this._list[index];
			}

			[DebuggerStepThrough]
			set
			{
				T itemOld = (T)(this._list as System.Collections.IList)[index];
				T itemNew = (T)value;

				if (RaiseChanging(index, itemOld, itemNew))
				{
					UnsubscribeFromNotifications(itemOld, index);
					(this._list as System.Collections.IList)[index] = value;
					SubscribeToNotifications(itemNew, index);
					RaiseChanged(index, itemOld, itemNew);
					RaiseListChange(index);
				}
			}
		}

		#endregion

		#endregion

		#region ICollection

		#region Methods

		[DebuggerStepThrough]
		public void CopyTo(Array array, int index)
		{
			((System.Collections.ICollection)this._list).CopyTo(array, index);
		}

		#endregion

		#region Properties

		public int Count
		{
			[DebuggerStepThrough]
			get { return this._list.Count; }
		}

		public bool IsSynchronized
		{
			[DebuggerStepThrough]
			get { return false; }
		}

		public object SyncRoot
		{
			[DebuggerStepThrough]
			get { return ((System.Collections.ICollection)this._list).SyncRoot; }
		}

		#endregion

		#endregion

		#region IEnumerable

		[DebuggerStepThrough]
		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
		{
			return (this._list as System.Collections.IEnumerable).GetEnumerator();
		}

		#endregion

		#region IObjectObserver

		[DebuggerStepThrough]
		public bool ObjectChanging(object target, object context)
		{
			return
				RaiseChanging(
					((ObjectObserverContext)context).Index,
					(T)target,
					(T)target);
		}

		[DebuggerStepThrough]
		public void ObjectChanged(object target, object context)
		{
			RaiseChanged(
				((ObjectObserverContext)context).Index,
				(T)target,
				(T)target);
		}

		#endregion

		#region ICancelAddNew

		[DebuggerStepThrough]
		public void CancelNew(int itemIndex)
		{
			if ((this._newItemIndex != -1) && (this._newItemIndex == itemIndex))
			{
				this.RemoveAt(this._newItemIndex);
			}
		}

		[DebuggerStepThrough]
		public void EndNew(int itemIndex)
		{
			if (this._newItemIndex == itemIndex)
			{
				this._newItemIndex = -1;
			}
		}

		#endregion

		#region IRaiseItemChangedEvents

		public bool RaisesItemChangedEvents
		{
			[DebuggerStepThrough]
			get { return true; }
		}

		#endregion
	}

	[Serializable]
	public sealed class BindingListEx<T> : BindingListExBase<T> where T : new()
	{
		#region BindingListEx<T>

		[DebuggerStepThrough]
		public BindingListEx()
			: base()
		{
		}

		[DebuggerStepThrough]
		public BindingListEx(int capacity)
			: base(capacity)
		{
		}

		[DebuggerStepThrough]
		public BindingListEx(IEnumerable<T> collection)
			: base(collection)
		{
		}

		[DebuggerStepThrough]
		public BindingListEx(IEnumerable<T> collection, int capacity)
			: base(collection, capacity)
		{
		}

		[DebuggerStepThrough]
		public BindingListEx(T[] array)
			: base(array)
		{
		}

		[DebuggerStepThrough]
		public BindingListEx(T[] array, bool attach)
			: base(array, attach)
		{
		}

		#endregion

		#region BindingListBase<T>

		[DebuggerStepThrough]
		protected override T CreateNew()
		{
			return new T();
		}

		#endregion
	}

	[Serializable]
	public sealed class BindingListExNoCreate<T> : BindingListExBase<T>
	{
		#region BindingListExNoCreate<T>

		[DebuggerStepThrough]
		public BindingListExNoCreate()
			: base()
		{
		}

		[DebuggerStepThrough]
		public BindingListExNoCreate(int capacity)
			: base(capacity)
		{
		}

		[DebuggerStepThrough]
		public BindingListExNoCreate(IEnumerable<T> collection)
			: base(collection)
		{
		}

		[DebuggerStepThrough]
		public BindingListExNoCreate(IEnumerable<T> collection, int capacity)
			: base(collection, capacity)
		{
		}

		[DebuggerStepThrough]
		public BindingListExNoCreate(T[] array)
			: base(array)
		{
		}

		[DebuggerStepThrough]
		public BindingListExNoCreate(T[] array, bool attach)
			: base(array, attach)
		{
		}

		#endregion

		#region BindingListBase<T>

		[DebuggerStepThrough]
		protected override T CreateNew()
		{
			return default(T);
		}

		#endregion
	}

	[Serializable]
	public sealed class BindingListExString : BindingListExBase<string>
	{
		#region BindingListExString

		[DebuggerStepThrough]
		public BindingListExString()
			: base()
		{
		}

		[DebuggerStepThrough]
		public BindingListExString(int capacity)
			: base(capacity)
		{
		}

		[DebuggerStepThrough]
		public BindingListExString(IEnumerable<string> collection)
			: base(collection)
		{
		}

		[DebuggerStepThrough]
		public BindingListExString(IEnumerable<string> collection, int capacity)
			: base(collection, capacity)
		{
		}

		[DebuggerStepThrough]
		public BindingListExString(string[] array)
			: base(array)
		{
		}

		[DebuggerStepThrough]
		public BindingListExString(string[] array, bool attach)
			: base(array, attach)
		{
		}

		#endregion

		#region BindingListBase<string>

		[DebuggerStepThrough]
		protected override string CreateNew()
		{
			return string.Empty;
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
