﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Collections Library                      */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;

namespace Nabu.Collections
{
	public class BindingListShadow<T> :
		IBindingListEx<T>,
		IBindingList,
		ICancelAddNew,
		IRaiseItemChangedEvents
	{
		#region BindingListShadow

		#region Constants

		private const string IndexingExceptionMessage = "Indexing is not supported.";
		private const string SortingExceptionMessage = "Sorting is not supported.";
		private const string SearchExceptionMessage = "Search is not supported.";

		#endregion

		#region Types

		private sealed class BindingListShadowEnumerator : IEnumerator<T>
		{
			#region BindingListShadowEnumerator

			#region Fields

			private BindingListShadow<T> _list;
			private int _count;
			private int _index;

			#endregion

			#region Methods

			[DebuggerStepThrough]
			public BindingListShadowEnumerator(BindingListShadow<T> list)
			{
				this._list = list;
				this._count = this._list.Count;
				this._index = -1;
			}

			#endregion

			#endregion

			#region IEnumerator<T>

			public T Current
			{
				get
				{
					if (this._list.Count != this._count)
					{
						throw new InvalidOperationException(ListEx<T>.SizeChangeExceptionMessage);
					}

					return this._list[this._index];
				}
			}

			#endregion

			#region IEnumerator

			object System.Collections.IEnumerator.Current
			{
				[DebuggerStepThrough]
				get
				{
					if (this._list.Count != this._count)
					{
						throw new InvalidOperationException(ListEx<T>.SizeChangeExceptionMessage);
					}

					return this._list[this._index];
				}
			}

			[DebuggerStepThrough]
			public bool MoveNext()
			{
				this._index++;

				return this._index < this._count;
			}

			[DebuggerStepThrough]
			public void Reset()
			{
				this._index = -1;
			}

			#endregion

			#region IDisposable

			public void Dispose()
			{
			}

			#endregion
		}

		#endregion

		#region Fields

		private IBindingListEx<T> _listEx;
		private IBindingList _list;
		private ICancelAddNew _cancel;
		private IListEx<int> _forwardIndices;
		private IListEx<int> _reverseIndices;

		#endregion

		#region Methods

		#region Utility

		private void CreateMapping()
		{
			for (int i = 0; i < this._listEx.Count; i++)
			{
				this._forwardIndices.Add(i);
				this._reverseIndices.Add(i);
			}
		}

		private void DestroyMapping()
		{
			this._forwardIndices.Clear();
			this._reverseIndices.Clear();
		}

		private void WidenMapping(int index)
		{
			for (int i = 0; i < this._forwardIndices.Count; i++)
			{
				if (this._forwardIndices[i] >= index)
				{
					this._forwardIndices[i] = this._forwardIndices[i] + 1;
				}
			}

			for (int i = 0; i < this._reverseIndices.Count; i++)
			{
				if (this._reverseIndices[i] >= this._reverseIndices[index])
				{
					this._reverseIndices[i] = this._reverseIndices[i] + 1;
				}
			}

			this._forwardIndices.Insert(this._reverseIndices[index], index);
			this._reverseIndices.Insert(index, this._reverseIndices[index]);
		}

		private void NarrowMapping(int index)
		{
			for (int i = 0; i < this._forwardIndices.Count; i++)
			{
				if (this._forwardIndices[i] > index)
				{
					this._forwardIndices[i] = this._forwardIndices[i] - 1;
				}
			}

			for (int i = 0; i < this._reverseIndices.Count; i++)
			{
				if (this._reverseIndices[i] > this._reverseIndices[index])
				{
					this._reverseIndices[i] = this._reverseIndices[i] - 1;
				}
			}

			this._forwardIndices.RemoveAt(this._reverseIndices[index]);
			this._reverseIndices.RemoveAt(index);
		}

		#endregion

		#region Event Handlers

		#region IBindingList Events

		private void List_ListChanged(object sender, ListChangedEventArgs e)
		{
			RaiseListChanged(
				e.ListChangedType,
				this._reverseIndices[e.NewIndex],
				this._reverseIndices[e.OldIndex]);
		}

		#endregion

		#region IBindingListEx<T> Events

		private void List_Reseting(object sender, CancelEventArgs e)
		{
			if (RaiseReseting())
			{
				DestroyMapping();
			}
			else
			{
				e.Cancel = true;
			}
		}

		private void List_Reseted(object sender, EventArgs e)
		{
			RaiseReseted();
		}

		private void List_Adding(object sender, BindingListExAddingRemovingEventArgs<T> e)
		{
			if (RaiseAdding(this._reverseIndices[e.Index], e.Value))
			{
				WidenMapping(e.Index);
			}
			else
			{
				e.Cancel = true;
			}
		}

		private void List_Added(object sender, BindingListExAddedRemovedEventArgs<T> e)
		{
			RaiseAdded(this._reverseIndices[e.Index], e.Value);
		}

		private void List_Removing(object sender, BindingListExAddingRemovingEventArgs<T> e)
		{
			if (RaiseRemoving(this._reverseIndices[e.Index], e.Value))
			{
				NarrowMapping(e.Index);
			}
			else
			{
				e.Cancel = true;
			}
		}

		private void List_Removed(object sender, BindingListExAddedRemovedEventArgs<T> e)
		{
			RaiseRemoved(this._reverseIndices[e.Index], e.Value);
		}

		private void List_Changing(object sender, BindingListExChangingEventArgs<T> e)
		{
			if (RaiseChanging(this._reverseIndices[e.Index], e.OldValue, e.NewValue))
			{

			}
			else
			{
				e.Cancel = true;
			}
		}

		private void List_Changed(object sender, BindingListExChangedEventArgs<T> e)
		{
			RaiseChanged(this._reverseIndices[e.Index], e.OldValue, e.NewValue);
		}

		private void List_Moving(object sender, BindingListExMovingEventArgs<T> e)
		{
			if (RaiseMoving(this._reverseIndices[e.Index1], this._reverseIndices[e.Index2]))
			{

			}
			else
			{
				e.Cancel = true;
			}
		}

		private void List_Moved(object sender, BindingListExMovedEventArgs<T> e)
		{
			RaiseMoved(this._reverseIndices[e.Index1], this._reverseIndices[e.Index2]);
		}

		#endregion

		#endregion

		#region Event Raisers

		#region IBindingList Events

		private void RaiseListChanged(ListChangedType type, int indexNew, int indexOld)
		{
			ListChangedEventArgs eventArgs = new ListChangedEventArgs(type, indexNew, indexOld);

			if (this._listChangedEventHandler != null)
			{
				this._listChangedEventHandler(this, eventArgs);
			}
		}

		private void RaiseListReset()
		{
			RaiseListChanged(ListChangedType.Reset, 0, 0);
		}

		private void RaiseListAdd(int index)
		{
			RaiseListChanged(ListChangedType.ItemAdded, index, 0);
		}

		private void RaiseListDelete(int index)
		{
			RaiseListChanged(ListChangedType.ItemDeleted, index, 0);
		}

		private void RaiseListChange(int index)
		{
			RaiseListChanged(ListChangedType.ItemChanged, index, 0);
		}

		private void RaiseListMove(int indexNew, int indexOld)
		{
			RaiseListChanged(ListChangedType.ItemMoved, indexNew, indexOld);
		}

		#endregion

		#region IBindingListEx<T> Events

		private bool RaiseReseting()
		{
			bool result;

			if (this._resetingHandler != null)
			{
				CancelEventArgs eventArgs = new CancelEventArgs();

				this._resetingHandler(this, eventArgs);

				result = !eventArgs.Cancel;
			}
			else
			{
				result = true;
			}

			return result;
		}

		private void RaiseReseted()
		{
			if (this._resetedHandler != null)
			{
				this._resetedHandler(this, EventArgs.Empty);
			}
		}

		private bool RaiseAdding(int index, T value)
		{
			bool result;

			if (this._addingHandler != null)
			{
				BindingListExAddingRemovingEventArgs<T> eventArgs =
					new BindingListExAddingRemovingEventArgs<T>(
						index,
						value);

				this._addingHandler(this, eventArgs);

				result = !eventArgs.Cancel;
			}
			else
			{
				result = true;
			}

			return result;
		}

		private void RaiseAdded(int index, T value)
		{
			if (this._addedHandler != null)
			{
				this._addedHandler(
					this,
					new BindingListExAddedRemovedEventArgs<T>(
						index,
						value));
			}
		}

		private bool RaiseRemoving(int index, T value)
		{
			bool result;

			if (this._removingHandler != null)
			{
				BindingListExAddingRemovingEventArgs<T> eventArgs =
					new BindingListExAddingRemovingEventArgs<T>(
						index,
						value);

				this._removingHandler(this, eventArgs);

				result = !eventArgs.Cancel;
			}
			else
			{
				result = true;
			}

			return result;
		}

		private void RaiseRemoved(int index, T value)
		{
			if (this._removedHandler != null)
			{
				this._removedHandler(
					this,
					new BindingListExAddedRemovedEventArgs<T>(
						index,
						value));
			}
		}

		private bool RaiseChanging(int index, T oldValue, T newValue)
		{
			bool result;

			if (this._changingHandler != null)
			{
				BindingListExChangingEventArgs<T> eventArgs =
					new BindingListExChangingEventArgs<T>(
						index,
						oldValue,
						newValue);

				this._changingHandler(this, eventArgs);

				result = !eventArgs.Cancel;
			}
			else
			{
				result = true;
			}

			return result;
		}

		private void RaiseChanged(int index, T oldValue, T newValue)
		{
			if (this._changedHandler != null)
			{
				this._changedHandler(
					this,
					new BindingListExChangedEventArgs<T>(
						index,
						oldValue,
						newValue));
			}
		}

		private bool RaiseMoving(int index1, int index2)
		{
			bool result;

			if (this._movingHandler != null)
			{
				BindingListExMovingEventArgs<T> eventArgs =
					new BindingListExMovingEventArgs<T>(
						index1,
						index2);

				this._movingHandler(this, eventArgs);

				result = !eventArgs.Cancel;
			}
			else
			{
				result = true;
			}

			return result;
		}

		private void RaiseMoved(int index1, int index2)
		{
			if (this._movedHandler != null)
			{
				this._movedHandler(
					this,
					new BindingListExMovedEventArgs<T>(
						index1,
						index2));
			}
		}

		#endregion

		#endregion

		#region Contructors

		[DebuggerStepThrough]
		protected BindingListShadow(IBindingListEx<T> list)
		{
			this._listEx = list;
			this._listEx.Reseting += new CancelEventHandler(List_Reseting);
			this._listEx.Reseted += new EventHandler(List_Reseted);
			this._listEx.Adding += new EventHandler<BindingListExAddingRemovingEventArgs<T>>(List_Adding);
			this._listEx.Added += new EventHandler<BindingListExAddedRemovedEventArgs<T>>(List_Added);
			this._listEx.Removing += new EventHandler<BindingListExAddingRemovingEventArgs<T>>(List_Removing);
			this._listEx.Removed += new EventHandler<BindingListExAddedRemovedEventArgs<T>>(List_Removed);
			this._listEx.Changing += new EventHandler<BindingListExChangingEventArgs<T>>(List_Changing);
			this._listEx.Changed += new EventHandler<BindingListExChangedEventArgs<T>>(List_Changed);
			this._listEx.Moving += new EventHandler<BindingListExMovingEventArgs<T>>(List_Moving);
			this._listEx.Moved += new EventHandler<BindingListExMovedEventArgs<T>>(List_Moved);
			this._list = (IBindingList)list;
			this._list.ListChanged += new ListChangedEventHandler(List_ListChanged);
			this._cancel = (ICancelAddNew)list;
			this._forwardIndices = new ListEx<int>(this._listEx.Count);
			this._reverseIndices = new ListEx<int>(this._listEx.Count);
			CreateMapping();
		}

		#endregion

		#endregion

		#endregion

		#region IBindingListEx<T>

		#region Methods

		[DebuggerStepThrough]
		public void Swap(int index1, int index2)
		{
			if (RaiseMoving(index1, index2))
			{
				int i1 = this._forwardIndices[index1];
				int i2 = this._forwardIndices[index2];

				int temp;

				temp = this._forwardIndices[index1];
				this._forwardIndices[index1] = this._forwardIndices[index2];
				this._forwardIndices[index2] = temp;
				temp = this._reverseIndices[i1];
				this._reverseIndices[i1] = this._reverseIndices[i2];
				this._reverseIndices[i2] = temp;

				RaiseMoved(index1, index2);
				RaiseListMove(index1, index2);
			}
		}

		#endregion

		#region Events

		#region Reseting

		[NonSerialized]
		private CancelEventHandler _resetingHandler;

		public event CancelEventHandler Reseting
		{
			add { this._resetingHandler += value; }
			remove { this._resetingHandler -= value; }
		}

		#endregion

		#region Reseted

		[NonSerialized]
		private EventHandler _resetedHandler;

		public event EventHandler Reseted
		{
			add { this._resetedHandler += value; }
			remove { this._resetedHandler -= value; }
		}

		#endregion

		#region Adding

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExAddingRemovingEventArgs<T>> _addingHandler;

		public event EventHandler<BindingListExAddingRemovingEventArgs<T>> Adding
		{
			add { this._addingHandler += value; }
			remove { this._addingHandler -= value; }
		}

		#endregion

		#region Added

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExAddedRemovedEventArgs<T>> _addedHandler;

		public event EventHandler<BindingListExAddedRemovedEventArgs<T>> Added
		{
			add { this._addedHandler += value; }
			remove { this._addedHandler -= value; }
		}

		#endregion

		#region Removing

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExAddingRemovingEventArgs<T>> _removingHandler;

		public event EventHandler<BindingListExAddingRemovingEventArgs<T>> Removing
		{
			add { this._removingHandler += value; }
			remove { this._removingHandler -= value; }
		}

		#endregion

		#region Removed

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExAddedRemovedEventArgs<T>> _removedHandler;

		public event EventHandler<BindingListExAddedRemovedEventArgs<T>> Removed
		{
			add { this._removedHandler += value; }
			remove { this._removedHandler -= value; }
		}

		#endregion

		#region Changing

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExChangingEventArgs<T>> _changingHandler;

		public event EventHandler<BindingListExChangingEventArgs<T>> Changing
		{
			add { this._changingHandler += value; }
			remove { this._changingHandler -= value; }
		}

		#endregion

		#region Changed

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExChangedEventArgs<T>> _changedHandler;

		public event EventHandler<BindingListExChangedEventArgs<T>> Changed
		{
			add { this._changedHandler += value; }
			remove { this._changedHandler -= value; }
		}

		#endregion

		#region Moving

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExMovingEventArgs<T>> _movingHandler;

		public event EventHandler<BindingListExMovingEventArgs<T>> Moving
		{
			add { this._movingHandler += value; }
			remove { this._movingHandler -= value; }
		}

		#endregion

		#region Moved

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		[NonSerialized]
		private EventHandler<BindingListExMovedEventArgs<T>> _movedHandler;

		public event EventHandler<BindingListExMovedEventArgs<T>> Moved
		{
			add { this._movedHandler += value; }
			remove { this._movedHandler -= value; }
		}

		#endregion

		#endregion

		#region Propeties

		public IListEx<T> InnerList
		{
			[DebuggerStepThrough]
			get { throw new NotImplementedException(); }
		}

		#endregion

		#endregion

		#region IListEx<T>

		#region Methods

		#region Capacity

		[DebuggerStepThrough]
		public void EnsureCapacity(int capacity)
		{
			this._listEx.EnsureCapacity(capacity);
		}

		[DebuggerStepThrough]
		public void TrimExcess(float ratio)
		{
			this._listEx.TrimExcess(ratio);
		}

		[DebuggerStepThrough]
		public void TrimExcess()
		{
			this._listEx.TrimExcess();
		}

		#endregion

		#region Range

		[DebuggerStepThrough]
		public IListEx<T> GetRange(int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex + count) > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			T[] array = new T[count];

			for (int i = 0; i < count; i++)
			{
				array[i] = this._listEx[this._forwardIndices[startIndex + i]];
			}

			return new ListEx<T>(array, true);
		}

		[DebuggerStepThrough]
		public void AddRange(IEnumerable<T> collection)
		{
			if (RaiseReseting())
			{
				this._listEx.AddRange(collection);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void AddRange(IListEx<T> list, int offset, int count)
		{
			if (RaiseReseting())
			{
				this._listEx.AddRange(list, offset, count);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void AddRange(IListEx<T> list)
		{
			if (RaiseReseting())
			{
				this._listEx.AddRange(list);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void AddRange(T[] array, int offset, int count)
		{
			if (RaiseReseting())
			{
				this._listEx.AddRange(array, offset, count);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void AddRange(T[] array)
		{
			if (RaiseReseting())
			{
				this._listEx.AddRange(array);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void InsertRange(int startIndex, IEnumerable<T> collection)
		{
			if (RaiseReseting())
			{
				this._listEx.InsertRange(startIndex, collection);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void InsertRange(int startIndex, IListEx<T> list, int offset, int count)
		{
			if (RaiseReseting())
			{
				this._listEx.InsertRange(startIndex, list, offset, count);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void InsertRange(int startIndex, IListEx<T> list)
		{
			if (RaiseReseting())
			{
				this._listEx.InsertRange(startIndex, list);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void InsertRange(int startIndex, T[] array, int offset, int count)
		{
			if (RaiseReseting())
			{
				this._listEx.InsertRange(startIndex, array, offset, count);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void InsertRange(int startIndex, T[] array)
		{
			if (RaiseReseting())
			{
				this._listEx.InsertRange(startIndex, array);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void RemoveRange(int startIndex, int count)
		{
			if (RaiseReseting())
			{
				this._listEx.RemoveRange(startIndex, count);
				RaiseReseted();
				RaiseListReset();
			}
		}

		#endregion

		#region Copy and Convert

		public void FromArray(T[] array, bool attach)
		{
			if (RaiseReseting())
			{
				this._listEx.FromArray(array, attach);
				RaiseReseted();
				RaiseListReset();
			}
		}

		public void FromArray(T[] array)
		{
			if (RaiseReseting())
			{
				this._listEx.FromArray(array);
				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public T[] ToArray()
		{
			T[] array = new T[this._listEx.Count];

			for (int i = 0; i < this._listEx.Count; i++)
			{
				array[i] = this._listEx[this._forwardIndices[i]];
			}

			return array;
		}

		[DebuggerStepThrough]
		public T[] DangerousGetArray()
		{
			throw new NotImplementedException();
		}

		[DebuggerStepThrough]
		public ArraySegment<T> ToArraySegment()
		{
			return new ArraySegment<T>(ToArray());
		}

		[DebuggerStepThrough]
		public void CopyTo(T[] array)
		{
			if (array == null)
			{
				throw new ArgumentNullException("array");
			}

			if (this._listEx.Count > array.Length)
			{
				throw new ArgumentOutOfRangeException("array");
			}

			for (int i = 0; i < this._listEx.Count; i++)
			{
				array[i] = this._listEx[this._forwardIndices[i]];
			}
		}

		[DebuggerStepThrough]
		public void CopyTo(int index, T[] array, int arrayIndex, int count)
		{
			if ((index < 0) || (index > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("index");
			}

			if (array == null)
			{
				throw new ArgumentNullException("array");
			}

			if ((arrayIndex < 0) || (arrayIndex > array.Length))
			{
				throw new ArgumentOutOfRangeException("arrayIndex");
			}

			if (
				(count < 0) ||
				((index + count) > this._listEx.Count) ||
				((arrayIndex + count) > array.Length))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			for (int i = 0; i < this._listEx.Count; i++)
			{
				array[i + index] = this._listEx[this._forwardIndices[arrayIndex + i]];
			}
		}

		[DebuggerStepThrough]
		public IListEx<TOutput> ConvertAll<TOutput>(Converter<T, TOutput> converter)
		{
			IListEx<TOutput> list = new ListEx<TOutput>(this._listEx.Count);

			for (int i = 0; i < this._listEx.Count; i++)
			{
				list.Add(converter(this._listEx[this._forwardIndices[i]]));
			}

			return list;
		}

		#endregion

		#region Reverse

		[DebuggerStepThrough]
		public void Reverse(int index, int count)
		{
			if (RaiseReseting())
			{
				this._forwardIndices.Reverse(index, count);

				for (int i = index; i < index + count; i++)
				{
					this._reverseIndices[this._forwardIndices[i]] = i;
				}

				RaiseReseted();
				RaiseListReset();
			}
		}

		[DebuggerStepThrough]
		public void Reverse()
		{
			Reverse(0, this._listEx.Count);
		}

		#endregion

		#region Sort

		[DebuggerStepThrough]
		public void Sort(int startIndex, int count)
		{
			throw new NotImplementedException();
		}

		[DebuggerStepThrough]
		public void Sort()
		{
			throw new NotImplementedException();
		}

		[DebuggerStepThrough]
		public void Sort(IComparer<T> comparer, int startIndex, int count)
		{
			throw new NotImplementedException();
		}

		[DebuggerStepThrough]
		public void Sort(IComparer<T> comparer)
		{
			throw new NotImplementedException();
		}

		[DebuggerStepThrough]
		public void Sort(Comparison<T> comparer, int startIndex, int count)
		{
			throw new NotImplementedException();
		}

		[DebuggerStepThrough]
		public void Sort(Comparison<T> comparer)
		{
			throw new NotImplementedException();
		}

		#endregion

		#region Search

		[DebuggerStepThrough]
		public int IndexOf(T item, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex + count) > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (item == null)
			{
				for (int i = startIndex; i < startIndex + count; i++)
				{
					if (this._listEx[this._forwardIndices[i]] == null)
					{
						return i;
					}
				}
			}
			else
			{
				for (int i = startIndex; i < startIndex + count; i++)
				{
					if (item.Equals(this._listEx[this._forwardIndices[i]]))
					{
						return i;
					}
				}
			}

			return -1;
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, int startIndex)
		{
			return IndexOf(item, startIndex, this._listEx.Count - startIndex);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, IComparer<T> comparer, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex + count) > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (item == null)
			{
				for (int i = startIndex; i < startIndex + count; i++)
				{
					if (this._listEx[this._forwardIndices[i]] == null)
					{
						return i;
					}
				}
			}
			else
			{
				for (int i = startIndex; i < startIndex + count; i++)
				{
					if (comparer.Compare(item, this._listEx[this._forwardIndices[i]]) == 0)
					{
						return i;
					}
				}
			}

			return -1;
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, IComparer<T> comparer, int startIndex)
		{
			return IndexOf(item, comparer, startIndex, this._listEx.Count - startIndex);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, IComparer<T> comparer)
		{
			return this._listEx.IndexOf(item, comparer, 0, this._listEx.Count);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, Comparison<T> comparer, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex + count) > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (item == null)
			{
				for (int i = startIndex; i < startIndex + count; i++)
				{
					if (this._listEx[this._forwardIndices[i]] == null)
					{
						return i;
					}
				}
			}
			else
			{
				for (int i = startIndex; i < startIndex + count; i++)
				{
					if (comparer(item, this._listEx[this._forwardIndices[i]]) == 0)
					{
						return i;
					}
				}
			}

			return -1;
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, Comparison<T> comparer, int startIndex)
		{
			return IndexOf(item, comparer, startIndex, this._listEx.Count - startIndex);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, Comparison<T> comparer)
		{
			return IndexOf(item, comparer, 0, this._listEx.Count);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex - count) > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (item == null)
			{
				for (int index = startIndex; index > startIndex - count; index--)
				{
					if (this._listEx[index] == null)
					{
						return index;
					}
				}
			}
			else
			{
				for (int index = startIndex; index > startIndex - count; index--)
				{
					if (item.Equals(this._listEx[index]))
					{
						return index;
					}
				}
			}

			return -1;
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, int startIndex)
		{
			return this._listEx.LastIndexOf(item, startIndex);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item)
		{
			return this._listEx.LastIndexOf(item);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, IComparer<T> comparer, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex - count) > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			for (int index = startIndex; index > startIndex - count; index--)
			{
				if (comparer.Compare(item, this._listEx[index]) == 0)
				{
					return index;
				}
			}

			return -1;
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, IComparer<T> comparer, int startIndex)
		{
			return LastIndexOf(item, comparer, startIndex, this._listEx.Count - startIndex);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, IComparer<T> comparer)
		{
			return LastIndexOf(item, comparer, 0, this._listEx.Count);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, Comparison<T> comparer, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex - count) > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			for (int index = startIndex; index > startIndex - count; index--)
			{
				if (comparer(item, this._listEx[index]) == 0)
				{
					return index;
				}
			}

			return -1;
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, Comparison<T> comparer, int startIndex)
		{
			return LastIndexOf(item, comparer, startIndex, this._listEx.Count - startIndex);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, Comparison<T> comparer)
		{
			return LastIndexOf(item, comparer, 0, this._listEx.Count);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, int startIndex, int count)
		{
			throw new NotImplementedException();
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, int startIndex)
		{
			throw new NotImplementedException();
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item)
		{
			int index = this._listEx.BinarySearch(item);

			if (index == -1)
			{
				return -1;
			}
			else
			{
				return this._reverseIndices[index];
			}
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, IComparer<T> comparer, int startIndex, int count)
		{
			throw new NotImplementedException();
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, IComparer<T> comparer, int startIndex)
		{
			throw new NotImplementedException();
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, IComparer<T> comparer)
		{
			int index = this._listEx.BinarySearch(item, comparer);

			if (index == -1)
			{
				return -1;
			}
			else
			{
				return this._reverseIndices[index];
			}
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, Comparison<T> comparer, int startIndex, int count)
		{
			throw new NotImplementedException();
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, Comparison<T> comparer, int startIndex)
		{
			throw new NotImplementedException();
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, Comparison<T> comparer)
		{
			int index = this._listEx.BinarySearch(item, comparer);

			if (index == -1)
			{
				return -1;
			}
			else
			{
				return this._reverseIndices[index];
			}
		}

		#endregion

		#region Predicates

		#region Exists

		[DebuggerStepThrough]
		public bool Exists(Predicate<T> match)
		{
			return this._listEx.Exists(match);
		}

		#endregion

		#region FindAll

		[DebuggerStepThrough]
		public IListEx<T> FindAll(Predicate<T> match, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex - count) > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			ListEx<T> list = new ListEx<T>();

			for (int i = startIndex; i < startIndex + count; i++)
			{
				if (match(this._listEx[this._forwardIndices[i]]))
				{
					list.Add(this._listEx[this._forwardIndices[i]]);
				}
			}

			return list;
		}

		[DebuggerStepThrough]
		public IListEx<T> FindAll(Predicate<T> match, int startIndex)
		{
			return FindAll(match, startIndex, this._listEx.Count - startIndex);
		}

		[DebuggerStepThrough]
		public IListEx<T> FindAll(Predicate<T> match)
		{
			return FindAll(match, 0, this._listEx.Count);
		}

		[DebuggerStepThrough]
		public IListEx<int> FindAllIndex(Predicate<T> match, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex - count) > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			ListEx<int> list = new ListEx<int>();

			for (int i = startIndex; i < startIndex + count; i++)
			{
				if (match(this._listEx[this._forwardIndices[i]]))
				{
					list.Add(i);
				}
			}

			return list;
		}

		[DebuggerStepThrough]
		public IListEx<int> FindAllIndex(Predicate<T> match, int startIndex)
		{
			return FindAllIndex(match, startIndex, this._listEx.Count - startIndex);
		}

		[DebuggerStepThrough]
		public IListEx<int> FindAllIndex(Predicate<T> match)
		{
			return FindAllIndex(match, 0, this._listEx.Count);
		}

		#endregion

		#region FindFirst

		[DebuggerStepThrough]
		public int FindFirstIndex(Predicate<T> match, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex - count) > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			for (int i = startIndex; i < startIndex + count; i++)
			{
				if (match(this._listEx[this._forwardIndices[i]]))
				{
					return i;
				}
			}

			return -1;
		}

		[DebuggerStepThrough]
		public int FindFirstIndex(Predicate<T> match, int startIndex)
		{
			return FindFirstIndex(match, startIndex, this._listEx.Count - startIndex);
		}

		[DebuggerStepThrough]
		public int FindFirstIndex(Predicate<T> match)
		{
			return FindFirstIndex(match, 0, this._listEx.Count);
		}

		#endregion

		#region FindLast

		[DebuggerStepThrough]
		public int FindLastIndex(Predicate<T> match, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if (
				(count < 0) ||
				((startIndex - count) < 0) ||
				((startIndex - count) > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (match == null)
			{
				throw new ArgumentNullException("match");
			}

			for (int index = startIndex; index > startIndex - count; --index)
			{
				if (match(this._listEx[this._forwardIndices[index]]))
				{
					return index;
				}
			}

			return -1;
		}

		[DebuggerStepThrough]
		public int FindLastIndex(Predicate<T> match, int startIndex)
		{
			return FindLastIndex(match, startIndex, this._listEx.Count - startIndex);
		}

		[DebuggerStepThrough]
		public int FindLastIndex(Predicate<T> match)
		{
			return FindLastIndex(match, 0, this._listEx.Count);
		}

		#endregion

		#region RemoveAll

		[DebuggerStepThrough]
		public int RemoveAll(Predicate<T> match, int startIndex, int count)
		{
			IListEx<int> list = FindAllIndex(match, startIndex, count);

			foreach (int index in list)
			{
				this._listEx.RemoveAt(this._forwardIndices[index]);
			}

			return list.Count;
		}

		[DebuggerStepThrough]
		public int RemoveAll(Predicate<T> match, int startIndex)
		{
			return RemoveAll(match, startIndex, this._listEx.Count - startIndex);
		}

		[DebuggerStepThrough]
		public int RemoveAll(Predicate<T> match)
		{
			return this._listEx.RemoveAll(match);
		}

		#endregion

		#region TrueForAll

		[DebuggerStepThrough]
		public bool TrueForAll(Predicate<T> match, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if (
				(count < 0) ||
				((startIndex + count) < 0) ||
				((startIndex + count) > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (match == null)
			{
				throw new ArgumentNullException("match");
			}

			for (int index = startIndex; index < startIndex + count; index++)
			{
				if (!match(this._listEx[this._forwardIndices[index]]))
				{
					return false;
				}
			}

			return true;
		}

		[DebuggerStepThrough]
		public bool TrueForAll(Predicate<T> match, int startIndex)
		{
			return TrueForAll(match, startIndex, this._listEx.Count - startIndex);
		}

		[DebuggerStepThrough]
		public bool TrueForAll(Predicate<T> match)
		{
			return this._listEx.TrueForAll(match);
		}

		#endregion

		#region FalseForAll

		[DebuggerStepThrough]
		public bool FalseForAll(Predicate<T> match, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if (
				(count < 0) ||
				((startIndex + count) < 0) ||
				((startIndex + count) > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (match == null)
			{
				throw new ArgumentNullException("match");
			}

			for (int index = startIndex; index < startIndex + count; index++)
			{
				if (match(this._listEx[this._forwardIndices[index]]))
				{
					return false;
				}
			}

			return true;
		}

		[DebuggerStepThrough]
		public bool FalseForAll(Predicate<T> match, int startIndex)
		{
			return FalseForAll(match, startIndex, this._listEx.Count - startIndex);
		}

		[DebuggerStepThrough]
		public bool FalseForAll(Predicate<T> match)
		{
			return this._listEx.FalseForAll(match);
		}

		#endregion

		#endregion

		#region Actions

		[DebuggerStepThrough]
		public void ForEach(Action<T> action, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if (
				(count < 0) ||
				((startIndex + count) < 0) ||
				((startIndex + count) > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			for (int index = startIndex; index < startIndex + count; index++)
			{
				action(this._listEx[this._forwardIndices[index]]);
			}
		}

		[DebuggerStepThrough]
		public void ForEach(Action<T> action, int startIndex)
		{
			ForEach(action, startIndex, this._listEx.Count - startIndex);
		}

		[DebuggerStepThrough]
		public void ForEach(Action<T> action)
		{
			this._listEx.ForEach(action);
		}

		[DebuggerStepThrough]
		public void ForEach(Action<T> action, Predicate<T> match, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if (
				(count < 0) ||
				((startIndex + count) < 0) ||
				((startIndex + count) > this._listEx.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			for (int index = startIndex; index < startIndex + count; index++)
			{
				if (match(this._listEx[this._forwardIndices[index]]))
				{
					action(this._listEx[this._forwardIndices[index]]);
				}
			}
		}

		[DebuggerStepThrough]
		public void ForEach(Action<T> action, Predicate<T> match, int startIndex)
		{
			ForEach(action, match, startIndex, this._listEx.Count - startIndex);
		}

		[DebuggerStepThrough]
		public void ForEach(Action<T> action, Predicate<T> match)
		{
			this._listEx.ForEach(action, match);
		}

		#endregion

		#endregion

		#region Properties

		public int Capacity
		{
			[DebuggerStepThrough]
			get { return this._listEx.Capacity; }
			[DebuggerStepThrough]
			set { this._listEx.Capacity = value; }
		}

		public T First
		{
			[DebuggerStepThrough]
			get { return this._listEx[this._forwardIndices[0]]; }
		}

		public T Last
		{
			[DebuggerStepThrough]
			get { return this._listEx[this._forwardIndices[this._listEx.Count - 1]]; }
		}

		public bool IsAddAllowed
		{
			[DebuggerStepThrough]
			get { return this._listEx.IsAddAllowed; }
			[DebuggerStepThrough]
			set { this._listEx.IsAddAllowed = value; }
		}

		public bool IsRemoveAllowed
		{
			[DebuggerStepThrough]
			get { return this._listEx.IsRemoveAllowed; }
			[DebuggerStepThrough]
			set { this._listEx.IsRemoveAllowed = value; }
		}

		public bool IsChangeAllowed
		{
			[DebuggerStepThrough]
			get { return this._listEx.IsChangeAllowed; }
			[DebuggerStepThrough]
			set { this._listEx.IsChangeAllowed = value; }
		}

		#endregion

		#endregion

		#region IList<T>

		#region Methods

		[DebuggerStepThrough]
		public int IndexOf(T item)
		{
			int index = this._listEx.IndexOf(item);

			if (index == -1)
			{
				return -1;
			}
			else
			{
				return this._reverseIndices[index];
			}
		}

		[DebuggerStepThrough]
		public void Insert(int index, T item)
		{
			this._listEx.Insert(this._forwardIndices[index], item);
		}

		#endregion

		#region Properties

		public T this[int index]
		{
			[DebuggerStepThrough]
			get { return this._listEx[this._forwardIndices[index]]; }
			[DebuggerStepThrough]
			set { this._listEx[this._forwardIndices[index]] = value; }
		}

		#endregion

		#endregion

		#region ICollection<T>

		[DebuggerStepThrough]
		public void Add(T item)
		{
			this._listEx.Add(item);
		}

		[DebuggerStepThrough]
		public bool Contains(T item)
		{
			return this._listEx.Contains(item);
		}

		[DebuggerStepThrough]
		public void CopyTo(T[] array, int arrayIndex)
		{
			if (array == null)
			{
				throw new ArgumentNullException("array");
			}

			if (
				(arrayIndex < 0) ||
				(arrayIndex > array.Length) ||
				((arrayIndex + this._listEx.Count) > array.Length))
			{
				throw new ArgumentOutOfRangeException("arrayIndex");
			}

			for (int i = 0; i < this._listEx.Count; i++)
			{
				array[arrayIndex + i] = this._listEx[this._forwardIndices[i]];
			}
		}

		[DebuggerStepThrough]
		public bool Remove(T item)
		{
			return this._listEx.Remove(item);
		}

		#endregion

		#region IEnumerable<T>

		[DebuggerStepThrough]
		public IEnumerator<T> GetEnumerator()
		{
			return new BindingListShadowEnumerator(this);
		}

		#endregion

		#region IBindingList

		#region Methods

		[DebuggerStepThrough]
		public void AddIndex(PropertyDescriptor property)
		{
			throw new NotSupportedException(IndexingExceptionMessage);
		}

		[DebuggerStepThrough]
		public object AddNew()
		{
			return this._list.AddNew();
		}

		[DebuggerStepThrough]
		public void ApplySort(PropertyDescriptor property, ListSortDirection direction)
		{
			throw new NotSupportedException(SortingExceptionMessage);
		}

		[DebuggerStepThrough]
		public int Find(PropertyDescriptor property, object key)
		{
			throw new NotSupportedException(SearchExceptionMessage);
		}

		[DebuggerStepThrough]
		public void RemoveIndex(PropertyDescriptor property)
		{
			throw new NotSupportedException(IndexingExceptionMessage);
		}

		[DebuggerStepThrough]
		public void RemoveSort()
		{
			throw new NotSupportedException(SortingExceptionMessage);
		}

		#endregion

		#region Properties

		public bool AllowNew
		{
			[DebuggerStepThrough]
			get { return this._listEx.IsAddAllowed; }
		}

		public bool AllowRemove
		{
			[DebuggerStepThrough]
			get { return this._listEx.IsRemoveAllowed; }
		}

		public bool AllowEdit
		{
			[DebuggerStepThrough]
			get { return this._listEx.IsChangeAllowed; }
		}

		public bool IsSorted
		{
			[DebuggerStepThrough]
			get { return false; }
		}

		public ListSortDirection SortDirection
		{
			[DebuggerStepThrough]
			get { throw new NotSupportedException(SortingExceptionMessage); }
		}

		public PropertyDescriptor SortProperty
		{
			[DebuggerStepThrough]
			get { throw new NotSupportedException(SortingExceptionMessage); }
		}

		public bool SupportsChangeNotification
		{
			[DebuggerStepThrough]
			get { return true; }
		}

		public bool SupportsSearching
		{
			[DebuggerStepThrough]
			get { return false; }
		}

		public bool SupportsSorting
		{
			[DebuggerStepThrough]
			get { return false; }
		}

		#endregion

		#region Events

		[NonSerialized]
		private ListChangedEventHandler _listChangedEventHandler;

		public event ListChangedEventHandler ListChanged
		{
			add { this._listChangedEventHandler += value; }
			remove { this._listChangedEventHandler -= value; }
		}

		#endregion

		#endregion

		#region IList

		#region Methods

		[DebuggerStepThrough]
		public int Add(object value)
		{
			int index = this._list.Add(value);

			if (index == -1)
			{
				return -1;
			}
			else
			{
				return this._reverseIndices[index];
			}
		}

		[DebuggerStepThrough]
		public void Remove(object value)
		{
			this._list.Remove(value);
		}

		[DebuggerStepThrough]
		public void RemoveAt(int index)
		{
			this._list.RemoveAt(this._forwardIndices[index]);
		}

		[DebuggerStepThrough]
		public void Clear()
		{
			this._list.Clear();
		}

		[DebuggerStepThrough]
		public bool Contains(object value)
		{
			return this._list.Contains(value);
		}

		[DebuggerStepThrough]
		public int IndexOf(object value)
		{
			int index = this._list.IndexOf(value);

			if (index == -1)
			{
				return -1;
			}
			else
			{
				return this._reverseIndices[index];
			}
		}

		[DebuggerStepThrough]
		public void Insert(int index, object value)
		{
			this._list.Insert(this._forwardIndices[index], value);
		}

		#endregion

		#region Properties

		public bool IsFixedSize
		{
			[DebuggerStepThrough]
			get { return this._list.IsFixedSize; }
		}

		public bool IsReadOnly
		{
			[DebuggerStepThrough]
			get { return this._list.IsReadOnly; }
		}

		object System.Collections.IList.this[int index]
		{
			[DebuggerStepThrough]
			get { return this._list[this._forwardIndices[index]]; }
			set { this._list[this._forwardIndices[index]] = value; }
		}

		#endregion

		#endregion

		#region ICollection

		#region Methods

		[DebuggerStepThrough]
		public void CopyTo(Array array, int index)
		{
			if (array == null)
			{
				throw new ArgumentNullException("array");
			}

			if (
				(index < 0) ||
				(index > array.Length) ||
				((index + this._list.Count) > array.Length))
			{
				throw new ArgumentOutOfRangeException("index");
			}

			for (int i = 0; i < this._listEx.Count; i++)
			{
				array.SetValue(this._listEx[this._forwardIndices[i]], index + i);
			}
		}

		#endregion

		#region Properties

		public int Count
		{
			[DebuggerStepThrough]
			get { return this._list.Count; }
		}

		public bool IsSynchronized
		{
			[DebuggerStepThrough]
			get { return false; }
		}

		public object SyncRoot
		{
			[DebuggerStepThrough]
			get { return this._list.SyncRoot; }
		}

		#endregion

		#endregion

		#region IEnumerable

		[DebuggerStepThrough]
		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
		{
			return new BindingListShadowEnumerator(this);
		}

		#endregion

		#region ICancelAddNew

		[DebuggerStepThrough]
		public void CancelNew(int itemIndex)
		{
			this._cancel.CancelNew(this._forwardIndices[itemIndex]);
		}

		[DebuggerStepThrough]
		public void EndNew(int itemIndex)
		{
			this._cancel.EndNew(this._forwardIndices[itemIndex]);
		}

		#endregion

		#region IRaiseItemChangedEvents

		public bool RaisesItemChangedEvents
		{
			[DebuggerStepThrough]
			get { return true; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
