﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Collections Library                      */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;

namespace Nabu.Collections
{
	public class BindingTree<T> : IBindingTree<T>
	{
		#region BindingTree<T>

		#region Types

		//public delegate U PropertyAccessor<U, V>(V obj);

#if !COMPACT_FRAMEWORK
		[System.Diagnostics.DebuggerDisplay("Data: {Data}; Collapsed: {IsCollapsed}; Visible: {IsVisible}")]
#endif
		private class TreeNode : ITreeNode<T>
		{
			#region TreeNode

			#region Fields

			private BindingTree<T> _tree;
			private TreeNode _parent;
			private TreeNode _prevSibling;
			private TreeNode _nextSibling;
			private TreeNode _firstChild;
			private TreeNode _lastChild;
			private T _data;
			private bool _isVisible;
			private bool _isCollapsed;
			private int _totalCount;
			private int _visibleCount;

			#endregion

			#region Methods

			#region Utility

			public void ChangeVisibleCount(int count)
			{
				this._visibleCount += count;

				if (this._isVisible && (!this._isCollapsed))
				{
					if (this._parent != null)
					{
						this._parent.ChangeVisibleCount(count);
					}
				}
			}

			public void ChangeTotalCount(int count)
			{
				this._totalCount += count;

				if (this._parent != null)
				{
					this._parent.ChangeTotalCount(count);
				}
			}

			public void SubscribeToPropertyChanged()
			{
				INotifyPropertyChanged notify = this._data as INotifyPropertyChanged;

				if (notify != null)
				{
					notify.PropertyChanged += new PropertyChangedEventHandler(Data_PropertyChanged);
				}
			}

			public void UnsubscribeFromPropertyChanged()
			{
				INotifyPropertyChanged notify = this._data as INotifyPropertyChanged;

				if (notify != null)
				{
					notify.PropertyChanged -= new PropertyChangedEventHandler(Data_PropertyChanged);
				}
			}

			#endregion

			#region Event Handlers

			private void Data_PropertyChanged(object sender, PropertyChangedEventArgs e)
			{
				this._tree.RaiseChanged(this);
			}

			#endregion

			#region Constructors

			[DebuggerStepThrough]
			public TreeNode(
				BindingTree<T> tree,
				TreeNode parent,
				TreeNode prevSibling,
				TreeNode nextSibling,
				T data,
				bool isVisible,
				bool isCollapsed)
			{
				this._tree = tree;
				this._parent = parent;
				this._prevSibling = prevSibling;
				this._nextSibling = nextSibling;
				this._firstChild = null;
				this._lastChild = null;
				this._data = data;
				this._isVisible = isVisible;
				this._isCollapsed = isCollapsed;
				this._totalCount = 1;
				this._visibleCount = isVisible ? 1 : 0;
			}

			#endregion

			#endregion

			#region Properties

			public TreeNode UnsafeParent
			{
				[DebuggerStepThrough]
				get { return this._parent; }
				[DebuggerStepThrough]
				set { this._parent = value; }
			}

			public TreeNode UnsafeNextSibling
			{
				[DebuggerStepThrough]
				get { return this._nextSibling; }
				[DebuggerStepThrough]
				set { this._nextSibling = value; }
			}

			public TreeNode UnsafePrevSibling
			{
				[DebuggerStepThrough]
				get { return this._prevSibling; }
				[DebuggerStepThrough]
				set { this._prevSibling = value; }
			}

			public TreeNode UnsafeFirstChild
			{
				[DebuggerStepThrough]
				get { return this._firstChild; }
				[DebuggerStepThrough]
				set { this._firstChild = value; }
			}

			public TreeNode UnsafeLastChild
			{
				[DebuggerStepThrough]
				get { return this._lastChild; }
				[DebuggerStepThrough]
				set { this._lastChild = value; }
			}

			#endregion

			#endregion

			#region ITreeNode<T>

			#region Methods

			public ITreeNode<T> AppendChild(T data)
			{
				return AppendChild(data, false);
			}

			public ITreeNode<T> PrependChild(T data)
			{
				return PrependChild(data, false);
			}

			public ITreeNode<T> AppendSibling(T data)
			{
				return AppendSibling(data, false);
			}

			public ITreeNode<T> PrependSibling(T data)
			{
				return PrependSibling(data, false);
			}

			public ITreeNode<T> AppendChild(T data, bool isCollapsed)
			{
				if (this._lastChild == null)
				{
					TreeNode treeNode = new TreeNode(this._tree, this, null, null, data, true, isCollapsed);

					if (this._tree.RaiseAdding(treeNode))
					{
						this._firstChild = treeNode;
						this._lastChild = treeNode;

						ChangeTotalCount(1);
						ChangeVisibleCount(1);

						this._tree.RaiseAdded(treeNode);

						return treeNode;
					}
					else
					{
						return null;
					}
				}
				else
				{
					TreeNode treeNode = new TreeNode(this._tree, this, this._lastChild, null, data, true, isCollapsed);

					if (this._tree.RaiseAdding(treeNode))
					{
						this._lastChild.UnsafeNextSibling = treeNode;
						this._lastChild = treeNode;

						ChangeTotalCount(1);

						if (this._isVisible && (!this._isCollapsed))
						{
							ChangeVisibleCount(1);
						}

						this._tree.RaiseAdded(treeNode);

						return treeNode;
					}
					else
					{
						return null;
					}
				}
			}

			public ITreeNode<T> PrependChild(T data, bool isCollapsed)
			{
				if (this._firstChild == null)
				{
					TreeNode treeNode = new TreeNode(this._tree, this, null, null, data, true, isCollapsed);

					if (this._tree.RaiseAdding(treeNode))
					{
						this._firstChild = treeNode;
						this._lastChild = treeNode;

						ChangeTotalCount(1);

						if (this._isVisible && (!this._isCollapsed))
						{
							ChangeVisibleCount(1);
						}

						this._tree.RaiseAdded(treeNode);

						return treeNode;
					}
					else
					{
						return null;
					}
				}
				else
				{
					TreeNode treeNode = new TreeNode(this._tree, this, null, this._firstChild, data, true, isCollapsed);

					if (this._tree.RaiseAdding(treeNode))
					{
						this._firstChild.UnsafePrevSibling = treeNode;
						this._firstChild = treeNode;

						ChangeTotalCount(1);

						if (this._isVisible && (!this._isCollapsed))
						{
							ChangeVisibleCount(1);
						}

						this._tree.RaiseAdded(treeNode);

						return treeNode;
					}
					else
					{
						return null;
					}
				}
			}

			public ITreeNode<T> AppendSibling(T data, bool isCollapsed)
			{
				if (this._nextSibling == null)
				{
					TreeNode treeNode = new TreeNode(this._tree, this._parent, this, null, data, true, isCollapsed);

					if (this._tree.RaiseAdding(treeNode))
					{
						this._nextSibling = treeNode;

						if (this._parent != null)
						{
							if (this._parent.LastChild == this)
							{
								this._parent.UnsafeLastChild = treeNode;
							}

							this._parent.ChangeTotalCount(1);
							this._parent.ChangeVisibleCount(1);
						}

						this._tree.RaiseAdded(treeNode);

						return treeNode;
					}
					else
					{
						return null;
					}
				}
				else
				{
					TreeNode treeNode = new TreeNode(this._tree, this._parent, this, this._nextSibling, data, true, isCollapsed);

					if (this._tree.RaiseAdding(treeNode))
					{
						this._nextSibling.UnsafePrevSibling = treeNode;
						this._nextSibling = treeNode;

						if (this._parent != null)
						{
							if (this._parent.LastChild == this)
							{
								this._parent.UnsafeLastChild = treeNode;
							}

							this._parent.ChangeTotalCount(1);
							this._parent.ChangeVisibleCount(1);
						}

						this._tree.RaiseAdded(treeNode);

						return treeNode;
					}
					else
					{
						return null;
					}
				}
			}

			public ITreeNode<T> PrependSibling(T data, bool isCollapsed)
			{
				if (this._prevSibling == null)
				{
					TreeNode treeNode = new TreeNode(this._tree, this._parent, null, this, data, true, isCollapsed);

					if (this._tree.RaiseAdding(treeNode))
					{
						this._prevSibling = treeNode;

						if (this._parent != null)
						{
							if (this._parent.FirstChild == this)
							{
								this._parent.UnsafeFirstChild = treeNode;
							}

							this._parent.ChangeTotalCount(1);
							this._parent.ChangeVisibleCount(1);
						}

						this._tree.RaiseAdded(treeNode);

						return treeNode;
					}
					else
					{
						return null;
					}
				}
				else
				{
					TreeNode treeNode = new TreeNode(this._tree, this._parent, this._prevSibling, this, data, true, isCollapsed);

					if (this._tree.RaiseAdding(treeNode))
					{
						this._prevSibling.UnsafeNextSibling = treeNode;
						this._prevSibling = treeNode;

						if (this._parent != null)
						{
							if (this._parent.FirstChild == this)
							{
								this._parent.UnsafeFirstChild = treeNode;
							}

							this._parent.ChangeTotalCount(1);
							this._parent.ChangeVisibleCount(1);
						}

						this._tree.RaiseAdded(treeNode);

						return treeNode;
					}
					else
					{
						return null;
					}
				}
			}

			public void Remove()
			{
				if (this._tree.RaiseRemoving(this))
				{
					if (this._prevSibling != null)
					{
						this._prevSibling.UnsafeNextSibling = this._nextSibling;
					}

					if (this._nextSibling != null)
					{
						this._nextSibling.UnsafePrevSibling = this._prevSibling;
					}

					if (this._parent != null)
					{
						if (this._parent.FirstChild == this)
						{
							this._parent.UnsafeFirstChild = this.UnsafeNextSibling;
						}

						if (this._parent.LastChild == this)
						{
							this._parent.UnsafeLastChild = this.UnsafePrevSibling;
						}

						this._parent.ChangeTotalCount(-1);
						this._parent.ChangeVisibleCount(this._isVisible ? -1 : 0);
					}

					this._tree.RaiseRemoved(this);
				}
			}

			public int GetCount(bool visibleNodes)
			{
				return visibleNodes ? VisibleCount : Count;
			}

			#endregion

			#region Properties

			public ITreeNode<T> Parent
			{
				[DebuggerStepThrough]
				get { return this._parent; }
			}

			public ITreeNode<T> NextSibling
			{
				[DebuggerStepThrough]
				get { return this._nextSibling; }
			}

			public ITreeNode<T> PrevSibling
			{
				[DebuggerStepThrough]
				get { return this._prevSibling; }
			}

			public ITreeNode<T> FirstChild
			{
				[DebuggerStepThrough]
				get { return this._firstChild; }
			}

			public ITreeNode<T> LastChild
			{
				[DebuggerStepThrough]
				get { return this._lastChild; }
			}

			public IEnumerable<ITreeNode<T>> Children
			{
				get
				{
					if (this._firstChild != null)
					{
						for (
							ITreeNode<T> child = this._firstChild;
							child != this._lastChild;
							child = child.NextSibling)
						{
							yield return child;
						}

						yield return this._lastChild;
					}

					yield break;
				}
			}

			public T Data
			{
				[DebuggerStepThrough]
				get
				{
					return this._data;
				}

				set
				{
					if (this._tree.RaiseChanging(this))
					{
						INotifyPropertyChanged nodePropertyChanged = this._data as INotifyPropertyChanged;

						if (nodePropertyChanged != null)
						{
							nodePropertyChanged.PropertyChanged -= new PropertyChangedEventHandler(Data_PropertyChanged);
						}

						this._data = value;

						this._tree.RaiseChanged(this);

						nodePropertyChanged = this._data as INotifyPropertyChanged;

						if (nodePropertyChanged != null)
						{
							nodePropertyChanged.PropertyChanged += new PropertyChangedEventHandler(Data_PropertyChanged);
						}
					}
				}
			}

			public bool IsCollapsed
			{
				[DebuggerStepThrough]
				get
				{
					return this._isCollapsed;
				}

				set
				{
					if (this._isCollapsed != value)
					{
						this._isCollapsed = value;

						if (this._parent != null)
						{
							if (this._isVisible)
							{
								if (this._isCollapsed)
								{
									this._parent.ChangeVisibleCount(this._visibleCount - 1);
								}
								else
								{
									this._parent.ChangeVisibleCount(1 - this._visibleCount);
								}
							}
						}
					}
				}
			}

			public bool IsVisible
			{
				[DebuggerStepThrough]
				get
				{
					return this._isVisible;
				}

				set
				{
					if (this._isVisible != value)
					{
						this._isVisible = value;

						if (this._parent != null)
						{
							if (this._isCollapsed)
							{
								if (this._isVisible)
								{
									this._parent.ChangeVisibleCount(1);
								}
								else
								{
									this._parent.ChangeVisibleCount(-1);
								}
							}
							else
							{
								if (this._isVisible)
								{
									this._parent.ChangeVisibleCount(this._visibleCount);
								}
								else
								{
									this._parent.ChangeVisibleCount(-this._visibleCount);
								}
							}
						}
					}
				}
			}

			public int Count
			{
				[DebuggerStepThrough]
				get { return this._totalCount; }
			}

			public int VisibleCount
			{
				[DebuggerStepThrough]
				get
				{
					if (this._isVisible)
					{
						if (this._isCollapsed)
						{
							return 1;
						}
						else
						{
							return this._visibleCount;
						}
					}
					else
					{
						return 0;
					}
				}
			}

			#endregion

			#endregion
		}

		#endregion

		#region Fields

		private TreeNode _root;

		#endregion

		#region Methods

		#region Constructors

		public BindingTree()
		{
			this._root = new TreeNode(this, null, null, null, default(T), true, false);
		}

		/*
		public static BindingTree<T> FromList<X, Y>(
			IList<X> list,
			Y rootID,
			PropertyAccessor<T, X> dataAccessor,
			PropertyAccessor<Y, X> idAccessor,
			PropertyAccessor<Y, X> parentIDAccessor)
		{
			BindingTree<T> tree = new BindingTree<T>();

			return tree;
		}
		*/

		#endregion

		#region Event Raisers

		#region IBindingTree<T> Events

		private bool RaiseAdding(ITreeNode<T> node)
		{
			if (this._addingHandler != null)
			{
				BindingTreeDoingEventArgs<T> eventArgs = new BindingTreeDoingEventArgs<T>(node);

				this._addingHandler(this, eventArgs);

				return !eventArgs.Cancel;
			}
			else
			{
				return true;
			}
		}

		private void RaiseAdded(ITreeNode<T> node)
		{
			((TreeNode)node).SubscribeToPropertyChanged();

			if (this._addedHandler != null)
			{
				this._addedHandler(this, new BindingTreeDoneEventArgs<T>(node));
			}
		}

		private bool RaiseRemoving(ITreeNode<T> node)
		{
			if (this._removingHandler != null)
			{
				BindingTreeDoingEventArgs<T> eventArgs = new BindingTreeDoingEventArgs<T>(node);

				this._removingHandler(this, eventArgs);

				return !eventArgs.Cancel;
			}
			else
			{
				return true;
			}
		}

		private void RaiseRemoved(ITreeNode<T> node)
		{
			((TreeNode)node).UnsubscribeFromPropertyChanged();

			if (this._removedHandler != null)
			{
				this._removedHandler(this, new BindingTreeDoneEventArgs<T>(node));
			}
		}

		private bool RaiseChanging(ITreeNode<T> node)
		{
			bool result;

			if (this._changingHandler != null)
			{
				BindingTreeDoingEventArgs<T> eventArgs = new BindingTreeDoingEventArgs<T>(node);

				this._changingHandler(this, eventArgs);

				result = !eventArgs.Cancel;
			}
			else
			{
				result = true;
			}

			if (result)
			{
				((TreeNode)node).UnsubscribeFromPropertyChanged();
			}

			return result;
		}

		private void RaiseChanged(ITreeNode<T> node)
		{
			((TreeNode)node).SubscribeToPropertyChanged();

			if (this._changedHandler != null)
			{
				this._changedHandler(this, new BindingTreeDoneEventArgs<T>(node));
			}
		}

		#endregion

		#endregion

		#region Enumerators

		private static IEnumerable<KeyValuePair<ITreeNode<T>, int>> Preorder(
			ITreeNode<T> root,
			int level,
			bool hiddenNodes)
		{
			if (root.IsVisible || hiddenNodes)
			{
				yield return new KeyValuePair<ITreeNode<T>, int>(root, level);
			}

			foreach (ITreeNode<T> child in root.Children)
			{
				if (child.IsVisible || hiddenNodes)
				{
					foreach (KeyValuePair<ITreeNode<T>, int> node in Preorder(child, level + 1, hiddenNodes))
					{
						yield return new KeyValuePair<ITreeNode<T>, int>(node.Key, node.Value);
					}
				}
			}

			yield break;
		}

		private static IEnumerable<KeyValuePair<ITreeNode<T>, int>> Postorder(
			ITreeNode<T> root,
			int level,
			bool hiddenNodes)
		{
			foreach (ITreeNode<T> child in root.Children)
			{
				if (child.IsVisible || hiddenNodes)
				{
					foreach (KeyValuePair<ITreeNode<T>, int> node in Postorder(child, level + 1, hiddenNodes))
					{
						yield return new KeyValuePair<ITreeNode<T>, int>(node.Key, node.Value);
					}
				}
			}

			if (root.IsVisible || hiddenNodes)
			{
				yield return new KeyValuePair<ITreeNode<T>, int>(root, level);
			}

			yield break;
		}

		private static IEnumerable<KeyValuePair<ITreeNode<T>, int>> Preorder(
			ITreeNode<T> root,
			int level,
			int offset,
			int count,
			bool hiddenNodes)
		{
			if (offset + count > 0)
			{
				if (offset <= 0)
				{
					if (root.IsVisible || hiddenNodes)
					{
						yield return new KeyValuePair<ITreeNode<T>, int>(root, level);
					}
				}

				offset--;

				foreach (ITreeNode<T> child in root.Children)
				{
					if (child.IsVisible || hiddenNodes)
					{
						int childCount = child.GetCount(!hiddenNodes);

						if (childCount > offset)
						{
							foreach (KeyValuePair<ITreeNode<T>, int> node in Preorder(child, level + 1, offset, count, hiddenNodes))
							{
								yield return new KeyValuePair<ITreeNode<T>, int>(node.Key, node.Value);
							}
						}

						offset -= childCount;
					}
				}
			}

			yield break;
		}

		private static IEnumerable<KeyValuePair<ITreeNode<T>, int>> Postorder(
			ITreeNode<T> root,
			int level,
			int offset,
			int count,
			bool hiddenNodes)
		{
			if (offset + count > 0)
			{
				foreach (ITreeNode<T> child in root.Children)
				{
					if (child.IsVisible || hiddenNodes)
					{
						int childCount = child.GetCount(!hiddenNodes);

						if (childCount > offset)
						{
							foreach (KeyValuePair<ITreeNode<T>, int> node in Postorder(child, level + 1, offset, count, hiddenNodes))
							{
								yield return new KeyValuePair<ITreeNode<T>, int>(node.Key, node.Value);
							}
						}

						offset -= childCount;
					}
				}

				if ((offset <= 0) && (offset + count > 0))
				{
					if (root.IsVisible || hiddenNodes)
					{
						yield return new KeyValuePair<ITreeNode<T>, int>(root, level);
					}
				}
			}

			yield break;
		}

		#endregion

		#endregion

		#endregion

		#region IBindingTree<T>

		#region Adding

		[NonSerialized]
		private EventHandler<BindingTreeDoingEventArgs<T>> _addingHandler;

		public event EventHandler<BindingTreeDoingEventArgs<T>> Adding
		{
			add { this._addingHandler += value; }
			remove { this._addingHandler -= value; }
		}

		#endregion

		#region Added

		[NonSerialized]
		private EventHandler<BindingTreeDoneEventArgs<T>> _addedHandler;

		public event EventHandler<BindingTreeDoneEventArgs<T>> Added
		{
			add { this._addedHandler += value; }
			remove { this._addedHandler -= value; }
		}

		#endregion

		#region Removing

		[NonSerialized]
		private EventHandler<BindingTreeDoingEventArgs<T>> _removingHandler;

		public event EventHandler<BindingTreeDoingEventArgs<T>> Removing
		{
			add { this._removingHandler += value; }
			remove { this._removingHandler -= value; }
		}

		#endregion

		#region Removed

		[NonSerialized]
		private EventHandler<BindingTreeDoneEventArgs<T>> _removedHandler;

		public event EventHandler<BindingTreeDoneEventArgs<T>> Removed
		{
			add { this._removedHandler += value; }
			remove { this._removedHandler -= value; }
		}

		#endregion

		#region Changing

		[NonSerialized]
		private EventHandler<BindingTreeDoingEventArgs<T>> _changingHandler;

		public event EventHandler<BindingTreeDoingEventArgs<T>> Changing
		{
			add { this._changingHandler += value; }
			remove { this._changingHandler -= value; }
		}

		#endregion

		#region Changed

		[NonSerialized]
		private EventHandler<BindingTreeDoneEventArgs<T>> _changedHandler;

		public event EventHandler<BindingTreeDoneEventArgs<T>> Changed
		{
			add { this._changedHandler += value; }
			remove { this._changedHandler -= value; }
		}

		#endregion

		#endregion

		#region ITree<T>

		#region Methods

		public IEnumerable<KeyValuePair<ITreeNode<T>, int>> Preorder()
		{
			return Preorder(this._root, 0, true);
		}

		public IEnumerable<KeyValuePair<ITreeNode<T>, int>> Postorder()
		{
			return Postorder(this._root, 0, true);
		}

		public IEnumerable<KeyValuePair<ITreeNode<T>, int>> Preorder(bool visibleNodes)
		{
			return Preorder(this._root, 0, !visibleNodes);
		}

		public IEnumerable<KeyValuePair<ITreeNode<T>, int>> Postorder(bool visibleNodes)
		{
			return Postorder(this._root, 0, !visibleNodes);
		}

		public IEnumerable<KeyValuePair<ITreeNode<T>, int>> Preorder(int offset, int count)
		{
			return Preorder(this._root, 0, offset, count, true);
		}

		public IEnumerable<KeyValuePair<ITreeNode<T>, int>> Postorder(int offset, int count)
		{
			return Postorder(this._root, 0, offset, count, true);
		}

		public IEnumerable<KeyValuePair<ITreeNode<T>, int>> Preorder(int offset, int count, bool visibleNodes)
		{
			return Preorder(this._root, 0, offset, count, !visibleNodes);
		}

		public IEnumerable<KeyValuePair<ITreeNode<T>, int>> Postorder(int offset, int count, bool visibleNodes)
		{
			return Postorder(this._root, 0, offset, count, !visibleNodes);
		}

		public int GetCount()
		{
			return this._root.GetCount(false);
		}

		public int GetCount(bool visibleNodes)
		{
			return this._root.GetCount(visibleNodes);
		}

		#endregion

		#region Properties

		public ITreeNode<T> Root
		{
			[DebuggerStepThrough]
			get { return this._root; }
		}

		public int Count
		{
			[DebuggerStepThrough]
			get { return this._root.Count; }
		}

		#endregion

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
