﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Collections Library                      */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Diagnostics;

namespace Nabu.Collections
{
	public sealed class SieveCache<TKey, TValue> : IDictionary<TKey, TValue> where TValue : class
	{
		#region SieveCache

		#region Constants

		private const int DefaultBoundaryFrom = 1024;
		private const int DefaultBoundaryTo = 4096;

		#endregion

		#region Types

		private class CacheItem
		{
			#region Fields

			private TKey _key;
			private WeakReference _valueReferenceWeak;
			private TValue _valueReferenceStrong;
			private int _referenceCount;
			private CacheItem _prev;
			private CacheItem _next;

			#endregion

			#region Methods

			private void MoveForward()
			{
				CacheItem itemPrev = this._prev;
				CacheItem itemNext = this._next;

				itemPrev.Next = itemNext;
				itemNext.Prev = itemPrev;

				this._next = itemNext.Next;
				itemNext.Next.Prev = this;

				itemNext.Next = this;
				this._prev = itemNext;
			}

			[DebuggerStepThrough]
			public CacheItem(TKey key, TValue value, bool own, int priority, CacheItem prev, CacheItem next)
			{
				this._key = key;
				this._valueReferenceWeak = new WeakReference(value);
				this._valueReferenceStrong = own ? value : null;
				this._referenceCount = priority;
				this._prev = prev;
				this._next = next;

				if (this._prev != null)
				{
					this._prev.Next = this;
				}

				if (this._next != null)
				{
					this._next.Prev = this;
				}
			}

			[DebuggerStepThrough]
			public void AddItemReference()
			{
				this._referenceCount++;

				while (this._referenceCount > this.Next.ReferenceCount)
				{
					MoveForward();
				}
			}

			[DebuggerStepThrough]
			public void ReleaseValueReference()
			{
				this._valueReferenceStrong = null;
			}

			#endregion

			#region Properties

			public TKey Key
			{
				[DebuggerStepThrough]
				get { return this._key; }
			}

			public TValue Value
			{
				[DebuggerStepThrough]
				get { return (TValue)this._valueReferenceWeak.Target; }
			}

			public int ReferenceCount
			{
				[DebuggerStepThrough]
				get { return this._referenceCount; }
			}

			public CacheItem Prev
			{
				[DebuggerStepThrough]
				get { return this._prev; }
				[DebuggerStepThrough]
				set { this._prev = value; }
			}

			public CacheItem Next
			{
				[DebuggerStepThrough]
				get { return this._next; }
				[DebuggerStepThrough]
				set { this._next = value; }
			}

			#endregion
		}

		#endregion

		#region Fields

		private Dictionary<TKey, CacheItem> _cache = new Dictionary<TKey, CacheItem>();
		private CacheItem _itemFirst = null;
		private CacheItem _itemLast = null;
		private int _boundaryFrom = 1024;
		private int _boundaryTo = 4096;

		#endregion

		#region Methods

		[DebuggerStepThrough]
		public SieveCache()
		{
			this._boundaryFrom = DefaultBoundaryFrom;
			this._boundaryTo = DefaultBoundaryTo;
			this._cache = new Dictionary<TKey, CacheItem>(this._boundaryTo);
			this._itemFirst = new CacheItem(default(TKey), null, false, int.MinValue, null, this._itemLast);
			this._itemLast = new CacheItem(default(TKey), null, false, int.MaxValue, this._itemFirst, null);
		}

		[DebuggerStepThrough]
		public SieveCache(int boundaryFrom, int boundaryTo)
		{
			if (boundaryFrom <= 1)
			{
				throw new ArgumentOutOfRangeException("boundaryFrom", "boundaryFrom must be greater than 1.");
			}

			if (boundaryTo <= boundaryFrom)
			{
				throw new ArgumentOutOfRangeException("boundaryTo", "boundaryTo must be greater than boundaryFrom.");
			}

			this._boundaryFrom = boundaryFrom;
			this._boundaryTo = boundaryTo;
			this._cache = new Dictionary<TKey, CacheItem>(this._boundaryTo);
			this._itemFirst = new CacheItem(default(TKey), null, false, int.MinValue, null, this._itemLast);
			this._itemLast = new CacheItem(default(TKey), null, false, int.MaxValue, this._itemFirst, null);
		}

		public void Add(TKey key, TValue value, bool own)
		{
			lock (this._cache)
			{
				if (this._cache.Count >= this._boundaryTo)
				{
					Compact();
				}

				this._cache[key] = new CacheItem(key, value, own, 0, this._itemLast.Prev, this._itemLast);
			}
		}

		public bool Remove(TKey key, bool own)
		{
			lock (this._cache)
			{
				CacheItem item;

				if (!this._cache.TryGetValue(key, out item))
				{
					return false;
				}

				item.ReleaseValueReference();

				if (own)
				{
					return true;
				}
				else
				{
					return this._cache.Remove(key);
				}
			}
		}

		public void Compact()
		{
			lock (this._cache)
			{
				CacheItem item = this._itemFirst.Next;

				for (int index = 0; index < this._boundaryTo - this._boundaryFrom; index++)
				{
					this._cache.Remove(item.Key);

					item = item.Next;
				}
			}
		}

		#endregion

		#region Properties

		public int BoundaryFrom
		{
			[DebuggerStepThrough]
			get { return this._boundaryFrom; }
		}

		public int BoundaryTo
		{
			[DebuggerStepThrough]
			get { return this._boundaryTo; }
		}

		#endregion

		#endregion

		#region IDictionary<TKey, TValue>

		#region Methods

		[DebuggerStepThrough]
		public void Add(TKey key, TValue value)
		{
			Add(key, value, false);
		}

		[DebuggerStepThrough]
		public bool Remove(TKey key)
		{
			return Remove(key, false);
		}

		[DebuggerStepThrough]
		public bool TryGetValue(TKey key, out TValue value)
		{
			lock (this._cache)
			{
				value = default(TValue);

				if (key == null)
				{
					return false;
				}

				CacheItem item;

				if (!this._cache.TryGetValue(key, out item))
				{
					return false;
				}

				value = item.Value;

				if (value == null)
				{
					Remove(key);

					return false;
				}
				else
				{
					item.AddItemReference();
				}
			}

			return true;
		}

		[DebuggerStepThrough]
		public bool ContainsKey(TKey key)
		{
			return this._cache.ContainsKey(key);
		}

		#endregion

		#region Properties

		public ICollection<TKey> Keys
		{
			[DebuggerStepThrough]
			get { return (ICollection<TKey>)this._cache.Keys; }
		}

		public ICollection<TValue> Values
		{
			[DebuggerStepThrough]
			get { throw new InvalidOperationException("SieveCache does not support Values property"); }
		}

		public TValue this[TKey key]
		{
			[DebuggerStepThrough]
			get
			{
				TValue value;

				TryGetValue(key, out value);

				return value;
			}

			[DebuggerStepThrough]
			set
			{
				Add(key, value, false);
			}
		}

		#endregion

		#endregion

		#region ICollection<KeyValuePair<TKey,TValue>>

		#region Methods

		[DebuggerStepThrough]
		public void Add(KeyValuePair<TKey, TValue> item)
		{
			Add(item.Key, item.Value);
		}

		[DebuggerStepThrough]
		public bool Remove(KeyValuePair<TKey, TValue> item)
		{
			TValue value;

			if (TryGetValue(item.Key, out value))
			{
				if (value == item.Value)
				{
					return Remove(item.Key);
				}
			}

			return false;
		}

		[DebuggerStepThrough]
		public void Clear()
		{
			lock (this._cache)
			{
				this._cache.Clear();
				this._itemFirst.Next = this._itemLast;
				this._itemLast.Prev = this._itemFirst;
			}
		}

		[DebuggerStepThrough]
		public void CopyTo(KeyValuePair<TKey, TValue>[] array, int arrayIndex)
		{
			((ICollection<KeyValuePair<TKey, TValue>>)this._cache).CopyTo(array, arrayIndex);
		}

		[DebuggerStepThrough]
		public bool Contains(KeyValuePair<TKey, TValue> item)
		{
			return ((ICollection<KeyValuePair<TKey, TValue>>)this._cache).Contains(item);
		}

		#endregion

		#region Properties

		public int Count
		{
			[DebuggerStepThrough]
			get { return ((ICollection<KeyValuePair<TKey, TValue>>)this._cache).Count; }
		}

		public bool IsReadOnly
		{
			[DebuggerStepThrough]
			get { return ((ICollection<KeyValuePair<TKey, TValue>>)this._cache).IsReadOnly; }
		}

		#endregion

		#endregion

		#region IEnumerable<KeyValuePair<TKey,TValue>>

		[DebuggerStepThrough]
		public IEnumerator<KeyValuePair<TKey, TValue>> GetEnumerator()
		{
			return ((IEnumerable<KeyValuePair<TKey, TValue>>)this._cache).GetEnumerator();
		}

		#endregion

		#region IEnumerable

		[DebuggerStepThrough]
		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
		{
			return ((System.Collections.IEnumerable)this._cache).GetEnumerator();
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
