﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Collections Library                      */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;

namespace Nabu.Collections
{
	[Serializable]
	public class ListEx<T> :
		IListEx<T>,
		System.Collections.IList
	{
		#region ListEx

		#region Constants

		[SuppressMessage("Microsoft.Design", "CA1000:DoNotDeclareStaticMembersOnGenericTypes")]
		internal const string AddExceptionMessage = "Cannot add element to fixed size collection.";
		[SuppressMessage("Microsoft.Design", "CA1000:DoNotDeclareStaticMembersOnGenericTypes")]
		internal const string RemoveExceptionMessage = "Cannot remove element from fixed size collection.";
		[SuppressMessage("Microsoft.Design", "CA1000:DoNotDeclareStaticMembersOnGenericTypes")]
		internal const string ResetExceptionMessage = "Cannot reset fixed size collection.";
		[SuppressMessage("Microsoft.Design", "CA1000:DoNotDeclareStaticMembersOnGenericTypes")]
		internal const string ChangeExceptionMessage = "Cannot change element of read-only collection.";
		[SuppressMessage("Microsoft.Design", "CA1000:DoNotDeclareStaticMembersOnGenericTypes")]
		internal const string SizeChangeExceptionMessage = "ListEx size changed during enumeration.";

		#endregion

		#region Types

		private sealed class ListExEnumerator : IEnumerator<T>
		{
			#region ListExEnumerator

			#region Fields

			private ListEx<T> _list;
			private int _count;
			private int _index;

			#endregion

			#region Methods

			[DebuggerStepThrough]
			public ListExEnumerator(ListEx<T> list)
			{
				this._list = list;
				this._count = this._list.Count;
				this._index = -1;
			}

			#endregion

			#endregion

			#region IEnumerator<T>

			public T Current
			{
				[DebuggerStepThrough]
				get
				{
					if (this._list.Count != this._count)
					{
						throw new InvalidOperationException(SizeChangeExceptionMessage);
					}

					return this._list[this._index];
				}
			}

			#endregion

			#region IEnumerator

			object System.Collections.IEnumerator.Current
			{
				[DebuggerStepThrough]
				get
				{
					if (this._list.Count != this._count)
					{
						throw new InvalidOperationException(SizeChangeExceptionMessage);
					}

					return this._list[this._index];
				}
			}

			[DebuggerStepThrough]
			public bool MoveNext()
			{
				this._index++;

				return this._index < this._count;
			}

			[DebuggerStepThrough]
			public void Reset()
			{
				this._index = -1;
			}

			#endregion

			#region IDisposable

			[DebuggerStepThrough]
			public void Dispose()
			{
			}

			#endregion
		}

		private sealed class ListExComparer : IComparer<T>
		{
			#region ListExComparer

			#region Fields

			private Comparison<T> _comparison;

			#endregion

			#region Methods

			[DebuggerStepThrough]
			public ListExComparer(Comparison<T> comparison)
			{
				this._comparison = comparison;
			}

			#endregion

			#endregion

			#region IComparer<T>

			[DebuggerStepThrough]
			public int Compare(T x, T y)
			{
				return this._comparison(x, y);
			}

			#endregion
		}

		#endregion

		#region Fields

		private T[] _array;
		private int _count;
		private bool _isAddAllowed;
		private bool _isRemoveAllowed;
		private bool _isChangeAllowed;

		#endregion

		#region Methods

		[DebuggerStepThrough]
		public ListEx(int capacity)
		{
			this._array = new T[capacity];
			this._count = 0;
			this._isAddAllowed = true;
			this._isRemoveAllowed = true;
			this._isChangeAllowed = true;
		}

		[DebuggerStepThrough]
		public ListEx()
			: this(0)
		{
		}

		[DebuggerStepThrough]
		public ListEx(IEnumerable<T> collection, int capacity)
		{
			this._array = new T[capacity];
			this._count = 0;
			this._isAddAllowed = true;
			this._isRemoveAllowed = true;
			this._isChangeAllowed = true;

			if (collection != null)
			{
				foreach (T item in collection)
				{
					Add(item);
				}
			}
		}

		[DebuggerStepThrough]
		public ListEx(IEnumerable<T> collection)
			: this(collection, 256)
		{
		}

		[DebuggerStepThrough]
		public ListEx(T[] array, bool attach)
		{
			if (attach)
			{
				this._array = array;
				this._count = this._array.Length;
			}
			else
			{
				this._array = new T[array.Length];
				Array.Copy(array, this._array, this._array.Length);
				this._count = 0;
			}

			this._isAddAllowed = true;
			this._isRemoveAllowed = true;
			this._isChangeAllowed = true;
		}

		[DebuggerStepThrough]
		public ListEx(T[] array)
			: this(array, false)
		{
		}

		#endregion

		#endregion

		#region IListEx<T>

		#region Methods

		#region Capacity

		[DebuggerStepThrough]
		public void EnsureCapacity(int capacity)
		{
			if (capacity >= this._array.Length)
			{
#if COMPACT_FRAMEWORK
				T[] array = new T[Math.Max(2 * this._array.Length, capacity)];

				Array.Copy(this._array, array, this._count);
				this._array = array;
#else
				Array.Resize<T>(ref this._array, Math.Max(2 * this._array.Length, capacity));
#endif
			}
		}

		[DebuggerStepThrough]
		public void TrimExcess(float ratio)
		{
			if (this._count < this._array.Length * ratio)
			{
#if COMPACT_FRAMEWORK
				T[] array = new T[this._count];

				Array.Copy(this._array, array, this._count);
				this._array = array;
#else
				Array.Resize<T>(ref this._array, this._count);
#endif
			}
		}

		[DebuggerStepThrough]
		public void TrimExcess()
		{
			TrimExcess(0.9F);
		}

		#endregion

		#region Range

		[DebuggerStepThrough]
		public IListEx<T> GetRange(int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex + count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			T[] array = new T[count];

			Array.Copy(this._array, startIndex, array, 0, array.Length);

			return new ListEx<T>(array, true);
		}

		[DebuggerStepThrough]
		public void AddRange(IEnumerable<T> collection)
		{
			if (!this._isAddAllowed)
			{
				throw new InvalidOperationException(AddExceptionMessage);
			}

			foreach (T item in collection)
			{
				Add(item);
			}
		}

		[DebuggerStepThrough]
		public void AddRange(IListEx<T> list, int offset, int count)
		{
			if (!this._isAddAllowed)
			{
				throw new InvalidOperationException(AddExceptionMessage);
			}

			EnsureCapacity(this._count + count);

			for (int index = offset; index < offset + count; index++)
			{
				this._array[this._count] = list[index];
				this._count++;
			}
		}

		[DebuggerStepThrough]
		public void AddRange(IListEx<T> list)
		{
			AddRange(list, 0, list.Count);
		}

		[DebuggerStepThrough]
		public void AddRange(T[] array, int offset, int count)
		{
			if (!this._isAddAllowed)
			{
				throw new InvalidOperationException(AddExceptionMessage);
			}

			EnsureCapacity(this._count + count);

			for (int index = offset; index < offset + count; index++)
			{
				this._array[this._count] = array[index];
				this._count++;
			}
		}

		[DebuggerStepThrough]
		public void AddRange(T[] array)
		{
			AddRange(array, 0, array.Length);
		}

		[DebuggerStepThrough]
		public void InsertRange(int startIndex, IEnumerable<T> collection)
		{
			if (!this._isAddAllowed)
			{
				throw new InvalidOperationException(AddExceptionMessage);
			}

			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			foreach (T item in collection)
			{
				EnsureCapacity(this._count + 1);
				Array.Copy(
					this._array,
					startIndex,
					this._array,
					startIndex + 1,
					this._count - startIndex);
				this._count++;
				this._array[startIndex] = item;
				startIndex++;
			}
		}

		[DebuggerStepThrough]
		public void InsertRange(int startIndex, IListEx<T> list, int offset, int count)
		{
			if (!this._isAddAllowed)
			{
				throw new InvalidOperationException(AddExceptionMessage);
			}

			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((startIndex < 0) || (startIndex > list.Count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex + count) > list.Count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (count > 0)
			{
				EnsureCapacity(this._count + count);

				Array.Copy(
					this._array,
					startIndex,
					this._array,
					startIndex + count,
					this._count - startIndex);

				list.CopyTo(0, this._array, startIndex, count);

				this._count += count;
			}
		}

		[DebuggerStepThrough]
		public void InsertRange(int startIndex, IListEx<T> list)
		{
			InsertRange(startIndex, list, 0, list.Count);
		}

		[DebuggerStepThrough]
		public void InsertRange(int startIndex, T[] array, int offset, int count)
		{
			if (!this._isAddAllowed)
			{
				throw new InvalidOperationException(AddExceptionMessage);
			}

			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((offset < 0) || (offset > array.Length))
			{
				throw new ArgumentOutOfRangeException("offset");
			}

			if ((count < 0) || ((startIndex + count) > array.Length))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (count > 0)
			{
				EnsureCapacity(this._count + count);

				Array.Copy(
					this._array,
					startIndex,
					this._array,
					startIndex + count,
					this._count - startIndex);
				Array.Copy(array, offset, this._array, startIndex, count);

				this._count += count;
			}
		}

		[DebuggerStepThrough]
		public void InsertRange(int startIndex, T[] array)
		{
			InsertRange(startIndex, array, 0, array.Length);
		}

		[DebuggerStepThrough]
		public void RemoveRange(int startIndex, int count)
		{
			if (!this._isRemoveAllowed)
			{
				throw new InvalidOperationException(RemoveExceptionMessage);
			}

			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex + count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (count > 0)
			{
				Array.Copy(
					this._array,
					startIndex + count,
					this._array,
					startIndex,
					this._count - (startIndex + count));
				this._count -= count;

				for (
					int indexClear = this._count;
					indexClear < this._count + count;
					indexClear++)
				{
					this._array[indexClear] = default(T);
				}
			}
		}

		#endregion

		#region Copy and Convert

		[DebuggerStepThrough]
		public void FromArray(T[] array, bool attach)
		{
			if (this.IsFixedSize)
			{
				throw new InvalidOperationException(ResetExceptionMessage);
			}

			if (attach)
			{
				this._array = array;
				this._count = this._array.Length;
			}
			else
			{
				this._array = new T[array.Length];
				Array.Copy(array, this._array, this._array.Length);
				this._count = this._array.Length;
			}
		}

		[DebuggerStepThrough]
		public void FromArray(T[] array)
		{
			FromArray(array, false);
		}

		[DebuggerStepThrough]
		public T[] ToArray()
		{
			T[] array = new T[this._count];

			Array.Copy(this._array, array, array.Length);

			return array;
		}

		[DebuggerStepThrough]
		public T[] DangerousGetArray()
		{
			return this._array;
		}

		[DebuggerStepThrough]
		public ArraySegment<T> ToArraySegment()
		{
			return new ArraySegment<T>(this._array, 0, this._count);
		}

		[DebuggerStepThrough]
		public void CopyTo(T[] array)
		{
			if (array == null)
			{
				throw new ArgumentNullException("array");
			}

			if (this._count > array.Length)
			{
				throw
					new ArgumentException(
						"Array is too small.",
						"array");
			}

			Array.Copy(this._array, 0, array, 0, this._count);
		}

		[DebuggerStepThrough]
		public void CopyTo(int startIndex, T[] array, int arrayIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if (array == null)
			{
				throw new ArgumentNullException("array");
			}

			if ((arrayIndex < 0) || (arrayIndex > array.Length))
			{
				throw new ArgumentOutOfRangeException("arrayIndex");
			}

			if (
				(count < 0) ||
				((startIndex + count) > this._count) ||
				((arrayIndex + count) > array.Length))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			Array.Copy(this._array, startIndex, array, arrayIndex, count);
		}

		[DebuggerStepThrough]
		public IListEx<TOutput> ConvertAll<TOutput>(Converter<T, TOutput> converter)
		{
			TOutput[] array = new TOutput[this._count];

			for (int index = 0; index < this._count; index++)
			{
				array[index] = converter(this._array[index]);
			}

			return new ListEx<TOutput>(array, true);
		}

		#endregion

		#region Reverse

		[DebuggerStepThrough]
		public void Reverse(int startIndex, int count)
		{
			if (!this._isChangeAllowed)
			{
				throw new InvalidOperationException(ChangeExceptionMessage);
			}

			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex + count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			Array.Reverse(this._array, startIndex, count);
		}

		[DebuggerStepThrough]
		public void Reverse()
		{
			if (!this._isChangeAllowed)
			{
				throw new InvalidOperationException(ChangeExceptionMessage);
			}

			Array.Reverse(this._array, 0, this._count);
		}

		#endregion

		#region Sort

		[DebuggerStepThrough]
		public void Sort(int startIndex, int count)
		{
			if (!this._isChangeAllowed)
			{
				throw new InvalidOperationException(ChangeExceptionMessage);
			}

			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex + count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			Array.Sort<T>(this._array, startIndex, count);
		}

		[DebuggerStepThrough]
		public void Sort()
		{
			Sort(0, this._count);
		}

		[DebuggerStepThrough]
		public void Sort(IComparer<T> comparer, int startIndex, int count)
		{
			if (!this._isChangeAllowed)
			{
				throw new InvalidOperationException(ChangeExceptionMessage);
			}

			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex + count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (comparer == null)
			{
				throw new ArgumentNullException("comparer");
			}

			Array.Sort<T>(this._array, startIndex, count, comparer);
		}

		[DebuggerStepThrough]
		public void Sort(IComparer<T> comparer)
		{
			Sort(comparer, 0, this._count);
		}

		[DebuggerStepThrough]
		public void Sort(Comparison<T> comparer, int startIndex, int count)
		{
			if (!this._isChangeAllowed)
			{
				throw new InvalidOperationException(ChangeExceptionMessage);
			}

			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex + count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (comparer == null)
			{
				throw new ArgumentNullException("comparer");
			}

			Array.Sort<T>(this._array, startIndex, count, new ListExComparer(comparer));
		}

		[DebuggerStepThrough]
		public void Sort(Comparison<T> comparer)
		{
			Sort(comparer, 0, this._count);
		}

		#endregion

		#region Search

		[DebuggerStepThrough]
		public int IndexOf(T item, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex + count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			return Array.IndexOf<T>(this._array, item, startIndex, count);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, int startIndex)
		{
			return IndexOf(item, startIndex, this._count - startIndex);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, IComparer<T> comparer, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex + count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			for (int index = startIndex; index < startIndex + count; index++)
			{
				if (comparer.Compare(item, this._array[index]) == 0)
				{
					return index;
				}
			}

			return -1;
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, IComparer<T> comparer, int startIndex)
		{
			return IndexOf(item, comparer, startIndex, this._count - startIndex);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, IComparer<T> comparer)
		{
			return IndexOf(item, comparer, 0, this._count);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, Comparison<T> comparer, int startIndex, int count)
		{
			return IndexOf(item, new ListExComparer(comparer), startIndex, count);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, Comparison<T> comparer, int startIndex)
		{
			return IndexOf(item, comparer, startIndex, this._count - startIndex);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item, Comparison<T> comparer)
		{
			return IndexOf(item, comparer, 0, this._count);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex - count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			return Array.LastIndexOf<T>(this._array, item, startIndex, count);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, int startIndex)
		{
			return LastIndexOf(item, startIndex, startIndex + 1);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item)
		{
			return LastIndexOf(item, this._count - 1, this._count);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, IComparer<T> comparer, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex - count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			for (int index = startIndex; index > startIndex - count; index--)
			{
				if (comparer.Compare(item, this._array[index]) == 0)
				{
					return index;
				}
			}

			return -1;
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, IComparer<T> comparer, int startIndex)
		{
			return LastIndexOf(item, comparer, startIndex, startIndex + 1);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, IComparer<T> comparer)
		{
			return LastIndexOf(item, comparer, this._count - 1, this._count);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, Comparison<T> comparer, int startIndex, int count)
		{
			return IndexOf(item, new ListExComparer(comparer), startIndex, count);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, Comparison<T> comparer, int startIndex)
		{
			return IndexOf(item, comparer, startIndex, startIndex + 1);
		}

		[DebuggerStepThrough]
		public int LastIndexOf(T item, Comparison<T> comparer)
		{
			return IndexOf(item, comparer, 0, this._count);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex + count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			return Array.BinarySearch<T>(this._array, startIndex, count, item);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, int startIndex)
		{
			return BinarySearch(item, startIndex, this._count - startIndex);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item)
		{
			return BinarySearch(item, 0, this._count);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, IComparer<T> comparer, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if ((count < 0) || ((startIndex + count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (comparer == null)
			{
				throw new ArgumentNullException("comparer");
			}

			return Array.BinarySearch<T>(this._array, startIndex, count, item, comparer);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, IComparer<T> comparer, int startIndex)
		{
			return BinarySearch(item, comparer, startIndex, this._count - startIndex);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, IComparer<T> comparer)
		{
			return BinarySearch(item, comparer, 0, this._count);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, Comparison<T> comparer, int startIndex, int count)
		{
			return BinarySearch(item, new ListExComparer(comparer), startIndex, count);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, Comparison<T> comparer, int startIndex)
		{
			return BinarySearch(item, comparer, startIndex, this._count - startIndex);
		}

		[DebuggerStepThrough]
		public int BinarySearch(T item, Comparison<T> comparer)
		{
			return BinarySearch(item, comparer, 0, this._count);
		}

		#endregion

		#region Predicates

		#region Exists

		[DebuggerStepThrough]
		public bool Exists(Predicate<T> match)
		{
			return FindFirstIndex(match) != -1;
		}

		#endregion

		#region FindAll

		[DebuggerStepThrough]
		public IListEx<T> FindAll(Predicate<T> match, int startIndex, int count)
		{
			ListEx<T> list = new ListEx<T>();

			for (
				int index = FindFirstIndex(match, startIndex, count);
				index != -1;
				index = FindFirstIndex(match, index + 1, startIndex + count - index - 1))
			{
				list.Add(this._array[index]);
			}

			return list;
		}

		[DebuggerStepThrough]
		public IListEx<T> FindAll(Predicate<T> match, int startIndex)
		{
			return FindAll(match, startIndex, this._count - startIndex);
		}

		[DebuggerStepThrough]
		public IListEx<T> FindAll(Predicate<T> match)
		{
			return FindAll(match, 0, this._count);
		}

		[DebuggerStepThrough]
		public IListEx<int> FindAllIndex(Predicate<T> match, int startIndex, int count)
		{
			ListEx<int> list = new ListEx<int>();

			for (
				int index = FindFirstIndex(match, startIndex, count);
				index != -1;
				index = FindFirstIndex(match, index + 1, startIndex + count - index - 1))
			{
				list.Add(index);
			}

			return list;
		}

		[DebuggerStepThrough]
		public IListEx<int> FindAllIndex(Predicate<T> match, int startIndex)
		{
			return FindAllIndex(match, startIndex, this._count - startIndex);
		}

		[DebuggerStepThrough]
		public IListEx<int> FindAllIndex(Predicate<T> match)
		{
			return FindAllIndex(match, 0, this._count);
		}

		#endregion

		#region FindFirst

		[DebuggerStepThrough]
		public int FindFirstIndex(Predicate<T> match, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if (
				(count < 0) ||
				((startIndex + count) < 0) ||
				((startIndex + count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (match == null)
			{
				throw new ArgumentNullException("match");
			}

			for (int index = startIndex; index < startIndex + count; index++)
			{
				if (match(this._array[index]))
				{
					return index;
				}
			}

			return -1;
		}

		[DebuggerStepThrough]
		public int FindFirstIndex(Predicate<T> match, int startIndex)
		{
			return FindFirstIndex(match, startIndex, this._count);
		}

		[DebuggerStepThrough]
		public int FindFirstIndex(Predicate<T> match)
		{
			return FindFirstIndex(match, 0, this._count);
		}

		#endregion

		#region FindLast

		[DebuggerStepThrough]
		public int FindLastIndex(Predicate<T> match, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if (
				(count < 0) ||
				((startIndex - count) < 0) ||
				((startIndex - count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (match == null)
			{
				throw new ArgumentNullException("match");
			}

			for (int index = startIndex; index > startIndex - count; --index)
			{
				if (match(this._array[index]))
				{
					return index;
				}
			}

			return -1;
		}

		[DebuggerStepThrough]
		public int FindLastIndex(Predicate<T> match, int startIndex)
		{
			return FindLastIndex(match, startIndex, this._count);
		}

		[DebuggerStepThrough]
		public int FindLastIndex(Predicate<T> match)
		{
			return FindLastIndex(match, 0, this._count);
		}

		#endregion

		#region RemoveAll

		[DebuggerStepThrough]
		public int RemoveAll(Predicate<T> match, int startIndex, int count)
		{
			IListEx<int> list = FindAllIndex(match, startIndex, count);

			if (list.Count == 0)
			{
				return 0;
			}

			int listIndex = 1;
			int arrayIndex = list[0];

			for (int index = list[0] + 1; index < this._count; index++)
			{
				if (listIndex < list.Count)
				{
					if (index < list[listIndex])
					{
						this._array[arrayIndex] = this._array[index];
						arrayIndex++;
					}
					else
					{
						listIndex++;
					}
				}
				else
				{
					Array.Copy(this._array, index, this._array, arrayIndex, this._count - index);
					break;
				}
			}

			this._count -= list.Count;

			for (
				int indexClear = this._count;
				indexClear < this._count + list.Count;
				indexClear++)
			{
				this._array[indexClear] = default(T);
			}

			return list.Count;
		}

		[DebuggerStepThrough]
		public int RemoveAll(Predicate<T> match, int startIndex)
		{
			return RemoveAll(match, startIndex, this._count - startIndex);
		}

		[DebuggerStepThrough]
		public int RemoveAll(Predicate<T> match)
		{
			return RemoveAll(match, 0, this._count);
		}

		#endregion

		#region TrueForAll

		[DebuggerStepThrough]
		public bool TrueForAll(Predicate<T> match, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if (
				(count < 0) ||
				((startIndex + count) < 0) ||
				((startIndex + count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (match == null)
			{
				throw new ArgumentNullException("match");
			}

			for (int index = startIndex; index < startIndex + count; index++)
			{
				if (!match(this._array[index]))
				{
					return false;
				}
			}

			return true;
		}

		[DebuggerStepThrough]
		public bool TrueForAll(Predicate<T> match, int startIndex)
		{
			return TrueForAll(match, startIndex, this._count - startIndex);
		}

		[DebuggerStepThrough]
		public bool TrueForAll(Predicate<T> match)
		{
			return TrueForAll(match, 0, this._count);
		}

		#endregion

		#region FalseForAll

		[DebuggerStepThrough]
		public bool FalseForAll(Predicate<T> match, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if (
				(count < 0) ||
				((startIndex + count) < 0) ||
				((startIndex + count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			if (match == null)
			{
				throw new ArgumentNullException("match");
			}

			for (int index = startIndex; index < startIndex + count; index++)
			{
				if (match(this._array[index]))
				{
					return false;
				}
			}

			return true;
		}

		[DebuggerStepThrough]
		public bool FalseForAll(Predicate<T> match, int startIndex)
		{
			return FalseForAll(match, startIndex, this._count - startIndex);
		}

		[DebuggerStepThrough]
		public bool FalseForAll(Predicate<T> match)
		{
			return FalseForAll(match, 0, this._count);
		}

		#endregion

		#endregion

		#region Actions

		[DebuggerStepThrough]
		public void ForEach(Action<T> action, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if (
				(count < 0) ||
				((startIndex + count) < 0) ||
				((startIndex + count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			for (int index = startIndex; index < startIndex + count; index++)
			{
				action(this._array[index]);
			}
		}

		[DebuggerStepThrough]
		public void ForEach(Action<T> action, int startIndex)
		{
			ForEach(action, startIndex, this._count - startIndex);
		}

		[DebuggerStepThrough]
		public void ForEach(Action<T> action)
		{
			ForEach(action, 0, this._count);
		}

		[DebuggerStepThrough]
		public void ForEach(Action<T> action, Predicate<T> match, int startIndex, int count)
		{
			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			if (
				(count < 0) ||
				((startIndex + count) < 0) ||
				((startIndex + count) > this._count))
			{
				throw new ArgumentOutOfRangeException("count");
			}

			for (int index = startIndex; index < startIndex + count; index++)
			{
				if (match(this._array[index]))
				{
					action(this._array[index]);
				}
			}
		}

		[DebuggerStepThrough]
		public void ForEach(Action<T> action, Predicate<T> match, int startIndex)
		{
			ForEach(action, match, startIndex, this._count - startIndex);
		}

		[DebuggerStepThrough]
		public void ForEach(Action<T> action, Predicate<T> match)
		{
			ForEach(action, match, 0, this._count);
		}

		#endregion

		#endregion

		#region Properties

		public int Capacity
		{
			[DebuggerStepThrough]
			get
			{
				return this._array.Length;
			}

			[DebuggerStepThrough]
			set
			{
				if (value < this._count)
				{
					throw new ArgumentOutOfRangeException("value");
				}

				EnsureCapacity(value);
			}
		}

		public T First
		{
			[DebuggerStepThrough]
			get { return this[0]; }
		}

		public T Last
		{
			[DebuggerStepThrough]
			get { return this[this._count - 1]; }
		}

		public bool IsAddAllowed
		{
			[DebuggerStepThrough]
			get { return this._isAddAllowed; }
			[DebuggerStepThrough]
			set { this._isAddAllowed = value; }
		}

		public bool IsRemoveAllowed
		{
			[DebuggerStepThrough]
			get { return this._isRemoveAllowed; }
			[DebuggerStepThrough]
			set { this._isRemoveAllowed = value; }
		}

		public bool IsChangeAllowed
		{
			[DebuggerStepThrough]
			get { return this._isChangeAllowed; }
			[DebuggerStepThrough]
			set { this._isChangeAllowed = value; }
		}

		#endregion

		#endregion

		#region IList<T>

		#region Methods

		[DebuggerStepThrough]
		public void Insert(int startIndex, T item)
		{
			if (!this._isAddAllowed)
			{
				throw new InvalidOperationException(AddExceptionMessage);
			}

			if ((startIndex < 0) || (startIndex > this._count))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			EnsureCapacity(this._count + 1);
			Array.Copy(this._array, startIndex, this._array, startIndex + 1, this._count - startIndex);
			this._count++;
			this._array[startIndex] = item;
		}

		[DebuggerStepThrough]
		public void RemoveAt(int startIndex)
		{
			RemoveRange(startIndex, 1);
		}

		[DebuggerStepThrough]
		public int IndexOf(T item)
		{
			return IndexOf(item, 0, this._count);
		}

		#endregion

		#region Properties

		public T this[int index]
		{
			[DebuggerStepThrough]
			get
			{
				return this._array[index];
			}

			[DebuggerStepThrough]
			set
			{
				if (!this._isChangeAllowed)
				{
					throw new InvalidOperationException(ChangeExceptionMessage);
				}

				this._array[index] = value;
			}
		}

		#endregion

		#endregion

		#region ICollection<T>

		#region Methods

		[DebuggerStepThrough]
		public void Add(T item)
		{
			if (!this._isAddAllowed)
			{
				throw new InvalidOperationException(AddExceptionMessage);
			}

			EnsureCapacity(this._count + 1);
			this._array[this._count] = item;
			this._count++;
		}

		[DebuggerStepThrough]
		public bool Remove(T item)
		{
			if (!this._isRemoveAllowed)
			{
				throw new InvalidOperationException(RemoveExceptionMessage);
			}

			int index = IndexOf(item);

			if (index == -1)
			{
				return false;
			}

			RemoveAt(index);

			return true;
		}

		[DebuggerStepThrough]
		public void Clear()
		{
			if (!this._isRemoveAllowed)
			{
				throw new InvalidOperationException(RemoveExceptionMessage);
			}

			this._count = 0;
		}

		[DebuggerStepThrough]
		public bool Contains(T item)
		{
			return IndexOf(item) != -1;
		}

		[DebuggerStepThrough]
		public void CopyTo(T[] array, int arrayIndex)
		{
			Array.Copy(this._array, 0, array, arrayIndex, this._count);
		}

		#endregion

		#region Properties

		public int Count
		{
			[DebuggerStepThrough]
			get
			{
				return this._count;
			}

			[DebuggerStepThrough]
			set
			{
				if (value > this._array.Length)
				{
					throw new ArgumentOutOfRangeException("value");
				}

				this._count = value;
			}
		}

		public bool IsReadOnly
		{
			[DebuggerStepThrough]
			get
			{
				return !(
					this._isAddAllowed ||
					this._isRemoveAllowed ||
					this._isChangeAllowed);
			}

			[DebuggerStepThrough]
			set
			{
				this._isAddAllowed = !value;
				this._isRemoveAllowed = !value;
				this._isChangeAllowed = !value;
			}
		}

		#endregion

		#endregion

		#region IEnumerable<T>

		[DebuggerStepThrough]
		public IEnumerator<T> GetEnumerator()
		{
			return new ListExEnumerator(this);
		}

		#endregion

		#region IList

		#region Methods

		[DebuggerStepThrough]
		public int Add(object value)
		{
			Add((T)value);

			return this._count - 1;
		}

		[DebuggerStepThrough]
		public void Insert(int startIndex, object value)
		{
			Insert(startIndex, (T)value);
		}

		[DebuggerStepThrough]
		public void Remove(object value)
		{
			Remove((T)value);
		}

		[DebuggerStepThrough]
		public bool Contains(object value)
		{
			return Contains((T)value);
		}

		[DebuggerStepThrough]
		public int IndexOf(object value)
		{
			return IndexOf((T)value);
		}

		#endregion

		#region Properties

		public bool IsFixedSize
		{
			[DebuggerStepThrough]
			get
			{
				return !(
					this._isAddAllowed ||
					this._isRemoveAllowed);
			}

			[DebuggerStepThrough]
			set
			{
				this._isAddAllowed = !value;
				this._isRemoveAllowed = !value;
			}
		}

		object System.Collections.IList.this[int index]
		{
			[DebuggerStepThrough]
			get
			{
				return this._array[index];
			}

			[DebuggerStepThrough]
			set
			{
				if (!this._isChangeAllowed)
				{
					throw new InvalidOperationException(ChangeExceptionMessage);
				}

				this._array[index] = (T)value;
			}
		}

		#endregion

		#endregion

		#region ICollection

		[DebuggerStepThrough]
		public void CopyTo(Array array, int startIndex)
		{
			if (array == null)
			{
				throw new ArgumentNullException("array");
			}

			if (
				(startIndex < 0) ||
				(startIndex > array.Length) ||
				((startIndex + this._count) > array.Length))
			{
				throw new ArgumentOutOfRangeException("startIndex");
			}

			Array.Copy(this._array, 0, array, startIndex, this._count);
		}

		public bool IsSynchronized
		{
			[DebuggerStepThrough]
			get { return false; }
		}

		public object SyncRoot
		{
			[DebuggerStepThrough]
			get { return this._array; }
		}

		#endregion

		#region IEnumerable

		[DebuggerStepThrough]
		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
		{
			return new ListExEnumerator(this);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
