﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Data Library                             */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;

namespace Nabu.Data
{
	public abstract class DiffBlobBase : DiffBase<byte>
	{
		#region Constants

		private const int BlockCount = 2048;

		#endregion

		#region Types

#if !COMPACT_FRAMEWORK
		[System.Diagnostics.DebuggerDisplay("Offset: {Offset}; Count: {Count};")]
#endif
		protected sealed class BinaryRange
		{
			#region Fields

			private byte[] _array;
			private int _offset;
			private int _count;
			private int _hash;

			#endregion

			#region Methods

			public BinaryRange(
				byte[] array,
				int offset,
				int count,
				int hash)
			{
				this._array = array;
				this._offset = offset;
				this._count = count;
				this._hash = hash;
			}

			public bool Equals(byte[] array, int offset, int count)
			{
				if (this._count != count)
				{
					return false;
				}
				else
				{
					for (int index = 0; index < this._count; index++)
					{
						if (this._array[this._offset + index] != array[offset + index])
						{
							return false;
						}
					}

					return true;
				}
			}

			#endregion

			#region Properties

			[SuppressMessage("Microsoft.Performance", "CA1819:PropertiesShouldNotReturnArrays")]
			public byte[] Array
			{
				[DebuggerStepThrough]
				get { return this._array; }
			}

			public int Offset
			{
				[DebuggerStepThrough]
				get { return this._offset; }
			}

			public int Count
			{
				[DebuggerStepThrough]
				get { return this._count; }
			}

			public int Hash
			{
				[DebuggerStepThrough]
				get { return this._hash; }
			}

			#endregion
		}

		protected sealed class BinaryRangeTraits : IEqualityComparer<BinaryRange>
		{
			#region IEqualityComparer<BinaryRange>

			public bool Equals(BinaryRange x, BinaryRange y)
			{
				return x.Equals(y.Array, y.Offset, y.Count);
			}

			public int GetHashCode(BinaryRange obj)
			{
				return obj.Offset ^ (obj.Count << 16) ^ obj.Array.GetHashCode();
			}

			#endregion
		}

		#endregion

		#region Fields

		private BinaryRange[] _oldRanges;
		private BinaryRange[] _newRanges;
		private int _oldContentOffset;
		private int _oldContentLength;
		private int _newContentOffset;
		private int _newContentLength;
		private int _blockSize;

		#endregion

		#region Methods

		#region Ranges

		private void BuildOldRanges()
		{
			this._oldRanges = new BinaryRange[(this._oldContentLength + this._blockSize - 1) / this._blockSize];

			int lastBlock = 0;
			int lastOffset = this._oldContentOffset;
			int hashA = 0;
			int hashB = 0;

			for (int index = this._oldContentOffset; index < this._oldContentOffset + this._blockSize; index++)
			{
				hashA = (hashA + base.OldContent[index]) % 0x10000;
				hashB = (hashB + (this._oldContentOffset + this._blockSize - index) * base.OldContent[index]) % 0x10000;
			}

			for (int index = this._oldContentOffset + this._blockSize; index < this._oldContentOffset + this._oldContentLength; index++)
			{
				int hash = hashA | (hashB << 16);

				if (((index - this._oldContentOffset) % this._blockSize) == 0)
				{
					this._oldRanges[lastBlock] =
						new BinaryRange(
							base.OldContent,
							index - this._blockSize,
							this._blockSize,
							hash);
					lastBlock++;
					lastOffset = index;
				}

				hashA = (hashA - base.OldContent[index - this._blockSize] + base.OldContent[index]) % 0x10000;
				hashB = (hashB - this._blockSize * base.OldContent[index - this._blockSize] + hashA) % 0x10000;
			}

			if (lastBlock < this._oldRanges.Length)
			{
				this._oldRanges[lastBlock] =
					new BinaryRange(
						base.OldContent,
						lastOffset,
						(this._oldContentOffset + this._oldContentLength) - lastOffset,
						0);
			}
		}

		private void BuildNewRanges()
		{
			Dictionary<int, BinaryRange> rangeMap = new Dictionary<int, BinaryRange>(this._oldRanges.Length);
			bool[] rangeAvailability = new bool[65536];

			foreach (BinaryRange range in this._oldRanges)
			{
				rangeMap[range.Hash] = range;
				rangeAvailability[((range.Hash >> 0) & 0xFFFF) ^ ((range.Hash >> 16) & 0xFFFF)] = true;
			}

			List<BinaryRange> result = new List<BinaryRange>(this._oldRanges.Length);
			int hashA = 0;
			int hashB = 0;

			for (int index = this._newContentOffset; index < this._newContentOffset + this._blockSize; index++)
			{
				hashA = (hashA + base.NewContent[index]) % 0x10000;
				hashB = (hashB + (this._newContentOffset + this._blockSize - index) * base.NewContent[index]) % 0x10000;
			}

			int lastOffset = this._newContentOffset;

			for (int index = this._newContentOffset + this._blockSize; index < this._newContentOffset + this._newContentLength; index++)
			{
				if (index - lastOffset >= this._blockSize)
				{
					int hash = hashA | (hashB << 16);

					if (rangeAvailability[((hash >> 0) & 0xFFFF) ^ ((hash >> 16) & 0xFFFF)])
					{
						BinaryRange range;

						if (rangeMap.TryGetValue(hash, out range))
						{
							if (range.Equals(base.NewContent, index - this._blockSize, this._blockSize))
							{
								if (lastOffset < index - this._blockSize)
								{
									result.Add(
										new BinaryRange(
											base.NewContent,
											lastOffset,
											(index - this._blockSize) - lastOffset,
											hash));
								}

								result.Add(
									new BinaryRange(
										base.NewContent,
										index - this._blockSize,
										this._blockSize,
										hash));

								lastOffset = index;
							}
						}
					}
				}

				hashA = (hashA - base.NewContent[index - this._blockSize] + base.NewContent[index]) % 0x10000;
				hashB = (hashB - this._blockSize * base.NewContent[index - this._blockSize] + hashA) % 0x10000;
			}

			if (lastOffset < this._newContentOffset + this._newContentLength)
			{
				result.Add(
					new BinaryRange(
						base.NewContent,
						lastOffset,
						(this._newContentOffset + this._newContentLength) - lastOffset,
						0));
			}

			this._newRanges = result.ToArray();
		}

		protected void BuildRanges(
			int oldContentOffset,
			int oldContentLength,
			int newContentOffset,
			int newContentLength)
		{
			if (
				(this._oldContentOffset != oldContentOffset) ||
				(this._oldContentLength != oldContentLength) ||
				(this._newContentOffset != newContentOffset) ||
				(this._newContentLength != newContentLength))
			{
				this._oldContentOffset = oldContentOffset;
				this._oldContentLength = oldContentLength;
				this._newContentOffset = newContentOffset;
				this._newContentLength = newContentLength;
				this._blockSize =
					Math.Max(
						oldContentLength / BlockCount,
						newContentLength / BlockCount);

				if (this._blockSize > 1)
				{
					BuildOldRanges();
					BuildNewRanges();
				}
			}
		}

		#endregion

		#region Diff

		protected void CalculateRanges(
			PatchOperationList<byte> poList,
			IDiff<BinaryRange> diff,
			int oldContentOffset,
			int oldContentLength,
			int newContentOffset,
			int newContentLength)
		{
			PatchOperationList<BinaryRange> poRange = new PatchOperationList<BinaryRange>();

			diff.BuildPatch(
				poRange,
				0,
				this._oldRanges.Length,
				0,
				this._newRanges.Length);

			foreach (PatchOperation<BinaryRange> po in poRange)
			{
				switch (po.Type)
				{
					case PatchOperationType.Delete:
					case PatchOperationType.Copy:
						{
							int length = 0;

							foreach (BinaryRange range in po.Data)
							{
								length += range.Count;
							}

							poList.Add(new PatchOperation<byte>(po.Type, length, null));
						}
						break;
					case PatchOperationType.Insert:
						{
							int length = 0;

							foreach (BinaryRange range in po.Data)
							{
								length += range.Count;
							}

							byte[] data = new byte[length];

							length = 0;

							foreach (BinaryRange range in po.Data)
							{
								Array.Copy(NewContent, range.Offset, data, length, range.Count);
								length += range.Count;
							}

							poList.Add(new PatchOperation<byte>(po.Type, length, data));
						}
						break;
				}
			}
		}

		#endregion

		#region Constructors

		protected DiffBlobBase(
			byte[] oldContent,
			byte[] newContent)
			: base(oldContent, newContent)
		{

		}

		#endregion

		#endregion

		#region Properties

		[SuppressMessage("Microsoft.Performance", "CA1819:PropertiesShouldNotReturnArrays")]
		protected BinaryRange[] OldRanges
		{
			get { return this._oldRanges; }
		}

		[SuppressMessage("Microsoft.Performance", "CA1819:PropertiesShouldNotReturnArrays")]
		protected BinaryRange[] NewRanges
		{
			get { return this._newRanges; }
		}

		protected int BlockSize
		{
			get { return this._blockSize; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
