﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Data Library                             */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;

namespace Nabu.Data
{
	public class DiffMyersBinary : DiffBase<byte>
	{
		#region DiffMyers

		#region Types

		private struct Point
		{
			public int X;
			public int Y;
		}

		#endregion

		#region Fields

		private int[] _forward;
		private int[] _reverse;
		private bool[] _oldMarker;
		private bool[] _newMarker;
		private int _oldMarkerOffset;
		private int _newMarkerOffset;
		private int _contentLength;

		#endregion

		#region Methods

		private Point OptimalMiddleSnake(
			int oldContentOffset,
			int oldContentLength,
			int newContentOffset,
			int newContentLength)
		{
			int forwardK = oldContentOffset - newContentOffset;
			int reverseK = (oldContentOffset + oldContentLength) - (newContentOffset + newContentLength);
			int forwardOffset = this._contentLength - forwardK;
			int reverseOffset = this._contentLength - reverseK;
			int delta = oldContentLength - newContentLength;
			int dMax = ((oldContentLength + newContentLength + 1) / 2);
			Point result = new Point();

			Array.Clear(this._forward, 0, this._forward.Length);
			Array.Clear(this._reverse, 0, this._reverse.Length);

			this._forward[this._contentLength + 1] = oldContentOffset;
			this._reverse[this._contentLength - 1] = oldContentOffset + oldContentLength;

			for (int d = 0; d <= dMax; d++)
			{
				for (int k = forwardK - d; k <= forwardK + d; k += 2)
				{
					if (k == forwardK - d)
					{
						result.X = this._forward[forwardOffset + k + 1];
					}
					else
					{
						int x1 = this._forward[forwardOffset + k - 1] + 1;
						int x2 = this._forward[forwardOffset + k + 1];

						if ((k < forwardK + d) && (x1 <= x2))
						{
							result.X = x2;
						}
						else
						{
							result.X = x1;
						}
					}

					result.Y = result.X - k;

					while (
						(result.X < oldContentOffset + oldContentLength) &&
						(result.Y < newContentOffset + newContentLength) &&
						(base.OldContent[result.X] == base.NewContent[result.Y]))
					{
						result.X++;
						result.Y++;
					}

					this._forward[forwardOffset + k] = result.X;

					if (
						((delta & 1) != 0) &&
						(reverseK - d < k) &&
						(reverseK + d > k))
					{
						if (this._reverse[reverseOffset + k] <= this._forward[forwardOffset + k])
						{
							result.X = this._forward[forwardOffset + k];
							result.Y = result.X - k;

							return result;
						}
					}
				}

				for (int k = reverseK - d; k <= reverseK + d; k += 2)
				{
					if (k == reverseK + d)
					{
						result.X = this._reverse[reverseOffset + k - 1];
					}
					else
					{
						int x1 = this._reverse[reverseOffset + k - 1];
						int x2 = this._reverse[reverseOffset + k + 1] - 1;

						if ((k > reverseK - d) && (x1 < x2))
						{
							result.X = x1;
						}
						else
						{
							result.X = x2;
						}
					}

					result.Y = result.X - k;

					while (
						(result.X > oldContentOffset) &&
						(result.Y > newContentOffset) &&
						(base.OldContent[result.X - 1] == base.NewContent[result.Y - 1]))
					{
						result.X--;
						result.Y--;
					}

					this._reverse[reverseOffset + k] = result.X;

					if (
						((delta & 1) == 0) &&
						(forwardK - d <= k) &&
						(forwardK + d >= k))
					{
						if (this._reverse[reverseOffset + k] <= this._forward[forwardOffset + k])
						{
							result.X = this._forward[forwardOffset + k];
							result.Y = result.X - k;

							return result;
						}
					}
				}
			}

			// This will never be reached.
			throw new Exception();
		}

		private void LongestCommonSubsequence(
			int oldContentOffset,
			int oldContentLength,
			int newContentOffset,
			int newContentLength)
		{
			while (
				(oldContentLength > 0) &&
				(newContentLength > 0) &&
				(base.OldContent[oldContentOffset] == base.NewContent[newContentOffset]))
			{
				oldContentOffset++;
				oldContentLength--;
				newContentOffset++;
				newContentLength--;
			}

			while (
				(oldContentLength > 0) &&
				(newContentLength > 0) &&
				(base.OldContent[oldContentOffset + oldContentLength - 1] == base.NewContent[newContentOffset + newContentLength - 1]))
			{
				oldContentLength--;
				newContentLength--;
			}

			if (oldContentLength == 0)
			{
				for (int index = newContentOffset; index < newContentOffset + newContentLength; index++)
				{
					this._newMarker[index - this._newMarkerOffset] = true;
				}
			}
			else if (newContentLength == 0)
			{
				for (int index = oldContentOffset; index < oldContentOffset + oldContentLength; index++)
				{
					this._oldMarker[index - this._oldMarkerOffset] = true;
				}
			}
			else
			{
				Point middle =
					OptimalMiddleSnake(
						oldContentOffset,
						oldContentLength,
						newContentOffset,
						newContentLength);

				LongestCommonSubsequence(
					oldContentOffset,
					middle.X - oldContentOffset,
					newContentOffset,
					middle.Y - newContentOffset);
				LongestCommonSubsequence(
					middle.X,
					oldContentLength + oldContentOffset - middle.X,
					middle.Y,
					newContentLength + newContentOffset - middle.Y);
			}
		}

		public DiffMyersBinary(
			byte[] oldContent,
			byte[] newContent)
			: base(oldContent, newContent)
		{
			this._oldMarker = new bool[0];
			this._newMarker = new bool[0];
			this._forward = new int[0];
			this._reverse = new int[0];
		}

		#endregion

		#endregion

		#region DiffBase

		protected override void Calculate(
			PatchOperationList<byte> poList,
			int oldContentOffset,
			int oldContentLength,
			int newContentOffset,
			int newContentLength)
		{
			this._contentLength = (oldContentLength + newContentLength + 1);

			if (this._forward.Length < 2 * this._contentLength + 2)
			{
				this._forward = new int[2 * this._contentLength + 2];
			}
			else
			{
				Array.Clear(this._forward, 0, this._forward.Length);
			}

			if (this._reverse.Length < 2 * this._contentLength + 2)
			{
				this._reverse = new int[2 * this._contentLength + 2];
			}
			else
			{
				Array.Clear(this._reverse, 0, this._reverse.Length);
			}

			if (this._oldMarker.Length < oldContentLength)
			{
				this._oldMarker = new bool[oldContentLength];
			}
			else
			{
				Array.Clear(this._oldMarker, 0, this._oldMarker.Length);
			}

			if (this._newMarker.Length < newContentLength)
			{
				this._newMarker = new bool[newContentLength];
			}
			else
			{
				Array.Clear(this._newMarker, 0, this._newMarker.Length);
			}

			this._oldMarkerOffset = oldContentOffset;
			this._newMarkerOffset = newContentOffset;

			LongestCommonSubsequence(
				oldContentOffset,
				oldContentLength,
				newContentOffset,
				newContentLength);

			int oldIndex = oldContentOffset;
			int newIndex = newContentOffset;

			while (
				(oldIndex < oldContentOffset + oldContentLength) ||
				(newIndex < newContentOffset + newContentLength))
			{
				while (
					(oldIndex < oldContentOffset + oldContentLength) &&
					this._oldMarker[oldIndex - this._oldMarkerOffset])
				{
					AppendPatchOperation(poList, PatchOperationType.Delete, 0);
					oldIndex++;
				}

				while (
					(newIndex < newContentOffset + newContentLength) &&
					this._newMarker[newIndex - this._newMarkerOffset])
				{
					AppendPatchOperation(poList, PatchOperationType.Insert, base.NewContent[newIndex]);
					newIndex++;
				}

				if (
					(oldIndex < oldContentOffset + oldContentLength) &&
					(newIndex < newContentOffset + newContentLength))
				{
					AppendPatchOperation(poList, PatchOperationType.Copy, 0);
					oldIndex++;
					newIndex++;
				}
			}
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
