﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Data Library                             */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.IO;
using System.Xml;
using System.Xml.Serialization;

namespace Nabu.Data
{
	public abstract class MruListBase<T> : ICollection<T>
	{
		#region MruListBase

		#region Fields

		private List<MruItem<T>> _list;
		private string _path;

		#endregion

		#region Methods

		private void Load()
		{
			if (File.Exists(this._path))
			{
				using (Stream stream =
					new FileStream(
						this._path,
						FileMode.Open,
						FileAccess.Read,
						FileShare.Read))
				{
					XmlReaderSettings xmlReaderSettings = new XmlReaderSettings();

					xmlReaderSettings.ConformanceLevel = ConformanceLevel.Document;

					using (XmlReader xmlReader = XmlReader.Create(stream, xmlReaderSettings))
					{
						XmlSerializer xmlSerializer = new XmlSerializer(typeof(List<MruItem<T>>));

						this._list = (List<MruItem<T>>)xmlSerializer.Deserialize(xmlReader);
					}
				}
			}
			else
			{
				this._list = new List<MruItem<T>>();
			}
		}

		private void Save()
		{
			using (Stream stream =
				new FileStream(
					this._path,
					FileMode.Create,
					FileAccess.Write,
					FileShare.None))
			{
				XmlWriterSettings xmlWriterSettings = new XmlWriterSettings();

				xmlWriterSettings.ConformanceLevel = ConformanceLevel.Document;
				xmlWriterSettings.Indent = true;
				xmlWriterSettings.IndentChars = "\t";

				using (XmlWriter xmlWriter = XmlWriter.Create(stream, xmlWriterSettings))
				{
					XmlSerializer xmlSerializer = new XmlSerializer(typeof(List<MruItem<T>>));
					XmlSerializerNamespaces xmlSerializerNamespaces = new XmlSerializerNamespaces();

					xmlSerializerNamespaces.Add("", "");
					xmlSerializer.Serialize(xmlWriter, this._list, xmlSerializerNamespaces);
				}
			}
		}

		protected abstract int Compare(T x, T y);

		public MruListBase(string name)
		{
			this._path =
				string.Format(
					"{0}\\{1}.xml",
					SpecialPath.CurrentUserAllMachinesSettingsPath,
					name);
			Load();
		}

		#endregion

		#endregion

		#region ICollection<T>

		#region Methods

		public void Add(T item)
		{
			int itemIndex = -1;

			for (int index = 0; index < this._list.Count; index++)
			{
				if (Compare(this._list[index].Value, item) == 0)
				{
					itemIndex = index;
					break;
				}
			}

			if (itemIndex == -1)
			{
				this._list.Add(new MruItem<T>(item, 1));
			}
			else
			{
				this._list[itemIndex].Priority++;
			}

			this._list.Sort(
				delegate(MruItem<T> x, MruItem<T> y)
				{
					return x.Priority - y.Priority;
				});

			Save();
		}

		public bool Remove(T item)
		{
			int itemIndex = -1;

			for (int index = 0; index < this._list.Count; index++)
			{
				if (Compare(this._list[index].Value, item) == 0)
				{
					itemIndex = index;
					break;
				}
			}

			if (itemIndex != -1)
			{
				this._list.RemoveAt(itemIndex);
				Save();

				return true;
			}
			else
			{
				return false;
			}
		}

		public void Clear()
		{
			this._list.Clear();
		}

		public bool Contains(T item)
		{
			for (int index = 0; index < this._list.Count; index++)
			{
				if (Compare(this._list[index].Value, item) == 0)
				{
					return true;
				}
			}

			return false;
		}

		public void CopyTo(T[] array, int arrayIndex)
		{
			for (int index = 0; index < this._list.Count; index++)
			{
				array[arrayIndex + index] = this._list[index].Value;
			}
		}

		#endregion

		#region Properties

		public int Count
		{
			get { return this._list.Count; }
		}

		public bool IsReadOnly
		{
			get { return false; }
		}

		#endregion

		#endregion

		#region IEnumerable<T>

		public IEnumerator<T> GetEnumerator()
		{
			T[] array = new T[this._list.Count];

			CopyTo(array, 0);

			return ((IEnumerable<T>)array).GetEnumerator();
		}

		#endregion

		#region IEnumerable

		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
		{
			T[] array = new T[this._list.Count];

			CopyTo(array, 0);

			return ((System.Collections.IEnumerable)array).GetEnumerator();
		}

		#endregion
	}

	public sealed class MruStringList : MruListBase<string>
	{
		#region MruStringList

		public MruStringList(string name)
			: base(name)
		{
		}

		#endregion

		#region MruListBase

		protected override int Compare(string x, string y)
		{
			return string.Compare(x, y, StringComparison.OrdinalIgnoreCase);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
