﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Diagnostics Library                      */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Runtime.InteropServices;
using System.Security.Permissions;
using System.Text;

using Microsoft.Win32;

namespace Nabu.Diagnostics
{
	[System.Diagnostics.DebuggerDisplay("{Description}")]
	public sealed class DeviceInstance
	{
		#region DeviceInstance

		#region Fields

		private IntPtr _hInfoSet;
		private DeviceInstanceInfo _instanceInfo;
		private Dictionary<DevicePropertyID, object> _properties;
		private IList<DeviceMemoryRange> _memoryRanges;
		private IList<DevicePortRange> _portRanges;
		private IList<DeviceDmaRange> _dmaRanges;
		private IList<DeviceIrqRange> _irqRanges;
		private IList<DeviceDriver> _drivers;
		private DeviceInstanceFlags _flags;
		private DeviceInstanceProblem _problem;

		#endregion

		#region Methods

		#region Constructors

		[DebuggerStepThrough]
		internal DeviceInstance(
			IntPtr hInfoSet,
			DeviceInstanceInfo instanceInfo)
		{
			this._hInfoSet = hInfoSet;
			this._instanceInfo = instanceInfo;
			this._properties = new Dictionary<DevicePropertyID, object>();

			if (NativeMethods.CM_Get_DevNode_Status(
				out this._flags,
				out this._problem,
				this._instanceInfo.InstanceID,
				0) == ConfigResult.SUCCESS)
			{
				if ((this._flags & DeviceInstanceFlags.HAS_PROBLEM) == 0)
				{
					this._problem = DeviceInstanceProblem.NO_PROBLEM;
				}

			}
			else
			{
				this._flags = DeviceInstanceFlags.HAS_PROBLEM;
				this._problem = DeviceInstanceProblem.NO_PROBLEM;
			}
		}

		#endregion

		#region Properties

		private byte[] GetByteArrayProperty(DevicePropertyID propertyID)
		{
			object value;
			byte[] result = null;

			if (this._properties.TryGetValue(propertyID, out value))
			{
				result = (byte[])value;
			}
			else
			{
				byte[] dataValue = new byte[1024];
				RegistryValueKind dataType;
				int requiredSize;

				if (NativeMethods.SetupDiGetDeviceRegistryPropertyW(
					this._hInfoSet,
					ref this._instanceInfo,
					propertyID,
					out dataType,
					dataValue,
					dataValue.Length,
					out requiredSize))
				{
					if (dataType != RegistryValueKind.Binary)
					{
						throw new InvalidCastException();
					}

					result = new byte[requiredSize];
					Array.Copy(dataValue, result, result.Length);
				}
				else
				{
					result = new byte[0];
				}

				this._properties.Add(propertyID, result);
			}

			return result;
		}

		private int GetIntegerProperty(DevicePropertyID propertyID)
		{
			object value;
			int result;

			if (this._properties.TryGetValue(propertyID, out value))
			{
				result = (int)value;
			}
			else
			{
				int dataValue;
				RegistryValueKind dataType;
				int requiredSize;

				if (NativeMethods.SetupDiGetDeviceRegistryPropertyW(
					this._hInfoSet,
					ref this._instanceInfo,
					propertyID,
					out dataType,
					out dataValue,
					4,
					out requiredSize))
				{
					if (dataType != RegistryValueKind.DWord)
					{
						throw new InvalidCastException();
					}

					result = dataValue;
				}
				else
				{
					result = -1;
				}

				this._properties.Add(propertyID, result);
			}

			return result;
		}

		private Guid GetGuidProperty(DevicePropertyID propertyID)
		{
			object value;
			Guid result;

			if (this._properties.TryGetValue(propertyID, out value))
			{
				result = (Guid)value;
			}
			else
			{
				byte[] dataValue = new byte[16];
				RegistryValueKind dataType;
				int requiredSize;

				if (NativeMethods.SetupDiGetDeviceRegistryPropertyW(
					this._hInfoSet,
					ref this._instanceInfo,
					propertyID,
					out dataType,
					dataValue,
					dataValue.Length,
					out requiredSize))
				{
					if (dataType != RegistryValueKind.Binary)
					{
						throw new InvalidCastException();
					}

					result = new Guid(dataValue);
				}
				else
				{
					result = Guid.Empty;
				}

				this._properties.Add(propertyID, result);
			}

			return result;
		}

		private string GetStringProperty(DevicePropertyID propertyID)
		{
			object value;
			string result;

			if (this._properties.TryGetValue(propertyID, out value))
			{
				result = (string)value;
			}
			else
			{
				StringBuilder dataValue = new StringBuilder(1024);
				RegistryValueKind dataType;
				int requiredSize;

				if (NativeMethods.SetupDiGetDeviceRegistryPropertyW(
					this._hInfoSet,
					ref this._instanceInfo,
					propertyID,
					out dataType,
					dataValue,
					dataValue.Capacity,
					out requiredSize))
				{
					if (dataType != RegistryValueKind.String)
					{
						throw new InvalidCastException();
					}

					result = dataValue.ToString();
				}
				else
				{
					result = null;
				}

				this._properties.Add(propertyID, result);
			}

			return result;
		}

		private string[] GetStringArrayProperty(DevicePropertyID propertyID)
		{
			object value;
			string[] result;

			if (this._properties.TryGetValue(propertyID, out value))
			{
				result = (string[])value;
			}
			else
			{
				char[] dataValue = new char[4096];
				RegistryValueKind dataType;
				int requiredSize;

				if (NativeMethods.SetupDiGetDeviceRegistryPropertyW(
					this._hInfoSet,
					ref this._instanceInfo,
					propertyID,
					out dataType,
					dataValue,
					dataValue.Length,
					out requiredSize))
				{
					if (dataType != RegistryValueKind.MultiString)
					{
						throw new InvalidCastException();
					}

					result =
						new string(dataValue, 0, requiredSize)
						.Split(
							new char[] { '\0' },
							StringSplitOptions.RemoveEmptyEntries);
				}
				else
				{
					result = new string[0];
				}

				this._properties.Add(propertyID, result);
			}

			return result;
		}

		#endregion

		#region Resources

		private void MemoryDataParser(NativeMemory data)
		{
			DeviceMemoryDescriptor descriptor = data.ToStructure<DeviceMemoryDescriptor>();
			int descriptorSize = Marshal.SizeOf(typeof(DeviceMemoryDescriptor));
			int rangeSize = Marshal.SizeOf(typeof(DeviceMemoryRangeDescriptor));

			if (descriptor.Size == rangeSize)
			{
				if (descriptor.Count == 0)
				{
					this._memoryRanges.Add(
						new DeviceMemoryRange(
							descriptor.Min,
							descriptor.Max,
							descriptor.Flags));
				}
				else
				{
					for (int index = 0; index < descriptor.Count; index++)
					{
						DeviceMemoryRangeDescriptor range =
							data.ToStructure<DeviceMemoryRangeDescriptor>(
								descriptorSize + index * rangeSize);

						this._memoryRanges.Add(
							new DeviceMemoryRange(
								range.Min,
								range.Max,
								range.Flags));
					}
				}
			}
		}

		private void PortDataParser(NativeMemory data)
		{
			DevicePortDescriptor descriptor = data.ToStructure<DevicePortDescriptor>();
			int descriptorSize = Marshal.SizeOf(typeof(DevicePortDescriptor));
			int rangeSize = Marshal.SizeOf(typeof(DevicePortRangeDescriptor));

			if (descriptor.Size == rangeSize)
			{
				if (descriptor.Count == 0)
				{
					this._portRanges.Add(
						new DevicePortRange(
							descriptor.Min,
							descriptor.Max,
							descriptor.Flags));
				}
				else
				{
					for (int index = 0; index < descriptor.Count; index++)
					{
						DevicePortRangeDescriptor range =
							data.ToStructure<DevicePortRangeDescriptor>(
								descriptorSize + index * rangeSize);

						this._portRanges.Add(
							new DevicePortRange(
								range.Min,
								range.Max,
								range.Flags));
					}
				}
			}
		}

		private void DmaDataParser(NativeMemory data)
		{
			DeviceDmaDescriptor descriptor = data.ToStructure<DeviceDmaDescriptor>();
			int descriptorSize = Marshal.SizeOf(typeof(DeviceDmaDescriptor));
			int rangeSize = Marshal.SizeOf(typeof(DeviceDmaRangeDescriptor));

			if (descriptor.Size == rangeSize)
			{
				if (descriptor.Count == 0)
				{
					this._dmaRanges.Add(
						new DeviceDmaRange(
							0,
							0,
							descriptor.Flags,
							descriptor.Channel));
				}
				else
				{
					for (int index = 0; index < descriptor.Count; index++)
					{
						DeviceDmaRangeDescriptor range =
							data.ToStructure<DeviceDmaRangeDescriptor>(
								descriptorSize + index * rangeSize);

						this._dmaRanges.Add(
							new DeviceDmaRange(
								range.Min,
								range.Max,
								range.Flags,
								descriptor.Channel));
					}
				}
			}
		}

		private void IrqDataParser(NativeMemory data)
		{
			DeviceIrqDescriptor descriptor = data.ToStructure<DeviceIrqDescriptor>();
			int descriptorSize = Marshal.SizeOf(typeof(DeviceIrqDescriptor));
			int rangeSize = Marshal.SizeOf(typeof(DeviceIrqRangeDescriptor));

			if (descriptor.Size == rangeSize)
			{
				if (descriptor.Count == 0)
				{
					this._irqRanges.Add(
						new DeviceIrqRange(
							0,
							0,
							descriptor.Flags,
							descriptor.Number,
							descriptor.CpuAffinity.ToInt64()));
				}
				else
				{
					for (int index = 0; index < descriptor.Count; index++)
					{
						DeviceIrqRangeDescriptor range =
							data.ToStructure<DeviceIrqRangeDescriptor>(
								descriptorSize + index * rangeSize);

						this._irqRanges.Add(
							new DeviceIrqRange(
								0,
								0,
								range.Flags,
								descriptor.Number,
								descriptor.CpuAffinity.ToInt64()));
					}
				}
			}
		}

		private void EnumerateResources()
		{
			this._memoryRanges = new List<DeviceMemoryRange>();
			this._portRanges = new List<DevicePortRange>();
			this._dmaRanges = new List<DeviceDmaRange>();
			this._irqRanges = new List<DeviceIrqRange>();

			IntPtr hConfiguration;
			IntPtr hNextConfiguration;

			if (NativeMethods.CM_Get_First_Log_Conf(
				out hNextConfiguration,
				this._instanceInfo.InstanceID,
				0x00000003 /* BOOT_LOG_CONF */) == ConfigResult.SUCCESS)
			{
				do
				{
					hConfiguration = hNextConfiguration;

					IntPtr hResource = hConfiguration;
					IntPtr hNextResource;
					DeviceResourceType resourceType;

					while (NativeMethods.CM_Get_Next_Res_Des(
						out hNextResource,
						hResource,
						DeviceResourceType.All,
						out resourceType,
						0) == ConfigResult.SUCCESS)
					{
						hResource = hNextResource;

						int dataSize;

						if (NativeMethods.CM_Get_Res_Des_Data_Size(
							out dataSize,
							hResource,
							0) == ConfigResult.SUCCESS)
						{
							using (NativeMemory data = new NativeMemory(dataSize))
							{
								NativeMethods.CM_Get_Res_Des_Data(hResource, data.Pointer, data.Size, 0);

								switch (resourceType)
								{
									case DeviceResourceType.Memory:
										MemoryDataParser(data);
										break;
									case DeviceResourceType.Port:
										PortDataParser(data);
										break;
									case DeviceResourceType.DMA:
										DmaDataParser(data);
										break;
									case DeviceResourceType.IRQ:
										IrqDataParser(data);
										break;
								}
							}
						}
					}

					NativeMethods.CM_Free_Res_Des_Handle(hResource);

				} while (
					NativeMethods.CM_Get_Next_Log_Conf(
						out hNextConfiguration,
						hConfiguration,
						0) == ConfigResult.SUCCESS);

				NativeMethods.CM_Free_Log_Conf_Handle(hConfiguration);
			}
		}

		[SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")]
		public ICollection<DeviceMemoryRange> GetMemoryResources()
		{
			if (this._memoryRanges == null)
			{
				EnumerateResources();
			}

			return this._memoryRanges;
		}

		[SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")]
		public ICollection<DevicePortRange> GetPortResources()
		{
			if (this._portRanges == null)
			{
				EnumerateResources();
			}

			return this._portRanges;
		}

		[SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")]
		public ICollection<DeviceDmaRange> GetDmaResources()
		{
			if (this._dmaRanges == null)
			{
				EnumerateResources();
			}

			return this._dmaRanges;
		}

		[SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")]
		public ICollection<DeviceIrqRange> GetIrqResources()
		{
			if (this._irqRanges == null)
			{
				EnumerateResources();
			}

			return this._irqRanges;
		}

		#endregion

		#region Drivers

		[SecurityPermission(SecurityAction.LinkDemand, Flags = SecurityPermissionFlag.UnmanagedCode)]
		private void EnumerateDrivers()
		{
			this._drivers = new List<DeviceDriver>();

			DeviceInstallParameters installParameters = new DeviceInstallParameters();

			installParameters.Size = Marshal.SizeOf(typeof(DeviceInstallParameters));

			if (NativeMethods.SetupDiGetDeviceInstallParamsW(
				this._hInfoSet,
				ref this._instanceInfo,
				ref installParameters))
			{
				installParameters.FlagsEx |= DeviceInstallFlagsEx.ALLOWEXCLUDEDDRVS | DeviceInstallFlagsEx.INSTALLEDDRIVER;

				if (NativeMethods.SetupDiSetDeviceInstallParamsW(
					this._hInfoSet,
					ref this._instanceInfo,
					ref installParameters))
				{
					if (NativeMethods.SetupDiBuildDriverInfoList(
							this._hInfoSet,
							ref this._instanceInfo,
							DeviceDriverType.Class))
					{
						DeviceDriverInfo driverInfo = new DeviceDriverInfo();
						int driverIndex = 0;

						driverInfo.Size = Marshal.SizeOf(typeof(DeviceDriverInfo));

						while (NativeMethods.SetupDiEnumDriverInfoW(
							this._hInfoSet,
							ref this._instanceInfo,
							DeviceDriverType.Class,
							driverIndex,
							ref driverInfo))
						{
							this._drivers.Add(
								new DeviceDriver(
									this._hInfoSet,
									this._instanceInfo,
									driverInfo));

							driverIndex++;
							driverInfo.Size = Marshal.SizeOf(typeof(DeviceDriverInfo));
						}
					}
				}
			}
		}

		[SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")]
		public ICollection<DeviceDriver> GetDrivers()
		{
			if (this._drivers == null)
			{
				EnumerateDrivers();
			}

			return this._drivers;
		}

		#endregion

		#endregion

		#region Properties

		public DeviceInstanceFlags Flags
		{
			[DebuggerStepThrough]
			get { return this._flags; }
		}

		public DeviceInstanceProblem Problem
		{
			[DebuggerStepThrough]
			get { return this._problem; }
		}

		public IDictionary<DevicePropertyID, object> Properties
		{
			get
			{
				GetStringProperty(DevicePropertyID.Description);
				GetStringArrayProperty(DevicePropertyID.HardwareIds);
				GetStringArrayProperty(DevicePropertyID.CompatibleIds);
				GetStringProperty(DevicePropertyID.Service);
				GetStringProperty(DevicePropertyID.Class);
				GetStringProperty(DevicePropertyID.ClassGuid);
				GetStringProperty(DevicePropertyID.Driver);
				GetIntegerProperty(DevicePropertyID.ConfigFlags);
				GetStringProperty(DevicePropertyID.Manufacturer);
				GetStringProperty(DevicePropertyID.FriendlyName);
				GetStringProperty(DevicePropertyID.LocationInformation);
				GetStringProperty(DevicePropertyID.PhysicalDeviceObjectName);
				GetIntegerProperty(DevicePropertyID.Capabilities);
				GetIntegerProperty(DevicePropertyID.UINumber);
				GetStringArrayProperty(DevicePropertyID.UpperFilters);
				GetStringArrayProperty(DevicePropertyID.LowerFilters);
				GetGuidProperty(DevicePropertyID.BusTypeGuid);
				GetIntegerProperty(DevicePropertyID.LegacyBusType);
				GetIntegerProperty(DevicePropertyID.BusNumber);
				GetStringProperty(DevicePropertyID.EnumeratorName);
				GetByteArrayProperty(DevicePropertyID.Security);
				GetStringProperty(DevicePropertyID.SecuritySDS);
				GetIntegerProperty(DevicePropertyID.DeviceType);
				GetStringProperty(DevicePropertyID.Exclusive);
				GetIntegerProperty(DevicePropertyID.Characteristics);
				GetIntegerProperty(DevicePropertyID.Address);
				GetIntegerProperty(DevicePropertyID.UINumberDescFormat);
				GetByteArrayProperty(DevicePropertyID.DevicePowerData);
				GetIntegerProperty(DevicePropertyID.RemovalPolicy);
				GetIntegerProperty(DevicePropertyID.RemovalPolicyHWDefault);
				GetIntegerProperty(DevicePropertyID.RemovalPolicyOverride);
				GetIntegerProperty(DevicePropertyID.InstallState);
				GetStringArrayProperty(DevicePropertyID.LocationPaths);

				return this._properties;
			}
		}

		public string Description
		{
			[DebuggerStepThrough]
			get { return GetStringProperty(DevicePropertyID.Description); }
		}

		[SuppressMessage("Microsoft.Performance", "CA1819:PropertiesShouldNotReturnArrays")]
		public string[] HardwareIDs
		{
			[DebuggerStepThrough]
			get { return GetStringArrayProperty(DevicePropertyID.HardwareIds); }
		}

		[SuppressMessage("Microsoft.Performance", "CA1819:PropertiesShouldNotReturnArrays")]
		public string[] CompatibleIDs
		{
			[DebuggerStepThrough]
			get { return GetStringArrayProperty(DevicePropertyID.CompatibleIds); }
		}

		public string Service
		{
			[DebuggerStepThrough]
			get { return GetStringProperty(DevicePropertyID.Service); }
		}

		public string Class
		{
			[DebuggerStepThrough]
			get { return GetStringProperty(DevicePropertyID.Class); }
		}

		public string ClassGuid
		{
			[DebuggerStepThrough]
			get { return GetStringProperty(DevicePropertyID.ClassGuid); }
		}

		public string Driver
		{
			[DebuggerStepThrough]
			get { return GetStringProperty(DevicePropertyID.Driver); }
		}

		public DeviceConfigFlags ConfigFlags
		{
			[DebuggerStepThrough]
			get { return (DeviceConfigFlags)GetIntegerProperty(DevicePropertyID.ConfigFlags); }
		}

		public string Manufacturer
		{
			[DebuggerStepThrough]
			get { return GetStringProperty(DevicePropertyID.Manufacturer); }
		}

		public string FriendlyName
		{
			[DebuggerStepThrough]
			get { return GetStringProperty(DevicePropertyID.FriendlyName); }
		}

		public string LocationInformation
		{
			[DebuggerStepThrough]
			get { return GetStringProperty(DevicePropertyID.LocationInformation); }
		}

		public string PhysicalDeviceObjectName
		{
			[DebuggerStepThrough]
			get { return GetStringProperty(DevicePropertyID.PhysicalDeviceObjectName); }
		}

		public DeviceCapabilityFlags Capabilities
		{
			[DebuggerStepThrough]
			get { return (DeviceCapabilityFlags)GetIntegerProperty(DevicePropertyID.Capabilities); }
		}

		public int UiNumber
		{
			[DebuggerStepThrough]
			get { return GetIntegerProperty(DevicePropertyID.UINumber); }
		}

		[SuppressMessage("Microsoft.Performance", "CA1819:PropertiesShouldNotReturnArrays")]
		public string[] UpperFilters
		{
			[DebuggerStepThrough]
			get { return GetStringArrayProperty(DevicePropertyID.UpperFilters); }
		}

		[SuppressMessage("Microsoft.Performance", "CA1819:PropertiesShouldNotReturnArrays")]
		public string[] LowerFilters
		{
			[DebuggerStepThrough]
			get { return GetStringArrayProperty(DevicePropertyID.LowerFilters); }
		}

		public Guid BusTypeGUID
		{
			[DebuggerStepThrough]
			get { return GetGuidProperty(DevicePropertyID.BusTypeGuid); }
		}

		public DeviceInterfaceType LegacyBusType
		{
			[DebuggerStepThrough]
			get { return (DeviceInterfaceType)GetIntegerProperty(DevicePropertyID.LegacyBusType); }
		}

		public int BusNumber
		{
			[DebuggerStepThrough]
			get { return GetIntegerProperty(DevicePropertyID.BusNumber); }
		}

		public string EnumeratorName
		{
			[DebuggerStepThrough]
			get { return GetStringProperty(DevicePropertyID.EnumeratorName); }
		}

		[SuppressMessage("Microsoft.Performance", "CA1819:PropertiesShouldNotReturnArrays")]
		public byte[] Security
		{
			[DebuggerStepThrough]
			get { return GetByteArrayProperty(DevicePropertyID.Security); }
		}

		public string SecuritySDS
		{
			[DebuggerStepThrough]
			get { return GetStringProperty(DevicePropertyID.SecuritySDS); }
		}

		public FileDeviceType DeviceType
		{
			[DebuggerStepThrough]
			get { return (FileDeviceType)GetIntegerProperty(DevicePropertyID.DeviceType); }
		}

		public string Exclusive
		{
			[DebuggerStepThrough]
			get { return GetStringProperty(DevicePropertyID.Exclusive); }
		}

		public string Characteristics
		{
			[DebuggerStepThrough]
			get { return GetStringProperty(DevicePropertyID.Characteristics); }
		}

		public int Address
		{
			[DebuggerStepThrough]
			get { return GetIntegerProperty(DevicePropertyID.Address); }
		}

		public int UiNumberDescFormat
		{
			[DebuggerStepThrough]
			get { return GetIntegerProperty(DevicePropertyID.UINumberDescFormat); }
		}

		[SuppressMessage("Microsoft.Performance", "CA1819:PropertiesShouldNotReturnArrays")]
		public byte[] DevicePowerData
		{
			[DebuggerStepThrough]
			get { return GetByteArrayProperty(DevicePropertyID.DevicePowerData); }
		}

		public DeviceRemovalPolicy RemovalPolicy
		{
			[DebuggerStepThrough]
			get { return (DeviceRemovalPolicy)GetIntegerProperty(DevicePropertyID.RemovalPolicy); }
		}

		public DeviceRemovalPolicy RemovalPolicyHwDefault
		{
			[DebuggerStepThrough]
			get { return (DeviceRemovalPolicy)GetIntegerProperty(DevicePropertyID.RemovalPolicyHWDefault); }
		}

		public DeviceRemovalPolicy RemovalPolicyOverride
		{
			[DebuggerStepThrough]
			get { return (DeviceRemovalPolicy)GetIntegerProperty(DevicePropertyID.RemovalPolicyOverride); }
		}

		public DeviceInstallState InstallState
		{
			[DebuggerStepThrough]
			get { return (DeviceInstallState)GetIntegerProperty(DevicePropertyID.InstallState); }
		}

		[SuppressMessage("Microsoft.Performance", "CA1819:PropertiesShouldNotReturnArrays")]
		public string[] LocationPaths
		{
			[DebuggerStepThrough]
			get { return GetStringArrayProperty(DevicePropertyID.LocationPaths); }
		}

		#endregion

		#endregion

		#region Object

		public override string ToString()
		{
			return Description;
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
