﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu Library                                  */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics.CodeAnalysis;
using System.Globalization;
using System.Security.Permissions;
using System.Text;

namespace Nabu.Diagnostics
{
	[CLSCompliant(false)]
	public sealed class FusionWrapper
	{
		#region Constants

		public static readonly Guid RefCountUninstallSubkey = new Guid("8cedc215-ac4b-488b-93c0-a50a49cb2fb8");
		public static readonly Guid RefCountFilePath = new Guid("b02f9d65-fb77-4f7a-afa5-b391309f11c9");
		public static readonly Guid RefCountOpaqueString = new Guid("2ec93463-b0c3-45e1-8364-327e96aea856");
		public static readonly Guid RefCountMsi = new Guid("25df0fc1-7f97-4070-add7-4b13bbfd7cb8");

		#endregion

		#region Methods

		#region Constructors

		private FusionWrapper()
		{
		}

		#endregion

		#region Wrappers

		public static IAssemblyCache CreateAssemblyCache()
		{
			IAssemblyCache assemblyCache;

			NativeMethods.CreateAssemblyCache(out assemblyCache, 0);

			return assemblyCache;
		}

		public static IAssemblyName CreateAssemblyName(string name)
		{
			if (name == null)
			{
				throw new ArgumentNullException("name");
			}

			IAssemblyName assemblyName;

			NativeMethods.CreateAssemblyNameObject(out assemblyName, name, 2, (IntPtr)0);

			return assemblyName;
		}

		public static string GetDisplayName(IAssemblyName assemblyName, AssemblyDisplayModes namePart)
		{
			if (assemblyName == null)
			{
				throw new ArgumentNullException("assemblyName");
			}

			int bufferSize = 512;
			StringBuilder buffer = new StringBuilder((int)bufferSize);

			NativeResult.ThrowOnComError(assemblyName.GetDisplayName(buffer, ref bufferSize, namePart));

			return buffer.ToString();
		}

		public static String GetName(IAssemblyName assemblyName)
		{
			if (assemblyName == null)
			{
				throw new ArgumentNullException("assemblyName");
			}

			int bufferSize = 512;
			StringBuilder buffer = new StringBuilder((int)bufferSize);

			NativeResult.ThrowOnComError(assemblyName.GetName(ref bufferSize, buffer));

			return buffer.ToString();
		}

		public static Version GetVersion(IAssemblyName assemblyName)
		{
			if (assemblyName == null)
			{
				throw new ArgumentNullException("assemblyName");
			}

			uint major;
			uint minor;

			NativeResult.ThrowOnComError(assemblyName.GetVersion(out major, out minor));

			return new Version((int)(major >> 16), (int)(major & 0xFFFF), (int)(minor >> 16), (int)(minor & 0xFFFF));
		}

		[SecurityPermission(SecurityAction.LinkDemand, Flags = SecurityPermissionFlag.UnmanagedCode)]
		public static byte[] GetPublicKeyToken(IAssemblyName assemblyName)
		{
			if (assemblyName == null)
			{
				throw new ArgumentNullException("assemblyName");
			}

			int bufferSize = 8;

			using (NativeMemory buffer = new NativeMemory(bufferSize))
			{
				NativeResult.ThrowOnComError(
					assemblyName.GetProperty(
						AssemblyNamePart.PublicKeyToken,
						buffer.Pointer,
						ref bufferSize));

				byte[] publicKeyToken = new byte[bufferSize];

				buffer.CopyTo(publicKeyToken, 0, bufferSize);

				return publicKeyToken;
			}
		}

		[SecurityPermission(SecurityAction.LinkDemand, Flags = SecurityPermissionFlag.UnmanagedCode)]
		public static byte[] GetPublicKey(IAssemblyName assemblyName)
		{
			if (assemblyName == null)
			{
				throw new ArgumentNullException("assemblyName");
			}

			int bufferSize = 512;

			using (NativeMemory buffer = new NativeMemory(bufferSize))
			{
				NativeResult.ThrowOnComError(
					assemblyName.GetProperty(
						AssemblyNamePart.PublicKey,
						buffer.Pointer,
						ref bufferSize));

				byte[] publicKey = new byte[bufferSize];

				buffer.CopyTo(publicKey, 0, bufferSize);

				return publicKey;
			}
		}

		[SecurityPermission(SecurityAction.LinkDemand, Flags = SecurityPermissionFlag.UnmanagedCode)]
		public static CultureInfo GetCulture(IAssemblyName assemblyName)
		{
			if (assemblyName == null)
			{
				throw new ArgumentNullException("assemblyName");
			}

			int bufferSize = 512;

			using (NativeMemory buffer = new NativeMemory(bufferSize))
			{
				NativeResult.ThrowOnComError(
					assemblyName.GetProperty(
						AssemblyNamePart.Culture,
						buffer.Pointer,
						ref bufferSize));

				return new CultureInfo(buffer.ToUnicodeString());
			}
		}

		public static IAssemblyEnum CreateGacEnum()
		{
			IAssemblyEnum assemblyEnum;

			NativeMethods.CreateAssemblyEnum(out assemblyEnum, (IntPtr)0, null, AssemblyCacheModes.Gac, (IntPtr)0);

			return assemblyEnum;
		}

		[SuppressMessage("Microsoft.Design", "CA1021", MessageId = "1#")]
		public static ComResult GetNextAssembly(IAssemblyEnum assemblyEnum, out IAssemblyName assemblyName)
		{
			if (assemblyEnum == null)
			{
				throw new ArgumentNullException("assemblyEnum");
			}

			return assemblyEnum.GetNextAssembly((IntPtr)0, out assemblyName, 0);
		}

		#endregion

		#region Properties

		public static string GacPath
		{
			get
			{
				int bufferSize = 512;
				StringBuilder buffer = new StringBuilder(bufferSize);

				NativeMethods.GetCachePath(AssemblyCacheModes.Gac, buffer, ref bufferSize);

				return buffer.ToString();
			}
		}

		public static string ZapPath
		{
			get
			{
				int bufferSize = 512;
				StringBuilder buffer = new StringBuilder(bufferSize);

				NativeMethods.GetCachePath(AssemblyCacheModes.Zap, buffer, ref bufferSize);

				return buffer.ToString();
			}
		}

		public static string DownloadPath
		{
			get
			{
				int bufferSize = 512;
				StringBuilder buffer = new StringBuilder(bufferSize);

				NativeMethods.GetCachePath(AssemblyCacheModes.Download, buffer, ref bufferSize);

				return buffer.ToString();
			}
		}

		#endregion

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
