﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Diagnostics Library                      */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics.CodeAnalysis;
using System.Runtime.InteropServices;
using System.Security.Permissions;

namespace Nabu.Diagnostics
{
	[CLSCompliant(false)]
	public sealed class ScheduledTask : DisposableBase
	{
		#region ScheduledTask

		#region Fields

		private ITask _task;
		private string _name;

		#endregion

		#region Methods

		internal ScheduledTask(ITask task, string name)
		{
			this._task = task;
			this._name = name;
		}

		public void Run()
		{
			base.CheckDisposed();

			this._task.Run();
		}

		public void Terminate()
		{
			base.CheckDisposed();

			this._task.Terminate();
		}

		#endregion

		#region Properties

		#region Name

		public string Name
		{
			get
			{
				base.CheckDisposed();

				return this._name;
			}
		}

		#endregion

		#region Next Run Time

		public DateTime NextRunTime
		{
			get
			{
				base.CheckDisposed();

				SystemTime time;

				this._task.GetNextRunTime(out time);

				return new DateTime(time.Year, time.Month, time.Day, time.Hour, time.Minute, time.Second);
			}
		}

		#endregion

		#region Last Run Time

		public DateTime LastRunTime
		{
			get
			{
				base.CheckDisposed();

				SystemTime time;

				this._task.GetMostRecentRunTime(out time);

				return new DateTime(time.Year, time.Month, time.Day, time.Hour, time.Minute, time.Second);
			}
		}

		#endregion

		#region Status

		public ComResult Status
		{
			get
			{
				base.CheckDisposed();

				ComResult status;

				this._task.GetStatus(out status);

				return status;
			}
		}

		#endregion

		#region Exit Code

		public int ExitCode
		{
			get
			{
				base.CheckDisposed();

				int exitCode;

				this._task.GetExitCode(out exitCode);

				return exitCode;
			}
		}

		#endregion

		#region Comment

		public string Comment
		{
			get
			{
				base.CheckDisposed();

				IntPtr lpComment;

				this._task.GetComment(out lpComment);

				return CoTaskMemory.PtrToStringAndFree(lpComment);
			}

			set
			{
				base.CheckDisposed();

				this._task.SetComment(value);
			}
		}

		#endregion

		#region Creator

		public string Creator
		{
			get
			{
				base.CheckDisposed();

				IntPtr lpCreator;

				this._task.GetCreator(out lpCreator);

				return CoTaskMemory.PtrToStringAndFree(lpCreator);
			}

			set
			{
				base.CheckDisposed();

				this._task.SetCreator(value);
			}
		}

		#endregion

		#region Word Item Data

		[SuppressMessage("Microsoft.Performance", "CA1819:PropertiesShouldNotReturnArrays")]
		public byte[] WorkItemData
		{
			[SecurityPermission(SecurityAction.LinkDemand, Flags = SecurityPermissionFlag.UnmanagedCode)]
			get
			{
				base.CheckDisposed();

				ushort dataLength;
				IntPtr lpData;

				this._task.GetWorkItemData(out dataLength, out lpData);

				if (lpData == IntPtr.Zero)
				{
					return null;
				}
				else
				{
					byte[] data = new byte[dataLength];

					Marshal.Copy(lpData, data, 0, dataLength);
					Marshal.FreeCoTaskMem(lpData);

					return data;
				}
			}

			set
			{
				base.CheckDisposed();

				this._task.SetWorkItemData((ushort)value.Length, value);
			}
		}

		#endregion

		#region Error Retry Count

		public ushort ErrorRetryCount
		{
			get
			{
				base.CheckDisposed();

				ushort errorRetryCount;

				this._task.GetErrorRetryCount(out errorRetryCount);

				return errorRetryCount;
			}

			set
			{
				base.CheckDisposed();

				this._task.SetErrorRetryCount(value);
			}
		}

		#endregion

		#region Error Retry Interval

		public ushort ErrorRetryInterval
		{
			get
			{
				base.CheckDisposed();

				ushort errorRetryInterval;

				this._task.GetErrorRetryInterval(out errorRetryInterval);

				return errorRetryInterval;
			}

			set
			{
				base.CheckDisposed();

				this._task.SetErrorRetryInterval(value);
			}
		}

		#endregion

		#region Flags

		public TaskFlags Flags
		{
			get
			{
				base.CheckDisposed();

				TaskFlags flags;

				this._task.GetFlags(out flags);

				return flags;
			}

			set
			{
				base.CheckDisposed();

				this._task.SetFlags(value);
			}
		}

		#endregion

		#region Application Name

		public string ApplicationName
		{
			get
			{
				base.CheckDisposed();

				IntPtr lpApplicationName;

				this._task.GetApplicationName(out lpApplicationName);

				return CoTaskMemory.PtrToStringAndFree(lpApplicationName);
			}

			set
			{
				base.CheckDisposed();

				this._task.SetApplicationName(value);
			}
		}

		#endregion

		#region Parameters

		public string Parameters
		{
			get
			{
				base.CheckDisposed();

				IntPtr lpParameters;

				this._task.GetParameters(out lpParameters);

				return CoTaskMemory.PtrToStringAndFree(lpParameters);
			}

			set
			{
				base.CheckDisposed();

				this._task.SetParameters(value);
			}
		}

		#endregion

		#region Working Directory

		public string WorkingDirectory
		{
			get
			{
				base.CheckDisposed();

				IntPtr lpWorkingDirectory;

				this._task.GetWorkingDirectory(out lpWorkingDirectory);

				return CoTaskMemory.PtrToStringAndFree(lpWorkingDirectory);
			}

			set
			{
				base.CheckDisposed();

				this._task.SetWorkingDirectory(value);
			}
		}

		#endregion

		#region Priority

		public int Priority
		{
			get
			{
				base.CheckDisposed();

				int prioriry;

				this._task.GetPriority(out prioriry);

				return prioriry;
			}

			set
			{
				base.CheckDisposed();

				this._task.SetPriority(value);
			}
		}

		#endregion

		#region Task Flags

		public uint TaskFlags
		{
			get
			{
				base.CheckDisposed();

				uint taskFlags;

				this._task.GetTaskFlags(out taskFlags);

				return taskFlags;
			}

			set
			{
				base.CheckDisposed();

				this._task.SetTaskFlags(value);
			}
		}

		#endregion

		#region Maximum Run Time

		public uint MaxRunTime
		{
			get
			{
				base.CheckDisposed();

				uint maxRunTime;

				this._task.GetMaxRunTime(out maxRunTime);

				return maxRunTime;
			}

			set
			{
				base.CheckDisposed();

				this._task.SetMaxRunTime(value);
			}
		}

		#endregion

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			Marshal.ReleaseComObject(this._task);
			this._task = null;

			base.Dispose(disposing);
		}

		#endregion
	}
}
/******************************************************************************/
/*                              END OF FILE                                   */
/******************************************************************************/
