﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Drawing Library                          */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;

namespace Nabu.Drawing
{
	public sealed class ColorSpaceConverter
	{
		#region Constants

		private const float OneSixth = 1.0F / 6.0F;
		private const float OneThird = 1.0F / 3.0F;
		private const float TwoThirds = 2.0F / 3.0F;
		private const float FiveSixths = 1.0F / 6.0F;

		#endregion

		#region Methods

		#region Constructors

		private ColorSpaceConverter()
		{
		}

		#endregion

		#region Utility

		private static void MinMax3(out float max, out float min, float a, float b, float c)
		{
			if (a > b)
			{
				if (a > c)
				{
					if (b > c)
					{
						// a b c
						max = a;
						min = c;
					}
					else
					{
						// a c b
						max = a;
						min = b;
					}
				}
				else
				{
					// c a b
					max = c;
					min = b;
				}
			}
			else
			{
				if (a > c)
				{
					// b a c
					max = b;
					min = c;
				}
				else
				{
					if (b > c)
					{
						// b c a
						max = b;
						min = a;
					}
					else
					{
						// c b a
						max = c;
						min = a;
					}
				}
			}
		}

		private static float RGBFromHue(float v1, float v2, float hue)
		{
			if (hue < 0.0F)
			{
				hue += 1.0F;
			}
			else if (hue > 1.0F)
			{
				hue -= 1.0F;
			}

			if (hue < OneSixth)
			{
				return v1 + 6.0F * (v2 - v1) * hue;
			}
			else if (hue < 0.5F)
			{
				return v2;
			}
			else if (hue < TwoThirds)
			{
				return v1 + 6.0F * (v2 - v1) * (TwoThirds - hue);
			}
			else
			{
				return v1;
			}
		}

		#endregion

		#region HSL

		/*
		public static ColorHSL HSLFromHSV(ColorHSV colorHSV)
		{
			return new ColorHSL();
		}
		*/

		public static ColorHSL HSLFromRGB(ColorRGB colorRGB)
		{
			float max;
			float min;

			MinMax3(out max, out min, colorRGB.R, colorRGB.G, colorRGB.B);

			float sum = max + min;
			float diff = max - min;

			float h;
			float s;
			float l = sum / 2.0F;

			if (max == min)
			{
				h = 0.0F;
				s = 0.0F;
			}
			else
			{
				if (max == colorRGB.R)
				{
					h = (colorRGB.G - colorRGB.B) / (6.0F * diff);
				}
				else if (max == colorRGB.G)
				{
					h = (colorRGB.B - colorRGB.R) / (6.0F * diff) + OneThird;
				}
				else
				{
					h = (colorRGB.R - colorRGB.G) / (6.0F * diff) + TwoThirds;
				}

				if (h < 0.0F)
				{
					h += 1.0F;
				}
				else if (h > 1.0F)
				{
					h -= 1.0F;
				}

				if (l <= 0.5F)
				{
					s = diff / sum;
				}
				else
				{
					s = diff / (2.0F - sum);
				}
			}

			return new ColorHSL(
				h,
				Math.Min(Math.Max(s, 0.0F), 1.0F),
				Math.Min(Math.Max(l, 0.0F), 1.0F));
		}

		/*
		public static ColorHSL HSLFromYUV(ColorYUV colorYUV)
		{
			return new ColorHSL();
		}
		*/

		#endregion

		#region HSV

		/*
		public static ColorHSV HSVFromHSL(ColorHSL colorHSL)
		{
			return new ColorHSV();
		}
		*/

		public static ColorHSV HSVFromRGB(ColorRGB colorRGB)
		{
			float max;
			float min;

			MinMax3(out max, out min, colorRGB.R, colorRGB.G, colorRGB.B);

			float diff = max - min;

			float h;
			float s;
			float v = max;

			if (max == min)
			{
				h = 0.0F;
			}
			else
			{
				if (max == colorRGB.R)
				{
					h = (colorRGB.G - colorRGB.B) / (6.0F * diff);
				}
				else if (max == colorRGB.G)
				{
					h = (colorRGB.B - colorRGB.R) / (6.0F * diff) + OneThird;
				}
				else
				{
					h = (colorRGB.R - colorRGB.G) / (6.0F * diff) + TwoThirds;
				}

				if (h < 0.0F)
				{
					h += 1.0F;
				}
				else if (h > 1.0F)
				{
					h -= 1.0F;
				}
			}

			if (max == 0.0F)
			{
				s = 0.0F;
			}
			else
			{
				s = diff / max;
			}

			return new ColorHSV(h, s, v);
		}

		/*
		public static ColorHSV HSVFromYUV(ColorYUV colorYUV)
		{
			return new ColorHSV();
		}
		*/

		#endregion

		#region RGB

		public static ColorRGB RGBFromHSL(ColorHSL colorHSL)
		{
			float r;
			float g;
			float b;

			if (colorHSL.S == 0.0F)
			{
				r = colorHSL.L;
				g = colorHSL.L;
				b = colorHSL.L;
			}
			else
			{
				float v2;

				if (colorHSL.L < 0.5F)
				{
					v2 = colorHSL.L * (1.0F + colorHSL.S);
				}
				else
				{
					v2 = (colorHSL.L + colorHSL.S) - (colorHSL.L * colorHSL.S);
				}

				float v1 = 2.0F * colorHSL.L - v2;

				r = RGBFromHue(v1, v2, colorHSL.H + OneThird);
				g = RGBFromHue(v1, v2, colorHSL.H);
				b = RGBFromHue(v1, v2, colorHSL.H - OneThird);
			}

			return new ColorRGB(
				Math.Min(Math.Max(r, 0.0F), 1.0F),
				Math.Min(Math.Max(g, 0.0F), 1.0F),
				Math.Min(Math.Max(b, 0.0F), 1.0F));
		}

		public static ColorRGB RGBFromHSV(ColorHSV colorHSV)
		{
			float r;
			float g;
			float b;

			if (colorHSV.S == 0.0F)
			{
				r = colorHSV.V;
				g = colorHSV.V;
				b = colorHSV.V;
			}
			else
			{
				float h = (int)(6.0F * colorHSV.H);
				float f = 6.0F * colorHSV.H - h;
				float p = colorHSV.V * (1.0F - colorHSV.S);
				float q = colorHSV.V * (1.0F - f * colorHSV.S);
				float t = colorHSV.V * (1.0F - (1.0F - f) * colorHSV.S);

				if (h < 1.0F)
				{
					r = colorHSV.V;
					g = t;
					b = p;
				}
				else if (h < 2.0F)
				{
					r = q;
					g = colorHSV.V;
					b = p;
				}
				else if (h < 3.0F)
				{
					r = p;
					g = colorHSV.V;
					b = t;
				}
				else if (h < 4.0F)
				{
					r = p;
					g = q;
					b = colorHSV.V;
				}
				else if (h < 5.0F)
				{
					r = t;
					g = p;
					b = colorHSV.V;
				}
				else
				{
					r = colorHSV.V;
					g = p;
					b = q;
				}
			}

			return new ColorRGB(
				Math.Min(Math.Max(r, 0.0F), 1.0F),
				Math.Min(Math.Max(g, 0.0F), 1.0F),
				Math.Min(Math.Max(b, 0.0F), 1.0F));
		}

		public static ColorRGB RGBFromYUV(ColorYUV colorYUV)
		{
			return new ColorRGB(
				Math.Min(Math.Max(colorYUV.Y - 0.000039457070707082786F * colorYUV.U + 1.1398279671717171F * colorYUV.V, 0.0F), 1.0F),
				Math.Min(Math.Max(colorYUV.Y - 0.39461016414141414F * colorYUV.U - 0.58050031565656568F * colorYUV.V, 0.0F), 1.0F),
				Math.Min(Math.Max(colorYUV.Y + 2.0319996843434343F * colorYUV.U - 0.00048137626262625125F * colorYUV.V, 0.0F), 1.0F));
		}

		#endregion

		#region YUV

		/*
		public static ColorYUV YUVFromHSL(ColorHSL colorHSL)
		{
			return new ColorYUV();
		}
		*/

		/*
		public static ColorYUV YUVFromHSV(ColorHSV colorHSV)
		{
			return new ColorYUV();
		}
		*/

		public static ColorYUV YUVFromRGB(ColorRGB colorRGB)
		{
			return new ColorYUV(
				Math.Min(Math.Max(0.299F * colorRGB.R + 0.587F * colorRGB.G + 0.114F * colorRGB.B, 0.000F), 1.000F),
				Math.Min(Math.Max(-0.147F * colorRGB.R - 0.289F * colorRGB.G + 0.436F * colorRGB.B, -0.436F), 0.436F),
				Math.Min(Math.Max(0.615F * colorRGB.R - 0.515F * colorRGB.G - 0.100F * colorRGB.B, -0.615F), 0.615F));
		}

		#endregion

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
