﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Drawing;

using Nabu.Collections;

namespace Nabu.Forms.Html
{
	[HtmlBehaviorName("docking-windows")]
	public sealed class HtmlDockingWindowsBehavior : HtmlBehaviorBase
	{
		#region HtmlDockingWindowsBehavior

		#region Constants

		private const HtmlTagArea ViewBorderBox = HtmlTagArea.ViewRelative | HtmlTagArea.BorderBox;
		private const string DragMarkerHtml =
@"<div.dock-drag>
	<div></div>
</div>";
		private const string DockMarkerCenterHtml =
@"<div>
	<table.dock-center cellspacing='0'>
		<tr>
			<td></td><td></td><td></td><td></td><td></td>
		</tr>
		<tr>
			<td></td><td></td><td><div#center-top .dock-target></td><td></td><td></td>
		</tr>
		<tr>
			<td></td><td><div#center-left .dock-target></td><td><div#center-merge .dock-target></td><td><div#center-right .dock-target></td><td></td>
		</tr>
		<tr>
			<td></td><td></td><td><div#center-bottom .dock-target></td><td></td><td></td>
		</tr>
		<tr>
			<td></td><td></td><td></td><td></td><td></td>
		</tr>
	</table>
	</div>";
		private const string DockMarkerEdgeHtml =
@"<div>
	<table.dock-edge cellspacing='0'>
	<tr>
		<td></td><td></td><td><div#edge-top .dock-target></td><td></td><td></td>
	</tr>
	<tr>
		<td></td><td></td><td></td><td></td><td></td>
	</tr>
	<tr>
		<td><div#edge-left .dock-target></td><td></td><td></td><td></td><td><div#edge-right .dock-target></td>
	</tr>
	<tr>
		<td></td><td></td><td></td><td></td><td></td>
	</tr>
	<tr>
		<td></td>
		<td></td>
		<td><div#edge-bottom .dock-target></td>
		<td></td>
		<td></td>
	</tr>
	</table>
</div>";

		#endregion

		#region Types

		private enum DragType
		{
			None,
			Single,
			Multiple
		}

		private abstract class HtmlTagGroup : DisposableBase
		{
			#region HtmlTagGroup

			#region Fields

			private DisposableList<HtmlTag> _tags;
			private bool _isTemporary;

			#endregion

			#region Methods

			protected HtmlTagGroup(
				DisposableList<HtmlTag> tags,
				bool isTemporary)
			{
				this._tags = tags;
				this._isTemporary = isTemporary;
			}

			protected HtmlTag GetTagFromPoint(HtmlTagArea area, Point point)
			{
				HtmlTag result = null;
				Rectangle resultBox = Rectangle.Empty;

				foreach (HtmlTag tag in this._tags)
				{
					Rectangle tagBox = tag.GetLocation(area);

					if (tagBox.Contains(point))
					{
						if (!resultBox.Contains(tagBox))
						{
							result = tag;
							resultBox = tagBox;
						}
					}
				}

				return result;
			}

			#endregion

			#region Properties

			public IList<HtmlTag> Tags
			{
				get { return this._tags; }
			}

			#endregion

			#endregion

			#region DisposableBase

			protected override void Dispose(bool disposing)
			{
				if (disposing)
				{
					if (this._isTemporary)
					{
						foreach (HtmlTag tag in this._tags)
						{
							if (!tag.IsDisposed)
							{
								tag.Delete();
							}
						}
					}

					this._tags.Dispose();
					this._tags = null;
				}

				base.Dispose(disposing);
			}

			#endregion
		}

		private sealed class DockContainerGroup : HtmlTagGroup
		{
			#region DockContainerGroup

			public DockContainerGroup(HtmlTag root)
				: base(root.Select("div.dock-container"), false)
			{
			}

			public HtmlTag GetTagFromPoint(Point point)
			{
				return GetTagFromPoint(ViewBorderBox, point);
			}

			public Rectangle GetBoxFromPoint(Point point)
			{
				using (HtmlTag container = GetTagFromPoint(ViewBorderBox, point))
				{
					if (container == null)
					{
						return Rectangle.Empty;
					}
					else
					{
						return container.GetLocation(ViewBorderBox);
					}
				}
			}

			#endregion
		}

		private sealed class DockMarkerGroup : HtmlTagGroup
		{
			#region DockMarkerGroup

			private static DisposableList<HtmlTag> CreateDockMarkers(HtmlTag root)
			{
				DisposableList<HtmlTag> dockMarkers = new DisposableList<HtmlTag>();

				dockMarkers.Add(HtmlTag.FromHtml(root, DockMarkerCenterHtml));
				dockMarkers.Add(HtmlTag.FromHtml(root, DockMarkerEdgeHtml));

				using (HtmlTag body = root.SelectFirst("body"))
				{
					foreach (HtmlTag dockMarker in dockMarkers)
					{
						body.Children.Add(dockMarker);
					}
				}

				return dockMarkers;
			}

			public DockMarkerGroup(HtmlTag root)
				: base(CreateDockMarkers(root), true)
			{
				base.Tags[0].SetStyle("visibility", "collapse");
			}

			public void MoveCenter(Rectangle position)
			{
				if (position.IsEmpty)
				{
					base.Tags[0].SetStyle("visibility", "collapse");
				}
				else
				{
					base.Tags[0].SetStyle("visibility", "visible");
					base.Tags[0].Move(position);
				}
			}

			#endregion
		}

		private sealed class DockTargetGroup : HtmlTagGroup
		{
			#region DockTargetGroup

			public DockTargetGroup(HtmlTag root)
				: base(root.Select("div.dock-target"), true)
			{
			}

			public HtmlTag GetTagFromPoint(Point point)
			{
				return GetTagFromPoint(ViewBorderBox, point);
			}

			public string GetIDFromPoint(Point point)
			{
				using (HtmlTag target = GetTagFromPoint(ViewBorderBox, point))
				{
					if (target == null)
					{
						return null;
					}
					else
					{
						return target.Attributes["id"];
					}
				}
			}

			#endregion
		}

		#endregion

		#region Fields

		private DockContainerGroup _dockContainers;
		private DockMarkerGroup _dockMarkers;
		private DockTargetGroup _dockTargets;
		private HtmlTag _dragContent;
		private HtmlTag _dragMarker;

		private Rectangle _startPosition;
		private Point _startPoint;
		private DragType _dragType;

		private IntPtr _hWindow;

		#endregion

		#region Methods

		#region Utility

		#region Drag Marker

		private static HtmlTag CreateDragMarker(HtmlTag root)
		{
			using (HtmlTag body = root.SelectFirst("body"))
			{
				HtmlTag marker = HtmlTag.FromHtml(root, DragMarkerHtml);

				body.Children.Add(marker);

				marker.SetStyle("visibility", "collapse");

				return marker;
			}
		}

		#endregion

		#region Content

		private void MoveSimpleContent(
			HtmlTag dockContainerOld,
			HtmlTag dockContainerNew,
			HtmlTag dockContainerRoot,
			HtmlTag dragContent,
			string dockTargetID)
		{
			switch (dockTargetID)
			{
				case "center-top":
				case "center-left":
				case "center-bottom":
				case "center-right":
					if (dockContainerNew != null)
					{
						DetachSimpleContent(dockContainerOld, dragContent);
						AttachSimpleContent(dockContainerNew, dragContent, dockTargetID.Substring(7));
					}
					break;
				case "center-merge":
					if (
						(dockContainerNew != null) &&
						(dockContainerNew != dockContainerOld))
					{
						DetachSimpleContent(dockContainerOld, dragContent);
						AttachSimpleContent(dockContainerNew, dragContent, dockTargetID.Substring(7));
					}
					break;
				case "edge-top":
				case "edge-left":
				case "edge-bottom":
				case "edge-right":
					DetachSimpleContent(dockContainerOld, dragContent);
					AttachSimpleContent(dockContainerRoot, dragContent, dockTargetID.Substring(5));
					break;
			}
		}

		private void MoveComplexContent(
			HtmlTag dockContainerOld,
			HtmlTag dockContainerNew,
			HtmlTag dockContainerRoot,
			HtmlTag dragContent,
			string dockTargetID)
		{
			switch (dockTargetID)
			{
				case "center-top":
				case "center-left":
				case "center-bottom":
				case "center-right":
					break;
				case "center-merge":
					break;
				case "edge-top":
				case "edge-left":
				case "edge-bottom":
				case "edge-right":
					break;
			}
		}

		private void DetachSimpleContent(HtmlTag dockContainer, HtmlTag dockContent)
		{
			string dockContentID = dockContent.Attributes["tab-panel-id"];

			using (HtmlTag tabPanelStrip = dockContent.Parent)
			{
				using (HtmlTag tabHeaderStrip = dockContainer.SelectLast(".tab-header-strip"))
				{
					int dockContentSiblingCount = tabPanelStrip.Children.Count - 1;
					int dockContentIndex = dockContent.Index;

					tabPanelStrip.Children.RemoveAt(dockContentIndex);
					tabHeaderStrip.Children[dockContentIndex].Delete();

					if (dockContentSiblingCount == 0)
					{
						using (HtmlTag container = dockContainer.Parent)
						{
							switch (container.Attributes["class"])
							{
								case "dock-container":
									{
										int dockContainerIndex = dockContainer.Index;

										if (dockContainerIndex == 0)
										{
											container.Children[2].Parent = container.Parent;
											container.Delete();
										}
										else if (dockContainerIndex == 2)
										{
											container.Children[0].Parent = container.Parent;
											container.Delete();
										}
									}
									break;
								case "dock-root":
									dockContainer.Delete();
									break;
							}
						}
					}
					else
					{
						if (dockContent.Index == 0)
						{
							using (HtmlButton button = new HtmlButton(tabHeaderStrip.Children[dockContentIndex]))
							{
								button.PerformClick();
							}
						}

						if (dockContentSiblingCount == 1)
						{
							tabHeaderStrip.SetStyle("visibility", "collapse");
						}
					}
				}
			}
		}

		private void AttachSimpleContent(HtmlTag dockContainer, HtmlTag dockContent, string targetID)
		{
			switch (targetID)
			{
				case "top":
				case "left":
				case "bottom":
				case "right":
					break;
				case "merge":
					using (HtmlTag tabPanelStrip = dockContainer.SelectFirst(".tab-panel-strip"))
					{
						using (HtmlTag tabHeaderStrip = dockContainer.SelectFirst(".tab-header-strip"))
						{
							using (HtmlTag dockContentTitle = dockContent.SelectFirst(".dock-content-title"))
							{
								string dockContainerID = dockContainer.Attributes["tab-id"];

								tabPanelStrip.Children.Add(dockContent);
								dockContent.Attributes["tab-id"] = dockContainerID;

								HtmlTag dockContentHeader = HtmlTag.Create("div");

								tabHeaderStrip.Children.Insert(
									tabHeaderStrip.Children.Count - 1,
									dockContentHeader);

								dockContentHeader.Attributes["class"] = "tab-header";
								dockContentHeader.Attributes["tab-id"] = dockContainerID;
								dockContentHeader.Attributes["tab-header-id"] = dockContent.Attributes["tab-panel-id"];
								dockContentHeader.InnerHtml = dockContentTitle.InnerHtml;

								using (HtmlButton button = new HtmlButton(dockContentHeader))
								{
									button.PerformClick();
								}

								if (tabPanelStrip.Children.Count > 1)
								{
									tabHeaderStrip.SetStyle("visibility", "visible");
								}
							}
						}
					}
					break;
			}
		}

		#endregion

		#region Dragging

		private void StartDragging(HtmlTag sender, HtmlMouseEventArgs eventArgs)
		{
			this._dragType = DragType.None;
			this._dragContent = null;

			if (eventArgs.MouseState == HtmlEventMouseState.Left)
			{
				HtmlTag dockContentTitle = eventArgs.Target.SelectParent(".dock-content-title");

				if (dockContentTitle != null)
				{
					HtmlTag dockContent = dockContentTitle.SelectParent(".dock-content");

					if (dockContent != null)
					{
						if (dockContent.Parent.Children.Count == 1)
						{
							this._dragType = DragType.Single;
							this._dragContent = dockContent;
						}
						else
						{
							this._dragType = DragType.Multiple;
							this._dragContent = dockContent.SelectParent(".dock-container");
						}
					}
				}
				else
				{
					HtmlTag dockContentHeader =
						eventArgs.Target.SelectParent(
							".dock-container > .tab-header-strip > .tab-header");

					if (dockContentHeader != null)
					{
						HtmlTag dockContentHeaderStrip =
							dockContentHeader.SelectParent(
								".dock-container");
						HtmlTag dockContent =
							dockContentHeaderStrip.SelectFirst(
								string.Format(
									".dock-content[tab-panel-id='{0}']",
									dockContentHeader.Attributes["tab-header-id"]));
						this._dragType = DragType.Single;
						this._dragContent = dockContent;
					}
				}
			}

			if (this._dragType != DragType.None)
			{
				using (HtmlTag root = sender.Root)
				{
					this._dragMarker = CreateDragMarker(root);
					this._dockMarkers = new DockMarkerGroup(root);
					this._dockTargets = new DockTargetGroup(root);
					this._dockContainers = new DockContainerGroup(root);
				}

				this._startPoint = eventArgs.PositionInDocument;
				this._startPosition = this._dragContent.GetLocation(ViewBorderBox);

				DoDragging(sender, eventArgs);

				eventArgs.IsHandled = true;
				sender.SetCapture();
			}
		}

		private void StopDragging(HtmlTag sender, HtmlMouseEventArgs eventArgs)
		{
			if (eventArgs.MouseState == HtmlEventMouseState.Left)
			{
				if (this._dragType != DragType.None)
				{
					string dockTargetID =
							this._dockTargets.GetIDFromPoint(
								eventArgs.PositionInDocument);

					if (dockTargetID != null)
					{
						if (this._dragType == DragType.Single)
						{
							using (HtmlTag dockContainerOld =
								this._dragContent.SelectParent(".dock-container"))
							{
								using (HtmlTag dockContainerNew =
									this._dockContainers.GetTagFromPoint(eventArgs.PositionInDocument))
								{
									using (HtmlTag dockContainerRoot = sender.SelectFirst(".dock-root"))
									{
										MoveSimpleContent(
											dockContainerOld,
											dockContainerNew,
											dockContainerRoot,
											this._dragContent,
											dockTargetID);
									}
								}
							}
						}
						else if (this._dragType == DragType.Multiple)
						{
							using (HtmlTag dockContainerNew =
								this._dockContainers.GetTagFromPoint(eventArgs.PositionInDocument))
							{
								using (HtmlTag dockContainerRoot = sender.SelectFirst(".dock-root"))
								{
									if (
										(this._dragContent != dockContainerNew) &&
										(this._dragContent != dockContainerRoot))
									{
										foreach (HtmlTag dockContent in this._dragContent.SelectFirst(".tab-panel-strip").Children)
										{
											MoveComplexContent(
												this._dragContent,
												dockContainerNew,
												dockContainerRoot,
												dockContent,
												dockTargetID);
										}
									}
								}
							}
						}
					}

					HtmlTag.ReleaseCapture();
					this._dockContainers.Dispose();
					this._dockContainers = null;
					this._dockTargets.Dispose();
					this._dockTargets = null;
					this._dockMarkers.Dispose();
					this._dockMarkers = null;
					this._dragMarker.Delete();
					this._dragMarker = null;
					this._dragContent = null;
					this._dragType = DragType.None;
					eventArgs.IsHandled = true;
				}
			}
		}

		private void DoDragging(HtmlTag sender, HtmlMouseEventArgs eventArgs)
		{
			if (this._dragType != DragType.None)
			{
				Rectangle dockContainerBox =
					this._dockContainers.GetBoxFromPoint(
						eventArgs.PositionInDocument);
				Rectangle viewBox =
					sender.GetLocation(
						ViewBorderBox);

				this._dockMarkers.MoveCenter(dockContainerBox);

				HtmlTag dockTarget =
					this._dockTargets.GetTagFromPoint(
						eventArgs.PositionInDocument);

				foreach (HtmlTag tag in this._dockTargets.Tags)
				{
					tag.SetState(HtmlTagState.Current, false, true);
				}

				Rectangle position = Rectangle.Empty;

				if (dockTarget == null)
				{
					position = this._startPosition;

					position.Offset(
						eventArgs.PositionInDocument.X - this._startPoint.X,
						eventArgs.PositionInDocument.Y - this._startPoint.Y);
				}
				else
				{
					dockTarget.SetState(HtmlTagState.Current, true, true);

					switch (dockTarget.Attributes["id"])
					{
						case "center-top":
							position =
								new Rectangle(
									dockContainerBox.X,
									dockContainerBox.Y,
									dockContainerBox.Width,
									dockContainerBox.Height / 2);
							break;
						case "center-left":
							position =
								new Rectangle(
									dockContainerBox.X,
									dockContainerBox.Y,
									dockContainerBox.Width / 2,
									dockContainerBox.Height);
							break;
						case "center-bottom":
							position =
								new Rectangle(
									dockContainerBox.X,
									dockContainerBox.Y + dockContainerBox.Height / 2,
									dockContainerBox.Width,
									dockContainerBox.Height / 2);
							break;
						case "center-right":
							position =
								new Rectangle(
									dockContainerBox.X + dockContainerBox.Width / 2,
									dockContainerBox.Y,
									dockContainerBox.Width / 2,
									dockContainerBox.Height);
							break;
						case "center-merge":
							position = dockContainerBox;
							break;
						case "edge-top":
							position =
								new Rectangle(
									viewBox.X,
									viewBox.Y,
									viewBox.Width,
									viewBox.Height / 4);
							break;
						case "edge-left":
							position =
								new Rectangle(
									viewBox.X,
									viewBox.Y,
									viewBox.Width / 4,
									viewBox.Height);
							break;
						case "edge-bottom":
							position =
								new Rectangle(
									viewBox.X,
									viewBox.Y + viewBox.Height * 3 / 4,
									viewBox.Width,
									viewBox.Height / 4);
							break;
						case "edge-right":
							position =
								new Rectangle(
									viewBox.X + viewBox.Width * 3 / 4,
									viewBox.Y,
									viewBox.Width / 4,
									viewBox.Height);
							break;
						default:

							break;
					}
				}

				this._dragMarker.Move(position);
				this._dragMarker.SetStyle("visibility", "visible");

				eventArgs.IsHandled = true;
			}
		}

		#endregion

		#endregion

		#region Constructors

		public HtmlDockingWindowsBehavior()
			: base(HtmlEventGroup.InitializationAndAllInput)
		{
		}

		#endregion

		#endregion

		#endregion

		#region HtmlBehaviorBase

		protected override void OnInitialization(HtmlTag sender, HtmlInitializationEventArgs e)
		{
			switch (e.Command)
			{
				case HtmlInitializationEventCommand.AttachBehavior:
					this._hWindow = sender.GetWindow(false);
					e.IsHandled = true;
					break;
			}

			base.OnInitialization(sender, e);
		}

		protected override void OnMouse(HtmlTag sender, HtmlMouseEventArgs e)
		{
			switch (e.Command)
			{
				case HtmlMouseEventCommand.Down:
					StartDragging(sender, e);
					break;
				case HtmlMouseEventCommand.Up:
					StopDragging(sender, e);
					break;
				case HtmlMouseEventCommand.Move:
					DoDragging(sender, e);
					break;
				case HtmlMouseEventCommand.Wheel:
					if (this._dragType != DragType.None)
					{
						e.IsHandled = true;
					}
					break;
				default:
					break;
			}

			base.OnMouse(sender, e);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
