﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Globalization;

using Nabu.Collections;

namespace Nabu.Forms.Html
{
	[HtmlBehaviorName("tab")]
	public sealed class HtmlTabBehavior : HtmlBehaviorBase
	{
		#region HtmlTabBehavior

		#region Constants

		private const string TabAttribute = "tab-id";
		private const string TabHeaderAttribute = "tab-header-id";
		private const string TabPanelAttribute = "tab-panel-id";

		#endregion

		#region Methods

		#region Utility

		private static HtmlTag GetTabHeader(HtmlTag tabStrip, HtmlTag tabHeader)
		{
			if (tabHeader == tabStrip)
			{
				return null;
			}

			if (tabHeader.Attributes.ContainsKey(TabHeaderAttribute))
			{
				return tabHeader;
			}
			else
			{
				return GetTabHeader(tabStrip, tabHeader.Parent);
			}
		}

		private static bool SetActivePage(
			HtmlTag tabStrip,
			HtmlTag tabHeader,
			bool forceUpdate)
		{
			string tabID = tabHeader.Attributes[TabAttribute];
			string pageID = tabHeader.Attributes[TabHeaderAttribute];

			using (HtmlTag tabHeaderOld =
				tabStrip.SelectFirst(
					string.Format(
						CultureInfo.InvariantCulture,
						"[" + TabAttribute + "='{0}'][" + TabHeaderAttribute + "][selected='true']",
						tabID)))
			{
				if ((tabHeaderOld != tabHeader) || forceUpdate)
				{
					tabStrip.Update(true);

					if (tabHeaderOld != null)
					{
						tabHeaderOld.SetState(HtmlTagState.Current, false, true);
						tabHeaderOld.Attributes["selected"] = null;
					}

					tabHeader.SetState(HtmlTagState.Current, true, true);
					tabHeader.Attributes["selected"] = "true";

					foreach (HtmlTag tabPanel in
						tabStrip.Select(
							string.Format(
								CultureInfo.InvariantCulture,
								"[" + TabAttribute + "='{0}'][" + TabPanelAttribute + "]",
								tabID)))
					{
						if (tabPanel.Attributes[TabPanelAttribute] != pageID)
						{
							tabPanel.SetState(HtmlTagState.Expanded, false, true);
							tabPanel.SetState(HtmlTagState.Disabled, true, true);
						}
						else
						{
							tabPanel.SetState(HtmlTagState.Expanded, true, true);
							tabPanel.SetState(HtmlTagState.Disabled, false, true);
						}
					}

					tabHeader.PostEvent(
						(int)HtmlBehaviorEventCommand.ButtonClick,
						tabHeader,
						IntPtr.Zero);

					return true;
				}
			}

			return false;
		}

		#endregion

		#region Constructors

		public HtmlTabBehavior()
			: base(
				HtmlEventGroup.InitializationAndMouse |
				HtmlEventGroup.InitializationAndKeyboard |
				HtmlEventGroup.InitializationAndBehavior)
		{
		}

		#endregion

		#endregion

		#endregion

		#region HtmlBehaviorBase

		protected override void OnInitialization(HtmlTag sender, HtmlInitializationEventArgs e)
		{
			switch (e.Command)
			{
				case HtmlInitializationEventCommand.AttachBehavior:
					{
						foreach (HtmlTag tabHeaderSelected in sender.Select("[" + TabHeaderAttribute + "][selected='true']"))
						{
							SetActivePage(sender, tabHeaderSelected, true);
						}
					}
					break;
			}

			base.OnInitialization(sender, e);
		}

		protected override void OnMouse(HtmlTag sender, HtmlMouseEventArgs e)
		{
			switch (e.Command)
			{
				case HtmlMouseEventCommand.Down:
					using (HtmlTag target = e.Target)
					{
						using (HtmlTag tabHeader = GetTabHeader(sender, target))
						{
							if (tabHeader != null)
							{
								e.IsHandled = SetActivePage(sender, tabHeader, false);
							}
						}
					}
					break;
			}

			base.OnMouse(sender, e);
		}

		protected override void OnKeyboard(HtmlTag sender, HtmlKeyboardEventArgs e)
		{
			switch (e.Command)
			{
				case HtmlKeyboardEventCommand.Down:
					switch (e.KeyCode)
					{
						case 9: // Tab
							break;
						case 37: // Left
							{
								string tabID = e.Target.Attributes[TabAttribute];

								using (DisposableList<HtmlTag> tabHeaders =
									sender.Select(
										string.Format(
											CultureInfo.InvariantCulture,
											"[" + TabAttribute + "='{0}'][" + TabHeaderAttribute + "]",
											tabID)))
								{
									for (int index = 0; index < tabHeaders.Count; index++)
									{
										if (tabHeaders[index].Attributes["selected"] == "true")
										{
											e.IsHandled =
												SetActivePage(
													sender,
													tabHeaders[(index + tabHeaders.Count - 1) % tabHeaders.Count],
													false);
											break;
										}
									}
								}
							}
							break;
						case 39: // Right
							{
								string tabID = e.Target.Attributes[TabAttribute];

								using (DisposableList<HtmlTag> tabHeaders =
									sender.Select(
										string.Format(
											CultureInfo.InvariantCulture,
											"[" + TabAttribute + "='{0}'][" + TabHeaderAttribute + "]",
											tabID)))
								{
									for (int index = 0; index < tabHeaders.Count; index++)
									{
										if (tabHeaders[index].Attributes["selected"] == "true")
										{
											e.IsHandled =
												SetActivePage(
													sender,
													tabHeaders[(index + 1) % tabHeaders.Count],
													false);
											break;
										}
									}
								}
							}
							break;
					}
					break;
			}

			base.OnKeyboard(sender, e);
		}

		protected override void OnBehavior(HtmlTag sender, HtmlBehaviorEventArgs e)
		{
			switch (e.Command)
			{
				case HtmlBehaviorEventCommand.CustomActivateChild:
					using (HtmlTag target = e.Target)
					{
						SetActivePage(sender, target, false);
						e.IsHandled = true;
					}
					break;
			}

			base.OnBehavior(sender, e);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
