﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Text;

namespace Nabu.Forms.Html
{
	public sealed class HtmlTemplateParser
	{
		#region Types

		private enum ParserState
		{
			#region Whitespace

			// |<tag>
			Whitespace_Before_OpenTagStart,

			// |</tag>
			Whitespace_Before_CloseTagStart,

			// <| tag>
			Whitespace_Before_OpenTagName,

			// </| tag>
			Whitespace_Before_CloseTagName,

			// <tag| > <tag| key=value>
			Whitespace_Before_OpenTagEnd_Or_AttributeName,

			// </tag| >
			Whitespace_Before_CloseTagEnd,

			// <tag key| =value>
			Whitespace_Before_AttributeEqual,

			// <tag key=| value>
			Whitespace_Before_AttributeValue,

			#endregion

			#region Tag

			// <|tag>
			OpenTag_Start1,

			// <|/tag>
			CloseTag_Start1,

			// </|tag>
			CloseTag_Start2,

			// <tag|>
			OpenTag_End,

			// </tag|>
			CloseTag_End,

			// < |tag>
			OpenTag_Name,

			// </ |tag>
			CloseTag_Name,

			#endregion

			#region Attribute

			// <tag |key=value>
			Attribute_Name,

			// <tag key|=value>
			Attribute_Equal,

			// <tag key=|'value'>
			Attribute_Value_StartSingleQuotes1,

			// <tag key='|value'>
			Attribute_Value_StartSingleQuotes2,

			// <tag key=|"value">
			Attribute_Value_StartDoubleQuotes1,

			// <tag key="|value">
			Attribute_Value_StartDoubleQuotes2,

			// <tag key='value|'>
			Attribute_Value_EndSingleQuotes,

			// <tag key="value|">
			Attribute_Value_EndDoubleQuotes,

			// <tag key=|value>
			Attribute_Value_NoQuotes,

			// <tag key='v|alue'>
			Attribute_Value_SingleQuotes,

			// <tag key="v|alue">
			Attribute_Value_DoubleQuotes,

			#endregion
		}

		#endregion

		#region Fields

		private IDictionary<string, HtmlTemplate> _templates;
		private List<string> _templateValue;
		private StringBuilder _templateValueLine;
		private StringBuilder _openTagName;
		private StringBuilder _closeTagName;
		private StringBuilder _closeTagText;
		private IDictionary<string, string> _attributes;
		private StringBuilder _attributeName;
		private StringBuilder _attributeValue;
		private ParserState _state;

		#endregion

		#region Methods

		private void AppendTemplateSymbol(char symbol)
		{
			switch (symbol)
			{
				case '\r':
					break;
				case '\n':
					this._templateValue.Add(this._templateValueLine.ToString());
					this._templateValueLine.Length = 0;
					break;
				default:
					this._templateValueLine.Append(symbol);
					break;
			}
		}

		private void CancelCloseTag()
		{
			foreach (char sym in this._closeTagText.ToString())
			{
				AppendTemplateSymbol(sym);
			}

			this._state = ParserState.Whitespace_Before_CloseTagStart;
		}

		private void ProcessSymbol(char symbol)
		{
			switch (this._state)
			{
				#region Whitespace_Before_OpenTagStart

				case ParserState.Whitespace_Before_OpenTagStart:
					switch (symbol)
					{
						case '<':
							this._state = ParserState.OpenTag_Start1;
							break;
						default:
							if (!char.IsWhiteSpace(symbol))
							{
								throw new ArgumentOutOfRangeException("symbol");
							}
							break;
					}
					break;

				#endregion

				#region Whitespace_Before_CloseTagStart

				case ParserState.Whitespace_Before_CloseTagStart:
					switch (symbol)
					{
						case '<':
							this._closeTagText.Length = 0;
							this._closeTagText.Append(symbol);
							this._state = ParserState.CloseTag_Start1;
							break;
						default:
							AppendTemplateSymbol(symbol);
							break;
					}
					break;

				#endregion

				#region Whitespace_Before_OpenTagName

				case ParserState.Whitespace_Before_OpenTagName:
					if (char.IsWhiteSpace(symbol))
					{
					}
					else if (char.IsLetterOrDigit(symbol))
					{
						this._openTagName.Length = 0;
						this._openTagName.Append(symbol);
						this._templateValue.Clear();
						this._attributes.Clear();
						this._state = ParserState.OpenTag_Name;
					}
					else
					{
						throw new ArgumentOutOfRangeException("symbol");
					}
					break;

				#endregion

				#region Whitespace_Before_CloseTagName

				case ParserState.Whitespace_Before_CloseTagName:
					if (char.IsWhiteSpace(symbol))
					{
						this._closeTagText.Append(symbol);
					}
					else if (char.IsLetterOrDigit(symbol))
					{
						this._closeTagText.Append(symbol);
						this._closeTagName.Length = 0;
						this._closeTagName.Append(symbol);
						this._state = ParserState.CloseTag_Name;
					}
					else
					{
						throw new ArgumentOutOfRangeException("symbol");
					}
					break;

				#endregion

				#region Whitespace_Before_OpenTagEnd_Or_AttributeName

				case ParserState.Whitespace_Before_OpenTagEnd_Or_AttributeName:
					switch (symbol)
					{
						case '>':
							this._state = ParserState.OpenTag_End;
							break;
						default:
							if (char.IsWhiteSpace(symbol))
							{
							}
							else if (char.IsLetterOrDigit(symbol))
							{
								this._attributeName.Length = 0;
								this._attributeName.Append(symbol);
								this._state = ParserState.Attribute_Name;
							}
							else
							{
								throw new ArgumentOutOfRangeException("symbol");
							}
							break;
					}
					break;

				#endregion

				#region Whitespace_Before_CloseTagEnd

				case ParserState.Whitespace_Before_CloseTagEnd:
					switch (symbol)
					{
						case '>':
							this._state = ParserState.CloseTag_End;
							break;
						default:
							if (!char.IsWhiteSpace(symbol))
							{
								throw new ArgumentOutOfRangeException("symbol");
							}
							break;
					}
					break;

				#endregion

				#region Whitespace_Before_AttributeEqual

				case ParserState.Whitespace_Before_AttributeEqual:
					switch (symbol)
					{
						case '=':
							this._state = ParserState.Attribute_Equal;
							break;
						case '>':
							this._state = ParserState.OpenTag_End;
							break;
						default:
							if (char.IsWhiteSpace(symbol))
							{
							}
							else if (char.IsLetterOrDigit(symbol))
							{
								this._state = ParserState.Attribute_Equal;
							}
							else
							{
								throw new ArgumentOutOfRangeException("symbol");
							}
							break;
					}
					break;

				#endregion

				#region Whitespace_Before_AttributeValue

				case ParserState.Whitespace_Before_AttributeValue:
					switch (symbol)
					{
						case '\'':
							this._attributeValue.Length = 0;
							this._state = ParserState.Attribute_Value_StartSingleQuotes1;
							break;
						case '"':
							this._attributeValue.Length = 0;
							this._state = ParserState.Attribute_Value_StartDoubleQuotes1;
							break;
						default:
							if (char.IsWhiteSpace(symbol))
							{
							}
							else if (char.IsLetterOrDigit(symbol))
							{
								this._attributeValue.Length = 0;
								this._attributeValue.Append(symbol);
								this._state = ParserState.Attribute_Value_NoQuotes;
							}
							else
							{
								throw new ArgumentOutOfRangeException("symbol");
							}
							break;
					}
					break;

				#endregion

				#region OpenTag_Start1

				case ParserState.OpenTag_Start1:
					switch (symbol)
					{
						default:
							if (char.IsWhiteSpace(symbol))
							{
								this._state = ParserState.Whitespace_Before_OpenTagName;
							}
							else if (char.IsLetterOrDigit(symbol))
							{
								this._openTagName.Length = 0;
								this._openTagName.Append(symbol);
								this._templateValue.Clear();
								this._attributes.Clear();
								this._state = ParserState.OpenTag_Name;
							}
							else
							{
								throw new ArgumentOutOfRangeException("symbol");
							}
							break;
					}
					break;

				#endregion

				#region CloseTag_Start1

				case ParserState.CloseTag_Start1:
					switch (symbol)
					{
						case '/':
							this._closeTagText.Append(symbol);
							this._state = ParserState.CloseTag_Start2;
							break;
						default:
							this._closeTagText.Append(symbol);
							CancelCloseTag();
							break;
					}
					break;

				#endregion

				#region CloseTag_Start2

				case ParserState.CloseTag_Start2:
					switch (symbol)
					{
						default:
							if (char.IsWhiteSpace(symbol))
							{
								this._closeTagText.Append(symbol);
								this._state = ParserState.Whitespace_Before_CloseTagName;
							}
							else if (char.IsLetterOrDigit(symbol))
							{
								this._closeTagText.Append(symbol);
								this._closeTagName.Length = 0;
								this._closeTagName.Append(symbol);
								this._state = ParserState.CloseTag_Name;
							}
							else
							{
								this._closeTagText.Append(symbol);
								CancelCloseTag();
							}
							break;
					}
					break;

				#endregion

				#region OpenTag_End

				case ParserState.OpenTag_End:
					switch (symbol)
					{
						case '<':
							this._closeTagText.Length = 0;
							this._closeTagText.Append(symbol);
							this._state = ParserState.CloseTag_Start1;
							break;
						default:
							AppendTemplateSymbol(symbol);
							this._state = ParserState.Whitespace_Before_CloseTagStart;
							break;
					}
					break;

				#endregion

				#region CloseTag_End

				case ParserState.CloseTag_End:
					switch (symbol)
					{
						case '<':
							this._state = ParserState.OpenTag_Start1;
							break;
						default:
							this._state = ParserState.Whitespace_Before_OpenTagStart;
							break;
					}
					break;

				#endregion

				#region OpenTag_Name

				case ParserState.OpenTag_Name:
					switch (symbol)
					{
						case '>':
							this._state = ParserState.OpenTag_End;
							break;
						default:
							if (char.IsWhiteSpace(symbol))
							{
								this._state = ParserState.Whitespace_Before_OpenTagEnd_Or_AttributeName;
							}
							else if (char.IsLetterOrDigit(symbol))
							{
								this._openTagName.Append(symbol);
							}
							else
							{
								throw new ArgumentOutOfRangeException("symbol");
							}
							break;
					}
					break;

				#endregion

				#region CloseTag_Name

				case ParserState.CloseTag_Name:
					if (char.IsWhiteSpace(symbol) || (symbol == '>'))
					{
						if (this._closeTagName.ToString() == this._openTagName.ToString())
						{
							this._templates.Add(
								this._openTagName.ToString(),
								new HtmlTemplate(
									this._openTagName.ToString(),
									this._templateValue.ToArray(),
									new Dictionary<string, string>(this._attributes)));

							if (symbol == '>')
							{
								this._state = ParserState.CloseTag_End;
							}
							else
							{
								this._state = ParserState.Whitespace_Before_CloseTagEnd;
							}
						}
						else
						{
							this._closeTagText.Append(symbol);
							CancelCloseTag();
						}
					}
					else if (char.IsLetterOrDigit(symbol))
					{
						this._closeTagText.Append(symbol);
						this._closeTagName.Append(symbol);
					}
					else
					{
						throw new ArgumentOutOfRangeException("symbol");
					}
					break;

				#endregion

				#region Attribute_Name

				case ParserState.Attribute_Name:
					switch (symbol)
					{
						case '=':
							this._state = ParserState.Attribute_Equal;
							break;
						default:
							if (char.IsWhiteSpace(symbol))
							{
								this._state = ParserState.Whitespace_Before_AttributeEqual;
							}
							else if (char.IsLetterOrDigit(symbol))
							{
								this._attributeName.Append(symbol);
							}
							else
							{
								throw new ArgumentOutOfRangeException("symbol");
							}
							break;
					}
					break;

				#endregion

				#region Attribute_Equal

				case ParserState.Attribute_Equal:
					switch (symbol)
					{
						case '\'':
							this._attributeValue.Length = 0;
							this._state = ParserState.Attribute_Value_StartSingleQuotes1;
							break;
						case '"':
							this._attributeValue.Length = 0;
							this._state = ParserState.Attribute_Value_StartDoubleQuotes1;
							break;
						default:
							if (char.IsWhiteSpace(symbol))
							{
								this._state = ParserState.Whitespace_Before_AttributeValue;
							}
							else if (char.IsLetterOrDigit(symbol))
							{
								this._attributeValue.Length = 0;
								this._attributeValue.Append(symbol);
								this._state = ParserState.Attribute_Value_NoQuotes;
							}
							else
							{
								throw new ArgumentOutOfRangeException("symbol");
							}
							break;
					}
					break;

				#endregion

				#region Attribute_Value_StartSingleQuotes1

				case ParserState.Attribute_Value_StartSingleQuotes1:
					switch (symbol)
					{
						case '\'':
							this._attributes.Add(
								this._attributeName.ToString(),
								this._attributeValue.ToString());
							this._state = ParserState.Attribute_Value_EndSingleQuotes;
							break;
						default:
							this._attributeValue.Length = 0;
							this._attributeValue.Append(symbol);
							this._state = ParserState.Attribute_Value_StartSingleQuotes2;
							break;
					}
					break;

				#endregion

				#region Attribute_Value_StartSingleQuotes2

				case ParserState.Attribute_Value_StartSingleQuotes2:
					switch (symbol)
					{
						case '\'':
							this._attributes.Add(
								this._attributeName.ToString(),
								this._attributeValue.ToString());
							this._state = ParserState.Attribute_Value_EndSingleQuotes;
							break;
						default:
							this._attributeValue.Append(symbol);
							this._state = ParserState.Attribute_Value_SingleQuotes;
							break;
					}
					break;

				#endregion

				#region Attribute_Value_StartDoubleQuotes

				case ParserState.Attribute_Value_StartDoubleQuotes1:
					switch (symbol)
					{
						case '"':
							this._attributes.Add(
								this._attributeName.ToString(),
								this._attributeValue.ToString());
							this._state = ParserState.Attribute_Value_EndDoubleQuotes;
							break;
						default:
							this._attributeValue.Length = 0;
							this._attributeValue.Append(symbol);
							this._state = ParserState.Attribute_Value_StartDoubleQuotes2;
							break;
					}
					break;

				#endregion

				#region Attribute_Value_StartDoubleQuotes2

				case ParserState.Attribute_Value_StartDoubleQuotes2:
					switch (symbol)
					{
						case '"':
							this._attributes.Add(
								this._attributeName.ToString(),
								this._attributeValue.ToString());
							this._state = ParserState.Attribute_Value_EndDoubleQuotes;
							break;
						default:
							this._attributeValue.Append(symbol);
							this._state = ParserState.Attribute_Value_DoubleQuotes;
							break;
					}
					break;

				#endregion

				#region Attribute_Value_EndSingleQuotes

				case ParserState.Attribute_Value_EndSingleQuotes:
					switch (symbol)
					{
						case '>':
							this._state = ParserState.OpenTag_End;
							break;
						default:
							if (char.IsWhiteSpace(symbol))
							{
								this._state = ParserState.Whitespace_Before_OpenTagEnd_Or_AttributeName;
							}
							else if (char.IsLetterOrDigit(symbol))
							{
								this._attributeName.Length = 0;
								this._attributeName.Append(symbol);
								this._state = ParserState.Attribute_Name;
							}
							else
							{
								throw new ArgumentOutOfRangeException("symbol");
							}
							break;
					}
					break;

				#endregion

				#region Attribute_Value_EndDoubleQuotes

				case ParserState.Attribute_Value_EndDoubleQuotes:
					switch (symbol)
					{
						case '>':
							this._state = ParserState.OpenTag_End;
							break;
						default:
							if (char.IsWhiteSpace(symbol))
							{
								this._state = ParserState.Whitespace_Before_OpenTagEnd_Or_AttributeName;
							}
							else if (char.IsLetterOrDigit(symbol))
							{
								this._attributeName.Length = 0;
								this._attributeName.Append(symbol);
								this._state = ParserState.Attribute_Name;
							}
							else
							{
								throw new ArgumentOutOfRangeException("symbol");
							}
							break;
					}
					break;

				#endregion

				#region Attribute_Value_NoQuotes

				case ParserState.Attribute_Value_NoQuotes:
					switch (symbol)
					{
						case '>':
							this._attributes.Add(
								this._attributeName.ToString(),
								this._attributeValue.ToString());
							this._state = ParserState.OpenTag_End;
							break;
						default:
							if (char.IsWhiteSpace(symbol))
							{
								this._attributes.Add(
									this._attributeName.ToString(),
									this._attributeValue.ToString());
								this._state = ParserState.Whitespace_Before_OpenTagEnd_Or_AttributeName;
							}
							else if (char.IsLetterOrDigit(symbol))
							{
								this._attributeValue.Append(symbol);
							}
							else
							{
								throw new ArgumentOutOfRangeException("symbol");
							}
							break;
					}
					break;

				#endregion

				#region Attribute_Value_SingleQuotes

				case ParserState.Attribute_Value_SingleQuotes:
					switch (symbol)
					{
						case '\'':
							this._attributes.Add(
								this._attributeName.ToString(),
								this._attributeValue.ToString());
							this._state = ParserState.Attribute_Value_EndSingleQuotes;
							break;
						default:
							this._attributeValue.Append(symbol);
							break;
					}
					break;

				#endregion

				#region Attribute_Value_DoubleQuotes

				case ParserState.Attribute_Value_DoubleQuotes:
					switch (symbol)
					{
						case '"':
							this._attributes.Add(
								this._attributeName.ToString(),
								this._attributeValue.ToString());
							this._state = ParserState.Attribute_Value_EndDoubleQuotes;
							break;
						default:
							this._attributeValue.Append(symbol);
							break;
					}
					break;

				#endregion
			}
		}

		public void Parse(string html)
		{
			this._templates = new Dictionary<string, HtmlTemplate>();
			this._templateValue = new List<string>();
			this._templateValueLine = new StringBuilder();
			this._openTagName = new StringBuilder();
			this._closeTagName = new StringBuilder();
			this._closeTagText = new StringBuilder();
			this._attributes = new Dictionary<string, string>();
			this._attributeName = new StringBuilder();
			this._attributeValue = new StringBuilder();
			this._state = ParserState.Whitespace_Before_OpenTagStart;

			foreach (char symbol in html)
			{
				ProcessSymbol(symbol);
			}

			ProcessSymbol('\r');
			ProcessSymbol('\n');
		}

		#endregion

		#region Properties

		public IDictionary<string, HtmlTemplate> Templates
		{
			get { return this._templates; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
