﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Text;
using System.Windows.Forms;

using Nabu.Collections;
using Nabu.Reflection;
using Nabu.Text;

namespace Nabu.Forms.Html
{
	public sealed class HtmlListViewBehaviorBinder<T>
	{
		#region Types

		private sealed class TemplateAccessor : ITemplateAccessor
		{
			#region FileTemplateAccessor

			#region Fields

			private IDictionary<string, HtmlTemplate> _templates;

			#endregion

			#region Methods

			public TemplateAccessor(IDictionary<string, HtmlTemplate> templates)
			{
				this._templates = templates;
			}

			#endregion

			#endregion

			#region ITemplateAccessor

			public string[] GetContent(string templateName)
			{
				HtmlTemplate template;

				if (this._templates.TryGetValue(templateName, out template))
				{
					return template.Value;
				}
				else
				{
					return new string[0];
				}
			}

			#endregion
		}

		private sealed class WorkerBaseSourceCodeUnit : ISourceCodeUnit
		{
			#region WorkerBaseSourceCodeUnit

			#region Fields

			private string _value;

			#endregion

			#region Methods

			public WorkerBaseSourceCodeUnit(
				string scopeName,
				string workerTypeName,
				string itemTypeName)
			{
				this._value =
					string.Format(
					CultureInfo.InvariantCulture,
@"namespace {0}
{{
	public abstract class {1}
	{{
		private {2}[] _itemList;
		private int _itemIndex;

		public {1}()
		{{
		}}

		public {2}[] ItemList
		{{
			get {{ return this._itemList; }}
			set {{ this._itemList = value; }}
		}}

		public int ItemIndex
		{{
			get {{ return this._itemIndex; }}
			set {{ this._itemIndex = value; }}
		}}

		public {2} ItemData
		{{
			get {{ return this._itemList[this._itemIndex]; }}
		}}
	}}
}}",
					scopeName,
					workerTypeName,
					itemTypeName);
			}

			#endregion

			#endregion

			#region ISourceCodeUnit

			public IEnumerable<string> Imports
			{
				get { return new string[0]; }
			}

			public IEnumerable<string> Modules
			{
				get { return new string[0]; }
			}

			public string Value
			{
				get { return this._value; }
			}

			public string Module
			{
				get { return "workerbase"; }
			}

			public string Language
			{
				get { return "CSharp"; }
			}

			#endregion
		}

		private sealed class HtmlTemplateEngine
		{
			#region Fields

			private HtmlTemplate _template;
			private IScriptExecutionEngine _engine;
			private string _workerTypeScopeName;
			private string _tagType;

			#endregion

			#region Methods

			#region Utility

			private static string GetTypeName(Type type)
			{
				StringBuilder result = new StringBuilder();

#if COMPACT_FRAMEWORK
			result.AppendFormat(
				CultureInfo.InvariantCulture,
				"{0}.",
				type.Namespace);
#else
				if (type.IsNested)
				{
					result.AppendFormat("{0}.", GetTypeName(type.DeclaringType));
				}
				else
				{
					result.AppendFormat("{0}.", type.Namespace);
				}
#endif

				if (type.IsGenericType)
				{
					result.Append(type.Name.Substring(0, type.Name.IndexOf('`')));

					Type[] genericArguments = type.GetGenericArguments();

					result.Append("<");
					result.Append(GetTypeName(genericArguments[0]));

					for (int index = 1; index < genericArguments.Length; index++)
					{
						result.AppendFormat(
							CultureInfo.InvariantCulture,
							", {0}",
							GetTypeName(genericArguments[index]));
					}

					result.Append(">");
				}
				else
				{
					result.Append(type.Name);
				}

				return result.ToString();
			}

			#endregion

			#region Constructors

			public HtmlTemplateEngine(
				IScriptExecutionHost host,
				IDictionary<string, HtmlTemplate> templates,
				string templateName,
				string workerTypeScopeName)
			{
				this._workerTypeScopeName = workerTypeScopeName;

				HtmlTemplate template;

				if (templates.TryGetValue(templateName, out template))
				{
					this._template = template;

					if (!this._template.Attributes.TryGetValue("-tag-type", out this._tagType))
					{
						this._tagType = "tr";
					}

					TemplateProcessor processor =
						new TemplateProcessor(
							new TemplateAccessor(templates));

					processor.Parse(templateName);
					processor.BaseType = templateName + "WorkerBase";

#if COMPACT_FRAMEWORK
					processor.AddAssembly(typeof(Form).Assembly.GetName().CodeBase);
					processor.AddAssembly(typeof(HtmlForm).Assembly.GetName().CodeBase);
					processor.AddAssembly(typeof(T).Assembly.GetName().CodeBase);
					processor.AddAssembly(typeof(BindingListEx<>).Assembly.GetName().CodeBase);
#else
					processor.AddAssembly(typeof(Form).Assembly.Location);
					processor.AddAssembly(typeof(HtmlForm).Assembly.Location);
					processor.AddAssembly(typeof(T).Assembly.Location);
					processor.AddAssembly(typeof(BindingListEx<>).Assembly.Location);
#endif
					processor.Generate(this._workerTypeScopeName, templateName + "Worker");

					this._engine =
						host.CreateEngine(
							new ISourceCodeUnit[]
							{
								processor.SourceCode,
								new WorkerBaseSourceCodeUnit(
									this._workerTypeScopeName,
									processor.BaseType,
									GetTypeName(typeof(T)))
							});
				}
			}

			#endregion

			#region Html

			private string GetEntityHtml(
				IBindingListEx<T> list,
				int itemIndex)
			{
				using (StringWriter output = new StringWriter(CultureInfo.InvariantCulture))
				{
					IntPtr cookie = IntPtr.Zero;

					try
					{
						cookie =
							this._engine.CreateInstance(
							this._workerTypeScopeName,
							this._template.Name + "Worker");

						this._engine.SetProperty(cookie, "Output", output);
						this._engine.SetProperty(cookie, "ItemList", list.DangerousGetArray());
						this._engine.SetProperty(cookie, "ItemIndex", itemIndex);

						bool isAllowed;

						try
						{
							isAllowed = (bool)this._engine.InvokeMethod(cookie, "PreProcess");
						}
						catch (MissingMethodException)
						{
							isAllowed = true;
						}

						if (isAllowed)
						{
							this._engine.InvokeMethod(cookie, "InProcess");

							try
							{
								this._engine.InvokeMethod(cookie, "PostProcess");
							}
							catch (MissingMethodException)
							{
							}
						}
					}
					finally
					{
						this._engine.DestroyInstance(cookie);
					}

					return output.ToString();
				}
			}

			#endregion

			#region Tag

			public void AddChildTag(
				HtmlTag parent,
				int tagIndex,
				IBindingListEx<T> list,
				int itemIndex)
			{
				if (this._engine != null)
				{
					HtmlTag child = HtmlTag.Create(this._tagType);

					parent.Children.Insert(tagIndex, child);
					child.InnerHtml =
						GetEntityHtml(
							list,
							itemIndex);

					foreach (KeyValuePair<string, string> attribute in this._template.Attributes)
					{
						child.Attributes.Add(attribute);
					}
				}
			}

			public void ChangeChildTag(
				HtmlTag parent,
				int tagIndex,
				IBindingListEx<T> list,
				int itemIndex)
			{
				if (this._engine != null)
				{
					parent.Children[tagIndex].InnerHtml =
						GetEntityHtml(
							list,
							itemIndex);
				}
			}

			#endregion

			#endregion
		}

		#endregion

		#region Fields

		private IScriptExecutionHost _host;
		private IDictionary<string, HtmlTemplateEngine> _engines;
		private string _workerTypeScopeName;
		private IBindingListEx<T> _list;
		private HtmlTag _tag;

		#endregion

		#region Methods

		#region Utility

		private void AddChildTag(
			string templateName,
			int tagIndex,
			int itemIndex)
		{
			this._engines[templateName].AddChildTag(
				this._tag,
				tagIndex,
				this._list,
				itemIndex);
		}

		private void ChangeChildTag(
			string templateName,
			int tagIndex,
			int itemIndex)
		{
			this._engines[templateName].ChangeChildTag(
				this._tag,
				tagIndex,
				this._list,
				itemIndex);
		}

		private void RemoveChildTag(int tagIndex)
		{
			this._tag.Children.RemoveAt(tagIndex);
		}

		private void SwapChildTag(
				int tagIndex1,
				int tagIndex2)
		{
			this._tag.Children[tagIndex1].Swap(this._tag.Children[tagIndex2]);
		}

		#endregion

		#region Event Handlers

		private void List_Added(object sender, BindingListExAddedRemovedEventArgs<T> e)
		{
			if (this._list.Count == 1)
			{
				RemoveChildTag(1);
			}

			if (e.Index > 0)
			{
				AddChildTag("SeparatorTemplate", 2 * e.Index + 0, -1);
			}

			AddChildTag("ItemTemplate", 2 * e.Index + 1, e.Index);
		}

		private void List_Removed(object sender, BindingListExAddedRemovedEventArgs<T> e)
		{
			if (e.Index < this._list.Count)
			{
				RemoveChildTag(2 * e.Index + 2);
				RemoveChildTag(2 * e.Index + 1);
			}
			else
			{
				RemoveChildTag(2 * e.Index + 1);

				if (e.Index > 0)
				{
					RemoveChildTag(2 * e.Index + 0);
				}
			}

			if (this._list.Count == 0)
			{
				AddChildTag("EmptyTemplate", 1, -1);
			}
		}

		private void List_Changed(object sender, BindingListExChangedEventArgs<T> e)
		{
			ChangeChildTag("ItemTemplate", 2 * e.Index + 1, e.Index);
		}

		private void List_Moved(object sender, BindingListExMovedEventArgs<T> e)
		{
			SwapChildTag(2 * e.Index1 + 1, 2 * e.Index2 + 1);
		}

		private void List_Reseted(object sender, EventArgs e)
		{
			this._tag.Children.Clear();

			AddChildTag("HeaderTemplate", int.MaxValue, -1);

			if (this._list.Count == 0)
			{
				AddChildTag("EmptyTemplate", int.MaxValue, -1);
			}
			else
			{
				AddChildTag("ItemTemplate", int.MaxValue, 0);

				for (int index = 1; index < this._list.Count; index++)
				{
					AddChildTag("SeparatorTemplate", int.MaxValue, -1);
					AddChildTag("ItemTemplate", int.MaxValue, index);
				}
			}

			AddChildTag("FooterTemplate", int.MaxValue, -1);
		}

		#endregion

		#region Constructor

		public HtmlListViewBehaviorBinder()
		{
		}

		public void Initialize(
			IDictionary<string, HtmlTemplate> templates,
			IBindingListEx<T> list,
			HtmlTag tag)
		{
			this._workerTypeScopeName =
				string.Format(
					CultureInfo.InvariantCulture,
					"NS_{0:N}",
					Guid.NewGuid());
#if COMPACT_FRAMEWORK
			//this._host = new PaxScriptExecutionHost();
#else
			this._host = new ManagedScriptIntegratedExecutionHost();
#endif
			this._engines = new Dictionary<string, HtmlTemplateEngine>();

			foreach (string templateName in new string[]
				{
					"HeaderTemplate",
					"FooterTemplate",
					"ItemTemplate",
					"SeparatorTemplate",
					"EmptyTemplate"
				})
			{
				this._engines.Add(
					templateName,
					new HtmlTemplateEngine(
						this._host,
						templates,
						templateName,
						this._workerTypeScopeName));
			}

			this._tag = tag;
			this._list = list;
			this._list.Added += new EventHandler<BindingListExAddedRemovedEventArgs<T>>(List_Added);
			this._list.Removed += new EventHandler<BindingListExAddedRemovedEventArgs<T>>(List_Removed);
			this._list.Changed += new EventHandler<BindingListExChangedEventArgs<T>>(List_Changed);
			this._list.Moved += new EventHandler<BindingListExMovedEventArgs<T>>(List_Moved);
			this._list.Reseted += new EventHandler(List_Reseted);

			List_Reseted(this._list, EventArgs.Empty);
		}

		#endregion

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
