﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Globalization;

namespace Nabu.Forms.Html
{
	public static class HtmlAttribute
	{
		/// <summary>
		/// Returns attribute value as string.
		/// </summary>
		/// <param name="tag">HtmlTag containing attribute.</param>
		/// <param name="name">Attribute name.</param>
		/// <param name="defaultValue">Value to return if attribute does not exist.</param>
		/// <returns>Attribute value.</returns>
		public static string GetString(HtmlTag tag, string name, string defaultValue)
		{
			string value = tag.Attributes[name];

			if (value == null)
			{
				return defaultValue;
			}
			else
			{
				return value;
			}
		}

		/// <summary>
		/// Returns attribute value as string. Default value is empty string.
		/// </summary>
		/// <param name="tag">HtmlTag containing attribute.</param>
		/// <param name="name">Attribute name.</param>
		/// <returns>Attribute value</returns>
		public static string GetString(HtmlTag tag, string name)
		{
			return GetString(tag, name, string.Empty);
		}

		/// <summary>
		/// Sets attribute value.
		/// </summary>
		/// <param name="tag">HtmlTag to add attribute to.</param>
		/// <param name="name">Attribute name.</param>
		/// <param name="value">Attribute value.</param>
		public static void SetString(HtmlTag tag, string name, string value)
		{
			tag.Attributes[name] = value;
		}

		/// <summary>
		/// Returns attribute value as signed 32-bit integer.
		/// </summary>
		/// <param name="tag">HtmlTag containing attribute.</param>
		/// <param name="name">Attribute name.</param>
		/// <param name="defaultValue">Value to return if attribute does not exist.</param>
		/// <returns>Attribute value.</returns>
		public static int GetInt32(HtmlTag tag, string name, int defaultValue)
		{
			string value = tag.Attributes[name];

			if (string.IsNullOrEmpty(value))
			{
				return defaultValue;
			}
			else
			{
#if COMPACT_FRAMEWORK
				try
				{
					return int.Parse(value);
				}
				catch
				{
					return defaultValue;
				}
#else
				int result;

				if (int.TryParse(
					value,
					NumberStyles.Integer,
					CultureInfo.InvariantCulture,
					out result))
				{
					return result;
				}
				else
				{
					return defaultValue;
				}
#endif
			}
		}

		/// <summary>
		/// Returns attribute value as signed 32-bit integer. Default value is zero.
		/// </summary>
		/// <param name="tag">HtmlTag containing attribute.</param>
		/// <param name="name">Attribute name.</param>
		/// <returns>Attribute value</returns>
		public static int GetInt32(HtmlTag tag, string name)
		{
			return GetInt32(tag, name, 0);
		}

		/// <summary>
		/// Sets attribute value.
		/// </summary>
		/// <param name="tag">HtmlTag to add attribute to.</param>
		/// <param name="name">Attribute name.</param>
		/// <param name="value">Attribute value.</param>
		public static void SetInt32(HtmlTag tag, string name, int value)
		{
			tag.Attributes[name] =
				value.ToString(
					CultureInfo.InvariantCulture);
		}

		/// <summary>
		/// Returns attribute value as boolean.
		/// </summary>
		/// <param name="tag">HtmlTag containing attribute.</param>
		/// <param name="name">Attribute name.</param>
		/// <param name="defaultValue">Value to return if attribute does not exist.</param>
		/// <returns>Attribute value.</returns>
		public static bool GetBoolean(HtmlTag tag, string name, bool defaultValue)
		{
			string value = tag.Attributes[name];

			if (string.IsNullOrEmpty(value))
			{
				return defaultValue;
			}
			else
			{
				return string.Compare(value, "true", StringComparison.OrdinalIgnoreCase) == 0;
			}
		}

		/// <summary>
		/// Returns attribute value as boolean. Default value is false.
		/// </summary>
		/// <param name="tag">HtmlTag containing attribute.</param>
		/// <param name="name">Attribute name.</param>
		/// <returns>Attribute value</returns>
		public static bool GetBoolean(HtmlTag tag, string name)
		{
			return GetBoolean(tag, name, false);
		}

		/// <summary>
		/// Sets attribute value.
		/// </summary>
		/// <param name="tag">HtmlTag to add attribute to.</param>
		/// <param name="name">Attribute name.</param>
		/// <param name="value">Attribute value.</param>
		public static void SetBoolean(HtmlTag tag, string name, bool value)
		{
			tag.Attributes[name] = (value ? "true" : null);
		}
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
