﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Text;

using Nabu.Collections;

namespace Nabu.Forms.Html
{
	public sealed class HtmlTextBox : IDisposable
	{
		#region HtmlTextBox

		#region Types

		private sealed class TextSelectionWriterContainer
		{
			#region Fields

			private ListEx<char> _list;

			#endregion

			#region Methods

			public TextSelectionWriterContainer()
			{
				this._list = new ListEx<char>();
			}

			public int Write(IntPtr context, int value)
			{
				this._list.Add((char)value);

				return 1;
			}

			#endregion

			#region Properties

			public ListEx<char> List
			{
				[DebuggerStepThrough]
				get { return this._list; }
			}

			#endregion
		}

		private sealed class HtmlSelectionWriterContainer
		{
			#region Fields

			private ListEx<byte> _list;

			#endregion

			#region Methods

			public HtmlSelectionWriterContainer()
			{
				this._list = new ListEx<byte>();
			}

			public int Write(IntPtr context, int value)
			{
				this._list.Add((byte)value);

				return 1;
			}

			#endregion

			#region Properties

			public ListEx<byte> List
			{
				[DebuggerStepThrough]
				get { return this._list; }
			}

			#endregion
		}

		#endregion

		#region Fields

		private HtmlTag _tag;

		#endregion

		#region Methods

		public HtmlTextBox(HtmlTag tag)
		{
			this._tag = tag;
		}

		public void ReplaceSelection(string value)
		{
			HtmlTextBoxValueMethodEventParameters parameters = new HtmlTextBoxValueMethodEventParameters();

			parameters.MethodID = HtmlTagMethodID.ReplaceSelection;

			try
			{
				parameters.lpText = Marshal.StringToBSTR(value);
				parameters.Length = value.Length;

				this._tag.CallMethod(parameters);
			}
			finally
			{
				Marshal.FreeBSTR(parameters.lpText);
			}
		}

		#endregion

		#region Properties

		#region Text

		public string Text
		{
			get
			{
				HtmlTextBoxValueMethodEventParameters parameters = new HtmlTextBoxValueMethodEventParameters();

				parameters.MethodID = HtmlTagMethodID.GetText;

				this._tag.CallMethod(ref parameters);

				if (parameters.Length == 0)
				{
					return string.Empty;
				}
				else
				{
					return Marshal.PtrToStringUni(parameters.lpText, parameters.Length);
				}
			}

			set
			{
				HtmlTextBoxValueMethodEventParameters parameters = new HtmlTextBoxValueMethodEventParameters();

				parameters.MethodID = HtmlTagMethodID.SetText;

				try
				{
					parameters.lpText = Marshal.StringToBSTR(value);
					parameters.Length = value.Length;

					this._tag.CallMethod(parameters);
				}
				finally
				{
					Marshal.FreeBSTR(parameters.lpText);
				}
			}
		}

		#endregion

		#region SelectedText

		public string SelectedText
		{
			get
			{
				HtmlTextSelectionMethodEventParameters parameters = new HtmlTextSelectionMethodEventParameters();
				TextSelectionWriterContainer selectionWriterContainer = new TextSelectionWriterContainer();
				HtmlSelectionWriter selectionWriter = new HtmlSelectionWriter(selectionWriterContainer.Write);

				parameters.MethodID = HtmlTagMethodID.GetSelectedText;
#if COMPACT_FRAMEWORK
				parameters.lpWriter = Marshal.GetFunctionPointerForDelegate(selectionWriter);
#else
				parameters.Writer = selectionWriter;
#endif

				this._tag.CallMethod(ref parameters);

				GC.KeepAlive(selectionWriter);

				return
					new string(
						selectionWriterContainer.List.DangerousGetArray(),
						0,
						selectionWriterContainer.List.Count);
			}
		}

		public string SelectedHtml
		{
			get
			{
				HtmlTextSelectionMethodEventParameters parameters = new HtmlTextSelectionMethodEventParameters();
				HtmlSelectionWriterContainer selectionWriterContainer = new HtmlSelectionWriterContainer();
				HtmlSelectionWriter selectionWriter = new HtmlSelectionWriter(selectionWriterContainer.Write);

				parameters.MethodID = HtmlTagMethodID.GetSelectedHTML;
#if COMPACT_FRAMEWORK
				parameters.lpWriter = Marshal.GetFunctionPointerForDelegate(selectionWriter);
#else
				parameters.Writer = selectionWriter;
#endif

				this._tag.CallMethod(ref parameters);

				GC.KeepAlive(selectionWriter);

				return
					Encoding.UTF8.GetString(
						selectionWriterContainer.List.DangerousGetArray(),
						0,
						selectionWriterContainer.List.Count);
			}
		}

		#endregion

		#region SelectionRange

		public Range<int> SelectionRange
		{
			get
			{
				HtmlTextBoxSelectionMethodEventParameters parameters = new HtmlTextBoxSelectionMethodEventParameters();

				parameters.MethodID = HtmlTagMethodID.GetSelection;

				this._tag.CallMethod(ref parameters);

				return new Range<int>(parameters.SelectionStart, parameters.SelectionEnd);
			}

			set
			{
				HtmlTextBoxSelectionMethodEventParameters parameters = new HtmlTextBoxSelectionMethodEventParameters();

				parameters.MethodID = HtmlTagMethodID.SetSelection;
				parameters.SelectionStart = value.Start;
				parameters.SelectionEnd = value.End;

				this._tag.CallMethod(parameters);
			}
		}

		#endregion

		#region CaretPosition

		public Rectangle CaretPosition
		{
			get
			{
				HtmlTextBoxCaretPositionMethodEventParameters parameters = new HtmlTextBoxCaretPositionMethodEventParameters();

				parameters.MethodID = HtmlTagMethodID.GetCaretPosition;
				this._tag.CallMethod(ref parameters);

				return parameters.Position;
			}
		}

		#endregion

		#endregion

		#endregion

		#region IDisposable

		public void Dispose()
		{
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
