﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics.CodeAnalysis;
using System.Drawing;
using System.Runtime.InteropServices;

using Nabu.Collections;
using Nabu.Drawing;

namespace Nabu.Forms.Html
{
	public sealed partial class HtmlTag
	{
		#region HtmlTag

		#region Fields

		private IntPtr _hTag;
		private string _type;
		private HtmlAttributeCollection _attributes;

		#endregion

		#region Methods

		#region Capture

		/// <summary>
		/// Captures mouse input.
		/// </summary>
		public void SetCapture()
		{
			base.CheckDisposed();

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.SetCapture(
					this._hTag));
		}

		/// <summary>
		/// Releases mouse input capture.
		/// </summary>
		public static void ReleaseCapture()
		{
			UnsafeNativeMethods.Window_ReleaseCapture();
		}

		#endregion

		#region Events

		/// <summary>
		/// Sends event to element.
		/// </summary>
		/// <param name="code">Event code.</param>
		/// <param name="source">Event source.</param>
		/// <param name="parameter">Unmanaged pointer to additional parameters.</param>
		/// <returns></returns>
		public bool SendEvent(int code, HtmlTag source, IntPtr parameter)
		{
			base.CheckDisposed();

			bool isHandled;

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.SendEvent(
					this._hTag,
					code,
					(source == null) ? IntPtr.Zero : source.Handle,
					parameter,
					out isHandled));

			return isHandled;
		}

		/// <summary>
		/// Posts event to element.
		/// </summary>
		/// <param name="code">Event code.</param>
		/// <param name="source">Event source.</param>
		/// <param name="parameter">Unmanaged pointer to additional parameters.</param>
		/// <returns></returns>
		public void PostEvent(int code, HtmlTag source, IntPtr parameter)
		{
			base.CheckDisposed();

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.PostEvent(
					this._hTag,
					code,
					(source == null) ? IntPtr.Zero : source.Handle,
					parameter));
		}

		#endregion

		#region Select

		private static bool OnSelect(
			IntPtr hTag,
			IntPtr param)
		{
			ICollection<HtmlTag> tags = (ICollection<HtmlTag>)(((GCHandle)param).Target);

			tags.Add(HtmlTag.FromHandle(hTag));

			return false;
		}

		private static bool OnSelectFirst(
			IntPtr hTag,
			IntPtr param)
		{
			IntPtrContainer container = (IntPtrContainer)(((GCHandle)param).Target);

			container.Value = hTag;

			return true;
		}

		private static bool OnSelectLast(
			IntPtr hTag,
			IntPtr param)
		{
			IntPtrContainer container = (IntPtrContainer)(((GCHandle)param).Target);

			container.Value = hTag;

			return false;
		}

		/// <summary>
		/// Returns list of elements matching CSS selector.
		/// </summary>
		/// <param name="cssSelector">CSS selector.</param>
		/// <returns>List of elements matching CSS selector</returns>
		public DisposableList<HtmlTag> Select(string cssSelector)
		{
			base.CheckDisposed();

			DisposableList<HtmlTag> tags = new DisposableList<HtmlTag>();
			GCHandle tagsHandle = GCHandle.Alloc(tags);

			try
			{
				DOM_TagEnumHandler selectHandler = new DOM_TagEnumHandler(OnSelect);

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.SelectTags(
						this._hTag,
						cssSelector,
						selectHandler,
						(IntPtr)tagsHandle));

				GC.KeepAlive(selectHandler);
			}
			finally
			{
				if ((IntPtr)tagsHandle != IntPtr.Zero)
				{
					tagsHandle.Free();
				}
			}

			return tags;
		}

		/// <summary>
		/// Returns first element matching CSS selector.
		/// </summary>
		/// <param name="cssSelector">CSS selector.</param>
		/// <returns>First element matching CSS selector.</returns>
		public HtmlTag SelectFirst(string cssSelector)
		{
			base.CheckDisposed();

			HtmlTag result = null;
			IntPtrContainer container = new IntPtrContainer();
			GCHandle containerHandle = GCHandle.Alloc(container);

			try
			{
				DOM_TagEnumHandler selectHandler = new DOM_TagEnumHandler(OnSelectFirst);

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.SelectTags(
						this._hTag,
						cssSelector,
						selectHandler,
						(IntPtr)containerHandle));

				GC.KeepAlive(selectHandler);

				if (container.IsAssigned)
				{
					result = HtmlTag.FromHandle(container.Value);
				}
			}
			finally
			{
				container = null;

				if ((IntPtr)containerHandle != IntPtr.Zero)
				{
					containerHandle.Free();
				}
			}

			return result;
		}

		/// <summary>
		/// Returns last element matching CSS selector.
		/// </summary>
		/// <param name="cssSelector">CSS selector.</param>
		/// <returns>Last element matching CSS selector.</returns>
		public HtmlTag SelectLast(string cssSelector)
		{
			base.CheckDisposed();

			HtmlTag result = null;
			IntPtrContainer container = new IntPtrContainer();
			GCHandle containerHandle = GCHandle.Alloc(container);

			try
			{
				DOM_TagEnumHandler selectHandler = new DOM_TagEnumHandler(OnSelectLast);

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.SelectTags(
						this._hTag,
						cssSelector,
						selectHandler,
						(IntPtr)containerHandle));

				GC.KeepAlive(selectHandler);

				if (container.IsAssigned)
				{
					result = HtmlTag.FromHandle(container.Value);
				}
			}
			finally
			{
				container = null;

				if ((IntPtr)containerHandle != IntPtr.Zero)
				{
					containerHandle.Free();
				}
			}

			return result;
		}

		#endregion

		#region State

		/// <summary>
		/// Returns if HtmlTag is in some state.
		/// </summary>
		/// <param name="state">State to check.</param>
		/// <returns>If HtmlTag is in some state</returns>
		public bool GetState(HtmlTagState state)
		{
			base.CheckDisposed();

			int tagState;

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.GetTagState(
					this._hTag,
					out tagState));

			return (tagState & (int)state) != 0;
		}

		/// <summary>
		/// Sets if HtmlTag is in some state.
		/// </summary>
		/// <param name="state">State to set.</param>
		/// <param name="value">State value.</param>
		/// <param name="redraw">Flag specifying if HtmlTag view should be updated to reflect new state value.</param>
		public void SetState(HtmlTagState state, bool value, bool redraw)
		{
			base.CheckDisposed();

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.SetTagState(
					this._hTag,
					value ? ((int)state) : 0,
					value ? 0 : ((int)state),
					redraw));
		}

		#endregion

		#region Style

#if COMPACT_FRAMEWORK

		private static int OnStyleEnumCF(
			ref HtmlTagStyle style,
			IntPtr param)
		{
			return OnStyleEnum(ref style, param) ? 1 : 0;
		}

#endif

		private static bool OnStyleEnum(
			ref HtmlTagStyle style,
			IntPtr param)
		{
			GCHandle stylesHandle = (GCHandle)param;

			((List<HtmlTagStyle>)stylesHandle.Target).Add(style);

			return true;
		}

		/// <summary>
		/// Returns style value by name.
		/// </summary>
		/// <param name="name">Style name.</param>
		/// <returns>Style value.</returns>
		public string GetStyle(string name)
		{
			base.CheckDisposed();

			IntPtr valuePtr;

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.GetStyleAttribute(
					this._hTag,
					name,
					out valuePtr));

			return valuePtr == IntPtr.Zero ? null : Marshal.PtrToStringUni(valuePtr);
		}

		/// <summary>
		/// Sets style value by name.
		/// </summary>
		/// <param name="name">Style name.</param>
		/// <param name="value">Style value.</param>
		public void SetStyle(string name, string value)
		{
			base.CheckDisposed();

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.SetStyleAttribute(
					this._hTag,
					name,
					value));
		}

		public void ClearAllStyles()
		{
			base.CheckDisposed();

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.SetStyleAttribute(
					this._hTag,
					null,
					null));
		}

		[SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")]
		public IEnumerable<HtmlTagStyle> GetStyleEnumerator()
		{
			base.CheckDisposed();

			IList<HtmlTagStyle> styles = new List<HtmlTagStyle>();
			GCHandle stylesHandle = GCHandle.Alloc(styles);

			try
			{
				DOM_StyleEnumHandler styleEnumHandler = new DOM_StyleEnumHandler(OnStyleEnum);

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.EnumerateStyles(
						this._hTag,
						styleEnumHandler,
						(IntPtr)stylesHandle));

				GC.KeepAlive(styleEnumHandler);
			}
			finally
			{
				stylesHandle.Free();
			}

			foreach (HtmlTagStyle style in styles)
			{
				yield return style;
			}

			yield break;
		}

		#endregion

		#region Symbols

#if COMPACT_FRAMEWORK
		private static int OnSymbolEnumCF(
			IntPtr param,
			IntPtr hTag,
			int position,
			int positionType,
			char code)
		{
			return OnSymbolEnum(param, hTag, position, positionType, code) ? 1 : 0;
		}
#endif

		private static bool OnSymbolEnum(
			IntPtr param,
			IntPtr hTag,
			int position,
			int positionType,
			char code)
		{
			if (positionType == 2)
			{
				GCHandle symbolsHandle = (GCHandle)param;
				Rectangle box;

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.GetSymbolRect(
						hTag,
						position,
						out box));

				((List<HtmlTagSymbol>)symbolsHandle.Target).Add(
					new HtmlTagSymbol(
						box,
						position,
						code));
			}

			return true;
		}

		[SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")]
		public IEnumerable<HtmlTagSymbol> GetSymbolEnumerator()
		{
			base.CheckDisposed();

			IList<HtmlTagSymbol> symbols = new List<HtmlTagSymbol>();
			GCHandle symbolsHandle = GCHandle.Alloc(symbols);

			try
			{
				DOM_SymbolEnumHandler symbolsEnumHandler = new DOM_SymbolEnumHandler(OnSymbolEnum);

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.EnumerateSymbols(
						this._hTag,
						symbolsEnumHandler,
						(IntPtr)symbolsHandle,
						true));

				GC.KeepAlive(symbolsEnumHandler);
			}
			finally
			{
				symbolsHandle.Free();
			}

			foreach (HtmlTagSymbol symbol in symbols)
			{
				yield return symbol;
			}

			yield break;
		}

		#endregion

		#region View

		/// <summary>
		/// Returns window handle for given HtmlTag.
		/// </summary>
		/// <param name="root">Flag specifying which window should be returned.</param>
		/// <returns>Window handle for given HtmlTag</returns>
		public IntPtr GetWindow(bool root)
		{
			base.CheckDisposed();

			IntPtr hWnd;

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.GetTagWindowView(
					this._hTag,
					out hWnd,
					root));

			return hWnd;
		}

		/// <summary>
		/// Returns tag location. Block elements only.
		/// </summary>
		/// <param name="areas">Area and reference point.</param>
		public Rectangle GetLocation(HtmlTagArea areas)
		{
			Rectangle location;

			base.CheckDisposed();

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.GetTagLocation(
					this._hTag,
					out location,
					areas));

			RectangleUtils.ManagedFromNative(ref location);

			return location;
		}

		/// <summary>
		/// Returns intrinsic width range.
		/// </summary>
		/// <returns>Intrinsic width range</returns>
		[SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")]
		public Range<int> GetIntrinsicWidth()
		{
			base.CheckDisposed();

			int minWidth;
			int maxWidth;

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.GetTagIntrinsicWidth(
					this._hTag,
					out minWidth,
					out maxWidth));

			return new Range<int>(minWidth, maxWidth);
		}

		/// <summary>
		/// Returns intrinsic height.
		/// </summary>
		/// <param name="width">Width</param>
		/// <returns>Intrinsic height</returns>
		public int GetIntrinsicHeight(int width)
		{
			base.CheckDisposed();

			int height;

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.GetTagIntrinsicHeight(
					this._hTag,
					width,
					out height));

			return height;
		}

		/// <summary>
		/// Returns scroll information.
		/// </summary>
		/// <returns>Scroll information</returns>
		[SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")]
		public HtmlTagScrollInfo GetScrollInfo()
		{
			base.CheckDisposed();

			Point position;
			Rectangle view;
			Size size;

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.GetScrollInfo(
					this._hTag,
					out position,
					out view,
					out size));

			RectangleUtils.ManagedFromNative(ref view);

			return
				new HtmlTagScrollInfo(
					position,
					view,
					size);
		}

		/// <summary>
		/// Sets element scroll information.
		/// </summary>
		/// <param name="position">Scroll position.</param>
		/// <param name="smooth">Mode.</param>
		public void SetScrollInfo(
			Point position,
			bool smooth)
		{
			base.CheckDisposed();

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.SetScrollInfo(
					this._hTag,
					position,
					smooth));
		}

		/// <summary>
		/// Scrolls view to ensure element visibility.
		/// </summary>
		/// <param name="flags">Flags specifying scroll mode.</param>
		public void ScrollToView(HtmlScrollBehavior flags)
		{
			base.CheckDisposed();

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.ScrollToView(
					this._hTag,
					flags));
		}

		/// <summary>
		/// Updates HtmlTag. This causes content size remeasure etc.
		/// </summary>
		/// <param name="redraw">Flag specifying if HtmlTag view should be updated.</param>
		public void Update(bool redraw)
		{
			base.CheckDisposed();

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.UpdateTag(
					this._hTag,
					redraw));
		}

		/// <summary>
		/// Moves HtmlTag from its normal place to the position defined by argument.
		/// </summary>
		/// <param name="position">HtmlTag position.</param>
		public void Move(Rectangle position)
		{
			base.CheckDisposed();

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.MoveTag(
					this._hTag,
					position.Left,
					position.Top,
					position.Width,
					position.Height));
		}

		#endregion

		#region Popup

		public void ShowPopup(HtmlTag anchor, HtmlPopupAlignment alignment)
		{
			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.ShowPopup(
					this._hTag,
					anchor.Handle,
					alignment));
		}

		public void ShowPopupAt(Point position, bool isAnimated)
		{
			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.ShowPopupAt(
					this._hTag,
					position,
					isAnimated));
		}

		public void HidePopup()
		{
			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.HidePopup(
					this._hTag));
		}

		public HtmlTag TrackPopup(Point position, bool isAnimated)
		{
			base.CheckDisposed();

			IntPtr hTagItem;

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.TrackPopupAt(
					this._hTag,
					position,
					isAnimated,
					out hTagItem));

			return new HtmlTag(hTagItem);
		}

		#endregion

		#endregion

		#region Properties

		/// <summary>
		/// Gets HtmlTag unmanaged handle.
		/// </summary>
		public IntPtr Handle
		{
			get
			{
				base.CheckDisposed();

				return this._hTag;
			}
		}

		/// <summary>
		/// Checks if HtmlTag unmanaged handle is valid.
		/// </summary>
		public bool IsValid
		{
			get
			{
				base.CheckDisposed();

				return this._hTag != IntPtr.Zero;
			}
		}

		/// <summary>
		/// Checks if HtmlTag is visible.
		/// </summary>
		public bool IsVisible
		{
			get
			{
				base.CheckDisposed();

				bool isVisible;

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.IsVisible(
						this._hTag,
						out isVisible));

				return isVisible;
			}
		}

		/// <summary>
		/// Checks if HtmlTag is enabled.
		/// </summary>
		public bool IsEnabled
		{
			get
			{
				base.CheckDisposed();

				bool isEnabled;

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.IsEnabled(
						this._hTag,
						out isEnabled));

				return isEnabled;
			}
		}

		/// <summary>
		/// Gets HtmlTag type: p, div, span etc.
		/// </summary>
		public string Type
		{
			get
			{
				base.CheckDisposed();

				if (this._type == null)
				{
					IntPtr typePtr;

					HtmlDomException.ThrowOnError(
						HtmlDynamicInterop.Dom.GetTagType(
							this._hTag,
							out typePtr));
#if COMPACT_FRAMEWORK
					this._type = typePtr == IntPtr.Zero ? null : UnsafeNativeMethods.PtrToStringAnsi(typePtr);
#else
					this._type = typePtr == IntPtr.Zero ? null : Marshal.PtrToStringAnsi(typePtr);
#endif
				}

				return this._type;
			}
		}

		/// <summary>
		/// Gets collection of HtmlTag attributes.
		/// </summary>
		public HtmlAttributeCollection Attributes
		{
			get
			{
				base.CheckDisposed();

				return this._attributes;
			}
		}

		/// <summary>
		/// Get control type, if HtmlTag represents control.
		/// </summary>
		public HtmlTagControlType ControlType
		{
			get
			{
				base.CheckDisposed();

				HtmlTagControlType controlType;

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.GetControlType(
						this._hTag,
						out controlType));

				return controlType;
			}
		}

		#endregion

		#endregion

		#region Object

		public override bool Equals(object obj)
		{
			base.CheckDisposed();

			HtmlTag tag = obj as HtmlTag;

			if (tag == null)
			{
				return false;
			}
			else
			{
				return this == tag;
			}
		}

		public override int GetHashCode()
		{
			base.CheckDisposed();

			return this._hTag.ToInt32();
		}

		public static bool operator ==(HtmlTag left, HtmlTag right)
		{
			if (((object)left) == null)
			{
				if (((object)right) == null)
				{
					return true;
				}
				else
				{
					return false;
				}
			}
			else
			{
				if (((object)right) == null)
				{
					return false;
				}
				else
				{
					left.CheckDisposed();
					right.CheckDisposed();

					return left.Handle == right.Handle;
				}
			}
		}

		public static bool operator !=(HtmlTag left, HtmlTag right)
		{
			if (((object)left) == null)
			{
				if (((object)right) == null)
				{
					return false;
				}
				else
				{
					return true;
				}
			}
			else
			{
				if (((object)right) == null)
				{
					return true;
				}
				else
				{
					left.CheckDisposed();
					right.CheckDisposed();

					return left.Handle != right.Handle;
				}
			}
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
