﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.ComponentModel;

namespace Nabu.Forms.Html
{
	public sealed partial class HtmlTag
	{
		#region Fields

		private HtmlTagCollection _children;

		#endregion

		#region Methods

		/// <summary>
		/// Swaps two sibling HtmlTags.
		/// </summary>
		/// <param name="tag">HtmlTag to swap with.</param>
		public void Swap(HtmlTag tag)
		{
			base.CheckDisposed();

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.SwapTags(
					this._hTag,
					tag.Handle));
		}

		/// <summary>
		/// Selects parent HtmlTag.
		/// </summary>
		/// <param name="cssSelector">CSS selector.</param>
		/// <param name="depth">Search depth. Use 0 to scan whole chain.</param>
		/// <returns>First parent element matching CSS selector.</returns>
		public HtmlTag SelectParent(string cssSelector, int depth)
		{
			base.CheckDisposed();

			IntPtr hParent;

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.SelectParentTag(
					this._hTag,
					cssSelector,
					depth,
					out hParent));

			if (hParent == IntPtr.Zero)
			{
				return null;
			}
			else
			{
				return HtmlTag.FromHandle(hParent);
			}
		}

		/// <summary>
		/// Selects parent HtmlTag.
		/// </summary>
		/// <param name="cssSelector">CSS selector.</param>
		/// <param name="depth">Search depth.</param>
		/// <returns>First parent element matching CSS selector.</returns>
		public HtmlTag SelectParent(string cssSelector)
		{
			base.CheckDisposed();

			return SelectParent(cssSelector, 0);
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets parent HtmlTag.
		/// </summary>
		public HtmlTag Parent
		{
			get
			{
				base.CheckDisposed();

				IntPtr hTag;

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.GetParentTag(
						this._hTag,
						out hTag));

				if (hTag == IntPtr.Zero)
				{
					return null;
				}
				else
				{
					return HtmlTag.FromHandle(hTag);
				}
			}

			set
			{
				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.DetachTag(
						this._hTag));
				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.InsertTag(
						this._hTag,
						value.Handle,
						int.MaxValue));
			}
		}

		/// <summary>
		/// Gets previous sibling HtmlTag.
		/// </summary>
		public HtmlTag PrevSibling
		{
			get
			{
				base.CheckDisposed();

				int index;

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.GetTagIndex(
						this._hTag,
						out index));

				if (index > 0)
				{
					IntPtr hParentTag;

					HtmlDomException.ThrowOnError(
						HtmlDynamicInterop.Dom.GetParentTag(
							this._hTag,
							out hParentTag));

					if (hParentTag != IntPtr.Zero)
					{
						IntPtr hTagChild;

						HtmlDomException.ThrowOnError(
							HtmlDynamicInterop.Dom.GetChildTag(
								hParentTag,
								index - 1,
								out hTagChild));

						if (hTagChild != IntPtr.Zero)
						{
							return HtmlTag.FromHandle(hTagChild);
						}
					}
				}

				return null;
			}
		}

		/// <summary>
		/// Gets next sibling HtmlTag.
		/// </summary>
		public HtmlTag NextSibling
		{
			get
			{
				base.CheckDisposed();

				int index;

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.GetTagIndex(
						this._hTag,
						out index));

				IntPtr hParentTag;

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.GetParentTag(
						this._hTag,
						out hParentTag));

				if (hParentTag != IntPtr.Zero)
				{
					int childrenCount;

					HtmlDomException.ThrowOnError(
						HtmlDynamicInterop.Dom.GetChildrenCount(
							hParentTag,
							out childrenCount));

					if (index < childrenCount - 1)
					{
						IntPtr hTagChild;

						HtmlDomException.ThrowOnError(
							HtmlDynamicInterop.Dom.GetChildTag(
								hParentTag,
								index + 1,
								out hTagChild));

						if (hTagChild != IntPtr.Zero)
						{
							return HtmlTag.FromHandle(hTagChild);
						}
					}
				}

				return null;
			}
		}

		/// <summary>
		/// Gets root HtmlTag.
		/// </summary>
		public HtmlTag Root
		{
			get
			{
				base.CheckDisposed();

				IntPtr hTag;
				IntPtr hTagRoot = this._hTag;

				do
				{
					hTag = hTagRoot;

					HtmlDomException.ThrowOnError(
						HtmlDynamicInterop.Dom.GetParentTag(
							hTag,
							out hTagRoot));
				} while (hTagRoot != IntPtr.Zero);

				return HtmlTag.FromHandle(hTag);
			}
		}

		/// <summary>
		/// Gets HtmlTag sibling index.
		/// </summary>
		public int Index
		{
			get
			{
				base.CheckDisposed();

				int index;

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.GetTagIndex(
						this._hTag,
						out index));

				return index;
			}
		}

		/// <summary>
		/// Gets collection of children HtmlTags.
		/// </summary>
		public HtmlTagCollection Children
		{
			get
			{
				base.CheckDisposed();

				return this._children;
			}
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
