﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/*                                                                            */
/* Weak References are experimental and may lead to unexpected side effects.  */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
//#define HTMLTAG_REFERENCE_DEFAULTS_TO_WEAK
/******************************************************************************/
using System;
using System.Runtime.InteropServices;
#if !COMPACT_FRAMEWORK
using System.Runtime.CompilerServices;
#endif

namespace Nabu.Forms.Html
{
	public sealed partial class HtmlTag
	{
		#region HtmlTag

		#region Constants

#if HTMLTAG_REFERENCE_DEFAULTS_TO_WEAK
		private const bool DefaultReferenceKind = false;
#else
		private const bool DefaultReferenceKind = true;
#endif

		private static int ExpandoSize = 2 * IntPtr.Size;
		private static int ExpandoFinalizeOffset = 0 * IntPtr.Size;
		private static int ExpandoEventSinkOffset = 1 * IntPtr.Size;

		#endregion

		#region Fields

		#region Static

		private static DOM_ExpandoRelease _expandoRelease;
		private static IntPtr _lpExpandoFinalizer;
		private static DOM_EventHandler _eventHandler;

		#endregion

		#region Instance

		private bool _isTagUsed;

		#endregion

		#endregion

		#region Methods

		#region Initialization

		#region Static Constructors

		static HtmlTag()
		{
			_expandoRelease = new DOM_ExpandoRelease(OnExpandoFinalize);
			_lpExpandoFinalizer = Marshal.GetFunctionPointerForDelegate(_expandoRelease);
#if COMPACT_FRAMEWORK
			_eventHandler = new DOM_EventHandler(OnEventCF);
#else
			_eventHandler = new DOM_EventHandler(OnEvent);
#endif
		}

		#endregion

		#region Instance Constructors

		private HtmlTag(IntPtr hTag, string type, bool isStrongReference)
		{
			this._hTag = hTag;
			this._attributes = new HtmlAttributeCollection(this);
			this._children = new HtmlTagCollection(this);
			this._type = type;

			if (isStrongReference)
			{
				UseTag();
			}
		}

		internal HtmlTag(IntPtr hTag, string type)
			: this(hTag, type, DefaultReferenceKind)
		{
		}

		internal HtmlTag(IntPtr hTag, bool isStrongReference)
			: this(hTag, null, isStrongReference)
		{
		}

		internal HtmlTag(IntPtr hTag)
			: this(hTag, DefaultReferenceKind)
		{
		}

		#endregion

		#region Create

		/// <summary>
		/// Returns new HtmlTag object.
		/// </summary>
		/// <param name="type">Tag type: p, div, span etc.</param>
		/// <param name="text">Tag inner text.</param>
		/// <param name="isStrongReference">Is returned object strong reference.</param>
		/// <returns>New HtmlTag object.</returns>
		public static HtmlTag Create(string type, string text, bool isStrongReference)
		{
			IntPtr hTag;

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.CreateTag(
					type,
					text,
					out hTag));

			return new HtmlTag(hTag, type);
		}

		/// <summary>
		/// Returns new HtmlTag object.
		/// </summary>
		/// <param name="type">Tag type: p, div, span etc.)</param>
		/// <param name="text">Tag inner text.</param>
		/// <returns>New HtmlTag object.</returns>
		public static HtmlTag Create(string type, string text)
		{
			return Create(type, text, DefaultReferenceKind);
		}

		/// <summary>
		/// Returns new HtmlTag object.
		/// </summary>
		/// <param name="type">Tag type: p, div, span etc.)</param>
		/// <returns>New HtmlTag object.</returns>
		public static HtmlTag Create(string type)
		{
			return Create(type, null, DefaultReferenceKind);
		}

		#endregion

		#region From

		/// <summary>
		/// Returns HtmlTag object for unmanaged handle.
		/// </summary>
		/// <param name="hTag">Unmanaged handle.</param>
		/// <param name="isStrongReference">Is returned object hard reference.</param>
		/// <returns>HtmlTag object for unmanaged handle.</returns>
		public static HtmlTag FromHandle(IntPtr hTag, bool isStrongReference)
		{
			return new HtmlTag(hTag, isStrongReference);
		}

		/// <summary>
		/// Returns HtmlTag object for unmanaged handle.
		/// </summary>
		/// <param name="hTag">Unmanaged handle.</param>
		/// <returns>HtmlTag object for unmanaged handle.</returns>
		public static HtmlTag FromHandle(IntPtr hTag)
		{
			return FromHandle(hTag, DefaultReferenceKind);
		}

		/// <summary>
		/// Creates HtmlTag for specified HTML fragment.
		/// </summary>
		/// <param name="rootTag">Root HtmlTag.</param>
		/// <param name="html">HTML fragment containing only one element.</param>
		/// <returns>HtmlTag object for HTML fragment.</returns>
		public static HtmlTag FromHtml(HtmlTag rootTag, string html)
		{
			using (HtmlTag headTag = rootTag.SelectFirst("head"))
			{
				int childrenCount = headTag.Children.Count;

				headTag.AppendHtml(html);

				HtmlTag htmlTag = headTag.Children[childrenCount];

				headTag.Children.RemoveAt(childrenCount);

				return htmlTag;
			}
		}

		#endregion

		#region Clone

		/// <summary>
		/// Returns clone of given HtmlTag.
		/// </summary>
		/// <param name="isStrongReference">Is returned object hard reference.</param>
		/// <returns>Clone of given HtmlTag</returns>
		public HtmlTag Clone(bool isStrongReference)
		{
			base.CheckDisposed();

			IntPtr hTag;

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.CloneTag(
					this._hTag,
					out hTag));

			return new HtmlTag(hTag, this._type, isStrongReference);
		}

		/// <summary>
		/// Returns clone of given HtmlTag.
		/// </summary>
		/// <returns>Clone of given HtmlTag</returns>
		public HtmlTag Clone()
		{
			base.CheckDisposed();

			return Clone(DefaultReferenceKind);
		}

		#endregion

		#region Delete

		/// <summary>
		/// Deletes HtmlTag.
		/// </summary>
		public void Delete()
		{
			base.CheckDisposed();

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.DeleteTag(
					this._hTag));

			UnuseTag();
			this._hTag = IntPtr.Zero;
		}

		#endregion

		#endregion

		#region Expando Callback

		private static void OnExpandoFinalize(
			IntPtr lpExpando,
			IntPtr hTag)
		{
			HtmlTag tag = HtmlTag.FromHandle(hTag, false);

			tag.TryDestroyExpando();
			tag.Dispose(true);
		}

		#endregion

		#region Event Dispatcher

#if COMPACT_FRAMEWORK
		private static int OnEventCF(
			IntPtr handlerID,
			IntPtr hTag,
			HtmlEventGroup eventGroup,
			IntPtr lpParameters)
		{
			return OnEvent(handlerID, hTag, eventGroup, lpParameters) ? 1 : 0;
		}
#endif

		private static bool OnEvent(
			IntPtr handlerID,
			IntPtr hTag,
			HtmlEventGroup eventGroup,
			IntPtr lpParameters)
		{
			using (HtmlTag sender = HtmlTag.FromHandle(hTag))
			{
				HtmlTagEventSink sink = sender.GetEventSink();

				return sink.DispatchEvent(sender, eventGroup, lpParameters);
			}
		}

		#endregion

		#region Reference Count

		private void UseTag()
		{
			base.CheckDisposed();

			if ((!this._isTagUsed) && (this._hTag != IntPtr.Zero))
			{
				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.UseTag(
						this._hTag));
				this._isTagUsed = true;
			}
		}

		private void UnuseTag()
		{
			base.CheckDisposed();

			if (this._isTagUsed)
			{
				this._isTagUsed = false;
				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.UnuseTag(
						this._hTag));
			}
		}

		#endregion

		#region Expando

		/******************************************************************************/
		/* internal struct HtmlTagExpando                                             */
		/* {                                                                          */
		/*   [MarshalAs(UnmanagedType.FunctionPtr)]                                   */
		/*   public DOM_ExpandoRelease Finalize;                                      */
		/*                                                                            */
		/*   [MarshalAs(UnmanagedType.SysInt)]                                        */
		/*   public GCHandle           EventSink;                                     */
		/* }                                                                          */
		/*                                                                            */
		/******************************************************************************/

		private IntPtr CreateExpando()
		{
			base.CheckDisposed();

#if !COMPACT_FRAMEWORK
			RuntimeHelpers.PrepareConstrainedRegions();
#endif
			IntPtr lpExpando;

			try
			{
			}
			finally
			{
				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.GetTagExpando(this._hTag, out lpExpando));

				if (lpExpando == IntPtr.Zero)
				{
					GCHandle gcHandle = GCHandle.Alloc(new HtmlTagEventSink(), GCHandleType.Normal);
					IntPtr gcHandlePointer = (IntPtr)gcHandle;

					lpExpando = Marshal.AllocHGlobal(ExpandoSize);
#if COMPACT_FRAMEWORK
					Marshal.WriteIntPtr(new IntPtr(lpExpando.ToInt32() + ExpandoFinalizeOffset), _lpExpandoFinalizer);
					Marshal.WriteIntPtr(new IntPtr(lpExpando.ToInt32() + ExpandoEventSinkOffset), gcHandlePointer);
#else
					Marshal.WriteIntPtr(lpExpando, ExpandoFinalizeOffset, _lpExpandoFinalizer);
					Marshal.WriteIntPtr(lpExpando, ExpandoEventSinkOffset, gcHandlePointer);
#endif
					HtmlDomException.ThrowOnError(
						HtmlDynamicInterop.Dom.SetTagExpando(this._hTag, lpExpando));

					HtmlDomException.ThrowOnError(
						HtmlDynamicInterop.Dom.AttachEventHandler(
							this._hTag,
							_eventHandler,
							0,
							HtmlEventGroup.All));
				}
			}

			return lpExpando;
		}

		private void TryDestroyExpando()
		{
			base.CheckDisposed();
			/*
			#if COMPACT_FRAMEWORK
						HtmlDomException.ThrowOnError(
							HtmlDynamicInterop.Dom.DetachEventHandler(
								this._hTag,
								Marshal.GetFunctionPointerForDelegate(_eventHandler),
								0));
			#else
								HtmlDomException.ThrowOnError(
									HtmlDynamicInterop.Dom.DetachEventHandler(
										this._hTag,
										_eventHandler,
										0));
			#endif
			*/
			IntPtr lpExpando;

			HtmlDomException.ThrowOnError(
				HtmlDynamicInterop.Dom.GetTagExpando(this._hTag, out lpExpando));

			if (lpExpando != IntPtr.Zero)
			{
#if COMPACT_FRAMEWORK
				IntPtr gcHandlePointer = Marshal.ReadIntPtr(new IntPtr(lpExpando.ToInt32() + ExpandoEventSinkOffset));
#else
				IntPtr gcHandlePointer = Marshal.ReadIntPtr(lpExpando, ExpandoEventSinkOffset);
#endif
				GCHandle gcHandle = (GCHandle)gcHandlePointer;
				HtmlTagEventSink eventSink = (HtmlTagEventSink)gcHandle.Target;

				gcHandle.Free();
				eventSink.Dispose();
				Marshal.FreeHGlobal(lpExpando);

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.SetTagExpando(this._hTag, IntPtr.Zero));
			}
		}

		private HtmlTagEventSink GetEventSink()
		{
			base.CheckDisposed();

			IntPtr lpExpando = CreateExpando();
#if COMPACT_FRAMEWORK
			IntPtr gcHandlePointer = Marshal.ReadIntPtr(new IntPtr(lpExpando.ToInt32() + ExpandoEventSinkOffset));
#else
			IntPtr gcHandlePointer = Marshal.ReadIntPtr(lpExpando, ExpandoEventSinkOffset);
#endif
			GCHandle gcHandle = (GCHandle)gcHandlePointer;

			return (HtmlTagEventSink)gcHandle.Target;
		}

		#endregion

		#endregion

		#region Events

		public event EventHandler<HtmlInitializationEventArgs> Initialization
		{
			add
			{
				base.CheckDisposed();

				GetEventSink().Initialization += value;
			}

			remove
			{
				base.CheckDisposed();

				GetEventSink().Initialization -= value;
			}
		}

		public event EventHandler<HtmlMouseEventArgs> Mouse
		{
			add
			{
				base.CheckDisposed();

				GetEventSink().Mouse += value;
			}

			remove
			{
				base.CheckDisposed();

				GetEventSink().Mouse -= value;
			}
		}

		public event EventHandler<HtmlKeyboardEventArgs> Keyboard
		{
			add
			{
				base.CheckDisposed();

				GetEventSink().Keyboard += value;
			}

			remove
			{
				base.CheckDisposed();

				GetEventSink().Keyboard -= value;
			}
		}

		public event EventHandler<HtmlFocusEventArgs> Focus
		{
			add
			{
				base.CheckDisposed();

				GetEventSink().Focus += value;
			}

			remove
			{
				base.CheckDisposed();

				GetEventSink().Focus -= value;
			}
		}

		public event EventHandler<HtmlScrollEventArgs> Scroll
		{
			add
			{
				base.CheckDisposed();

				GetEventSink().Scroll += value;
			}

			remove
			{
				base.CheckDisposed();

				GetEventSink().Scroll -= value;
			}
		}

		public event EventHandler<HtmlTimerEventArgs> Timer
		{
			add
			{
				base.CheckDisposed();

				GetEventSink().Timer += value;
			}

			remove
			{
				base.CheckDisposed();

				GetEventSink().Timer -= value;
			}
		}

		public event EventHandler<EventArgs> Size
		{
			add
			{
				base.CheckDisposed();

				GetEventSink().Size += value;
			}

			remove
			{
				base.CheckDisposed();

				GetEventSink().Size -= value;
			}
		}

		public event EventHandler<HtmlDrawEventArgs> Draw
		{
			add
			{
				base.CheckDisposed();

				GetEventSink().Draw += value;
			}

			remove
			{
				base.CheckDisposed();

				GetEventSink().Draw -= value;
			}
		}

		public event EventHandler<HtmlDataEventArgs> Data
		{
			add
			{
				base.CheckDisposed();

				GetEventSink().Data += value;
			}

			remove
			{
				base.CheckDisposed();

				GetEventSink().Data -= value;
			}
		}

		public event EventHandler<HtmlBehaviorEventArgs> Behavior
		{
			add
			{
				base.CheckDisposed();

				GetEventSink().Behavior += value;
			}

			remove
			{
				base.CheckDisposed();

				GetEventSink().Behavior -= value;
			}
		}

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			UnuseTag();
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
