﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics.CodeAnalysis;
using System.Runtime.InteropServices;

namespace Nabu.Forms.Html
{
	public sealed partial class HtmlTag
	{
		#region Call

		/// <summary>
		/// Calls HtmlTag method.
		/// </summary>
		/// <typeparam name="T">Structure specifying method parameters.</typeparam>
		/// <param name="parameters">Method parameters.</param>
		internal void CallMethod<T>(T parameters)
		{
			base.CheckDisposed();

			using (NativeMemory buffer = new NativeMemory(Marshal.SizeOf(parameters)))
			{
				buffer.FromStructure(parameters, false);

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.CallBehaviorMethod(
						this._hTag,
						buffer.Pointer));
			}
		}

		/// <summary>
		/// Calls HtmlTag method. Modifications in method parameters are reflected.
		/// </summary>
		/// <typeparam name="T">Structure specifying method parameters.</typeparam>
		/// <param name="parameters">Method parameters.</param>
		internal void CallMethod<T>(ref T parameters)
		{
			base.CheckDisposed();

			using (NativeMemory buffer = new NativeMemory(Marshal.SizeOf(parameters)))
			{
				buffer.FromStructure(parameters, false);

				HtmlDomException.ThrowOnError(
					HtmlDynamicInterop.Dom.CallBehaviorMethod(
						this._hTag,
						buffer.Pointer));

				parameters = buffer.ToStructure<T>();
			}
		}

		#endregion

		#region TryCall

		/// <summary>
		/// Tries to call HtmlTag method.
		/// </summary>
		/// <typeparam name="T">Structure specifying method parameters.</typeparam>
		/// <param name="parameters">Method parameters.</param>
		/// <returns>Error code if any.</returns>
		internal int TryCallMethod<T>(T parameters)
		{
			base.CheckDisposed();

			using (NativeMemory buffer = new NativeMemory(Marshal.SizeOf(parameters)))
			{
				buffer.FromStructure(parameters, false);

				return
					HtmlDynamicInterop.Dom.CallBehaviorMethod(
						this._hTag,
						buffer.Pointer);
			}
		}

		/// <summary>
		/// Tries to call HtmlTag method. Modifications in method parameters are reflected.
		/// </summary>
		/// <typeparam name="T">Structure specifying method parameters.</typeparam>
		/// <param name="parameters">Method parameters.</param>
		/// <returns>Error code if any.</returns>
		internal int TryCallMethod<T>(ref T parameters)
		{
			base.CheckDisposed();

			using (NativeMemory buffer = new NativeMemory(Marshal.SizeOf(parameters)))
			{
				buffer.FromStructure(parameters, false);

				int result =
					HtmlDynamicInterop.Dom.CallBehaviorMethod(
						this._hTag,
						buffer.Pointer);

				if (result == (int)HtmlResult.DOM_Success)
				{
					parameters = buffer.ToStructure<T>();
				}
				else
				{
					parameters = default(T);
				}

				return result;
			}
		}

		#endregion

		#region XCall

		private object XCall(
			string name,
			Type returnType,
			object[] args)
		{
			base.CheckDisposed();

			HtmlXCallMethodEventParameters parameters = new HtmlXCallMethodEventParameters();
			int valueSize = Marshal.SizeOf(typeof(JsonValueData));

			using (NativeMemory bufferArguments = new NativeMemory(args.Length * valueSize))
			{
				parameters.ArgumentCount = args.Length;
				parameters.lpArgumentValues = bufferArguments.Pointer;

				try
				{
					for (int index = 0; index < args.Length; index++)
					{
						JsonValueData argument = JsonValue.GetValueFromObject(args[index]);

						bufferArguments.FromStructure(argument, false, index * valueSize);
					}

					try
					{
						using (NativeMemory bufferName = NativeMemory.FromComStringAnsi(name))
						{
							parameters.MethodID = HtmlTagMethodID.XCall;
							parameters.lpMethodName = bufferName.Pointer;

							CallMethod<HtmlXCallMethodEventParameters>(ref parameters);
						}

						return JsonValue.GetObjectFromValue(parameters.ReturnValue);
					}
					finally
					{
						HtmlDynamicInterop.Json.Clear(ref parameters.ReturnValue);
					}
				}
				finally
				{
					for (int index = 0; index < args.Length; index++)
					{
						JsonValueData argument =
							(JsonValueData)Marshal.PtrToStructure(
								new IntPtr(parameters.lpArgumentValues.ToInt64() + index * valueSize),
								typeof(JsonValueData));

						HtmlDynamicInterop.Json.Clear(ref argument);
					}
				}
			}
		}

		/// <summary>
		/// Calls behavior method of behavior associated with HtmlTag.
		/// </summary>
		/// <typeparam name="TR">Return type.</typeparam>
		/// <param name="name">Method name.</param>
		/// <returns>Method return value.</returns>
		[SuppressMessage("Microsoft.Design", "CA1004:GenericMethodsShouldProvideTypeParameter")]
		public TR XCall<TR>(string name)
		{
			return
				(TR)XCall(
					name,
					typeof(TR),
					new object[0]);
		}

		/// <summary>
		/// Calls behavior method of behavior associated with HtmlTag.
		/// </summary>
		/// <typeparam name="TR">Return type.</typeparam>
		/// <typeparam name="T1">First parameter type.</typeparam>
		/// <param name="name">Method name.</param>
		/// <param name="a1">First parameter value.</param>
		/// <returns>Method return value.</returns>
		[SuppressMessage("Microsoft.Design", "CA1004:GenericMethodsShouldProvideTypeParameter")]
		public TR XCall<TR, T1>(string name, T1 a1)
		{
			return
				(TR)XCall(
					name,
					typeof(TR),
					new object[] { a1 });
		}

		/// <summary>
		/// Calls behavior method of behavior associated with HtmlTag.
		/// </summary>
		/// <typeparam name="TR">Return type.</typeparam>
		/// <typeparam name="T1">First parameter type.</typeparam>
		/// <typeparam name="T2">Second parameter type.</typeparam>
		/// <param name="name">Method name.</param>
		/// <param name="a1">First parameter value.</param>
		/// <param name="a2">Second parameter value.</param>
		/// <returns>Method return value.</returns>
		[SuppressMessage("Microsoft.Design", "CA1004:GenericMethodsShouldProvideTypeParameter")]
		public TR XCall<TR, T1, T2>(string name, T1 a1, T2 a2)
		{
			return
				(TR)XCall(
					name,
					typeof(TR),
					new object[] { a1, a2 });
		}

		/// <summary>
		/// Calls behavior method of behavior associated with HtmlTag.
		/// </summary>
		/// <typeparam name="TR">Return type.</typeparam>
		/// <typeparam name="T1">First parameter type.</typeparam>
		/// <typeparam name="T2">Second parameter type.</typeparam>
		/// <typeparam name="T3">Third parameter type.</typeparam>
		/// <param name="name">Method name.</param>
		/// <param name="a1">First parameter value.</param>
		/// <param name="a2">Second parameter value.</param>
		/// <param name="a3">Third parameter value.</param>
		/// <returns>Method return value.</returns>
		[SuppressMessage("Microsoft.Design", "CA1004:GenericMethodsShouldProvideTypeParameter")]
		public TR XCall<TR, T1, T2, T3>(string name, T1 a1, T2 a2, T3 a3)
		{
			return
				(TR)XCall(
					name,
					typeof(TR),
					new object[] { a1, a2, a3 });
		}

		/// <summary>
		/// Calls behavior method of behavior associated with HtmlTag.
		/// </summary>
		/// <param name="name">Method name.</param>
		/// <param name="args">Method parameters.</param>
		/// <returns>Method return value.</returns>
		public object XCall(string name, params object[] args)
		{
			return
				XCall(
					name,
					typeof(object),
					args);
		}

		#endregion

		#region ManagedCall

		private object ManagedCall(
			string name,
			Type returnType,
			object[] args)
		{
			base.CheckDisposed();

			HtmlManagedCallMethodEventParameters parameters = new HtmlManagedCallMethodEventParameters();

			using (NativeMemory bufferArguments = new NativeMemory(IntPtr.Size * args.Length))
			{
				parameters.ArgumentCount = args.Length;
				parameters.lpArgumentValues = bufferArguments.Pointer;

				try
				{
					for (int index = 0; index < args.Length; index++)
					{
						IntPtr argumentValue =
							(IntPtr)GCHandle.Alloc(
								args[index],
								GCHandleType.Normal);

						bufferArguments.WriteIntPtr(IntPtr.Size * index, argumentValue);
					}

					try
					{
						parameters.MethodID = HtmlTagMethodID.ManagedCall;
						parameters.MethodName = name;

						CallMethod<HtmlManagedCallMethodEventParameters>(ref parameters);

						return ((GCHandle)parameters.ReturnValue).Target;
					}
					finally
					{
						((GCHandle)parameters.ReturnValue).Free();
					}
				}
				finally
				{
					for (int index = 0; index < args.Length; index++)
					{
						IntPtr handle =
							Marshal.ReadIntPtr(
								new IntPtr(parameters.lpArgumentValues.ToInt64() + IntPtr.Size * index));

						if (handle != IntPtr.Zero)
						{
							((GCHandle)handle).Free();
						}
					}
				}
			}
		}

		/// <summary>
		/// Calls managed method of managed behavior associated with HtmlTag.
		/// </summary>
		/// <typeparam name="TR">Return type.</typeparam>
		/// <param name="name">>Method name.</param>
		/// <returns>Method return value.</returns>
		[SuppressMessage("Microsoft.Design", "CA1004:GenericMethodsShouldProvideTypeParameter")]
		public TR ManagedFunc<TR>(string name)
		{
			return
				(TR)ManagedCall(
					name,
					typeof(TR),
					new object[0]);
		}

		/// <summary>
		///  Calls managed method, returning void, of managed behavior associated with HtmlTag.
		/// </summary>
		/// <param name="name">Method return value.</param>
		public void ManagedProc(string name)
		{
			ManagedCall(
				name,
				typeof(void),
				new object[0]);
		}

		/// <summary>
		/// Calls managed method of managed behavior associated with HtmlTag.
		/// </summary>
		/// <typeparam name="TR">Return type.</typeparam>
		/// <typeparam name="T1">First parameter type.</typeparam>
		/// <param name="name">Method name.</param>
		/// <param name="a1">First parameter value.</param>
		/// <returns>Method return value.</returns>
		[SuppressMessage("Microsoft.Design", "CA1004:GenericMethodsShouldProvideTypeParameter")]
		public TR ManagedFunc<TR, T1>(string name, T1 a1)
		{
			return
				(TR)ManagedCall(
					name,
					typeof(TR),
					new object[] { a1 });
		}

		/// <summary>
		/// Calls managed method, returning void, of managed behavior associated with HtmlTag.
		/// </summary>
		/// <typeparam name="T1">First parameter type.</typeparam>
		/// <param name="name">Method name.</param>
		/// <param name="a1">First parameter value.</param>
		public void ManagedProc<T1>(string name, T1 a1)
		{
			ManagedCall(
				name,
				typeof(void),
				new object[] { a1 });
		}

		/// <summary>
		/// Calls managed method of managed behavior associated with HtmlTag.
		/// </summary>
		/// <typeparam name="TR">Return type.</typeparam>
		/// <typeparam name="T1">First parameter type.</typeparam>
		/// <typeparam name="T2">Second parameter type.</typeparam>
		/// <param name="name">Method name.</param>
		/// <param name="a1">First parameter value.</param>
		/// <param name="a2">Second parameter value.</param>
		/// <returns>Method return value.</returns>
		[SuppressMessage("Microsoft.Design", "CA1004:GenericMethodsShouldProvideTypeParameter")]
		public TR ManagedFunc<TR, T1, T2>(string name, T1 a1, T2 a2)
		{
			return
				(TR)ManagedCall(
					name,
					typeof(TR),
					new object[] { a1, a2 });
		}

		/// <summary>
		/// Calls managed method, returning void, of managed behavior associated with HtmlTag.
		/// </summary>
		/// <typeparam name="T1">First parameter type.</typeparam>
		/// <typeparam name="T2">Second parameter type.</typeparam>
		/// <param name="name">Method name.</param>
		/// <param name="a1">First parameter value.</param>
		/// <param name="a2">Second parameter value.</param>
		public void ManagedProc<T1, T2>(string name, T1 a1, T2 a2)
		{
			ManagedCall(
				name,
				typeof(void),
				new object[] { a1, a2 });
		}

		/// <summary>
		/// Calls managed method of managed behavior associated with HtmlTag.
		/// </summary>
		/// <typeparam name="TR">Return type.</typeparam>
		/// <typeparam name="T1">First parameter type.</typeparam>
		/// <typeparam name="T2">Second parameter type.</typeparam>
		/// <typeparam name="T3">Third parameter type.</typeparam>
		/// <param name="name">Method name.</param>
		/// <param name="a1">First parameter value.</param>
		/// <param name="a2">Second parameter value.</param>
		/// <param name="a3">Third parameter value.</param>
		/// <returns>Method return value.</returns>
		[SuppressMessage("Microsoft.Design", "CA1004:GenericMethodsShouldProvideTypeParameter")]
		public TR ManagedFunc<TR, T1, T2, T3>(string name, T1 a1, T2 a2, T3 a3)
		{
			return
				(TR)ManagedCall(
					name,
					typeof(TR),
					new object[] { a1, a2, a3 });
		}

		/// <summary>
		/// Calls managed method, returning void, of managed behavior associated with HtmlTag.
		/// </summary>
		/// <typeparam name="T1">First parameter type.</typeparam>
		/// <typeparam name="T2">Second parameter type.</typeparam>
		/// <typeparam name="T3">Third parameter type.</typeparam>
		/// <param name="name">Method name.</param>
		/// <param name="a1">First parameter value.</param>
		/// <param name="a2">Second parameter value.</param>
		/// <param name="a3">Third parameter value.</param>
		public void ManagedProc<T1, T2, T3>(string name, T1 a1, T2 a2, T3 a3)
		{
			ManagedCall(
				name,
				typeof(void),
				new object[] { a1, a2, a3 });
		}

		/// <summary>
		/// Calls managed method of managed behavior associated with HtmlTag.
		/// </summary>
		/// <param name="name">Method name.</param>
		/// <param name="args">Method parameters.</param>
		/// <returns>Method return value.</returns>
		public object ManagedFunc(string name, params object[] args)
		{
			return
				ManagedCall(
					name,
					typeof(object),
					args);
		}

		/// <summary>
		/// Calls managed method, returning void, of managed behavior associated with HtmlTag.
		/// </summary>
		/// <param name="name">Method name.</param>
		/// <param name="args">Method parameters.</param>
		public void ManagedProc(string name, params object[] args)
		{
			ManagedCall(
				name,
				typeof(void),
				args);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
