﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Nabu.Forms.Html
{
	public abstract class HtmlEventSinkBase : DisposableBase
	{
		#region Fields

		private HtmlEventGroup _eventGroup;

		#endregion

		#region Methods

		#region Event Handlers

		protected virtual void OnInitialization(HtmlTag sender, HtmlInitializationEventArgs e)
		{
		}

		protected virtual void OnMouse(HtmlTag sender, HtmlMouseEventArgs e)
		{
			using (HtmlTag target = e.Target)
			{
				if (
					(e.Command == (HtmlMouseEventCommand.IsSinking | HtmlMouseEventCommand.Up)) &&
					target.GetState(HtmlTagState.Pressed))
				{
					HtmlMouseEventArgs eventArgsSinking =
						new HtmlMouseEventArgs(
							e.Parameters,
							HtmlMouseEventCommand.IsSinking | HtmlMouseEventCommand.Click);

					OnMouse(sender, eventArgsSinking);

					if (!eventArgsSinking.IsHandled)
					{
						HtmlMouseEventArgs eventArgsBubbling =
							new HtmlMouseEventArgs(
								e.Parameters,
								HtmlMouseEventCommand.IsBubbling | HtmlMouseEventCommand.Click);

						OnMouse(sender, eventArgsBubbling);

						if (eventArgsBubbling.IsHandled)
						{
							HtmlMouseEventArgs eventArgsHandled =
								new HtmlMouseEventArgs(
									e.Parameters,
									HtmlMouseEventCommand.IsHandled | HtmlMouseEventCommand.Click);

							OnMouse(sender, eventArgsHandled);
						}
					}
				}
			}
		}

		protected virtual void OnKeyboard(HtmlTag sender, HtmlKeyboardEventArgs e)
		{
		}

		protected virtual void OnFocus(HtmlTag sender, HtmlFocusEventArgs e)
		{
		}

		protected virtual void OnScroll(HtmlTag sender, HtmlScrollEventArgs e)
		{
		}

		protected virtual void OnTimer(HtmlTag sender, HtmlTimerEventArgs e)
		{
		}

		protected virtual void OnSize(HtmlTag sender, EventArgs e)
		{
		}

		protected virtual void OnDraw(HtmlTag sender, HtmlDrawEventArgs e)
		{
		}

		protected virtual void OnData(HtmlTag sender, HtmlDataEventArgs e)
		{
		}

		protected virtual void OnBehavior(HtmlTag sender, HtmlBehaviorEventArgs e)
		{
		}

		protected virtual void OnMethod(HtmlTag sender, HtmlMethodEventArgs e)
		{
		}

		#endregion

		#region Constructors

		protected HtmlEventSinkBase(HtmlEventGroup eventGroup)
		{
			this._eventGroup = eventGroup;
		}

		#endregion

		#region Event Dispatcher

		public bool DispatchEvent(
			HtmlTag sender,
			HtmlEventGroup eventGroup,
			IntPtr lpParameters)
		{
			switch (eventGroup)
			{
				case HtmlEventGroup.Initialization:
					if ((this._eventGroup & HtmlEventGroup.NoInitialization) == 0)
					{
						HtmlInitializationEventParameters parameters =
							(HtmlInitializationEventParameters)Marshal.PtrToStructure(
								lpParameters,
								typeof(HtmlInitializationEventParameters));
						HtmlInitializationEventArgs eventArgs = new HtmlInitializationEventArgs(parameters);

						OnInitialization(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							return true;
						}
					}
					break;
				case HtmlEventGroup.InitializationAndMouse:
					if ((this._eventGroup & HtmlEventGroup.InitializationAndMouse) != 0)
					{
						HtmlMouseEventParameters parameters =
						(HtmlMouseEventParameters)Marshal.PtrToStructure(
							lpParameters,
							typeof(HtmlMouseEventParameters));
						HtmlMouseEventArgs eventArgs = new HtmlMouseEventArgs(parameters);

						OnMouse(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							Marshal.StructureToPtr(eventArgs.Parameters, lpParameters, true);

							return true;
						}
					}
					break;
				case HtmlEventGroup.InitializationAndKeyboard:
					if ((this._eventGroup & HtmlEventGroup.InitializationAndKeyboard) != 0)
					{
						HtmlKeyboardEventParameters parameters =
						(HtmlKeyboardEventParameters)Marshal.PtrToStructure(
							lpParameters,
							typeof(HtmlKeyboardEventParameters));
						HtmlKeyboardEventArgs eventArgs = new HtmlKeyboardEventArgs(parameters);

						OnKeyboard(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							Marshal.StructureToPtr(eventArgs.Parameters, lpParameters, true);

							return true;
						}
					}
					break;
				case HtmlEventGroup.InitializationAndFocus:
					if ((this._eventGroup & HtmlEventGroup.InitializationAndFocus) != 0)
					{
						HtmlFocusEventParameters parameters =
							(HtmlFocusEventParameters)Marshal.PtrToStructure(
								lpParameters,
								typeof(HtmlFocusEventParameters));
						HtmlFocusEventArgs eventArgs = new HtmlFocusEventArgs(parameters);

						OnFocus(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							Marshal.StructureToPtr(eventArgs.Parameters, lpParameters, true);

							return true;
						}
					}
					break;
				case HtmlEventGroup.InitializationAndScroll:
					if ((this._eventGroup & HtmlEventGroup.InitializationAndScroll) != 0)
					{
						HtmlScrollEventParameters parameters =
							(HtmlScrollEventParameters)Marshal.PtrToStructure(
								lpParameters,
								typeof(HtmlScrollEventParameters));
						HtmlScrollEventArgs eventArgs = new HtmlScrollEventArgs(parameters);

						OnScroll(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							Marshal.StructureToPtr(eventArgs.Parameters, lpParameters, true);

							return true;
						}
					}
					break;
				case HtmlEventGroup.InitializationAndTimer:
					if ((this._eventGroup & HtmlEventGroup.InitializationAndTimer) != 0)
					{
						HtmlTimerEventParameters parameters =
							(HtmlTimerEventParameters)Marshal.PtrToStructure(
								lpParameters,
								typeof(HtmlTimerEventParameters));
						HtmlTimerEventArgs eventArgs = new HtmlTimerEventArgs(parameters);

						OnTimer(sender, eventArgs);
					}
					break;
				case HtmlEventGroup.InitializationAndSize:
					if ((this._eventGroup & HtmlEventGroup.InitializationAndSize) != 0)
					{
						OnSize(sender, EventArgs.Empty);
					}
					break;
				case HtmlEventGroup.InitializationAndDraw:
					if ((this._eventGroup & HtmlEventGroup.InitializationAndDraw) != 0)
					{
						HtmlDrawEventParameters parameters =
							(HtmlDrawEventParameters)Marshal.PtrToStructure(
								lpParameters,
								typeof(HtmlDrawEventParameters));
						HtmlDrawEventArgs eventArgs = new HtmlDrawEventArgs(parameters);

						OnDraw(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							Marshal.StructureToPtr(eventArgs.Parameters, lpParameters, true);

							return true;
						}
					}
					break;
				case HtmlEventGroup.InitializationAndData:
					if ((this._eventGroup & HtmlEventGroup.InitializationAndData) != 0)
					{
						HtmlDataEventParameters parameters =
							(HtmlDataEventParameters)Marshal.PtrToStructure(
								lpParameters,
								typeof(HtmlDataEventParameters));
						HtmlDataEventArgs eventArgs = new HtmlDataEventArgs(parameters);

						OnData(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							Marshal.StructureToPtr(eventArgs.Parameters, lpParameters, true);

							return true;
						}
					}
					break;
				case HtmlEventGroup.InitializationAndBehavior:
					if ((this._eventGroup & HtmlEventGroup.InitializationAndBehavior) != 0)
					{
						HtmlBehaviorEventParameters parameters =
							(HtmlBehaviorEventParameters)Marshal.PtrToStructure(
								lpParameters,
								typeof(HtmlBehaviorEventParameters));
						HtmlBehaviorEventArgs eventArgs = new HtmlBehaviorEventArgs(parameters);

						OnBehavior(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							Marshal.StructureToPtr(eventArgs.Parameters, lpParameters, true);

							return true;
						}
					}
					break;
				case HtmlEventGroup.InitializationAndMethod:
					if ((this._eventGroup & HtmlEventGroup.InitializationAndMethod) != 0)
					{
						HtmlMethodEventParameters parameters =
							(HtmlMethodEventParameters)Marshal.PtrToStructure(
								lpParameters,
								typeof(HtmlMethodEventParameters));

						HtmlMethodEventArgs eventArgs =
							new HtmlMethodEventArgs(
								parameters.MethodID,
								lpParameters);

						OnMethod(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							return true;
						}
					}
					break;
			}

			return false;
		}

		#endregion

		#endregion

		#region Properties

		public HtmlEventGroup EventGroup
		{
			[DebuggerStepThrough]
			get { return this._eventGroup; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
