﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Drawing;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;

using Nabu.Collections;

namespace Nabu.Forms.Html
{
	public abstract class HtmlMethodEventSinkBase : HtmlEventSinkBase
	{
		#region HtmlMethodEventSinkBase

		#region Fields

		private IntPtr _lpStringBufferData;
		private int _stringBufferLength;

		#endregion

		#region Methods

		#region Utility

		private void SetStringBuffer(string value)
		{
			if (value == null)
			{
				value = string.Empty;
			}

			this._lpStringBufferData = Marshal.StringToBSTR(value);
			this._stringBufferLength = value.Length;
		}

		#endregion

		#region Event Handlers

		protected virtual void OnMethodClick(HtmlTag sender, EventArgs e)
		{
		}

		protected virtual void OnMethodGetText(HtmlTag sender, HandleMutableEventArgs<string> e)
		{
		}

		protected virtual void OnMethodSetText(HtmlTag sender, HandleEventArgs<string> e)
		{
		}

		protected virtual void OnMethodGetSelection(HtmlTag sender, HandleRangeMutableEventArgs<int> e)
		{
		}

		protected virtual void OnMethodSetSelection(HtmlTag sender, HandleRangeEventArgs<int> e)
		{
		}

		protected virtual void OnMethodReplaceSelection(HtmlTag sender, HandleEventArgs<string> e)
		{
		}

		protected virtual void OnMethodGetScrollValue(HtmlTag sender, HtmlMethodMutableScrollEventArgs e)
		{
		}

		protected virtual void OnMethodSetScrollValue(HtmlTag sender, HtmlMethodScrollEventArgs e)
		{
		}

		protected virtual void OnMethodGetCaretPosition(HtmlTag sender, HandleMutableEventArgs<Rectangle> e)
		{
		}

		protected virtual void OnMethodGetSelectedText(HtmlTag sender, HandleMutableEventArgs<string> e)
		{
		}

		protected virtual void OnMethodGetSelectedHtml(HtmlTag sender, HandleMutableEventArgs<string> e)
		{
		}

		protected virtual void OnMethodXCall(HtmlTag sender, HtmlMethodCallEventArgs e)
		{
			foreach (MethodInfo methodInfo in this.GetType().GetMethods())
			{
				// Cannot XCall abstract method, constructor, special method or generic method
				if (
					(!methodInfo.IsAbstract) &&
					(!methodInfo.IsConstructor) &&
					(!methodInfo.ContainsGenericParameters) &&
					(!methodInfo.IsSpecialName))
				{
					// Method signature must be 'object (object[])'
					if (methodInfo.ReturnType == typeof(object))
					{
						ParameterInfo[] parameters = methodInfo.GetParameters();

						if (
							(parameters.Length == 1) &&
#if !COMPACT_FRAMEWORK
 (!parameters[0].IsOut) &&
#endif
 (parameters[0].ParameterType == typeof(object[])))
						{
							HtmlXCallMethodAttribute[] attributes =
								(HtmlXCallMethodAttribute[])methodInfo.GetCustomAttributes(
									typeof(HtmlXCallMethodAttribute),
									true);

							foreach (HtmlXCallMethodAttribute attribute in attributes)
							{
								if (string.Compare(e.MethodName, attribute.Name, StringComparison.OrdinalIgnoreCase) == 0)
								{
									e.ReturnValue =
										methodInfo.Invoke(
											this,
											new object[] { e.Arguments });
									e.IsHandled = true;
									return;
								}
							}
						}
					}
				}
			}
		}

		protected virtual void OnMethodManagedCall(HtmlTag sender, HtmlMethodCallEventArgs e)
		{
			foreach (MethodInfo methodInfo in this.GetType().GetMethods())
			{
				// Cannot ManagedCall abstract method, constructor, special method or generic method
				if (
					(!methodInfo.IsAbstract) &&
					(!methodInfo.IsConstructor) &&
					(!methodInfo.ContainsGenericParameters) &&
					(!methodInfo.IsSpecialName))
				{
					ParameterInfo[] parameters = methodInfo.GetParameters();

					if (e.Arguments.Length == parameters.Length)
					{
						bool typeMatch = true;

						for (int index = 0; index < parameters.Length; index++)
						{
							if (!parameters[index].ParameterType.IsAssignableFrom(e.Arguments[index].GetType()))
							{
								typeMatch = false;
								break;
							}
						}

						if (typeMatch)
						{
							if (methodInfo.ReturnType == typeof(void))
							{
								methodInfo.Invoke(
									this,
									e.Arguments);
							}
							else
							{
								e.ReturnValue =
									methodInfo.Invoke(
										this,
										e.Arguments);
							}

							e.IsHandled = true;
							return;
						}
					}
				}
			}
		}

		#endregion

		#region Constructors

		protected HtmlMethodEventSinkBase(HtmlEventGroup eventGroup)
			: base(eventGroup)
		{
		}

		#endregion

		#endregion

		#endregion

		#region HtmlEventSinkBase

		protected override void OnMethod(HtmlTag sender, HtmlMethodEventArgs e)
		{
			switch (e.MethodID)
			{
				case HtmlTagMethodID.Click:
					{
						HandleEventArgs eventArgs = new HandleEventArgs();

						OnMethodClick(sender, eventArgs);

						eventArgs.IsHandled = e.IsHandled;
					}
					break;
				case HtmlTagMethodID.GetText:
					{
						HandleMutableEventArgs<string> eventArgs =
							new HandleMutableEventArgs<string>();

						OnMethodGetText(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							e.IsHandled = true;

							HtmlTextBoxValueMethodEventParameters parameters =
								new HtmlTextBoxValueMethodEventParameters();

							if (eventArgs.Value == null)
							{
								eventArgs.Value = string.Empty;
							}

							SetStringBuffer(eventArgs.Value);

							parameters.MethodID = e.MethodID;
							parameters.lpText = this._lpStringBufferData;
							parameters.Length = this._stringBufferLength;

							Marshal.StructureToPtr(parameters, e.Parameters, false);
						}
					}
					break;
				case HtmlTagMethodID.SetText:
					{
						HtmlTextBoxValueMethodEventParameters parameters =
							(HtmlTextBoxValueMethodEventParameters)Marshal.PtrToStructure(
								e.Parameters,
								typeof(HtmlTextBoxValueMethodEventParameters));

						HandleEventArgs<string> eventArgs =
							new HandleEventArgs<string>(
								Marshal.PtrToStringUni(
									parameters.lpText,
									parameters.Length));

						OnMethodSetText(sender, eventArgs);

						eventArgs.IsHandled = e.IsHandled;
					}
					break;
				case HtmlTagMethodID.GetSelection:
					{
						HandleRangeMutableEventArgs<int> eventArgs =
							new HandleRangeMutableEventArgs<int>();

						OnMethodGetSelection(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							e.IsHandled = true;

							HtmlTextBoxSelectionMethodEventParameters parameters =
								new HtmlTextBoxSelectionMethodEventParameters();

							parameters.MethodID = e.MethodID;
							parameters.SelectionStart = eventArgs.Offset;
							parameters.SelectionEnd = eventArgs.Offset + eventArgs.Length;

							Marshal.StructureToPtr(parameters, e.Parameters, false);
						}
					}
					break;
				case HtmlTagMethodID.SetSelection:
					{
						HtmlTextBoxSelectionMethodEventParameters parameters =
							(HtmlTextBoxSelectionMethodEventParameters)Marshal.PtrToStructure(
								e.Parameters,
								typeof(HtmlTextBoxSelectionMethodEventParameters));

						HandleRangeEventArgs<int> eventArgs =
							new HandleRangeEventArgs<int>(
								parameters.SelectionStart,
								parameters.SelectionEnd - parameters.SelectionStart);

						OnMethodSetSelection(sender, eventArgs);

						eventArgs.IsHandled = e.IsHandled;
					}
					break;
				case HtmlTagMethodID.ReplaceSelection:
					{
						HtmlTextBoxValueMethodEventParameters parameters =
							(HtmlTextBoxValueMethodEventParameters)Marshal.PtrToStructure(
								e.Parameters,
								typeof(HtmlTextBoxValueMethodEventParameters));

						HandleEventArgs<string> eventArgs =
							new HandleEventArgs<string>(
								Marshal.PtrToStringUni(
									parameters.lpText,
									parameters.Length));

						OnMethodReplaceSelection(sender, eventArgs);

						eventArgs.IsHandled = e.IsHandled;
					}
					break;
				case HtmlTagMethodID.GetScrollValue:
					{
						HtmlMethodMutableScrollEventArgs eventArgs =
							new HtmlMethodMutableScrollEventArgs();

						OnMethodGetScrollValue(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							e.IsHandled = true;

							HtmlScrollBarValueMethodEventParameters parameters =
								new HtmlScrollBarValueMethodEventParameters();

							parameters.MethodID = e.MethodID;
							parameters.Value = eventArgs.Value;
							parameters.MinValue = eventArgs.MinValue;
							parameters.MaxValue = eventArgs.MaxValue;
							parameters.PageValue = eventArgs.PageValue;
							parameters.StepValue = eventArgs.StepValue;

							Marshal.StructureToPtr(parameters, e.Parameters, false);
						}
					}
					break;
				case HtmlTagMethodID.SetScrollValue:
					{
						HtmlScrollBarValueMethodEventParameters parameters =
							(HtmlScrollBarValueMethodEventParameters)Marshal.PtrToStructure(
								e.Parameters,
								typeof(HtmlScrollBarValueMethodEventParameters));

						HtmlMethodScrollEventArgs eventArgs =
							new HtmlMethodScrollEventArgs(
								parameters.Value,
								parameters.MinValue,
								parameters.MaxValue,
								parameters.PageValue,
								parameters.StepValue);

						OnMethodSetScrollValue(sender, eventArgs);

						eventArgs.IsHandled = e.IsHandled;
					}
					break;
				case HtmlTagMethodID.GetCaretPosition:
					{
						HandleMutableEventArgs<Rectangle> eventArgs =
							new HandleMutableEventArgs<Rectangle>();

						OnMethodGetCaretPosition(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							e.IsHandled = true;

							HtmlTextBoxCaretPositionMethodEventParameters parameters =
								new HtmlTextBoxCaretPositionMethodEventParameters();

							parameters.MethodID = e.MethodID;
							parameters.Position = eventArgs.Value;

							Marshal.StructureToPtr(parameters, e.Parameters, false);
						}
					}
					break;
				case HtmlTagMethodID.GetSelectedText:
					{
						HandleMutableEventArgs<string> eventArgs =
							new HandleMutableEventArgs<string>();

						OnMethodGetSelectedText(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							e.IsHandled = true;

							HtmlTextSelectionMethodEventParameters parameters =
								(HtmlTextSelectionMethodEventParameters)Marshal.PtrToStructure(
									e.Parameters,
									typeof(HtmlTextSelectionMethodEventParameters));

							if (eventArgs.Value == null)
							{
								eventArgs.Value = string.Empty;
							}

#if COMPACT_FRAMEWORK
							foreach (char symbol in eventArgs.Value)
							{
								MarshalCF.CallUnmanagedStdCallFunction(
									parameters.lpWriter,
									e.Parameters,
									symbol);
							}
#else
							foreach (char symbol in eventArgs.Value)
							{
								parameters.Writer(e.Parameters, symbol);
							}
#endif
						}
					}
					break;
				case HtmlTagMethodID.GetSelectedHTML:
					{
						HandleMutableEventArgs<string> eventArgs =
							new HandleMutableEventArgs<string>();

						OnMethodGetSelectedHtml(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							e.IsHandled = true;

							HtmlTextSelectionMethodEventParameters parameters =
								(HtmlTextSelectionMethodEventParameters)Marshal.PtrToStructure(
									e.Parameters,
									typeof(HtmlTextSelectionMethodEventParameters));

							if (eventArgs.Value == null)
							{
								eventArgs.Value = string.Empty;
							}

							byte[] buffer = Encoding.UTF8.GetBytes(eventArgs.Value);

#if COMPACT_FRAMEWORK
							foreach (char value in buffer)
							{
								MarshalCF.CallUnmanagedStdCallFunction(
									parameters.lpWriter,
									e.Parameters,
									value);
							}
#else
							foreach (char value in buffer)
							{
								parameters.Writer(e.Parameters, value);
							}
#endif
						}
					}
					break;
				case HtmlTagMethodID.XCall:
					{
						HtmlXCallMethodEventParameters parameters =
							(HtmlXCallMethodEventParameters)Marshal.PtrToStructure(
								e.Parameters,
								typeof(HtmlXCallMethodEventParameters));

						ListEx<object> arguments = new ListEx<object>(parameters.ArgumentCount);
						int valueSize = Marshal.SizeOf(typeof(JsonValueData));

						for (int index = 0; index < parameters.ArgumentCount; index++)
						{
							JsonValueData value =
								(JsonValueData)Marshal.PtrToStructure(
									new IntPtr(parameters.lpArgumentValues.ToInt64() + index * valueSize),
									typeof(JsonValueData));

							arguments.Add(JsonValue.GetObjectFromValue(value));
						}

						HtmlMethodCallEventArgs eventArgs =
							new HtmlMethodCallEventArgs(
#if COMPACT_FRAMEWORK
								MarshalCF.PtrToStringAnsi(parameters.lpMethodName),
#else
Marshal.PtrToStringAnsi(parameters.lpMethodName),
#endif
 arguments.ToArray());

						OnMethodXCall(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							e.IsHandled = true;

							parameters.ReturnValue = JsonValue.GetValueFromObject(eventArgs.ReturnValue);

							Marshal.StructureToPtr(parameters, e.Parameters, false);
						}
					}
					break;
				case HtmlTagMethodID.ManagedCall:
					{
						HtmlManagedCallMethodEventParameters parameters =
							(HtmlManagedCallMethodEventParameters)Marshal.PtrToStructure(
								e.Parameters,
								typeof(HtmlManagedCallMethodEventParameters));

						ListEx<object> arguments = new ListEx<object>(parameters.ArgumentCount);

						for (int index = 0; index < parameters.ArgumentCount; index++)
						{
							GCHandle argumentHandle =
								(GCHandle)Marshal.ReadIntPtr(
									new IntPtr(parameters.lpArgumentValues.ToInt64() + IntPtr.Size * index));

							arguments.Add(argumentHandle.Target);
						}

						HtmlMethodCallEventArgs eventArgs =
							new HtmlMethodCallEventArgs(
								parameters.MethodName,
								arguments.ToArray());

						OnMethodManagedCall(sender, eventArgs);

						if (eventArgs.IsHandled)
						{
							e.IsHandled = true;

							parameters.ReturnValue =
								(IntPtr)GCHandle.Alloc(
									eventArgs.ReturnValue,
									GCHandleType.Normal);

							Marshal.StructureToPtr(parameters, e.Parameters, false);
						}
					}
					break;
			}
		}

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			Marshal.FreeBSTR(this._lpStringBufferData);
			this._lpStringBufferData = IntPtr.Zero;

			base.Dispose(disposing);
		}

		#endregion
	};
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
