﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Globalization;

namespace Nabu.Forms.Html
{
	public delegate byte[] ObjectLoadFailureHandler(
		Uri resourceUrl,
		HtmlResourceType resourceUriType,
		Exception exception);
	public delegate byte[] ProtocolLoadFailureHandler(
		string resourceProtocol,
		HtmlResourceType resourceUriType);

	public static class HtmlProtocolManager
	{
		#region Fields

		private static Dictionary<string, HtmlProtocolBase> _protocols = new Dictionary<string, HtmlProtocolBase>(8);
		private static Dictionary<string, string> _urlRedirection = new Dictionary<string, string>(8);
		private static ObjectLoadFailureHandler _objectLoadFailureHandler;
		private static ProtocolLoadFailureHandler _protocolLoadFailureHandler;

		#endregion

		#region Methods

		#region Constructors

		static HtmlProtocolManager()
		{
#if COMPACT_FRAMEWORK
			RegisterProtocol(new HtmlProtocolConst());
			RegisterProtocol(new HtmlProtocolHttp());
			RegisterProtocol(new HtmlProtocolFile());
			RegisterProtocol(new HtmlProtocolResX());
#else
			RegisterProtocol(new HtmlProtocolConst());
			RegisterProtocol(new HtmlProtocolHttp());
			RegisterProtocol(new HtmlProtocolFtp());
			RegisterProtocol(new HtmlProtocolFile());
			RegisterProtocol(new HtmlProtocolRes());
			RegisterProtocol(new HtmlProtocolResX());
#endif
			_urlRedirection["res:activity.gif"] =
				"res:HTMLayout.dll#GIF/ACTIVITY";
			_urlRedirection["res:edit-copy.png"] =
				"res:HTMLayout.dll#PNG/EDIT-COPY";
			_urlRedirection["res:edit-cut.png"] =
				"res:HTMLayout.dll#PNG/EDIT-CUT";
			_urlRedirection["res:edit-paste.png"] =
				"res:HTMLayout.dll#PNG/EDIT-PASTE";
			_urlRedirection["res:edit-undo.png"] =
				"res:HTMLayout.dll#PNG/EDIT-UNDO";
			_urlRedirection["res:edit-redo.png"] =
				"res:HTMLayout.dll#PNG/EDIT-REDO";
			_urlRedirection["res:behavior-edit-menu.htm"] =
				"res:HTMLayout.dll#RT_HTML/BEHAVIOR-EDIT-MENU";
			_urlRedirection["res:behavior-richtext-menu.htm"] =
				"res:HTMLayout.dll#RT_HTML/BEHAVIOR-RICHTEXT-MENU";
			_urlRedirection["res:behavior-text-menu.htm"] =
				"res:HTMLayout.dll#RT_HTML/BEHAVIOR-TEXT-MENU";
		}

		#endregion

		#region Fail Resources

		public static byte[] OnObjectLoadFailure(
			Uri resourceUrl,
			HtmlResourceType resourceUriType,
			Exception exception)
		{
			if (_objectLoadFailureHandler != null)
			{
				return _objectLoadFailureHandler(resourceUrl, resourceUriType, exception);
			}
			else
			{
#if !COMPACT_FRAMEWORK
				switch (resourceUriType)
				{
					case HtmlResourceType.Html:
						return HtmlProtocolBase.BinaryDataFromString(
							string.Format(
								CultureInfo.InvariantCulture,
								Properties.Resources.FailureObjectHtml,
								resourceUrl,
								exception.Message));
					case HtmlResourceType.StyleSheet:
						return HtmlProtocolBase.BinaryDataFromString(Properties.Resources.FailureObjectCss);
					case HtmlResourceType.Image:
						return Properties.Resources.FailureObjectBitmap;
				}
#endif
			}

			return new byte[0];
		}

		public static byte[] OnProtocolLoadFailure(
			string resourceProtocol,
			HtmlResourceType resourceUriType)
		{
			if (_protocolLoadFailureHandler != null)
			{
				return _protocolLoadFailureHandler(resourceProtocol, resourceUriType);
			}
			else
			{
#if !COMPACT_FRAMEWORK
				switch (resourceUriType)
				{
					case HtmlResourceType.Html:
						return HtmlProtocolBase.BinaryDataFromString(
							string.Format(
								CultureInfo.InvariantCulture,
								Properties.Resources.FailureProtocolHtml,
								resourceProtocol));
					case HtmlResourceType.StyleSheet:
						return HtmlProtocolBase.BinaryDataFromString(Properties.Resources.FailureProtocolCss);
					case HtmlResourceType.Image:
						return Properties.Resources.FailureProtocolBitmap;
				}
#endif
			}

			return new byte[0];
		}

		#endregion

		#region Protocols

		public static void RegisterProtocol(HtmlProtocolBase htmlProtocol)
		{
			foreach (string htmlProtocolName in htmlProtocol.Names)
			{
				_protocols.Add(htmlProtocolName, htmlProtocol);
			}
		}

		public static void UnregisterProtocol(HtmlProtocolBase htmlProtocol)
		{
			foreach (string htmlProtocolName in htmlProtocol.Names)
			{
				_protocols.Remove(htmlProtocolName);
			}
		}

		public static HtmlProtocolBase GetProtocol(string htmlProtocolName)
		{
			HtmlProtocolBase protocol;

			if (_protocols.TryGetValue(htmlProtocolName, out protocol))
			{
				return protocol;
			}
			else
			{
				return null;
			}
		}

		#endregion

		#region URL Redirection

		[SuppressMessage("Microsoft.Design", "CA1055:UriReturnValuesShouldNotBeStrings")]
		[SuppressMessage("Microsoft.Design", "CA1054:UriParametersShouldNotBeStrings", MessageId = "0#")]
		[SuppressMessage("Microsoft.Design", "CA1054:UriParametersShouldNotBeStrings", MessageId = "1#")]
		public static void SetUrlRedirection(string urlFrom, string urlTo)
		{
			if ((urlTo == null) || (urlFrom == urlTo))
			{
				_urlRedirection.Remove(urlFrom);
			}
			else
			{
				_urlRedirection[urlFrom] = urlTo;
			}
		}

		[SuppressMessage("Microsoft.Design", "CA1055:UriReturnValuesShouldNotBeStrings")]
		public static Uri GetUrlRedirection(Uri urlFrom)
		{
			string pathTo = null;

			if (_urlRedirection.TryGetValue(urlFrom.AbsoluteUri, out pathTo))
			{
				return new Uri(pathTo);
			}
			else
			{
				return urlFrom;
			}
		}

		#endregion

		#endregion

		#region Properties

		public static ObjectLoadFailureHandler ObjectLoadFailure
		{
			get { return _objectLoadFailureHandler; }
			set { _objectLoadFailureHandler = value; }
		}

		public static ProtocolLoadFailureHandler ProtocolLoadFailure
		{
			get { return _protocolLoadFailureHandler; }
			set { _protocolLoadFailureHandler = value; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
