﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics;
using System.Drawing;
using System.Globalization;
using System.Resources;
using System.Runtime.InteropServices;
using System.Text;
using System.Text.RegularExpressions;

namespace Nabu.Forms.Html
{
	public sealed class HtmlProtocolRes : HtmlProtocolBase
	{
		#region HtmlProtocolRes

		#region Constants

		private readonly string[] _names = new string[] { "res" };

		#endregion

		#region Fields

		private Regex _regex = new Regex(@"#([\w\d]+)/([\w\d\-\.]+)(\:[\w\d]+|)", RegexOptions.Compiled);

		#endregion

		#region Methods

		#region IDFromName

		private static short IDFromName(
			string name)
		{
			short id;

			if (name.StartsWith("0x", StringComparison.OrdinalIgnoreCase))
			{
#if COMPACT_FRAMEWORK
				id = short.Parse(name, NumberStyles.AllowHexSpecifier, NumberFormatInfo.CurrentInfo);
#else
				if (!short.TryParse(name, NumberStyles.AllowHexSpecifier, NumberFormatInfo.CurrentInfo, out id))
				{
					id = 0;
				}
#endif
			}
			else
			{
#if COMPACT_FRAMEWORK
				id = short.Parse(name, NumberStyles.None, NumberFormatInfo.CurrentInfo);
#else
				if (!short.TryParse(name, NumberStyles.None, NumberFormatInfo.CurrentInfo, out id))
				{
					id = 0;
				}
#endif
			}

			return id;
		}

		#endregion

		#region LoadIcon

		private static byte[] LoadIcon(IntPtr hModule, short objectID, short localeID)
		{
			IntPtr hImage = IntPtr.Zero;

			try
			{
				hImage = NativeMethods.Image_Load(hModule, new IntPtr(objectID), ImageType.Icon, 0, 0, 0);

				using (Image image = Image.FromHbitmap(hImage))
				{
					return BinaryDataFromImage(image);
				}
			}
			finally
			{
				NativeMethods.Icon_Delete(hImage);
			}
		}

		private static byte[] LoadIcon(IntPtr hModule, string objectID, short localeID)
		{
			IntPtr hImage = IntPtr.Zero;

			try
			{
				hImage = NativeMethods.Image_Load(hModule, objectID, ImageType.Icon, 0, 0, 0);

				using (Image image = Image.FromHbitmap(hImage))
				{
					return BinaryDataFromImage(image);
				}
			}
			finally
			{
				NativeMethods.Icon_Delete(hImage);
			}
		}

		#endregion

		#region LoadCursor

		private static byte[] LoadCursor(IntPtr hModule, short objectID, short localeID)
		{
			IntPtr hImage = IntPtr.Zero;

			try
			{
				hImage = NativeMethods.Image_Load(hModule, new IntPtr(objectID), ImageType.Cursor, 0, 0, 0);

				using (Image image = Image.FromHbitmap(hImage))
				{
					return BinaryDataFromImage(image);
				}
			}
			finally
			{
				NativeMethods.Icon_Delete(hImage);
			}
		}

		private static byte[] LoadCursor(IntPtr hModule, string objectID, short localeID)
		{
			IntPtr hImage = IntPtr.Zero;

			try
			{
				hImage = NativeMethods.Image_Load(hModule, objectID, ImageType.Cursor, 0, 0, 0);

				using (Image image = Image.FromHbitmap(hImage))
				{
					return BinaryDataFromImage(image);
				}
			}
			finally
			{
				NativeMethods.Icon_Delete(hImage);
			}
		}

		#endregion

		#region LoadBitmap

		private static byte[] LoadBitmap(IntPtr hModule, short objectID, short localeID)
		{
			IntPtr hImage = IntPtr.Zero;

			try
			{
				hImage = NativeMethods.Image_Load(hModule, new IntPtr(objectID), ImageType.Bitmap, 0, 0, 0);

				using (Image image = Image.FromHbitmap(hImage))
				{
					return BinaryDataFromImage(image);
				}
			}
			finally
			{
				NativeMethods.Bitmap_Delete(hImage);
			}
		}

		private static byte[] LoadBitmap(IntPtr hModule, string objectID, short localeID)
		{
			IntPtr hImage = IntPtr.Zero;

			try
			{
				hImage = NativeMethods.Image_Load(hModule, objectID, ImageType.Bitmap, 0, 0, 0);

				using (Image image = Image.FromHbitmap(hImage))
				{
					return BinaryDataFromImage(image);
				}
			}
			finally
			{
				NativeMethods.Bitmap_Delete(hImage);
			}
		}

		#endregion

		#region LoadString

		private static byte[] LoadString(IntPtr hModule, short objectID, short localeID)
		{
			StringBuilder buffer = new StringBuilder(1024);

			if (NativeMethods.String_Load(hModule, objectID, buffer, buffer.Capacity) == 0)
			{
				return new byte[0];
			}
			else
			{
				return Encoding.GetEncoding(localeID).GetBytes(buffer.ToString());
			}
		}

		#endregion

		#region LoadCustom

		private static byte[] LoadCustom(IntPtr hModule, ResourceType resourceTypeID, short objectID, short localeID)
		{
			IntPtr hResourceInfo =
				NativeMethods.Resource_Find(
					hModule,
					(IntPtr)resourceTypeID,
					objectID,
					localeID);

			if (hResourceInfo != IntPtr.Zero)
			{
				IntPtr hResource = NativeMethods.Resource_Load(hModule, hResourceInfo);

				if (hResource != IntPtr.Zero)
				{
					IntPtr resourceAddress = NativeMethods.Resource_Lock(hResource);
					int resourceSize = NativeMethods.Resource_Size(hModule, hResourceInfo);
					byte[] resourceBuffer = new byte[resourceSize];

					Marshal.Copy(resourceAddress, resourceBuffer, 0, resourceBuffer.Length);

					return resourceBuffer;
				}
			}

			return new byte[0];
		}

		private static byte[] LoadCustom(IntPtr hModule, ResourceType resourceTypeID, string objectName, short localeID)
		{
			IntPtr hResourceInfo =
				NativeMethods.Resource_Find(
					hModule,
					(IntPtr)resourceTypeID,
					objectName,
					localeID);

			if (hResourceInfo != IntPtr.Zero)
			{
				IntPtr hResource = NativeMethods.Resource_Load(hModule, hResourceInfo);

				if (hResource != IntPtr.Zero)
				{
					IntPtr resourceAddress = NativeMethods.Resource_Lock(hResource);
					int resourceSize = NativeMethods.Resource_Size(hModule, hResourceInfo);
					byte[] resourceBuffer = new byte[resourceSize];

					Marshal.Copy(resourceAddress, resourceBuffer, 0, resourceBuffer.Length);

					return resourceBuffer;
				}
			}

			return new byte[0];
		}

		private static byte[] LoadCustom(IntPtr hModule, string resourceTypeName, short objectID, short localeID)
		{
			IntPtr hResourceInfo = NativeMethods.Resource_Find(hModule, resourceTypeName, objectID, localeID);

			if (hResourceInfo != IntPtr.Zero)
			{
				IntPtr hResource = NativeMethods.Resource_Load(hModule, hResourceInfo);

				if (hResource != IntPtr.Zero)
				{
					IntPtr resourceAddress = NativeMethods.Resource_Lock(hResource);
					int resourceSize = NativeMethods.Resource_Size(hModule, hResourceInfo);
					byte[] resourceBuffer = new byte[resourceSize];

					Marshal.Copy(resourceAddress, resourceBuffer, 0, resourceBuffer.Length);

					return resourceBuffer;
				}
			}

			return new byte[0];
		}

		private static byte[] LoadCustom(IntPtr hModule, string resourceTypeName, string objectName, short localeID)
		{
			IntPtr hResourceInfo = NativeMethods.Resource_Find(hModule, resourceTypeName, objectName, localeID);

			if (hResourceInfo != IntPtr.Zero)
			{
				IntPtr hResource = NativeMethods.Resource_Load(hModule, hResourceInfo);

				if (hResource != IntPtr.Zero)
				{
					IntPtr resourceAddress = NativeMethods.Resource_Lock(hResource);
					int resourceSize = NativeMethods.Resource_Size(hModule, hResourceInfo);
					byte[] resourceBuffer = new byte[resourceSize];

					Marshal.Copy(resourceAddress, resourceBuffer, 0, resourceBuffer.Length);

					return resourceBuffer;
				}
			}

			return new byte[0];
		}

		#endregion

		#endregion

		#endregion

		#region IHtmlProtocol

		#region Methods

		public override byte[] LoadResource(
			Uri resourceUri,
			HtmlResourceType resourceUriType)
		{
			Match matchResource = _regex.Match(resourceUri.Fragment);

			if (matchResource.Success)
			{
				string resourceFileName = resourceUri.LocalPath;
				string resourceTypeName = matchResource.Groups[1].Value;
				string resourceObjectName = matchResource.Groups[2].Value;
				string resourceLocaleName = (matchResource.Groups[3].Length > 0) ? matchResource.Groups[3].Value.Substring(1) : string.Empty;

				IntPtr hModule = NativeMethods.Module_GetHandle(resourceFileName);

				if (hModule == IntPtr.Zero)
				{
					hModule = NativeMethods.Module_Load(resourceFileName, IntPtr.Zero, 0x00000002 /* LOAD_LIBRARY_AS_DATAFILE */);
				}

				if (hModule != IntPtr.Zero)
				{
					short resourceObjectID = IDFromName(resourceObjectName);
					short resourceLocaleID = IDFromName(resourceLocaleName);
					ResourceType resourceTypeID = ResourceType.Unknown;

					switch (resourceTypeName)
					{
						case "RT_CURSOR":
							resourceTypeID = ResourceType.Cursor;
							break;
						case "RT_BITMAP":
							resourceTypeID = ResourceType.Bitmap;
							break;
						case "RT_ICON":
							resourceTypeID = ResourceType.Icon;
							break;
						case "RT_STRING":
							resourceTypeID = ResourceType.String;
							break;
						case "RT_ANICURSOR":
							resourceTypeID = ResourceType.AniCursor;
							break;
						case "RT_ANIICON":
							resourceTypeID = ResourceType.AniIcon;
							break;
						case "RT_HTML":
							resourceTypeID = ResourceType.Html;
							break;
						case "RT_FILE":
							resourceTypeID = ResourceType.File;
							break;
						default:
							resourceTypeID = ResourceType.Unknown;
							break;
					}

#if COMPACT_FRAMEWORK
					short[] localeIDs = (resourceLocaleID == 0) ?
						new short[] {
							(short)CultureInfo.CurrentUICulture.LCID,
							(short)CultureInfo.InvariantCulture.LCID,
							0x0409
						}
						:
						new short[]
						{
							resourceLocaleID,
							(short)CultureInfo.CurrentUICulture.LCID,
							(short)CultureInfo.InvariantCulture.LCID,
							0x0409
						};
#else
					short[] localeIDs = (resourceLocaleID == 0) ?
						new short[]
						{
							(short)CultureInfo.CurrentUICulture.LCID,
							(short)CultureInfo.InstalledUICulture.LCID,
							(short)CultureInfo.InvariantCulture.LCID,
							0x0409
						}
						:
						new short[]
						{
							resourceLocaleID,
							(short)CultureInfo.CurrentUICulture.LCID,
							(short)CultureInfo.InstalledUICulture.LCID,
							(short)CultureInfo.InvariantCulture.LCID,
							0x0409
						};
#endif

					foreach (short localeID in localeIDs)
					{
						byte[] resourceData = null;

						if (resourceObjectID != 0)
						{
							switch (resourceTypeID)
							{
								case ResourceType.Cursor:
								case ResourceType.AniCursor:
									resourceData = LoadCursor(hModule, resourceObjectID, localeID);
									break;
								case ResourceType.Icon:
								case ResourceType.AniIcon:
									resourceData = LoadIcon(hModule, resourceObjectID, localeID);
									break;
								case ResourceType.Bitmap:
									resourceData = LoadBitmap(hModule, resourceObjectID, localeID);
									break;
								case ResourceType.String:
									resourceData = LoadString(hModule, resourceObjectID, localeID);
									break;
								case ResourceType.Html:
								case ResourceType.File:
									resourceData = LoadCustom(hModule, resourceTypeID, resourceObjectID, localeID);
									break;
								case ResourceType.Unknown:
									resourceData = LoadCustom(hModule, resourceTypeName, resourceObjectID, localeID);
									break;
							}
						}
						else
						{
							switch (resourceTypeID)
							{
								case ResourceType.Cursor:
								case ResourceType.AniCursor:
									resourceData = LoadCursor(hModule, resourceObjectName, localeID);
									break;
								case ResourceType.Icon:
								case ResourceType.AniIcon:
									resourceData = LoadIcon(hModule, resourceObjectName, localeID);
									break;
								case ResourceType.Bitmap:
									resourceData = LoadBitmap(hModule, resourceObjectName, localeID);
									break;
								case ResourceType.String:
									//resourceBuffer = LoadString(hModule, resourceObjectName, localeID);
									break;
								case ResourceType.Html:
								case ResourceType.File:
									resourceData = LoadCustom(hModule, resourceTypeID, resourceObjectName, localeID);
									break;
								case ResourceType.Unknown:
									resourceData = LoadCustom(hModule, resourceTypeName, resourceObjectName, localeID);
									break;
							}
						}

						if ((resourceData != null) && (resourceData.Length != 0))
						{
							return resourceData;
						}
					}
				}
			}

			throw new MissingManifestResourceException();
		}

		#endregion

		#region Properties

		public override string[] Names
		{
			[DebuggerStepThrough]
			get { return this._names; }
		}

		#endregion

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
