﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections;
using System.Collections.Generic;
using System.Runtime.InteropServices;

namespace Nabu.Forms.Html
{
	internal static class JsonValue
	{
		#region Constants

		private const string JsonTypeExceptionMessage = "Unsupported JsonValueType.";
		private const string ManagedTypeExceptionMessage = "Unsupported m,anaged type.";

		#endregion

		#region Methods

		public static object GetObjectFromValue(JsonValueData value)
		{
			object result = null;

			JsonValueType valueType;
			JsonValueUnit valueUnit;

			HtmlValueException.ThrowOnError(
				HtmlDynamicInterop.Json.GetType(
					ref value,
					out valueType,
					out valueUnit));

			switch (valueType)
			{
				case JsonValueType.Undefined:
					break;
				case JsonValueType.Null:
					break;
				case JsonValueType.Boolean:
					{
						int typedValue;

						HtmlValueException.ThrowOnError(
							HtmlDynamicInterop.Json.GetIntData(
								ref value,
								out typedValue));

						result = (typedValue != 0);
					}
					break;
				case JsonValueType.Integer:
					{
						int typedValue;

						HtmlValueException.ThrowOnError(
							HtmlDynamicInterop.Json.GetIntData(
								ref value,
								out typedValue));

						result = typedValue;
					}
					break;
				case JsonValueType.Floating:
					{
						double typedValue;

						HtmlValueException.ThrowOnError(
							HtmlDynamicInterop.Json.GetFloatData(
								ref value,
								out typedValue));

						result = typedValue;
					}
					break;
				case JsonValueType.String:
					{
						IntPtr typedValueData;
						int typedValueSize;

						HtmlValueException.ThrowOnError(
							HtmlDynamicInterop.Json.GetStringData(
								ref value,
								out typedValueData,
								out typedValueSize));

						result = Marshal.PtrToStringUni(typedValueData, typedValueSize);
					}
					break;
				case JsonValueType.FileTime:
					{
						long typedValue;

						HtmlValueException.ThrowOnError(
							HtmlDynamicInterop.Json.GetLongData(
								ref value,
								out typedValue));

						result = DateTime.FromFileTime(typedValue);
					}
					break;
				case JsonValueType.Currency:
					{
						long intermediateValue;

						HtmlValueException.ThrowOnError(
							HtmlDynamicInterop.Json.GetLongData(
								ref value,
								out intermediateValue));

						decimal typedValue = (decimal)intermediateValue;

						result = typedValue / 10000;
					}
					break;
				//case JsonValueType.Length:
				//	break;
				case JsonValueType.Array:
					{
						int elementCount;

						HtmlValueException.ThrowOnError(
							HtmlDynamicInterop.Json.GetElementCount(
								ref value,
								out elementCount));

						object[] array = new object[elementCount];

						for (int index = 0; index < elementCount; index++)
						{
							JsonValueData elementValue = new JsonValueData();

							try
							{
								HtmlValueException.ThrowOnError(
									HtmlDynamicInterop.Json.GetElement(
										ref value,
										index,
										out elementValue));

								array[index] = GetObjectFromValue(elementValue);
							}
							finally
							{
								HtmlValueException.ThrowOnError(
									HtmlDynamicInterop.Json.Clear(
										ref elementValue));
							}
						}

						result = array;
					}
					break;
				case JsonValueType.Map:
					{
						int elementCount;

						HtmlValueException.ThrowOnError(
							HtmlDynamicInterop.Json.GetElementCount(
								ref value,
								out elementCount));

						IDictionary<object, object> map = new Dictionary<object, object>(elementCount);

						for (int index = 0; index < elementCount; index++)
						{
							JsonValueData elementKey = new JsonValueData();
							JsonValueData elementValue = new JsonValueData();

							try
							{
								HtmlValueException.ThrowOnError(
									HtmlDynamicInterop.Json.GetElementKey(
										ref value,
										index,
										out elementKey));

								HtmlValueException.ThrowOnError(
									HtmlDynamicInterop.Json.GetElement(
										ref value,
										index,
										out elementValue));

								map.Add(
									GetObjectFromValue(elementKey),
									GetObjectFromValue(elementValue));
							}
							finally
							{
								HtmlValueException.ThrowOnError(
									HtmlDynamicInterop.Json.Clear(
										ref elementKey));
								HtmlValueException.ThrowOnError(
									HtmlDynamicInterop.Json.Clear(
										ref elementValue));
							}
						}

						result = map;
					}
					break;
				case JsonValueType.Binary:
					{
						IntPtr intermediateValueData;
						int intermediateValueSize;

						HtmlValueException.ThrowOnError(
							HtmlDynamicInterop.Json.GetBinaryData(
								ref value,
								out intermediateValueData,
								out intermediateValueSize));

						byte[] typedValue = new byte[intermediateValueSize];

						Marshal.Copy(intermediateValueData, typedValue, 0, intermediateValueSize);

						result = typedValue;
					}
					break;
				default:
					throw new NotImplementedException(JsonTypeExceptionMessage);
			}

			return result;
		}

		public static JsonValueData GetValueFromObject(object value)
		{
			JsonValueData result = new JsonValueData();

			if (value != null)
			{
				Type type = value.GetType();

				if (type == typeof(bool))
				{
					HtmlValueException.ThrowOnError(
						HtmlDynamicInterop.Json.Init(
							out result));

					HtmlValueException.ThrowOnError(
						HtmlDynamicInterop.Json.SetIntData(
							ref result,
							((bool)value) ? 1 : 0,
							JsonValueType.Boolean,
							JsonValueUnit.Undefined));
				}
				else if (type == typeof(int))
				{
					HtmlValueException.ThrowOnError(
						HtmlDynamicInterop.Json.Init(
							out result));

					HtmlValueException.ThrowOnError(
						HtmlDynamicInterop.Json.SetIntData(
							ref result,
							(int)value,
							JsonValueType.Integer,
							JsonValueUnit.Undefined));
				}
				else if (type == typeof(double))
				{
					HtmlValueException.ThrowOnError(
						HtmlDynamicInterop.Json.Init(
							out result));

					HtmlValueException.ThrowOnError(
						HtmlDynamicInterop.Json.SetFloatData(
							ref result,
							(double)value,
							JsonValueType.Floating,
							JsonValueUnit.Undefined));
				}
				else if (type == typeof(string))
				{
					HtmlValueException.ThrowOnError(
						HtmlDynamicInterop.Json.Init(
							out result));

					string stringValue = (string)value;

					HtmlValueException.ThrowOnError(
						HtmlDynamicInterop.Json.SetStringData(
							ref result,
							stringValue,
							stringValue.Length,
							JsonValueUnit.Undefined));
				}
				else if (type == typeof(DateTime))
				{
					HtmlValueException.ThrowOnError(
						HtmlDynamicInterop.Json.Init(
							out result));

					HtmlValueException.ThrowOnError(
						HtmlDynamicInterop.Json.SetLongData(
							ref result,
							((DateTime)value).ToFileTime(),
							JsonValueType.FileTime,
							JsonValueUnit.Undefined));
				}
				else if (type == typeof(decimal))
				{
					HtmlValueException.ThrowOnError(
						HtmlDynamicInterop.Json.Init(
							out result));

					HtmlValueException.ThrowOnError(
						HtmlDynamicInterop.Json.SetLongData(
							ref result,
							(long)(10000 * (decimal)value),
							JsonValueType.Currency,
							JsonValueUnit.Undefined));
				}
				else if (typeof(IDictionary).IsAssignableFrom(type))
				{
					IDictionary typedValue = (IDictionary)value;

					HtmlValueException.ThrowOnError(
						HtmlDynamicInterop.Json.Init(
							out result));

					foreach (DictionaryEntry element in typedValue)
					{
						JsonValueData keyData = GetValueFromObject(element.Key);
						JsonValueData valueData = GetValueFromObject(element.Value);

						HtmlValueException.ThrowOnError(
							HtmlDynamicInterop.Json.SetElementValue(
								ref result,
								ref keyData,
								ref valueData));
					}
				}
				else if (typeof(IEnumerable).IsAssignableFrom(type))
				{
					if (type == typeof(byte[]))
					{
						byte[] typedValue = (byte[])value;

						HtmlValueException.ThrowOnError(
							HtmlDynamicInterop.Json.Init(
								out result));

						HtmlValueException.ThrowOnError(
							HtmlDynamicInterop.Json.SetBinaryData(
								ref result,
								typedValue,
								typedValue.Length,
								JsonValueType.Binary,
								JsonValueUnit.Undefined));
					}
					else
					{
						IEnumerable typedValue = (IEnumerable)value;

						HtmlValueException.ThrowOnError(
							HtmlDynamicInterop.Json.Init(
								out result));

						int index = 0;

						foreach (object element in typedValue)
						{
							JsonValueData elementData = GetValueFromObject(element);

							HtmlValueException.ThrowOnError(
								HtmlDynamicInterop.Json.SetElement(
									ref result,
									index,
									ref elementData));
							index++;
						}
					}
				}
				else
				{
					throw new NotImplementedException(ManagedTypeExceptionMessage);
				}
			}

			return result;
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
