﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Drawing;
using System.Runtime.InteropServices;

namespace Nabu.Forms.Html
{
	public class HtmlPrinter :
		DisposableBase,
		IHtmlHost
	{
		#region HtmlPrinter

		#region Fields

		private IHtmlHostNotificationHandler _notificationHandler;
		private IntPtr _hPrint;
		private GCHandle _gcHandle;
		private Uri _baseUrl;
		private string _html = "<html><head></head><body></body></html>";
		private int _pageNumber;
		private int _pageIndex;
		private HtmlPrintLoadDataHandler _loadDataHandler;

		#endregion

		#region Methods

		#region Callbacks

		private bool LoadDataCallback(
			IntPtr hPrint,
			IntPtr hTag,
			string resourceUrl)
		{
			base.CheckDisposed();

			byte[] resourceData;

			if (_notificationHandler.OnResourceLoading(
				new Uri(resourceUrl),
				HtmlResourceType.Unknown,
				out resourceData))
			{
				HtmlPrintException.ThrowOnError(
					HtmlDynamicInterop.Print.ResourceReady(
						hPrint,
						resourceUrl,
						resourceData,
						resourceData.Length));

				return true;
			}
			else
			{
				return false;
			}
		}

		#endregion

		#region Utility

		private void UpdateHtml(byte[] htmlData)
		{
			OnDocumentResetting(EventArgs.Empty);
			OnDocumentReseted(EventArgs.Empty);

			if (this._baseUrl != null)
			{
				HtmlPrintException.ThrowOnError(
					HtmlDynamicInterop.Print.LoadHtml(
						this._hPrint,
						this._baseUrl.AbsolutePath,
						htmlData,
						htmlData.Length));
			}
			else
			{
				HtmlPrintException.ThrowOnError(
					HtmlDynamicInterop.Print.LoadHtml(
						this._hPrint,
						null,
						htmlData,
						htmlData.Length));
			}

			HtmlPrintException.ThrowOnError(
				HtmlDynamicInterop.Print.SetMediaType(
					this._hPrint,
					"print"));
		}

		#endregion

		#region Events Raisers

		protected virtual void OnDocumentResetting(EventArgs e)
		{
			if (this.DocumentResetting != null)
			{
				this.DocumentResetting(this, e);
			}
		}

		protected virtual void OnDocumentReseted(EventArgs e)
		{
			if (this.DocumentReseted != null)
			{
				this.DocumentReseted(this, e);
			}
		}

		protected virtual void OnHtmlHandleCreating(EventArgs e)
		{
			if (this.HtmlHandleCreating != null)
			{
				this.HtmlHandleCreating(this, e);
			}
		}

		protected virtual void OnHtmlHandleCreated(EventArgs e)
		{
			if (this.HtmlHandleCreated != null)
			{
				this.HtmlHandleCreated(this, e);
			}
		}

		protected virtual void OnHtmlHandleDestroying(EventArgs e)
		{
			if (this.HtmlHandleDestroying != null)
			{
				this.HtmlHandleDestroying(this, e);
			}
		}

		protected virtual void OnHtmlHandleDestroyed(EventArgs e)
		{
			if (this.HtmlHandleDestroyed != null)
			{
				this.HtmlHandleDestroyed(this, e);
			}
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Creates new HtmlPrinter object.
		/// </summary>
		public HtmlPrinter()
		{
			this._notificationHandler = new HtmlPrinterNotificationHandler();
			this._hPrint = HtmlDynamicInterop.Print.CreateInstance();
			this._gcHandle = GCHandle.Alloc(this, GCHandleType.Normal);
			this._pageNumber = -1;
			this._pageIndex = -1;
			this._loadDataHandler = LoadDataCallback;

			HtmlPrintException.ThrowOnError(
				HtmlDynamicInterop.Print.SetMediaType(
					this._hPrint,
					"print"));
			HtmlPrintException.ThrowOnError(
				HtmlDynamicInterop.Print.SetResourceLoadCallback(
					this._hPrint,
					this._loadDataHandler));
			HtmlPrintException.ThrowOnError(
				HtmlDynamicInterop.Print.SetCallback(
					this._hPrint,
					HtmlPrinterNotificationDispatcher.NotificationCallback,
					(IntPtr)this._gcHandle));

			OnHtmlHandleCreating(EventArgs.Empty);
			OnHtmlHandleCreated(EventArgs.Empty);

			UpdateHtml(HtmlProtocolBase.BinaryDataFromString(this._html));
		}

		#endregion

		#region Loading

		/// <summary>
		/// Loads HTML code from URI.
		/// </summary>
		/// <param name="url">URI to load HTML code from.</param>
		public void LoadHtml(Uri uri)
		{
			base.CheckDisposed();

			if (this.IsHtmlHandleCreated)
			{
				byte[] htmlData;

				this.NotificationHandler.OnResourceLoading(
					uri,
					HtmlResourceType.Html,
					out htmlData);

				UpdateHtml(htmlData);

				using (HtmlTag root = this.Root)
				{
					if (root == null)
					{
						this._html = string.Empty;
					}
					else
					{
						this._html = root.OuterHtml;
					}
				}
			}
		}

		#endregion

		#region Printing

		public void BeginPrint()
		{
			base.CheckDisposed();

			this._pageNumber = -1;
			this._pageIndex = 0;
		}

		public void EndPrint()
		{
			base.CheckDisposed();

			this._pageNumber = -1;
			this._pageIndex = -1;
		}

		public void MeasureDocument(
			IntPtr hDC,
			int widthInPixels,
			Size sizeInUnits)
		{
			base.CheckDisposed();

			HtmlPrintException.ThrowOnError(
				HtmlDynamicInterop.Print.Measure(
					this._hPrint,
					hDC,
					widthInPixels,
					sizeInUnits.Width,
					sizeInUnits.Height,
					out this._pageNumber));
		}

		public void RenderNextPage(
			IntPtr hDC,
			Point offset)
		{
			base.CheckDisposed();

			HtmlPrintException.ThrowOnError(
				HtmlDynamicInterop.Print.RenderOnDC(
					this._hPrint,
					hDC,
					offset.X,
					offset.Y,
					this._pageIndex));

			this._pageIndex++;
		}

		#endregion

		#endregion

		#region Properties

		/// <summary>
		/// Document base URL.
		/// </summary>
		public Uri BaseUrl
		{
			get
			{
				base.CheckDisposed();

				return this._baseUrl;
			}

			set
			{
				base.CheckDisposed();

				if (this._baseUrl != value)
				{
					this._baseUrl = value;
					UpdateHtml(HtmlProtocolBase.BinaryDataFromString(this._html));
				}
			}
		}

		/// <summary>
		/// HTML code.
		/// </summary>
		/// <param name="html">HTML code.</param>
		public string Html
		{
			get
			{
				base.CheckDisposed();

				using (HtmlTag root = this.Root)
				{
					if (root == null)
					{
						return string.Empty;
					}
					else
					{
						return root.OuterHtml;
					}
				}
			}

			set
			{
				base.CheckDisposed();

				this._html = (value == null) ? string.Empty : value;
				UpdateHtml(HtmlProtocolBase.BinaryDataFromString(this._html));
			}
		}

		/// <summary>
		/// Root HTML tag.
		/// </summary>
		public HtmlTag Root
		{
			get
			{
				base.CheckDisposed();

				IntPtr hTag;

				if (HtmlDynamicInterop.Print.GetRootTag(
					this._hPrint,
					out hTag) == (int)HtmlResult.Print_Success)
				{
					return HtmlTag.FromHandle(hTag);
				}
				else
				{
					return null;
				}
			}
		}

		public int PageNumber
		{
			get
			{
				base.CheckDisposed();

				return this._pageNumber;
			}
		}

		public int PageIndex
		{
			get
			{
				base.CheckDisposed();

				return this._pageIndex;
			}
		}

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			OnHtmlHandleDestroying(EventArgs.Empty);

			HtmlPrintException.ThrowOnError(
				HtmlDynamicInterop.Print.SetCallback(
					this._hPrint,
					null,
					IntPtr.Zero));

			if (this._hPrint != IntPtr.Zero)
			{
				HtmlPrintException.ThrowOnError(
					HtmlDynamicInterop.Print.DestroyInstance(
						this._hPrint));
				this._hPrint = IntPtr.Zero;
			}

			if (this._gcHandle.IsAllocated)
			{
				this._gcHandle.Free();
			}

			OnHtmlHandleDestroyed(EventArgs.Empty);

			base.Dispose(disposing);
		}

		#endregion

		#region IHtmlHost

		#region Properties

		public virtual IHtmlHostNotificationHandler NotificationHandler
		{
			get
			{
				base.CheckDisposed();

				return this._notificationHandler;
			}
		}

		public IntPtr HtmlHandle
		{
			get
			{
				base.CheckDisposed();

				return this._hPrint;
			}
		}

		public bool IsHtmlHandleCreated
		{
			get
			{
				base.CheckDisposed();

				return true;
			}
		}

		#endregion

		#region Events

		public event EventHandler DocumentResetting;
		public event EventHandler DocumentReseted;
		public event EventHandler HtmlHandleCreating;
		public event EventHandler HtmlHandleCreated;
		public event EventHandler HtmlHandleDestroying;
		public event EventHandler HtmlHandleDestroyed;

		#endregion

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
