﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Drawing;
using System.Drawing.Printing;

namespace Nabu.Forms.Html
{
	public sealed class HtmlPrintHelper<T> : DisposableBase where T : HtmlPrinter, new()
	{
		#region HtmlPrintHelper

		#region Fields

		private T _printer = new T();

		#endregion

		#region Methods

		#region Constructors

		/// <summary>
		/// Creates new HtmlPrinter object.
		/// </summary>
		public HtmlPrintHelper()
		{
		}

		#endregion

		#region Loading

		/// <summary>
		/// Loads HTML code from URL.
		/// </summary>
		/// <param name="url">URL to load HTML code from.</param>
		public void LoadHtml(Uri url)
		{
			this._printer.LoadHtml(url);
		}

		#endregion

		#region Printing

		public void BeginPrint(object sender, PrintEventArgs e)
		{
			this._printer.BeginPrint();
		}

		public void EndPrint(object sender, PrintEventArgs e)
		{
			this._printer.EndPrint();
		}

		public void PrintPage(object sender, PrintPageEventArgs e)
		{
			IntPtr hDC = IntPtr.Zero;

			try
			{
				hDC = e.Graphics.GetHdc();

				PageSettings pageSettings = e.PageSettings;
				Point printerResolution = new Point(
					pageSettings.PrinterResolution.X,
					pageSettings.PrinterResolution.Y == 0 ? pageSettings.PrinterResolution.X : pageSettings.PrinterResolution.Y);
				Rectangle printableArea = new Rectangle(
					pageSettings.Bounds.Left + pageSettings.Margins.Left,
					pageSettings.Bounds.Top + pageSettings.Margins.Top,
					pageSettings.Bounds.Width - (pageSettings.Margins.Left + pageSettings.Margins.Right),
					pageSettings.Bounds.Height - (pageSettings.Margins.Top + pageSettings.Margins.Bottom));

				if (this._printer.PageNumber == -1)
				{
					this._printer.MeasureDocument(
						hDC,
						(int)(printableArea.Width * 96.0F / 100.0F),
						new Size(
							(int)(printableArea.Width * printerResolution.X / 100.0F),
							(int)(printableArea.Height * printerResolution.Y / 100.0F)));
				}

				this._printer.RenderNextPage(
					hDC,
					new Point(
						(int)(printableArea.Left * printerResolution.X / 100.0F),
						(int)(printableArea.Top * printerResolution.Y / 100.0F)));

				e.HasMorePages = this._printer.PageIndex != this._printer.PageNumber;
			}
			finally
			{
				if (hDC != IntPtr.Zero)
				{
					e.Graphics.ReleaseHdc(hDC);
				}
			}
		}

		#endregion

		#endregion

		#region Properties

		/// <summary>
		/// Document base URL.
		/// </summary>
		public Uri BaseUrl
		{
			get
			{
				base.CheckDisposed();

				return this._printer.BaseUrl;
			}

			set
			{
				base.CheckDisposed();

				this._printer.BaseUrl = value;
			}
		}

		/// <summary>
		/// HTML code.
		/// </summary>
		/// <param name="html">HTML code.</param>
		public string Html
		{
			get
			{
				base.CheckDisposed();

				return this._printer.Html;
			}

			set
			{
				base.CheckDisposed();

				this._printer.Html = value;
			}
		}

		/// <summary>
		/// Root HTML tag.
		/// </summary>
		public HtmlTag Root
		{
			get
			{
				base.CheckDisposed();

				return this._printer.Root;
			}
		}

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				this._printer.Dispose();
			}

			base.Dispose(disposing);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
