﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Text;
using System.ComponentModel.Design;

namespace Nabu.Forms.Html
{
	[Designer(typeof(HtmlComponentDesigner))]
	public sealed class HtmlComponent : Component
	{
		#region Constants

		private const string FormReassignExceptionMessage = "Form property cannot be reassigned.";
		private const string TagExceptionMessage = "Component is not bound to HTML tag.";

		#endregion

		#region Fields

		private HtmlForm _htmlForm;
		private HtmlTag _htmlTag;
		private string _bindingID;

		#endregion

		#region Methods

		#region Utility

		private void BindEventHandlers()
		{
			using (HtmlTag rootTag = this._htmlForm.RootTag)
			{
				this._htmlTag = rootTag.SelectFirst(string.Format("[id='{0}']", _bindingID));

				if (this._htmlTag != null)
				{
					this._htmlTag.Initialization += new EventHandler<HtmlInitializationEventArgs>(HtmlTag_Initialization);
					this._htmlTag.Mouse += new EventHandler<HtmlMouseEventArgs>(HtmlTag_Mouse);
					this._htmlTag.Keyboard += new EventHandler<HtmlKeyboardEventArgs>(HtmlTag_Keyboard);
					this._htmlTag.Focus += new EventHandler<HtmlFocusEventArgs>(HtmlTag_Focus);
					this._htmlTag.Scroll += new EventHandler<HtmlScrollEventArgs>(HtmlTag_Scroll);
					this._htmlTag.Timer += new EventHandler<HtmlTimerEventArgs>(HtmlTag_Timer);
					this._htmlTag.Size += new EventHandler<EventArgs>(HtmlTag_Size);
					this._htmlTag.Draw += new EventHandler<HtmlDrawEventArgs>(HtmlTag_Draw);
					this._htmlTag.Data += new EventHandler<HtmlDataEventArgs>(HtmlTag_Data);
					this._htmlTag.Behavior += new EventHandler<HtmlBehaviorEventArgs>(HtmlTag_Behavior);
				}
			}
		}

		private void UnbindEventHandlers()
		{
			if (this._htmlTag != null)
			{
				this._htmlTag.Initialization -= new EventHandler<HtmlInitializationEventArgs>(HtmlTag_Initialization);
				this._htmlTag.Mouse -= new EventHandler<HtmlMouseEventArgs>(HtmlTag_Mouse);
				this._htmlTag.Keyboard -= new EventHandler<HtmlKeyboardEventArgs>(HtmlTag_Keyboard);
				this._htmlTag.Focus -= new EventHandler<HtmlFocusEventArgs>(HtmlTag_Focus);
				this._htmlTag.Scroll -= new EventHandler<HtmlScrollEventArgs>(HtmlTag_Scroll);
				this._htmlTag.Timer -= new EventHandler<HtmlTimerEventArgs>(HtmlTag_Timer);
				this._htmlTag.Size -= new EventHandler<EventArgs>(HtmlTag_Size);
				this._htmlTag.Draw -= new EventHandler<HtmlDrawEventArgs>(HtmlTag_Draw);
				this._htmlTag.Data -= new EventHandler<HtmlDataEventArgs>(HtmlTag_Data);
				this._htmlTag.Behavior -= new EventHandler<HtmlBehaviorEventArgs>(HtmlTag_Behavior);
			}
		}

		private void CheckBinding()
		{
			if (this._htmlTag == null)
			{
				throw new InvalidOperationException(TagExceptionMessage);
			}
		}

		#endregion

		#region Event Handlers

		private void Form_HtmlHandleCreated(object sender, EventArgs e)
		{
			BindEventHandlers();
		}

		private void Form_HtmlHandleDestroying(object sender, EventArgs e)
		{
			UnbindEventHandlers();
		}

		private void HtmlTag_Initialization(object sender, HtmlInitializationEventArgs e)
		{
			if (this.Initialization != null)
			{
				this.Initialization(sender, e);
			}
		}

		private void HtmlTag_Mouse(object sender, HtmlMouseEventArgs e)
		{
			if (this.Mouse != null)
			{
				this.Mouse(sender, e);
			}
		}

		private void HtmlTag_Keyboard(object sender, HtmlKeyboardEventArgs e)
		{
			if (this.Keyboard != null)
			{
				this.Keyboard(sender, e);
			}
		}

		private void HtmlTag_Focus(object sender, HtmlFocusEventArgs e)
		{
			if (this.Focus != null)
			{
				this.Focus(sender, e);
			}
		}

		private void HtmlTag_Scroll(object sender, HtmlScrollEventArgs e)
		{
			if (this.Scroll != null)
			{
				this.Scroll(sender, e);
			}
		}

		private void HtmlTag_Timer(object sender, HtmlTimerEventArgs e)
		{
			if (this.Timer != null)
			{
				this.Timer(sender, e);
			}
		}

		private void HtmlTag_Size(object sender, EventArgs e)
		{
			if (this.Size != null)
			{
				this.Size(sender, e);
			}
		}

		private void HtmlTag_Draw(object sender, HtmlDrawEventArgs e)
		{
			if (this.Draw != null)
			{
				this.Draw(sender, e);
			}
		}

		private void HtmlTag_Data(object sender, HtmlDataEventArgs e)
		{
			if (this.Data != null)
			{
				this.Data(sender, e);
			}
		}

		private void HtmlTag_Behavior(object sender, HtmlBehaviorEventArgs e)
		{
			if (this.Behavior != null)
			{
				this.Behavior(sender, e);
			}

			switch (e.Command)
			{
				case HtmlBehaviorEventCommand.ButtonClick:
					if (this.Clicked != null)
					{
						this.Clicked(this, e);
					}
					break;
				case HtmlBehaviorEventCommand.ButtonStateChanged:
				case HtmlBehaviorEventCommand.EditValueChanged:
				case HtmlBehaviorEventCommand.SelectStateChanged:
					if (this.Changed != null)
					{
						this.Changed(this, e);
					}
					break;
			}
		}

		#endregion

		#region Constructors

		public HtmlComponent()
		{
		}

		public HtmlComponent(IContainer container)
		{
			container.Add(this);
		}

		#endregion

		#endregion

		#region Properties

		#region Designer

		[Browsable(false)]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[EditorBrowsable(EditorBrowsableState.Never)]
		public HtmlForm HtmlForm
		{
			get
			{
				if ((this._htmlForm == null) && base.DesignMode)
				{
					IDesignerHost designerHost = this.GetService(typeof(IDesignerHost)) as IDesignerHost;

					if (designerHost != null)
					{
						this._htmlForm = designerHost.RootComponent as HtmlForm;
					}
				}

				return this._htmlForm;
			}

			set
			{
				if (!DesignMode)
				{
					if (
						(this._htmlForm != null) &&
						(this._htmlForm != value))
					{
						throw new InvalidOperationException(FormReassignExceptionMessage);
					}
				}

				this._htmlForm = value;

				if (this._htmlForm != null)
				{
					this._htmlForm.HtmlHandleCreated += new EventHandler(Form_HtmlHandleCreated);
					this._htmlForm.HtmlHandleDestroying += new EventHandler(Form_HtmlHandleDestroying);
				}
			}
		}

		[Browsable(false)]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Visible)]
		[EditorBrowsable(EditorBrowsableState.Never)]
		public string BindingID
		{
			get { return this._bindingID; }
			set { this._bindingID = value; }
		}

		#endregion

		#region HtmlTag Wrapping

		[Browsable(false)]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public string this[string name]
		{
			get
			{
				CheckBinding();

				return this._htmlTag.Attributes[name];
			}

			set
			{
				CheckBinding();

				this._htmlTag.Attributes[name] = value;
			}
		}

		[Browsable(false)]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public string Text
		{
			get
			{
				CheckBinding();

				return this._htmlTag.InnerText;
			}

			set
			{
				CheckBinding();

				this._htmlTag.InnerText = value;
			}
		}

		[Browsable(false)]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public string Html
		{
			get
			{
				CheckBinding();

				return this._htmlTag.InnerHtml;
			}

			set
			{
				CheckBinding();

				this._htmlTag.InnerHtml = value;
			}
		}

		[Browsable(false)]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public bool IsEnabled
		{
			get
			{
				return !this._htmlTag.GetState(HtmlTagState.Disabled);
			}

			set
			{
				this._htmlTag.SetState(HtmlTagState.Disabled, !value, true);
			}
		}

		[Browsable(false)]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public bool IsVisible
		{
			get
			{
				CheckBinding();

				return this._htmlTag.IsVisible;
			}
		}

		[Browsable(false)]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public object Value
		{
			get
			{
				CheckBinding();

				using (HtmlControl htmlControl = new HtmlControl(this._htmlTag))
				{
					return htmlControl.Value;
				}
			}

			set
			{
				CheckBinding();

				using (HtmlControl htmlControl = new HtmlControl(this._htmlTag))
				{
					htmlControl.Value = value;
				}
			}
		}

		[Browsable(false)]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public HtmlTag Tag
		{
			get
			{
				CheckBinding();

				return this._htmlTag;
			}
		}

		#endregion

		#endregion

		#region Events

		public event EventHandler<HtmlInitializationEventArgs> Initialization;
		public event EventHandler<HtmlMouseEventArgs> Mouse;
		public event EventHandler<HtmlKeyboardEventArgs> Keyboard;
		public event EventHandler<HtmlFocusEventArgs> Focus;
		public event EventHandler<HtmlScrollEventArgs> Scroll;
		public event EventHandler<HtmlTimerEventArgs> Timer;
		public event EventHandler<EventArgs> Size;
		public event EventHandler<HtmlDrawEventArgs> Draw;
		public event EventHandler<HtmlDataEventArgs> Data;
		public event EventHandler<HtmlBehaviorEventArgs> Behavior;

		public event EventHandler<HtmlBehaviorEventArgs> Clicked;
		public event EventHandler<HtmlBehaviorEventArgs> Changed;

		#endregion
	}
}
