﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.ComponentModel;
#if !COMPACT_FRAMEWORK
using System.ComponentModel.Design;
#endif
using System.Diagnostics;
using System.Drawing;
#if !COMPACT_FRAMEWORK
using System.Drawing.Design;
#endif
using System.Runtime.InteropServices;
using System.Windows.Forms;

using Nabu.Collections;
using Nabu.Drawing;

namespace Nabu.Forms.Html
{
#if !COMPACT_FRAMEWORK
	[Designer(
		typeof(HtmlFormDesigner),
		typeof(IRootDesigner))]
#endif
	public class HtmlForm :
		Form,
		IHtmlHost
	{
		#region HtmlForm

		#region Fields

		private IHtmlHostNotificationHandler _notificationHandler;
		private Uri _baseUrl;
		private string _html = "<html><head></head><body></body></html>";
		private HtmlLayoutMode _layoutMode;
		private HtmlScrollMode _scrollMode;
		private HtmlFontMode _fontSmoothingMode;
		private GCHandle _gcHandle;
#if COMPACT_FRAMEWORK
		private bool _isHtmlHandleCreated = false;
		private IntPtr _hHtmlHandle;
		private bool _isLoaded = false;
#else
		private Padding _bodyMargin;
#endif

		#endregion

		#region Methods

		#region Utility

		private void UpdateHtml(byte[] htmlData)
		{
			OnDocumentResetting(EventArgs.Empty);
			OnDocumentReseted(EventArgs.Empty);

			if (this._baseUrl != null)
			{
				HtmlDynamicInterop.Window.LoadHtmlEx(
					this.HtmlHandle,
					htmlData,
					htmlData.Length,
					Uri.UnescapeDataString(this._baseUrl.AbsoluteUri));
			}
			else
			{
				HtmlDynamicInterop.Window.LoadHtml(
					this.HtmlHandle,
					htmlData,
					htmlData.Length);
			}

			HtmlWindowEngine.SetLayoutMode(this.HtmlHandle, this._layoutMode);
			HtmlWindowEngine.SetScrollMode(this.HtmlHandle, this._scrollMode);
			HtmlWindowEngine.SetFontSmoothingMode(this.HtmlHandle, this._fontSmoothingMode);
		}

#if !COMPACT_FRAMEWORK
		private void UpdateBodyMargin()
		{
			using (HtmlTag root = this.RootTag)
			{
				if (root != null)
				{
					using (HtmlTag body = root.SelectFirst("body"))
					{
						body.SetStyle(
							"margin",
							string.Format(
								"{0}px {1}px {2}px {3}px",
								this._bodyMargin.Top,
								this._bodyMargin.Right,
								this._bodyMargin.Bottom,
								this._bodyMargin.Left));
					}
				}
			}
		}
#endif

		#endregion

		#region Events Raisers

		protected virtual void OnDocumentResetting(EventArgs e)
		{
			if (this.DocumentResetting != null)
			{
				this.DocumentResetting(this, e);
			}
		}

		protected virtual void OnDocumentReseted(EventArgs e)
		{
			if (this.DocumentReseted != null)
			{
				this.DocumentReseted(this, e);
			}
		}

		protected virtual void OnHtmlHandleCreating(EventArgs e)
		{
			if (this.HtmlHandleCreating != null)
			{
				this.HtmlHandleCreating(this, e);
			}
		}

		protected virtual void OnHtmlHandleCreated(EventArgs e)
		{
			if (this.HtmlHandleCreated != null)
			{
				this.HtmlHandleCreated(this, e);
			}
		}

		protected virtual void OnHtmlHandleDestroying(EventArgs e)
		{
			if (this.HtmlHandleDestroying != null)
			{
				this.HtmlHandleDestroying(this, e);
			}
		}

		protected virtual void OnHtmlHandleDestroyed(EventArgs e)
		{
			if (this.HtmlHandleDestroyed != null)
			{
				this.HtmlHandleDestroyed(this, e);
			}
		}

		#endregion

		#region Constructors

		public HtmlForm()
		{
#if !COMPACT_FRAMEWORK
			base.SetStyle(
				ControlStyles.AllPaintingInWmPaint | ControlStyles.Selectable | ControlStyles.UserPaint,
				true);
#endif
		}

		#endregion

		#region Load

		public void LoadHtml(Uri uri)
		{
			if (this.IsHtmlHandleCreated)
			{
				byte[] htmlData;

				this.NotificationHandler.OnResourceLoading(
					uri,
					HtmlResourceType.Html,
					out htmlData);

				UpdateHtml(htmlData);

				using (HtmlTag root = this.RootTag)
				{
					if (root == null)
					{
						this._html = string.Empty;
					}
					else
					{
						this._html = root.OuterHtml;
					}
				}
			}
		}

		#endregion

		#region Update

		public bool ForceUpdate(bool redraw)
		{
			if (redraw)
			{
				return HtmlDynamicInterop.Window.ForceUpdate(this.HtmlHandle);
			}
			else
			{
				return HtmlDynamicInterop.Window.CommitUpdate(this.HtmlHandle);
			}
		}

		#endregion

		#region Tag

		public HtmlTag GetTagFromPoint(Point point)
		{
			return HtmlWindowEngine.GetTagFromPoint(this.HtmlHandle, point);
		}

		#endregion

		#region Subscriptions

		public void Subscribe(string cssSelector, EventHandler<HtmlInitializationEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Initialization += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlInitializationEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Initialization -= handler;
					}
				}
			}
		}

		public void Subscribe(string cssSelector, EventHandler<HtmlMouseEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Mouse += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlMouseEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Mouse -= handler;
					}
				}
			}
		}

		public void Subscribe(string cssSelector, EventHandler<HtmlKeyboardEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Keyboard += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlKeyboardEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Keyboard -= handler;
					}
				}
			}
		}

		public void Subscribe(string cssSelector, EventHandler<HtmlFocusEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Focus += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlFocusEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Focus -= handler;
					}
				}
			}
		}

		public void Subscribe(string cssSelector, EventHandler<HtmlScrollEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Scroll += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlScrollEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Scroll -= handler;
					}
				}
			}
		}

		public void Subscribe(string cssSelector, EventHandler<HtmlDrawEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Draw += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlDrawEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Draw -= handler;
					}
				}
			}
		}

		public void Subscribe(string cssSelector, EventHandler<HtmlDataEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Data += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlDataEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Data -= handler;
					}
				}
			}
		}

		public void Subscribe(string cssSelector, EventHandler<HtmlBehaviorEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Behavior += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlBehaviorEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Behavior -= handler;
					}
				}
			}
		}

		#endregion

		#endregion

		#region Properties

		#region Html

#if !COMPACT_FRAMEWORK
		[Category("Appearance")]
		[SettingsBindable(true)]
#endif
		public Uri BaseUrl
		{
			get
			{
				return this._baseUrl;
			}

			set
			{
				if (this._baseUrl != value)
				{
					this._baseUrl = value;

					if (this.IsHtmlHandleCreated)
					{
						UpdateHtml(HtmlProtocolBase.BinaryDataFromString(this._html));
					}
				}
			}
		}

#if !COMPACT_FRAMEWORK
		[Category("Appearance")]
		[Editor("System.ComponentModel.Design.MultilineStringEditor, System.Design", typeof(UITypeEditor))]
		[SettingsBindable(true)]
#endif
		public string Html
		{
			get
			{
#if COMPACT_FRAMEWORK
				if (this.IsHtmlHandleCreated)
				{
					using (HtmlTag root = this.RootTag)
					{
						if (root == null)
						{
							return string.Empty;
						}
						else
						{
							return root.OuterHtml;
						}
					}
				}
				else
				{
					return this._html;
				}
#else
				if (base.DesignMode)
				{
					return this._html;
				}
				else
				{
					if (base.IsHandleCreated)
					{
						using (HtmlTag root = this.RootTag)
						{
							if (root == null)
							{
								return string.Empty;
							}
							else
							{
								return root.OuterHtml;
							}
						}
					}
					else
					{
						return this._html;
					}
				}
#endif
			}

			set
			{
				this._html = (value == null) ? string.Empty : value;

				if (this.IsHtmlHandleCreated)
				{
					UpdateHtml(HtmlProtocolBase.BinaryDataFromString(this._html));
				}
			}
		}

#if !COMPACT_FRAMEWORK
		[Browsable(false)]
#endif
		public string SelectedHtml
		{
			get
			{
				if (this.IsHtmlHandleCreated)
				{
					return HtmlWindowEngine.GetSelectedHtml(this.HtmlHandle);
				}
				else
				{
					return null;
				}
			}
		}

		#endregion

		#region Tags

#if !COMPACT_FRAMEWORK
		[Browsable(false)]
#endif
		public HtmlTag RootTag
		{
			get
			{
				if (this.IsHtmlHandleCreated)
				{
					return HtmlWindowEngine.GetRootTag(this.HtmlHandle);
				}
				else
				{
					return null;
				}
			}
		}

#if !COMPACT_FRAMEWORK
		[Browsable(false)]
#endif
		public HtmlTag FocusTag
		{
			get
			{
				if (this.IsHtmlHandleCreated)
				{
					return HtmlWindowEngine.GetFocusTag(this.HtmlHandle);
				}
				else
				{
					return null;
				}
			}
		}

		#endregion

		#region Options

#if !COMPACT_FRAMEWORK
		[Category("Behavior")]
		[SettingsBindable(true)]
#endif
		public HtmlLayoutMode LayoutMode
		{
			get
			{
				return this._layoutMode;
			}

			set
			{
				if (this._layoutMode != value)
				{
					this._layoutMode = value;

					if (this.IsHtmlHandleCreated)
					{
						HtmlWindowEngine.SetLayoutMode(this.HtmlHandle, this._layoutMode);
					}
				}
			}
		}

#if !COMPACT_FRAMEWORK
		[Category("Behavior")]
		[SettingsBindable(true)]
#endif
		public HtmlScrollMode ScrollMode
		{
			get
			{
				return this._scrollMode;
			}

			set
			{
				if (this._scrollMode != value)
				{
					this._scrollMode = value;

					if (this.IsHtmlHandleCreated)
					{
						HtmlWindowEngine.SetScrollMode(this.HtmlHandle, this._scrollMode);
					}
				}
			}
		}

#if !COMPACT_FRAMEWORK
		[Category("Behavior")]
		[SettingsBindable(true)]
#endif
		public HtmlFontMode FontSmoothingMode
		{
			get
			{
				return this._fontSmoothingMode;
			}

			set
			{
				if (this._fontSmoothingMode != value)
				{
					this._fontSmoothingMode = value;

					if (this.IsHtmlHandleCreated)
					{
						HtmlWindowEngine.SetFontSmoothingMode(this.HtmlHandle, this._fontSmoothingMode);
					}
				}
			}
		}

		#endregion

		#endregion

		#endregion

		#region Control

#if COMPACT_FRAMEWORK
		protected override void OnHandleCreated(EventArgs e)
		{
			base.OnHandleCreated(e);

			OnHtmlHandleCreating(EventArgs.Empty);

			IntPtr hModule = NativeMethods.Module_GetHandle(null);

			this._hHtmlHandle =
				NativeMethods.Window_Create(
					0,
					HtmlDynamicInterop.Window.GetUnicodeClassName(),
					null,
					WindowStyle.Child,
					0, 0, 0, 0,
					base.Handle,
					IntPtr.Zero,
					hModule,
					IntPtr.Zero);
			this._isHtmlHandleCreated = true;
			this._gcHandle = GCHandle.Alloc(this, GCHandleType.Normal);
			NativeMethods.Window_SetLong(
				HtmlHandle,
				WindowLong.UserData,
				(IntPtr)this._gcHandle);
			HtmlDynamicInterop.Window.SetCallback(
				HtmlHandle,
				HtmlWindowNotificationDispatcher.NotificationCallback,
				(IntPtr)this._gcHandle);
			UpdateHtml(HtmlProtocolBase.BinaryDataFromString(this._html));
			HtmlWindowEngine.SetLayoutMode(this.HtmlHandle, this._layoutMode);
			HtmlWindowEngine.SetScrollMode(this.HtmlHandle, this._scrollMode);
			HtmlWindowEngine.SetFontSmoothingMode(this.HtmlHandle, this._fontSmoothingMode);
			OnHtmlHandleCreated(EventArgs.Empty);
		}

		protected override void OnClosed(EventArgs e)
		{
			OnHtmlHandleDestroying(EventArgs.Empty);

			base.OnClosed(e);
		}

		protected override void OnHandleDestroyed(EventArgs e)
		{
			HtmlDynamicInterop.Window.SetCallback(
				HtmlHandle,
				null,
				IntPtr.Zero);
			this._isHtmlHandleCreated = false;
			NativeMethods.Window_Destroy(this._hHtmlHandle);
			this._gcHandle.Free();
			OnHtmlHandleDestroyed(EventArgs.Empty);

			base.OnHandleDestroyed(e);
#if COMPACT_FRAMEWORK
			if (this._isLoaded)
			{
				NativeMethods.Window_PostQuitMessage(0);
			}
#endif
		}

		protected override void OnLoad(EventArgs e)
		{
			this._isLoaded = true;

			base.OnLoad(e);
		}

		protected override void OnResize(EventArgs e)
		{
			base.OnResize(e);

			NativeMethods.Window_SetPosition(
				this.HtmlHandle,
				IntPtr.Zero,
				0, 0, base.Width, base.Height,
				WindowPositionFlags.NoZOrder | WindowPositionFlags.FrameChanged | WindowPositionFlags.ShowWindow);
		}
#else
		protected override void WndProc(ref Message message)
		{
			switch ((WindowMessage)message.Msg)
			{
				case WindowMessage.WM_CREATE:
					if (!HtmlWindowNotificationDispatcher.WindowProc(ref message))
					{
						base.WndProc(ref message);

						OnHtmlHandleCreating(EventArgs.Empty);

						this._gcHandle = GCHandle.Alloc(this, GCHandleType.Normal);
						NativeMethods.Window_SetLong(
							HtmlHandle,
							WindowLong.UserData,
							(IntPtr)this._gcHandle);
						HtmlDynamicInterop.Window.SetCallback(
							HtmlHandle,
							HtmlWindowNotificationDispatcher.NotificationCallback,
							(IntPtr)this._gcHandle);

						UpdateHtml(HtmlProtocolBase.BinaryDataFromString(this._html));
						HtmlWindowEngine.SetLayoutMode(this.HtmlHandle, this._layoutMode);
						HtmlWindowEngine.SetScrollMode(this.HtmlHandle, this._scrollMode);
						HtmlWindowEngine.SetFontSmoothingMode(this.HtmlHandle, this._fontSmoothingMode);
						UpdateBodyMargin();
						OnHtmlHandleCreated(EventArgs.Empty);
					}
					break;
				case WindowMessage.WM_SETTEXT:
					base.WndProc(ref message);
					break;
				case WindowMessage.WM_CLOSE:
					if (!HtmlWindowNotificationDispatcher.WindowProc(ref message))
					{
						OnHtmlHandleDestroying(EventArgs.Empty);

						base.WndProc(ref message);
					}
					break;
				case WindowMessage.WM_DESTROY:
					if (!HtmlWindowNotificationDispatcher.WindowProc(ref message))
					{
						HtmlDynamicInterop.Window.SetCallback(
							HtmlHandle,
							null,
							IntPtr.Zero);
						this._gcHandle.Free();
						OnHtmlHandleDestroyed(EventArgs.Empty);

						base.WndProc(ref message);
					}
					break;
				default:
					if (base.DesignMode)
					{
						if (
							(
								((WindowMessage)message.Msg >= WindowMessage.WM_MOUSEFIRST) &&
								((WindowMessage)message.Msg <= WindowMessage.WM_MOUSELAST)) ||
							(
								((WindowMessage)message.Msg >= WindowMessage.WM_KEYFIRST) &&
								((WindowMessage)message.Msg <= WindowMessage.WM_KEYLAST)))
						{
							base.WndProc(ref message);
						}
						else
						{
							if (!HtmlWindowNotificationDispatcher.WindowProc(ref message))
							{
								base.WndProc(ref message);
							}
						}
					}
					else
					{
						if (!HtmlWindowNotificationDispatcher.WindowProc(ref message))
						{
							base.WndProc(ref message);
						}
					}
					break;
			}
		}

		protected override void OnLayout(LayoutEventArgs e)
		{
			this._bodyMargin = base.Padding;

			foreach (Control control in base.Controls)
			{
				switch (control.Dock)
				{
					case DockStyle.Top:
						this._bodyMargin.Top = Math.Max(this._bodyMargin.Top, control.Bottom);
						break;
					case DockStyle.Bottom:
						this._bodyMargin.Bottom = Math.Max(this._bodyMargin.Bottom, base.ClientSize.Height - control.Top);
						break;
					case DockStyle.Left:
						this._bodyMargin.Left = Math.Max(this._bodyMargin.Left, control.Right);
						break;
					case DockStyle.Right:
						this._bodyMargin.Right = Math.Max(this._bodyMargin.Right, base.ClientSize.Width - control.Left);
						break;
				}
			}

			if (this.IsHtmlHandleCreated)
			{
				UpdateBodyMargin();
			}

			base.OnLayout(e);
		}

#endif

		#endregion

		#region IHtmlHost

		#region Properties

#if !COMPACT_FRAMEWORK
		[Browsable(false)]
#endif
		public virtual IHtmlHostNotificationHandler NotificationHandler
		{
			[DebuggerStepThrough]
			get
			{
				if (this._notificationHandler == null)
				{
					this._notificationHandler = new HtmlWindowNotificationHandler(this);
				}

				return this._notificationHandler;
			}
		}

		public virtual IntPtr HtmlHandle
		{
			get
			{
#if COMPACT_FRAMEWORK
				return this._hHtmlHandle;
#else
				return base.Handle;
#endif
			}
		}

		public virtual bool IsHtmlHandleCreated
		{
			get
			{
#if COMPACT_FRAMEWORK
				return this._isHtmlHandleCreated;
#else
				return base.IsHandleCreated;
#endif
			}
		}

		#endregion

		#region Events

		public event EventHandler DocumentResetting;
		public event EventHandler DocumentReseted;
		public event EventHandler HtmlHandleCreating;
		public event EventHandler HtmlHandleCreated;
		public event EventHandler HtmlHandleDestroying;
		public event EventHandler HtmlHandleDestroyed;

		#endregion

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
