﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Design;
using System.Runtime.InteropServices;
using System.Windows.Forms;

using Nabu.Collections;

namespace Nabu.Forms.Html
{
	public class HtmlView :
		Control,
		IHtmlHost
	{
		#region HtmlView

		#region Fields

		private IHtmlHostNotificationHandler _notificationHandler;
		private Uri _baseUrl;
		private string _html = "<html><head></head><body></body></html>";
		private HtmlLayoutMode _layoutMode;
		private HtmlScrollMode _scrollMode;
		private HtmlFontMode _fontSmoothingMode;
		private GCHandle _gcHandle;

		#endregion

		#region Methods

		#region Utility

		private void UpdateHtml(byte[] htmlData)
		{
			OnDocumentResetting(EventArgs.Empty);
			OnDocumentReseted(EventArgs.Empty);

			if (this._baseUrl != null)
			{
				HtmlDynamicInterop.Window.LoadHtmlEx(
					this.HtmlHandle,
					htmlData,
					htmlData.Length,
					Uri.UnescapeDataString(this._baseUrl.AbsoluteUri));
			}
			else
			{
				HtmlDynamicInterop.Window.LoadHtml(
					this.HtmlHandle,
					htmlData,
					htmlData.Length);
			}

			HtmlWindowEngine.SetLayoutMode(this.HtmlHandle, this._layoutMode);
			HtmlWindowEngine.SetScrollMode(this.HtmlHandle, this._scrollMode);
			HtmlWindowEngine.SetFontSmoothingMode(this.HtmlHandle, this._fontSmoothingMode);
		}

		#endregion

		#region Event Raisers

		protected virtual void OnDocumentResetting(EventArgs e)
		{
			if (this.DocumentResetting != null)
			{
				this.DocumentResetting(this, e);
			}
		}

		protected virtual void OnDocumentReseted(EventArgs e)
		{
			if (this.DocumentReseted != null)
			{
				this.DocumentReseted(this, e);
			}
		}

		protected virtual void OnHtmlHandleCreating(EventArgs e)
		{
			if (this.HtmlHandleCreating != null)
			{
				this.HtmlHandleCreating(this, e);
			}
		}

		protected virtual void OnHtmlHandleCreated(EventArgs e)
		{
			if (this.HtmlHandleCreated != null)
			{
				this.HtmlHandleCreated(this, e);
			}
		}

		protected virtual void OnHtmlHandleDestroying(EventArgs e)
		{
			if (this.HtmlHandleDestroying != null)
			{
				this.HtmlHandleDestroying(this, e);
			}
		}

		protected virtual void OnHtmlHandleDestroyed(EventArgs e)
		{
			if (this.HtmlHandleDestroyed != null)
			{
				this.HtmlHandleDestroyed(this, e);
			}
		}

		#endregion

		#region Constructors

		public HtmlView()
		{
			base.SetStyle(
				ControlStyles.AllPaintingInWmPaint | ControlStyles.Selectable | ControlStyles.UserPaint,
				true);
		}

		public HtmlView(IContainer container)
		{
			container.Add(this);

			base.SetStyle(
				ControlStyles.AllPaintingInWmPaint | ControlStyles.Selectable | ControlStyles.UserPaint,
				true);
		}

		#endregion

		#region Load

		public void LoadHtml(Uri uri)
		{
			if (this.IsHtmlHandleCreated)
			{
				byte[] htmlData;

				this.NotificationHandler.OnResourceLoading(
					uri,
					HtmlResourceType.Html,
					out htmlData);

				UpdateHtml(htmlData);

				using (HtmlTag root = this.RootTag)
				{
					if (root == null)
					{
						this._html = string.Empty;
					}
					else
					{
						this._html = root.OuterHtml;
					}
				}
			}
		}

		#endregion

		#region Update

		public bool ForceUpdate(bool redraw)
		{
			if (redraw)
			{
				return HtmlDynamicInterop.Window.ForceUpdate(this.HtmlHandle);
			}
			else
			{
				return HtmlDynamicInterop.Window.CommitUpdate(this.HtmlHandle);
			}
		}

		#endregion

		#region Tag

		public HtmlTag FromPoint(Point point)
		{
			return HtmlWindowEngine.GetTagFromPoint(this.HtmlHandle, point);
		}

		#endregion

		#region Subscriptions

		public void Subscribe(string cssSelector, EventHandler<HtmlInitializationEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Initialization += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlInitializationEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Initialization -= handler;
					}
				}
			}
		}

		public void Subscribe(string cssSelector, EventHandler<HtmlMouseEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Mouse += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlMouseEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Mouse -= handler;
					}
				}
			}
		}

		public void Subscribe(string cssSelector, EventHandler<HtmlKeyboardEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Keyboard += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlKeyboardEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Keyboard -= handler;
					}
				}
			}
		}

		public void Subscribe(string cssSelector, EventHandler<HtmlFocusEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Focus += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlFocusEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Focus -= handler;
					}
				}
			}
		}

		public void Subscribe(string cssSelector, EventHandler<HtmlScrollEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Scroll += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlScrollEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Scroll -= handler;
					}
				}
			}
		}

		public void Subscribe(string cssSelector, EventHandler<HtmlDrawEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Draw += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlDrawEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Draw -= handler;
					}
				}
			}
		}

		public void Subscribe(string cssSelector, EventHandler<HtmlDataEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Data += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlDataEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Data -= handler;
					}
				}
			}
		}

		public void Subscribe(string cssSelector, EventHandler<HtmlBehaviorEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Behavior += handler;
					}
				}
			}
		}

		public void Unsubscribe(string cssSelector, EventHandler<HtmlBehaviorEventArgs> handler)
		{
			using (HtmlTag root = this.RootTag)
			{
				using (DisposableList<HtmlTag> list = root.Select(cssSelector))
				{
					foreach (HtmlTag tag in list)
					{
						tag.Behavior -= handler;
					}
				}
			}
		}

		#endregion

		#endregion

		#region Properties

		#region Html

		[Category("Appearance")]
		[SettingsBindable(true)]
		public Uri BaseUrl
		{
			get
			{
				return this._baseUrl;
			}

			set
			{
				if (this._baseUrl != value)
				{
					this._baseUrl = value;

					if (this.IsHtmlHandleCreated)
					{
						UpdateHtml(HtmlProtocolBase.BinaryDataFromString(this._html));
					}
				}
			}
		}

		[Category("Appearance")]
		[Editor("System.ComponentModel.Design.MultilineStringEditor, System.Design", typeof(UITypeEditor))]
		[SettingsBindable(true)]
		public string Html
		{
			get
			{
				if (base.DesignMode)
				{
					return this._html;
				}
				else
				{
					if (base.IsHandleCreated)
					{
						using (HtmlTag root = this.RootTag)
						{
							if (root == null)
							{
								return string.Empty;
							}
							else
							{
								return root.OuterHtml;
							}
						}
					}
					else
					{
						return this._html;
					}
				}
			}

			set
			{
				this._html = (value == null) ? string.Empty : value;

				if (this.IsHtmlHandleCreated)
				{
					UpdateHtml(HtmlProtocolBase.BinaryDataFromString(this._html));
				}
			}
		}

		[Browsable(false)]
		public string SelectedHtml
		{
			get
			{
				if (this.IsHtmlHandleCreated)
				{
					return HtmlWindowEngine.GetSelectedHtml(this.HtmlHandle);
				}
				else
				{
					return null;
				}
			}
		}

		#endregion

		#region Tags

		[Browsable(false)]
		public HtmlTag RootTag
		{
			get
			{
				if (this.IsHtmlHandleCreated)
				{
					return HtmlWindowEngine.GetRootTag(this.HtmlHandle);
				}
				else
				{
					return null;
				}
			}
		}

		[Browsable(false)]
		public HtmlTag FocusTag
		{
			get
			{
				if (this.IsHtmlHandleCreated)
				{
					return HtmlWindowEngine.GetFocusTag(this.HtmlHandle);
				}
				else
				{
					return null;
				}
			}
		}

		#endregion

		#region Options

		[Category("Behavior")]
		[SettingsBindable(true)]
		public HtmlLayoutMode LayoutMode
		{
			get
			{
				return this._layoutMode;
			}

			set
			{
				if (this._layoutMode != value)
				{
					this._layoutMode = value;

					if (this.IsHtmlHandleCreated)
					{
						HtmlWindowEngine.SetLayoutMode(this.HtmlHandle, this._layoutMode);
					}
				}
			}
		}

		[Category("Behavior")]
		[SettingsBindable(true)]
		public HtmlScrollMode ScrollMode
		{
			get
			{
				return this._scrollMode;
			}

			set
			{
				if (this._scrollMode != value)
				{
					this._scrollMode = value;

					if (this.IsHtmlHandleCreated)
					{
						HtmlWindowEngine.SetScrollMode(this.HtmlHandle, this._scrollMode);
					}
				}
			}
		}

		[Category("Behavior")]
		[SettingsBindable(true)]
		public HtmlFontMode FontSmoothingMode
		{
			get
			{
				return this._fontSmoothingMode;
			}

			set
			{
				if (this._fontSmoothingMode != value)
				{
					this._fontSmoothingMode = value;

					if (this.IsHtmlHandleCreated)
					{
						HtmlWindowEngine.SetFontSmoothingMode(this.HtmlHandle, this._fontSmoothingMode);
					}
				}
			}
		}

		#endregion

		#endregion

		#endregion

		#region Control

		protected override void WndProc(ref Message message)
		{
			switch ((WindowMessage)message.Msg)
			{
				case WindowMessage.WM_CREATE:
					if (!HtmlWindowNotificationDispatcher.WindowProc(ref message))
					{
						base.WndProc(ref message);

						OnHtmlHandleCreating(EventArgs.Empty);

						this._gcHandle = GCHandle.Alloc(this, GCHandleType.Normal);
						NativeMethods.Window_SetLong(
							HtmlHandle,
							WindowLong.UserData,
							(IntPtr)this._gcHandle);
						HtmlDynamicInterop.Window.SetCallback(
							HtmlHandle,
							HtmlWindowNotificationDispatcher.NotificationCallback,
							(IntPtr)this._gcHandle);

						UpdateHtml(HtmlProtocolBase.BinaryDataFromString(this._html));
						HtmlWindowEngine.SetLayoutMode(this.HtmlHandle, this._layoutMode);
						HtmlWindowEngine.SetScrollMode(this.HtmlHandle, this._scrollMode);
						HtmlWindowEngine.SetFontSmoothingMode(this.HtmlHandle, this._fontSmoothingMode);
						OnHtmlHandleCreated(EventArgs.Empty);
					}
					break;
				case WindowMessage.WM_SETTEXT:
					base.WndProc(ref message);
					break;
				case WindowMessage.WM_CLOSE:
					if (!HtmlWindowNotificationDispatcher.WindowProc(ref message))
					{
						OnHtmlHandleDestroying(EventArgs.Empty);

						base.WndProc(ref message);
					}
					break;
				case WindowMessage.WM_DESTROY:
					if (!HtmlWindowNotificationDispatcher.WindowProc(ref message))
					{
						HtmlDynamicInterop.Window.SetCallback(
							HtmlHandle,
							null,
							IntPtr.Zero);
						this._gcHandle.Free();
						OnHtmlHandleDestroyed(EventArgs.Empty);

						base.WndProc(ref message);
					}
					break;
				default:
					if (!HtmlWindowNotificationDispatcher.WindowProc(ref message))
					{
						base.WndProc(ref message);
					}
					break;
			}
		}

		#endregion

		#region IHtmlHost

		#region Properties

		[Browsable(false)]
		public virtual IHtmlHostNotificationHandler NotificationHandler
		{
			[DebuggerStepThrough]
			get
			{
				if (this._notificationHandler == null)
				{
					this._notificationHandler = new HtmlWindowNotificationHandler(this);
				}

				return this._notificationHandler;
			}
		}

		public virtual IntPtr HtmlHandle
		{
			get { return base.Handle; }
		}

		public virtual bool IsHtmlHandleCreated
		{
			get { return base.IsHandleCreated; }
		}

		#endregion

		#region Events

		public event EventHandler DocumentResetting;
		public event EventHandler DocumentReseted;
		public event EventHandler HtmlHandleCreating;
		public event EventHandler HtmlHandleCreated;
		public event EventHandler HtmlHandleDestroying;
		public event EventHandler HtmlHandleDestroyed;

		#endregion

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
