﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Drawing;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.Windows.Forms.Design;

namespace Nabu.Forms.Html
{
	public sealed class HtmlFormDesigner : DocumentDesigner
	{
		#region HtmlFormDesigner

		#region Types

		private sealed class TagDescritor
		{
			#region Fields

			private IDictionary<string, string> _eventHandlers = new SortedList<string, string>();

			#endregion

			#region Methods

			public TagDescritor()
			{
			}

			#endregion

			#region Properties

			public IDictionary<string, string> EventHandlers
			{
				get { return this._eventHandlers; }
			}

			#endregion
		}

		#endregion

		#region Fields

		private Regex _regExIdentifier = new Regex(@"^[a-zA-Z0-9_]+$");
		private IDesignerHost _designerHost;
		private IComponentChangeService _componentChangeService;
		private IEventBindingService _eventBindingService;
		private ISelectionService _selectionService;
		private HtmlForm _htmlForm;
		private bool _isRebinding;

		#endregion

		#region Methods

		#region Utility

		private void RebindComponents()
		{
			this._isRebinding = true;

			List<HtmlComponent> tagComponents = new List<HtmlComponent>();

			foreach (IComponent component in this._designerHost.Container.Components)
			{
				HtmlComponent htmlComponent = component as HtmlComponent;

				if (htmlComponent != null)
				{
					tagComponents.Add(htmlComponent);
				}
			}

			Dictionary<string, TagDescritor> tagDescriptors = new Dictionary<string, TagDescritor>();

			foreach (HtmlComponent htmlComponent in tagComponents)
			{
				TagDescritor tagDescriptor = new TagDescritor();

				foreach (EventDescriptor tagEvent in TypeDescriptor.GetEvents(typeof(HtmlComponent)))
				{
					if (tagEvent.IsBrowsable)
					{
						PropertyDescriptor tagEventProperty = this._eventBindingService.GetEventProperty(tagEvent);
						string tagEventHandler = tagEventProperty.GetValue(htmlComponent) as string;

						tagDescriptor.EventHandlers[tagEvent.Name] = tagEventHandler;
					}
				}

				tagDescriptors[htmlComponent.Site.Name] = tagDescriptor;
				this._designerHost.DestroyComponent(htmlComponent);
			}

			HtmlForm htmlForm = (HtmlForm)this._designerHost.RootComponent as HtmlForm;

			foreach (HtmlTag tag in htmlForm.RootTag.Select("[id]"))
			{
				string tagID = tag.Attributes["id"];

				if (this._regExIdentifier.IsMatch(tagID))
				{
					using (DesignerTransaction transaction = this._designerHost.CreateTransaction())
					{
						HtmlComponent htmlComponent = (HtmlComponent)this._designerHost.CreateComponent(typeof(HtmlComponent), tagID);

						htmlComponent.BindingID = tagID;

						TagDescritor tagDescriptor;

						if (tagDescriptors.TryGetValue(tagID, out tagDescriptor))
						{
							foreach (EventDescriptor tagEvent in TypeDescriptor.GetEvents(typeof(HtmlComponent)))
							{
								if (tagEvent.IsBrowsable)
								{
									string tagEventHandler;

									if (tagDescriptor.EventHandlers.TryGetValue(tagEvent.Name, out tagEventHandler))
									{
										PropertyDescriptor tagEventProperty = this._eventBindingService.GetEventProperty(tagEvent);

										tagEventProperty.SetValue(htmlComponent, tagEventHandler);
									}
								}
							}
						}

						transaction.Commit();
					}
				}
			}

			this._isRebinding = false;
		}

		#endregion

		#region Event Handlers

		private void ComponentChange_ComponentChanged(object sender, ComponentChangedEventArgs e)
		{
			HtmlForm htmlForm = e.Component as HtmlForm;
			HtmlComponent htmlComponent = e.Component as HtmlComponent;
			PropertyDescriptor propertyDescriptor = e.Member as PropertyDescriptor;
			EventDescriptor eventDescriptor = e.Member as EventDescriptor;

			if (htmlForm != null)
			{
				if (propertyDescriptor != null)
				{
					switch (propertyDescriptor.Name)
					{
						case "Html":
							RebindComponents();
							break;
					}
				}
			}
			else if (htmlComponent != null)
			{
				if (eventDescriptor != null)
				{
					if (!this._isRebinding)
					{
						this._eventBindingService.ShowCode(htmlComponent, eventDescriptor);
					}
				}
			}
		}

		private void HtmlForm_MouseClick(object sender, MouseEventArgs e)
		{
			bool isComponentSelected = false;

			using (HtmlTag htmlTag = this._htmlForm.GetTagFromPoint(e.Location))
			{
				HtmlTag componentTag = htmlTag;

				while (
					(componentTag != null) &&
					(!componentTag.Attributes.ContainsKey("id")))
				{
					componentTag = componentTag.Parent;
				}

				if (componentTag != null)
				{
					string bindingID = componentTag.Attributes["id"];

					foreach (IComponent component in this._designerHost.Container.Components)
					{
						HtmlComponent htmlComponent = component as HtmlComponent;

						if (htmlComponent != null)
						{
							if (htmlComponent.BindingID == bindingID)
							{
								this._selectionService.SetSelectedComponents(new IComponent[] { component });
								isComponentSelected = true;

								break;
							}
						}
					}
				}
			}

			if (!isComponentSelected)
			{
				this._selectionService.SetSelectedComponents(new IComponent[] { this._htmlForm });
			}
		}

		#endregion

		#region Constructors

		public HtmlFormDesigner()
		{
		}

		#endregion

		#endregion

		#endregion

		#region DocumentDesigner

		public override void Initialize(IComponent component)
		{
			base.Initialize(component);

			this._designerHost = (IDesignerHost)base.GetService(typeof(IDesignerHost));
			this._componentChangeService = (IComponentChangeService)base.GetService(typeof(IComponentChangeService));
			this._eventBindingService = (IEventBindingService)base.GetService(typeof(IEventBindingService));
			this._selectionService = (ISelectionService)base.GetService(typeof(ISelectionService));

			this._componentChangeService.ComponentChanged += new ComponentChangedEventHandler(ComponentChange_ComponentChanged);

			this._htmlForm = (HtmlForm)component;
			this._htmlForm.MouseClick += new System.Windows.Forms.MouseEventHandler(HtmlForm_MouseClick);
		}

		protected override bool GetHitTest(Point pt)
		{
			return true;
		}

		#endregion

		#region IDisposable

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				this._componentChangeService.ComponentChanged -= new ComponentChangedEventHandler(ComponentChange_ComponentChanged);

				this._eventBindingService = null;
				this._componentChangeService = null;
			}

			base.Dispose(disposing);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
