﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms.Html Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Drawing;

namespace Nabu.Forms.Html
{
	public class HtmlWindowless :
		DisposableBase,
		IHtmlHost
	{
		#region HtmlWindowless

		#region Fields

		private IHtmlHostNotificationHandler _notificationHandler;
		private IntPtr _hEngine;
		private Uri _baseUrl;
		private string _html = "<html><head></head><body></body></html>";

		#endregion

		#region Methods

		#region Utility

		private void UpdateHtml(byte[] htmlData)
		{
			OnDocumentResetting();
			OnDocumentReseted();

			if (this._baseUrl != null)
			{
				HtmlPrintException.ThrowOnError(
					HtmlDynamicInterop.Windowless.LoadHtml(
						this._hEngine,
						this._baseUrl.AbsoluteUri,
						htmlData,
						htmlData.Length));
			}
			else
			{
				HtmlPrintException.ThrowOnError(
					HtmlDynamicInterop.Windowless.LoadHtml(
						this._hEngine,
						null,
						htmlData,
						htmlData.Length));
			}
		}

		#endregion

		#region Events Raisers

		protected virtual void OnDocumentResetting()
		{
			if (this.DocumentResetting != null)
			{
				this.DocumentResetting(this, EventArgs.Empty);
			}
		}

		protected virtual void OnDocumentReseted()
		{
			if (this.DocumentReseted != null)
			{
				this.DocumentReseted(this, EventArgs.Empty);
			}
		}

		protected virtual void OnHtmlHandleCreating(EventArgs e)
		{
			if (this.HtmlHandleCreating != null)
			{
				this.HtmlHandleCreating(this, e);
			}
		}

		protected virtual void OnHtmlHandleCreated(EventArgs e)
		{
			if (this.HtmlHandleCreated != null)
			{
				this.HtmlHandleCreated(this, e);
			}
		}

		protected virtual void OnHtmlHandleDestroying(EventArgs e)
		{
			if (this.HtmlHandleDestroying != null)
			{
				this.HtmlHandleDestroying(this, e);
			}
		}

		protected virtual void OnHtmlHandleDestroyed(EventArgs e)
		{
			if (this.HtmlHandleDestroyed != null)
			{
				this.HtmlHandleDestroyed(this, e);
			}
		}

		#endregion

		#region Constructors

		public HtmlWindowless()
		{
			this._hEngine = HtmlDynamicInterop.Windowless.CreateInstance();
			this._notificationHandler = new HtmlWindowlessNotificationHandler();
		}

		#endregion

		#region Loading

		public void LoadHtml(Uri uri)
		{
			base.CheckDisposed();

			byte[] htmlData;

			this._notificationHandler.OnResourceLoading(
				uri,
				HtmlResourceType.Html,
				out htmlData);

			UpdateHtml(htmlData);

			using (HtmlTag root = this.Root)
			{
				if (root == null)
				{
					this._html = string.Empty;
				}
				else
				{
					this._html = root.OuterHtml;
				}
			}
		}

		#endregion

		#region Update

		public void Update()
		{
			base.CheckDisposed();

			HtmlDynamicInterop.Windowless.UpdateView(this._hEngine);
		}

		#endregion

		#region Measure

		public void Measure(int width, int height)
		{
			base.CheckDisposed();

			HtmlPrintException.ThrowOnError(
				HtmlDynamicInterop.Windowless.Measure(
					this._hEngine,
					width,
					height));
		}

		#endregion

		#region Render

		public void RenderOnDC(IntPtr hDC, Rectangle area)
		{
			base.CheckDisposed();

			HtmlPrintException.ThrowOnError(
				HtmlDynamicInterop.Windowless.RenderOnDC(
					this._hEngine,
					hDC,
					area.X,
					area.Y,
					area.Width,
					area.Height));
		}

		public void RenderOnBitmap(IntPtr hBitmap, Rectangle area)
		{
			base.CheckDisposed();

			HtmlPrintException.ThrowOnError(
				HtmlDynamicInterop.Windowless.RenderOnBitmap(
					this._hEngine,
					hBitmap,
					area.X,
					area.Y,
					area.Width,
					area.Height));
		}

		#endregion

		#endregion

		#region Properties

		public Uri BaseUrl
		{
			get
			{
				base.CheckDisposed();

				return this._baseUrl;
			}

			set
			{
				base.CheckDisposed();

				if (this._baseUrl != value)
				{
					this._baseUrl = value;

					byte[] htmlData = HtmlProtocolBase.BinaryDataFromString(this._html);

					HtmlPrintException.ThrowOnError(
						HtmlDynamicInterop.Windowless.LoadHtml(
							this._hEngine,
							this._baseUrl.AbsoluteUri,
							htmlData,
							htmlData.Length));
				}
			}
		}

		public string Html
		{
			get
			{
				base.CheckDisposed();

				using (HtmlTag root = this.Root)
				{
					if (root == null)
					{
						return string.Empty;
					}
					else
					{
						return root.OuterHtml;
					}
				}
			}

			set
			{
				base.CheckDisposed();

				this._html = (value == null) ? string.Empty : value;

				byte[] htmlData = HtmlProtocolBase.BinaryDataFromString(this._html);

				HtmlPrintException.ThrowOnError(
					HtmlDynamicInterop.Windowless.LoadHtml(
						this._hEngine,
						this._baseUrl.AbsoluteUri,
						htmlData,
						htmlData.Length));
			}
		}

		public HtmlTag Root
		{
			get
			{
				base.CheckDisposed();

				IntPtr hTag;

				if (HtmlDynamicInterop.Windowless.GetRootTag(
					this._hEngine,
					out hTag) == (int)HtmlResult.Print_Success)
				{
					return HtmlTag.FromHandle(hTag);
				}
				else
				{
					return null;
				}
			}
		}

		public HtmlTag FocusTag
		{
			get
			{
				base.CheckDisposed();

				if (this.IsHtmlHandleCreated)
				{
					IntPtr hTag;
					bool isViewActive;

					if (HtmlDynamicInterop.Windowless.GetFocusTag(
						this._hEngine,
						out hTag,
						out isViewActive) == (int)HtmlResult.Print_Success)
					{
						return HtmlTag.FromHandle(hTag);
					}
				}

				return null;
			}
		}

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			OnHtmlHandleDestroying(EventArgs.Empty);

			if (this._hEngine != IntPtr.Zero)
			{
				HtmlPrintException.ThrowOnError(
					HtmlDynamicInterop.Windowless.DestroyInstance(
						this._hEngine));

				this._hEngine = IntPtr.Zero;
			}

			OnHtmlHandleDestroyed(EventArgs.Empty);

			base.Dispose(disposing);
		}

		#endregion

		#region IHtmlHost

		#region Properties

		public virtual IHtmlHostNotificationHandler NotificationHandler
		{
			get
			{
				base.CheckDisposed();

				return this._notificationHandler;
			}
		}

		public IntPtr HtmlHandle
		{
			get
			{
				base.CheckDisposed();

				return this._hEngine;
			}
		}

		public bool IsHtmlHandleCreated
		{
			get
			{
				base.CheckDisposed();

				return true;
			}
		}

		#endregion

		#region Events

		public event EventHandler DocumentResetting;
		public event EventHandler DocumentReseted;
		public event EventHandler HtmlHandleCreating;
		public event EventHandler HtmlHandleCreated;
		public event EventHandler HtmlHandleDestroying;
		public event EventHandler HtmlHandleDestroyed;

		#endregion

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
