﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms Library                            */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Security.Permissions;
using System.Windows.Forms;
using System.Windows.Forms.VisualStyles;

using Nabu.Drawing;

namespace Nabu.Forms
{
	public sealed partial class ColorPickerBox : Control
	{
		#region ColorPickerBox

		#region Fields

		private string _palette = string.Empty;
		private Color _value = Color.White;
		private Size _formSize = new Size();
		private bool _dropDownState;

		#endregion

		#region Methods

		private void RaiseValueChanged()
		{
			if (this.ValueChanged != null)
			{
				this.ValueChanged(this, EventArgs.Empty);
			}
		}

		private void FormColorPicker_FormClosed(object sender, FormClosedEventArgs e)
		{
			FormColorPicker formColorPicker = (FormColorPicker)sender;

			this._formSize = formColorPicker.Size;

			if (formColorPicker.DialogResult == DialogResult.OK)
			{
				this._palette = formColorPicker.Palette;
				this._value = formColorPicker.Color;
				RaiseValueChanged();
			}

			this._dropDownState = false;
			base.Refresh();
			base.Focus();
		}

		private void DrawContent(Graphics graphics)
		{
			Rectangle rectContent = this.RectContent;
			Rectangle rectFocus = rectContent;
			Rectangle rectColor = rectContent;

			rectFocus.Inflate(-1, -1);
			rectColor.Inflate(-2, -2);

			if (base.Focused)
			{
				ControlPaint.DrawFocusRectangle(graphics, rectFocus);
			}

			graphics.FillRectangle(new SolidBrush(this._value), rectColor);
		}

		private void DrawDropDown(Graphics graphics)
		{
			Rectangle rectButton = this.RectButton;

			if (Application.RenderWithVisualStyles && ComboBoxRenderer.IsSupported)
			{
				ComboBoxRenderer.DrawDropDownButton(
					graphics,
					rectButton,
					this._dropDownState ? ComboBoxState.Pressed : ComboBoxState.Normal);
			}
			else
			{
				IntPtr hDC = graphics.GetHdc();

				Rectangle rectButtonNative = rectButton;

				RectangleUtils.NativeFromManaged(ref rectButtonNative);

				NativeMethods.FrameControl_Draw(
					hDC,
					ref rectButtonNative,
					FrameControlType.Scroll,
					FrameControlState.ScrollDown | (this._dropDownState ? FrameControlState.Pushed : 0));

				graphics.ReleaseHdc(hDC);
			}
		}

		private void ShowForm()
		{
			FormColorPicker formColorPicker = new FormColorPicker();

			Point location = base.Parent.PointToScreen(this.Location);

			location.Offset(0, base.Size.Height);

			formColorPicker.Location = location;
			formColorPicker.Palette = this._palette;
			formColorPicker.ColorInitial = this._value;
			formColorPicker.FormClosed += FormColorPicker_FormClosed;

			if (!this._formSize.IsEmpty)
			{
				formColorPicker.Size = this._formSize;
			}

			formColorPicker.Show(this);
		}

		public ColorPickerBox()
		{
		}

		#endregion

		#region Properties

		private Rectangle RectContent
		{
			get
			{
				Rectangle rectClient = base.ClientRectangle;

				return new Rectangle(
					rectClient.X,
					rectClient.Y,
					rectClient.Width - rectClient.Height,
					rectClient.Height);
			}
		}

		private Rectangle RectButton
		{
			get
			{
				Rectangle rectClient = base.ClientRectangle;

				return new Rectangle(
					rectClient.X + rectClient.Width - rectClient.Height,
					rectClient.Y,
					rectClient.Height,
					rectClient.Height);
			}
		}

		[Category("Behavior")]
		public string Palette
		{
			get
			{
				return this._palette;
			}

			set
			{
				this._palette = value;
			}
		}

		[Category("Behavior")]
		public Color Value
		{
			get
			{
				return this._value;
			}

			set
			{
				if (this._value != value)
				{
					this._value = value;

					RaiseValueChanged();
				}
			}
		}

		#endregion

		#region Events

		[Category("Property Changed")]
		public event EventHandler ValueChanged;

		#endregion

		#endregion

		#region Control

		#region Methods

		protected override void OnPaint(PaintEventArgs pe)
		{
			DrawContent(pe.Graphics);
			DrawDropDown(pe.Graphics);
		}

		protected override void OnGotFocus(EventArgs e)
		{
			base.Refresh();
			base.OnGotFocus(e);
		}

		protected override void OnLostFocus(EventArgs e)
		{
			base.Refresh();
			base.OnLostFocus(e);
		}

		protected override void OnKeyDown(KeyEventArgs e)
		{
			switch (e.KeyCode)
			{
				case Keys.Space:
					this._dropDownState = true;

					using (Graphics graphics = base.CreateGraphics())
					{
						DrawDropDown(graphics);
					}
					break;
			}

			base.OnKeyDown(e);
		}

		protected override void OnKeyUp(KeyEventArgs e)
		{
			switch (e.KeyCode)
			{
				case Keys.Space:
					ShowForm();
					break;
			}
			base.OnKeyUp(e);
		}

		protected override void OnMouseDown(MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
			{
				Rectangle rectButton = this.RectButton;

				if (rectButton.Contains(new Point(e.X, e.Y)))
				{
					this._dropDownState = true;

					using (Graphics graphics = base.CreateGraphics())
					{
						DrawDropDown(graphics);
					}

					base.Capture = true;
				}
			}

			base.OnMouseDown(e);
		}

		protected override void OnMouseUp(MouseEventArgs e)
		{
			if ((e.Button == MouseButtons.Left) && (base.Capture))
			{
				Rectangle rectButton = this.RectButton;

				if (rectButton.Contains(new Point(e.X, e.Y)))
				{
					ShowForm();
				}
				else
				{
					this._dropDownState = false;
					base.Refresh();
				}

				base.Capture = false;
			}

			base.OnMouseUp(e);
		}

		protected override void OnMouseMove(MouseEventArgs e)
		{
			if (base.Capture)
			{
				Rectangle rectButton = this.RectButton;

				if (rectButton.Contains(new Point(e.X, e.Y)))
				{
					if (!this._dropDownState)
					{
						this._dropDownState = true;

						using (Graphics graphics = base.CreateGraphics())
						{
							DrawDropDown(graphics);
						}
					}
				}
				else
				{
					if (this._dropDownState)
					{
						this._dropDownState = false;

						using (Graphics graphics = base.CreateGraphics())
						{
							DrawDropDown(graphics);
						}
					}
				}
			}

			base.OnMouseMove(e);
		}

		#endregion

		#region Properties

		protected override CreateParams CreateParams
		{
			[SecurityPermission(SecurityAction.LinkDemand, Flags = SecurityPermissionFlag.UnmanagedCode)]
			get
			{
				CreateParams createParams = base.CreateParams;

				createParams.ExStyle |= (int)WindowStyleExtended.ClientEdge;

				return createParams;
			}
		}

		#endregion

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
