﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms Library                            */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics.CodeAnalysis;
using System.Windows.Forms;

namespace Nabu.Forms
{
	public class BindingListView : ListView
	{
		#region BindingListView

		#region Fields

		private IBindingList _dataSource;

		#endregion

		#region Methods

		#region Utility

		private ListViewItem GetBoundItem(int index, object value)
		{
			BindingListViewBindEventArgs eventArgs = new BindingListViewBindEventArgs(index, value);

			if (this.BindItem != null)
			{
				this.BindItem(this, eventArgs);
			}

			return eventArgs.Item;
		}

		#endregion

		#region Event Handlers

		private void OnReset()
		{
			base.SelectedIndices.Clear();
			base.Items.Clear();

			if (this._dataSource != null)
			{
				for (int index = 0; index < this._dataSource.Count; index++)
				{
					base.Items.Add(GetBoundItem(index, this._dataSource[index]));
				}
			}
		}

		private void OnAdd(int index)
		{
			base.Items.Insert(index, GetBoundItem(index, this._dataSource[index]));
		}

		private void OnRemove(int index)
		{
			if (base.Items.Count > index)
			{
				base.Items.RemoveAt(index);
			}
		}

		private void OnChange(int index)
		{
			if (base.Items.Count > index)
			{
				base.Items[index] = GetBoundItem(index, this._dataSource[index]);
			}
		}

		private void OnMove(int index1, int index2)
		{
			if ((base.Items.Count > index1) && (base.Items.Count > index2))
			{
				base.Items[index1] = GetBoundItem(index1, this._dataSource[index1]);
				base.Items[index2] = GetBoundItem(index2, this._dataSource[index2]);
			}
		}

		private void OnListChanged(object sender, ListChangedEventArgs e)
		{
			switch (e.ListChangedType)
			{
				case ListChangedType.Reset:
					OnReset();
					break;
				case ListChangedType.ItemAdded:
					OnAdd(e.NewIndex);
					break;
				case ListChangedType.ItemDeleted:
					OnRemove(e.NewIndex);
					break;
				case ListChangedType.ItemMoved:
					OnMove(e.OldIndex, e.NewIndex);
					break;
				case ListChangedType.ItemChanged:
					OnChange(e.NewIndex);
					break;
			}
		}

		#endregion

		#region Event Handler Proxies

		private void OnListChangedProxy(object sender, ListChangedEventArgs e)
		{
			if (base.InvokeRequired)
			{
				base.Invoke(new ListChangedEventHandler(OnListChanged), sender, e);
			}
			else
			{
				OnListChanged(sender, e);
			}
		}

		#endregion

		#region Constructors

		public BindingListView()
		{
		}

		#endregion

		#endregion

		#region Properties

		[SuppressMessage("Microsoft.Usage", "CA2227:CollectionPropertiesShouldBeReadOnly")]
		public IBindingList DataSource
		{
			get
			{
				return this._dataSource;
			}

			set
			{
				if (this._dataSource != null)
				{
					this._dataSource.ListChanged -= new ListChangedEventHandler(OnListChangedProxy);
				}

				this._dataSource = value;
				OnReset();

				if (this._dataSource != null)
				{
					this._dataSource.ListChanged += new ListChangedEventHandler(OnListChangedProxy);
				}
			}
		}

		[Browsable(false)]
		public int SelectedIndex
		{
			get
			{
				if (base.SelectedIndices.Count == 1)
				{
					return base.SelectedIndices[0];
				}
				else
				{
					return -1;
				}
			}

			set
			{
				base.SelectedIndices.Clear();

				if (value != -1)
				{
					base.SelectedIndices.Add(value);
				}
			}
		}

		public object SelectedItem
		{
			get
			{
				if (base.SelectedIndices.Count == 1)
				{
					return this._dataSource[base.SelectedIndices[0]];
				}
				else
				{
					return null;
				}
			}
		}

		#endregion

		#region Events

		public event EventHandler<BindingListViewBindEventArgs> BindItem;

		#endregion

		#endregion

		#region Control

		protected override void OnHandleDestroyed(EventArgs e)
		{
			DataSource = null;
			base.OnHandleDestroyed(e);
		}

		#endregion
	}
}
