﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms Library                            */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Design;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Windows.Forms;

namespace Nabu.Forms
{
	public sealed class ProgressDisk : Control
	{
		#region ProgressDisk

		#region Fields

		private int _pieceCount = 12;
		private float _pieceWidth = 0.9F;
		private float _pieceCircularity = 0.66666667F;
		private float _pieceAngularity = 0.33333333F;
		private float _activePieceScale = 1.2F;
		private float _innerCircleRadius = 10.0F;
		private float _outerCircleRadius = 50.0F;
		private float _innerBorderRadius = 15.0F;
		private float _outerBorderRadius = 45.0F;
		private int _pieceTrain = 12;
		private bool _symmetricTrain;
		private int _currentPiece;
		private Bitmap[] _bitmaps;
		private Color _colorFrom = Color.Transparent;
		private Color _colorTo = Color.FromKnownColor(KnownColor.HotTrack);

		#endregion

		#region Methods

		private static Color AverageColor(
			Color colorA,
			Color colorB,
			float weightA)
		{
			return Color.FromArgb(
				(int)(colorA.A * weightA + colorB.A * (1.0F - weightA)),
				(int)(colorA.R * weightA + colorB.R * (1.0F - weightA)),
				(int)(colorA.G * weightA + colorB.G * (1.0F - weightA)),
				(int)(colorA.B * weightA + colorB.B * (1.0F - weightA)));
		}

		private static float AverageNumber(
			float valueA,
			float valueB,
			float weightA)
		{
			return valueA * weightA + valueB * (1.0F - weightA);
		}

		private void UpdateBitmaps()
		{
			if (this._pieceCount > 0)
			{
				this._bitmaps = new Bitmap[this._pieceCount];

				Size clientSize = base.ClientSize;
				PointF clientCenter =
					new PointF(
						clientSize.Width / 2.0F,
						clientSize.Height / 2.0F);
				float angleSize = (3.141593F / this._pieceCount) * this._pieceWidth;
				float angleStep = 6.283185F / this._pieceCount;

				for (int indexBitmap = 0; indexBitmap < this._pieceCount; indexBitmap++)
				{
					this._bitmaps[indexBitmap] =
						new Bitmap(
							clientSize.Width,
							clientSize.Height,
							PixelFormat.Format32bppArgb);

					using (Graphics graphics = Graphics.FromImage(this._bitmaps[indexBitmap]))
					{
						graphics.FillRectangle(
							new SolidBrush(base.BackColor),
							base.ClientRectangle);

						for (int indexCurrent = 0; indexCurrent < this._pieceCount; indexCurrent++)
						{
							float angleCenter = (indexCurrent + indexBitmap) * angleStep - 1.570796F;
							float anglePieceLeft = angleCenter - this._pieceCircularity * angleSize;
							float anglePieceRight = angleCenter + this._pieceCircularity * angleSize;
							float angleBorderLeft = angleCenter - angleSize;
							float angleBorderRight = angleCenter + angleSize;
							Color pieceColor;
							float pieceScale;

							if (this._symmetricTrain)
							{
								int pieceTrain =
									Math.Min(
										(this._pieceCount - indexCurrent),
										indexCurrent) % this._pieceCount;

								if (pieceTrain > this._pieceTrain)
								{
									pieceColor = this._colorFrom;
									pieceScale = 1.0F / this._activePieceScale;
								}
								else
								{
									float weight = (float)pieceTrain / (float)this._pieceTrain;

									pieceColor =
										AverageColor(
											this._colorFrom,
											this._colorTo,
											weight);
									pieceScale =
										AverageNumber(
											1.0F / this._activePieceScale,
											1.0F,
											weight);
								}
							}
							else
							{
								int pieceTrain = (this._pieceCount - indexCurrent) % this._pieceCount;

								if (pieceTrain > this._pieceTrain)
								{
									pieceColor = this._colorFrom;
									pieceScale = 1.0F / this._activePieceScale;
								}
								else
								{
									float weight = (float)pieceTrain / (float)this._pieceTrain;

									pieceColor =
										AverageColor(
											this._colorFrom,
											this._colorTo,
											weight);
									pieceScale =
										AverageNumber(
											1.0F / this._activePieceScale,
											1.0F,
											weight);
								}
							}

							PointF[] points = new PointF[]
							{
								new PointF(
									clientCenter.X + pieceScale * this._outerCircleRadius * (float)Math.Cos(anglePieceRight),
									clientCenter.Y + pieceScale * this._outerCircleRadius * (float)Math.Sin(anglePieceRight)),
								new PointF(
									clientCenter.X + pieceScale * this._outerBorderRadius * (float)Math.Cos(angleBorderRight),
									clientCenter.Y + pieceScale * this._outerBorderRadius * (float)Math.Sin(angleBorderRight)),
								new PointF(
									clientCenter.X + pieceScale * this._innerBorderRadius * (float)Math.Cos(angleBorderRight),
									clientCenter.Y + pieceScale * this._innerBorderRadius * (float)Math.Sin(angleBorderRight)),
								new PointF(
									clientCenter.X + pieceScale * this._innerCircleRadius * (float)Math.Cos(anglePieceRight),
									clientCenter.Y + pieceScale * this._innerCircleRadius * (float)Math.Sin(anglePieceRight)),
								new PointF(
									clientCenter.X + pieceScale * this._innerCircleRadius * (float)Math.Cos(anglePieceLeft),
									clientCenter.Y + pieceScale * this._innerCircleRadius * (float)Math.Sin(anglePieceLeft)),
								new PointF(
									clientCenter.X + pieceScale * this._innerBorderRadius * (float)Math.Cos(angleBorderLeft),
									clientCenter.Y + pieceScale * this._innerBorderRadius * (float)Math.Sin(angleBorderLeft)),
								new PointF(
									clientCenter.X + pieceScale * this._outerBorderRadius * (float)Math.Cos(angleBorderLeft),
									clientCenter.Y + pieceScale * this._outerBorderRadius * (float)Math.Sin(angleBorderLeft)),
								new PointF(
									clientCenter.X + pieceScale * this._outerCircleRadius * (float)Math.Cos(anglePieceLeft),
									clientCenter.Y + pieceScale * this._outerCircleRadius * (float)Math.Sin(anglePieceLeft)),
							};

							graphics.CompositingQuality = CompositingQuality.HighQuality;
							graphics.InterpolationMode = InterpolationMode.HighQualityBicubic;
							graphics.SmoothingMode = SmoothingMode.HighQuality;
							graphics.FillClosedCurve(
								new SolidBrush(pieceColor),
								points,
								FillMode.Winding,
								this._pieceAngularity);
						}
					}
				}
			}
			else
			{
				this._bitmaps = null;
			}
		}

		public ProgressDisk()
		{
			base.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
		}

		#endregion

		#region Properties

		[Category("Layout")]
		public int PieceCount
		{
			get
			{
				return this._pieceCount;
			}

			set
			{
				this._pieceCount = value;
				this._bitmaps = null;
				base.Refresh();
			}
		}

		[Category("Layout")]
		public float PieceWidth
		{
			get
			{
				return this._pieceWidth;
			}

			set
			{
				this._pieceWidth = value;
				this._bitmaps = null;
				base.Refresh();
			}
		}

		[Category("Layout")]
		public float PieceCircularity
		{
			get
			{
				return this._pieceCircularity;
			}

			set
			{
				this._pieceCircularity = value;
				this._bitmaps = null;
				base.Refresh();
			}
		}

		[Category("Layout")]
		public float PieceAngularity
		{
			get
			{
				return this._pieceAngularity;
			}

			set
			{
				this._pieceAngularity = value;
				this._bitmaps = null;
				base.Refresh();
			}
		}

		[Category("Layout")]
		public float InnerCircleRadius
		{
			get
			{
				return this._innerCircleRadius;
			}

			set
			{
				this._innerCircleRadius = value;
				this._bitmaps = null;
				base.Refresh();
			}
		}

		[Category("Layout")]
		public float OuterCircleRadius
		{
			get
			{
				return this._outerCircleRadius;
			}

			set
			{
				this._outerCircleRadius = value;
				this._bitmaps = null;
				base.Refresh();
			}
		}

		[Category("Layout")]
		public float InnerBorderRadius
		{
			get
			{
				return this._innerBorderRadius;
			}

			set
			{
				this._innerBorderRadius = value;
				this._bitmaps = null;
				base.Refresh();
			}
		}

		[Category("Layout")]
		public float OuterBorderRadius
		{
			get
			{
				return this._outerBorderRadius;
			}

			set
			{
				this._outerBorderRadius = value;
				this._bitmaps = null;
				base.Refresh();
			}
		}

		[Category("Layout")]
		public float ActivePieceScale
		{
			get
			{
				return this._activePieceScale;
			}

			set
			{
				this._activePieceScale = value;
				this._bitmaps = null;
				base.Refresh();
			}
		}

		[Category("Layout")]
		public int PieceTrain
		{
			get
			{
				return this._pieceTrain;
			}

			set
			{
				this._pieceTrain = value;
				this._bitmaps = null;
				base.Refresh();
			}
		}

		[Category("Layout")]
		public bool SymmetricTrain
		{
			get
			{
				return this._symmetricTrain;
			}

			set
			{
				this._symmetricTrain = value;
				this._bitmaps = null;
				base.Refresh();
			}
		}

		[Category("Appearance")]
		public Color ColorFrom
		{
			get
			{
				return this._colorFrom;
			}

			set
			{
				this._colorFrom = value;
				this._bitmaps = null;
				base.Refresh();
			}
		}

		[Category("Appearance")]
		public Color ColorTo
		{
			get
			{
				return this._colorTo;
			}

			set
			{
				this._colorTo = value;
				this._bitmaps = null;
				base.Refresh();
			}
		}

		public int CurrentPiece
		{
			get
			{
				return this._currentPiece;
			}

			set
			{
				this._currentPiece = value;
				base.Refresh();
			}
		}

		#endregion

		#endregion

		#region Control

		#region Methods

		protected override void OnClientSizeChanged(EventArgs e)
		{
			this._bitmaps = null;
			base.Refresh();
			base.OnClientSizeChanged(e);
		}

		protected override void OnForeColorChanged(EventArgs e)
		{
			this._bitmaps = null;
			base.Refresh();
			base.OnForeColorChanged(e);
		}

		protected override void OnBackColorChanged(EventArgs e)
		{
			this._bitmaps = null;
			base.Refresh();
			base.OnBackColorChanged(e);
		}

		protected override void OnTextChanged(EventArgs e)
		{
			base.Refresh();
			base.OnTextChanged(e);
		}

		protected override void OnPaint(PaintEventArgs e)
		{
			if (this._bitmaps == null)
			{
				UpdateBitmaps();
			}

			e.Graphics.CompositingQuality = CompositingQuality.HighQuality;
			e.Graphics.InterpolationMode = InterpolationMode.HighQualityBicubic;
			e.Graphics.SmoothingMode = SmoothingMode.HighQuality;
			e.Graphics.DrawImage(
				this._bitmaps[this._currentPiece % this._bitmaps.Length],
				Point.Empty);

			StringFormat stringFormat = new StringFormat(StringFormatFlags.NoWrap);

			stringFormat.Alignment = StringAlignment.Center;
			stringFormat.LineAlignment = StringAlignment.Center;
			stringFormat.Trimming = StringTrimming.None;

			e.Graphics.DrawString(
				base.Text,
				base.Font,
				new SolidBrush(base.ForeColor),
				(RectangleF)base.ClientRectangle,
				stringFormat);
		}

		#endregion

		#region Properties

		[Editor("System.ComponentModel.Design.MultilineStringEditor, System.Design, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a", typeof(UITypeEditor))]
		public override string Text
		{
			get { return base.Text; }
			set { base.Text = value; }
		}

		#endregion

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
