﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Forms Library                            */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.ComponentModel;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Security.Permissions;
using System.Windows.Forms;

namespace Nabu.Forms
{
	public abstract class ScrollableView : Control
	{
		#region ScrollableView

		#region Fields

		private Size _contentSize;
		private Point _contentPosition;
		private float _contentZoom;
		private float _minimumZoom;
		private float _maximumZoom;
		private Size _scrollRange;
		private Size _scrollPage;

		#endregion

		#region Methods

		#region Utility

		private void SetScrollInfo(
			ScrollBarID barID,
			int barMin,
			int barMax,
			int barPage,
			bool redraw)
		{
			ScrollInfo scrollInfo = new ScrollInfo();

			scrollInfo.Size = Marshal.SizeOf(scrollInfo);
			scrollInfo.Mask = ScrollBarInfoFlag.Range | ScrollBarInfoFlag.Page;
			scrollInfo.Min = barMin;
			scrollInfo.Max = barMax;
			scrollInfo.Page = barPage;

			NativeMethods.Scroll_SetInfo(
				base.Handle,
				barID,
				ref scrollInfo,
				redraw);
		}

		private void SetScrollPosition(
			ScrollBarID barID,
			int newPosition)
		{
			switch (barID)
			{
				case ScrollBarID.Horizontal:
					{
						newPosition = Math.Max(0, Math.Min(newPosition, this._scrollRange.Width - this._scrollPage.Width));

						if (this._contentPosition.X != newPosition)
						{
							ScrollInfo scrollInfo = new ScrollInfo();

							scrollInfo.Size = Marshal.SizeOf(scrollInfo);
							scrollInfo.Mask = ScrollBarInfoFlag.Position;
							scrollInfo.Position = newPosition;

							NativeMethods.Scroll_SetInfo(
								base.Handle,
								ScrollBarID.Horizontal,
								ref scrollInfo,
								true);

							this._contentPosition.X = newPosition;
						}
					}
					break;
				case ScrollBarID.Vertical:
					{
						newPosition = Math.Max(0, Math.Min(newPosition, this._scrollRange.Height - this._scrollPage.Height));

						if (this._contentPosition.Y != newPosition)
						{
							ScrollInfo scrollInfo = new ScrollInfo();

							scrollInfo.Size = Marshal.SizeOf(scrollInfo);
							scrollInfo.Mask = ScrollBarInfoFlag.Position;
							scrollInfo.Position = newPosition;

							NativeMethods.Scroll_SetInfo(
								base.Handle,
								ScrollBarID.Vertical,
								ref scrollInfo,
								true);

							this._contentPosition.Y = newPosition;
						}
					}
					break;
			}
		}

		private void UpdateScrollInfo()
		{
			if (base.IsHandleCreated && (!this._contentSize.IsEmpty))
			{
				this._scrollPage =
					new Size(
						base.ClientSize.Width,
						base.ClientSize.Height);

				SizeF rangeMargin =
					new SizeF(
						this._scrollPage.Width * (this._contentZoom - 1) / this._contentZoom,
						this._scrollPage.Height * (this._contentZoom - 1) / this._contentZoom);

				this._scrollRange =
					new Size(
						(int)(this._contentSize.Width + rangeMargin.Width + 0.4999F),
						(int)(this._contentSize.Height + rangeMargin.Height + 0.4999F));

				SetScrollInfo(
					ScrollBarID.Horizontal,
					0,
					this._scrollRange.Width,
					this._scrollPage.Width,
					base.IsHandleCreated);
				SetScrollInfo(
					ScrollBarID.Vertical,
					0,
					this._scrollRange.Height,
					this._scrollPage.Height,
					base.IsHandleCreated);

				RectangleF viewPort = this.ViewPort;

				if (viewPort.Left < 0)
				{
					SetScrollPosition(
						ScrollBarID.Horizontal,
						(int)(this._contentPosition.X - viewPort.Left));
				}

				if (viewPort.Right > this._contentSize.Width)
				{
					SetScrollPosition(
						ScrollBarID.Horizontal,
						(int)(this._contentPosition.X + viewPort.Right - this._contentSize.Width));
				}

				if (viewPort.Top < 0)
				{
					SetScrollPosition(
						ScrollBarID.Vertical,
						(int)(this._contentPosition.Y - viewPort.Top));
				}

				if (viewPort.Bottom > this._contentSize.Height)
				{
					SetScrollPosition(
						ScrollBarID.Vertical,
						(int)(this._contentPosition.Y + viewPort.Bottom - this._contentSize.Height));
				}
			}
		}

		#endregion

		#region Event Raisers

		protected virtual void OnContentSizeChanged(EventArgs e)
		{
			if (this.ContentSizeChanged != null)
			{
				this.ContentSizeChanged(this, e);
			}
		}

		protected virtual void OnContentZoomChanged(EventArgs e)
		{
			if (this.ContentZoomChanged != null)
			{
				this.ContentZoomChanged(this, e);
			}
		}

		protected virtual void OnContentPositionChanged(EventArgs e)
		{
			if (this.ContentPositionChanged != null)
			{
				this.ContentPositionChanged(this, e);
			}
		}

		protected virtual void OnScroll(ScrollEventArgs e)
		{
			if (this.Scroll != null)
			{
				this.Scroll(this, e);
			}
		}

		protected virtual void OnNonClientMouseDown(NonClientMouseEventArgs e)
		{
			if (this.NonClientMouseDown != null)
			{
				this.NonClientMouseDown(this, e);
			}
		}

		protected virtual void OnNonClientMouseUp(NonClientMouseEventArgs e)
		{
			if (this.NonClientMouseUp != null)
			{
				this.NonClientMouseUp(this, e);
			}
		}

		protected virtual void OnNonClientMouseLeave(EventArgs e)
		{
			if (this.NonClientMouseLeave != null)
			{
				this.NonClientMouseLeave(this, e);
			}
		}

		#endregion

		#region Constructors

		public ScrollableView()
		{
			this._contentZoom = 1.0000F;
			this._minimumZoom = 0.0625F;
			this._maximumZoom = 16.000F;
		}

		#endregion

		#region Actions

		public void SetContentZoom(Point visibleCenter, float contentZoom)
		{
			contentZoom = Math.Max(this._minimumZoom, Math.Min(contentZoom, this._maximumZoom));

			if (contentZoom != this._contentZoom)
			{
				PointF oldContentCenter =
					new PointF(
						this._contentPosition.X + visibleCenter.X / this._contentZoom,
						this._contentPosition.Y + visibleCenter.Y / this._contentZoom);
				float oldContentZoom = this._contentZoom;

				SetScrollPosition(
					ScrollBarID.Horizontal,
					(int)(this._contentPosition.X + visibleCenter.X * (contentZoom - this._contentZoom) / (contentZoom * this._contentZoom)));
				SetScrollPosition(
					ScrollBarID.Vertical,
					(int)(this._contentPosition.Y + visibleCenter.Y * (contentZoom - this._contentZoom) / (contentZoom * this._contentZoom)));

				this._contentZoom = contentZoom;
				UpdateScrollInfo();
				OnContentZoomChanged(EventArgs.Empty);
			}
		}

		#endregion

		#endregion

		#region Properties

		[Browsable(false)]
		[Category("Scroll")]
		public Size ContentSize
		{
			get
			{
				return this._contentSize;
			}

			set
			{
				if (this._contentSize != value)
				{
					this._contentSize = value;
					UpdateScrollInfo();
					OnContentSizeChanged(EventArgs.Empty);
				}
			}
		}

		[Category("Scroll")]
		public Point ContentPosition
		{
			get
			{
				return this._contentPosition;
			}

			set
			{
				SetScrollPosition(ScrollBarID.Horizontal, value.X);
				SetScrollPosition(ScrollBarID.Vertical, value.Y);
				UpdateScrollInfo();
				OnContentPositionChanged(EventArgs.Empty);
			}
		}

		[Category("Scroll")]
		public float ContentZoom
		{
			get
			{
				return this._contentZoom;
			}

			set
			{
				value = Math.Max(this._minimumZoom, Math.Min(value, this._maximumZoom));

				if (this._contentZoom != value)
				{
					UpdateScrollInfo();
					OnContentZoomChanged(EventArgs.Empty);
				}
			}
		}

		[Category("Scroll")]
		public float MinimumZoom
		{
			get
			{
				return this._minimumZoom;
			}

			set
			{
				if (this._minimumZoom != value)
				{
					this._minimumZoom = value;

					if (this._contentZoom < this._minimumZoom)
					{
						this._contentZoom = this._minimumZoom;

						UpdateScrollInfo();
						OnContentZoomChanged(EventArgs.Empty);
					}
				}
			}
		}

		[Category("Scroll")]
		public float MaximumZoom
		{
			get
			{
				return this._maximumZoom;
			}

			set
			{
				if (this._maximumZoom != value)
				{
					this._maximumZoom = value;

					if (this._contentZoom > this._maximumZoom)
					{
						this._contentZoom = this._maximumZoom;

						UpdateScrollInfo();
						OnContentZoomChanged(EventArgs.Empty);
					}
				}
			}
		}

		[Browsable(false)]
		public RectangleF ViewPort
		{
			get
			{
				return
					new RectangleF(
						this._contentPosition.X,
						this._contentPosition.Y,
						base.ClientSize.Width / this._contentZoom,
						base.ClientSize.Height / this._contentZoom);
			}
		}

		#endregion

		#region Events

		[Category("Scroll")]
		public event EventHandler ContentSizeChanged;
		[Category("Scroll")]
		public event EventHandler ContentPositionChanged;
		[Category("Scroll")]
		public event EventHandler ContentZoomChanged;
		[Category("Scroll")]
		public event EventHandler<ScrollEventArgs> Scroll;
		[Category("NonClient")]
		public event EventHandler<NonClientMouseEventArgs> NonClientMouseDown;
		[Category("NonClient")]
		public event EventHandler<NonClientMouseEventArgs> NonClientMouseUp;
		[Category("NonClient")]
		public event EventHandler NonClientMouseLeave;

		#endregion

		#endregion

		#region Control

		#region Methods

		protected override void OnHandleCreated(EventArgs e)
		{
			NativeMethods.ScrollBar_Show(base.Handle, ScrollBarID.Both, true);
			UpdateScrollInfo();

			base.OnHandleCreated(e);
		}

		protected override void OnSizeChanged(EventArgs e)
		{
			UpdateScrollInfo();

			base.OnSizeChanged(e);
		}

		[SecurityPermission(SecurityAction.LinkDemand, Flags = SecurityPermissionFlag.UnmanagedCode)]
		protected override void WndProc(ref Message message)
		{
			Point scrollPosition = this._contentPosition;

			switch ((WindowMessage)message.Msg)
			{
				case WindowMessage.WM_HSCROLL:
					switch ((ScrollBarNotificationCode)(((int)message.WParam) & 0xFFFF))
					{
						case ScrollBarNotificationCode.LineLeft:
							SetScrollPosition(ScrollBarID.Horizontal, scrollPosition.X - 1);

							if (scrollPosition.X != this._contentPosition.X)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.SmallDecrement,
										scrollPosition.X,
										this._contentPosition.X,
										ScrollOrientation.HorizontalScroll));
							}
							break;
						case ScrollBarNotificationCode.LineRight:
							SetScrollPosition(ScrollBarID.Horizontal, scrollPosition.X + 1);

							if (scrollPosition.X != this._contentPosition.X)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.SmallIncrement,
										scrollPosition.X,
										this._contentPosition.X,
										ScrollOrientation.HorizontalScroll));
							}
							break;
						case ScrollBarNotificationCode.PageLeft:
							SetScrollPosition(ScrollBarID.Horizontal, scrollPosition.X - this._scrollPage.Width);

							if (scrollPosition.X != this._contentPosition.X)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.LargeDecrement,
										scrollPosition.X,
										this._contentPosition.X,
										ScrollOrientation.HorizontalScroll));
							}
							break;
						case ScrollBarNotificationCode.PageRight:
							SetScrollPosition(ScrollBarID.Horizontal, scrollPosition.X + this._scrollPage.Width);

							if (scrollPosition.X != this._contentPosition.X)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.LargeIncrement,
										scrollPosition.X,
										this._contentPosition.X,
										ScrollOrientation.HorizontalScroll));
							}
							break;
						case ScrollBarNotificationCode.ThumbPosition:
							SetScrollPosition(ScrollBarID.Horizontal, (((int)message.WParam) >> 16));

							if (scrollPosition.X != this._contentPosition.X)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.ThumbPosition,
										scrollPosition.X,
										this._contentPosition.X,
										ScrollOrientation.HorizontalScroll));
							}
							break;
						case ScrollBarNotificationCode.ThumbTrack:
							SetScrollPosition(ScrollBarID.Horizontal, (((int)message.WParam) >> 16));

							if (scrollPosition.X != this._contentPosition.X)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.ThumbTrack,
										scrollPosition.X,
										this._contentPosition.X,
										ScrollOrientation.HorizontalScroll));
							}
							break;
						case ScrollBarNotificationCode.Left:
							SetScrollPosition(ScrollBarID.Horizontal, 0);

							if (scrollPosition.X != this._contentPosition.X)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.ThumbTrack,
										scrollPosition.X,
										this._contentPosition.X,
										ScrollOrientation.HorizontalScroll));
							}
							break;
						case ScrollBarNotificationCode.Right:
							SetScrollPosition(ScrollBarID.Horizontal, this._scrollRange.Width);

							if (scrollPosition.X != this._contentPosition.X)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.ThumbTrack,
										scrollPosition.X,
										this._contentPosition.X,
										ScrollOrientation.HorizontalScroll));
							}
							break;
						case ScrollBarNotificationCode.End:
							OnScroll(
								new ScrollEventArgs(
									ScrollEventType.EndScroll,
									this._contentPosition.X,
									ScrollOrientation.HorizontalScroll));
							break;
					}
					break;
				case WindowMessage.WM_VSCROLL:
					switch ((ScrollBarNotificationCode)(((int)message.WParam) & 0xFFFF))
					{
						case ScrollBarNotificationCode.LineLeft:
							SetScrollPosition(ScrollBarID.Vertical, scrollPosition.Y - 1);

							if (scrollPosition.Y != this._contentPosition.Y)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.SmallDecrement,
										scrollPosition.Y,
										this._contentPosition.Y,
										ScrollOrientation.VerticalScroll));
							}
							break;
						case ScrollBarNotificationCode.LineRight:
							SetScrollPosition(ScrollBarID.Vertical, scrollPosition.Y + 1);

							if (scrollPosition.Y != this._contentPosition.Y)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.SmallIncrement,
										scrollPosition.Y,
										this._contentPosition.Y,
										ScrollOrientation.VerticalScroll));
							}
							break;
						case ScrollBarNotificationCode.PageLeft:
							SetScrollPosition(ScrollBarID.Vertical, scrollPosition.Y - this._scrollPage.Height);

							if (scrollPosition.Y != this._contentPosition.Y)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.LargeDecrement,
										scrollPosition.Y,
										this._contentPosition.Y,
										ScrollOrientation.VerticalScroll));
							}
							break;
						case ScrollBarNotificationCode.PageRight:
							SetScrollPosition(ScrollBarID.Vertical, scrollPosition.Y + this._scrollPage.Height);

							if (scrollPosition.Y != this._contentPosition.Y)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.LargeIncrement,
										scrollPosition.Y,
										this._contentPosition.Y,
										ScrollOrientation.VerticalScroll));
							}
							break;
						case ScrollBarNotificationCode.ThumbPosition:
							SetScrollPosition(ScrollBarID.Vertical, (((int)message.WParam) >> 16));

							if (scrollPosition.Y != this._contentPosition.Y)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.ThumbPosition,
										scrollPosition.Y,
										this._contentPosition.Y,
										ScrollOrientation.VerticalScroll));
							}
							break;
						case ScrollBarNotificationCode.ThumbTrack:
							SetScrollPosition(ScrollBarID.Vertical, (((int)message.WParam) >> 16));

							if (scrollPosition.Y != this._contentPosition.Y)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.ThumbTrack,
										scrollPosition.Y,
										this._contentPosition.Y,
										ScrollOrientation.VerticalScroll));
							}
							break;
						case ScrollBarNotificationCode.Left:
							SetScrollPosition(ScrollBarID.Vertical, 0);

							if (scrollPosition.Y != this._contentPosition.Y)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.First,
										scrollPosition.Y,
										this._contentPosition.Y,
										ScrollOrientation.VerticalScroll));
							}
							break;
						case ScrollBarNotificationCode.Right:
							SetScrollPosition(ScrollBarID.Vertical, this._scrollRange.Height);

							if (scrollPosition.Y != this._contentPosition.Y)
							{
								OnScroll(
									new ScrollEventArgs(
										ScrollEventType.Last,
										scrollPosition.Y,
										this._contentPosition.Y,
										ScrollOrientation.VerticalScroll));
							}
							break;
						case ScrollBarNotificationCode.End:
							OnScroll(
								new ScrollEventArgs(
									ScrollEventType.EndScroll,
									this._contentPosition.Y,
									ScrollOrientation.VerticalScroll));
							break;
					}
					break;
				case WindowMessage.WM_NCLBUTTONDOWN:
					OnNonClientMouseDown(
						new NonClientMouseEventArgs(
							MouseButtons.Left,
							0,
							(message.LParam.ToInt32() >> 16),
							(message.LParam.ToInt32() & 0xFFFF),
							0,
							(HitTestCode)((int)message.WParam)));
					break;
				case WindowMessage.WM_NCLBUTTONUP:
					OnNonClientMouseUp(
						new NonClientMouseEventArgs(
							MouseButtons.Left,
							0,
							(message.LParam.ToInt32() >> 16),
							(message.LParam.ToInt32() & 0xFFFF),
							0,
							(HitTestCode)((int)message.WParam)));
					break;
				case WindowMessage.WM_NCRBUTTONDOWN:
					OnNonClientMouseDown(
						new NonClientMouseEventArgs(
							MouseButtons.Right,
							0,
							(message.LParam.ToInt32() >> 16),
							(message.LParam.ToInt32() & 0xFFFF),
							0,
							(HitTestCode)((int)message.WParam)));
					break;
				case WindowMessage.WM_NCRBUTTONUP:
					OnNonClientMouseUp(
						new NonClientMouseEventArgs(
							MouseButtons.Right,
							0,
							(message.LParam.ToInt32() >> 16),
							(message.LParam.ToInt32() & 0xFFFF),
							0,
							(HitTestCode)((int)message.WParam)));
					break;
				case WindowMessage.WM_NCMOUSELEAVE:
					OnNonClientMouseLeave(EventArgs.Empty);
					break;
			}

			base.WndProc(ref message);
		}

		#endregion

		#region Properties

		protected override CreateParams CreateParams
		{
			[SecurityPermission(SecurityAction.LinkDemand, Flags = SecurityPermissionFlag.UnmanagedCode)]
			get
			{
				CreateParams createParams = base.CreateParams;

				createParams.Style |= (int)(
					WindowStyle.HScroll |
					WindowStyle.VScroll);

				return createParams;
			}
		}

		#endregion

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
