﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.IO Library                               */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics;
using System.IO;

namespace Nabu.IO
{
	internal sealed class MSCabFileStream : Stream
	{
		#region MSCabFileStream

		#region Constants

		private const string CompressionExceptionMessage = "Unsupported compression type.";

		#endregion

		#region Fields

		private Stream _stream;
		private StreamHelper _helper;
		private int _dataOffset;
		// private MSCabCompressionType _compressionType;
		private int _decompressedSize;
		private int _dataLeft;
		private byte _dataExtraDataSize;

		#endregion

		#region Methods

		private void ReadDataHeader()
		{
			// CRC
			this._helper.ReadU4();
			// Compressed Data
			this._helper.ReadU2();
			// Decompressed Data
			this._dataLeft = this._helper.ReadU2();
			// Extra Data
			this._stream.Seek(this._dataExtraDataSize, SeekOrigin.Current);
		}

		[DebuggerStepThrough]
		public MSCabFileStream(
			Stream stream,
			int dataOffset,
			MSCabCompressionType compressionType,
			int decompressedOffset,
			int decompressedSize,
			byte dataExtraDataSize)
		{
			if (compressionType != MSCabCompressionType.None)
			{
				throw new NotSupportedException(CompressionExceptionMessage);
			}

			this._stream = stream;
			this._helper = new StreamHelper(this._stream);
			this._dataOffset = dataOffset;
			// this._compressionType = compressionType;
			this._decompressedSize = decompressedSize;
			this._dataExtraDataSize = dataExtraDataSize;
			this._stream.Position = this._dataOffset;

			while (decompressedOffset > 0)
			{
				if (this._dataLeft == 0)
				{
					ReadDataHeader();
				}

				if (this._dataLeft > decompressedOffset)
				{
					this._stream.Seek(decompressedOffset, SeekOrigin.Current);
					this._dataLeft -= decompressedOffset;
					decompressedOffset = 0;
				}
				else
				{
					this._stream.Seek(this._dataLeft, SeekOrigin.Current);
					decompressedOffset -= this._dataLeft;
					this._dataLeft = 0;
				}
			}
		}

		#endregion

		#endregion

		#region Stream

		#region Methods

		[DebuggerStepThrough]
		public override int Read(byte[] buffer, int offset, int count)
		{
			if (this._dataLeft == 0)
			{
				ReadDataHeader();
			}

			int result =
				this._stream.Read(
					buffer,
					offset,
					Math.Min(this._dataLeft, count));

			this._dataLeft -= result;

			return result;
		}

		[DebuggerStepThrough]
		public override void Write(byte[] buffer, int offset, int count)
		{
			throw new NotSupportedException();
		}

		[DebuggerStepThrough]
		public override void Flush()
		{
			throw new NotSupportedException();
		}

		[DebuggerStepThrough]
		public override long Seek(long offset, SeekOrigin origin)
		{
			throw new NotSupportedException();
		}

		[DebuggerStepThrough]
		public override void SetLength(long value)
		{
			throw new NotSupportedException();
		}

		#endregion

		#region Properties

		public override bool CanRead
		{
			[DebuggerStepThrough]
			get { return true; }
		}

		public override bool CanSeek
		{
			[DebuggerStepThrough]
			get { return false; }
		}

		public override bool CanWrite
		{
			[DebuggerStepThrough]
			get { return false; }
		}

		public override long Length
		{
			[DebuggerStepThrough]
			get { return this._decompressedSize; }
		}

		public override long Position
		{
			[DebuggerStepThrough]
			get { throw new NotSupportedException(); }
			[DebuggerStepThrough]
			set { throw new NotSupportedException(); }
		}

		#endregion

		#endregion

		#region IDisposable

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				this._helper.Dispose();
			}

			base.Dispose(disposing);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
