﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.IO Library                               */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics;
using System.IO;

namespace Nabu.IO
{
	public sealed class SubStream : Stream
	{
		#region SubStream

		#region Constants

		private const string PositionExceptionMessage = "Position is out of range.";

		#endregion

		#region Fields

		private Stream _stream;
		private long _offset;
		private long _length;

		#endregion

		#region Methods

		[DebuggerStepThrough]
		public SubStream(Stream stream, long offset, long length)
		{
			this._stream = stream;
			this._offset = offset;
			this._length = length;
		}

		#endregion

		#region Properties

		public Stream Stream
		{
			[DebuggerStepThrough]
			get { return this._stream; }
		}

		public long Offset
		{
			[DebuggerStepThrough]
			get { return this._offset; }
		}

		#endregion

		#endregion

		#region Stream

		#region Methods

		public override int Read(byte[] buffer, int offset, int count)
		{
			if (
				(this._stream.Position < this._offset) ||
				((this._stream.Position > this._offset + this._length)))
			{
				throw new InvalidOperationException(PositionExceptionMessage);
			}

			count = Math.Min(count, (int)(this._length + this._offset - this._stream.Position));

			return this._stream.Read(buffer, offset, count);
		}

		public override void Write(byte[] buffer, int offset, int count)
		{
			if (
				(this._stream.Position < this._offset) ||
				((this._stream.Position + count > this._offset + this._length)))
			{
				throw new InvalidOperationException(PositionExceptionMessage);
			}

			this._stream.Write(buffer, offset, count);
		}

		public override long Seek(long offset, SeekOrigin origin)
		{
			switch (origin)
			{
				case SeekOrigin.Begin:
					return this._stream.Seek(
						this._offset + offset,
						SeekOrigin.Begin);
				case SeekOrigin.Current:
					return this._stream.Seek(
						offset,
						SeekOrigin.Current);
				case SeekOrigin.End:
					return this._stream.Seek(
						this._offset + this._length - this._stream.Length + offset,
						SeekOrigin.End);
			}

			return 0;
		}

		public override void Flush()
		{
			this._stream.Flush();
		}

		public override void SetLength(long value)
		{
			throw new NotImplementedException();
		}

		#endregion

		#region Properties

		public override bool CanRead
		{
			[DebuggerStepThrough]
			get { return this._stream.CanRead; }
		}

		public override bool CanSeek
		{
			[DebuggerStepThrough]
			get { return this._stream.CanSeek; }
		}

		public override bool CanWrite
		{
			[DebuggerStepThrough]
			get { return this._stream.CanWrite; }
		}

		public override long Length
		{
			[DebuggerStepThrough]
			get { return this._length; }
		}

		public override long Position
		{
			[DebuggerStepThrough]
			get { return this._stream.Position - this._offset; }
			[DebuggerStepThrough]
			set { this._stream.Position = value + this._offset; }
		}

		#endregion

		#endregion

		#region IDisposable

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				this._stream = null;
			}

			base.Dispose(disposing);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
