﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.IO Library                               */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.IO;
using System.IO.Compression;
using System.Text;

namespace Nabu.IO
{
	[CLSCompliant(false)]
	public abstract class ZipArchiveReaderFileHeaderEventArgs : EventArgs
	{
		#region Fields

		private ushort _decompressorVersion;
		private ZipFileFlags _flags;
		private ZipCompressionMethod _compressionMethod;
		private DateTime _creationTime;
		private DateTime _modificationTime;
		private DateTime _accessTime;
		// private uint _crc32;
		private long _compressedSize;
		private long _decompressedSize;
		private uint _diskNumber;
		private string _filePath;

		#endregion

		#region Methods

		protected void ReadDecompressionData(StreamHelper streamHelper)
		{
			if (streamHelper == null)
			{
				throw new ArgumentNullException("streamHelper");
			}

			this._decompressorVersion = streamHelper.ReadU2();
			this._flags = (ZipFileFlags)streamHelper.ReadU2();
			this._compressionMethod = (ZipCompressionMethod)streamHelper.ReadU2();
		}

		protected void ReadCompressionData(StreamHelper streamHelper)
		{
			if (streamHelper == null)
			{
				throw new ArgumentNullException("streamHelper");
			}

			/* this._crc32 = */
			streamHelper.ReadU4();
			this._compressedSize = streamHelper.ReadU4();
			this._decompressedSize = streamHelper.ReadU4();
		}

		protected void ReadFileTime(StreamHelper streamHelper)
		{
			if (streamHelper == null)
			{
				throw new ArgumentNullException("streamHelper");
			}

			ushort time = streamHelper.ReadU2();
			ushort date = streamHelper.ReadU2();

			this._modificationTime =
				new DateTime(
					1980 + ((date >> 9) & 0x3F),
					(date >> 5) & 0x0F,
					(date >> 0) & 0x1F,
					(time >> 11) & 0x1F,
					(time >> 5) & 0x3F,
					2 * ((time >> 0) & 0x1F));
		}

		protected void ReadFilePath(StreamHelper streamHelper, ushort filePathLength)
		{
			if (streamHelper == null)
			{
				throw new ArgumentNullException("streamHelper");
			}

			byte[] filePathData = streamHelper.ReadArray(filePathLength);

			if ((this._flags & ZipFileFlags.FileIsUtf8) != 0)
			{
				this._filePath = Encoding.UTF8.GetString(filePathData);
			}
			else
			{
				this._filePath = Encoding.ASCII.GetString(filePathData);
			}
		}

		protected void ReadExtraData(Stream stream, ushort extraDataLength)
		{
			if (stream == null)
			{
				throw new ArgumentNullException("stream");
			}

			if (extraDataLength > 0)
			{
				using (Stream extraStream = new SubStream(stream, stream.Position, extraDataLength))
				{
					using (StreamHelper extraHelper = new StreamHelper(extraStream))
					{
						while (extraStream.Position < extraStream.Length)
						{
							ushort headerID = extraHelper.ReadU2();

							switch (headerID)
							{
								case ZipExtraDataSignature.Zip64:
									{
										ushort blockSize = extraHelper.ReadU2();

										this._decompressedSize = extraHelper.ReadI8();
										this._compressedSize = extraHelper.ReadI8();

										/* long localHeaderOffset = */
										extraHelper.ReadI8();

										this._diskNumber = extraHelper.ReadU4();
									}
									break;
								case ZipExtraDataSignature.NTFS:
									{
										ushort blockSize = extraHelper.ReadU2();
										/* uint reserved = */
										extraHelper.ReadU4();

										using (Stream ntfsStream = new SubStream(stream, stream.Position, blockSize - 4))
										{
											using (StreamHelper ntfsHelper = new StreamHelper(ntfsStream))
											{
												for (; ntfsStream.Position < ntfsStream.Length; )
												{
													ushort tagID = ntfsHelper.ReadU2();
													ushort tagSize = ntfsHelper.ReadU2();

													switch (tagID)
													{
														case 1:
															this._modificationTime = DateTime.FromFileTime(ntfsHelper.ReadI8());
															this._accessTime = DateTime.FromFileTime(ntfsHelper.ReadI8());
															this._creationTime = DateTime.FromFileTime(ntfsHelper.ReadI8());
															break;
														default:
															ntfsStream.Seek(tagSize, SeekOrigin.Current);
															break;
													}
												}
											}
										}
									}
									break;
								default:
									extraStream.Seek(0, SeekOrigin.End);
									break;
							}
						}
					}

					extraStream.Seek(0, SeekOrigin.End);
				}
			}
		}

		#endregion

		#region Properties

		#region Decompressor

		public ushort DecompressorVersion
		{
			[DebuggerStepThrough]
			get { return this._decompressorVersion; }
		}

		public ZipCompressionMethod CompressionMethod
		{
			[DebuggerStepThrough]
			get { return this._compressionMethod; }
		}

		public bool IsEncrypted
		{
			[DebuggerStepThrough]
			get { return (this._flags & ZipFileFlags.FileIsEncrypted) != 0; }
		}

		public bool IsStronglyEncrypted
		{
			[DebuggerStepThrough]
			get { return (this._flags & ZipFileFlags.EncryptionIsStrong) != 0; }
		}

		public bool IsDataDescriptorUsed
		{
			[DebuggerStepThrough]
			get { return (this._flags & ZipFileFlags.DataDescriptorIsUsed) != 0; }
		}

		#endregion

		#region Compressor

		public long CompressedSize
		{
			[DebuggerStepThrough]
			get { return this._compressedSize; }
		}

		public long DecompressedSize
		{
			[DebuggerStepThrough]
			get { return this._decompressedSize; }
		}

		#endregion

		#region File

		public string FilePath
		{
			[DebuggerStepThrough]
			get { return this._filePath; }
		}

		public DateTime CreationTime
		{
			[DebuggerStepThrough]
			get { return this._creationTime; }
		}

		public DateTime ModificationTime
		{
			[DebuggerStepThrough]
			get { return this._modificationTime; }
		}

		public DateTime AccessTime
		{
			[DebuggerStepThrough]
			get { return this._accessTime; }
		}

		public uint DiskNumber
		{
			[DebuggerStepThrough]
			get { return this._diskNumber; }
			[DebuggerStepThrough]
			protected set { this._diskNumber = value; }
		}

		#endregion

		#endregion
	}

	[CLSCompliant(false)]
	public sealed class ZipArchiveReaderFileLocalHeaderEventArgs :
		ZipArchiveReaderFileHeaderEventArgs,
		IDisposable
	{
		#region ZipArchiveReaderFileLocalHeaderEventArgs

		#region Constants

		private const string CompressionExceptionMessage = "Unsupported compression type.";

		#endregion

		#region Fields

		private SubStream _stream;

		#endregion

		#region Methods

		[DebuggerStepThrough]
		internal ZipArchiveReaderFileLocalHeaderEventArgs(Stream stream)
		{
			using (StreamHelper streamHelper = new StreamHelper(stream))
			{
				ReadDecompressionData(streamHelper);
				ReadFileTime(streamHelper);
				ReadCompressionData(streamHelper);

				ushort filePathLength = streamHelper.ReadU2();
				ushort extraDataLength = streamHelper.ReadU2();

				ReadFilePath(streamHelper, filePathLength);
				ReadExtraData(stream, extraDataLength);
			}

			this._stream = new SubStream(stream, stream.Position, base.CompressedSize);
		}

		[SuppressMessage("Microsoft.Design", "CA1024:UsePropertiesWhereAppropriate")]
		public Stream GetDecompressedStream()
		{
			this._stream.Position = 0;

			if (base.CompressionMethod == ZipCompressionMethod.Store)
			{
				return this._stream;
			}
			else if (base.CompressionMethod == ZipCompressionMethod.Deflated)
			{
				return new DeflateStream(this._stream, CompressionMode.Decompress, true);
			}
			else
			{
				throw new NotSupportedException(CompressionExceptionMessage);
			}
		}

		#endregion

		#endregion

		#region IDisposable

		public void Dispose()
		{
			if (this._stream != null)
			{
				this._stream.Dispose();
				GC.SuppressFinalize(this);
			}
		}

		#endregion
	}

	[CLSCompliant(false)]
	public sealed class ZipArchiveReaderFileCentralHeaderEventArgs : ZipArchiveReaderFileHeaderEventArgs
	{
		#region Fields

		private ushort _compressorVersion;
		private uint _localHeaderOffset;

		#endregion

		#region Methods

		[DebuggerStepThrough]
		internal ZipArchiveReaderFileCentralHeaderEventArgs(Stream stream)
		{
			using (StreamHelper streamHelper = new StreamHelper(stream))
			{
				this._compressorVersion = streamHelper.ReadU2();

				ReadDecompressionData(streamHelper);
				ReadFileTime(streamHelper);
				ReadCompressionData(streamHelper);

				ushort filePathLength = streamHelper.ReadU2();
				ushort extraDataLength = streamHelper.ReadU2();
				ushort fileCommentLength = streamHelper.ReadU2();

				base.DiskNumber = streamHelper.ReadU2();

				/* ushort internalAttributes = */
				streamHelper.ReadU2();
				/* uint externalAttributes = */
				streamHelper.ReadU4();

				this._localHeaderOffset = streamHelper.ReadU4();

				ReadFilePath(streamHelper, filePathLength);
				ReadExtraData(stream, extraDataLength);

				/* byte[] fileComment = */
				streamHelper.ReadArray(fileCommentLength);
			}
		}

		#endregion

		#region Properties

		public ushort CompressorVersion
		{
			[DebuggerStepThrough]
			get { return this._compressorVersion; }
		}

		public uint LocalHeaderOffset
		{
			[DebuggerStepThrough]
			get { return this._localHeaderOffset; }
		}

		#endregion
	}

	[CLSCompliant(false)]
	public sealed class ZipArchiveFooterEventArgs : EventArgs
	{
		#region Fields

		private ushort _currentDiskNumber;
		private ushort _startDiskNumber;
		private ushort _diskEntryNumber;
		private ushort _totalEntryNumber;
		// private uint _size;
		// private uint _offset;

		#endregion

		#region Methods

		[DebuggerStepThrough]
		internal ZipArchiveFooterEventArgs(Stream stream)
		{
			using (StreamHelper streamHelper = new StreamHelper(stream))
			{
				this._currentDiskNumber = streamHelper.ReadU2();
				this._startDiskNumber = streamHelper.ReadU2();
				this._diskEntryNumber = streamHelper.ReadU2();
				this._totalEntryNumber = streamHelper.ReadU2();
				/* this._size = */
				streamHelper.ReadU4();
				/* this._offset = */
				streamHelper.ReadU4();

				ushort archiveCommentLength = streamHelper.ReadU2();

				/* byte[] archiveComment = */
				streamHelper.ReadArray(archiveCommentLength);
			}
		}

		#endregion

		#region Properties

		public ushort CurrentDiskNumber
		{
			[DebuggerStepThrough]
			get { return this._currentDiskNumber; }
		}

		public ushort StartDiskNumber
		{
			[DebuggerStepThrough]
			get { return this._startDiskNumber; }
		}

		public ushort DiskEntryNumber
		{
			[DebuggerStepThrough]
			get { return this._diskEntryNumber; }
		}

		public ushort TotalEntryNumber
		{
			[DebuggerStepThrough]
			get { return this._totalEntryNumber; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
