﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.IO Library                               */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text;

using Nabu.Collections;

namespace Nabu.IO
{
	public sealed class ZipArchiveWriter : DisposableBase
	{
		#region ZipArchiveWriter

		#region Fields

		private DisposableList<ZipArchiveWriterEntry> _entries;

		#endregion

		#region Methods

		public ZipArchiveWriter()
		{
			this._entries = new DisposableList<ZipArchiveWriterEntry>();
		}

		public ZipArchiveWriter(int capacity)
		{
			this._entries = new DisposableList<ZipArchiveWriterEntry>(capacity);
		}

		private void WriteLocalData(Stream stream)
		{
			foreach (ZipArchiveWriterEntry entry in this._entries)
			{
				entry.WriteLocalHeader(stream);
				entry.WriteData(stream);
			}
		}

		private void WriteCentralData(Stream stream, string archiveComment)
		{
			long centalDirectoryFrom = stream.Position;

			foreach (ZipArchiveWriterEntry entry in this._entries)
			{
				entry.WriteCentralHeader(stream);
			}

			long centralDirectoryTo = stream.Position;
			byte[] archiveCommentData = Encoding.ASCII.GetBytes(archiveComment);

			using (StreamHelper streamHelper = new StreamHelper(stream))
			{
				// Signature
				streamHelper.WriteU4(ZipBlockSignature.CentralEnd);
				// Current Disk Number
				streamHelper.WriteU2(0);
				// Central Start Disk Number
				streamHelper.WriteU2(0);
				// Disk Entries
				streamHelper.WriteU2((ushort)this._entries.Count);
				// Total Entries
				streamHelper.WriteU2((ushort)this._entries.Count);
				// Directory Size
				streamHelper.WriteU4((uint)(centralDirectoryTo - centalDirectoryFrom));
				// Directory Offset
				streamHelper.WriteU4((uint)centalDirectoryFrom);
				// Comment Length
				streamHelper.WriteU2((ushort)archiveCommentData.Length);
				// Comment Data
				streamHelper.WriteArray(archiveCommentData);
			}
		}

		public void Write(Stream stream, string comment)
		{
			WriteLocalData(stream);
			WriteCentralData(stream, comment);
			this._entries.Clear();
		}

		public void Write(Stream stream)
		{
			Write(stream, string.Empty);
		}

		#endregion

		#region Properties

		public IList<ZipArchiveWriterEntry> Entries
		{
			[DebuggerStepThrough]
			get { return this._entries; }
		}

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				this._entries.Dispose();
			}

			base.Dispose(disposing);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
