﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.IO Library                               */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics;
using System.IO;
using System.Text;

namespace Nabu.IO
{
	public sealed class ZipArchiveWriterEntry : DisposableBase
	{
		#region ZipArchiveWriterEntry

		#region Fields

		private Stream _externalStream;
		private long _externalSize;

		private string _internalPath;
		private string _internalComment;
		private long _internalSize;
		private DateTime _internalCreationTime;
		private DateTime _internalModificationTime;
		private DateTime _internalAccessTime;
		private uint _internalAttributes;

		private byte[] _internalPathData;
		private byte[] _internalCommentData;
		private bool _isUtf8;
		private bool _isZip64;
		private long _localHeaderOffset;
		private uint _crc32;

		#endregion

		#region Methods

		[DebuggerStepThrough]
		public ZipArchiveWriterEntry(
			string externalPath,
			string internalPath,
			string internalComment)
		{
			FileInfo fileInfo = new FileInfo(externalPath);

			this._externalStream =
				new FileStream(
					externalPath,
					FileMode.Open,
					FileAccess.Read,
					FileShare.Read);
			this._externalSize = fileInfo.Length;
			this._internalPath = internalPath;
			this._internalComment = internalComment;
			this._internalCreationTime = fileInfo.CreationTime;
			this._internalModificationTime = fileInfo.LastWriteTime;
			this._internalAccessTime = fileInfo.LastAccessTime;
			this._internalAttributes = (uint)fileInfo.Attributes;
			this._internalPathData = Encoding.UTF8.GetBytes(this._internalPath);
			this._internalCommentData = Encoding.UTF8.GetBytes(this._internalComment);
			this._isUtf8 = !
				(Array.TrueForAll(this._internalPathData, delegate(byte x) { return x < 128; }) &&
				Array.TrueForAll(this._internalCommentData, delegate(byte x) { return x < 128; }));
			this._isZip64 = this._externalSize >= 0x80000000;
		}

		[DebuggerStepThrough]
		public ZipArchiveWriterEntry(string externalPath, string internalPath)
			: this(externalPath, internalPath, string.Empty)
		{
		}

		[DebuggerStepThrough]
		public ZipArchiveWriterEntry(
			Stream externalStream,
			string internalPath,
			string internalComment)
		{
			DateTime now = DateTime.Now;

			this._externalStream = externalStream;
			this._externalSize = externalStream.Length;
			this._internalPath = internalPath;
			this._internalComment = internalComment;
			this._internalCreationTime = now;
			this._internalModificationTime = now;
			this._internalAccessTime = now;
			// this._internalAttributes = 0;
			this._internalPathData = Encoding.UTF8.GetBytes(this._internalPath);
			this._internalCommentData = Encoding.UTF8.GetBytes(this._internalComment);
			this._isUtf8 = !
				(Array.TrueForAll(this._internalPathData, delegate(byte x) { return x < 128; }) &&
				Array.TrueForAll(this._internalCommentData, delegate(byte x) { return x < 128; }));
			this._isZip64 = this._externalSize >= 0x80000000;
		}

		[DebuggerStepThrough]
		public ZipArchiveWriterEntry(
			Stream externalStream,
			string internalPath)
			: this(externalStream, internalPath, string.Empty)
		{
		}

		[DebuggerStepThrough]
		public ZipArchiveWriterEntry(
			byte[] externalData,
			int index,
			int count,
			string internalPath,
			string internalComment)
		{
			if (externalData == null)
			{
				throw new ArgumentNullException("externalData");
			}

			DateTime now = DateTime.Now;

			this._externalStream = new MemoryStream(externalData, index, count, false);
			this._externalSize = count;
			this._internalPath = internalPath;
			this._internalComment = internalComment;
			this._internalCreationTime = now;
			this._internalModificationTime = now;
			this._internalAccessTime = now;
			// this._internalAttributes = 0;
			this._internalPathData = Encoding.UTF8.GetBytes(this._internalPath);
			this._internalCommentData = Encoding.UTF8.GetBytes(this._internalComment);
			this._isUtf8 = !
				(Array.TrueForAll(this._internalPathData, delegate(byte x) { return x < 128; }) &&
				Array.TrueForAll(this._internalCommentData, delegate(byte x) { return x < 128; }));
			this._isZip64 = this._externalSize >= 0x80000000;
		}

		[DebuggerStepThrough]
		public ZipArchiveWriterEntry(
			byte[] externalData,
			string internalPath,
			string internalComment)
			: this(externalData, 0, externalData.Length, internalPath, internalComment)
		{
		}

		[DebuggerStepThrough]
		public ZipArchiveWriterEntry(
			byte[] externalData,
			string internalPath)
			: this(externalData, 0, externalData.Length, internalPath, string.Empty)
		{
		}

		internal void WriteLocalHeader(Stream stream)
		{
			this._localHeaderOffset = stream.Position;

			using (StreamHelper streamHelper = new StreamHelper(stream))
			{
				// +0x00 Signature
				streamHelper.WriteU4(ZipBlockSignature.LocalFileHeader);
				// +0x04 Decompressor version
				streamHelper.WriteU2(20);
				// +0x06 Flags
				streamHelper.WriteU2((ushort)((this._isUtf8 ? ZipFileFlags.FileIsUtf8 : ZipFileFlags.None)));
				// +0x08 Compression Method
				streamHelper.WriteU2((ushort)ZipCompressionMethod.Store);
				// +0x0A Modification Time
				streamHelper.WriteU2((ushort)
					(((this._internalModificationTime.Hour) << 11) |
					((this._internalModificationTime.Minute) << 5) |
					(this._internalModificationTime.Second / 2)));
				// +0x0C Modification Date
				streamHelper.WriteU2((ushort)
					(((this._internalModificationTime.Year - 1980) << 9) |
					((this._internalModificationTime.Month) << 5) |
					(this._internalModificationTime.Day)));
				// +0x0E CRC32
				streamHelper.WriteU4(0);
				// +0x12 Compressed Size
				streamHelper.WriteU4(this._isZip64 ? 0xFFFFFFFF : 0);
				// +0x16 Decompressed Size
				streamHelper.WriteU4(this._isZip64 ? 0xFFFFFFFF : (uint)this._externalSize);
				// +0x1A File Name Length
				streamHelper.WriteU2((ushort)this._internalPathData.Length);
				// +0x1C Extra Data Length
				streamHelper.WriteU2((ushort)((this._isZip64 ? 32 : 0) + 36));
				// +0x1E File Name
				streamHelper.WriteArray(this._internalPathData);

				// Zip64 Extra Data
				if (this._isZip64)
				{
					// +0x1E+FileName Extra Data Type
					streamHelper.WriteU2(1);
					// +0x20+FileName Extra Data Size
					streamHelper.WriteU2(28);
					// +0x22+FileName Uncompressed Size
					streamHelper.WriteI8(this._externalSize);
					// +0x2A+FileName Compressed Size
					streamHelper.WriteI8(0);
					// +0x32+FileName Local Header Offset
					streamHelper.WriteI8(this._localHeaderOffset);
					// +0x3A+FileName Disk Number
					streamHelper.WriteU4(0);
				}

				// NTFS Extra Data

				// Extra Data Type
				streamHelper.WriteU2(10);
				// Extra Data Size
				streamHelper.WriteU2(32);
				// Reserved
				streamHelper.WriteU4(0);
				// Tag
				streamHelper.WriteU2(1);
				// Size
				streamHelper.WriteU2(24);
				// Modification Time
				streamHelper.WriteI8(this._internalModificationTime.ToFileTime());
				// Access Time
				streamHelper.WriteI8(this._internalAccessTime.ToFileTime());
				// Creation Time
				streamHelper.WriteI8(this._internalCreationTime.ToFileTime());
			}
		}

		internal void WriteCentralHeader(Stream stream)
		{
			using (StreamHelper streamHelper = new StreamHelper(stream))
			{
				// +0x00 Signature
				streamHelper.WriteU4(ZipBlockSignature.CentralFileHeader);
				// +0x04 Compressor version
				streamHelper.WriteU2(20);
				// +0x06 Decompressor version
				streamHelper.WriteU2(20);
				// +0x08 Flags
				streamHelper.WriteU2((ushort)((this._isUtf8 ? ZipFileFlags.FileIsUtf8 : ZipFileFlags.None)));
				// +0x0A Compression Method
				streamHelper.WriteU2((ushort)ZipCompressionMethod.Store);
				// +0x0C Modification Time
				streamHelper.WriteU2((ushort)
					(((this._internalModificationTime.Hour) << 11) |
					((this._internalModificationTime.Minute) << 5) |
					(this._internalModificationTime.Second / 2)));
				// +0x0E Modification Date
				streamHelper.WriteU2((ushort)
					(((this._internalModificationTime.Year - 1980) << 9) |
					((this._internalModificationTime.Month) << 5) |
					(this._internalModificationTime.Day)));
				// +0x10 CRC32
				streamHelper.WriteU4(this._crc32);
				// +0x14 Compressed Size
				streamHelper.WriteU4(this._isZip64 ? 0xFFFFFFFF : (uint)this._internalSize);
				// +0x18 Decompressed Size
				streamHelper.WriteU4(this._isZip64 ? 0xFFFFFFFF : (uint)this._externalSize);
				// +0x1C File Name Length
				streamHelper.WriteU2((ushort)this._internalPathData.Length);
				// +0x1E Extra Data Length
				streamHelper.WriteU2((ushort)((this._isZip64 ? 32 : 0) + 36));
				// +0x20 File Comment Length
				streamHelper.WriteU2((ushort)this._internalCommentData.Length);
				// +0x22 Disk Number
				streamHelper.WriteU2(0);
				// +0x24 Internal Attributes
				streamHelper.WriteU2(0);
				// +0x26 External Attributes
				streamHelper.WriteU4(this._internalAttributes);
				// +0x2A Local Header Offset
				streamHelper.WriteU4(this._isZip64 ? 0xFFFFFFFF : (uint)this._localHeaderOffset);
				// +0x28 File Name
				streamHelper.WriteArray(this._internalPathData);

				// Zip64 Extra Data
				if (this._isZip64)
				{
					// +0x28+FileName Extra Data Type
					streamHelper.WriteU2(1);
					// +0x30+FileName Extra Data Size
					streamHelper.WriteU2(28);
					// +0x32+FileName Uncompressed Size
					streamHelper.WriteI8(this._externalSize);
					// +0x3A+FileName Compressed Size
					streamHelper.WriteI8(this._internalSize);
					// +0x42+FileName Local Header Offset
					streamHelper.WriteI8(this._localHeaderOffset);
					// +0x4A+FileName Disk Number
					streamHelper.WriteU4(0);
				}

				// NTFS Extra Data

				// Extra Data Type
				streamHelper.WriteU2(10);
				// Extra Data Size
				streamHelper.WriteU2(32);
				// Reserved
				streamHelper.WriteU4(0);
				// Tag
				streamHelper.WriteU2(1);
				// Size
				streamHelper.WriteU2(24);
				// Modification Time
				streamHelper.WriteI8(this._internalModificationTime.ToFileTime());
				// Access Time
				streamHelper.WriteI8(this._internalAccessTime.ToFileTime());
				// Creation Time
				streamHelper.WriteI8(this._internalCreationTime.ToFileTime());

				// File Comment
				streamHelper.WriteArray(this._internalCommentData);
			}
		}

		internal void WriteData(Stream stream)
		{
			long dataFrom = stream.Position;

			using (Crc32Stream crcStream = new Crc32Stream(this._externalStream))
			{
				byte[] buffer = new byte[Math.Min(this._externalSize, 16 * 1024 * 1024)];

				while (this._externalStream.Position < this._externalSize)
				{
					int count = (int)Math.Min(this._externalSize - this._externalStream.Position, buffer.Length);

					crcStream.Read(buffer, 0, count);
					stream.Write(buffer, 0, count);
				}

				this._crc32 = crcStream.Crc32Read;
			}

			long dataTo = stream.Position;

			this._internalSize = dataTo - dataFrom;

			using (StreamHelper streamHelper = new StreamHelper(stream))
			{
				stream.Seek(this._localHeaderOffset + 0x0E, SeekOrigin.Begin);
				streamHelper.WriteU4((uint)this._crc32);

				if (this._isZip64)
				{
					stream.Seek(this._localHeaderOffset + 0x2A + this._internalPathData.Length, SeekOrigin.Begin);
					streamHelper.WriteI8(this._internalSize);
				}
				else
				{
					stream.Seek(this._localHeaderOffset + 0x12, SeekOrigin.Begin);
					streamHelper.WriteU4((uint)this._internalSize);
				}
			}

			stream.Seek(dataTo, SeekOrigin.Begin);
		}

		#endregion

		#region Properties

		public long ExternalSize
		{
			[DebuggerStepThrough]
			get { return this._externalSize; }
			[DebuggerStepThrough]
			set { this._externalSize = value; }
		}

		public string InternalPath
		{
			[DebuggerStepThrough]
			get { return this._internalPath; }
		}

		public DateTime InternalCreationTime
		{
			[DebuggerStepThrough]
			get { return this._internalCreationTime; }
			[DebuggerStepThrough]
			set { this._internalCreationTime = value; }
		}

		public DateTime InternalModificationTime
		{
			[DebuggerStepThrough]
			get { return this._internalModificationTime; }
			[DebuggerStepThrough]
			set { this._internalModificationTime = value; }
		}

		public DateTime InternalAccessTime
		{
			[DebuggerStepThrough]
			get { return this._internalAccessTime; }
			[DebuggerStepThrough]
			set { this._internalAccessTime = value; }
		}

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				this._externalStream.Dispose();
			}

			base.Dispose(disposing);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
