﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Logging Library                          */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing.Design;

using Nabu.Data;

namespace Nabu.Logging.LogWriters
{
	public abstract class DatabaseLogWriterBase : ILogWriter
	{
		#region DatabaseLogWriterBase

		#region Constants

		protected const string TemplateLineDefault = "\r\n";
		protected const string TemplateIndentDefault = "\t";

		protected struct ObjectIndex
		{
			public const int TableEntry = 0;
			public const int TableData = 1;
			public const int TableStack = 2;

			public const int ColumnEntryEntryID = 3;
			public const int ColumnEntryLevel = 4;
			public const int ColumnEntrySource = 5;
			public const int ColumnEntryDateTime = 6;
			public const int ColumnEntryThreadID = 7;
			public const int ColumnDataDataID = 8;
			public const int ColumnDataEntryID = 9;
			public const int ColumnDataDatum = 10;
			public const int ColumnStackStackID = 11;
			public const int ColumnStackEntryID = 12;
			public const int ColumnStackAssemblyName = 13;
			public const int ColumnStackMethodName = 14;
			public const int ColumnStackFileName = 15;
			public const int ColumnStackLineNumber = 16;
			public const int ColumnStackColumnNumber = 17;

			public const int StoredProcedureEntry = 18;
			public const int StoredProcedureData = 19;
			public const int StoredProcedureStack = 20;

			public const int TotalCount = 21;
		}

		protected struct CommandIndex
		{
			public const int InsertIntoEntry = 0;
			public const int InsertIntoData = 1;
			public const int InsertIntoStack = 2;
			public const int CreateTableEntry = 3;
			public const int CreateTableData = 4;
			public const int CreateTableStack = 5;
			public const int CreateStoredProcInsertIntoEntry = 6;
			public const int CreateStoredProcInsertIntoData = 7;
			public const int CreateStoredProcInsertIntoStack = 8;

			public const int TotalCount = 9;
		}

		protected string[] TemplateObjectListDefault =
		{
			"nabu_log_entry",
			"nabu_log_data",
			"nabu_log_stack",
			"entry_entry_id",
			"entry_level",
			"entry_source",
			"entry_datetime",
			"entry_thread_id",
			"data_data_id",
			"data_entry_id",
			"data_datum",
			"stack_stack_id",
			"stack_entry_id",
			"stack_assembly_name",
			"stack_method_name",
			"stack_file_name",
			"stack_line_number",
			"stack_column_number",
			"sp_log_entry",
			"sp_log_data",
			"sp_log_stack",
		};

		#endregion

		#region Fields

		private string _lineTemplate;
		private string _indentTemplate;
		private string[] _objectTemplates;
		private string[] _commandTemplates;

		#endregion

		#region Methods

		protected void UpdateCommandTemplates()
		{
			for (int index = 0; index < this._commandTemplates.Length; index++)
			{
				this._commandTemplates[index] = this._commandTemplates[index]
					.Replace("%nabu_log_entry%", this._objectTemplates[ObjectIndex.TableEntry])
					.Replace("%nabu_log_data%", this._objectTemplates[ObjectIndex.TableData])
					.Replace("%nabu_log_stack%", this._objectTemplates[ObjectIndex.TableStack])
					.Replace("%entry_entry_id%", this._objectTemplates[ObjectIndex.ColumnEntryEntryID])
					.Replace("%entry_level%", this._objectTemplates[ObjectIndex.ColumnEntryLevel])
					.Replace("%entry_source%", this._objectTemplates[ObjectIndex.ColumnEntrySource])
					.Replace("%entry_datetime%", this._objectTemplates[ObjectIndex.ColumnEntryDateTime])
					.Replace("%entry_thread_id%", this._objectTemplates[ObjectIndex.ColumnEntryThreadID])
					.Replace("%data_data_id%", this._objectTemplates[ObjectIndex.ColumnDataDataID])
					.Replace("%data_entry_id%", this._objectTemplates[ObjectIndex.ColumnDataEntryID])
					.Replace("%data_datum%", this._objectTemplates[ObjectIndex.ColumnDataDatum])
					.Replace("%stack_stack_id%", this._objectTemplates[ObjectIndex.ColumnStackStackID])
					.Replace("%stack_entry_id%", this._objectTemplates[ObjectIndex.ColumnStackEntryID])
					.Replace("%stack_assembly_name%", this._objectTemplates[ObjectIndex.ColumnStackAssemblyName])
					.Replace("%stack_method_name%", this._objectTemplates[ObjectIndex.ColumnStackMethodName])
					.Replace("%stack_file_name%", this._objectTemplates[ObjectIndex.ColumnStackFileName])
					.Replace("%stack_line_number%", this._objectTemplates[ObjectIndex.ColumnStackLineNumber])
					.Replace("%stack_column_number%", this._objectTemplates[ObjectIndex.ColumnStackColumnNumber])
					.Replace("%sp_log_entry%", this._objectTemplates[ObjectIndex.StoredProcedureEntry])
					.Replace("%sp_log_data%", this._objectTemplates[ObjectIndex.StoredProcedureData])
					.Replace("%sp_log_stack%", this._objectTemplates[ObjectIndex.StoredProcedureStack]);
			}
		}

		protected DatabaseLogWriterBase(string[] templateCommandList)
		{
			if (templateCommandList == null)
			{
				throw new ArgumentNullException("templateCommandList");
			}

			this._lineTemplate = TemplateLineDefault;
			this._indentTemplate = TemplateIndentDefault;
			this._objectTemplates = TemplateObjectListDefault;
			this._commandTemplates = templateCommandList;
		}

		protected DatabaseLogWriterBase(
			string[] templateCommandList,
			string[] templateObjectList)
		{
			if (templateCommandList == null)
			{
				throw new ArgumentNullException("templateCommandList");
			}

			if (templateObjectList == null)
			{
				throw new ArgumentNullException("templateCommandList");
			}

			this._lineTemplate = TemplateLineDefault;
			this._indentTemplate = TemplateIndentDefault;
			this._objectTemplates = templateObjectList;
			this._commandTemplates = templateCommandList;
		}

		protected DatabaseLogWriterBase(
			string templateLine,
			string templateIndent,
			string[] templateCommandList,
			string[] templateObjectList)
		{
			if (templateLine == null)
			{
				throw new ArgumentNullException("templateLine");
			}

			if (templateIndent == null)
			{
				throw new ArgumentNullException("templateIndent");
			}

			if (templateCommandList == null)
			{
				throw new ArgumentNullException("templateCommandList");
			}

			if (templateObjectList == null)
			{
				throw new ArgumentNullException("templateObjectList");
			}

			this._lineTemplate = templateLine;
			this._indentTemplate = templateIndent;
			this._objectTemplates = templateObjectList;
			this._commandTemplates = templateCommandList;
		}

		protected string GetDatumString(object datum)
		{
			return Singleton<LogFormatterDefault>.Instance.GetLogString(datum, this._lineTemplate, this._indentTemplate, 0);
		}

		#endregion

		#region Properties

		[Editor("System.ComponentModel.Design.MultilineStringEditor, System.Design, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a", typeof(UITypeEditor))]
		public string LineTemplate
		{
			get { return this._lineTemplate; }
		}

		[Editor("System.ComponentModel.Design.MultilineStringEditor, System.Design, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a", typeof(UITypeEditor))]
		public string IndentTemplate
		{
			get { return this._indentTemplate; }
		}

		[Editor("System.Windows.Forms.Design.StringArrayEditor, System.Design, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a", typeof(UITypeEditor))]
		public string[] ObjectTemplates
		{
			get { return this._objectTemplates; }
		}

		[Editor("System.Windows.Forms.Design.StringArrayEditor, System.Design, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a", typeof(UITypeEditor))]
		public string[] CommandTemplates
		{
			get { return this._commandTemplates; }
		}

		public string TableData
		{
			get { return this._objectTemplates[ObjectIndex.TableData]; }
			set { this._objectTemplates[ObjectIndex.TableData] = value; }
		}

		public string TableEntry
		{
			get { return this._objectTemplates[ObjectIndex.TableEntry]; }
			set { this._objectTemplates[ObjectIndex.TableEntry] = value; }
		}

		public string TableStack
		{
			get { return this._objectTemplates[ObjectIndex.TableStack]; }
			set { this._objectTemplates[ObjectIndex.TableStack] = value; }
		}

		public string ColumnDataDataID
		{
			get { return this._objectTemplates[ObjectIndex.ColumnDataDataID]; }
			set { this._objectTemplates[ObjectIndex.ColumnDataDataID] = value; }
		}

		public string ColumnDataDatum
		{
			get { return this._objectTemplates[ObjectIndex.ColumnDataDatum]; }
			set { this._objectTemplates[ObjectIndex.ColumnDataDatum] = value; }
		}

		public string ColumnDataEntryID
		{
			get { return this._objectTemplates[ObjectIndex.ColumnDataEntryID]; }
			set { this._objectTemplates[ObjectIndex.ColumnDataEntryID] = value; }
		}

		public string ColumnEntryDateTime
		{
			get { return this._objectTemplates[ObjectIndex.ColumnEntryDateTime]; }
			set { this._objectTemplates[ObjectIndex.ColumnEntryDateTime] = value; }
		}

		public string ColumnEntryEntryID
		{
			get { return this._objectTemplates[ObjectIndex.ColumnEntryEntryID]; }
			set { this._objectTemplates[ObjectIndex.ColumnEntryEntryID] = value; }
		}

		public string ColumnEntryLevel
		{
			get { return this._objectTemplates[ObjectIndex.ColumnEntryLevel]; }
			set { this._objectTemplates[ObjectIndex.ColumnEntryLevel] = value; }
		}

		public string ColumnEntrySource
		{
			get { return this._objectTemplates[ObjectIndex.ColumnEntrySource]; }
			set { this._objectTemplates[ObjectIndex.ColumnEntrySource] = value; }
		}

		public string ColumnEntryThreadID
		{
			get { return this._objectTemplates[ObjectIndex.ColumnEntryThreadID]; }
			set { this._objectTemplates[ObjectIndex.ColumnEntryThreadID] = value; }
		}

		public string ColumnStackAssemblyName
		{
			get { return this._objectTemplates[ObjectIndex.ColumnStackAssemblyName]; }
			set { this._objectTemplates[ObjectIndex.ColumnStackAssemblyName] = value; }
		}

		public string ColumnStackColumnNumber
		{
			get { return this._objectTemplates[ObjectIndex.ColumnStackColumnNumber]; }
			set { this._objectTemplates[ObjectIndex.ColumnStackColumnNumber] = value; }
		}

		public string ColumnStackEntryID
		{
			get { return this._objectTemplates[ObjectIndex.ColumnStackEntryID]; }
			set { this._objectTemplates[ObjectIndex.ColumnStackEntryID] = value; }
		}

		public string ColumnStackFileName
		{
			get { return this._objectTemplates[ObjectIndex.ColumnStackFileName]; }
			set { this._objectTemplates[ObjectIndex.ColumnStackFileName] = value; }
		}

		public string ColumnStackLineNumber
		{
			get { return this._objectTemplates[ObjectIndex.ColumnStackLineNumber]; }
			set { this._objectTemplates[ObjectIndex.ColumnStackLineNumber] = value; }
		}

		public string ColumnStackMethodName
		{
			get { return this._objectTemplates[ObjectIndex.ColumnStackMethodName]; }
			set { this._objectTemplates[ObjectIndex.ColumnStackMethodName] = value; }
		}

		public string ColumnStackStackID
		{
			get { return this._objectTemplates[ObjectIndex.ColumnStackStackID]; }
			set { this._objectTemplates[ObjectIndex.ColumnStackStackID] = value; }
		}

		public string StoredProcedureData
		{
			get { return this._objectTemplates[ObjectIndex.StoredProcedureData]; }
			set { this._objectTemplates[ObjectIndex.StoredProcedureData] = value; }
		}

		public string StoredProcedureEntry
		{
			get { return this._objectTemplates[ObjectIndex.StoredProcedureEntry]; }
			set { this._objectTemplates[ObjectIndex.StoredProcedureEntry] = value; }
		}

		public string StoredProcedureStack
		{
			get { return this._objectTemplates[ObjectIndex.StoredProcedureStack]; }
			set { this._objectTemplates[ObjectIndex.StoredProcedureStack] = value; }
		}

		#endregion

		#endregion

		#region ILogWriter

		public abstract void Initialize();
		public abstract void Terminate();
		public abstract void Log(LogLevel level, LogSource source, DateTime dateTime, int threadID, object[] data, IList<string[]> stackFrameList);

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
