﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Logging Library                          */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Data;
using System.Data.OleDb;

namespace Nabu.Logging.LogWriters
{
	public sealed class MicrosoftAccessDatabaseLogWriter :
		DatabaseLogWriterBase, IDisposable
	{
		#region MicrosoftAccessDatabaseLogWriter

		#region Constants

		private const string TemplateConnectionString = "Provider=Microsoft.Jet.OLEDB.4.0;Data Source=%filename%;Jet OLEDB:Database Password=%password%;%additional%";

		private static string[] TemplateCommandListDefault =
		{
			"%sp_log_entry%",

			"%sp_log_data%",

			"%sp_log_stack%",

			"CREATE TABLE %nabu_log_entry% ( " +
			"%entry_entry_id% GUID NOT NULL PRIMARY KEY, " +
			"%entry_level% TEXT NOT NULL, " +
			"%entry_source% TEXT NOT NULL, " +
			"%entry_datetime% DATETIME NOT NULL, " +
			"%entry_thread_id% INTEGER NOT NULL)",

			"CREATE TABLE %nabu_log_data% ( " +
			"%data_data_id% GUID NOT NULL PRIMARY KEY, " +
			"%data_entry_id% GUID NOT NULL REFERENCES %nabu_log_entry%(%entry_entry_id%), " +
			"%data_datum% TEXT NOT NULL)",

			"CREATE TABLE %nabu_log_stack% ( " +
			"%stack_stack_id% GUID NOT NULL PRIMARY KEY, " +
			"%stack_entry_id% GUID NOT NULL REFERENCES %nabu_log_entry%(%entry_entry_id%), " +
			"%stack_assembly_name% TEXT NOT NULL, " +
			"%stack_method_name% TEXT NOT NULL, " +
			"%stack_file_name% TEXT NOT NULL, " +
			"%stack_line_number% INTEGER NOT NULL, " +
			"%stack_column_number% INTEGER NOT NULL)",

			"CREATE PROCEDURE %sp_log_entry%" +
			"( " +
			" ParamEntryEntryID GUID, " +
			" ParamEntryLevel TEXT, " +
			" ParamEntrySource TEXT, " +
			" ParamEntryDateTime DATETIME, " +
			" ParamEntryThreadID INTEGER " +
			") " +
			"AS " +
			"INSERT INTO %nabu_log_entry% " +
			"(%entry_entry_id%, %entry_level%, %entry_source%, %entry_datetime%, %entry_thread_id%) " +
			"VALUES (ParamEntryEntryID, ParamEntryLevel, ParamEntrySource, ParamEntryDateTime, ParamEntryThreadID)",

			"CREATE PROCEDURE %sp_log_data% " +
			"( " +
			" ParamDataDataID GUID, " +
			" ParamDataEntryID GUID, " +
			" ParamDataDatum TEXT " +
			") " +
			"AS " +
			"INSERT INTO %nabu_log_data% " +
			"(%data_data_id%, %data_entry_id%, %data_datum%) " +
			"VALUES (ParamDataDataID, ParamDataEntryID, ParamDataDatum)",

			"CREATE PROCEDURE %sp_log_stack% " +
			"( " +
			" ParamStackStackID GUID, " +
			" ParamStackEntryID GUID, " +
			" ParamStackAssemblyName TEXT, " +
			" ParamStackMethodName TEXT, " +
			" ParamStackFileName TEXT, " +
			" ParamStackLineNumber INTEGER, " +
			" ParamStackColumnNumber INTEGER " +
			") " +
			"AS " +
			"INSERT INTO %nabu_log_stack% " +
			"(%stack_stack_id%, %stack_entry_id%, %stack_assembly_name%, %stack_method_name%, %stack_file_name%, %stack_line_number%, %stack_column_number%) " +
			"VALUES (ParamStackStackID, ParamStackEntryID, ParamStackAssemblyName, ParamStackMethodName, ParamStackFileName, ParamStackLineNumber, ParamStackColumnNumber)"
		};

		#endregion

		#region Fields

		private OleDbConnection _connection;
		private OleDbCommand _commandEntry;
		private OleDbCommand _commandData;
		private OleDbCommand _commandStack;
		private string _connectionString;

		#endregion

		#region Methods

		public MicrosoftAccessDatabaseLogWriter()
			: base(TemplateCommandListDefault)
		{
			this._connectionString = string.Empty;
		}

		public MicrosoftAccessDatabaseLogWriter(
			string connectionString)
			: base(TemplateCommandListDefault)
		{
			if (connectionString == null)
			{
				throw new ArgumentNullException("connectionString");
			}

			this._connectionString = connectionString;
		}

		public MicrosoftAccessDatabaseLogWriter(
			string filename,
			string password)
			: base(TemplateCommandListDefault)
		{
			if (filename == null)
			{
				throw new ArgumentNullException("filename");
			}

			if (password == null)
			{
				throw new ArgumentNullException("password");
			}

			this._connectionString = TemplateConnectionString
				.Replace("%filename%", filename)
				.Replace("%password%", password);
		}

		public MicrosoftAccessDatabaseLogWriter(
			string connectionString,
			string[] templateObjectList)
			: base(TemplateCommandListDefault, templateObjectList)
		{
			if (connectionString == null)
			{
				throw new ArgumentNullException("connectionString");
			}

			this._connectionString = connectionString;
		}

		public MicrosoftAccessDatabaseLogWriter(
			string filename,
			string password,
			string[] templateObjectList)
			: base(TemplateCommandListDefault, templateObjectList)
		{
			if (filename == null)
			{
				throw new ArgumentNullException("filename");
			}

			if (password == null)
			{
				throw new ArgumentNullException("password");
			}

			this._connectionString = TemplateConnectionString
				.Replace("%filename%", filename)
				.Replace("%password%", password);
		}

		public MicrosoftAccessDatabaseLogWriter(
			string connectionString,
			string[] templateObjectList,
			string templateLine,
			string templateIndent)
			: base(templateLine, templateIndent, TemplateCommandListDefault, templateObjectList)
		{
			if (connectionString == null)
			{
				throw new ArgumentNullException("connectionString");
			}

			this._connectionString = connectionString;
		}

		public MicrosoftAccessDatabaseLogWriter(
			string filename,
			string password,
			string[] templateObjectList,
			string templateLine,
			string templateIndent)
			: base(templateLine, templateIndent, TemplateCommandListDefault, templateObjectList)
		{
			if (filename == null)
			{
				throw new ArgumentNullException("filename");
			}

			if (password == null)
			{
				throw new ArgumentNullException("password");
			}

			this._connectionString = TemplateConnectionString
				.Replace("%filename%", filename)
				.Replace("%password%", password);
		}

		~MicrosoftAccessDatabaseLogWriter()
		{
		}

		private IList<string> GetObjectList(string collectionName, string propertyName)
		{
			DataTable schema = this._connection.GetSchema(collectionName);

			int indexColumnProperty = 0;

			foreach (DataColumn column in schema.Columns)
			{
				if (column.ColumnName == propertyName)
				{
					indexColumnProperty = column.Ordinal;
				}
			}

			IList<string> objects = new List<string>(schema.Rows.Count);

			foreach (DataRow row in schema.Rows)
			{
				objects.Add(row[indexColumnProperty] as string);
			}

			return objects;
		}

		private void CreateObject(IList<string> objectList, string objectName, string objectCreateCommand)
		{
			if (!objectList.Contains(objectName))
			{
				using (OleDbCommand command = new OleDbCommand(objectCreateCommand, this._connection))
				{
					command.ExecuteNonQuery();
				}
			}
		}

		#endregion

		#region Properties

		public string ConnectionString
		{
			get { return this._connectionString; }
			set { this._connectionString = value; }
		}

		#endregion

		#endregion

		#region ILogWriter

		public override void Initialize()
		{
			try
			{
				base.UpdateCommandTemplates();

				this._connection = new OleDbConnection(this._connectionString);
				this._connection.Open();

				IList<string> tableList = GetObjectList("Tables", "TABLE_NAME");

				CreateObject(tableList, this.ObjectTemplates[ObjectIndex.TableEntry], this.CommandTemplates[CommandIndex.CreateTableEntry]);
				CreateObject(tableList, this.ObjectTemplates[ObjectIndex.TableData], this.CommandTemplates[CommandIndex.CreateTableData]);
				CreateObject(tableList, this.ObjectTemplates[ObjectIndex.TableStack], this.CommandTemplates[CommandIndex.CreateTableStack]);

				IList<string> storedProcedureList = GetObjectList("Procedures", "PROCEDURE_NAME");

				CreateObject(storedProcedureList, this.ObjectTemplates[ObjectIndex.StoredProcedureEntry], this.CommandTemplates[CommandIndex.CreateStoredProcInsertIntoEntry]);
				CreateObject(storedProcedureList, this.ObjectTemplates[ObjectIndex.StoredProcedureData], this.CommandTemplates[CommandIndex.CreateStoredProcInsertIntoData]);
				CreateObject(storedProcedureList, this.ObjectTemplates[ObjectIndex.StoredProcedureStack], this.CommandTemplates[CommandIndex.CreateStoredProcInsertIntoStack]);

				this._commandEntry = this._connection.CreateCommand();
				this._commandEntry.CommandText = this.CommandTemplates[CommandIndex.InsertIntoEntry];
				this._commandEntry.CommandType = CommandType.StoredProcedure;
				this._commandEntry.Parameters.Add("ParamEntryEntryID", OleDbType.Guid);
				this._commandEntry.Parameters.Add("ParamEntryLevel", OleDbType.BSTR);
				this._commandEntry.Parameters.Add("ParamEntrySource", OleDbType.BSTR);
				this._commandEntry.Parameters.Add("ParamEntryDateTime", OleDbType.Date);
				this._commandEntry.Parameters.Add("ParamEntryThreadID", OleDbType.Integer);
				this._commandEntry.Prepare();

				this._commandData = this._connection.CreateCommand();
				this._commandData.CommandText = this.CommandTemplates[CommandIndex.InsertIntoData];
				this._commandData.CommandType = CommandType.StoredProcedure;
				this._commandData.Parameters.Add("ParamDataDataID", OleDbType.Guid);
				this._commandData.Parameters.Add("ParamDataEntryID", OleDbType.Guid);
				this._commandData.Parameters.Add("ParamDataDatum", OleDbType.BSTR);
				this._commandData.Prepare();

				this._commandStack = this._connection.CreateCommand();
				this._commandStack.CommandText = this.CommandTemplates[CommandIndex.InsertIntoStack];
				this._commandStack.CommandType = CommandType.StoredProcedure;
				this._commandStack.Parameters.Add("ParamStackStackID", OleDbType.Guid);
				this._commandStack.Parameters.Add("ParamStackEntryID", OleDbType.Guid);
				this._commandStack.Parameters.Add("ParamStackAssemblyName", OleDbType.BSTR);
				this._commandStack.Parameters.Add("ParamStackMethodName", OleDbType.BSTR);
				this._commandStack.Parameters.Add("ParamStackFileName", OleDbType.BSTR);
				this._commandStack.Parameters.Add("ParamStackLineNumber", OleDbType.Integer);
				this._commandStack.Parameters.Add("ParamStackColumnNumber", OleDbType.Integer);
				this._commandStack.Prepare();
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		public override void Terminate()
		{
			try
			{
				if (this._commandEntry != null)
				{
					this._commandEntry.Dispose();
					this._commandEntry = null;
				}

				if (this._commandData != null)
				{
					this._commandData.Dispose();
					this._commandData = null;
				}

				if (this._commandStack != null)
				{
					this._commandStack.Dispose();
					this._commandStack = null;
				}

				if (this._connection != null)
				{
					this._connection.Dispose();
					this._connection = null;
				}
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		public override void Log(LogLevel level, LogSource source, DateTime dateTime, int threadID, object[] data, IList<string[]> stackFrameList)
		{
			if (level == null)
			{
				throw new ArgumentNullException("level");
			}

			if (source == null)
			{
				throw new ArgumentNullException("source");
			}

			if (data == null)
			{
				throw new ArgumentNullException("data");
			}

			if (stackFrameList == null)
			{
				throw new ArgumentNullException("stackFrameList");
			}

			try
			{
				Guid entryID = Guid.NewGuid();

				using (OleDbTransaction transaction = this._connection.BeginTransaction(IsolationLevel.ReadUncommitted))
				{
					this._commandEntry.Parameters["ParamEntryEntryID"].Value = entryID;
					this._commandEntry.Parameters["ParamEntryLevel"].Value = level.ToString();
					this._commandEntry.Parameters["ParamEntrySource"].Value = source.ToString();
					this._commandEntry.Parameters["ParamEntryDateTime"].Value = dateTime;
					this._commandEntry.Parameters["ParamEntryThreadID"].Value = threadID;
					this._commandEntry.Transaction = transaction;
					this._commandEntry.ExecuteNonQuery();

					foreach (object datum in data)
					{
						this._commandData.Parameters["ParamDataDataID"].Value = Guid.NewGuid();
						this._commandData.Parameters["ParamDataEntryID"].Value = entryID;
						this._commandData.Parameters["ParamDataDatum"].Value = GetDatumString(datum);
						this._commandData.Transaction = transaction;
						this._commandData.ExecuteNonQuery();
					}

					foreach (string[] stackFrame in stackFrameList)
					{
						this._commandStack.Parameters["ParamStackStackID"].Value = Guid.NewGuid();
						this._commandStack.Parameters["ParamStackEntryID"].Value = entryID;
						this._commandStack.Parameters["ParamStackAssemblyName"].Value = stackFrame[LogStackFrameElement.AssemblyName];
						this._commandStack.Parameters["ParamStackMethodName"].Value = stackFrame[LogStackFrameElement.MethodName];
						this._commandStack.Parameters["ParamStackFileName"].Value = stackFrame[LogStackFrameElement.FileName];
						this._commandStack.Parameters["ParamStackLineNumber"].Value = stackFrame[LogStackFrameElement.LineNumber];
						this._commandStack.Parameters["ParamStackColumnNumber"].Value = stackFrame[LogStackFrameElement.ColumnNumber];
						this._commandStack.Transaction = transaction;
						this._commandStack.ExecuteNonQuery();
					}

					transaction.Commit();
				}
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		#endregion

		#region IDisposable

		public void Dispose()
		{
			Terminate();
			GC.SuppressFinalize(this);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
