﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Logging Library                          */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace Nabu.Logging.LogWriters
{
	public sealed class TextFileLogWriter :
		TextLogWriterBase, IDisposable
	{
		#region TextFileLogWriter

		#region Fields

		private string _pathTemplate = "\\Log.txt";
		private TextWriter _textWriter;
		private Encoding _encoding = Encoding.UTF8;

		#endregion

		#region Methods

		public TextFileLogWriter()
		{
		}

		public TextFileLogWriter(string templatePath)
		{
			if (templatePath == null)
			{
				throw new ArgumentNullException("templatePath");
			}

			this._pathTemplate = templatePath;
		}

		public TextFileLogWriter(string templatePath, Encoding encoding)
		{
			if (templatePath == null)
			{
				throw new ArgumentNullException("templatePath");
			}

			if (encoding == null)
			{
				throw new ArgumentNullException("encoding");
			}

			this._pathTemplate = templatePath;
			this._encoding = encoding;
		}

		public TextFileLogWriter(string templatePath, string templateLine, string templateIndent)
			: base(templateLine, templateIndent)
		{
			if (templatePath == null)
			{
				throw new ArgumentNullException("templatePath");
			}

			this._pathTemplate = templatePath;
		}

		public TextFileLogWriter(string templatePath, Encoding encoding, string templateLine, string templateIndent)
			: base(templateLine, templateIndent)
		{
			if (templatePath == null)
			{
				throw new ArgumentNullException("templatePath");
			}

			if (encoding == null)
			{
				throw new ArgumentNullException("encoding");
			}

			this._pathTemplate = templatePath;
			this._encoding = encoding;
		}

		~TextFileLogWriter()
		{
		}

		#endregion

		#region Properties

		public string PathTemplate
		{
			get { return this._pathTemplate; }
			set { this._pathTemplate = value; }
		}

		public Encoding Encoding
		{
			get { return this._encoding; }
		}

		#endregion

		#endregion

		#region ILogWriter

		public override void Initialize()
		{
			try
			{
				string path = LogFilePathExpander.Expand(this._pathTemplate);

				if (this._pathTemplate.StartsWith(new string(Path.DirectorySeparatorChar, 1)) || this._pathTemplate.StartsWith(new string(Path.AltDirectorySeparatorChar, 1)))
				{
					path = SpecialPath.CurrentUserCurrentMachineSettingsPath + path;
				}

				Directory.CreateDirectory(Path.GetDirectoryName(path));
				this._textWriter = new StreamWriter(new FileStream(path, FileMode.Append, FileAccess.Write, FileShare.Read), this._encoding);
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		public override void Terminate()
		{
			try
			{
				if (this._textWriter != null)
				{
					this._textWriter.Close();
					this._textWriter = null;
				}
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		public override void Log(LogLevel level, LogSource source, DateTime dateTime, int threadID, object[] data, IList<string[]> stackFrameList)
		{
			if (level == null)
			{
				throw new ArgumentNullException("level");
			}

			if (source == null)
			{
				throw new ArgumentNullException("source");
			}

			if (data == null)
			{
				throw new ArgumentNullException("data");
			}

			if (stackFrameList == null)
			{
				throw new ArgumentNullException("stackFrameList");
			}

			try
			{
				this._textWriter.Write(base.GetLogString(level, source, dateTime, threadID, data, stackFrameList));
				this._textWriter.Flush();
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		#endregion

		#region IDisposable

		public void Dispose()
		{
			Terminate();
			GC.SuppressFinalize(this);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
