﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Logging Library                          */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing.Design;
using System.Text;

using Nabu.Data;

namespace Nabu.Logging.LogWriters
{
	public abstract class TextLogWriterBase : ILogWriter
	{
		#region TextLogWriterBase

		#region Constants

		protected const string TemplateLineDefault = "\r\n";
		protected const string TemplateIndentDefault = "\t";
		protected const string TemplateEntryDefault =
			"Level: %level%; Source: %source%; Date/Time: %date_time%; ThreadID: %thread_id%;\n" +
			"Data:\n" +
			"%data%" +
			"Stack Trace:\n" +
			"%stack%\n";
		protected const string TemplateDatumDefault =
			"%datum%\n";
		protected const string TemplateStackDefault =
			"\t%method_name% in \"%file_name%\":%line_number%:%column_number% (%assembly_name%)\n";
		#endregion

		#region Fields

		private string _lineTemplate;
		private string _indentTemplate;
		private string _entryTemplate;
		private string _dataTemplate;
		private string _stackTemplate;

		#endregion

		#region Methods

		protected TextLogWriterBase()
		{
			this._lineTemplate = TemplateLineDefault;
			this._indentTemplate = TemplateIndentDefault;
			this._entryTemplate = TemplateEntryDefault
				.Replace("\n", this._lineTemplate)
				.Replace("\t", this._indentTemplate);
			this._dataTemplate = TemplateDatumDefault
				.Replace("\n", this._lineTemplate)
				.Replace("\t", this._indentTemplate);
			this._stackTemplate = TemplateStackDefault
				.Replace("\n", this._lineTemplate)
				.Replace("\t", this._indentTemplate);
		}

		protected TextLogWriterBase(
			string templateLine,
			string templateIndent)
		{
			if (templateLine == null)
			{
				throw new ArgumentNullException("templateLine");
			}

			if (templateIndent == null)
			{
				throw new ArgumentNullException("templateIndent");
			}

			this._lineTemplate = templateLine;
			this._indentTemplate = templateIndent;
			this._entryTemplate = TemplateEntryDefault
				.Replace("\n", this._lineTemplate)
				.Replace("\t", this._indentTemplate);
			this._dataTemplate = TemplateDatumDefault
				.Replace("\n", this._lineTemplate)
				.Replace("\t", this._indentTemplate);
			this._stackTemplate = TemplateStackDefault
				.Replace("\n", this._lineTemplate)
				.Replace("\t", this._indentTemplate);
		}

		protected TextLogWriterBase(
			string templateLine,
			string templateIndent,
			string templateEntry,
			string templateData,
			string templateStack)
		{
			if (templateLine == null)
			{
				throw new ArgumentNullException("templateLine");
			}

			if (templateIndent == null)
			{
				throw new ArgumentNullException("templateIndent");
			}

			if (templateEntry == null)
			{
				throw new ArgumentNullException("templateEntry");
			}

			if (templateData == null)
			{
				throw new ArgumentNullException("templateData");
			}

			if (templateStack == null)
			{
				throw new ArgumentNullException("templateStack");
			}

			this._lineTemplate = templateLine;
			this._indentTemplate = templateIndent;
			this._entryTemplate = templateEntry
				.Replace("\n", this._lineTemplate)
				.Replace("\t", this._indentTemplate);
			this._dataTemplate = templateData
				.Replace("\n", this._lineTemplate)
				.Replace("\t", this._indentTemplate);
			this._stackTemplate = templateStack
				.Replace("\n", this._lineTemplate)
				.Replace("\t", this._indentTemplate);
		}

		protected string GetLogString(LogLevel level, LogSource source, DateTime dateTime, int threadID, object[] data, IList<string[]> stackFrameList)
		{
			StringBuilder dataBuilder = new StringBuilder();

			foreach (object datum in data)
			{
				dataBuilder.Append(
					this._dataTemplate
						.Replace("%datum%", Singleton<LogFormatterDefault>.Instance.GetLogString(datum, this._lineTemplate, this._indentTemplate, 1)));
			}

			StringBuilder stackFrameBuilder = new StringBuilder();

			foreach (string[] stackFrame in stackFrameList)
			{
				stackFrameBuilder.Append(
					this._stackTemplate
						.Replace("%assembly_name%", stackFrame[LogStackFrameElement.AssemblyName])
						.Replace("%method_name%", stackFrame[LogStackFrameElement.MethodName])
						.Replace("%file_name%", stackFrame[LogStackFrameElement.FileName])
						.Replace("%line_number%", stackFrame[LogStackFrameElement.LineNumber])
						.Replace("%column_number%", stackFrame[LogStackFrameElement.ColumnNumber]));
			}

			return this._entryTemplate
				.Replace("%level%", level.ToString())
				.Replace("%source%", source.ToString())
				.Replace("%date_time%", Singleton<LogFormatterDefault>.Instance.GetLogString(dateTime, this._lineTemplate, this._indentTemplate, 0))
				.Replace("%thread_id%", threadID.ToString())
				.Replace("%stack%", stackFrameBuilder.ToString())
				.Replace("%data%", dataBuilder.ToString());
		}

		#endregion

		#region Properties

		[Editor("System.ComponentModel.Design.MultilineStringEditor, System.Design, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a", typeof(UITypeEditor))]
		public string LineTemplate
		{
			get { return this._lineTemplate; }
			set { this._lineTemplate = value; }
		}

		[Editor("System.ComponentModel.Design.MultilineStringEditor, System.Design, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a", typeof(UITypeEditor))]
		public string IndentTemplate
		{
			get { return this._indentTemplate; }
			set { this._indentTemplate = value; }
		}

		[Editor("System.ComponentModel.Design.MultilineStringEditor, System.Design, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a", typeof(UITypeEditor))]
		public string EntryTemplate
		{
			get { return this._entryTemplate; }
			set { this._entryTemplate = value; }
		}

		[Editor("System.ComponentModel.Design.MultilineStringEditor, System.Design, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a", typeof(UITypeEditor))]
		public string DataTemplate
		{
			get { return this._dataTemplate; }
			set { this._dataTemplate = value; }
		}

		[Editor("System.ComponentModel.Design.MultilineStringEditor, System.Design, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a", typeof(UITypeEditor))]
		public string StackTemplate
		{
			get { return this._stackTemplate; }
			set { this._stackTemplate = value; }
		}

		#endregion

		#endregion

		#region ILogWriter

		public abstract void Initialize();
		public abstract void Terminate();
		public abstract void Log(LogLevel level, LogSource source, DateTime dateTime, int threadID, object[] data, IList<string[]> stackFrameList);

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
