﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Logging Library                          */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace Nabu.Logging.LogWriters
{
	public sealed class TextMultiFileLogWriter : TextLogWriterBase
	{
		#region TextMultiFileLogWriter

		#region Fields

		private string _pathTemplate = "\\Log-%source%-%level%.txt";
		private Dictionary<string, TextWriter> _textWriterMap = new Dictionary<string, TextWriter>();
		private Encoding _encoding = Encoding.UTF8;

		#endregion

		#region Methods

		public TextMultiFileLogWriter()
		{
		}

		public TextMultiFileLogWriter(string templatePath)
		{
			if (templatePath == null)
			{
				throw new ArgumentNullException("templatePath");
			}

			this._pathTemplate = templatePath;
		}

		public TextMultiFileLogWriter(string templatePath, Encoding encoding)
		{
			if (templatePath == null)
			{
				throw new ArgumentNullException("templatePath");
			}

			if (encoding == null)
			{
				throw new ArgumentNullException("encoding");
			}

			this._pathTemplate = templatePath;
			this._encoding = encoding;
		}

		public TextMultiFileLogWriter(string templatePath, string templateLine, string templateIndent)
			: base(templateLine, templateIndent)
		{
			if (templatePath == null)
			{
				throw new ArgumentNullException("templatePath");
			}

			this._pathTemplate = templatePath;
		}

		public TextMultiFileLogWriter(string templatePath, Encoding encoding, string templateLine, string templateIndent)
			: base(templateLine, templateIndent)
		{
			if (templatePath == null)
			{
				throw new ArgumentNullException("templatePath");
			}

			if (encoding == null)
			{
				throw new ArgumentNullException("encoding");
			}

			this._pathTemplate = templatePath;
			this._encoding = encoding;
		}

		#endregion

		#region Properties

		public string PathTemplate
		{
			get { return this._pathTemplate; }
			set { this._pathTemplate = value; }
		}

		public Encoding Encoding
		{
			get { return this._encoding; }
		}

		#endregion

		#endregion

		#region ILogWriter

		public override void Initialize()
		{
			try
			{
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		public override void Terminate()
		{
			try
			{
				foreach (TextWriter textWriter in this._textWriterMap.Values)
				{
					textWriter.Close();
				}
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		public override void Log(LogLevel level, LogSource source, DateTime dateTime, int threadID, object[] data, IList<string[]> stackFrameList)
		{
			if (level == null)
			{
				throw new ArgumentNullException("level");
			}

			if (source == null)
			{
				throw new ArgumentNullException("source");
			}

			if (data == null)
			{
				throw new ArgumentNullException("data");
			}

			if (stackFrameList == null)
			{
				throw new ArgumentNullException("stackFrameList");
			}

			try
			{
				string path = LogFilePathExpander.Expand(
					this._pathTemplate
						.Replace("%level%", level.ToString())
						.Replace("%source%", source.ToString()));

				if (path.StartsWith(new string(Path.DirectorySeparatorChar, 1)) || path.StartsWith(new string(Path.AltDirectorySeparatorChar, 1)))
				{
					path = SpecialPath.CurrentUserCurrentMachineSettingsPath + path;
				}

				Directory.CreateDirectory(Path.GetDirectoryName(path));

				TextWriter textWriter;

				if (!this._textWriterMap.TryGetValue(path, out textWriter))
				{
					Directory.CreateDirectory(Path.GetDirectoryName(path));

					textWriter = new StreamWriter(new FileStream(path, FileMode.Append, FileAccess.Write, FileShare.Read), this._encoding);

					this._textWriterMap.Add(path, textWriter);
				}

				textWriter.Write(base.GetLogString(level, source, dateTime, threadID, data, stackFrameList));
				textWriter.Flush();
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
