﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Logging Library                          */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace Nabu.Logging.LogWriters
{
	public sealed class XmlFileLogWriter :
		XmlLogWriterBase, IDisposable
	{
		#region XmlFileLogWriter

		#region Fields

		private string _wrapperPathTemplate = "\\Log.xml";
		private string _dataPathTemplate = "\\Log-Data.txt";
		private TextWriter _textWriterData;

		#endregion

		#region Methods

		public XmlFileLogWriter()
		{
		}

		public XmlFileLogWriter(
			string templatePathData,
			string templatePathWrapper)
		{
			this._dataPathTemplate = templatePathData;
			this._wrapperPathTemplate = templatePathWrapper;
		}

		public XmlFileLogWriter(
			string templatePathData,
			string templatePathWrapper,
			string templateLine,
			string templateIndent)
			: base(templateLine, templateIndent)
		{
			this._dataPathTemplate = templatePathData;
			this._wrapperPathTemplate = templatePathWrapper;
		}

		public XmlFileLogWriter(
			string templatePathData,
			string templatePathWrapper,
			string templateLine,
			string templateIndent,
			string[] templateTagList)
			: base(templateLine, templateIndent, templateTagList)
		{
			this._dataPathTemplate = templatePathData;
			this._wrapperPathTemplate = templatePathWrapper;
		}

		~XmlFileLogWriter()
		{
		}

		#endregion

		#region Properties

		public string WrapperPathTemplate
		{
			get { return this._wrapperPathTemplate; }
			set { this._wrapperPathTemplate = value; }
		}

		public string DataPathTemplate
		{
			get { return this._dataPathTemplate; }
			set { this._dataPathTemplate = value; }
		}

		#endregion

		#endregion

		#region ILogWriter

		public override void Initialize()
		{
			try
			{
				this._dataPathTemplate = LogFilePathExpander.Expand(this._dataPathTemplate);

				if (this._dataPathTemplate.StartsWith(new string(Path.DirectorySeparatorChar, 1)) || this._dataPathTemplate.StartsWith(new string(Path.AltDirectorySeparatorChar, 1)))
				{
					this._dataPathTemplate = SpecialPath.CurrentUserCurrentMachineSettingsPath + this._dataPathTemplate;
				}

				Directory.CreateDirectory(Path.GetDirectoryName(this._dataPathTemplate));

				this._wrapperPathTemplate = LogFilePathExpander.Expand(this._wrapperPathTemplate);

				if (this._wrapperPathTemplate.StartsWith(new string(Path.DirectorySeparatorChar, 1)) || this._wrapperPathTemplate.StartsWith(new string(Path.AltDirectorySeparatorChar, 1)))
				{
					this._wrapperPathTemplate = SpecialPath.CurrentUserCurrentMachineSettingsPath + this._wrapperPathTemplate;
				}

				Directory.CreateDirectory(Path.GetDirectoryName(this._wrapperPathTemplate));

				using (TextWriter textWriterWrapper = new StreamWriter(
					new FileStream(
						this.WrapperPathTemplate,
						FileMode.Create,
						FileAccess.Write,
						FileShare.Read),
					Encoding.UTF8))
				{
					textWriterWrapper.WriteLine(
							"<?xml version=\"1.0\" encoding=\"utf-8\"?>{0}" +
							"<!DOCTYPE nabu_log [<!ENTITY {2} SYSTEM \"{3}\">]>{0}" +
							"<{4}>{0}" +
							"{1}&{2};{0}" +
							"</{4}>{0}",
							this.LineTemplate,
							this.IndentTemplate,
							this.TagTemplates[TagIndex.Entry],
							(new Uri(this._wrapperPathTemplate))
								.MakeRelativeUri(new Uri(_dataPathTemplate))
								.OriginalString
								.Replace(Path.AltDirectorySeparatorChar, Path.DirectorySeparatorChar)
								.Replace("&", "&amp;"),
							this.TagTemplates[TagIndex.Root]);
					textWriterWrapper.Close();
				}

				this._textWriterData = new StreamWriter(new FileStream(this._dataPathTemplate, FileMode.Append, FileAccess.Write, FileShare.Read), Encoding.UTF8);
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		public override void Terminate()
		{
			try
			{
				if (this._textWriterData != null)
				{
					this._textWriterData.Close();
				}
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		public override void Log(LogLevel level, LogSource source, DateTime dateTime, int threadID, object[] data, IList<string[]> stackFrameList)
		{
			if (level == null)
			{
				throw new ArgumentNullException("level");
			}

			if (source == null)
			{
				throw new ArgumentNullException("source");
			}

			if (data == null)
			{
				throw new ArgumentNullException("data");
			}

			if (stackFrameList == null)
			{
				throw new ArgumentNullException("stackFrameList");
			}

			try
			{
				this._textWriterData.Write(base.GetLogString(level, source, dateTime, threadID, data, stackFrameList));
				this._textWriterData.Flush();
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		#endregion

		#region IDisposable

		public void Dispose()
		{
			Terminate();
			GC.SuppressFinalize(this);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
