﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Logging Library                          */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing.Design;

namespace Nabu.Logging.LogWriters
{
	public abstract class XmlLogWriterBase : TextLogWriterBase
	{
		#region XmlLogWriterBase

		#region Constants

		public struct TagIndex
		{
			public const int Root = 0;
			public const int Entry = 1;
			public const int Data = 2;
			public const int Stack = 3;
			public const int Level = 4;
			public const int Source = 5;
			public const int DateTime = 6;
			public const int ThreadID = 7;
			public const int Datum = 8;
			public const int StackFrame = 9;
			public const int AssemblyName = 10;
			public const int MethodName = 11;
			public const int FileName = 12;
			public const int LineNumber = 13;
			public const int ColumnNumber = 14;

			public const int TotalCount = 15;
		}

		private static string[] TemplateTagListDefault =
		{
			/*  0 */ "log",
			/*  1 */ "entry",
			/*  2 */ "data",
			/*  3 */ "stack",
			/*  4 */ "level",
			/*  5 */ "source",
			/*  6 */ "date_time",
			/*  7 */ "thread_id",
			/*  8 */ "datum",
			/*  9 */ "stack_frame",
			/* 10 */ "assembly_name",
			/* 11 */ "method_name",
			/* 12 */ "file_name",
			/* 13 */ "line_number",
			/* 14 */ "column_number"
		};

		#endregion

		#region Fields

		private string[] _tagTemplates;

		#endregion

		#region Methods

		private static string GetEntryTemplate(string[] templateTagList)
		{
			return string.Format(
				"<{1} {4}=\"%level%\" {5}=\"%source%\" {6}=\"%date_time%\" {7}=\"%thread_id%\">\n" +
				"\t<{2}>\n" +
				"%data%" +
				"\t</{2}>\n" +
				"\t<{3}>\n" +
				"%stack%" +
				"\t</{3}>\n" +
				"</{1}>\n",
				templateTagList);
		}

		private static string GetDataTemplate(string[] templateTagList)
		{
			return string.Format(
				"\t\t<{8}>%datum%</{8}>\n",
				templateTagList);
		}

		private static string GetStackTemplate(string[] templateTagList)
		{
			return string.Format(
				"\t\t<{9}>\n" +
				"\t\t\t<{10}>%assembly_name%</{10}>\n" +
				"\t\t\t<{11}>%method_name%</{11}>\n" +
				"\t\t\t<{12}>%file_name%</{12}>\n" +
				"\t\t\t<{13}>%line_number%</{13}>\n" +
				"\t\t\t<{14}>%column_number%</{14}>\n" +
				"\t\t</{9}>\n",
				templateTagList);
		}

		protected XmlLogWriterBase()
			: base(TemplateLineDefault, TemplateIndentDefault, GetEntryTemplate(TemplateTagListDefault), GetDataTemplate(TemplateTagListDefault), GetStackTemplate(TemplateTagListDefault))
		{
			this._tagTemplates = TemplateTagListDefault;
		}

		protected XmlLogWriterBase(
			string templateLine,
			string templateIndent)
			: base(templateLine, templateIndent, GetEntryTemplate(TemplateTagListDefault), GetDataTemplate(TemplateTagListDefault), GetStackTemplate(TemplateTagListDefault))
		{
			if (templateLine == null)
			{
				throw new ArgumentNullException("templateLine");
			}

			if (templateIndent == null)
			{
				throw new ArgumentNullException("templateIndent");
			}

			this._tagTemplates = TemplateTagListDefault;
		}

		protected XmlLogWriterBase(
			string templateLine,
			string templateIndent,
			string[] templateTagList)
			: base(templateLine, templateIndent, GetEntryTemplate(templateTagList), GetDataTemplate(templateTagList), GetStackTemplate(templateTagList))
		{
			if (templateLine == null)
			{
				throw new ArgumentNullException("templateLine");
			}

			if (templateIndent == null)
			{
				throw new ArgumentNullException("templateIndent");
			}

			if (templateTagList == null)
			{
				throw new ArgumentNullException("templateTagList");
			}

			this._tagTemplates = templateTagList;
		}

		#endregion

		#region Properties

		[Editor("System.Windows.Forms.Design.StringArrayEditor, System.Design, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b03f5f7f11d50a3a", typeof(UITypeEditor))]
		public string[] TagTemplates
		{
			get { return this._tagTemplates; }
		}

		public string TagRoot
		{
			get { return this._tagTemplates[TagIndex.Root]; }
			set { this._tagTemplates[TagIndex.Root] = value; }
		}

		public string TagEntry
		{
			get { return this._tagTemplates[TagIndex.Entry]; }
			set { this._tagTemplates[TagIndex.Entry] = value; }
		}

		public string TagData
		{
			get { return this._tagTemplates[TagIndex.Data]; }
			set { this._tagTemplates[TagIndex.Data] = value; }
		}

		public string TagStack
		{
			get { return this._tagTemplates[TagIndex.Stack]; }
			set { this._tagTemplates[TagIndex.Stack] = value; }
		}

		public string TagLevel
		{
			get { return this._tagTemplates[TagIndex.Level]; }
			set { this._tagTemplates[TagIndex.Level] = value; }
		}

		public string TagSource
		{
			get { return this._tagTemplates[TagIndex.Source]; }
			set { this._tagTemplates[TagIndex.Source] = value; }
		}

		public string TagDateTime
		{
			get { return this._tagTemplates[TagIndex.DateTime]; }
			set { this._tagTemplates[TagIndex.DateTime] = value; }
		}

		public string TagThreadID
		{
			get { return this._tagTemplates[TagIndex.ThreadID]; }
			set { this._tagTemplates[TagIndex.ThreadID] = value; }
		}

		public string TagDatum
		{
			get { return this._tagTemplates[TagIndex.Datum]; }
			set { this._tagTemplates[TagIndex.Datum] = value; }
		}

		public string TagStackFrame
		{
			get { return this._tagTemplates[TagIndex.StackFrame]; }
			set { this._tagTemplates[TagIndex.StackFrame] = value; }
		}

		public string TagAssemblyName
		{
			get { return this._tagTemplates[TagIndex.AssemblyName]; }
			set { this._tagTemplates[TagIndex.AssemblyName] = value; }
		}

		public string TagMethodName
		{
			get { return this._tagTemplates[TagIndex.MethodName]; }
			set { this._tagTemplates[TagIndex.MethodName] = value; }
		}

		public string TagFileName
		{
			get { return this._tagTemplates[TagIndex.FileName]; }
			set { this._tagTemplates[TagIndex.FileName] = value; }
		}

		public string TagLineNumber
		{
			get { return this._tagTemplates[TagIndex.LineNumber]; }
			set { this._tagTemplates[TagIndex.LineNumber] = value; }
		}

		public string TagColumnNumber
		{
			get { return this._tagTemplates[TagIndex.ColumnNumber]; }
			set { this._tagTemplates[TagIndex.ColumnNumber] = value; }
		}

		#endregion

		#endregion

		#region ILogWriter

		public abstract override void Initialize();
		public abstract override void Terminate();
		public abstract override void Log(LogLevel level, LogSource source, DateTime dateTime, int threadID, object[] data, IList<string[]> stackFrameList);

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
