﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Logging Library                          */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Text;

namespace Nabu.Logging.LogWriters
{
	public sealed class XmlMultiFileLogWriter : XmlLogWriterBase
	{
		#region LogWriterXmlMultiFile

		#region Fields

		private string _wrapperPathTemplate = "\\Log-{0}-{1}.xml";
		private string _dataPathTemplate = "\\Log-{0}-{1}-Data.txt";
		private Dictionary<string, TextWriter> _textWriterMap = new Dictionary<string, TextWriter>();

		#endregion

		#region Methods

		public XmlMultiFileLogWriter()
		{
		}

		public XmlMultiFileLogWriter(string templatePathData, string templatePathWrapper)
		{
			if (templatePathData == null)
			{
				throw new ArgumentNullException("templatePathData");
			}

			if (templatePathWrapper == null)
			{
				throw new ArgumentNullException("templatePathWrapper");
			}

			this._dataPathTemplate = templatePathData;
			this._wrapperPathTemplate = templatePathWrapper;
		}

		public XmlMultiFileLogWriter(string templatePathData, string templatePathWrapper, string templateLine, string templateIndent)
			: base(templateLine, templateIndent)
		{
			if (templatePathData == null)
			{
				throw new ArgumentNullException("templatePathData");
			}

			if (templatePathWrapper == null)
			{
				throw new ArgumentNullException("templatePathWrapper");
			}

			this._dataPathTemplate = templatePathData;
			this._wrapperPathTemplate = templatePathWrapper;
		}

		public XmlMultiFileLogWriter(string templatePathData, string templatePathWrapper, string templateLine, string templateIndent, string[] templateTagList)
			: base(templateLine, templateIndent, templateTagList)
		{
			if (templatePathData == null)
			{
				throw new ArgumentNullException("templatePathData");
			}

			if (templatePathWrapper == null)
			{
				throw new ArgumentNullException("templatePathWrapper");
			}

			this._dataPathTemplate = templatePathData;
			this._wrapperPathTemplate = templatePathWrapper;
		}

		#endregion

		#region Properties

		public string WrapperPathTemplate
		{
			get { return this._wrapperPathTemplate; }
			set { this._wrapperPathTemplate = value; }
		}

		public string DataPathTemplate
		{
			get { return this._dataPathTemplate; }
			set { this._dataPathTemplate = value; }
		}

		#endregion

		#endregion

		#region ILogWriter

		public override void Initialize()
		{
			try
			{
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		public override void Terminate()
		{
			try
			{
				foreach (TextWriter textWriter in this._textWriterMap.Values)
				{
					textWriter.Close();
				}
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		public override void Log(LogLevel level, LogSource source, DateTime dateTime, int threadID, object[] data, IList<string[]> stackFrameList)
		{
			if (level == null)
			{
				throw new ArgumentNullException("level");
			}

			if (source == null)
			{
				throw new ArgumentNullException("source");
			}

			if (data == null)
			{
				throw new ArgumentNullException("data");
			}

			if (stackFrameList == null)
			{
				throw new ArgumentNullException("stackFrameList");
			}

			try
			{
				string dataPath = LogFilePathExpander.Expand(
					this._dataPathTemplate
						.Replace("%level%", level.ToString())
						.Replace("%source%", source.ToString()));

				if (dataPath.StartsWith(new string(Path.DirectorySeparatorChar, 1)) || dataPath.StartsWith(new string(Path.AltDirectorySeparatorChar, 1)))
				{
					dataPath = SpecialPath.CurrentUserCurrentMachineSettingsPath + dataPath;
				}

				TextWriter textWriterData;

				if (!this._textWriterMap.TryGetValue(dataPath, out textWriterData))
				{
					Directory.CreateDirectory(Path.GetDirectoryName(dataPath));

					string wrapperPath = LogFilePathExpander.Expand(
						this._wrapperPathTemplate
							.Replace("%level%", level.ToString())
							.Replace("%source%", source.ToString()));

					if (wrapperPath.StartsWith(new string(Path.DirectorySeparatorChar, 1)) || wrapperPath.StartsWith(new string(Path.AltDirectorySeparatorChar, 1)))
					{
						wrapperPath = SpecialPath.CurrentUserCurrentMachineSettingsPath + wrapperPath;
					}

					Directory.CreateDirectory(Path.GetDirectoryName(wrapperPath));
					Directory.CreateDirectory(Path.GetDirectoryName(dataPath));

					using (TextWriter textWriterWrapper = new StreamWriter(new FileStream(wrapperPath, FileMode.Create, FileAccess.Write, FileShare.Read), Encoding.UTF8))
					{
						textWriterWrapper.WriteLine(string.Format(
							CultureInfo.InvariantCulture.NumberFormat,
							"<?xml version=\"1.0\" encoding=\"utf-8\"?>{0}" +
							"<!DOCTYPE nabu_log [<!ENTITY {2} SYSTEM \"{3}\">]>{0}" +
							"<{4}>{0}" +
							"{1}&{2};{0}" +
							"</{4}>{0}",
							this.LineTemplate,
							this.IndentTemplate,
							this.TagTemplates[TagIndex.Entry],
							(new Uri(wrapperPath))
								.MakeRelativeUri(new Uri(dataPath))
								.OriginalString
								.Replace(Path.AltDirectorySeparatorChar, Path.DirectorySeparatorChar)
								.Replace("&", "&amp;"),
							this.TagTemplates[TagIndex.Root]));
						textWriterWrapper.Close();
					}

					textWriterData = new StreamWriter(new FileStream(dataPath, FileMode.Append, FileAccess.Write, FileShare.Read), Encoding.UTF8);

					this._textWriterMap.Add(dataPath, textWriterData);
				}

				textWriterData.Write(base.GetLogString(level, source, dateTime, threadID, data, stackFrameList));
				textWriterData.Flush();
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
