﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Logging Library                          */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.ComponentModel;

using Nabu.Collections;

namespace Nabu.Logging
{
	/// <summary>
	/// Default <see cref="ILogFormatter"/> implementation.
	/// </summary>
	public class LogFormatterDefault : ILogFormatter
	{
		#region LogFormatterDefault

		#region Fields

		private SieveCache<object, string> _stringMap = new SieveCache<object, string>();
		private Dictionary<Type, ILogFormatter> _formatterMap = new Dictionary<Type, ILogFormatter>();
		private string[] _templateList = LogFormatterTemplate.DefaultTemplateList;

		#endregion

		#region Methods

		private LogFormatterDefault()
		{
			Add(typeof(Guid), new LogFormatterGuid());
			Add(typeof(DateTime), new LogFormatterDateTime());
			Add(typeof(char), new LogFormatterChar());
			Add(typeof(string), new LogFormatterString());
			Add(typeof(Exception), new LogFormatterException());
		}

		public void Add(Type type, ILogFormatter formatter)
		{
			formatter.Initialize(this._templateList);

			this._formatterMap.Add(type, formatter);
		}

		public void Remove(Type type)
		{
			this._formatterMap.Remove(type);
		}

		#endregion

		#endregion

		#region ILogFormatter

		public void Initialize(string[] templateList)
		{
			this._templateList = templateList;
		}

		public string GetLogString(object value, string lineString, string indentString, int indentValue)
		{
			string logString = null;

			if (!this._stringMap.TryGetValue(value, out logString))
			{
				Type type = value.GetType();
				ILogFormatter formatter;

				if (!this._formatterMap.TryGetValue(type, out formatter))
				{
					foreach (Attribute attribute in TypeDescriptor.GetAttributes(value))
					{
						LogFormatterAttribute logFormatterAttribute = attribute as LogFormatterAttribute;

						if (logFormatterAttribute != null)
						{
							formatter = logFormatterAttribute.TypeLogFormatter.GetConstructor(Type.EmptyTypes).Invoke(null) as ILogFormatter;
							formatter.Initialize(this._templateList);
						}
					}
				}

				if (formatter != null)
				{
					logString = formatter.GetLogString(value, lineString, indentString, indentValue);
				}
				else
				{
					foreach (KeyValuePair<Type, ILogFormatter> typeFormatter in this._formatterMap)
					{
						if (typeFormatter.Key.IsAssignableFrom(type))
						{
							formatter = typeFormatter.Value;
							break;
						}
					}

					if (formatter != null)
					{
						this._formatterMap.Add(type, formatter);
						logString = formatter.GetLogString(value, lineString, indentString, indentValue);
					}
					else
					{
						string indent = LogFormatterIndent.GetIndent(indentString, indentValue);

						if (type.IsValueType)
						{
							if (type == typeof(bool))
							{
								logString = string.Format(
									this._templateList[LogFormatterTemplate.Boolean], lineString, indent,
									((bool)value) ? this._templateList[LogFormatterTemplate.BooleanTrue] : this._templateList[LogFormatterTemplate.BooleanFalse]);
							}
							else if (type == typeof(sbyte))
							{
								logString = string.Format(this._templateList[LogFormatterTemplate.Int8], lineString, indent, value);
							}
							else if (type == typeof(byte))
							{
								logString = string.Format(this._templateList[LogFormatterTemplate.UInt8], lineString, indent, value);
							}
							else if (type == typeof(short))
							{
								logString = string.Format(this._templateList[LogFormatterTemplate.Int16], lineString, indent, value);
							}
							else if (type == typeof(ushort))
							{
								logString = string.Format(this._templateList[LogFormatterTemplate.UInt16], lineString, indent, value);
							}
							else if (type == typeof(int))
							{
								logString = string.Format(this._templateList[LogFormatterTemplate.Int32], lineString, indent, value);
							}
							else if (type == typeof(uint))
							{
								logString = string.Format(this._templateList[LogFormatterTemplate.UInt32], lineString, indent, value);
							}
							else if (type == typeof(long))
							{
								logString = string.Format(this._templateList[LogFormatterTemplate.Int64], lineString, indent, value);
							}
							else if (type == typeof(ulong))
							{
								logString = string.Format(this._templateList[LogFormatterTemplate.UInt64], lineString, indent, value);
							}
							else if (type == typeof(float))
							{
								logString = string.Format(this._templateList[LogFormatterTemplate.Single], lineString, indent, value);
							}
							else if (type == typeof(double))
							{
								logString = string.Format(this._templateList[LogFormatterTemplate.Double], lineString, indent, value);
							}
							else if (type.IsEnum)
							{
								logString = string.Format(this._templateList[LogFormatterTemplate.Enum], lineString, indent, Enum.GetName(type, value));
							}
							else
							{
								logString = string.Format(this._templateList[LogFormatterTemplate.Struct], lineString, indent, value);
							}
						}
						else if (type.IsClass)
						{
							logString = string.Format(this._templateList[LogFormatterTemplate.Class], lineString, indent, value);
						}
						else if (type.IsInterface)
						{
							logString = string.Format(this._templateList[LogFormatterTemplate.Interface], lineString, indent, value);
						}
						else
						{
							logString = string.Format(this._templateList[LogFormatterTemplate.Default], lineString, indent, value);
						}
					}
				}
			}

			return logString;
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
