﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Logging Library                          */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Configuration;
using System.Diagnostics;
using System.Reflection;
using System.Threading;

using Nabu.Collections;
using Nabu.Data;
using Nabu.Reflection;

namespace Nabu.Logging
{
	/// <summary>
	/// Logging manager. Allows add and remove log writers, write log entries.
	/// </summary>
	public sealed class LogManager
	{
		#region Constants

		private const int StackFrameCountLogger = 2;

		#endregion

		#region Fields

		private Dictionary<LogLevel, Dictionary<LogSource, List<ILogWriter>>> _writerMap = new Dictionary<LogLevel, Dictionary<LogSource, List<ILogWriter>>>();
		private Dictionary<ILogWriter, int> _writerReferences = new Dictionary<ILogWriter, int>();
		private Dictionary<LogLevel, Dictionary<LogSource, bool>> _modeMap = new Dictionary<LogLevel, Dictionary<LogSource, bool>>();

		private SieveCache<MethodBase, string> _methodNameMap = new SieveCache<MethodBase, string>();
		private ProgrammingLanguage _language = ProgrammingLanguage.CSharp;

		private Queue<LogAsyncRequest> _queueAsync = new Queue<LogAsyncRequest>();
		private Thread _threadAsync;
		private object _sync = new object();

		#endregion

		#region Methods

		private void CurrentDomain_DomainUnload(object sender, EventArgs e)
		{
			try
			{
				if (this._threadAsync != null)
				{
					this._threadAsync.Join();
				}
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
			}
		}

		private void ThreadProcAsyncLog()
		{
			for (; ; )
			{
				if (this._queueAsync.Count > 0)
				{
					lock (this._sync)
					{
						if (this._queueAsync.Count > 0)
						{
							LogAsyncRequest asyncRequest = this._queueAsync.Dequeue();

							lock (asyncRequest.Writer)
							{
								asyncRequest.Writer.Log(asyncRequest.Level, asyncRequest.Source, asyncRequest.DateTime, asyncRequest.ThreadID, asyncRequest.Data, asyncRequest.StackFrames);
							}
						}
					}
				}
				else
				{
					this._threadAsync = null;

					return;
				}
			}
		}

		private void InternalAddWriter(ILogWriter writer, LogLevel[] levelList, LogSource[] sourceList)
		{
			lock (this._sync)
			{
				if (levelList == null)
				{
					levelList = LogLevel.SinkList;
				}

				if (sourceList == null)
				{
					sourceList = LogSource.SinkList;
				}

				foreach (LogLevel level in levelList)
				{
					Dictionary<LogSource, List<ILogWriter>> writerMapLevel;

					if (!this._writerMap.TryGetValue(level, out writerMapLevel))
					{
						writerMapLevel = new Dictionary<LogSource, List<ILogWriter>>();
						this._writerMap.Add(level, writerMapLevel);
					}

					foreach (LogSource source in sourceList)
					{
						List<ILogWriter> logWriters;

						if (!writerMapLevel.TryGetValue(source, out logWriters))
						{
							logWriters = new List<ILogWriter>();
							writerMapLevel.Add(source, logWriters);
						}

						if (!logWriters.Contains(writer))
						{
							int referenceCount;

							if (this._writerReferences.TryGetValue(writer, out referenceCount))
							{
								this._writerReferences[writer] = referenceCount + 1;
							}
							else
							{
								this._writerReferences[writer] = 1;

								writer.Initialize();
							}

							logWriters.Add(writer);
						}
					}
				}
			}
		}

		private void InternalRemoveWriter(ILogWriter writer, LogLevel[] levels, LogSource[] sources)
		{
			lock (this._sync)
			{
				foreach (LogLevel level in levels)
				{
					Dictionary<LogSource, List<ILogWriter>> writerMapLevel;

					if (this._writerMap.TryGetValue(level, out writerMapLevel))
					{
						foreach (LogSource source in sources)
						{
							List<ILogWriter> logWriters;

							if (writerMapLevel.TryGetValue(source, out logWriters))
							{
								if (logWriters.Contains(writer))
								{
									int referenceCount = this._writerReferences[writer];

									if (referenceCount == 1)
									{
										logWriters.Remove(writer);

										this._writerReferences.Remove(writer);
									}
									else
									{
										this._writerReferences[writer] = referenceCount - 1;
									}
								}

								if (logWriters.Count == 0)
								{
									writerMapLevel.Remove(source);
								}
							}

							if (writerMapLevel.Count == 0)
							{
								this._writerMap.Remove(level);
							}
						}
					}
				}
			}
		}

		private string GetMethodName(MethodBase methodInfo)
		{
			string methodName = null;

			if (!this._methodNameMap.TryGetValue(methodInfo, out methodName))
			{
				methodName = MemberDeclaration.GetDeclaration(methodInfo, this._language);

				this._methodNameMap.Add(methodInfo, methodName, true);
			}

			return methodName;
		}

		private void InternalSetMode(LogLevel[] levelList, LogSource[] sourceList, bool isAsync)
		{
			foreach (LogLevel level in levelList)
			{
				Dictionary<LogSource, bool> modeMapLevel;

				if (!this._modeMap.TryGetValue(level, out modeMapLevel))
				{
					modeMapLevel = new Dictionary<LogSource, bool>();
					this._modeMap.Add(level, modeMapLevel);
				}

				foreach (LogSource source in sourceList)
				{
					modeMapLevel[source] = isAsync;
				}
			}
		}

		private void InternalLog(LogLevel level, LogSource source, object[] data)
		{
			try
			{
				List<ILogWriter> logWriters = null;
				Dictionary<LogSource, List<ILogWriter>> writerMapList;

				if (this._writerMap.TryGetValue(level, out writerMapList))
				{
					if (writerMapList.TryGetValue(source, out logWriters))
					{
					}
					else
					{
						if (writerMapList.TryGetValue(LogSource.Sink, out logWriters))
						{
						}
						else
						{
							logWriters = null;
						}
					}
				}
				else
				{
					if (this._writerMap.TryGetValue(LogLevel.Sink, out writerMapList))
					{
						if (writerMapList.TryGetValue(source, out logWriters))
						{
						}
						else
						{
							if (writerMapList.TryGetValue(LogSource.Sink, out logWriters))
							{
							}
							else
							{
								logWriters = null;
							}
						}
					}
					else
					{
						logWriters = null;
					}
				}

				if (logWriters != null)
				{
					StackTrace stackTrace = new StackTrace(true);
					IList<string[]> stackFrames = new List<string[]>(16);

					for (int frameIndex = StackFrameCountLogger; frameIndex < stackTrace.FrameCount; frameIndex++)
					{
						StackFrame stackTraceFrame = stackTrace.GetFrame(frameIndex);
						MethodBase methodInfo = stackTraceFrame.GetMethod();
						string[] stackFrame = new string[LogStackFrameElement.TotalCount]
						{
							/* LogStackFrameElement.AssemblyName */ methodInfo.DeclaringType.Assembly.Location,
							/* LogStackFrameElement.MethodName   */ GetMethodName(methodInfo),
							/* LogStackFrameElement.FileName     */ stackTraceFrame.GetFileName(),
							/* LogStackFrameElement.LineNumber   */ stackTraceFrame.GetFileLineNumber().ToString(),
							/* LogStackFrameElement.ColumnNumber */ stackTraceFrame.GetFileColumnNumber().ToString()
						};

						for (int index = 0; index < stackFrame.Length; index++)
						{
							if (stackFrame[index] == null)
							{
								stackFrame[index] = string.Empty;
							}
						}

						stackFrames.Add(stackFrame);
					}

					DateTime dateTime = DateTime.Now;

					bool isAsync = false;
					Dictionary<LogSource, bool> modeMapLevel;

					if (this._modeMap.TryGetValue(level, out modeMapLevel))
					{
						if (modeMapLevel.TryGetValue(source, out isAsync))
						{
						}
						else
						{
							isAsync = false;
						}
					}

					if (isAsync)
					{
						foreach (ILogWriter writer in logWriters)
						{
							lock (this._sync)
							{
								this._queueAsync.Enqueue(new LogAsyncRequest(writer, level, source, dateTime, Thread.CurrentThread.ManagedThreadId, data, stackFrames));
							}
						}

						lock (this._sync)
						{
							if (this._threadAsync == null)
							{
								this._threadAsync = new Thread(new ThreadStart(ThreadProcAsyncLog));
								this._threadAsync.Priority = ThreadPriority.Lowest;
								this._threadAsync.Start();
							}
						}
					}
					else
					{
						foreach (ILogWriter writer in logWriters)
						{
							lock (writer)
							{
								writer.Log(level, source, dateTime, Thread.CurrentThread.ManagedThreadId, data, stackFrames);
							}
						}
					}
				}
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		private static string ProcessTypeName(string typeName)
		{
			if (typeName.IndexOf('.') != -1)
			{
				return typeName;
			}
			else
			{
				if (typeName.EndsWith("LogWriter"))
				{
					return string.Format("Nabu.Logging.LogWriters.{0}, Nabu.Logging.LogWriters", typeName);
				}
				else
				{
					switch (typeName.ToLower())
					{
						case "bool":
							return "System.Boolean";
						case "char":
							return "System.Char";
						case "byte":
							return "System.Byte";
						case "sbyte":
							return "System.SByte";
						case "short":
							return "System.Int16";
						case "ushort":
							return "System.UInt16";
						case "int":
							return "System.Int32";
						case "uint":
							return "System.UInt32";
						case "long":
							return "System.Int64";
						case "ulong":
							return "System.UInt64";
						case "decimal":
							return "System.Decimal";
						case "float":
							return "System.Single";
						case "double":
							return "System.Double";
						case "intptr":
							return "System.IntPtr";
						default:
							return string.Format("System.{0}", typeName);
					}
				}
			}
		}

		private void ProcessConfigurationWriterData(LogXmlSection.LogWriterData writerData, LogLevel[] levelList, LogSource[] sourceList)
		{
			try
			{
				Type writerType = Type.GetType(ProcessTypeName(writerData.Type), true, true);

				List<Type> parameterTypes = new List<Type>();
				List<object> parameterValues = new List<object>();
				List<KeyValuePair<PropertyInfo, object>> properties = new List<KeyValuePair<PropertyInfo, object>>();

				foreach (LogXmlSection.LogParameterData parameterData in writerData.Parameters)
				{
					if (string.IsNullOrEmpty(parameterData.Name))
					{
						Type parameterType = Type.GetType(ProcessTypeName(parameterData.Type), true, true);

						parameterTypes.Add(parameterType);
						parameterValues.Add(Convert.ChangeType(parameterData.Value, parameterType));
					}
					else
					{
						PropertyInfo propertyInfo = writerType.GetProperty(parameterData.Name);

						properties.Add(new KeyValuePair<PropertyInfo, object>(propertyInfo, Convert.ChangeType(parameterData.Value, propertyInfo.PropertyType)));
					}
				}

				ILogWriter writer = (ILogWriter)writerType.GetConstructor(parameterTypes.ToArray()).Invoke(parameterValues.ToArray());

				foreach (KeyValuePair<PropertyInfo, object> property in properties)
				{
					property.Key.SetValue(writer, property.Value, null);
				}

				InternalAddWriter(writer, levelList, sourceList);
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		private void ProcessConfigurationGroupData(LogXmlSection.LogGroupData groupData)
		{
			foreach (LogXmlSection.LogWriterData writerData in groupData.Writers)
			{
				ProcessConfigurationWriterData(writerData, groupData.Levels, groupData.Sources);
			}
		}

		private void ProcessConfigurationSectionData(LogXmlSection.LogSectionData logSectionData)
		{
			if (logSectionData != null)
			{
				foreach (LogXmlSection.LogGroupData groupData in logSectionData.Groups)
				{
					ProcessConfigurationGroupData(groupData);
				}
			}
		}

		private LogManager()
		{
			try
			{
				ProcessConfigurationSectionData((LogXmlSection.LogSectionData)ConfigurationManager.GetSection("log_nabu"));

				AppDomain.CurrentDomain.DomainUnload += new EventHandler(CurrentDomain_DomainUnload);
			}
			catch (Exception exception)
			{
#if TRACE
				LogWriterInternal.LogException(exception);
#endif
				throw;
			}
		}

		public static void AddWriter(ILogWriter writer, LogLevel[] levelList, LogSource[] sourceList)
		{
			Singleton<LogManager>.Instance.InternalAddWriter(writer, levelList, sourceList);
		}

		public static void RemoveWriter(ILogWriter writer, LogLevel[] levelList, LogSource[] sourceList)
		{
			Singleton<LogManager>.Instance.InternalRemoveWriter(writer, levelList, sourceList);
		}

		public static void SetMode(LogLevel[] levelList, LogSource[] sourceList, bool isAsync)
		{
			Singleton<LogManager>.Instance.InternalSetMode(levelList, sourceList, isAsync);
		}

		public static void Log(params object[] data)
		{
			Singleton<LogManager>.Instance.InternalLog(LogLevel.Unknown, LogSource.Unknown, data);
		}

		public static void Log(LogLevel level, params object[] data)
		{
			Singleton<LogManager>.Instance.InternalLog(level, LogSource.Unknown, data);
		}

		public static void Log(LogSource source, params object[] data)
		{
			Singleton<LogManager>.Instance.InternalLog(LogLevel.Unknown, source, data);
		}

		public static void Log(LogLevel level, LogSource source, params object[] data)
		{
			Singleton<LogManager>.Instance.InternalLog(level, source, data);
		}

		#endregion

		#region Properties

		private ProgrammingLanguage InternalLanguage
		{
			get
			{
				return this._language;
			}

			set
			{
				lock (this._sync)
				{
					this._language = value;
					this._methodNameMap.Clear();
				}
			}
		}

		public static ProgrammingLanguage Language
		{
			get { return Singleton<LogManager>.Instance.InternalLanguage; }
			set { Singleton<LogManager>.Instance.InternalLanguage = value; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
