﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Mobile Library                           */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.IO;

namespace Nabu.Mobile
{
	public sealed class MobileFileStream : Stream
	{
		#region MobileFileStream

		#region Fields

		private MobileSession _session;
		private string _path;
		private IntPtr _hFile;
		private FileAccess _access;

		#endregion

		#region Methods

		[DebuggerStepThrough]
		internal MobileFileStream(
			MobileSession session,
			string path,
			FileMode mode,
			FileAccess access,
			FileShare share)
		{
			this._session = session;
			this._session.RegisterSessionObject(this);
			this._access = access;
			this._path = path;
			this._hFile =
				this._session.RapiSession.CeCreateFile(
					this._path,
					((uint)(this._access & FileAccess.Read) * 0x80000000) + ((uint)(this._access & FileAccess.Write) * 0x20000000),
					share,
					IntPtr.Zero,
					mode,
					(FileAttributes)0,
					IntPtr.Zero);
		}

		#endregion

		#region Properties

		public MobileFileTimes Times
		{
			[SuppressMessage("Microsoft.Design", "CA1065:DoNotRaiseExceptionsInUnexpectedLocations")]
			get
			{
				long creation;
				long lastAccess;
				long lastWrite;

				if (!this._session.RapiSession.CeGetFileTime(
					this._hFile,
					out creation,
					out lastAccess,
					out lastWrite))
				{
					throw new MobileException(this._session, "CeGetFileTime", this._path);
				}

				creation = ((creation & 0xFFFFFFFF) << 32) + (creation >> 32);
				lastAccess = ((lastAccess & 0xFFFFFFFF) << 32) + (lastAccess >> 32);
				lastWrite = ((lastWrite & 0xFFFFFFFF) << 32) + (lastWrite >> 32);

				return
					new MobileFileTimes(
						DateTime.FromFileTime(creation),
						DateTime.FromFileTime(lastAccess),
						DateTime.FromFileTime(lastWrite));
			}

			set
			{
				long creation = value.Creation.ToFileTime();
				long lastAccess = value.LastAccess.ToFileTime();
				long lastWrite = value.LastWrite.ToFileTime();

				creation = ((creation & 0xFFFFFFFF) << 32) + (creation >> 32);
				lastAccess = ((lastAccess & 0xFFFFFFFF) << 32) + (lastAccess >> 32);
				lastWrite = ((lastWrite & 0xFFFFFFFF) << 32) + (lastWrite >> 32);

				if (!this._session.RapiSession.CeSetFileTime(
					this._hFile,
					ref creation,
					ref lastAccess,
					ref lastWrite))
				{
					throw new MobileException(this._session, "CeSetFileTime", this._path);
				}
			}
		}

		#endregion

		#endregion

		#region Stream

		#region Methods

		unsafe public override int Read(byte[] buffer, int offset, int count)
		{
			int bytesRead;

			fixed (byte* lpDataBuffer = &buffer[offset])
			{
				if (!this._session.RapiSession.CeReadFile(
					this._hFile,
					lpDataBuffer,
					count,
					out bytesRead,
					IntPtr.Zero))
				{
					throw new MobileException(this._session, "CeReadFile", this._path);
				}
			}

			return bytesRead;
		}

		unsafe public override void Write(byte[] buffer, int offset, int count)
		{
			int bytesWritten = 0;

			for (int index = 0; index < count; index += bytesWritten)
			{
				fixed (byte* lpDataBuffer = &buffer[offset + index])
				{
					if (!this._session.RapiSession.CeWriteFile(
						this._hFile,
						lpDataBuffer,
						Math.Min(count - index, 0x10000),
						out bytesWritten,
						IntPtr.Zero))
					{
						throw new MobileException(this._session, "CeWriteFile", this._path);
					}
				}
			}
		}

		public override long Seek(long offset, SeekOrigin origin)
		{
			int high32 = (int)(offset >> 32);
			int low32 =
				this._session.RapiSession.CeSetFilePointer(
					this._hFile,
					(int)(offset & 0xFFFFFFFF),
					ref high32,
					origin);

			return (high32 << 32) + low32;
		}

		public override void SetLength(long value)
		{
			long position = this.Position;

			this.Position = value;
			this._session.RapiSession.CeSetEndOfFile(this._hFile);
			this.Position = position;
		}

		public override void Flush()
		{
		}

		#endregion

		#region Properties

		public override bool CanRead
		{
			[DebuggerStepThrough]
			get { return (this._access & FileAccess.Read) != 0; }
		}

		public override bool CanWrite
		{
			[DebuggerStepThrough]
			get { return (this._access & FileAccess.Write) != 0; }
		}

		public override bool CanSeek
		{
			[DebuggerStepThrough]
			get { return true; }
		}

		public override long Length
		{
			get
			{
				int high32;
				int low32 =
					this._session.RapiSession.CeGetFileSize(
						this._hFile,
						out high32);

				return (high32 << 32) + low32;
			}
		}

		public override long Position
		{
			get
			{
				int high32 = 0;
				int low32 =
					this._session.RapiSession.CeSetFilePointer(
						this._hFile,
						0,
						ref high32,
						SeekOrigin.Current);

				return (high32 << 32) + low32;
			}

			set
			{
				int high32 = (int)(value >> 32);
				/* int low32 = */
				this._session.RapiSession.CeSetFilePointer(
					this._hFile,
					(int)(value & 0xFFFFFFFF),
					ref high32,
					SeekOrigin.Begin);
			}
		}

		#endregion

		#endregion

		#region IDisposable

		protected override void Dispose(bool disposing)
		{
			if (
				(this._hFile != IntPtr.Zero) &&
				(this._hFile != new IntPtr(-1)))
			{
				this._session.UnregisterSessionObject(this);
				this._session.RapiSession.CeCloseHandle(this._hFile);
			}

			base.Dispose(disposing);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
