﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Mobile Library                           */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Text;

namespace Nabu.Mobile
{
	public sealed class MobileRegistryKey : DisposableBase
	{
		#region MobileRegistryKey

		#region Fields

		private MobileSession _session;
		private MobileRegistryKey _parentKey;
		private IntPtr _hKey;
		private MobileRegistryRights _rights;
		private string _name;
		private string _class;

		#endregion

		#region Methods

		#region Utility

		internal void CheckBound()
		{
			if (this._hKey == IntPtr.Zero)
			{
				if (this._parentKey == null)
				{
					throw new InvalidOperationException("_hKey");
				}
				else
				{
					IntPtr hKey;
					SystemResult result =
						this._session.RapiSession.CeRegOpenKeyEx(
							this._parentKey.Handle,
							this._name,
							0,
							this._rights,
							out hKey);

					NativeResult.ThrowOnWin32Error(result);

					this._hKey = hKey;
				}
			}
		}

		internal void CheckUnbound()
		{
			if (this._hKey != IntPtr.Zero)
			{
				throw new InvalidOperationException("_hKey");
			}
		}

		#endregion

		#region Constructors

		internal MobileRegistryKey(
			MobileSession session,
			MobileRegistryKey parentKey,
			IntPtr hKey,
			MobileRegistryRights rights,
			string name,
			string class_)
		{
			this._session = session;
			this._session.RegisterSessionObject(this);
			this._parentKey = parentKey;
			this._hKey = hKey;
			this._rights = rights;
			this._name = name;
			this._class = class_;
		}

		public MobileRegistryKey(
			string name,
			string class_)
		{
			this._name = name;
			this._class = class_;
		}

		#endregion

		#region Operations

		public void Close()
		{
			if (!IsPredefined)
			{
				if (this._hKey != IntPtr.Zero)
				{
					this._session.RapiSession.CeRegCloseKey(this._hKey);
					this._hKey = IntPtr.Zero;
				}
			}
		}

		#endregion

		#endregion

		#region Properties

		internal MobileSession Session
		{
			get
			{
				return this._session;
			}

			set
			{
				if (this._session != null)
				{
					this._session.UnregisterSessionObject(this);
				}

				this._session = value;

				if (this._session != null)
				{
					this._session.RegisterSessionObject(this);
				}
			}
		}

		internal MobileRegistryKey ParentKey
		{
			get
			{
				base.CheckDisposed();

				return this._parentKey;
			}

			set
			{
				base.CheckDisposed();

				this._parentKey = value;
			}
		}

		internal IntPtr Handle
		{
			get
			{
				base.CheckDisposed();

				return this._hKey;
			}

			set
			{
				base.CheckDisposed();

				this._hKey = value;
			}
		}

		internal MobileRegistryRights Rights
		{
			get
			{
				base.CheckDisposed();

				return this._rights;
			}

			set
			{
				base.CheckDisposed();

				this._rights = value;
			}
		}

		public MobileRegistryKeyCollection Keys
		{
			get
			{
				base.CheckDisposed();
				CheckBound();

				return new MobileRegistryKeyCollection(this._session, this);
			}
		}

		public MobileRegistryValueCollection Values
		{
			get
			{
				base.CheckDisposed();
				CheckBound();

				return new MobileRegistryValueCollection(this._session, this);
			}
		}

		public string Name
		{
			get
			{
				base.CheckDisposed();

				return this._name;
			}
		}

		public string Class
		{
			get
			{
				base.CheckDisposed();
				CheckBound();

				if (this._class == null)
				{
					StringBuilder class_ = new StringBuilder(256);
					int classLength = class_.Capacity;
					int subKeyCount;
					int subKeyMaxNameLength;
					int subKeyMaxClassLength;
					int valueCount;
					int valueMaxNameLength;
					int valueMaxDataLength;
					SystemResult result =
						this._session.RapiSession.CeRegQueryInfoKey(
							this._hKey,
							class_,
							ref classLength,
							IntPtr.Zero,
							out subKeyCount,
							out subKeyMaxNameLength,
							out subKeyMaxClassLength,
							out valueCount,
							out valueMaxNameLength,
							out valueMaxDataLength,
							IntPtr.Zero,
							IntPtr.Zero);

					NativeResult.ThrowOnWin32Error(result);

					class_.Length = classLength;

					this._class = class_.ToString();
				}

				return this._class;
			}
		}

		public bool IsPredefined
		{
			get
			{
				base.CheckDisposed();

				foreach (MobileRegistryKeys key in Enum.GetValues(typeof(MobileRegistryKeys)))
				{
					if ((this._hKey.ToInt64() & 0xFFFFFFFF) == (int)key)
					{
						return true;
					}
				}

				return false;
			}
		}

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			Close();
			this._session.UnregisterSessionObject(this);
			base.Dispose(disposing);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
