﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Mobile Library                           */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Text;

namespace Nabu.Mobile
{
	public sealed class MobileRegistryKeyCollection :
		DisposableBase,
		ICollection<MobileRegistryKey>
	{
		#region MobileRegistryKeyCollection

		#region Fields

		private MobileSession _session;
		private MobileRegistryKey _parentKey;
		private IDictionary<string, MobileRegistryKey> _keys;

		#endregion

		#region Methods

		#region Utility

		private void BuildKeyCollection()
		{
			if (this._keys == null)
			{
				StringBuilder class_ = new StringBuilder(256);
				int classLength = class_.Capacity;
				int subKeyCount;
				int subKeyMaxNameLength;
				int subKeyMaxClassLength;
				int valueCount;
				int valueMaxNameLength;
				int valueMaxDataLength;
				SystemResult result =
					this._session.RapiSession.CeRegQueryInfoKey(
						this._parentKey.Handle,
						class_,
						ref classLength,
						IntPtr.Zero,
						out subKeyCount,
						out subKeyMaxNameLength,
						out subKeyMaxClassLength,
						out valueCount,
						out valueMaxNameLength,
						out valueMaxDataLength,
						IntPtr.Zero,
						IntPtr.Zero);

				NativeResult.ThrowOnWin32Error(result);

				StringBuilder subKeyName = new StringBuilder(subKeyMaxNameLength + 1);
				StringBuilder subKeyClass = new StringBuilder(subKeyMaxClassLength + 1);
				int subKeyNameLength;
				int subKeyClassLength;

				this._keys = new SortedList<string, MobileRegistryKey>(subKeyCount);

				for (int index = 0; index < subKeyCount; index++)
				{
					subKeyNameLength = subKeyName.Capacity;
					subKeyClassLength = subKeyName.Capacity;

					result =
						this._session.RapiSession.CeRegEnumKeyEx(
							this._parentKey.Handle,
							index,
							subKeyName,
							ref subKeyNameLength,
							IntPtr.Zero,
							subKeyClass,
							ref subKeyClassLength,
							IntPtr.Zero);

					NativeResult.ThrowOnWin32Error(result);

					subKeyName.Length = subKeyNameLength;
					subKeyClass.Length = subKeyClassLength;

					this._keys.Add(
						new KeyValuePair<string, MobileRegistryKey>(
							subKeyName.ToString(),
							new MobileRegistryKey(
								this._session,
								this._parentKey,
								IntPtr.Zero,
								this._parentKey.Rights,
								subKeyName.ToString(),
								subKeyClass.ToString())));
				}
			}
		}

		private void AddKey(MobileRegistryKey item)
		{
			if (item.Rights == MobileRegistryRights.Undefined)
			{
				item.Rights = this._parentKey.Rights;
			}

			IntPtr hKey;
			int disposition;
			SystemResult result =
				this._session.RapiSession.CeRegCreateKeyEx(
					this._parentKey.Handle,
					item.Name,
					0,
					item.Class,
					0,
					item.Rights,
					IntPtr.Zero,
					out hKey,
					out disposition);

			NativeResult.ThrowOnWin32Error(result);

			item.Handle = hKey;
			item.Session = this._session;

			if (this._keys != null)
			{
				this._keys.Add(item.Name, item);
			}
		}

		public bool RemoveKey(string name)
		{
			SystemResult result =
				this._session.RapiSession.CeRegDeleteKey(
					this._parentKey.Handle,
					name);

			NativeResult.ThrowOnWin32Error(result);

			if (this._keys != null)
			{
				return this._keys.Remove(name);
			}

			return true;
		}

		public bool ContainsKey(string name)
		{
			BuildKeyCollection();

			return this._keys.ContainsKey(name);
		}

		#endregion

		#region Constructors

		internal MobileRegistryKeyCollection(
			MobileSession session,
			MobileRegistryKey parentKey)
		{
			this._session = session;
			this._session.RegisterSessionObject(this);
			this._parentKey = parentKey;
		}

		#endregion

		#region Operations

		public bool Remove(string name)
		{
			base.CheckDisposed();

			return RemoveKey(name);
		}

		#endregion

		#endregion

		#region Properties

		public MobileRegistryKey this[string name]
		{
			get
			{
				base.CheckDisposed();

				return this._keys[name];
			}
		}

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			this._session.UnregisterSessionObject(this);
			base.Dispose(disposing);
		}

		#endregion

		#region ICollection<MobileRegistryKey>

		#region Methods

		public void Add(MobileRegistryKey item)
		{
			base.CheckDisposed();

			item.CheckUnbound();

			AddKey(item);
		}

		public bool Remove(MobileRegistryKey item)
		{
			base.CheckDisposed();

			item.CheckBound();

			return RemoveKey(item.Name);
		}

		public void Clear()
		{
			base.CheckDisposed();

			throw new NotImplementedException();
		}

		public bool Contains(MobileRegistryKey item)
		{
			base.CheckDisposed();

			return ContainsKey(item.Name);
		}

		public void CopyTo(MobileRegistryKey[] array, int arrayIndex)
		{
			base.CheckDisposed();

			throw new NotImplementedException();
		}

		#endregion

		#region Properties

		public int Count
		{
			get
			{
				base.CheckDisposed();

				BuildKeyCollection();

				return this._keys.Count;
			}
		}

		public bool IsReadOnly
		{
			get
			{
				base.CheckDisposed();

				return false;
			}
		}

		#endregion

		#endregion

		#region IEnumerable<MobileRegistryKey>

		public IEnumerator<MobileRegistryKey> GetEnumerator()
		{
			base.CheckDisposed();

			BuildKeyCollection();

			return this._keys.Values.GetEnumerator();
		}

		#endregion

		#region IEnumerable

		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
		{
			return ((IEnumerable<MobileRegistryKey>)this).GetEnumerator();
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
