﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Mobile Library                           */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Win32;

namespace Nabu.Mobile
{
	public sealed class MobileRegistryValueCollection :
		DisposableBase,
		IDictionary<string, object>
	{
		#region MobileRegistryValueCollection

		#region Fields

		private MobileSession _session;
		private MobileRegistryKey _key;
		private IDictionary<string, object> _values;

		#endregion

		#region Methods

		#region Utility

		private static void PackValue(
			out byte[] valueData,
			out RegistryValueKind valueType,
			object valueObject)
		{
			if (valueObject.GetType() == typeof(string))
			{
				string value = (string)valueObject;

				valueData = new byte[Encoding.Unicode.GetByteCount(value) + 2];
				Encoding.Unicode.GetBytes(value, 0, value.Length, valueData, 0);
				valueType = RegistryValueKind.String;
			}
			else if (valueObject.GetType() == typeof(byte[]))
			{
				valueData = (byte[])valueObject;
				valueType = RegistryValueKind.Binary;
			}
			else if (valueObject.GetType() == typeof(int))
			{
				int value = (int)valueObject;

				valueData = new byte[]
				{
					(byte)((value >> 0) & 0xFF),
					(byte)((value >> 8) & 0xFF),
					(byte)((value >> 16) & 0xFF),
					(byte)((value >> 24) & 0xFF)
				};
				valueType = RegistryValueKind.DWord;
			}
			else if (valueObject.GetType() == typeof(string[]))
			{
				string[] values = (string[])valueObject;
				int capacity = 0;

				foreach (string value in values)
				{
					capacity = capacity + value.Length + 1;
				}

				StringBuilder stringBuilder = new StringBuilder(2 * capacity + 2);

				foreach (string value in values)
				{
					stringBuilder.Append(value);
					stringBuilder.Append('\0');
				}

				stringBuilder.Append('\0');

				valueData = Encoding.Unicode.GetBytes(stringBuilder.ToString());
				valueType = RegistryValueKind.String;
			}
			else if (valueObject.GetType() == typeof(long))
			{
				int value = (int)valueObject;

				valueData = new byte[]
				{
					(byte)((value >> 0)& 0xFF),
					(byte)((value >> 8)& 0xFF),
					(byte)((value >> 16)& 0xFF),
					(byte)((value >> 24)& 0xFF),
					(byte)((value >> 32)& 0xFF),
					(byte)((value >> 40)& 0xFF),
					(byte)((value >> 48)& 0xFF),
					(byte)((value >> 56)& 0xFF)
				};

				valueType = RegistryValueKind.DWord;
			}
			else
			{
				valueData = null;
				valueType = RegistryValueKind.Unknown;
			}
		}

		private static void UnpackValue(
			out object valueObject,
			byte[] valueData,
			int valueDataLength,
			RegistryValueKind valueType)
		{
			switch (valueType)
			{
				case RegistryValueKind.String:
					{
						int length;

						if (valueDataLength > valueData.Length)
						{
							length = (valueData.Length / 2);
						}
						else
						{
							length = valueDataLength;
						}

						length = (length == 0) ? 0 : length - 2;

						valueObject =
							Encoding.Unicode.GetString(
								valueData,
								0,
								length);
					}
					break;
				case RegistryValueKind.ExpandString:
					{
						int length;

						if (valueDataLength > valueData.Length)
						{
							length = (valueData.Length / 2);
						}
						else
						{
							length = valueDataLength;
						}

						length = (length == 0) ? 0 : length - 2;

						valueObject =
							Encoding.Unicode.GetString(
								valueData,
								0,
								length);
					}
					break;
				case RegistryValueKind.Binary:
					{
						byte[] result = new byte[valueDataLength];

						Array.Copy(valueData, result, valueDataLength);

						valueObject = result;
					}
					break;
				case RegistryValueKind.DWord:
					{
						int result =
							(valueData[3] << 24) |
							(valueData[2] << 16) |
							(valueData[1] << 8) |
							(valueData[0] << 0);

						valueObject = result;
					}
					break;
				case RegistryValueKind.MultiString:
					{
						int length;

						if (valueDataLength > valueData.Length)
						{
							length = (valueData.Length / 2);
						}
						else
						{
							length = valueDataLength;
						}

						length = (length == 0) ? 0 : length - 2;

						valueObject =
							Encoding.Unicode.GetString(
								valueData,
								0,
								length).Split('\0');
					}
					break;
				case RegistryValueKind.QWord:
					{
						long result =
							(valueData[7] << 56) |
							(valueData[6] << 48) |
							(valueData[5] << 40) |
							(valueData[4] << 32) |
							(valueData[3] << 24) |
							(valueData[2] << 16) |
							(valueData[1] << 8) |
							(valueData[0] << 0);

						valueObject = result;
					}
					break;
				default:
					valueObject = null;
					break;
			}
		}

		private void BuildValueCollection()
		{
			if (this._values == null)
			{
				StringBuilder class_ = new StringBuilder(256);
				int classLength = class_.Capacity;
				int subKeyCount;
				int subKeyMaxNameLength;
				int subKeyMaxClassLength;
				int valueCount;
				int valueMaxNameLength;
				int valueMaxDataLength;
				SystemResult result =
					this._session.RapiSession.CeRegQueryInfoKey(
						this._key.Handle,
						class_,
						ref classLength,
						IntPtr.Zero,
						out subKeyCount,
						out subKeyMaxNameLength,
						out subKeyMaxClassLength,
						out valueCount,
						out valueMaxNameLength,
						out valueMaxDataLength,
						IntPtr.Zero,
						IntPtr.Zero);

				NativeResult.ThrowOnWin32Error(result);

				StringBuilder valueName = new StringBuilder(valueMaxNameLength + 1);
				int valueNameLength;
				// RAPI2 over 32bit ActiveSync 4.5 returns incorrect value data
				// length for DisplayName at HKEY_LOCAL_MACHINE\\Comm\\VMini key
				// for HP iPAQ 614. I have no idea about other models and versions.
				// 64bit Vista is OK
				byte[] valueData = new byte[Math.Max(1024, 2 * valueMaxDataLength)];
				int valueDataLength;

				this._values = new SortedList<string, object>(valueCount);

				for (int index = 0; index < valueCount; index++)
				{
					valueNameLength = valueName.Capacity;
					valueDataLength = valueData.Length;

					RegistryValueKind valueType;

					result =
						this._session.RapiSession.CeRegEnumValue(
							this._key.Handle,
							index,
							valueName,
							ref valueNameLength,
							IntPtr.Zero,
							out valueType,
							valueData,
							ref valueDataLength);

					NativeResult.ThrowOnWin32Error(result);

					valueName.Length = valueNameLength;

					object valueObject;

					UnpackValue(
						out valueObject,
						valueData,
						valueDataLength,
						valueType);

					this._values.Add(
						valueName.ToString(),
						valueObject);
				}
			}
		}

		public object GetValue(string name)
		{
			RegistryValueKind valueType;
			byte[] valueData;
			int valueDataLength = 0;

			SystemResult result =
				this._session.RapiSession.CeRegQueryValueEx(
					this._key.Handle,
					name,
					IntPtr.Zero,
					out valueType,
					null,
					ref valueDataLength);

			if (
				(result == SystemResult.ERROR_MORE_DATA) ||
				(result == SystemResult.ERROR_SUCCESS))
			{
				valueData = new byte[valueDataLength];

				result =
					this._session.RapiSession.CeRegQueryValueEx(
						this._key.Handle,
						name,
						IntPtr.Zero,
						out valueType,
						valueData,
						ref valueDataLength);

				NativeResult.ThrowOnWin32Error(result);

				object value;

				UnpackValue(out value, valueData, valueDataLength, valueType);

				return value;
			}
			else
			{
				return null;
			}
		}

		public void SetValue(string name, object value)
		{
			RegistryValueKind valueType;
			byte[] valueData;

			PackValue(out valueData, out valueType, value);

			SystemResult result =
				this._session.RapiSession.CeRegSetValueEx(
					this._key.Handle,
					name,
					IntPtr.Zero,
					valueType,
					valueData,
					valueData.Length);

			NativeResult.ThrowOnWin32Error(result);
		}

		#endregion

		#region Constructors

		internal MobileRegistryValueCollection(
			MobileSession session,
			MobileRegistryKey key)
		{
			this._session = session;
			this._key = key;
			this._session.RegisterSessionObject(this);
		}

		#endregion

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			this._session.UnregisterSessionObject(this);
			base.Dispose(disposing);
		}

		#endregion

		#region IDictionary<string, object>

		#region Methods

		public void Add(string key, object value)
		{
			base.CheckDisposed();

			byte[] valueData;
			RegistryValueKind valueType;

			PackValue(out valueData, out valueType, value);

			SystemResult result =
				this._session.RapiSession.CeRegSetValueEx(
					this._key.Handle,
					key,
					IntPtr.Zero,
					valueType,
					valueData,
					valueData.Length);

			NativeResult.ThrowOnWin32Error(result);

			if (this._values != null)
			{
				this._values.Add(key, value);
			}
		}

		public bool Remove(string key)
		{
			base.CheckDisposed();

			SystemResult result =
				this._session.RapiSession.CeRegDeleteValue(
					this._key.Handle,
					key);

			NativeResult.ThrowOnWin32Error(result);

			if (this._values != null)
			{
				this._values.Remove(key);
			}

			return true;
		}

		public bool ContainsKey(string key)
		{
			base.CheckDisposed();

			BuildValueCollection();

			return this._values.ContainsKey(key);
		}

		public bool TryGetValue(string key, out object value)
		{
			base.CheckDisposed();

			value = GetValue(key);

			return value != null;
		}

		#endregion

		#region Properties

		public object this[string key]
		{
			get
			{
				base.CheckDisposed();

				return GetValue(key);
			}

			set
			{
				base.CheckDisposed();

				SetValue(key, value);
			}
		}

		public ICollection<string> Keys
		{
			get
			{
				base.CheckDisposed();

				BuildValueCollection();

				return this._values.Keys;
			}
		}

		public ICollection<object> Values
		{
			get
			{
				base.CheckDisposed();

				BuildValueCollection();

				return this._values.Values;
			}
		}

		#endregion

		#endregion

		#region ICollection<KeyValuePair<string, object>>

		#region Methods

		public void Add(KeyValuePair<string, object> item)
		{
			base.CheckDisposed();

			Add(item.Key, item.Value);
		}

		public bool Remove(KeyValuePair<string, object> item)
		{
			base.CheckDisposed();

			return Remove(item.Key);
		}

		public void Clear()
		{
			base.CheckDisposed();

			throw new NotImplementedException();
		}

		public bool Contains(KeyValuePair<string, object> item)
		{
			base.CheckDisposed();

			return ContainsKey(item.Key);
		}

		public void CopyTo(KeyValuePair<string, object>[] array, int arrayIndex)
		{
			base.CheckDisposed();

			throw new NotImplementedException();
		}

		#endregion

		#region Properties

		public int Count
		{
			get
			{
				base.CheckDisposed();

				BuildValueCollection();

				return this._values.Count;
			}
		}

		public bool IsReadOnly
		{
			get
			{
				base.CheckDisposed();

				return false;
			}
		}

		#endregion

		#endregion

		#region IEnumerable<KeyValuePair<string,object>>

		public IEnumerator<KeyValuePair<string, object>> GetEnumerator()
		{
			base.CheckDisposed();

			BuildValueCollection();

			return this._values.GetEnumerator();
		}

		#endregion

		#region IEnumerable

		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
		{
			return ((IEnumerable<KeyValuePair<string, object>>)this).GetEnumerator();
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
