﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Mobile Library                           */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;

using Nabu.Collections;

namespace Nabu.Mobile
{
	public sealed class MobileSession : DisposableBase
	{
		#region MobileSession

		#region Fields

		private IRapiSession _rapiSession;
		private DisposableList<IDisposable> _sessionObjects;

		#endregion

		#region Methods

		#region Utility

		internal void RegisterSessionObject(IDisposable sessionObject)
		{
			this._sessionObjects.Add(sessionObject);
		}

		internal void UnregisterSessionObject(IDisposable sessionObject)
		{
			this._sessionObjects.Remove(sessionObject);
		}

		#endregion

		#region Constructors

		internal MobileSession(IRapiSession rapiSession)
		{
			this._rapiSession = rapiSession;
			this._rapiSession.CeRapiInit();
			this._sessionObjects = new DisposableList<IDisposable>();
		}

		#endregion

		#region System

		public int GetSystemMetrics(MobileSystemMetrics index)
		{
			return this._rapiSession.CeGetSystemMetrics(index);
		}

		public int GetDesktopDeviceCaps(MobileDeviceCaps index)
		{
			return this._rapiSession.CeGetDesktopDeviceCaps(index);
		}

		public void CreateProcess(
			string path,
			string commandLine)
		{
			ProcessInformation processInformation;

			if (!this._rapiSession.CeCreateProcess(
				path,
				commandLine,
				IntPtr.Zero,
				IntPtr.Zero,
				false,
				0,
				IntPtr.Zero,
				IntPtr.Zero,
				IntPtr.Zero,
				out processInformation))
			{
				throw new MobileException(this, "CeCreateProcess", path);
			}
		}

		public bool CheckPassword(string password)
		{
			return this._rapiSession.CeCheckPassword(password);
		}

		#endregion

		#region Files

		public IEnumerable<MobileFileFindInfo> GetFiles(string mask)
		{
			base.CheckDisposed();

			return new MobileFileEnumerator(this, mask);
		}

		public FileAttributes GetFileAttributes(string path)
		{
			base.CheckDisposed();

			return this._rapiSession.CeGetFileAttributes(path);
		}

		public void SetFileAttributes(
			string path,
			FileAttributes attributes)
		{
			if (!this._rapiSession.CeSetFileAttributes(
				path,
				attributes))
			{
				throw new MobileException(this, "CeSetFileAttributes", path);
			}
		}

		public bool FileExists(string path)
		{
			base.CheckDisposed();

			return (int)this._rapiSession.CeGetFileAttributes(path) != -1;
		}

		public void MoveFile(
			string sourcePath,
			string destinationPath)
		{
			if (!this._rapiSession.CeMoveFile(
				sourcePath,
				destinationPath))
			{
				throw new MobileException(this, "CeMoveFile", sourcePath, destinationPath);
			}
		}

		public void CopyFile(
			string sourcePath,
			string destinationPath,
			bool failIfExists)
		{
			if (!this._rapiSession.CeCopyFile(
				sourcePath,
				destinationPath,
				failIfExists))
			{
				throw new MobileException(this, "CeCopyFile", sourcePath, destinationPath);
			}
		}

		public void DeleteFile(string path)
		{
			if (!this._rapiSession.CeDeleteFile(path))
			{
				throw new MobileException(this, "CeDeleteFile", path);
			}
		}

		public MobileFileStream CreateFileStream(
			string path,
			FileMode mode,
			FileAccess access,
			FileShare share)
		{
			base.CheckDisposed();

			return
				new MobileFileStream(
					this,
					path,
					mode,
					access,
					share);
		}

		public void CreateDirectory(string path)
		{
			if (!FileExists(path))
			{
				if (!this._rapiSession.CeCreateDirectory(
					path,
					IntPtr.Zero))
				{
					throw new MobileException(this, "CeCreateDirectory", path);
				}
			}
		}

		public void DeleteDirectory(string path)
		{
			if (!this._rapiSession.CeRemoveDirectory(path))
			{
				throw new MobileException(this, "CeRemoveDirectory", path);
			}
		}

		public void CreateShortcut(
			string shortcutPath,
			string targetPath)
		{
			if (!this._rapiSession.CeSHCreateShortcut(
				shortcutPath,
				targetPath))
			{
				throw new MobileException(this, "CeSHCreateShortcut", shortcutPath, targetPath);
			}
		}

		public string GetShortcutTargetPath(
			string shortcutPath)
		{
			StringBuilder targetPath = new StringBuilder(260);

			if (!this._rapiSession.CeSHGetShortcutTarget(
				shortcutPath,
				targetPath,
				targetPath.Capacity))
			{
				throw new MobileException(this, "CeSHGetShortcutTarget", shortcutPath);
			}

			return targetPath.ToString();
		}

		public string GetSpecialFolderPath(MobileSpecialFolder folder)
		{
			StringBuilder folderPath = new StringBuilder(260);
			int folderPathLength;

			switch (folder)
			{
				case MobileSpecialFolder.Temp:
					folderPathLength =
						this._rapiSession.CeGetTempPath(
							folderPath.Capacity,
							folderPath);

					if (folderPathLength == 0)
					{
						throw new MobileException(this, "CeGetTempPath");
					}
					break;
				default:
					folderPathLength =
						this._rapiSession.CeGetSpecialFolderPath(
						folder,
						folderPath.Capacity,
						folderPath);

					if (folderPathLength == 0)
					{
						throw new MobileException(this, "CeGetTempPath");
					}
					break;
			}

			folderPath.Length = folderPathLength;

			return folderPath.ToString();
		}

		#endregion

		#region Registry

		public MobileRegistryKey OpenRegistryKey(
			MobileRegistryKeys root,
			string path,
			MobileRegistryRights rights)
		{
			if (path == null)
			{
				return
					new MobileRegistryKey(
						this,
						null,
						new IntPtr((int)root),
						rights,
						root.ToString(),
						null);
			}
			else
			{
				IntPtr hKey;
				SystemResult result =
					this._rapiSession.CeRegOpenKeyEx(
						new IntPtr((int)root),
						path,
						0,
						rights,
						out hKey);

				NativeResult.ThrowOnWin32Error(result);

				int slashIndex = path.LastIndexOf('\\');

				return
					new MobileRegistryKey(
						this,
						null,
						hKey,
						rights,
						(slashIndex == -1) ? path : path.Substring(slashIndex),
						null);
			}
		}

		public MobileRegistryKey OpenRegistryKey(
			MobileRegistryKeys root,
			string path)
		{
			return OpenRegistryKey(root, path, MobileRegistryRights.All);
		}

		public MobileRegistryKey OpenRegistryKey(
			MobileRegistryKeys root,
			MobileRegistryRights rights)
		{
			return OpenRegistryKey(root, null, rights);
		}

		public MobileRegistryKey OpenRegistryKey(
			MobileRegistryKeys root)
		{
			return OpenRegistryKey(root, null, MobileRegistryRights.All);
		}

		#endregion

		#region RAW

		/*

		#region Registry

		[PreserveSig]
		ComResult CeRegEnumValue(
			[In] IntPtr hKey,
			[In] int index,
			[In][Out][MarshalAs(UnmanagedType.LPWStr)] StringBuilder name,
			[In][Out] ref int nameLength,
			[In] IntPtr reserved,
			[Out][MarshalAs(UnmanagedType.U4)] out RegistryValueKind type,
			[In] IntPtr lpDataBuffer,
			[In][Out] ref int dataSize);

		[PreserveSig]
		ComResult CeRegDeleteValue(
			[In] IntPtr hKey,
			[In][MarshalAs(UnmanagedType.BStr)] string name);

		[PreserveSig]
		ComResult CeRegQueryInfoKey(
			[In] IntPtr hKey,
			[In][Out][MarshalAs(UnmanagedType.LPWStr)] StringBuilder class_,
			[In][Out] ref int classLength,
			[In] IntPtr reserved,
			[Out] out int subKeyCount,
			[Out] out int subKeyMaxNameLength,
			[Out] out int subKeyMaxClassLength,
			[Out] out int valueCount,
			[Out] out int valueMaxNameLength,
			[Out] out int valueMaxDataLength,
			[In] IntPtr securityDescriptor,
			[In] IntPtr lastWriteTime);

		[PreserveSig]
		ComResult CeRegQueryValueEx(
			[In] IntPtr hKey,
			[In][MarshalAs(UnmanagedType.BStr)] string name,
			[In] IntPtr reserved,
			[Out][MarshalAs(UnmanagedType.U4)] out RegistryValueKind type,
			[In] IntPtr lpDataBuffer,
			[In][Out] ref int dataSize);

		[PreserveSig]
		ComResult CeRegSetValueEx(
		[In] IntPtr hKey,
			[In][MarshalAs(UnmanagedType.BStr)] string name,
			[In] IntPtr reserved,
			[In][MarshalAs(UnmanagedType.U4)] RegistryValueKind type,
			[In] IntPtr lpDataBuffer,
			[In] int dataSize);

		#endregion
*/
		#endregion

		#endregion

		#region Properties

		#region Utility

		internal IRapiSession RapiSession
		{
			get
			{
				base.CheckDisposed();

				return this._rapiSession;
			}
		}

		#endregion

		#region System

		public ComResult LastHostError
		{
			get
			{
				base.CheckDisposed();

				return this._rapiSession.CeGetLastError();
			}
		}

		public ComResult LastDeviceError
		{
			get
			{
				base.CheckDisposed();

				return this._rapiSession.CeRapiGetError();
			}
		}

		public MobileSystemInfo SystemInfo
		{
			get
			{
				base.CheckDisposed();

				SystemInfo systemInfo;

				this._rapiSession.CeGetSystemInfo(out systemInfo);

				return
					new MobileSystemInfo(
						systemInfo.ProcessorArchitecture,
						systemInfo.ProcessorType,
						systemInfo.ProcessorLevel,
						systemInfo.ProcessorRevision,
						systemInfo.ProcessorNumber,
						systemInfo.ActiveProcessorMask,
						systemInfo.MinimumApplicationAddress,
						systemInfo.MaximumApplicationAddress,
						systemInfo.PageSize,
						systemInfo.AllocationGranularity);
			}
		}

		public OperatingSystem Version
		{
			[SuppressMessage("Microsoft.Design", "CA1065:DoNotRaiseExceptionsInUnexpectedLocations")]
			get
			{
				base.CheckDisposed();

				OSVersionInfo version = new OSVersionInfo();

				version.Size = Marshal.SizeOf(typeof(OSVersionInfo));

				if (!this._rapiSession.CeGetVersionEx(
					ref version))
				{
					throw new MobileException(this, "CeGetVersionEx");
				}

				return
					new OperatingSystem(
						(PlatformID)version.PlatformID,
						new Version(
							version.MajorVersion,
							version.MinorVersion,
							version.BuildNumber));
			}
		}

		#endregion

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			this._sessionObjects.Dispose();

			if (this._rapiSession != null)
			{
				this._rapiSession.CeRapiUninit();
			}

			base.Dispose(disposing);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
