﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Net Library                              */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Net.Security;
using System.Net.Sockets;
using System.Security.Cryptography.X509Certificates;

using Nabu.Threading;

namespace Nabu.Net
{
	[SuppressMessage("Microsoft.Design", "CA1005:AvoidExcessiveParametersOnGenericTypes")]
	public abstract class TcpNodeBase<TSocket, TSession, TCallback> : DisposableBase
		where TSocket : ITcpNodeSocket
		where TSession : NetworkSessionBase<TSocket>
		where TCallback : ITcpNodeCallback<TSession>
	{
		#region TcpNodeBase

		#region Types

		protected sealed class TcpSecurityCallback : ITcpSecurityCallback
		{
			#region TcpSecurityCallback

			#region Fields

			private TSession _session;
			private TCallback _callback;

			#endregion

			#region Methods

			public TcpSecurityCallback(
				TSession session,
				TCallback callback)
			{
				this._session = session;
				this._callback = callback;
			}

			#endregion

			#endregion

			#region ITcpSecurityCallback

			public bool OnValidateRemoteCertificate(
				X509Certificate certificate,
				X509Chain chain,
				SslPolicyErrors sslPolicyErrors)
			{
				return
					this._callback.OnSslValidateCertificate(
						this._session,
						certificate,
						chain,
						sslPolicyErrors);
			}

			public X509Certificate OnSelectLocalCertificate(
				string targetHost,
				X509CertificateCollection localCertificates,
				X509Certificate remoteCertificate,
				string[] acceptableIssuers)
			{
				return
					this._callback.OnSslSelectCertificate(
						this._session,
						targetHost,
						localCertificates,
						remoteCertificate,
						acceptableIssuers);
			}

			#endregion
		}

		#endregion

		#region Fields

		private TCallback _callback;
		private ReverseSemaphore _semaphore;

		#endregion

		#region Methods

		private void ProcessIO(TSession session)
		{
			base.CheckDisposed();

			if ((session.Socket == null) || (!session.Socket.IsConnected))
			{
				throw new SocketException((int)SystemResult.ERROR_CONNECTION_INVALID);
			}

			if (session.IsClosing)
			{
				throw new SocketException((int)SystemResult.ERROR_CONNECTION_ABORTED);
			}
			else
			{
				if (session.SendSize > 0)
				{
					Send(session);
				}
				else
				{
					Receive(session);
				}
			}
		}

		[SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
		private void OnReceive(IAsyncResult asyncResult)
		{
			base.CheckDisposed();

			TSession session = (TSession)asyncResult.AsyncState;

			try
			{
				session.IsReceiving = false;

				int received = session.Socket.EndReceive(asyncResult);

				if (received == 0)
				{
					throw new SocketException((int)SystemResult.ERROR_CONNECTION_INVALID);
				}
				else
				{
					session.ReceiveBuffer.Count += received;
					/** /
					Console.ForegroundColor = ConsoleColor.Green;
					Console.WriteLine(
						System.Text.Encoding.ASCII.GetString(
							session.ReceiveBuffer.DangerousGetArray(),
							session.ReceiveOffset,
							session.ReceiveSize).Replace("\r\n", "\\r\\n\r\n"));
					Console.ForegroundColor = ConsoleColor.Gray;
					/**/
					int receiveSize;

					for (; ; )
					{
						receiveSize = session.ReceiveSize;
						OnReceive(session);
						session.ShrinkReceiveBuffer();

						if ((session.ReceiveSize == receiveSize) || (session.ReceiveSize == 0))
						{
							break;
						}
					}

					ProcessIO(session);
				}
			}
			catch (Exception exception)
			{
				OnException(session, exception);
			}
		}

		[SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
		private void OnSend(IAsyncResult asyncResult)
		{
			base.CheckDisposed();

			TSession session = (TSession)asyncResult.AsyncState;

			try
			{
				session.Socket.EndSend(asyncResult);
				session.IsSending = false;
				/** /
				Console.ForegroundColor = ConsoleColor.Blue;
				Console.WriteLine(
					System.Text.Encoding.ASCII.GetString(
						session.SendBuffer.DangerousGetArray(),
						session.SendOffset,
						session.SendSize).Replace("\r\n", "\\r\\n\r\n"));
				Console.ForegroundColor = ConsoleColor.Gray;
				/**/
				session.SendOffset += session.SendSize;

				int sendSize;

				for (; ; )
				{
					sendSize = session.SendSize;
					session.ShrinkSendBuffer();
					OnSend(session);

					if ((session.SendSize == sendSize) || (session.SendSize == 0))
					{
						break;
					}
				}

				ProcessIO(session);
			}
			catch (Exception exception)
			{
				OnException(session, exception);
			}
		}

		private void Receive(TSession session)
		{
			base.CheckDisposed();

			if (!session.IsReceiving)
			{
				lock (session)
				{
					if (!session.IsReceiving)
					{
						session.IsReceiving = true;

						session.Socket.BeginReceive(
							session.ReceiveBuffer.DangerousGetArray(),
							session.ReceiveOffset,
							session.ReceiveBuffer.Capacity - session.ReceiveOffset,
							OnReceive,
							session);
					}
				}
			}
		}

		private void Send(TSession session)
		{
			base.CheckDisposed();

			if (!session.IsSending)
			{
				lock (session)
				{
					if (!session.IsSending)
					{
						session.IsSending = true;
						session.Socket.BeginSend(
							session.SendBuffer.DangerousGetArray(),
							session.SendOffset,
							session.SendSize,
							OnSend,
							session);
					}
				}
			}
		}

		private void CloseSession(TSession session)
		{
			if (session != null)
			{
				base.CheckDisposed();

				this._semaphore.Decrement();

				session.Socket.Dispose();
				session.Socket = default(TSocket);
				session = null;

				this._callback.OnCloseSession(session);
			}
		}

		protected abstract TSession CreateSession(object customData);
		protected abstract void OnReceive(TSession session);
		protected abstract void OnSend(TSession session);

		protected void OnException(
			TSession session,
			Exception exception)
		{
			session.RaiseException(exception);
			CloseSession(session);
		}

		protected void RunSession(TSession session)
		{
			this._semaphore.Increment();
			ProcessIO(session);
		}

		protected TcpNodeBase(TCallback callback)
		{
			if (callback == null)
			{
				throw new ArgumentNullException("callback");
			}

			this._callback = callback;
			this._semaphore = new ReverseSemaphore();
		}

		public void WaitForSessions()
		{
			this._semaphore.WaitOne();
		}

		public void WaitForSessions(int timeout, bool exitContext)
		{
			this._semaphore.WaitOne(timeout, exitContext);
		}

		#endregion

		#region Properties

		protected TCallback Callback
		{
			[DebuggerStepThrough]
			get
			{
				base.CheckDisposed();

				return this._callback;
			}
		}

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				this._semaphore.Dispose();
			}

			base.Dispose(disposing);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
