﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Net Library                              */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Security.Authentication;

namespace Nabu.Net
{
	public abstract class TcpNodeSocketBase :
		DisposableBase,
		ITcpNodeSocket
	{
		#region TcpNodeSocketBase

		#region Fields

		private Socket _socket;
		private NetworkStream _rawStream;
		private TcpNodeStream _sslStream;
		private Stream _currentStream;

		#endregion

		#region Methods

		[DebuggerStepThrough]
		protected TcpNodeSocketBase()
		{
			this._socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
		}

		[DebuggerStepThrough]
		protected TcpNodeSocketBase(Socket socket)
		{
			this._socket = socket;
		}

		#endregion

		#region Properties

		protected Socket Socket
		{
			[DebuggerStepThrough]
			get { return this._socket; }
		}

		protected NetworkStream RawStream
		{
			[DebuggerStepThrough]
			get { return this._rawStream; }
			[DebuggerStepThrough]
			set { this._rawStream = value; }
		}

		protected TcpNodeStream SslStream
		{
			[DebuggerStepThrough]
			get { return this._sslStream; }
			[DebuggerStepThrough]
			set { this._sslStream = value; }
		}

		protected Stream CurrentStream
		{
			[DebuggerStepThrough]
			get { return this._currentStream; }
			[DebuggerStepThrough]
			set { this._currentStream = value; }
		}

		#endregion

		#endregion

		#region ITcpNodeSocket

		#region Methods

		[DebuggerStepThrough]
		public IAsyncResult BeginSend(
			byte[] buffer,
			int offset,
			int size,
			AsyncCallback callback,
			object state)
		{
			base.CheckDisposed();

			return
				this._currentStream.BeginWrite(
					buffer,
					offset,
					size,
					callback,
					state);
		}

		[DebuggerStepThrough]
		public void EndSend(IAsyncResult asyncResult)
		{
			base.CheckDisposed();

			this._currentStream.EndWrite(asyncResult);
		}

		[DebuggerStepThrough]
		public IAsyncResult BeginReceive(
			byte[] buffer,
			int offset,
			int size,
			AsyncCallback callback,
			object state)
		{
			base.CheckDisposed();

			return
				this._currentStream.BeginRead(
					buffer,
					offset,
					size,
					callback,
					state);
		}

		[DebuggerStepThrough]
		public int EndReceive(IAsyncResult asyncResult)
		{
			base.CheckDisposed();

			return this._currentStream.EndRead(asyncResult);
		}

		#endregion

		#region Properties

		public bool IsConnected
		{
			[DebuggerStepThrough]
			get
			{
				base.CheckDisposed();

				return this._socket.Connected;
			}
		}

		public bool IsEncrypted
		{
			[DebuggerStepThrough]
			get
			{
				base.CheckDisposed();

				return (this._sslStream == null) ? false : this._sslStream.IsEncrypted;
			}
		}

		public bool IsSigned
		{
			[DebuggerStepThrough]
			get
			{
				base.CheckDisposed();

				return (this._sslStream == null) ? false : this._sslStream.IsSigned;
			}
		}

		public bool IsAuthenticated
		{
			[DebuggerStepThrough]
			get
			{
				base.CheckDisposed();

				return (this._sslStream == null) ? false : this._sslStream.IsAuthenticated;
			}
		}

		public bool IsMutuallyAuthenticated
		{
			[DebuggerStepThrough]
			get
			{
				base.CheckDisposed();

				return (this._sslStream == null) ? false : this._sslStream.IsMutuallyAuthenticated;
			}
		}

		public SslProtocols SslProtocol
		{
			[DebuggerStepThrough]
			get
			{
				base.CheckDisposed();

				return (this._sslStream == null) ? SslProtocols.None : this._sslStream.SslProtocol;
			}
		}

		public EndPoint LocalEndPoint
		{
			[DebuggerStepThrough]
			get
			{
				base.CheckDisposed();

				return this._socket.LocalEndPoint;
			}
		}

		public EndPoint RemoteEndPoint
		{
			[DebuggerStepThrough]
			get
			{
				base.CheckDisposed();

				return this._socket.RemoteEndPoint;
			}
		}

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				if (this._rawStream != null)
				{
					this._rawStream.Close();
					this._rawStream = null;
				}

				if (this._sslStream != null)
				{
					this._sslStream.Close();
					this._sslStream = null;
				}

				if (this._socket != null)
				{
					this._socket.Close();
					this._socket = null;
				}
			}

			base.Dispose(disposing);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
