﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Net Library                              */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics;
using System.IO;
using System.Net.Security;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;

namespace Nabu.Net
{
	public sealed class TcpNodeStream : Stream
	{
		#region TcpNodeStream

		#region Fields

		private ITcpSecurityCallback _securityCallback;
		private SslStream _stream;

		#endregion

		#region Methods

		[DebuggerStepThrough]
		private bool OnValidateRemoteCertificate(
			object sender,
			X509Certificate certificate,
			X509Chain chain,
			SslPolicyErrors sslPolicyErrors)
		{
			return
				this._securityCallback.OnValidateRemoteCertificate(
					certificate,
					chain,
					sslPolicyErrors);
		}

		[DebuggerStepThrough]
		private X509Certificate OnSelectLocalCertificate(
			object sender,
			string targetHost,
			X509CertificateCollection localCertificates,
			X509Certificate remoteCertificate,
			string[] acceptableIssuers)
		{
			return
				this._securityCallback.OnSelectLocalCertificate(
					targetHost,
					localCertificates,
					remoteCertificate,
					acceptableIssuers);
		}

		[DebuggerStepThrough]
		public TcpNodeStream(
			ITcpSecurityCallback securityCallback,
			Stream stream)
		{
			this._securityCallback = securityCallback;
			this._stream = new SslStream(
				stream,
				true,
				new RemoteCertificateValidationCallback(OnValidateRemoteCertificate),
				new LocalCertificateSelectionCallback(OnSelectLocalCertificate));
		}

		#endregion

		#endregion

		#region SslStream

		#region Methods

		[DebuggerStepThrough]
		public void AuthenticateAsClient(
			string targetHost,
			X509CertificateCollection clientCertificates,
			SslProtocols enabledSslProtocols,
			bool checkCertificateRevocation)
		{
			this._stream.AuthenticateAsClient(
				targetHost,
				clientCertificates,
				enabledSslProtocols,
				checkCertificateRevocation);
		}

		[DebuggerStepThrough]
		public void AuthenticateAsServer(
			X509Certificate serverCertificate,
			bool clientCertificateRequired,
			SslProtocols enabledSslProtocols,
			bool checkCertificateRevocation)
		{
			this._stream.AuthenticateAsServer(
				serverCertificate,
				clientCertificateRequired,
				enabledSslProtocols,
				checkCertificateRevocation);
		}

		#endregion

		#region Properties

		public bool IsEncrypted
		{
			[DebuggerStepThrough]
			get { return this._stream.IsEncrypted; }
		}

		public bool IsSigned
		{
			[DebuggerStepThrough]
			get { return this._stream.IsSigned; }
		}

		public bool IsAuthenticated
		{
			[DebuggerStepThrough]
			get { return this._stream.IsAuthenticated; }
		}

		public bool IsMutuallyAuthenticated
		{
			[DebuggerStepThrough]
			get { return this._stream.IsMutuallyAuthenticated; }
		}

		public SslProtocols SslProtocol
		{
			[DebuggerStepThrough]
			get { return this._stream.SslProtocol; }
		}

		#endregion

		#endregion

		#region Stream

		#region Methods

		[DebuggerStepThrough]
		public override IAsyncResult BeginRead(byte[] buffer, int offset, int count, AsyncCallback callback, object state)
		{
			return this._stream.BeginRead(buffer, offset, count, callback, state);
		}

		[DebuggerStepThrough]
		public override int EndRead(IAsyncResult asyncResult)
		{
			return this._stream.EndRead(asyncResult);
		}

		[DebuggerStepThrough]
		public override IAsyncResult BeginWrite(byte[] buffer, int offset, int count, AsyncCallback callback, object state)
		{
			return this._stream.BeginWrite(buffer, offset, count, callback, state);
		}

		[DebuggerStepThrough]
		public override void EndWrite(IAsyncResult asyncResult)
		{
			this._stream.EndWrite(asyncResult);
		}

		[DebuggerStepThrough]
		public override int Read(byte[] buffer, int offset, int count)
		{
			return this._stream.Read(buffer, offset, count);
		}

		[DebuggerStepThrough]
		public override int ReadByte()
		{
			return this._stream.ReadByte();
		}

		[DebuggerStepThrough]
		public override void Write(byte[] buffer, int offset, int count)
		{
			this._stream.Read(buffer, offset, count);
		}

		[DebuggerStepThrough]
		public override void WriteByte(byte value)
		{
			this._stream.WriteByte(value);
		}

		[DebuggerStepThrough]
		public override long Seek(long offset, SeekOrigin origin)
		{
			return this._stream.Seek(offset, origin);
		}

		[DebuggerStepThrough]
		public override void SetLength(long value)
		{
			this._stream.SetLength(value);
		}

		[DebuggerStepThrough]
		public override void Flush()
		{
			this._stream.Flush();
		}

		[DebuggerStepThrough]
		public override void Close()
		{
			this._stream.Close();
		}

		#endregion

		#region Properties

		public override long Length
		{
			[DebuggerStepThrough]
			get { return this._stream.Length; }
		}

		public override long Position
		{
			[DebuggerStepThrough]
			get { return this._stream.Position; }
			[DebuggerStepThrough]
			set { this._stream.Position = value; }
		}

		public override bool CanRead
		{
			[DebuggerStepThrough]
			get { return this._stream.CanRead; }
		}

		public override bool CanWrite
		{
			[DebuggerStepThrough]
			get { return this._stream.CanWrite; }
		}

		public override bool CanSeek
		{
			[DebuggerStepThrough]
			get { return this._stream.CanSeek; }
		}

		public override bool CanTimeout
		{
			[DebuggerStepThrough]
			get { return this._stream.CanTimeout; }
		}

		public override int ReadTimeout
		{
			[DebuggerStepThrough]
			get { return this._stream.ReadTimeout; }
		}

		public override int WriteTimeout
		{
			[DebuggerStepThrough]
			get { return this._stream.WriteTimeout; }
		}

		#endregion

		#endregion

		#region IDisposable

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				this._stream.Dispose();
				this._stream = null;
			}
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
