﻿/******************************************************************************/
using System;
using System.Diagnostics.CodeAnalysis;
using System.Net;
using System.Net.Sockets;

namespace Nabu.Net
{
	[SuppressMessage("Microsoft.Design", "CA1005:AvoidExcessiveParametersOnGenericTypes")]
	public abstract class UdpClientBase<TSocket, TSession, TCallback> :
		UdpNodeBase<TSocket, TSession, TCallback>
		where TSocket : IUdpClientSocket, new()
		where TSession : NetworkSessionBase<TSocket>
		where TCallback : IUdpNodeCallback<TSession>
	{
		#region Fields

		private IPEndPoint _endPoint;

		#endregion

		#region Methods

		private void ProcessIO(TSession session)
		{
			base.CheckDisposed();

			if (session.Socket == null)
			{
				throw new SocketException((int)SystemResult.ERROR_CONNECTION_INVALID);
			}

			if (session.IsClosing)
			{
				throw new SocketException((int)SystemResult.ERROR_CONNECTION_ABORTED);
			}
			else
			{
				Send(session, this._endPoint);
			}
		}

		private void Send(TSession session, IPEndPoint endPoint)
		{
			base.CheckDisposed();

			if (!session.IsSending)
			{
				lock (session)
				{
					if (!session.IsSending)
					{
						session.IsSending = true;

						session.Socket.BeginSendTo(
							endPoint,
							session.SendBuffer.DangerousGetArray(),
							session.SendOffset,
							session.SendSize,
							OnSend,
							session);
					}
				}
			}
		}

		[SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
		private void OnSend(IAsyncResult result)
		{
			base.CheckDisposed();

			TSession session = (TSession)result.AsyncState;

			try
			{
				session.SendOffset +=
					session.Socket.EndSendTo(result);
				session.IsSending = false;
				session.ShrinkSendBuffer();

				OnSend(session, out this._endPoint);
				ProcessIO(session);
			}
			catch (Exception exception)
			{
				OnException(session, exception);
			}
		}

		[SuppressMessage("Microsoft.Design", "CA1021:AvoidOutParameters", MessageId = "1#")]
		protected abstract void OnSend(TSession session, out IPEndPoint endPoint);

		protected UdpClientBase(TCallback callback)
			: base(callback)
		{
		}

		[SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
		public void OpenSession(object customData)
		{
			base.CheckDisposed();

			TSession session = CreateSession(customData);

			try
			{
				session.Socket = new TSocket();

				base.Callback.OnOpenSession(session);

				OnSend(session, out this._endPoint);
				ProcessIO(session);
			}
			catch (Exception exception)
			{
				OnException(session, exception);
			}
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
