﻿/******************************************************************************/
using System;
using System.Diagnostics.CodeAnalysis;
using System.Net;
using System.Net.Sockets;

namespace Nabu.Net
{
	[SuppressMessage("Microsoft.Design", "CA1005:AvoidExcessiveParametersOnGenericTypes")]
	public abstract class UdpServerBase<TSocket, TSession, TCallback> :
		UdpNodeBase<TSocket, TSession, TCallback>
		where TSocket : IUdpServerSocket, new()
		where TSession : NetworkSessionBase<TSocket>
		where TCallback : IUdpNodeCallback<TSession>
	{
		private void ProcessIO(TSession session)
		{
			base.CheckDisposed();

			if (session.Socket == null)
			{
				throw new SocketException((int)SystemResult.ERROR_CONNECTION_INVALID);
			}

			if (session.IsClosing)
			{
				throw new SocketException((int)SystemResult.ERROR_CONNECTION_ABORTED);
			}
			else
			{
				Receive(session);
			}
		}

		private void Receive(TSession session)
		{
			base.CheckDisposed();

			if (!session.IsReceiving)
			{
				lock (session)
				{
					if (!session.IsReceiving)
					{
						session.IsReceiving = true;

						session.Socket.BeginReceiveFrom(
							session.ReceiveBuffer.DangerousGetArray(),
							session.ReceiveOffset,
							session.ReceiveBuffer.Capacity - session.ReceiveOffset,
							OnReceive,
							session);
					}
				}
			}
		}

		[SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
		private void OnReceive(IAsyncResult result)
		{
			base.CheckDisposed();

			TSession session = (TSession)result.AsyncState;

			try
			{
				IPEndPoint endPoint;

				session.ReceiveBuffer.Count +=
					session.Socket.EndReceiveFrom(
						out endPoint,
						result);
				session.IsReceiving = false;

				if (session.ReceiveSize == 0)
				{
					throw new SocketException((int)SystemResult.ERROR_CONNECTION_INVALID);
				}

				OnReceive(session, endPoint);
				session.ShrinkReceiveBuffer();
				ProcessIO(session);
			}
			catch (Exception exception)
			{
				OnException(session, exception);
			}
		}

		protected abstract void OnReceive(TSession session, IPEndPoint endPoint);

		protected UdpServerBase(TCallback callback)
			: base(callback)
		{
		}

		[SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
		public void Start(
			string hostname,
			int port,
			object customData)
		{
			base.CheckDisposed();

			TSession session = CreateSession(customData);

			try
			{
				session.Socket = new TSocket();
				session.Socket.Bind(hostname, port);

				base.Callback.OnOpenSession(session);

				ProcessIO(session);
			}
			catch (Exception exception)
			{
				OnException(session, exception);
			}
		}
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
