﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Net Library                              */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Text.RegularExpressions;

namespace Nabu.Net
{
	public class Pop3ClientBase<TSocket> :
		TcpClientBase<TSocket, Pop3ClientSession<TSocket>, IPop3ClientCallback<TSocket>>
			where TSocket : ITcpClientSocket, new()
	{
		#region Pop3ClientBase<TSocket>

		#region Fields

		private Regex _regExKeyValue = new Regex(@"\s*(\S+)\s+(\S+)\s*", RegexOptions.Compiled);

		#endregion

		#region Methods

		private void RaiseError(
			Pop3ClientSession<TSocket> session,
			string text)
		{
			base.Callback.OnError(
				session,
				session.States.First.Value,
				text);
		}

		private void GetRequestAndSendCommand(Pop3ClientSession<TSocket> session)
		{
			Pop3ClientRequest request;

			base.Callback.OnRequest(
				session,
				out request);

			switch (request.Type)
			{
				case Pop3ClientRequestType.Stat:
					TextProtocol.Write<TSocket>(session, "STAT\r\n");
					session.States.AddLast(Pop3ClientState.Stat);
					break;
				case Pop3ClientRequestType.List:
					if (request.MessageIndex == 0)
					{
						TextProtocol.Write<TSocket>(session, "LIST\r\n");
						session.States.AddLast(Pop3ClientState.List_All);
					}
					else
					{
						TextProtocol.Write<TSocket>(session, "LIST " + request.MessageIndex + "\r\n");
						session.States.AddLast(Pop3ClientState.List_One);
					}
					break;
				case Pop3ClientRequestType.Uidl:
					if (request.MessageIndex == 0)
					{
						TextProtocol.Write<TSocket>(session, "UIDL\r\n");
						session.States.AddLast(Pop3ClientState.Uidl_All);
					}
					else
					{
						TextProtocol.Write<TSocket>(session, "UIDL " + (request.MessageIndex + 1) + "\r\n");
						session.States.AddLast(Pop3ClientState.Uidl_One);
					}
					break;
				case Pop3ClientRequestType.Retrive:
					TextProtocol.Write<TSocket>(session, "RETR " + (request.MessageIndex + 1) + "\r\n");
					session.States.AddLast(Pop3ClientState.Retr);
					break;
				case Pop3ClientRequestType.Delete:
					TextProtocol.Write<TSocket>(session, "DELE " + (request.MessageIndex + 1) + "\r\n");
					session.States.AddLast(Pop3ClientState.Dele);
					break;
				case Pop3ClientRequestType.Reset:
					TextProtocol.Write<TSocket>(session, "RSET\r\n");
					session.States.AddLast(Pop3ClientState.Rset);
					break;
				case Pop3ClientRequestType.Top:
					TextProtocol.Write<TSocket>(session, "TOP " + (request.MessageIndex + 1) + " " + request.LineIndex + "\r\n");
					session.States.AddLast(Pop3ClientState.Top);
					break;
				case Pop3ClientRequestType.Quit:
					TextProtocol.Write<TSocket>(session, "QUIT\r\n");
					session.States.AddLast(Pop3ClientState.Quit);
					break;
			}
		}

		public Pop3ClientBase(IPop3ClientCallback<TSocket> callback)
			: base(callback)
		{
		}

		#endregion

		#endregion

		#region TcpClientBase<TSocket>

		protected override Pop3ClientSession<TSocket> CreateSession(object customData)
		{
			Pop3ClientSession<TSocket> session = new Pop3ClientSession<TSocket>(customData);

			session.States.AddLast(Pop3ClientState.None);

			string hostname;
			string username;
			string password;

			base.Callback.OnCredentials(
				session,
				out hostname,
				out username,
				out password);

			session.Hostname = hostname;
			session.Username = username;
			session.Password = password;

			return session;
		}

		protected override void OnSend(Pop3ClientSession<TSocket> session)
		{
			switch (session.States.Last.Value)
			{
				#region None

				case Pop3ClientState.None:
					break;

				#endregion

				#region STLS

				case Pop3ClientState.STLS:
					break;

				#endregion

				#region Capa

				case Pop3ClientState.Capa:
					break;

				#endregion

				#region User

				case Pop3ClientState.User:
					if (session.GetOptionSupported(Pop3ClientOptions.ModePipelining))
					{
						TextProtocol.Write<TSocket>(session, "PASS " + session.Password + "\r\n");
						session.States.AddLast(Pop3ClientState.Pass);
					}
					break;

				#endregion

				#region Pass

				case Pop3ClientState.Pass:
					break;

				#endregion

				#region Stat

				case Pop3ClientState.Stat:
					break;

				#endregion

				#region List_One

				case Pop3ClientState.List_One:
					break;

				#endregion

				#region List_All

				case Pop3ClientState.List_All:
					break;

				#endregion

				#region Uidl_One

				case Pop3ClientState.Uidl_One:
					break;

				#endregion

				#region Uidl_All

				case Pop3ClientState.Uidl_All:
					break;

				#endregion

				#region Retr

				case Pop3ClientState.Retr:
					break;

				#endregion

				#region Top

				case Pop3ClientState.Top:
					break;

				#endregion

				#region Dele

				case Pop3ClientState.Dele:
					break;

				#endregion

				#region Rset

				case Pop3ClientState.Rset:
					break;

				#endregion

				#region Quit

				case Pop3ClientState.Quit:
					break;

				#endregion
			}
		}

		protected override void OnReceive(Pop3ClientSession<TSocket> session)
		{
			switch (session.States.First.Value)
			{
				#region None

				case Pop3ClientState.None:
					session.SingleLineParser.Parse(session);

					if (session.SingleLineParser.IsValid)
					{
						if (session.SingleLineParser.IsComplete)
						{
							if (session.SingleLineParser.IsSuccessful)
							{
								TextProtocol.Write<TSocket>(session, "CAPA\r\n");
								session.States.AddLast(Pop3ClientState.Capa);
								session.States.RemoveFirst();
								session.SingleLineParser.Reset();
							}
							else
							{
								RaiseError(session, session.SingleLineParser.Line);
							}
						}
					}
					else
					{
						RaiseError(session, session.SingleLineParser.Line);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region STLS

				case Pop3ClientState.STLS:
					session.SingleLineParser.Parse(session);

					if (session.SingleLineParser.IsValid)
					{
						if (session.SingleLineParser.IsComplete)
						{
							if (session.SingleLineParser.IsSuccessful)
							{
								base.SslAuthenticate(session, session.Hostname);
								TextProtocol.Write<TSocket>(session, "CAPA\r\n");
								session.States.AddLast(Pop3ClientState.Capa);
								session.States.RemoveFirst();
								session.SingleLineParser.Reset();
							}
							else
							{
								RaiseError(session, session.SingleLineParser.Line);
							}
						}
					}
					else
					{
						RaiseError(session, session.SingleLineParser.Line);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Capa

				case Pop3ClientState.Capa:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							if (session.MultiLineParser.IsSuccessful)
							{
								for (int index = 1; index < session.MultiLineParser.Lines.Count; index++)
								{
									string line = session.MultiLineParser.Lines[index];

									if (string.Compare(line, "TOP", StringComparison.OrdinalIgnoreCase) == 0)
									{
										session.SetOptionSupported(Pop3ClientOptions.CommandTop);
									}
									else if (string.Compare(line, "UIDL", StringComparison.OrdinalIgnoreCase) == 0)
									{
										session.SetOptionSupported(Pop3ClientOptions.CommandUidl);
									}
									else if (string.Compare(line, "STLS", StringComparison.OrdinalIgnoreCase) == 0)
									{
										session.SetOptionSupported(Pop3ClientOptions.CommandSTLS);
									}
									else if (string.Compare(line, "USER", StringComparison.OrdinalIgnoreCase) == 0)
									{
										session.SetOptionSupported(Pop3ClientOptions.AuthenticationUser);
									}
									else if (string.Compare(line, "PIPELINING", StringComparison.OrdinalIgnoreCase) == 0)
									{
										session.SetOptionSupported(Pop3ClientOptions.ModePipelining);
									}
									else if (line.StartsWith("LOGIN-DELAY", StringComparison.OrdinalIgnoreCase))
									{
										session.SetOptionSupported(Pop3ClientOptions.OptionLoginDelay);
										session.OptionLoginDelay = TextProtocol.ParseInteger(line.Substring(12));
									}
								}

								if (session.GetOptionSupported(Pop3ClientOptions.CommandSTLS) && (!session.Socket.IsEncrypted))
								{
									TextProtocol.Write<TSocket>(session, "STLS\r\n");
									session.States.AddLast(Pop3ClientState.STLS);
									session.States.RemoveFirst();
								}
								else if (session.GetOptionSupported(Pop3ClientOptions.AuthenticationUser))
								{
									TextProtocol.Write<TSocket>(session, "USER " + session.Username + "\r\n");
									session.States.AddLast(Pop3ClientState.User);
									session.States.RemoveFirst();
								}
								else
								{
									RaiseError(
										session,
										"Unsupported authentication.");
									session.IsClosing = true;
								}

								session.MultiLineParser.Reset();
							}
							else
							{
								TextProtocol.Write<TSocket>(session, "USER " + session.Username + "\r\n");
								session.States.AddLast(Pop3ClientState.User);
								session.States.RemoveFirst();
								session.MultiLineParser.Reset();
							}
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region User

				case Pop3ClientState.User:
					session.SingleLineParser.Parse(session);

					if (session.SingleLineParser.IsValid)
					{
						if (session.SingleLineParser.IsComplete)
						{
							if (session.SingleLineParser.IsSuccessful)
							{
								if (!session.GetOptionSupported(Pop3ClientOptions.ModePipelining))
								{
									TextProtocol.Write<TSocket>(session, "PASS " + session.Password + "\r\n");
									session.States.AddLast(Pop3ClientState.Pass);
								}

								session.States.RemoveFirst();
								session.SingleLineParser.Reset();
							}
							else
							{
								RaiseError(session, session.SingleLineParser.Line);
							}
						}
					}
					else
					{
						RaiseError(session, session.SingleLineParser.Line);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Pass

				case Pop3ClientState.Pass:
					session.SingleLineParser.Parse(session);

					if (session.SingleLineParser.IsValid)
					{
						if (session.SingleLineParser.IsComplete)
						{
							if (session.SingleLineParser.IsSuccessful)
							{
								GetRequestAndSendCommand(session);
								session.States.RemoveFirst();
								session.SingleLineParser.Reset();
							}
							else
							{
								RaiseError(session, session.SingleLineParser.Line);
							}
						}
					}
					else
					{
						RaiseError(session, session.SingleLineParser.Line);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Stat

				case Pop3ClientState.Stat:
					session.SingleLineParser.Parse(session);

					if (session.SingleLineParser.IsValid)
					{
						if (session.SingleLineParser.IsComplete)
						{
							if (session.SingleLineParser.IsSuccessful)
							{
								Match match = _regExKeyValue.Match(session.SingleLineParser.Line);

								if (match.Success)
								{
									int messageNumber = TextProtocol.ParseInteger(match.Groups[1].Value);
									int messageSize = TextProtocol.ParseInteger(match.Groups[2].Value);

									base.Callback.OnStat(
										session,
										messageNumber,
										messageSize);

									GetRequestAndSendCommand(session);
									session.States.RemoveFirst();
									session.SingleLineParser.Reset();
								}
								else
								{
									RaiseError(session, session.SingleLineParser.Line);
									session.IsClosing = true;
								}
							}
							else
							{
								RaiseError(session, session.SingleLineParser.Line);
							}
						}
					}
					else
					{
						RaiseError(session, session.SingleLineParser.Line);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region List_One

				case Pop3ClientState.List_One:
					session.SingleLineParser.Parse(session);

					if (session.SingleLineParser.IsValid)
					{
						if (session.SingleLineParser.IsComplete)
						{
							if (session.SingleLineParser.IsSuccessful)
							{
								SortedList<int, int> messageSizes = new SortedList<int, int>();

								Match match = _regExKeyValue.Match(session.SingleLineParser.Line);

								if (match.Success)
								{
									int messageIndex = TextProtocol.ParseInteger(match.Groups[1].Value);
									int messageSize = TextProtocol.ParseInteger(match.Groups[2].Value);

									session.BlockParser.SetCapacity(messageSize);

									messageSizes.Add(messageIndex, messageSize);
								}
								else
								{
									RaiseError(session, session.SingleLineParser.Line);
									session.IsClosing = true;
								}

								base.Callback.OnList(
									session,
									messageSizes);

								GetRequestAndSendCommand(session);
								session.States.RemoveFirst();
								session.SingleLineParser.Reset();
							}
							else
							{
								RaiseError(session, session.SingleLineParser.Line);
							}
						}
					}
					else
					{
						RaiseError(session, session.SingleLineParser.Line);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region List_All

				case Pop3ClientState.List_All:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							if (session.MultiLineParser.IsSuccessful)
							{
								SortedList<int, int> messageSizes = new SortedList<int, int>();

								int maxMessageSize = 0;

								for (int index = 1; index < session.MultiLineParser.Lines.Count; index++)
								{
									Match match = _regExKeyValue.Match(session.MultiLineParser.Lines[index]);

									if (match.Success)
									{
										int messageIndex = TextProtocol.ParseInteger(match.Groups[1].Value);
										int messageSize = TextProtocol.ParseInteger(match.Groups[2].Value);

										messageSizes.Add(messageIndex, messageSize);

										maxMessageSize = Math.Max(maxMessageSize, messageSize);
									}
									else
									{
										RaiseError(session, session.MultiLineParser.Lines[index]);
										session.IsClosing = true;
									}
								}

								session.BlockParser.SetCapacity(maxMessageSize);

								base.Callback.OnList(
									session,
									messageSizes);

								GetRequestAndSendCommand(session);
								session.States.RemoveFirst();
								session.MultiLineParser.Reset();
							}
							else
							{
								RaiseError(session, session.MultiLineParser.FirstLine);
							}
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Uidl_One

				case Pop3ClientState.Uidl_One:
					session.SingleLineParser.Parse(session);

					if (session.SingleLineParser.IsValid)
					{
						if (session.SingleLineParser.IsComplete)
						{
							if (session.SingleLineParser.IsSuccessful)
							{
								SortedList<int, string> messageUIDs = new SortedList<int, string>();
								Match match = _regExKeyValue.Match(session.SingleLineParser.Line);

								if (match.Success)
								{
									int messageIndex = TextProtocol.ParseInteger(match.Groups[1].Value);
									string messageUID = match.Groups[2].Value;

									messageUIDs.Add(messageIndex, messageUID);
								}
								else
								{
									RaiseError(session, session.SingleLineParser.Line);
									session.IsClosing = true;
								}

								base.Callback.OnUidl(
									session,
									messageUIDs);
								GetRequestAndSendCommand(session);
								session.States.RemoveFirst();
								session.SingleLineParser.Reset();
							}
							else
							{
								RaiseError(session, session.SingleLineParser.Line);
							}
						}
					}
					else
					{
						RaiseError(session, session.SingleLineParser.Line);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Uidl_All

				case Pop3ClientState.Uidl_All:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							if (session.MultiLineParser.IsSuccessful)
							{
								SortedList<int, string> messageUIDs = new SortedList<int, string>();

								for (int index = 1; index < session.MultiLineParser.Lines.Count; index++)
								{
									Match match = _regExKeyValue.Match(session.MultiLineParser.Lines[index]);

									if (match.Success)
									{
										int messageIndex = TextProtocol.ParseInteger(match.Groups[1].Value);
										string messageUID = match.Groups[2].Value;

										messageUIDs.Add(messageIndex, messageUID);
									}
									else
									{
										RaiseError(session, session.MultiLineParser.Lines[index]);
										session.IsClosing = true;
									}
								}

								base.Callback.OnUidl(
									session,
									messageUIDs);
								GetRequestAndSendCommand(session);
								session.States.RemoveFirst();
								session.MultiLineParser.Reset();
							}
							else
							{
								RaiseError(session, session.MultiLineParser.FirstLine);
							}
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Retr

				case Pop3ClientState.Retr:
					session.BlockParser.Parse(session);

					if (session.BlockParser.IsValid)
					{
						if (session.BlockParser.IsComplete)
						{
							if (session.BlockParser.IsSuccessful)
							{
								base.Callback.OnRetrive(
									session,
									session.BlockParser.Data);

								GetRequestAndSendCommand(session);
								session.States.RemoveFirst();
								session.BlockParser.Reset();
							}
							else
							{
								RaiseError(session, session.BlockParser.Line);
							}
						}
					}
					else
					{
						RaiseError(session, session.BlockParser.Line);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Top

				case Pop3ClientState.Top:
					session.BlockParser.Parse(session);

					if (session.BlockParser.IsValid)
					{
						if (session.BlockParser.IsComplete)
						{
							if (session.BlockParser.IsSuccessful)
							{
								base.Callback.OnTop(
									session,
									session.BlockParser.Data);

								GetRequestAndSendCommand(session);
								session.States.RemoveFirst();
								session.BlockParser.Reset();
							}
							else
							{
								RaiseError(session, session.BlockParser.Line);
							}
						}
					}
					else
					{
						RaiseError(session, session.BlockParser.Line);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Dele

				case Pop3ClientState.Dele:
					session.SingleLineParser.Parse(session);

					if (session.SingleLineParser.IsValid)
					{
						if (session.SingleLineParser.IsComplete)
						{
							if (session.SingleLineParser.IsSuccessful)
							{
								GetRequestAndSendCommand(session);
								session.States.RemoveFirst();
								session.SingleLineParser.Reset();
							}
							else
							{
								RaiseError(session, session.SingleLineParser.Line);
							}
						}
					}
					else
					{
						RaiseError(session, session.SingleLineParser.Line);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Rset

				case Pop3ClientState.Rset:
					session.SingleLineParser.Parse(session);

					if (session.SingleLineParser.IsValid)
					{
						if (session.SingleLineParser.IsComplete)
						{
							if (session.SingleLineParser.IsSuccessful)
							{
								GetRequestAndSendCommand(session);
								session.States.RemoveFirst();
								session.SingleLineParser.Reset();
							}
							else
							{
								RaiseError(session, session.SingleLineParser.Line);
							}
						}
					}
					else
					{
						RaiseError(session, session.SingleLineParser.Line);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Quit

				case Pop3ClientState.Quit:
					session.SingleLineParser.Parse(session);

					if (session.SingleLineParser.IsValid)
					{
						if (session.SingleLineParser.IsComplete)
						{
							if (session.SingleLineParser.IsSuccessful)
							{
								session.IsClosing = true;
								session.SingleLineParser.Reset();
							}
							else
							{
								RaiseError(session, session.SingleLineParser.Line);
							}
						}
					}
					else
					{
						RaiseError(session, session.SingleLineParser.Line);
						session.IsClosing = true;
					}
					break;

				#endregion
			}
		}

		#endregion
	}

	public sealed class Pop3Client : Pop3ClientBase<TcpClientSocket>
	{
		public Pop3Client(IPop3ClientCallback<TcpClientSocket> callback)
			: base(callback)
		{
		}
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
