﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Net Library                              */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;

using Nabu.Collections;

namespace Nabu.Net
{
	public sealed class Pop3ClientSession<TSocket> : NetworkSessionBase<TSocket>
		where TSocket : ITcpClientSocket, new()
	{
		#region Types

		internal sealed class SingleLineResponseParser
		{
			#region Types

			private enum State
			{
				Default,
				Invalid,
				Plus,
				PlusO,
				PlusOK,
				PlusOKSymbol,
				PlusOKSymbolCr,
				PlusOKSymbolCrLf,
				Minus,
				MinusE,
				MinusER,
				MinusERR,
				MinusERRSymbol,
				MinusERRSymbolCr,
				MinusERRSymbolCrLf,
			}

			#endregion

			#region Fields

			private State _state = State.Default;
			private StringBuilder _line = new StringBuilder(1024);

			#endregion

			#region Methods

			[DebuggerStepThrough]
			public SingleLineResponseParser()
			{
			}

			[DebuggerStepThrough]
			public void Reset()
			{
				this._state = State.Default;
				this._line.Length = 0;
			}

			public void Parse(Pop3ClientSession<TSocket> session)
			{
				while (session.ReceiveSize > 0)
				{
					char symbol = (char)session.ReceiveStream.ReadByte();

					switch (this._state)
					{
						case State.Default:
							switch (symbol)
							{
								case '+':
									this._state = State.Plus;
									break;
								case '-':
									this._state = State.Minus;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.Invalid:

							return;
						case State.Plus:
							switch (symbol)
							{
								case 'O':
									this._state = State.PlusO;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.PlusO:
							switch (symbol)
							{
								case 'K':
									this._state = State.PlusOK;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.PlusOK:
							switch (symbol)
							{
								case '\r':
									this._state = State.PlusOKSymbolCr;
									break;
								default:
									this._state = State.PlusOKSymbol;
									break;
							}
							break;
						case State.PlusOKSymbol:
							switch (symbol)
							{
								case '\r':
									this._state = State.PlusOKSymbolCr;
									break;
								default:
									this._line.Append(symbol);
									break;
							}
							break;
						case State.PlusOKSymbolCr:
							switch (symbol)
							{
								case '\n':
									this._state = State.PlusOKSymbolCrLf;

									return;
								default:
									this._state = State.Invalid;

									return;
							}
						case State.PlusOKSymbolCrLf:
							switch (symbol)
							{
								default:
									this._state = State.Invalid;

									return;
							}
						case State.Minus:
							switch (symbol)
							{
								case 'E':
									this._state = State.MinusE;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.MinusE:
							switch (symbol)
							{
								case 'R':
									this._state = State.MinusER;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.MinusER:
							switch (symbol)
							{
								case 'R':
									this._state = State.MinusERR;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.MinusERR:
							switch (symbol)
							{
								case '\r':
									this._state = State.MinusERRSymbolCr;
									break;
								default:
									this._state = State.MinusERRSymbol;
									break;
							}
							break;
						case State.MinusERRSymbol:
							switch (symbol)
							{
								case '\r':
									this._state = State.MinusERRSymbolCr;
									break;
								default:
									this._line.Append(symbol);
									break;
							}
							break;
						case State.MinusERRSymbolCr:
							switch (symbol)
							{
								case '\n':
									this._state = State.MinusERRSymbolCrLf;

									return;
								default:
									this._state = State.Invalid;

									return;
							}
						case State.MinusERRSymbolCrLf:
							this._state = State.Invalid;

							return;
					}
				}
			}

			#endregion

			#region Properties

			public bool IsValid
			{
				[DebuggerStepThrough]
				get { return this._state != State.Invalid; }
			}

			public bool IsComplete
			{
				[DebuggerStepThrough]
				get
				{
					return
						(this._state == State.PlusOKSymbolCrLf) ||
						(this._state == State.MinusERRSymbolCrLf);
				}
			}

			public bool IsSuccessful
			{
				[DebuggerStepThrough]
				get
				{
					return
						(this._state == State.PlusOKSymbolCrLf);
				}
			}

			public string Line
			{
				[DebuggerStepThrough]
				get { return this._line.ToString(); }
			}

			#endregion
		}

		internal sealed class MultiLineResponseParser
		{
			#region Types

			private enum State
			{
				Default,
				Invalid,
				Plus,
				PlusO,
				PlusOK,
				PlusOKSymbol,
				PlusOKSymbolCr,
				PlusOKSymbolCrLf,
				PlusOKSymbolCrLfDot,
				PlusOKSymbolCrLfDotCr,
				PlusOKSymbolCrLfDotCrLf,
				Minus,
				MinusE,
				MinusER,
				MinusERR,
				MinusERRSymbol,
				MinusERRSymbolCr,
				MinusERRSymbolCrLf,
			}

			#endregion

			#region Fields

			private State _state = State.Default;
			private StringBuilder _line = new StringBuilder(1024);
			private List<string> _lines = new List<string>();

			#endregion

			#region Methods

			[DebuggerStepThrough]
			public MultiLineResponseParser()
			{
			}

			[DebuggerStepThrough]
			public void Reset()
			{
				this._state = State.Default;
				this._line.Length = 0;
				this._lines.Clear();
			}

			public void Parse(Pop3ClientSession<TSocket> session)
			{
				while (session.ReceiveSize > 0)
				{
					char symbol = (char)session.ReceiveStream.ReadByte();

					switch (this._state)
					{
						case State.Default:
							switch (symbol)
							{
								case '+':
									this._state = State.Plus;
									break;
								case '-':
									this._state = State.Minus;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.Invalid:

							return;
						case State.Plus:
							switch (symbol)
							{
								case 'O':
									this._state = State.PlusO;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.PlusO:
							switch (symbol)
							{
								case 'K':
									this._state = State.PlusOK;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.PlusOK:
							switch (symbol)
							{
								case '\r':
									this._state = State.PlusOKSymbolCr;
									break;
								default:
									this._state = State.PlusOKSymbol;
									break;
							}
							break;
						case State.PlusOKSymbol:
							switch (symbol)
							{
								case '\r':
									this._state = State.PlusOKSymbolCr;
									break;
								default:
									this._line.Append(symbol);
									break;
							}
							break;
						case State.PlusOKSymbolCr:
							switch (symbol)
							{
								case '\n':
									this._state = State.PlusOKSymbolCrLf;
									this._lines.Add(this._line.ToString());
									this._line.Length = 0;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.PlusOKSymbolCrLf:
							switch (symbol)
							{
								case '\r':
									this._state = State.PlusOKSymbolCr;
									break;
								case '.':
									this._state = State.PlusOKSymbolCrLfDot;
									this._line.Append(symbol);
									break;
								default:
									this._state = State.PlusOKSymbol;
									this._line.Append(symbol);
									break;
							}
							break;
						case State.PlusOKSymbolCrLfDot:
							switch (symbol)
							{
								case '\r':
									this._state = State.PlusOKSymbolCrLfDotCr;
									break;
								default:
									this._state = State.PlusOKSymbol;
									this._line.Append(symbol);
									break;
							}
							break;
						case State.PlusOKSymbolCrLfDotCr:
							switch (symbol)
							{
								case '\n':
									this._state = State.PlusOKSymbolCrLfDotCrLf;

									return;
								default:
									this._state = State.Invalid;

									return;
							}
						case State.PlusOKSymbolCrLfDotCrLf:
							this._state = State.Invalid;

							return;
						case State.Minus:
							switch (symbol)
							{
								case 'E':
									this._state = State.MinusE;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.MinusE:
							switch (symbol)
							{
								case 'R':
									this._state = State.MinusER;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.MinusER:
							switch (symbol)
							{
								case 'R':
									this._state = State.MinusERR;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.MinusERR:
							switch (symbol)
							{
								case '\r':
									this._state = State.MinusERRSymbolCr;
									break;
								default:
									this._state = State.MinusERRSymbol;
									break;
							}
							break;
						case State.MinusERRSymbol:
							switch (symbol)
							{
								case '\r':
									this._state = State.MinusERRSymbolCr;
									break;
								default:
									this._line.Append(symbol);
									break;
							}
							break;
						case State.MinusERRSymbolCr:
							switch (symbol)
							{
								case '\n':
									this._state = State.MinusERRSymbolCrLf;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.MinusERRSymbolCrLf:
							this._state = State.Invalid;

							return;
					}
				}
			}

			#endregion

			#region Properties

			public bool IsValid
			{
				[DebuggerStepThrough]
				get { return this._state != State.Invalid; }
			}

			public bool IsComplete
			{
				[DebuggerStepThrough]
				get
				{
					return
						(this._state == State.PlusOKSymbolCrLfDotCrLf) ||
						(this._state == State.MinusERRSymbolCrLf);
				}
			}

			public bool IsSuccessful
			{
				[DebuggerStepThrough]
				get
				{
					return
						(this._state == State.PlusOKSymbolCrLfDotCrLf);
				}
			}

			public string FirstLine
			{
				[DebuggerStepThrough]
				get
				{
					return this._lines.Count > 0 ? this._lines[0] : string.Empty;
				}
			}

			public IList<string> Lines
			{
				[DebuggerStepThrough]
				get { return this._lines; }
			}

			#endregion
		}

		internal sealed class BlockResponseParser
		{
			#region Types

			private enum State
			{
				Default,
				Invalid,
				Plus,
				PlusO,
				PlusOK,
				PlusOKSymbol,
				PlusOKSymbolCr,
				PlusOKSymbolCrLf,
				PlusOKSymbol2,
				PlusOKSymbolCr2,
				PlusOKSymbolCrLf2,
				PlusOKSymbolCrLfDot,
				PlusOKSymbolCrLfDotCr,
				PlusOKSymbolCrLfDotCrLf,
				Minus,
				MinusE,
				MinusER,
				MinusERR,
				MinusERRSymbol,
				MinusERRSymbolCr,
				MinusERRSymbolCrLf,
			}

			#endregion

			#region Fields

			private State _state = State.Default;
			private StringBuilder _line = new StringBuilder(1024);
			private ListEx<byte> _data = new ListEx<byte>(1024 * 1024);

			#endregion

			#region Methods

			[DebuggerStepThrough]
			public BlockResponseParser()
			{
			}

			[DebuggerStepThrough]
			public void Reset()
			{
				this._state = State.Default;
				this._data.Clear();
			}

			[DebuggerStepThrough]
			public void SetCapacity(int capacity)
			{
				if (this._data.Capacity < capacity)
				{
					this._data.Capacity = capacity;
				}
			}

			public void Parse(Pop3ClientSession<TSocket> session)
			{
				while (session.ReceiveSize > 0)
				{
					int value = session.ReceiveStream.ReadByte();
					char symbol = (char)value;

					switch (this._state)
					{
						case State.Default:
							switch (symbol)
							{
								case '+':
									this._state = State.Plus;
									break;
								case '-':
									this._state = State.Minus;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.Invalid:

							return;
						case State.Plus:
							switch (symbol)
							{
								case 'O':
									this._state = State.PlusO;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.PlusO:
							switch (symbol)
							{
								case 'K':
									this._state = State.PlusOK;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.PlusOK:
							switch (symbol)
							{
								case '\r':
									this._state = State.PlusOKSymbolCr;
									break;
								default:
									this._state = State.PlusOKSymbol;
									break;
							}
							break;
						case State.PlusOKSymbol:
							switch (symbol)
							{
								case '\r':
									this._state = State.PlusOKSymbolCr;
									break;
								default:
									this._line.Append(symbol);
									break;
							}
							break;
						case State.PlusOKSymbolCr:
							switch (symbol)
							{
								case '\n':
									this._state = State.PlusOKSymbolCrLf;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.PlusOKSymbolCrLf:
							switch (symbol)
							{
								case '.':
									this._state = State.PlusOKSymbolCrLfDot;
									break;
								default:
									this._data.Add((byte)value);
									this._state = State.PlusOKSymbol2;
									break;
							}
							break;
						case State.PlusOKSymbol2:
							switch (symbol)
							{
								case '\r':
									this._data.Add((byte)value);
									this._state = State.PlusOKSymbolCr2;
									break;
								default:
									this._data.Add((byte)value);
									break;
							}
							break;
						case State.PlusOKSymbolCr2:
							switch (symbol)
							{
								case '\n':
									this._data.Add((byte)value);
									this._state = State.PlusOKSymbolCrLf2;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.PlusOKSymbolCrLf2:
							switch (symbol)
							{
								case '\r':
									this._data.Add((byte)value);
									this._state = State.PlusOKSymbolCr2;
									break;
								case '.':
									this._data.Add((byte)value);
									this._state = State.PlusOKSymbolCrLfDot;
									break;
								default:
									this._data.Add((byte)value);
									this._state = State.PlusOKSymbol2;
									break;
							}
							break;
						case State.PlusOKSymbolCrLfDot:
							switch (symbol)
							{
								case '\r':
									this._data.RemoveAt(this._data.Count - 1);
									this._state = State.PlusOKSymbolCrLfDotCr;
									break;
								default:
									this._data.Add((byte)value);
									this._state = State.PlusOKSymbol2;
									break;
							}
							break;
						case State.PlusOKSymbolCrLfDotCr:
							switch (symbol)
							{
								case '\n':
									this._state = State.PlusOKSymbolCrLfDotCrLf;

									return;
								default:
									this._state = State.Invalid;

									return;
							}
						case State.PlusOKSymbolCrLfDotCrLf:
							this._state = State.Invalid;

							return;
						case State.Minus:
							switch (symbol)
							{
								case 'E':
									this._state = State.MinusE;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.MinusE:
							switch (symbol)
							{
								case 'R':
									this._state = State.MinusER;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.MinusER:
							switch (symbol)
							{
								case 'R':
									this._state = State.MinusERR;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.MinusERR:
							switch (symbol)
							{
								case '\r':
									this._state = State.MinusERRSymbolCr;
									break;
								default:
									this._state = State.MinusERRSymbol;
									break;
							}
							break;
						case State.MinusERRSymbol:
							switch (symbol)
							{
								case '\r':
									this._state = State.MinusERRSymbolCr;
									break;
								default:
									this._line.Append(symbol);
									break;
							}
							break;
						case State.MinusERRSymbolCr:
							switch (symbol)
							{
								case '\n':
									this._state = State.MinusERRSymbolCrLf;

									return;
								default:
									this._state = State.Invalid;

									return;
							}
						case State.MinusERRSymbolCrLf:
							this._state = State.Invalid;

							return;
					}
				}
			}

			#endregion

			#region Properties

			public bool IsValid
			{
				[DebuggerStepThrough]
				get { return this._state != State.Invalid; }
			}

			public bool IsComplete
			{
				[DebuggerStepThrough]
				get
				{
					return
						(this._state == State.PlusOKSymbolCrLfDotCrLf) ||
						(this._state == State.MinusERRSymbolCrLf);
				}
			}

			public bool IsSuccessful
			{
				[DebuggerStepThrough]
				get { return (this._state == State.PlusOKSymbolCrLfDotCrLf); }
			}

			public string Line
			{
				[DebuggerStepThrough]
				get
				{
					return this._line.ToString();
				}
			}

			public ArraySegment<byte> Data
			{
				[DebuggerStepThrough]
				get { return this._data.ToArraySegment(); }
			}

			#endregion
		}

		#endregion

		#region Fields

		private LinkedList<Pop3ClientState> _states = new LinkedList<Pop3ClientState>();
		private SingleLineResponseParser _singleLineParser = new SingleLineResponseParser();
		private MultiLineResponseParser _multiLineParser = new MultiLineResponseParser();
		private BlockResponseParser _blockParser = new BlockResponseParser();
		private string _hostname;
		private string _username;
		private string _password;
		private Pop3ClientOptions _options;
		private int _optionLoginDelay;
		private object _customData;

		#endregion

		#region Methods

		[DebuggerStepThrough]
		public Pop3ClientSession(object customData)
		{
			this._customData = customData;
		}

		internal void ResetOptions()
		{
			this._options = 0;
		}

		internal void SetOptionSupported(Pop3ClientOptions option)
		{
			this._options |= option;
		}

		[DebuggerStepThrough]
		public bool GetOptionSupported(Pop3ClientOptions option)
		{
			return (this._options & option) != 0;
		}

		[DebuggerStepThrough]
		public void SetOptionUnsupported(Pop3ClientOptions option)
		{
			this._options &= ~option;
		}

		[DebuggerStepThrough]
		public int GetOptionValue(Pop3ClientOptions option)
		{
			switch (option)
			{
				case Pop3ClientOptions.OptionLoginDelay:
					return this._optionLoginDelay;
			}

			return 0;
		}

		#endregion

		#region Properties

		internal LinkedList<Pop3ClientState> States
		{
			[DebuggerStepThrough]
			get { return this._states; }
			[DebuggerStepThrough]
			set { this._states = value; }
		}

		internal SingleLineResponseParser SingleLineParser
		{
			[DebuggerStepThrough]
			get { return this._singleLineParser; }
		}

		internal MultiLineResponseParser MultiLineParser
		{
			[DebuggerStepThrough]
			get { return this._multiLineParser; }
		}

		internal BlockResponseParser BlockParser
		{
			[DebuggerStepThrough]
			get { return this._blockParser; }
		}

		internal string Hostname
		{
			[DebuggerStepThrough]
			get { return this._hostname; }
			[DebuggerStepThrough]
			set { this._hostname = value; }
		}

		internal string Username
		{
			[DebuggerStepThrough]
			get { return this._username; }
			[DebuggerStepThrough]
			set { this._username = value; }
		}

		internal string Password
		{
			[DebuggerStepThrough]
			get { return this._password; }
			[DebuggerStepThrough]
			set { this._password = value; }
		}

		internal int OptionLoginDelay
		{
			[DebuggerStepThrough]
			get { return this._optionLoginDelay; }
			[DebuggerStepThrough]
			set { this._optionLoginDelay = value; }
		}

		public object CustomData
		{
			[DebuggerStepThrough]
			get { return this._customData; }
			[DebuggerStepThrough]
			set { this._customData = value; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
