﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Net Library                              */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;

namespace Nabu.Net
{
	public class Pop3ServerBase<TSocketListen, TSocketIO> :
		TcpServerBase<TSocketListen, TSocketIO, Pop3ServerSession<TSocketIO>, IPop3ServerCallback<TSocketIO>>
		where TSocketListen : ITcpServerSocketListen, new()
		where TSocketIO : ITcpServerSocketIO
	{
		#region Pop3ServerBase

		#region Constants

		private const string PositiveWelcome = "+OK Server ready\r\n";
		private const string PositiveNOOP = "+OK Nothing done\r\n";
		private const string PositiveSTLS = "+OK Ready to start TLS\r\n";
		private const string PositiveUSER = "+OK Password required\r\n";
		private const string PositivePASS = "+OK Logged in\r\n";
		private const string PositiveCAPA = "+OK Capability list follows\r\n";
		private const string PositiveLIST = "+OK Message list follows\r\n";
		private const string PositiveUIDL = "+OK Unique-ID list follows\r\n";
		private const string PositiveRETR = "+OK Message follows\r\n";
		private const string PositiveTOP = "+OK Message follows\r\n";
		private const string PositiveDELE = "+OK Message deleted\r\n";
		private const string PositiveRSET = "+OK Reset state\r\n";
		private const string PositiveQUIT = "+OK Server quiting\r\n";

		private const string NegativeCommand = "-ERR Unknown command\r\n";
		private const string NegativeParameterExpected = "-ERR Invalid syntax, parameter expected\r\n";
		private const string NegativeParameterUnexpected = "-ERR Invalid syntax, parameter unexpected\r\n";
		private const string NegativeOneParameterExpected = "-ERR Invalid syntax, no more than one parameter expected\r\n";
		private const string NegativeParameterNotIndex = "-ERR Invalid syntax, parameter must be positive integer\r\n";
		private const string NegativeParameterInvalidIndex = "-ERR Invalid message index\r\n";
		private const string NegativeCommandSequence = "-ERR Invalid command sequence\r\n";
		private const string NegativePASS = "-ERR Invalid username or password\r\n";

		#endregion

		#region Methods

		private void RaiseError(
			Pop3ServerSession<TSocketIO> session,
			string text)
		{
			base.Callback.OnError(
				session,
				session.States.First.Value,
				text);
			session.IsClosing = true;
		}

		public Pop3ServerBase(IPop3ServerCallback<TSocketIO> callback)
			: base(callback)
		{
		}

		#endregion

		#endregion

		#region TcpServerBase

		protected override Pop3ServerSession<TSocketIO> CreateSession(object customData)
		{
			Pop3ServerSession<TSocketIO> session = new Pop3ServerSession<TSocketIO>();

			session.States.AddLast(Pop3ServerState.Default);
			TextProtocol.Write(session, PositiveWelcome);

			return session;
		}

		protected override void OnReceive(Pop3ServerSession<TSocketIO> session)
		{
			switch (session.States.Last.Value)
			{
				#region Default

				case Pop3ServerState.Default:
					session.CommandParser.Parse(session);

					if (session.CommandParser.IsValid)
					{
						if (session.CommandParser.IsComplete)
						{
							switch (session.CommandParser.Command)
							{
								case "NOOP":
									if (session.CommandParser.FirstArgument != null)
									{
										TextProtocol.Write(session, NegativeParameterUnexpected);
									}
									else
									{
										TextProtocol.Write(session, PositiveNOOP);
									}
									break;
								case "CAPA":
									if (session.CommandParser.FirstArgument != null)
									{
										TextProtocol.Write(session, NegativeParameterUnexpected);
									}
									else
									{
										TextProtocol.Write(session, PositiveCAPA);
										TextProtocol.Write(session, "IMPLEMENTATION Nabu.Net.Pop3Server\r\n");
										TextProtocol.Write(session, "PIPELINING\r\n");

										if (!session.Socket.IsEncrypted)
										{
											TextProtocol.Write(session, "STLS\r\n");
										}

										TextProtocol.Write(session, "TOP\r\n");
										TextProtocol.Write(session, "UIDL\r\n");
										TextProtocol.Write(session, "USER\r\n");
										TextProtocol.Write(session, ".\r\n");
									}
									break;
								case "STLS":
									if (session.CommandParser.FirstArgument != null)
									{
										TextProtocol.Write(session, NegativeParameterUnexpected);
									}
									else
									{
										TextProtocol.Write(session, PositiveSTLS);
										session.States.AddFirst(Pop3ServerState.STLS);
										session.States.RemoveLast();
									}
									break;
								case "USER":
									if (session.CommandParser.FirstArgument == null)
									{
										TextProtocol.Write(session, NegativeParameterExpected);
									}
									else
									{
										session.Username = session.CommandParser.FirstArgument;
										TextProtocol.Write(session, PositiveUSER);
										session.States.AddFirst(Pop3ServerState.AuthPass);
										session.States.RemoveLast();
									}
									break;
								case "PASS":
								case "STAT":
								case "LIST":
								case "UIDL":
								case "RETR":
								case "TOP":
								case "DELE":
									TextProtocol.Write(session, NegativeCommandSequence);
									break;
								case "RSET":
									if (session.CommandParser.FirstArgument != null)
									{
										TextProtocol.Write(session, NegativeParameterUnexpected);
									}
									else
									{
										base.Callback.OnRset(session);
										TextProtocol.Write(session, PositiveRSET);
									}
									break;
								case "QUIT":
									if (session.CommandParser.FirstArgument != null)
									{
										TextProtocol.Write(session, NegativeParameterUnexpected);
									}
									else
									{
										base.Callback.OnQuit(session);
										TextProtocol.Write(session, PositiveQUIT);
										session.States.AddFirst(Pop3ServerState.Quit);
										session.States.RemoveLast();
									}
									break;
								default:
									TextProtocol.Write(session, NegativeCommand);
									break;
							}

							session.CommandParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.CommandParser.Command);
					}
					break;

				#endregion

				#region AuthPass

				case Pop3ServerState.AuthPass:
					session.CommandParser.Parse(session);

					if (session.CommandParser.IsValid)
					{
						if (session.CommandParser.IsComplete)
						{
							switch (session.CommandParser.Command)
							{
								case "NOOP":
									if (session.CommandParser.FirstArgument != null)
									{
										TextProtocol.Write(session, NegativeParameterUnexpected);
									}
									else
									{
										TextProtocol.Write(session, PositiveNOOP);
									}
									break;
								case "PASS":
									if (session.CommandParser.FirstArgument == null)
									{
										TextProtocol.Write(session, NegativeParameterExpected);
									}
									else
									{
										session.Password = session.CommandParser.FirstArgument;

										bool isValid;

										base.Callback.OnCredentials(session, out isValid);

										if (isValid)
										{
											TextProtocol.Write(session, PositivePASS);
											session.States.AddFirst(Pop3ServerState.Transaction);
											session.States.RemoveLast();
										}
										else
										{
											session.Username = null;
											session.Password = null;
											TextProtocol.Write(session, NegativePASS);
											session.States.AddFirst(Pop3ServerState.Default);
											session.States.RemoveLast();
										}
									}
									break;
								case "STAT":
								case "LIST":
								case "UIDL":
								case "RETR":
								case "TOP":
								case "DELE":
									TextProtocol.Write(session, NegativeCommandSequence);
									break;
								case "RSET":
									if (session.CommandParser.FirstArgument != null)
									{
										TextProtocol.Write(session, NegativeParameterUnexpected);
									}
									else
									{
										base.Callback.OnRset(session);
										TextProtocol.Write(session, PositiveRSET);
										session.States.AddFirst(Pop3ServerState.Default);
										session.States.RemoveLast();
									}
									break;
								default:
									TextProtocol.Write(session, NegativeCommand);
									break;
							}
							session.CommandParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.CommandParser.Command);
					}
					break;

				#endregion

				#region STLS

				case Pop3ServerState.STLS:
					{
						X509CertificateCollection certificates;
						SslProtocols enabledProtocols;
						bool mutualAuthenticationRequired;
						bool checkCertificateRevokation;

						base.Callback.OnSslAuthenticate(
							session,
							out certificates,
							out enabledProtocols,
							out mutualAuthenticationRequired,
							out checkCertificateRevokation);
						session.Socket.AuthenticateAsServer(
							certificates[0],
							mutualAuthenticationRequired,
							enabledProtocols,
							checkCertificateRevokation);
						session.States.AddFirst(Pop3ServerState.Default);
						session.States.RemoveLast();
					}
					break;

				#endregion

				#region Transaction

				case Pop3ServerState.Transaction:
					session.CommandParser.Parse(session);

					if (session.CommandParser.IsValid)
					{
						if (session.CommandParser.IsComplete)
						{
							switch (session.CommandParser.Command)
							{
								case "NOOP":
									if (session.CommandParser.FirstArgument != null)
									{
										TextProtocol.Write(session, NegativeParameterUnexpected);
									}
									else
									{
										TextProtocol.Write(session, PositiveNOOP);
									}
									break;
								case "CAPA":
									if (session.CommandParser.FirstArgument != null)
									{
										TextProtocol.Write(session, NegativeParameterUnexpected);
									}
									else
									{
										TextProtocol.Write(session, PositiveCAPA);
										TextProtocol.Write(session, "PIPELINING\r\n");
										TextProtocol.Write(session, "TOP\r\n");
										TextProtocol.Write(session, "UIDL\r\n");
										TextProtocol.Write(session, "IMPLEMENTATION Nabu.Net.Pop3Server\r\n");
										TextProtocol.Write(session, ".\r\n");
									}
									break;
								case "USER":
								case "PASS":
									TextProtocol.Write(session, NegativeCommandSequence);
									break;
								case "STAT":
									if (session.CommandParser.FirstArgument != null)
									{
										TextProtocol.Write(session, NegativeParameterUnexpected);
									}
									else
									{
										int totalCount;
										int totalSize;

										base.Callback.OnStat(
											session,
											out totalCount,
											out totalSize);

										TextProtocol.Write(session, "+OK " + totalCount + " " + totalSize + "\r\n");
									}
									break;
								case "LIST":
									if (session.CommandParser.FirstArgument == null)
									{
										IDictionary<int, int> messages;

										base.Callback.OnList(
											session,
											out messages);

										TextProtocol.Write(session, PositiveLIST);

										foreach (KeyValuePair<int, int> message in messages)
										{
											TextProtocol.Write(session, (message.Key + 1) + " " + message.Value + "\r\n");
										}

										TextProtocol.Write(session, ".\r\n");
									}
									else
									{
										if (session.CommandParser.SecondArgument != null)
										{
											TextProtocol.Write(session, NegativeOneParameterExpected);
										}
										else
										{
											int index;

											if (
												!int.TryParse(session.CommandParser.FirstArgument, out index) ||
												(index <= 0))
											{
												TextProtocol.Write(session, NegativeParameterNotIndex);
											}
											else
											{
												int size;
												bool isFound;

												base.Callback.OnList(
													session,
													index - 1,
													out isFound,
													out size);

												if (isFound)
												{
													TextProtocol.Write(session, "+OK " + index + " " + size + "\r\n");
												}
												else
												{
													TextProtocol.Write(session, NegativeParameterInvalidIndex);
												}
											}
										}
									}
									break;
								case "UIDL":
									if (session.CommandParser.FirstArgument == null)
									{
										IDictionary<int, string> messages;

										base.Callback.OnUidl(
											session,
											out messages);

										TextProtocol.Write(session, PositiveUIDL);

										foreach (KeyValuePair<int, string> message in messages)
										{
											TextProtocol.Write(session, (message.Key + 1) + " " + message.Value + "\r\n");
										}

										TextProtocol.Write(session, ".\r\n");
									}
									else
									{
										if (session.CommandParser.SecondArgument != null)
										{
											TextProtocol.Write(session, NegativeOneParameterExpected);
										}
										else
										{
											int index;

											if (
												!int.TryParse(session.CommandParser.FirstArgument, out index) ||
												(index <= 0))
											{
												TextProtocol.Write(session, NegativeParameterNotIndex);
											}
											else
											{
												string uid;
												bool isFound;

												base.Callback.OnUidl(
													session,
													index - 1,
													out isFound,
													out uid);

												if (isFound)
												{
													TextProtocol.Write(session, "+OK " + index + " " + uid + "\r\n");
												}
												else
												{
													TextProtocol.Write(session, NegativeParameterInvalidIndex);
												}
											}
										}
									}
									break;
								case "RETR":
									if (session.CommandParser.FirstArgument == null)
									{
										TextProtocol.Write(session, NegativeParameterExpected);
									}
									else
									{
										int index;

										if (
											!int.TryParse(session.CommandParser.FirstArgument, out index) ||
											(index <= 0))
										{
											TextProtocol.Write(session, NegativeParameterNotIndex);
										}
										else
										{
											bool isFound;
											ArraySegment<byte> segment;

											base.Callback.OnRetr(
												session,
												index - 1,
												out isFound,
												out segment);

											if (isFound)
											{
												TextProtocol.Write(session, PositiveRETR);
												session.SendStream.Write(
													segment.Array,
													segment.Offset,
													segment.Count);
												TextProtocol.Write(session, "\r\n.\r\n");
											}
											else
											{
												TextProtocol.Write(session, NegativeParameterInvalidIndex);
											}
										}
									}
									break;
								case "TOP":
									if (session.CommandParser.FirstArgument == null)
									{
										TextProtocol.Write(session, NegativeParameterExpected);
									}
									else
									{
										int index;

										if (
											!int.TryParse(session.CommandParser.FirstArgument, out index) ||
											(index <= 0))
										{
											TextProtocol.Write(session, NegativeParameterNotIndex);
										}
										else
										{
											bool isFound;
											ArraySegment<byte> segment;

											base.Callback.OnRetr(
												session,
												index - 1,
												out isFound,
												out segment);

											if (isFound)
											{
												TextProtocol.Write(session, PositiveTOP);
												session.SendStream.Write(
													segment.Array,
													segment.Offset,
													segment.Count);
												TextProtocol.Write(session, "\r\n.\r\n");
											}
											else
											{
												TextProtocol.Write(session, NegativeParameterInvalidIndex);
											}
										}
									}
									break;
								case "DELE":
									if (session.CommandParser.FirstArgument == null)
									{
										TextProtocol.Write(session, NegativeParameterExpected);
									}
									else
									{
										int index;

										if (
											!int.TryParse(session.CommandParser.FirstArgument, out index) ||
											(index <= 0))
										{
											TextProtocol.Write(session, NegativeParameterNotIndex);
										}
										else
										{
											bool isFound;

											base.Callback.OnDele(
												session,
												index - 1,
												out isFound);

											if (isFound)
											{
												TextProtocol.Write(session, PositiveDELE);
											}
											else
											{
												TextProtocol.Write(session, NegativeParameterInvalidIndex);
											}
										}
									}
									break;
								case "RSET":
									base.Callback.OnRset(session);
									TextProtocol.Write(session, PositiveRSET);
									break;
								case "QUIT":
									base.Callback.OnQuit(session);
									TextProtocol.Write(session, PositiveQUIT);
									session.States.AddFirst(Pop3ServerState.Quit);
									session.States.RemoveLast();
									break;
								default:
									TextProtocol.Write(session, NegativeCommand);
									break;
							}

							session.CommandParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.CommandParser.Command);
					}
					break;

				#endregion

				#region Quit

				case Pop3ServerState.Quit:
					session.IsClosing = true;
					break;

				#endregion
			}
		}

		protected override void OnSend(Pop3ServerSession<TSocketIO> session)
		{
			switch (session.States.Last.Value)
			{
				#region Default

				case Pop3ServerState.Default:
					break;

				#endregion

				#region AuthPass

				case Pop3ServerState.AuthPass:
					break;

				#endregion

				#region STLS

				case Pop3ServerState.STLS:
					break;

				#endregion

				#region Transaction

				case Pop3ServerState.Transaction:
					break;

				#endregion

				#region Quit

				case Pop3ServerState.Quit:
					session.IsClosing = true;
					break;

				#endregion
			}
		}

		#endregion
	}

	public sealed class Pop3Server : Pop3ServerBase<TcpServerSocketListen, TcpServerSocketIO>
	{
		public Pop3Server(IPop3ServerCallback<TcpServerSocketIO> callback)
			: base(callback)
		{
		}
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
