﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Net Library                              */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Text;

namespace Nabu.Net
{
	public sealed class Pop3ServerSession<TSocket> : NetworkSessionBase<TSocket>
		where TSocket : ITcpServerSocketIO
	{
		#region Types

		internal sealed class CommandRequestParser
		{
			#region Types

			private enum State
			{
				Default,
				Invalid,
				Command,
				CommandSpace,
				FirstArgument,
				FirstArgumentSpace,
				SecondArgument,
				Cr,
				CrLf
			}

			#endregion

			#region Fields

			private State _state = State.Default;
			private StringBuilder _line = new StringBuilder(1024);
			private string _command = string.Empty;
			private string _firstArgument = null;
			private string _secondArgument = null;

			#endregion

			#region Methods

			public CommandRequestParser()
			{
			}

			public void Reset()
			{
				this._state = State.Default;
				this._line.Length = 0;
				this._command = string.Empty;
				this._firstArgument = null;
				this._secondArgument = null;
			}

			public void Parse(Pop3ServerSession<TSocket> session)
			{
				while (session.ReceiveSize > 0)
				{
					char symbol = (char)session.ReceiveStream.ReadByte();

					switch (this._state)
					{
						case State.Default:
							if (
								((symbol >= 'A') && (symbol <= 'Z')) ||
								((symbol >= 'a') && (symbol <= 'z')))
							{
								this._line.Append(char.ToUpper(symbol, CultureInfo.InvariantCulture));
								this._state = State.Command;
							}
							else
							{
								this._state = State.Invalid;

								return;
							}
							break;
						case State.Invalid:

							return;
						case State.Command:
							if (
								((symbol >= 'A') && (symbol <= 'Z')) ||
								((symbol >= 'a') && (symbol <= 'z')))
							{
								this._line.Append(char.ToUpper(symbol, CultureInfo.InvariantCulture));
							}
							else if (symbol == ' ')
							{
								this._command = this._line.ToString();
								this._firstArgument = string.Empty;
								this._line.Length = 0;
								this._state = State.CommandSpace;
							}
							else if (symbol == '\r')
							{
								this._command = this._line.ToString();
								this._line.Length = 0;
								this._state = State.Cr;
							}
							else
							{
								this._state = State.Invalid;

								return;
							}
							break;
						case State.CommandSpace:
							if ((symbol >= (char)0x21) && (symbol <= (char)0x7E))
							{
								this._line.Append(symbol);
								this._state = State.FirstArgument;
							}
							else if (symbol == '\r')
							{
								this._firstArgument = string.Empty;
								this._state = State.Cr;
							}
							else
							{
								this._state = State.Invalid;

								return;
							}
							break;
						case State.FirstArgument:
							if ((symbol >= (char)0x21) && (symbol <= (char)0x7E))
							{
								this._line.Append(symbol);
							}
							else if (symbol == ' ')
							{
								switch (this._command)
								{
									case "PASS":
									case "USER":
										this._line.Append(symbol);
										break;
									default:
										this._firstArgument = this._line.ToString();
										this._secondArgument = string.Empty;
										this._line.Length = 0;
										this._state = State.FirstArgumentSpace;
										break;
								}
							}
							else if (symbol == '\r')
							{
								this._firstArgument = this._line.ToString();
								this._line.Length = 0;
								this._state = State.Cr;
							}
							else
							{
								this._state = State.Invalid;

								return;
							}
							break;
						case State.FirstArgumentSpace:
							if ((symbol >= (char)0x21) && (symbol <= (char)0x7E))
							{
								this._line.Append(symbol);
								this._state = State.SecondArgument;
							}
							else
							{
								this._state = State.Invalid;

								return;
							}
							break;
						case State.SecondArgument:
							if ((symbol >= (char)0x21) && (symbol <= (char)0x7E))
							{
								this._line.Append(symbol);
							}
							else if (symbol == '\r')
							{
								this._secondArgument = this._line.ToString();
								this._line.Length = 0;
								this._state = State.Cr;
							}
							else
							{
								this._state = State.Invalid;

								return;
							}
							break;
						case State.Cr:
							if (symbol == '\n')
							{
								this._state = State.CrLf;

								return;
							}
							else
							{
								this._state = State.Invalid;

								return;
							}
						case State.CrLf:
							switch (symbol)
							{
								default:
									this._state = State.Invalid;
									break;
							}
							break;
					}
				}
			}

			#endregion

			#region Properties

			public bool IsValid
			{
				[DebuggerStepThrough]
				get { return this._state != State.Invalid; }
			}

			public bool IsComplete
			{
				[DebuggerStepThrough]
				get
				{
					return this._state == State.CrLf;
				}
			}

			public string Command
			{
				[DebuggerStepThrough]
				get { return this._command; }
			}

			public string FirstArgument
			{
				[DebuggerStepThrough]
				get { return this._firstArgument; }
			}

			public string SecondArgument
			{
				[DebuggerStepThrough]
				get { return this._secondArgument; }
			}

			#endregion
		}

		#endregion

		#region Fields

		private LinkedList<Pop3ServerState> _states = new LinkedList<Pop3ServerState>();
		private CommandRequestParser _commandParser = new CommandRequestParser();
		private string _hostname;
		private string _username;
		private string _password;

		#endregion

		#region Methods

		[DebuggerStepThrough]
		public Pop3ServerSession()
		{
		}

		#endregion

		#region Properties

		internal LinkedList<Pop3ServerState> States
		{
			[DebuggerStepThrough]
			get { return this._states; }
			[DebuggerStepThrough]
			set { this._states = value; }
		}

		internal CommandRequestParser CommandParser
		{
			[DebuggerStepThrough]
			get { return this._commandParser; }
		}

		public string Hostname
		{
			[DebuggerStepThrough]
			get { return this._hostname; }
			[DebuggerStepThrough]
			set { this._hostname = value; }
		}

		public string Username
		{
			[DebuggerStepThrough]
			get { return this._username; }
			[DebuggerStepThrough]
			set { this._username = value; }
		}

		public string Password
		{
			[DebuggerStepThrough]
			get { return this._password; }
			[DebuggerStepThrough]
			set { this._password = value; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
