﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Net Library                              */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Text;

namespace Nabu.Net
{
	public class SmtpClientBase<TSocket> :
		TcpClientBase<TSocket, SmtpClientSession<TSocket>, ISmtpClientCallback<TSocket>>
			where TSocket : ITcpClientSocket, new()
	{
		#region SmtpClientBase<TSocket>

		private void RaiseError(
			SmtpClientSession<TSocket> session,
			string text,
			int code)
		{
			base.Callback.OnError(
				session,
				session.States.First.Value,
				text,
				code);
		}

		private void GetRequestAndSendCommand(SmtpClientSession<TSocket> session)
		{
			SmtpClientRequest request;

			base.Callback.OnRequest(
				session,
				out request);

			switch (request.Type)
			{
				case SmtpClientRequestType.Send:
					{
						session.Addresses = request.DestinationAddresses.GetEnumerator();
						TextProtocol.Write<TSocket>(session, "MAIL FROM:" + request.SourceAddress + "\r\n");
						session.States.AddLast(SmtpClientState.Mail);
					}
					break;
				case SmtpClientRequestType.Reset:
					TextProtocol.Write<TSocket>(session, "RSET\r\n");
					session.States.AddLast(SmtpClientState.Rset);
					break;
				case SmtpClientRequestType.Verify:
					{
						TextProtocol.Write<TSocket>(session, "VRFY " + request.SourceAddress + "\r\n");
						session.States.AddLast(SmtpClientState.Vrfy);
					}
					break;
				case SmtpClientRequestType.Expand:
					{
						TextProtocol.Write<TSocket>(session, "EXPN " + request.SourceAddress + "\r\n");
						session.States.AddLast(SmtpClientState.Expn);
					}
					break;
				case SmtpClientRequestType.Quit:
					TextProtocol.Write<TSocket>(session, "QUIT\r\n");
					session.States.AddLast(SmtpClientState.Quit);
					break;
			}
		}

		public SmtpClientBase(ISmtpClientCallback<TSocket> callback)
			: base(callback)
		{
		}

		#endregion

		#region TcpClientBase<TSocket>

		protected override SmtpClientSession<TSocket> CreateSession(object sessionUserData)
		{
			SmtpClientSession<TSocket> session = new SmtpClientSession<TSocket>(sessionUserData);

			session.States.AddLast(SmtpClientState.None);

			string hostname;
			string username;
			string password;

			base.Callback.OnCredentials(
				session,
				out hostname,
				out username,
				out password);

			session.Hostname = hostname;
			session.Username = username;
			session.Password = password;

			return session;
		}

		protected override void OnSend(SmtpClientSession<TSocket> session)
		{
			switch (session.States.Last.Value)
			{
				#region None

				case SmtpClientState.None:
					break;

				#endregion

				#region Ehlo

				case SmtpClientState.Ehlo:
					break;

				#endregion

				#region Helo

				case SmtpClientState.Helo:
					break;

				#endregion

				#region StartTLS

				case SmtpClientState.StartTLS:
					break;

				#endregion

				#region AuthLogin

				case SmtpClientState.AuthLogin:
					if (session.GetOptionSupported(SmtpClientOptions.ModePipelining))
					{
						TextProtocol.Write<TSocket>(
							session,
							Convert.ToBase64String(
								Encoding.UTF8.GetBytes(session.Username)) +
							"\r\n");
						session.States.AddLast(SmtpClientState.AuthLoginUsername);
					}
					break;

				#endregion

				#region AuthLoginUsername

				case SmtpClientState.AuthLoginUsername:
					if (session.GetOptionSupported(SmtpClientOptions.ModePipelining))
					{
						TextProtocol.Write<TSocket>(
							session,
							Convert.ToBase64String(
								Encoding.UTF8.GetBytes(session.Password)) +
							"\r\n");
						session.States.AddLast(SmtpClientState.AuthLoginPassword);
					}
					break;

				#endregion

				#region AuthLoginPassword

				case SmtpClientState.AuthLoginPassword:
					break;

				#endregion

				#region Mail

				case SmtpClientState.Mail:
					if (session.GetOptionSupported(SmtpClientOptions.ModePipelining))
					{
						if (session.Addresses.MoveNext())
						{
							TextProtocol.Write<TSocket>(session, "RCPT TO:" + session.Addresses.Current + "\r\n");
							session.States.AddLast(SmtpClientState.Rcpt);
						}
						else
						{
							TextProtocol.Write<TSocket>(session, "DATA\r\n");
							session.States.AddLast(SmtpClientState.DataStart);
						}
					}
					break;

				#endregion

				#region Rcpt

				case SmtpClientState.Rcpt:
					if (session.GetOptionSupported(SmtpClientOptions.ModePipelining))
					{
						if (session.Addresses.MoveNext())
						{
							TextProtocol.Write<TSocket>(session, "RCPT TO:" + session.Addresses.Current + "\r\n");
							session.States.AddLast(SmtpClientState.Rcpt);
						}
						else
						{
							TextProtocol.Write<TSocket>(session, "DATA\r\n");
							session.States.AddLast(SmtpClientState.DataStart);
						}
					}
					break;

				#endregion

				#region DataStart

				case SmtpClientState.DataStart:
					break;

				#endregion

				#region Data

				case SmtpClientState.Data:
					break;

				#endregion

				#region DataContinue

				case SmtpClientState.DataContinue:
					break;

				#endregion

				#region Rset

				case SmtpClientState.Rset:
					break;

				#endregion

				#region Vrfy

				case SmtpClientState.Vrfy:
					break;

				#endregion

				#region Expn

				case SmtpClientState.Expn:
					break;

				#endregion

				#region Quit

				case SmtpClientState.Quit:
					break;

				#endregion
			}
		}

		protected override void OnReceive(SmtpClientSession<TSocket> session)
		{
			switch (session.States.First.Value)
			{
				#region None

				case SmtpClientState.None:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							switch (session.MultiLineParser.Code)
							{
								case 220: // <domain> Service ready
									TextProtocol.Write<TSocket>(session, "EHLO localhost\r\n");
									session.States.AddLast(SmtpClientState.Ehlo);
									session.States.RemoveFirst();
									break;
								case 554: // Transaction failed
								default:
									RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
									break;
							}

							session.MultiLineParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Ehlo

				case SmtpClientState.Ehlo:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							switch (session.MultiLineParser.Code)
							{
								case 250: // Requested mail action okay, completed
									session.ResetOptions();

									foreach (string line in session.MultiLineParser.Lines)
									{
										if (string.Compare(line, "STARTTLS", StringComparison.OrdinalIgnoreCase) == 0)
										{
											session.SetOptionSupported(SmtpClientOptions.CommandStartTLS);
										}
										else if (line.StartsWith("AUTH ", StringComparison.OrdinalIgnoreCase))
										{
											if (line.IndexOf("LOGIN", StringComparison.OrdinalIgnoreCase) != -1)
											{
												session.SetOptionSupported(SmtpClientOptions.AuthenticationLogin);
											}
										}
										else if (string.Compare(line, "PIPELINING", StringComparison.OrdinalIgnoreCase) == 0)
										{
											session.SetOptionSupported(SmtpClientOptions.ModePipelining);
										}
										else if (string.Compare(line, "8BITMIME", StringComparison.OrdinalIgnoreCase) == 0)
										{
											session.SetOptionSupported(SmtpClientOptions.ModeMime8Bit);
										}
										else if (string.Compare(line, "ENHANCEDSTATUSCODES", StringComparison.OrdinalIgnoreCase) == 0)
										{
											session.SetOptionSupported(SmtpClientOptions.OptionEnhancedStatusCodes);
										}
										else if (string.Compare(line, "DSN", StringComparison.OrdinalIgnoreCase) == 0)
										{
											session.SetOptionSupported(SmtpClientOptions.OptionDSN);
										}
										else if (line.StartsWith("SIZE ", StringComparison.OrdinalIgnoreCase))
										{
											session.SetOptionSupported(SmtpClientOptions.OptionMaxMessageSize);
											session.OptionMaxMessageSize = TextProtocol.ParseInteger(line.Substring(5));
										}
									}

									if (session.GetOptionSupported(SmtpClientOptions.CommandStartTLS) && (!session.Socket.IsEncrypted))
									{
										TextProtocol.Write<TSocket>(session, "STARTTLS\r\n");
										session.States.AddLast(SmtpClientState.StartTLS);
										session.States.RemoveFirst();
									}
									else if (session.GetOptionSupported(SmtpClientOptions.AuthenticationLogin))
									{
										TextProtocol.Write<TSocket>(session, "AUTH LOGIN\r\n");
										session.States.AddLast(SmtpClientState.AuthLogin);
										session.States.RemoveFirst();
									}
									else
									{
										GetRequestAndSendCommand(session);
										session.States.RemoveFirst();
									}
									break;
								case 500: // Command not recognized: EHLO
									TextProtocol.Write<TSocket>(session, "HELO\r\n");
									session.States.AddLast(SmtpClientState.Helo);
									session.States.RemoveFirst();
									break;
								case 504: // Command parameter not implemented
								case 550: // Requested mail action not taken: mailbox unavailable
								default:
									RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
									break;
							}

							session.MultiLineParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Helo

				case SmtpClientState.Helo:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							switch (session.MultiLineParser.Code)
							{
								case 250: // Requested mail action okay, completed
									GetRequestAndSendCommand(session);
									session.States.RemoveFirst();
									break;
								case 504: // Command parameter not implemented
								case 550: // Requested mail action not taken: mailbox unavailable
								default:
									RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
									break;
							}

							session.MultiLineParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region StartTLS

				case SmtpClientState.StartTLS:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							switch (session.MultiLineParser.Code)
							{
								case 220: // Go ahead
									base.SslAuthenticate(session, session.Hostname);
									TextProtocol.Write<TSocket>(session, "EHLO localhost\r\n");
									session.States.AddLast(SmtpClientState.Ehlo);
									session.States.RemoveFirst();
									break;
								default:
									RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
									break;
							}

							session.MultiLineParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region AuthLogin

				case SmtpClientState.AuthLogin:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							switch (session.MultiLineParser.Code)
							{
								case 334: // VXNlcm5hbWU6
									if (!session.GetOptionSupported(SmtpClientOptions.ModePipelining))
									{
										TextProtocol.Write<TSocket>(
											session,
											Convert.ToBase64String(
												Encoding.UTF8.GetBytes(session.Username)) +
											"\r\n");
										session.States.AddLast(SmtpClientState.AuthLoginUsername);
									}

									session.States.RemoveFirst();
									break;
								default:
									RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
									break;
							}

							session.MultiLineParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region AuthLoginUsername

				case SmtpClientState.AuthLoginUsername:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							switch (session.MultiLineParser.Code)
							{
								case 334: // UGFzc3dvcmQ6
									if (!session.GetOptionSupported(SmtpClientOptions.ModePipelining))
									{
										TextProtocol.Write<TSocket>(
											session,
											Convert.ToBase64String(
												Encoding.UTF8.GetBytes(session.Password)) +
												"\r\n");
										session.States.AddLast(SmtpClientState.AuthLoginUsername);
									}

									session.States.RemoveFirst();
									break;
								default:
									RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
									break;
							}

							session.MultiLineParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region AuthLoginPassword

				case SmtpClientState.AuthLoginPassword:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							switch (session.MultiLineParser.Code)
							{
								case 235:
									GetRequestAndSendCommand(session);
									session.States.RemoveFirst();
									break;
								default:
									RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
									break;
							}

							session.MultiLineParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Mail

				case SmtpClientState.Mail:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							switch (session.MultiLineParser.Code)
							{
								case 250: // Requested mail action okay, completed
									if (!session.GetOptionSupported(SmtpClientOptions.ModePipelining))
									{
										if (session.Addresses.MoveNext())
										{
											TextProtocol.Write<TSocket>(session, "RCPT TO:" + session.Addresses.Current + "\r\n");
											session.States.AddLast(SmtpClientState.Rcpt);
										}
										else
										{
											TextProtocol.Write<TSocket>(session, "DATA\r\n");
											session.States.AddLast(SmtpClientState.DataStart);
										}
									}

									session.States.RemoveFirst();
									break;
								case 451: // Requested action aborted: error in processing
								case 452: // Requested action not taken: insufficient system storage
								case 503: // Bad sequence of commands
								case 550: // Requested mail action not taken: mailbox unavailable
								case 552: // Requested mail action aborted: exceeded storage allocation
								case 553: // Requested action not taken: mailbox name not allowed
								default:
									RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
									break;
							}

							session.MultiLineParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Rcpt

				case SmtpClientState.Rcpt:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							switch (session.MultiLineParser.Code)
							{
								case 250: // Requested mail action okay, completed
								case 251: // User not local; will forward to <forward-path>
									if (!session.GetOptionSupported(SmtpClientOptions.ModePipelining))
									{
										if (session.Addresses.MoveNext())
										{
											TextProtocol.Write<TSocket>(session, "RCPT TO:" + session.Addresses.Current + "\r\n");
											session.States.AddLast(SmtpClientState.Rcpt);
										}
										else
										{
											TextProtocol.Write<TSocket>(session, "DATA\r\n");
											session.States.AddLast(SmtpClientState.DataStart);
										}
									}

									session.States.RemoveFirst();
									break;
								case 450: // Requested mail action not taken: mailbox unavailable
								case 451: // Requested action aborted: error in processing
								case 452: // Requested action not taken: insufficient system storage
								case 503: // Bad sequence of commands
								case 550: // Requested mail action not taken: mailbox unavailable
								case 551: // User not local; please try <forward-path>
								case 552: // Requested mail action aborted: exceeded storage allocation
								case 553: // Requested action not taken: mailbox name not allowed
								default:
									RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
									break;
							}

							session.MultiLineParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region DataStart

				case SmtpClientState.DataStart:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							switch (session.MultiLineParser.Code)
							{
								//case 250: // Requested mail action okay, completed
								//	break;
								case 354: // Start mail input; end with <CRLF>.<CRLF>
									{
										ArraySegment<byte> segment;
										bool isLast;

										base.Callback.OnMessage(
											session,
											out segment,
											out isLast);

										session.SendStream.Write(
											segment.Array,
											segment.Offset,
											segment.Count);

										if (isLast)
										{
											session.States.AddLast(SmtpClientState.DataContinue);
											session.States.RemoveFirst();
										}
										else
										{
											session.States.AddLast(SmtpClientState.Data);
											session.States.RemoveFirst();
										}
									}
									break;
								case 451: // Requested action aborted: error in processing
								case 452: // Requested action not taken: insufficient system storage
								case 503: // Bad sequence of commands
								case 552: // Requested mail action aborted: exceeded storage allocation
								case 554: // Transaction failed
								default:
									RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
									break;
							}

							session.MultiLineParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Data

				case SmtpClientState.Data:
					{
						ArraySegment<byte> segment;
						bool isLast;

						base.Callback.OnMessage(
							session,
							out segment,
							out isLast);

						session.SendStream.Write(
							segment.Array,
							segment.Offset,
							segment.Count);

						if (isLast)
						{
							session.States.AddLast(SmtpClientState.DataContinue);
							session.States.RemoveFirst();
						}
					}
					break;

				#endregion

				#region DataContinue

				case SmtpClientState.DataContinue:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							switch (session.MultiLineParser.Code)
							{
								case 250: // Requested mail action okay, completed
									GetRequestAndSendCommand(session);
									session.States.RemoveFirst();
									break;
								//case 354: // Start mail input; end with <CRLF>.<CRLF>
								//	break;
								case 451: // Requested action aborted: error in processing
								case 452: // Requested action not taken: insufficient system storage
								case 503: // Bad sequence of commands
								case 552: // Requested mail action aborted: exceeded storage allocation
								case 554: // Transaction failed
								default:
									RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
									break;
							}

							session.MultiLineParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Rset

				case SmtpClientState.Rset:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							switch (session.MultiLineParser.Code)
							{
								case 250: // Requested mail action okay, completed
									GetRequestAndSendCommand(session);
									session.States.RemoveFirst();
									break;
								default:
									RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
									break;
							}

							session.MultiLineParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Vrfy

				case SmtpClientState.Vrfy:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							switch (session.MultiLineParser.Code)
							{
								case 250: // Requested mail action okay, completed
									base.Callback.OnVerify(
										session,
										true,
										true);
									GetRequestAndSendCommand(session);
									session.States.RemoveFirst();
									break;
								case 251: // User not local; will forward to <forward-path>
									base.Callback.OnVerify(
										session,
										true,
										false);
									GetRequestAndSendCommand(session);
									session.States.RemoveFirst();
									break;
								case 252: // Cannot VRFY user, but will accept message and attempt
									base.Callback.OnVerify(
										session,
										false,
										false);
									GetRequestAndSendCommand(session);
									session.States.RemoveFirst();
									break;
								case 502: // Command not implemented
								case 504: // Command parameter not implemented
								case 550: // Requested mail action not taken: mailbox unavailable
								case 551: // User not local; please try <forward-path>
								case 553: // Requested action not taken: mailbox name not allowed
								default:
									RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
									break;
							}

							session.MultiLineParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
						session.IsClosing = true;
					}
					break;

				#endregion

				#region Expn

				case SmtpClientState.Expn:
					break;

				#endregion

				#region Quit

				case SmtpClientState.Quit:
					session.MultiLineParser.Parse(session);

					if (session.MultiLineParser.IsValid)
					{
						if (session.MultiLineParser.IsComplete)
						{
							switch (session.MultiLineParser.Code)
							{
								case 221: // <domain> Service closing transmission channel
									session.IsClosing = true;
									break;
								default:
									RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
									break;
							}

							session.MultiLineParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.MultiLineParser.FirstLine, session.MultiLineParser.Code);
						session.IsClosing = true;
					}
					break;

				#endregion
			}
		}

		#endregion
	}

	public sealed class SmtpClient : SmtpClientBase<TcpClientSocket>
	{
		public SmtpClient(ISmtpClientCallback<TcpClientSocket> callback)
			: base(callback)
		{
		}
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
