﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Net Library                              */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System.Diagnostics;
using System.Collections.Generic;
using System.Text;

namespace Nabu.Net
{
	public sealed class SmtpClientSession<TSocket> : NetworkSessionBase<TSocket>
		where TSocket : ITcpClientSocket, new()
	{
		#region Types

		internal sealed class MultiLineResponseParser
		{
			#region Types

			public enum State
			{
				Default,
				Invalid,
				Digit1,
				Digit2,
				Digit3,
				Line,
				LineCr,
				LastLine,
				LastLineCr,
				LastLineCrLf,
			}

			#endregion

			#region Fields

			private State _state = State.Default;
			private StringBuilder _line = new StringBuilder();
			private List<string> _lines = new List<string>();
			private int _code = -1;
			private int _codeLine;

			#endregion

			#region Methods

			[DebuggerStepThrough]
			public MultiLineResponseParser()
			{
			}

			[DebuggerStepThrough]
			public void Reset()
			{
				this._state = State.Default;
				this._line.Length = 0;
				this._lines.Clear();
				this._code = -1;
				this._codeLine = 0;
			}

			public void Parse(SmtpClientSession<TSocket> session)
			{
				while (session.ReceiveSize > 0)
				{
					char symbol = (char)session.ReceiveStream.ReadByte();

					switch (this._state)
					{
						case State.Default:
							switch (symbol)
							{
								case '0':
								case '1':
								case '2':
								case '3':
								case '4':
								case '5':
								case '6':
								case '7':
								case '8':
								case '9':
									this._state = State.Digit1;
									this._codeLine = 100 * (symbol - (byte)'0');
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.Invalid:
							break;
						case State.Digit1:
							switch (symbol)
							{
								case '0':
								case '1':
								case '2':
								case '3':
								case '4':
								case '5':
								case '6':
								case '7':
								case '8':
								case '9':
									this._state = State.Digit2;
									this._codeLine += 10 * (symbol - (byte)'0');
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.Digit2:
							switch (symbol)
							{
								case '0':
								case '1':
								case '2':
								case '3':
								case '4':
								case '5':
								case '6':
								case '7':
								case '8':
								case '9':
									this._state = State.Digit3;
									this._codeLine += (symbol - (byte)'0');
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.Digit3:
							switch (symbol)
							{
								case '-':
									if ((this._code != -1) && (this._code != this._codeLine))
									{
										this._state = State.Invalid;

										return;
									}
									else
									{
										this._state = State.Line;
										this._code = this._codeLine;
									}
									break;
								case ' ':
									if ((this._code != -1) && (this._code != this._codeLine))
									{
										this._state = State.Invalid;

										return;
									}
									else
									{
										this._state = State.LastLine;
										this._code = this._codeLine;
									}
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.Line:
							switch (symbol)
							{
								case '\r':
									this._state = State.LineCr;
									break;
								default:
									this._line.Append(symbol);
									break;
							}
							break;
						case State.LineCr:
							switch (symbol)
							{
								case '\n':
									this._lines.Add(this._line.ToString());
									this._line.Length = 0;
									this._state = State.Default;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.LastLine:
							switch (symbol)
							{
								case '\r':
									this._state = State.LastLineCr;
									break;
								default:
									this._line.Append(symbol);
									break;
							}
							break;
						case State.LastLineCr:
							switch (symbol)
							{
								case '\n':
									this._lines.Add(this._line.ToString());
									this._line.Length = 0;
									this._state = State.LastLineCrLf;

									return;
								default:
									this._state = State.Invalid;

									return;
							}
						case State.LastLineCrLf:
							this._state = State.Invalid;

							return;
					}
				}
			}

			#endregion

			#region Properties

			public bool IsValid
			{
				[DebuggerStepThrough]
				get { return this._state != State.Invalid; }
			}

			public bool IsComplete
			{
				[DebuggerStepThrough]
				get { return this._state == State.LastLineCrLf; }
			}

			public int Code
			{
				[DebuggerStepThrough]
				get { return this._code; }
			}

			public string FirstLine
			{
				[DebuggerStepThrough]
				get { return (this._lines.Count > 0) ? this._lines[0] : string.Empty; }
			}

			public IList<string> Lines
			{
				[DebuggerStepThrough]
				get { return this._lines; }
			}

			#endregion
		}

		#endregion

		#region Fields

		private LinkedList<SmtpClientState> _states = new LinkedList<SmtpClientState>();
		private MultiLineResponseParser _multiLineParser = new MultiLineResponseParser();
		private IEnumerator<string> _addresses;
		private string _hostname;
		private string _username;
		private string _password;
		private SmtpClientOptions _options;
		private int _optionMaxMessageSize;
		private object _customData;

		#endregion

		#region Methods

		public SmtpClientSession(object customData)
		{
			this._customData = customData;
		}

		internal void ResetOptions()
		{
			this._options = 0;
		}

		internal void SetOptionSupported(SmtpClientOptions option)
		{
			this._options |= option;
		}

		public bool GetOptionSupported(SmtpClientOptions option)
		{
			return (this._options & option) != 0;
		}

		public void SetOptionUnsupported(SmtpClientOptions option)
		{
			this._options &= ~option;
		}

		public int GetOptionValue(SmtpClientOptions option)
		{
			switch (option)
			{
				case SmtpClientOptions.OptionMaxMessageSize:
					return this._optionMaxMessageSize;
			}

			return 0;
		}

		#endregion

		#region Properties

		internal LinkedList<SmtpClientState> States
		{
			[DebuggerStepThrough]
			get { return this._states; }
			[DebuggerStepThrough]
			set { this._states = value; }
		}

		internal MultiLineResponseParser MultiLineParser
		{
			[DebuggerStepThrough]
			get { return this._multiLineParser; }
		}

		internal IEnumerator<string> Addresses
		{
			[DebuggerStepThrough]
			get { return this._addresses; }
			[DebuggerStepThrough]
			set { this._addresses = value; }
		}

		internal string Hostname
		{
			[DebuggerStepThrough]
			get { return this._hostname; }
			[DebuggerStepThrough]
			set { this._hostname = value; }
		}

		internal string Username
		{
			[DebuggerStepThrough]
			get { return this._username; }
			[DebuggerStepThrough]
			set { this._username = value; }
		}

		internal string Password
		{
			[DebuggerStepThrough]
			get { return this._password; }
			[DebuggerStepThrough]
			set { this._password = value; }
		}

		internal int OptionMaxMessageSize
		{
			[DebuggerStepThrough]
			get { return this._optionMaxMessageSize; }
			[DebuggerStepThrough]
			set { this._optionMaxMessageSize = value; }
		}

		public object CustomData
		{
			[DebuggerStepThrough]
			get { return this._customData; }
			[DebuggerStepThrough]
			set { this._customData = value; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
