﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Net Library                              */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;
using System.Text;

namespace Nabu.Net
{
	public class SmtpServerBase<TSocketListen, TSocketIO> :
		TcpServerBase<TSocketListen, TSocketIO, SmtpServerSession<TSocketIO>, ISmtpServerCallback<TSocketIO>>
		where TSocketListen : ITcpServerSocketListen, new()
		where TSocketIO : ITcpServerSocketIO
	{
		#region SmtpServerBase

		#region Constants

		private const string PositiveNOOP = "200 2.0.0 OK\r\n";
		private const string PositiveStartTLS = "220 2.0.0 Ready to start TLS\r\n";
		private const string PositiveAuthLoginUser = "334 VXNlcm5hbWU6\r\n";
		private const string PositiveAuthLoginPass = "334 UGFzc3dvcmQ6\r\n";
		private const string PositiveAuth = "235 2.0.0 Authentication successful\r\n";
		private const string PositiveMailFrom = "250 2.1.0 Sender accepted\r\n";
		private const string PositiveRcptTo = "250 2.1.0 Recipient accepted\r\n";
		private const string PositiveDataStart = "354 Enter mail, end with CRLF.CRLF\r\n";
		private const string PositiveDataEnd = "250 2.0.0 Message accepted for delivery\r\n";
		private const string PositiveQuit = "221 2.0.0 SMTP closing connection\r\n";

		private const string NegativeCommand = "500 5.5.1 Invalid syntax, command unrecognized\r\n";
		private const string NegativeCommandSequence = "503 5.5.1 Invalid command sequence\r\n";
		private const string NegativeParameterExpected = "500 5.5.1 Invalid syntax, parameter expected\r\n";
		private const string NegativeParameterUnexpected = "500 5.5.1 Invalid syntax, parameter unexpected\r\n";
		private const string NegativeAuthenticationMethod = "504 5.7.6 Invalid authentication method\r\n";
		private const string NegativeAuth = "535 5.7.0 Authentication failed\r\n";
		private const string NegativeBase64 = "500 5.5.1 Syntax error, invalid BASE64 data\r\n";

		#endregion

		#region Fields

		private string _name;

		#endregion

		#region Methods

		private void RaiseError(
			SmtpServerSession<TSocketIO> session,
			string text)
		{
			base.Callback.OnError(
				session,
				session.States.First.Value,
				text);
			session.IsClosing = true;
		}

		public SmtpServerBase(
			ISmtpServerCallback<TSocketIO> callback,
			string name)
			: base(callback)
		{
			this._name = name;
		}

		#endregion

		#endregion

		#region TcpServerBase

		protected override SmtpServerSession<TSocketIO> CreateSession(object customData)
		{
			SmtpServerSession<TSocketIO> session = new SmtpServerSession<TSocketIO>();

			session.States.AddLast(SmtpServerState.Default);
			TextProtocol.Write(session, "220 " + this._name + " ESTMP ready\r\n");

			return session;
		}

		protected override void OnReceive(SmtpServerSession<TSocketIO> session)
		{
			switch (session.States.Last.Value)
			{
				#region Default

				case SmtpServerState.Default:
					session.CommandParser.Parse(session);

					if (session.CommandParser.IsValid)
					{
						if (session.CommandParser.IsComplete)
						{
							if (session.CommandParser.IsSuccessful)
							{
								switch (session.CommandParser.Command)
								{
									case "NOOP":
										if (session.CommandParser.Argument != null)
										{
											TextProtocol.Write(session, NegativeParameterUnexpected);
										}
										else
										{
											TextProtocol.Write(session, PositiveNOOP);
										}
										break;
									case "STARTTLS":
										if (session.CommandParser.Argument != null)
										{
											TextProtocol.Write(session, NegativeParameterUnexpected);
										}
										else
										{
											TextProtocol.Write(session, PositiveStartTLS);
											session.States.AddLast(SmtpServerState.StartTLS);
											session.States.RemoveFirst();
										}
										break;
									case "HELO":
										TextProtocol.Write(session, "250 " + this._name + "\r\n");
										break;
									case "EHLO":
										if (session.CommandParser.Argument == null)
										{
											TextProtocol.Write(session, NegativeParameterExpected);
										}
										else
										{
											session.Hostname = session.CommandParser.Argument;

											TextProtocol.Write(session, "250-" + this._name + "\r\n");
											TextProtocol.Write(session, "250-8BITMIME\r\n");
											TextProtocol.Write(session, "250-AUTH LOGIN\r\n");
											TextProtocol.Write(session, "250-DSN\r\n");
											TextProtocol.Write(session, "250-ENHANCEDSTATUSCODES\r\n");
											TextProtocol.Write(session, "250-PIPELINING\r\n");

											if (!session.Socket.IsEncrypted)
											{
												TextProtocol.Write(session, "250-STARTTLS\r\n");
											}

											TextProtocol.Write(session, "250 HELP\r\n");
										}
										break;
									case "AUTH":
										if (session.CommandParser.Argument == null)
										{
											TextProtocol.Write(session, NegativeParameterExpected);
										}
										else
										{
											if (session.Username != null)
											{
												TextProtocol.Write(session, NegativeCommandSequence);
											}
											else
											{
												switch (session.CommandParser.Argument)
												{
													case "LOGIN":
														TextProtocol.Write(session, PositiveAuthLoginUser);
														session.States.AddLast(SmtpServerState.AuthLoginUsername);
														session.States.RemoveFirst();
														break;
													default:
														TextProtocol.Write(session, NegativeAuthenticationMethod);
														break;
												}
											}
										}
										break;
									case "MAIL FROM":
										if (session.CommandParser.Argument == null)
										{
											TextProtocol.Write(session, NegativeParameterExpected);
										}
										else
										{
											session.Senders.Add(session.CommandParser.Argument);
											TextProtocol.Write(session, PositiveMailFrom);
											session.States.AddLast(SmtpServerState.MessageAddresses);
											session.States.RemoveFirst();
										}
										break;
									case "RCPT TO":
										if (session.CommandParser.Argument == null)
										{
											TextProtocol.Write(session, NegativeParameterExpected);
										}
										else
										{
											session.Recipients.Add(session.CommandParser.Argument);
											TextProtocol.Write(session, PositiveRcptTo);
											session.States.AddLast(SmtpServerState.MessageAddresses);
											session.States.RemoveFirst();
										}
										break;
									case "DATA":
										TextProtocol.Write(session, NegativeCommandSequence);
										break;
									case "RSET":
										base.Callback.OnRset(session);
										session.Senders.Clear();
										session.Recipients.Clear();
										break;
									case "QUIT":
										base.Callback.OnQuit(session);
										TextProtocol.Write(session, PositiveQuit);
										session.States.AddLast(SmtpServerState.Quit);
										session.States.RemoveFirst();
										break;
									default:
										TextProtocol.Write(session, NegativeCommand);
										break;
								}

								session.CommandParser.Reset();
							}
							else
							{
								RaiseError(session, session.CommandParser.Command);
							}
						}
					}
					else
					{
						RaiseError(session, session.CommandParser.Command);
					}
					break;

				#endregion

				#region StartTLS

				case SmtpServerState.StartTLS:
					{
						X509CertificateCollection certificates;
						SslProtocols enabledProtocols;
						bool mutualAuthenticationRequired;
						bool checkCertificateRevokation;

						base.Callback.OnSslAuthenticate(
							session,
							out certificates,
							out enabledProtocols,
							out mutualAuthenticationRequired,
							out checkCertificateRevokation);
						session.Socket.AuthenticateAsServer(
							certificates[0],
							mutualAuthenticationRequired,
							enabledProtocols,
							checkCertificateRevokation);
						session.States.AddFirst(SmtpServerState.Default);
						session.States.RemoveLast();
					}
					break;

				#endregion

				#region AuthLoginUsername

				case SmtpServerState.AuthLoginUsername:
					session.AuthLoginParser.Parse(session);

					if (session.AuthLoginParser.IsValid)
					{
						if (session.AuthLoginParser.IsComplete)
						{
							try
							{
								session.Username =
									Encoding.ASCII.GetString(
										Convert.FromBase64String(session.AuthLoginParser.Value));
								TextProtocol.Write(session, PositiveAuthLoginPass);
								session.States.AddLast(SmtpServerState.AuthLoginPassword);
								session.States.RemoveFirst();
							}
							catch (FormatException)
							{
								TextProtocol.Write(session, NegativeBase64);
							}

							session.AuthLoginParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.CommandParser.Command);
					}
					break;

				#endregion

				#region AuthLoginPassword

				case SmtpServerState.AuthLoginPassword:
					session.AuthLoginParser.Parse(session);

					if (session.AuthLoginParser.IsValid)
					{
						if (session.AuthLoginParser.IsComplete)
						{
							try
							{
								session.Password =
									Encoding.ASCII.GetString(
										Convert.FromBase64String(session.AuthLoginParser.Value));

								bool isValid;

								base.Callback.OnCredentials(session, out isValid);

								if (isValid)
								{
									TextProtocol.Write(session, PositiveAuth);
								}
								else
								{
									session.Username = null;
									session.Password = null;
									TextProtocol.Write(session, NegativeAuth);
								}

								session.States.AddLast(SmtpServerState.Default);
								session.States.RemoveFirst();
							}
							catch (FormatException)
							{
								TextProtocol.Write(session, NegativeBase64);
							}

							session.AuthLoginParser.Reset();
						}
					}
					else
					{
						RaiseError(session, session.CommandParser.Command);
					}
					break;

				#endregion

				#region MessageAddresses

				case SmtpServerState.MessageAddresses:
					session.CommandParser.Parse(session);

					if (session.CommandParser.IsValid)
					{
						if (session.CommandParser.IsComplete)
						{
							if (session.CommandParser.IsSuccessful)
							{
								switch (session.CommandParser.Command)
								{
									case "HELO":
									case "EHLO":
									case "AUTH":
										TextProtocol.Write(session, NegativeCommandSequence);
										break;
									case "MAIL FROM":
										if (session.CommandParser.Argument == null)
										{
											TextProtocol.Write(session, NegativeParameterExpected);
										}
										else
										{
											session.Senders.Add(session.CommandParser.Argument);
											TextProtocol.Write(session, PositiveMailFrom);
										}
										break;
									case "RCPT TO":
										if (session.CommandParser.Argument == null)
										{
											TextProtocol.Write(session, NegativeParameterExpected);
										}
										else
										{
											session.Recipients.Add(session.CommandParser.Argument);
											TextProtocol.Write(session, PositiveRcptTo);
										}
										break;
									case "DATA":
										TextProtocol.Write(session, PositiveDataStart);
										session.States.AddLast(SmtpServerState.MessageData);
										session.States.RemoveFirst();
										break;
									case "RSET":
										base.Callback.OnRset(session);
										session.Senders.Clear();
										session.Recipients.Clear();
										session.States.AddLast(SmtpServerState.Default);
										session.States.RemoveFirst();
										break;
									case "QUIT":
										base.Callback.OnQuit(session);
										TextProtocol.Write(session, PositiveQuit);
										session.States.AddLast(SmtpServerState.Quit);
										session.States.RemoveFirst();
										break;
									default:
										TextProtocol.Write(session, NegativeCommand);
										break;
								}

								session.CommandParser.Reset();
							}
							else
							{
								RaiseError(session, session.CommandParser.Command);
							}
						}
					}
					else
					{
						RaiseError(session, session.CommandParser.Command);
					}
					break;

				#endregion

				#region MessageData

				case SmtpServerState.MessageData:
					session.BlockParser.Parse(session);

					if (session.BlockParser.IsValid)
					{
						if (session.BlockParser.IsComplete)
						{
							base.Callback.OnSend(
								session,
								session.Senders,
								session.Recipients,
								session.BlockParser.Data);
							TextProtocol.Write(session, PositiveDataEnd);
							session.States.AddLast(SmtpServerState.Default);
							session.States.RemoveFirst();
							session.BlockParser.Reset();
						}
					}
					else
					{
						RaiseError(session, "Invalid data");
					}
					break;

				#endregion

				#region Quit

				case SmtpServerState.Quit:
					break;

				#endregion
			}
		}

		protected override void OnSend(SmtpServerSession<TSocketIO> session)
		{
			switch (session.States.Last.Value)
			{
				#region Default

				case SmtpServerState.Default:
					break;

				#endregion

				#region STLS

				case SmtpServerState.StartTLS:
					break;

				#endregion

				#region AuthLoginUsername

				case SmtpServerState.AuthLoginUsername:
					break;

				#endregion

				#region AuthLoginPassword

				case SmtpServerState.AuthLoginPassword:
					break;

				#endregion

				#region MessageAddresses

				case SmtpServerState.MessageAddresses:
					break;

				#endregion

				#region MessageData

				case SmtpServerState.MessageData:
					break;

				#endregion

				#region Quit

				case SmtpServerState.Quit:
					session.IsClosing = true;
					break;

				#endregion
			}
		}

		#endregion
	}

	public sealed class SmtpServer : SmtpServerBase<TcpServerSocketListen, TcpServerSocketIO>
	{
		public SmtpServer(
			ISmtpServerCallback<TcpServerSocketIO> callback,
			string name)
			: base(callback, name)
		{
		}
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
