﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Net Library                              */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Text;

using Nabu.Collections;

namespace Nabu.Net
{
	public sealed class SmtpServerSession<TSocket> : NetworkSessionBase<TSocket>
		where TSocket : ITcpServerSocketIO
	{
		#region Types

		internal sealed class CommandRequestParser
		{
			#region Types

			private enum State
			{
				Default,
				Invalid,
				Command,
				CommandDelimiter,
				Argument,
				Cr,
				CrLf
			}

			#endregion

			#region Fields

			private State _state = State.Default;
			private StringBuilder _line = new StringBuilder(1024);
			private string _command = string.Empty;
			private string _argument = null;

			#endregion

			#region Methods

			public CommandRequestParser()
			{
			}

			public void Reset()
			{
				this._state = State.Default;
				this._line.Length = 0;
				this._command = string.Empty;
				this._argument = null;
			}

			public void Parse(SmtpServerSession<TSocket> session)
			{
				while (session.ReceiveSize > 0)
				{
					char symbol = (char)session.ReceiveStream.ReadByte();

					switch (this._state)
					{
						case State.Default:
							if (
								((symbol >= 'A') && (symbol <= 'Z')) ||
								((symbol >= 'a') && (symbol <= 'z')))
							{
								this._line.Append(char.ToUpper(symbol, CultureInfo.InvariantCulture));
								this._state = State.Command;
							}
							else
							{
								this._state = State.Invalid;

								return;
							}
							break;
						case State.Invalid:

							return;
						case State.Command:
							if (
								((symbol >= 'A') && (symbol <= 'Z')) ||
								((symbol >= 'a') && (symbol <= 'z')))
							{
								this._line.Append(char.ToUpper(symbol, CultureInfo.InvariantCulture));
							}
							else if (symbol == ' ')
							{
								switch (this._line.ToString())
								{
									case "HELO":
									case "EHLO":
									case "AUTH":
										this._command = this._line.ToString();
										this._argument = string.Empty;
										this._line.Length = 0;
										this._state = State.CommandDelimiter;
										break;
									default:
										this._line.Append(char.ToUpper(symbol, CultureInfo.InvariantCulture));
										break;
								}
							}
							else if (symbol == ':')
							{
								this._command = this._line.ToString();
								this._argument = string.Empty;
								this._line.Length = 0;
								this._state = State.CommandDelimiter;
							}
							else if (symbol == '\r')
							{
								this._command = this._line.ToString();
								this._line.Length = 0;
								this._state = State.Cr;
							}
							else
							{
								this._state = State.Invalid;

								return;
							}
							break;
						case State.CommandDelimiter:
							if ((symbol >= (char)0x20) && (symbol <= (char)0x7F))
							{
								this._line.Append(symbol);
								this._state = State.Argument;
							}
							else if (symbol == '\r')
							{
								this._argument = string.Empty;
								this._state = State.Cr;
							}
							else
							{
								this._state = State.Invalid;

								return;
							}
							break;
						case State.Argument:
							if (symbol >= (char)0x20)
							{
								this._line.Append(symbol);
							}
							else if (symbol == '\r')
							{
								this._argument = this._line.ToString();
								this._line.Length = 0;
								this._state = State.Cr;
							}
							else
							{
								this._state = State.Invalid;

								return;
							}
							break;
						case State.Cr:
							if (symbol == '\n')
							{
								this._state = State.CrLf;

								return;
							}
							else
							{
								this._state = State.Invalid;

								return;
							}
						case State.CrLf:
							switch (symbol)
							{
								default:
									this._state = State.Invalid;
									break;
							}
							break;
					}
				}
			}

			#endregion

			#region Properties

			public bool IsValid
			{
				[DebuggerStepThrough]
				get { return this._state != State.Invalid; }
			}

			public bool IsComplete
			{
				[DebuggerStepThrough]
				get
				{
					return this._state == State.CrLf;
				}
			}

			public bool IsSuccessful
			{
				[DebuggerStepThrough]
				get
				{
					return this._state == State.CrLf;
				}
			}

			public string Command
			{
				[DebuggerStepThrough]
				get { return this._command; }
			}

			public string Argument
			{
				[DebuggerStepThrough]
				get { return this._argument; }
			}

			#endregion
		}

		internal sealed class AuthLoginRequestParser
		{
			#region Types

			private enum State
			{
				Default,
				Invalid,
				Cr,
				CrLf
			}

			#endregion

			#region Fields

			private State _state = State.Default;
			private StringBuilder _line = new StringBuilder(64);

			#endregion

			#region Methods

			public AuthLoginRequestParser()
			{
			}

			public void Reset()
			{
				this._state = State.Default;
				this._line.Length = 0;
			}

			public void Parse(SmtpServerSession<TSocket> session)
			{
				while (session.ReceiveSize > 0)
				{
					char symbol = (char)session.ReceiveStream.ReadByte();

					switch (this._state)
					{
						case State.Default:
							if ((symbol >= (char)0x21) && (symbol <= (char)0x7E))
							{
								this._line.Append(symbol);
							}
							else if (symbol == '\r')
							{
								this._state = State.Cr;
							}
							else
							{
								this._state = State.Invalid;

								return;
							}
							break;
						case State.Invalid:

							return;
						case State.Cr:
							if (symbol == '\n')
							{
								this._state = State.CrLf;

								return;
							}
							else
							{
								this._state = State.Invalid;

								return;
							}
						case State.CrLf:
							switch (symbol)
							{
								default:
									this._state = State.Invalid;
									break;
							}
							break;
					}
				}
			}

			#endregion

			#region Properties

			public bool IsValid
			{
				[DebuggerStepThrough]
				get { return this._state != State.Invalid; }
			}

			public bool IsComplete
			{
				[DebuggerStepThrough]
				get { return this._state == State.CrLf; }
			}

			public string Value
			{
				[DebuggerStepThrough]
				get { return this._line.ToString(); }
			}

			#endregion
		}

		internal sealed class BlockRequestParser
		{
			#region Types

			private enum State
			{
				Default,
				Invalid,
				Cr,
				CrLf,
				CrLfDot,
				CrLfDotCr,
				CrLfDotCrLf
			}

			#endregion

			#region Fields

			private State _state = State.Default;
			private ListEx<byte> _data = new ListEx<byte>(1024 * 1024);

			#endregion

			#region Methods

			public BlockRequestParser()
			{
			}

			public void Reset()
			{
				this._state = State.Default;
				this._data.Clear();
			}

			public void Parse(SmtpServerSession<TSocket> session)
			{
				while (session.ReceiveSize > 0)
				{
					int value = session.ReceiveStream.ReadByte();
					char symbol = (char)value;

					switch (this._state)
					{
						case State.Default:
							switch (symbol)
							{
								case '\r':
									this._data.Add((byte)value);
									this._state = State.Cr;
									break;
								default:
									this._data.Add((byte)value);
									break;
							}
							break;
						case State.Invalid:

							return;
						case State.Cr:
							switch (symbol)
							{
								case '\n':
									this._data.Add((byte)value);
									this._state = State.CrLf;
									break;
								default:
									this._state = State.Invalid;

									return;
							}
							break;
						case State.CrLf:
							switch (symbol)
							{
								case '\r':
									this._data.Add((byte)value);
									this._state = State.Cr;
									break;
								case '.':
									this._data.Add((byte)value);
									this._state = State.CrLfDot;
									break;
								default:
									this._data.Add((byte)value);
									this._state = State.Default;
									break;
							}
							break;
						case State.CrLfDot:
							switch (symbol)
							{
								case '\r':
									this._state = State.CrLfDotCr;
									this._data.RemoveAt(this._data.Count - 1);
									break;
								default:
									this._data.Add((byte)value);
									this._state = State.Default;
									break;
							}
							break;
						case State.CrLfDotCr:
							switch (symbol)
							{
								case '\n':
									this._state = State.CrLfDotCrLf;

									return;
								default:
									this._state = State.Invalid;

									return;
							}
						case State.CrLfDotCrLf:
							this._state = State.Invalid;

							return;
					}
				}
			}

			#endregion

			#region Properties

			public bool IsValid
			{
				[DebuggerStepThrough]
				get { return this._state != State.Invalid; }
			}

			public bool IsComplete
			{
				[DebuggerStepThrough]
				get { return this._state == State.CrLfDotCrLf; }
			}

			public ArraySegment<byte> Data
			{
				[DebuggerStepThrough]
				get { return this._data.ToArraySegment(); }
			}

			#endregion
		}

		#endregion

		#region Fields

		private LinkedList<SmtpServerState> _states = new LinkedList<SmtpServerState>();
		private CommandRequestParser _commandParser = new CommandRequestParser();
		private AuthLoginRequestParser _authLoginParser = new AuthLoginRequestParser();
		private BlockRequestParser _blockParser = new BlockRequestParser();
		private List<string> _senders = new List<string>();
		private List<string> _recipients = new List<string>();
		private string _hostname;
		private string _username;
		private string _password;

		#endregion

		#region Methods

		public SmtpServerSession()
		{
		}

		#endregion

		#region Properties

		internal LinkedList<SmtpServerState> States
		{
			[DebuggerStepThrough]
			get { return this._states; }
			[DebuggerStepThrough]
			set { this._states = value; }
		}

		internal CommandRequestParser CommandParser
		{
			[DebuggerStepThrough]
			get { return this._commandParser; }
		}

		internal AuthLoginRequestParser AuthLoginParser
		{
			[DebuggerStepThrough]
			get { return this._authLoginParser; }
		}

		internal BlockRequestParser BlockParser
		{
			[DebuggerStepThrough]
			get { return this._blockParser; }
		}

		internal List<string> Senders
		{
			[DebuggerStepThrough]
			get { return this._senders; }
		}

		internal List<string> Recipients
		{
			[DebuggerStepThrough]
			get { return this._recipients; }
		}

		public string Hostname
		{
			[DebuggerStepThrough]
			get { return this._hostname; }
			[DebuggerStepThrough]
			set { this._hostname = value; }
		}

		public string Username
		{
			[DebuggerStepThrough]
			get { return this._username; }
			[DebuggerStepThrough]
			set { this._username = value; }
		}

		public string Password
		{
			[DebuggerStepThrough]
			get { return this._password; }
			[DebuggerStepThrough]
			set { this._password = value; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
