﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Ole Library                              */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System.IO;

using Nabu;

namespace System.Runtime.InteropServices.ComTypes
{
	public sealed class OleStreamWrapper :
		DisposableBase,
		IStream
	{
		#region OleStreamWrapper

		#region Constants

		private const string CommitExceptionMessage = "Commit is not implemented.";
		private const string RevertExceptionMessage = "Revert is not implemented.";
		private const string LockRegionExceptionMessage = "LockRegion is not implemented.";
		private const string UnlockRegionExceptionMessage = "UnlockRegion is not implemented.";

		#endregion

		#region Fields

		private Stream _stream;
		private FileStream _streamFile;
		private bool _streamOwner;

		#endregion

		#region Methods

		public OleStreamWrapper(Stream stream)
		{
			if (stream == null)
			{
				throw new ArgumentNullException("stream");
			}

			this._stream = stream;
			this._streamFile = this._stream as FileStream;
		}

		public OleStreamWrapper(Stream stream, bool streamOwner)
		{
			if (stream == null)
			{
				throw new ArgumentNullException("stream");
			}

			this._stream = stream;
			this._streamFile = this._stream as FileStream;
			this._streamOwner = streamOwner;
		}

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			if (this._streamOwner)
			{
				this._stream.Dispose();
				this._stream = null;
			}
		}

		#endregion

		#region IStream

		public void Clone(out IStream stream)
		{
			base.CheckDisposed();

			stream = new OleStreamWrapper(this._stream);
		}

		public void Commit(int flags)
		{
			base.CheckDisposed();

			throw new COMException(CommitExceptionMessage, (int)ComResult.E_NOTIMPL);
		}

		public void CopyTo(IStream stream, long bytesToCopy, IntPtr lpBytesRead, IntPtr lpBytesWritten)
		{
			base.CheckDisposed();

			byte[] buffer = new byte[(bytesToCopy < 1024 * 1024) ? bytesToCopy : 1024 * 1024];

			Marshal.WriteInt32(lpBytesRead, 0);
			Marshal.WriteInt32(lpBytesWritten, 0);

			using (NativeMemory bytesWrittenOnce = new NativeMemory(8))
			{
				long count = 0;

				for (long offset = 0; offset < bytesToCopy; offset += count)
				{
					count = ((offset + buffer.Length > bytesToCopy) ? (bytesToCopy - offset) : buffer.Length);
					count = this._stream.Read(buffer, 0, (int)count);

					if (count == 0)
					{
						break;
					}

					Marshal.WriteInt64(lpBytesRead, offset);

					stream.Write(buffer, (int)count, bytesWrittenOnce.Pointer);

					if (bytesWrittenOnce.ReadInt64(0) != count)
					{
						break;
					}

					Marshal.WriteInt64(lpBytesWritten, offset);
				}

				Marshal.WriteInt64(lpBytesRead, bytesToCopy);
				Marshal.WriteInt64(lpBytesWritten, bytesToCopy);
			}
		}

		public void LockRegion(long offset, long size, int lockType)
		{
			base.CheckDisposed();

			if (this._streamFile != null)
			{
				this._streamFile.Lock(offset, size);
			}
			else
			{
				throw new COMException(LockRegionExceptionMessage, (int)ComResult.E_NOTIMPL);
			}
		}

		public void Read(byte[] buffer, int bufferSize, IntPtr lpBytesRead)
		{
			base.CheckDisposed();

			Marshal.WriteInt32(lpBytesRead, this._stream.Read(buffer, 0, bufferSize));
		}

		public void Revert()
		{
			base.CheckDisposed();

			throw new COMException(RevertExceptionMessage, (int)ComResult.E_NOTIMPL);
		}

		public void Seek(long offset, int origin, IntPtr lpPosition)
		{
			base.CheckDisposed();

			switch (origin)
			{
				case 0:
					Marshal.WriteInt64(lpPosition, this._stream.Seek(offset, SeekOrigin.Begin));
					break;
				case 1:
					Marshal.WriteInt64(lpPosition, this._stream.Seek(offset, SeekOrigin.Current));
					break;
				case 2:
					Marshal.WriteInt64(lpPosition, this._stream.Seek(offset, SeekOrigin.End));
					break;
			}
		}

		public void SetSize(long size)
		{
			base.CheckDisposed();

			this._stream.SetLength(size);
		}

		public void Stat(out STATSTG statStg, int flags)
		{
			base.CheckDisposed();

			statStg = new STATSTG();
			statStg.pwcsName = null;

			if (flags == 0)
			{
				if (this._streamFile != null)
				{
					statStg.pwcsName = this._streamFile.Name;
				}
			}

			statStg.type = 2;
			statStg.cbSize = this._stream.Length;
			statStg.mtime = new FILETIME();
			statStg.ctime = new FILETIME();
			statStg.atime = new FILETIME();

			if (this._streamFile != null)
			{
				FileInfo fileInfo = new FileInfo(this._streamFile.Name);

				long mTime = fileInfo.LastWriteTime.ToFileTime();
				long cTime = fileInfo.CreationTime.ToFileTime();
				long aTime = fileInfo.LastAccessTime.ToFileTime();

				statStg.mtime.dwLowDateTime = (int)(mTime & 0xFFFFFFFF);
				statStg.mtime.dwHighDateTime = (int)(mTime >> 32);
				statStg.ctime.dwLowDateTime = (int)(cTime & 0xFFFFFFFF);
				statStg.ctime.dwHighDateTime = (int)(cTime >> 32);
				statStg.atime.dwLowDateTime = (int)(aTime & 0xFFFFFFFF);
				statStg.atime.dwHighDateTime = (int)(aTime >> 32);
			}

			statStg.grfMode = (int)((this._stream.CanRead ? 0x80000000L : 0) | (this._stream.CanWrite ? 0x40000000L : 0));
			statStg.grfLocksSupported = (this._streamFile != null) ? 3 : 0;
		}

		public void UnlockRegion(long offset, long size, int lockType)
		{
			base.CheckDisposed();

			if (this._streamFile != null)
			{
				this._streamFile.Unlock(offset, size);
			}
			else
			{
				throw new COMException(UnlockRegionExceptionMessage, (int)ComResult.E_NOTIMPL);
			}
		}

		public void Write(byte[] buffer, int bufferSize, IntPtr lpBytesWritten)
		{
			base.CheckDisposed();

			this._stream.Write(buffer, 0, bufferSize);
			Marshal.WriteInt32(lpBytesWritten, bufferSize);
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
