﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Reflection Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics;
using System.Reflection;
using System.Reflection.Emit;
using System.Runtime.InteropServices;

namespace Nabu.Reflection
{
	public sealed class DynamicMethodBuilder
	{
		#region Fields

		private DynamicMethod _dynamicMethod;
		private ILGenerator _ilGenerator;

		#endregion

		#region Methods

		#region Constructors

		public DynamicMethodBuilder(
			Module owner,
			string name,
			Type returnType,
			params Type[] argumentTypes)
		{
			this._dynamicMethod = new DynamicMethod(
				name,
				returnType,
				argumentTypes,
				owner);
			this._ilGenerator = this._dynamicMethod.GetILGenerator();
		}

		public DynamicMethodBuilder(
			Type owner,
			string name,
			Type returnType,
			params Type[] argumentTypes)
		{
			this._dynamicMethod = new DynamicMethod(
				name,
				returnType,
				argumentTypes,
				owner);
			this._ilGenerator = this._dynamicMethod.GetILGenerator();
		}

		public DynamicMethodBuilder(
			Module owner,
			string name,
			Type returnType,
			Type[] argumentTypes,
			bool skipVisibilityChecks,
			bool isGlobal)
		{
			if (isGlobal)
			{
				this._dynamicMethod = new DynamicMethod(
					name,
					MethodAttributes.Public | MethodAttributes.Static,
					CallingConventions.Standard,
					returnType,
					argumentTypes,
					owner,
					skipVisibilityChecks);
			}
			else
			{
				this._dynamicMethod = new DynamicMethod(
					name,
					returnType,
					argumentTypes,
					owner,
					skipVisibilityChecks);
			}

			this._ilGenerator = this._dynamicMethod.GetILGenerator();
		}

		public DynamicMethodBuilder(
			Type owner,
			string name,
			Type returnType,
			Type[] argumentTypes,
			bool skipVisibilityChecks,
			bool isGlobal)
		{
			if (isGlobal)
			{
				this._dynamicMethod = new DynamicMethod(
					name,
					MethodAttributes.Public | MethodAttributes.Static,
					CallingConventions.Standard,
					returnType,
					argumentTypes,
					owner,
					skipVisibilityChecks);
			}
			else
			{
				this._dynamicMethod = new DynamicMethod(
					name,
					returnType,
					argumentTypes,
					owner,
					skipVisibilityChecks);
			}

			this._ilGenerator = this._dynamicMethod.GetILGenerator();
		}

		#endregion

		#region Parameters

		public ParameterBuilder DefineParameter(
			int position,
			ParameterAttributes attributes,
			string name)
		{
			return this._dynamicMethod.DefineParameter(position, attributes, name);
		}

		#endregion

		#region Generation

		#region Exceptions

		public DynamicMethodBuilder BeginExceptionBlock()
		{
			this._ilGenerator.BeginExceptionBlock();

			return this;
		}

		public DynamicMethodBuilder BeginCatchBlock(Type exceptionType)
		{
			this._ilGenerator.BeginCatchBlock(exceptionType);

			return this;
		}

		public DynamicMethodBuilder BeginFinallyBlock()
		{
			this._ilGenerator.BeginFinallyBlock();

			return this;
		}

		public DynamicMethodBuilder BeginFaultBlock()
		{
			this._ilGenerator.BeginFaultBlock();

			return this;
		}

		public DynamicMethodBuilder BeginExceptFilterBlock()
		{
			this._ilGenerator.BeginExceptFilterBlock();

			return this;
		}

		public DynamicMethodBuilder EndExceptionBlock()
		{
			this._ilGenerator.EndExceptionBlock();

			return this;
		}

		public DynamicMethodBuilder Throw(Type exceptionType)
		{
			this._ilGenerator.ThrowException(exceptionType);

			return this;
		}

		public DynamicMethodBuilder ReThrow()
		{
			this._ilGenerator.Emit(OpCodes.Rethrow);

			return this;
		}

		public DynamicMethodBuilder Leave(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Leave, label);

			return this;
		}

		public DynamicMethodBuilder LeaveShort(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Leave_S, label);

			return this;
		}

		#endregion

		#region Scopes

		public DynamicMethodBuilder BeginScope()
		{
			this._ilGenerator.BeginScope();

			return this;
		}

		public DynamicMethodBuilder EndScope()
		{
			this._ilGenerator.EndScope();

			return this;
		}

		public DynamicMethodBuilder UsingNamespace(string name)
		{
			this._ilGenerator.UsingNamespace(name);

			return this;
		}

		#endregion

		#region Locals

		public LocalBuilder DeclareLocal(Type variableType)
		{
			return this._ilGenerator.DeclareLocal(variableType);
		}

		#endregion

		#region Labels

		public Label DefineLabel()
		{
			return this._ilGenerator.DefineLabel();
		}

		public Label MarkLabel()
		{
			Label label = DefineLabel();

			this._ilGenerator.MarkLabel(label);

			return label;
		}

		public DynamicMethodBuilder MarkLabel(Label label)
		{
			this._ilGenerator.MarkLabel(label);

			return this;
		}

		#endregion

		#region Operations

		#region Mathmatical

		#region Add

		public DynamicMethodBuilder Add()
		{
			this._ilGenerator.Emit(OpCodes.Add);

			return this;
		}

		public DynamicMethodBuilder AddSignedChecked()
		{
			this._ilGenerator.Emit(OpCodes.Add_Ovf);

			return this;
		}

		public DynamicMethodBuilder AddUnsignedChecked()
		{
			this._ilGenerator.Emit(OpCodes.Add_Ovf_Un);
			return this;
		}

		#endregion

		#region Sub

		public DynamicMethodBuilder Sub()
		{
			this._ilGenerator.Emit(OpCodes.Sub);

			return this;
		}

		public DynamicMethodBuilder SubSignedChecked()
		{
			this._ilGenerator.Emit(OpCodes.Sub_Ovf);

			return this;
		}

		public DynamicMethodBuilder SubUnsignedChecked()
		{
			this._ilGenerator.Emit(OpCodes.Sub_Ovf_Un);

			return this;
		}

		#endregion

		#region Mul

		public DynamicMethodBuilder Mul()
		{
			this._ilGenerator.Emit(OpCodes.Mul);

			return this;
		}

		public DynamicMethodBuilder MulSignedChecked()
		{
			this._ilGenerator.Emit(OpCodes.Mul_Ovf);

			return this;
		}

		public DynamicMethodBuilder MulUnsignedChecked()
		{
			this._ilGenerator.Emit(OpCodes.Mul_Ovf_Un);

			return this;
		}

		#endregion

		#region Div

		public DynamicMethodBuilder DivSigned()
		{
			this._ilGenerator.Emit(OpCodes.Div);

			return this;
		}

		public DynamicMethodBuilder DivUnsigned()
		{
			this._ilGenerator.Emit(OpCodes.Div_Un);

			return this;
		}

		#endregion

		#region Rem

		public DynamicMethodBuilder RemSigned()
		{
			this._ilGenerator.Emit(OpCodes.Rem);

			return this;
		}

		public DynamicMethodBuilder RemUnsigned()
		{
			this._ilGenerator.Emit(OpCodes.Rem_Un);

			return this;
		}

		#endregion

		#region Negative

		public DynamicMethodBuilder Negative()
		{
			this._ilGenerator.Emit(OpCodes.Neg);

			return this;
		}

		#endregion

		#endregion

		#region Bitwise

		public DynamicMethodBuilder And()
		{
			this._ilGenerator.Emit(OpCodes.And);

			return this;
		}

		public DynamicMethodBuilder Not()
		{
			this._ilGenerator.Emit(OpCodes.Not);

			return this;
		}

		public DynamicMethodBuilder Or()
		{
			this._ilGenerator.Emit(OpCodes.Or);

			return this;
		}

		public DynamicMethodBuilder Xor()
		{
			this._ilGenerator.Emit(OpCodes.Xor);

			return this;
		}

		public DynamicMethodBuilder Shl()
		{
			this._ilGenerator.Emit(OpCodes.Shl);

			return this;
		}

		public DynamicMethodBuilder ShrSigned()
		{
			this._ilGenerator.Emit(OpCodes.Shr);

			return this;
		}

		public DynamicMethodBuilder ShrUnsigned()
		{
			this._ilGenerator.Emit(OpCodes.Shr_Un);

			return this;
		}

		#endregion

		#region Test

		public DynamicMethodBuilder TestIfEqual()
		{
			this._ilGenerator.Emit(OpCodes.Ceq);

			return this;
		}

		public DynamicMethodBuilder TestIfGreaterSigned()
		{
			this._ilGenerator.Emit(OpCodes.Cgt);

			return this;
		}

		public DynamicMethodBuilder TestIfGreaterUnsigned()
		{
			this._ilGenerator.Emit(OpCodes.Cgt_Un);

			return this;
		}

		public DynamicMethodBuilder TestIfLessSigned()
		{
			this._ilGenerator.Emit(OpCodes.Clt);

			return this;
		}

		public DynamicMethodBuilder TestIfLessUnsigned()
		{
			this._ilGenerator.Emit(OpCodes.Clt_Un);

			return this;
		}

		public DynamicMethodBuilder TestIfFinite()
		{
			this._ilGenerator.Emit(OpCodes.Ckfinite);

			return this;
		}

		public DynamicMethodBuilder TestIfOfType(Type objectType)
		{
			this._ilGenerator.Emit(OpCodes.Isinst, objectType);

			return this;
		}

		#endregion

		#region Jump

		#region Jump

		public DynamicMethodBuilder Jump_Far(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Br, label);

			return this;
		}

		public DynamicMethodBuilder Jump_Near(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Br_S, label);

			return this;
		}

		#endregion

		#region JumpIfFalseZeroNull

		public DynamicMethodBuilder JumpIfFalseZeroNull_Far(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Brfalse, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfFalseZeroNull_Near(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Brfalse_S, label);

			return this;
		}

		#endregion

		#region JumpIfTrueNonZeroNonNull

		public DynamicMethodBuilder JumpIfTrueNonZeroNonNull_Far(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Brtrue, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfTrueNonZeroNonNull_Near(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Brtrue_S, label);

			return this;
		}

		#endregion

		#region JumpIfEqual

		public DynamicMethodBuilder JumpIfEqual_Far(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Beq, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfEqual_Near(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Beq_S, label);

			return this;
		}

		#endregion

		#region JumpIfNotEqual

		public DynamicMethodBuilder JumpIfNotEqualUnsigned_Far(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Bne_Un, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfNotEqualUnsigned_Near(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Bne_Un_S, label);

			return this;
		}

		#endregion

		#region JumpIfGreater

		public DynamicMethodBuilder JumpIfGreaterSigned_Far(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Bgt, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfGreaterSigned_Near(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Bgt_S, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfGreaterUnsigned_Far(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Bgt_Un, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfGreaterUnsigned_Near(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Bgt_Un_S, label);

			return this;
		}

		#endregion

		#region JumpIfGreaterOrEqual

		public DynamicMethodBuilder JumpIfGreaterOrEqualSigned_Far(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Bge, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfGreaterOrEqualSigned_Near(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Bge_S, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfGreaterOrEqualUnsigned_Far(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Bge_Un, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfGreaterOrEqualUnsigned_Near(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Bge_Un_S, label);

			return this;
		}

		#endregion

		#region JumpIfLess

		public DynamicMethodBuilder JumpIfLessSigned_Far(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Blt, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfLessSigned_Near(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Blt_S, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfLessUnsigned_Far(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Blt_Un, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfLessUnsigned_Near(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Blt_Un_S, label);

			return this;
		}

		#endregion

		#region JumpIfLessOrEqual

		public DynamicMethodBuilder JumpIfLessOrEqualSigned_Far(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Ble, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfLessOrEqualSigned_Near(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Ble_S, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfLessOrEqualUnsigned_Far(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Ble_Un, label);

			return this;
		}

		public DynamicMethodBuilder JumpIfLessrOrEqualUnsigned_Near(Label label)
		{
			this._ilGenerator.Emit(OpCodes.Ble_Un_S, label);

			return this;
		}

		#endregion

		#endregion

		#region Call

		#region Call

		public DynamicMethodBuilder Call(MethodInfo method)
		{
			if (method.IsStatic)
			{
				this._ilGenerator.EmitCall(OpCodes.Call, method, null);
			}
			else
			{
				this._ilGenerator.EmitCall(OpCodes.Callvirt, method, null);
			}

			return this;
		}

		public DynamicMethodBuilder Call(Type classType, string methodName, params Type[] argumentTypes)
		{
			return Call(classType.GetMethod(methodName, argumentTypes));
		}

		#endregion

		#region CallMethod

		public DynamicMethodBuilder CallMethod(MethodInfo method)
		{
			this._ilGenerator.Emit(OpCodes.Call, method);

			return this;
		}

		public DynamicMethodBuilder CallMethod(ConstructorInfo constructor)
		{
			this._ilGenerator.Emit(OpCodes.Call, constructor);

			return this;
		}

		public DynamicMethodBuilder CallMethod(Type classType, string methodName, params Type[] argumentTypes)
		{
			return CallMethod(classType.GetMethod(methodName, argumentTypes));
		}

		public DynamicMethodBuilder CallMethod(Type classType, params Type[] argumentTypes)
		{
			return CallMethod(classType.GetConstructor(argumentTypes));
		}

		#endregion

		#region CallPointer

		public DynamicMethodBuilder CallPointer(CallingConvention callingConvention, Type returnType, params Type[] argumentTypes)
		{
			this._ilGenerator.EmitCalli(OpCodes.Calli, callingConvention, returnType, argumentTypes);

			return this;
		}

		#endregion

		#region CallVirtualMethod

		public DynamicMethodBuilder CallVirtualMethod(MethodInfo method)
		{
			this._ilGenerator.Emit(OpCodes.Callvirt, method);

			return this;
		}

		public DynamicMethodBuilder CallVirtualMethod(ConstructorInfo constructor)
		{
			this._ilGenerator.Emit(OpCodes.Callvirt, constructor);

			return this;
		}

		public DynamicMethodBuilder CallVirtualMethod(Type classType, string methodName, params Type[] argumentTypes)
		{
			return CallVirtualMethod(classType.GetMethod(methodName, argumentTypes));
		}

		public DynamicMethodBuilder CallVirtualMethod(Type classType, params Type[] argumentTypes)
		{
			return CallVirtualMethod(classType.GetConstructor(argumentTypes));
		}

		#endregion

		#region CallVirtualMethodConstrained

		public DynamicMethodBuilder CallVirtualMethodConstrained(Type constrainedType, MethodInfo method)
		{
			this._ilGenerator.Emit(OpCodes.Constrained, constrainedType);
			this._ilGenerator.Emit(OpCodes.Callvirt, method);

			return this;
		}

		public DynamicMethodBuilder CallVirtualMethodConstrained(Type constrainedType, ConstructorInfo constructor)
		{
			this._ilGenerator.Emit(OpCodes.Constrained, constrainedType);
			this._ilGenerator.Emit(OpCodes.Callvirt, constructor);

			return this;
		}

		public DynamicMethodBuilder CallVirtualMethodConstrained(Type constrainedType, Type classType, string methodName, params Type[] argumentTypes)
		{
			return CallVirtualMethodConstrained(constrainedType, classType.GetMethod(methodName, argumentTypes));
		}

		public DynamicMethodBuilder CallVirtualMethodConstrained(Type constrainedType, Type classType, params Type[] argumentTypes)
		{
			return CallVirtualMethodConstrained(constrainedType, classType.GetConstructor(argumentTypes));
		}

		#endregion

		#region Jump

		public DynamicMethodBuilder JumpToMethod(MethodInfo method)
		{
			this._ilGenerator.Emit(OpCodes.Jmp, method);

			return this;
		}

		public DynamicMethodBuilder JumpToMethod(Type classType, string methodName, params Type[] argumentTypes)
		{
			this._ilGenerator.Emit(OpCodes.Jmp, classType.GetMethod(methodName, argumentTypes));

			return this;
		}

		#endregion

		#region Ret

		public DynamicMethodBuilder Ret()
		{
			this._ilGenerator.Emit(OpCodes.Ret);

			return this;
		}

		#endregion

		#endregion

		#region Boxing

		public DynamicMethodBuilder Box(Type targetType)
		{
			this._ilGenerator.Emit(OpCodes.Box, targetType);

			return this;
		}

		public DynamicMethodBuilder UnBox(Type targetType)
		{
			this._ilGenerator.Emit(OpCodes.Unbox, targetType);

			return this;
		}

		public DynamicMethodBuilder UnBoxAny(Type targetType)
		{
			this._ilGenerator.Emit(OpCodes.Unbox_Any, targetType);

			return this;
		}

		#endregion

		#region Cast

		#region CastToType

		public DynamicMethodBuilder CastToType(Type targetType)
		{
			if (targetType == typeof(sbyte))
			{
				this._ilGenerator.Emit(OpCodes.Conv_I1);
			}
			else if (targetType == typeof(sbyte))
			{
				this._ilGenerator.Emit(OpCodes.Conv_U1);
			}
			else if (targetType == typeof(short))
			{
				this._ilGenerator.Emit(OpCodes.Conv_I2);
			}
			else if (targetType == typeof(ushort))
			{
				this._ilGenerator.Emit(OpCodes.Conv_U2);
			}
			else if (targetType == typeof(int))
			{
				this._ilGenerator.Emit(OpCodes.Conv_I4);
			}
			else if (targetType == typeof(uint))
			{
				this._ilGenerator.Emit(OpCodes.Conv_U4);
			}
			else if (targetType == typeof(long))
			{
				this._ilGenerator.Emit(OpCodes.Conv_I8);
			}
			else if (targetType == typeof(ulong))
			{
				this._ilGenerator.Emit(OpCodes.Conv_U8);
			}
			else if (targetType == typeof(float))
			{
				this._ilGenerator.Emit(OpCodes.Conv_R4);
			}
			else if (targetType == typeof(double))
			{
				this._ilGenerator.Emit(OpCodes.Conv_R8);
			}
			else
			{
				this._ilGenerator.Emit(OpCodes.Castclass, targetType);
			}

			return this;
		}

		#endregion

		#region CastToTypeReference

		public DynamicMethodBuilder CastToTypeReference(Type targetType)
		{
			this._ilGenerator.Emit(OpCodes.Mkrefany, targetType);

			return this;
		}

		#endregion

		#region CastToIntChecked

		public DynamicMethodBuilder CastToSignedIntChecked()
		{
			this._ilGenerator.Emit(OpCodes.Conv_Ovf_I);

			return this;
		}

		public DynamicMethodBuilder CastToSignedInt8Checked()
		{
			this._ilGenerator.Emit(OpCodes.Conv_Ovf_I1);

			return this;
		}

		public DynamicMethodBuilder CastToSignedInt16Checked()
		{
			this._ilGenerator.Emit(OpCodes.Conv_Ovf_I2);

			return this;
		}

		public DynamicMethodBuilder CastToSignedInt32Checked()
		{
			this._ilGenerator.Emit(OpCodes.Conv_Ovf_I4);

			return this;
		}

		public DynamicMethodBuilder CastToSignedInt64Checked()
		{
			this._ilGenerator.Emit(OpCodes.Conv_Ovf_I8);

			return this;
		}

		#endregion

		#region CastToUIntChecked

		public DynamicMethodBuilder CastToSignedUIntChecked()
		{
			this._ilGenerator.Emit(OpCodes.Conv_Ovf_U);

			return this;
		}

		public DynamicMethodBuilder CastToSignedUInt8Checked()
		{
			this._ilGenerator.Emit(OpCodes.Conv_Ovf_U1);

			return this;
		}

		public DynamicMethodBuilder CastToSignedUInt16Checked()
		{
			this._ilGenerator.Emit(OpCodes.Conv_Ovf_U2);

			return this;
		}

		public DynamicMethodBuilder CastToSignedUInt32Checked()
		{
			this._ilGenerator.Emit(OpCodes.Conv_Ovf_U4);

			return this;
		}

		public DynamicMethodBuilder CastToSignedUInt64Checked()
		{
			this._ilGenerator.Emit(OpCodes.Conv_Ovf_U8);

			return this;
		}

		#endregion

		#region CastToFloat

		public DynamicMethodBuilder CastToFloat4Unsigned()
		{
			this._ilGenerator.Emit(OpCodes.Conv_R_Un);

			return this;
		}

		#endregion

		#endregion

		#region Copy

		public DynamicMethodBuilder CopyBytes()
		{
			this._ilGenerator.Emit(OpCodes.Cpblk);

			return this;
		}

		public DynamicMethodBuilder CopyObject(Type objectType)
		{
			this._ilGenerator.Emit(OpCodes.Cpobj, objectType);

			return this;
		}

		public DynamicMethodBuilder DuplicateObject()
		{
			this._ilGenerator.Emit(OpCodes.Dup);

			return this;
		}

		#endregion

		#region Initialization

		public DynamicMethodBuilder InitBytes()
		{
			this._ilGenerator.Emit(OpCodes.Initblk);

			return this;
		}

		public DynamicMethodBuilder InitObject(Type objectType)
		{
			this._ilGenerator.Emit(OpCodes.Initobj, objectType);

			return this;
		}

		#endregion

		#region Load

		#region LoadConstant

		public DynamicMethodBuilder LoadConstant(int value)
		{
			switch (value)
			{
				case -1:
					this._ilGenerator.Emit(OpCodes.Ldc_I4_M1);
					break;
				case 0:
					this._ilGenerator.Emit(OpCodes.Ldc_I4_0);
					break;
				case 1:
					this._ilGenerator.Emit(OpCodes.Ldc_I4_1);
					break;
				case 2:
					this._ilGenerator.Emit(OpCodes.Ldc_I4_2);
					break;
				case 3:
					this._ilGenerator.Emit(OpCodes.Ldc_I4_3);
					break;
				case 4:
					this._ilGenerator.Emit(OpCodes.Ldc_I4_4);
					break;
				case 5:
					this._ilGenerator.Emit(OpCodes.Ldc_I4_5);
					break;
				case 6:
					this._ilGenerator.Emit(OpCodes.Ldc_I4_6);
					break;
				case 7:
					this._ilGenerator.Emit(OpCodes.Ldc_I4_7);
					break;
				case 8:
					this._ilGenerator.Emit(OpCodes.Ldc_I4_8);
					break;
				default:
					if ((value <= byte.MaxValue) && (value >= byte.MinValue))
					{
						this._ilGenerator.Emit(OpCodes.Ldc_I4_S, (byte)value);
					}
					else
					{
						this._ilGenerator.Emit(OpCodes.Ldc_I4, value);
					}
					break;
			}

			return this;
		}

		public DynamicMethodBuilder LoadConstant(long value)
		{
			this._ilGenerator.Emit(OpCodes.Ldc_I8, value);

			return this;
		}

		public DynamicMethodBuilder LoadConstant(float value)
		{
			this._ilGenerator.Emit(OpCodes.Ldc_R4, value);

			return this;
		}

		public DynamicMethodBuilder LoadConstant(double value)
		{
			this._ilGenerator.Emit(OpCodes.Ldc_R8, value);

			return this;
		}

		public DynamicMethodBuilder LoadNull()
		{
			this._ilGenerator.Emit(OpCodes.Ldnull);

			return this;
		}

		#endregion

		#region LoadArgument

		public DynamicMethodBuilder LoadArgument(int index)
		{
			switch (index)
			{
				case 0:
					this._ilGenerator.Emit(OpCodes.Ldarg_0);
					break;
				case 1:
					this._ilGenerator.Emit(OpCodes.Ldarg_1);
					break;
				case 2:
					this._ilGenerator.Emit(OpCodes.Ldarg_2);
					break;
				case 3:
					this._ilGenerator.Emit(OpCodes.Ldarg_3);
					break;
				default:
					if (index <= byte.MaxValue)
					{
						this._ilGenerator.Emit(OpCodes.Ldarg_S, (byte)index);
					}
					else
					{
						this._ilGenerator.Emit(OpCodes.Ldarg, (short)index);
					}
					break;
			}

			return this;
		}

		#endregion

		#region LoadArgumentAddress

		public DynamicMethodBuilder LoadArgumentAddress(int index)
		{
			if (index <= byte.MaxValue)
			{
				this._ilGenerator.Emit(OpCodes.Ldarga_S, (byte)index);
			}
			else
			{
				this._ilGenerator.Emit(OpCodes.Ldarga, (short)index);
			}

			return this;
		}

		#endregion

		#region LoadLocal

		public DynamicMethodBuilder LoadLocal(int index)
		{
			switch (index)
			{
				case 0:
					this._ilGenerator.Emit(OpCodes.Ldloc_0);
					break;
				case 1:
					this._ilGenerator.Emit(OpCodes.Ldloc_1);
					break;
				case 2:
					this._ilGenerator.Emit(OpCodes.Ldloc_2);
					break;
				case 3:
					this._ilGenerator.Emit(OpCodes.Ldloc_3);
					break;
				default:
					if (index <= byte.MaxValue)
					{
						this._ilGenerator.Emit(OpCodes.Ldloc_S, (byte)index);
					}
					else
					{
						this._ilGenerator.Emit(OpCodes.Ldloc, (short)index);
					}
					break;
			}

			return this;
		}

		#endregion

		#region LoadLocalAddress

		public DynamicMethodBuilder LoadLocalAddress(int index)
		{
			if (index <= byte.MaxValue)
			{
				this._ilGenerator.Emit(OpCodes.Ldloca_S, (byte)index);
			}
			else
			{
				this._ilGenerator.Emit(OpCodes.Ldloca, (short)index);
			}

			return this;
		}

		#endregion

		#region LoadElement

		public DynamicMethodBuilder LoadElement(Type elementType)
		{
			if (elementType == typeof(sbyte))
			{
				this._ilGenerator.Emit(OpCodes.Ldelem_I1);
			}
			else if (elementType == typeof(byte))
			{
				this._ilGenerator.Emit(OpCodes.Ldelem_U1);
			}
			else if (elementType == typeof(short))
			{
				this._ilGenerator.Emit(OpCodes.Ldelem_I2);
			}
			else if (elementType == typeof(ushort))
			{
				this._ilGenerator.Emit(OpCodes.Ldelem_U2);
			}
			else if (elementType == typeof(int))
			{
				this._ilGenerator.Emit(OpCodes.Ldelem_I4);
			}
			else if (elementType == typeof(uint))
			{
				this._ilGenerator.Emit(OpCodes.Ldelem_U4);
			}
			else if (elementType == typeof(long))
			{
				this._ilGenerator.Emit(OpCodes.Ldelem_I8);
			}
			else if (elementType == typeof(float))
			{
				this._ilGenerator.Emit(OpCodes.Ldelem_R4);
			}
			else if (elementType == typeof(double))
			{
				this._ilGenerator.Emit(OpCodes.Ldelem_R8);
			}
			else if (elementType == typeof(object))
			{
				this._ilGenerator.Emit(OpCodes.Ldelem_Ref);
			}
			else
			{
				this._ilGenerator.Emit(OpCodes.Ldelem, elementType);
			}

			return this;
		}

		#endregion

		#region LoadElementAddress

		public DynamicMethodBuilder LoadElementAddress(Type elementType)
		{
			this._ilGenerator.Emit(OpCodes.Ldelema, elementType);

			return this;
		}

		#endregion

		#region LoadField

		public DynamicMethodBuilder LoadField(FieldInfo field)
		{
			this._ilGenerator.Emit(OpCodes.Ldfld, field);

			return this;
		}

		public DynamicMethodBuilder LoadField(Type classType, string fieldName)
		{
			this._ilGenerator.Emit(OpCodes.Ldfld, classType.GetField(fieldName));

			return this;
		}

		#endregion

		#region LoadFieldAddress

		public DynamicMethodBuilder LoadFieldAddress(FieldInfo field)
		{
			this._ilGenerator.Emit(OpCodes.Ldflda, field);

			return this;
		}

		public DynamicMethodBuilder LoadFieldAddress(Type classType, string fieldName)
		{
			this._ilGenerator.Emit(OpCodes.Ldflda, classType.GetField(fieldName));

			return this;
		}

		#endregion

		#region LoadStaticField

		public DynamicMethodBuilder LoadStaticField(FieldInfo field)
		{
			this._ilGenerator.Emit(OpCodes.Ldsfld, field);

			return this;
		}

		public DynamicMethodBuilder LoadStaticField(Type classType, string fieldName)
		{
			this._ilGenerator.Emit(OpCodes.Ldsfld, classType.GetField(fieldName));

			return this;
		}

		#endregion

		#region LoadStaticFieldAddress

		public DynamicMethodBuilder LoadStaticFieldAddress(FieldInfo field)
		{
			this._ilGenerator.Emit(OpCodes.Ldsflda, field);

			return this;
		}

		public DynamicMethodBuilder LoadStaticFieldAddress(Type classType, string fieldName)
		{
			this._ilGenerator.Emit(OpCodes.Ldsflda, classType.GetField(fieldName));

			return this;
		}

		#endregion

		#region LoadFromAddress

		public DynamicMethodBuilder LoadFromAddress(Type elementType)
		{
			if (elementType == typeof(sbyte))
			{
				this._ilGenerator.Emit(OpCodes.Ldind_I1);
			}
			else if (elementType == typeof(byte))
			{
				this._ilGenerator.Emit(OpCodes.Ldind_U1);
			}
			else if (elementType == typeof(short))
			{
				this._ilGenerator.Emit(OpCodes.Ldind_I2);
			}
			else if (elementType == typeof(ushort))
			{
				this._ilGenerator.Emit(OpCodes.Ldind_U2);
			}
			else if (elementType == typeof(int))
			{
				this._ilGenerator.Emit(OpCodes.Ldind_I4);
			}
			else if (elementType == typeof(uint))
			{
				this._ilGenerator.Emit(OpCodes.Ldind_U4);
			}
			else if (elementType == typeof(long))
			{
				this._ilGenerator.Emit(OpCodes.Ldelem_I8);
			}
			else if (elementType == typeof(float))
			{
				this._ilGenerator.Emit(OpCodes.Ldind_R4);
			}
			else if (elementType == typeof(double))
			{
				this._ilGenerator.Emit(OpCodes.Ldind_R8);
			}
			else
			{
				this._ilGenerator.Emit(OpCodes.Ldind_Ref);
			}

			return this;
		}

		#endregion

		#region LoadArrayLength

		public DynamicMethodBuilder LoadArrayLength()
		{
			this._ilGenerator.Emit(OpCodes.Ldlen);

			return this;
		}

		#endregion

		#region LoadObject

		public DynamicMethodBuilder LoadObject(Type objectType)
		{
			this._ilGenerator.Emit(OpCodes.Ldobj, objectType);

			return this;
		}

		#endregion

		#region LoadString

		public DynamicMethodBuilder LoadObject(string value)
		{
			this._ilGenerator.Emit(OpCodes.Ldstr, value);

			return this;
		}

		#endregion

		#region LoadToken

		public DynamicMethodBuilder LoadToken(Type type)
		{
			this._ilGenerator.Emit(OpCodes.Ldtoken, type);

			return this;
		}

		public DynamicMethodBuilder LoadToken(MethodInfo method)
		{
			this._ilGenerator.Emit(OpCodes.Ldtoken, method);

			return this;
		}

		public DynamicMethodBuilder LoadToken(FieldInfo field)
		{
			this._ilGenerator.Emit(OpCodes.Ldtoken, field);

			return this;
		}

		#endregion

		#endregion

		#region Store

		#region StoreArgument

		public DynamicMethodBuilder StoreArgument(int index)
		{
			if (index <= byte.MaxValue)
			{
				this._ilGenerator.Emit(OpCodes.Starg_S, (byte)index);
			}
			else
			{
				this._ilGenerator.Emit(OpCodes.Starg, (short)index);
			}

			return this;
		}

		#endregion

		#region StoreLocal

		public DynamicMethodBuilder StoreLocal(int index)
		{
			switch (index)
			{
				case 0:
					this._ilGenerator.Emit(OpCodes.Stloc_0);
					break;
				case 1:
					this._ilGenerator.Emit(OpCodes.Stloc_1);
					break;
				case 2:
					this._ilGenerator.Emit(OpCodes.Stloc_2);
					break;
				case 3:
					this._ilGenerator.Emit(OpCodes.Stloc_3);
					break;
				default:
					if (index <= byte.MaxValue)
					{
						this._ilGenerator.Emit(OpCodes.Stloc_S, (byte)index);
					}
					else
					{
						this._ilGenerator.Emit(OpCodes.Stloc, (short)index);
					}
					break;
			}

			return this;
		}

		#endregion

		#region StoreElement

		public DynamicMethodBuilder StoreElement(Type elementType)
		{
			if (elementType == typeof(sbyte))
			{
				this._ilGenerator.Emit(OpCodes.Stelem_I1);
			}
			else if (elementType == typeof(short))
			{
				this._ilGenerator.Emit(OpCodes.Stelem_I2);
			}
			else if (elementType == typeof(int))
			{
				this._ilGenerator.Emit(OpCodes.Stelem_I4);
			}
			else if (elementType == typeof(long))
			{
				this._ilGenerator.Emit(OpCodes.Stelem_I8);
			}
			else if (elementType == typeof(float))
			{
				this._ilGenerator.Emit(OpCodes.Stelem_R4);
			}
			else if (elementType == typeof(double))
			{
				this._ilGenerator.Emit(OpCodes.Stelem_R8);
			}
			else if (elementType == typeof(object))
			{
				this._ilGenerator.Emit(OpCodes.Stelem_Ref);
			}
			else
			{
				this._ilGenerator.Emit(OpCodes.Stelem, elementType);
			}

			return this;
		}

		#endregion

		#region StoreField

		public DynamicMethodBuilder StoreField(FieldInfo field)
		{
			this._ilGenerator.Emit(OpCodes.Stfld, field);

			return this;
		}

		public DynamicMethodBuilder StoreField(Type classType, string fieldName)
		{
			this._ilGenerator.Emit(OpCodes.Stfld, classType.GetField(fieldName));

			return this;
		}

		#endregion

		#region StoreStaticField

		public DynamicMethodBuilder StoreStaticField(FieldInfo field)
		{
			this._ilGenerator.Emit(OpCodes.Stsfld, field);

			return this;
		}

		public DynamicMethodBuilder StoreStaticField(Type classType, string fieldName)
		{
			this._ilGenerator.Emit(OpCodes.Stsfld, classType.GetField(fieldName));

			return this;
		}

		#endregion

		#region StoreToAddress

		public DynamicMethodBuilder StoreToAddress(Type elementType)
		{
			if (elementType == typeof(sbyte))
			{
				this._ilGenerator.Emit(OpCodes.Stind_I1);
			}
			else if (elementType == typeof(short))
			{
				this._ilGenerator.Emit(OpCodes.Stind_I2);
			}
			else if (elementType == typeof(int))
			{
				this._ilGenerator.Emit(OpCodes.Stind_I4);
			}
			else if (elementType == typeof(long))
			{
				this._ilGenerator.Emit(OpCodes.Stelem_I8);
			}
			else if (elementType == typeof(float))
			{
				this._ilGenerator.Emit(OpCodes.Stind_R4);
			}
			else if (elementType == typeof(double))
			{
				this._ilGenerator.Emit(OpCodes.Stind_R8);
			}
			else
			{
				this._ilGenerator.Emit(OpCodes.Stind_Ref);
			}

			return this;
		}

		#endregion

		#region StoreObject

		public DynamicMethodBuilder StoreObject(Type objectType)
		{
			this._ilGenerator.Emit(OpCodes.Stobj, objectType);

			return this;
		}

		#endregion

		#endregion

		#region Alloc

		public DynamicMethodBuilder AllocLocal()
		{
			this._ilGenerator.Emit(OpCodes.Localloc);

			return this;
		}

		public DynamicMethodBuilder AllocArray(Type elementType)
		{
			this._ilGenerator.Emit(OpCodes.Newarr, elementType);

			return this;
		}

		public DynamicMethodBuilder AllocObject(ConstructorInfo constructorInfo)
		{
			this._ilGenerator.Emit(OpCodes.Newobj, constructorInfo);

			return this;
		}

		public DynamicMethodBuilder AllocObject(Type classType, params Type[] argumentTypes)
		{
			this._ilGenerator.Emit(OpCodes.Newobj, classType.GetConstructor(argumentTypes));

			return this;
		}

		#endregion

		#region Pop

		public DynamicMethodBuilder Pop()
		{
			this._ilGenerator.Emit(OpCodes.Pop);

			return this;
		}

		#endregion

		#region SizeOf

		public DynamicMethodBuilder SizeOf(Type valueType)
		{
			this._ilGenerator.Emit(OpCodes.Sizeof, valueType);

			return this;
		}

		#endregion

		#region Debugging

		public DynamicMethodBuilder Break()
		{
			this._ilGenerator.Emit(OpCodes.Break);

			return this;
		}

		#endregion

		#endregion

		#endregion

		#region Invokation

		public Delegate CreateDelegate(Type delegateType)
		{
			return this._dynamicMethod.CreateDelegate(delegateType);
		}

		public Delegate CreateDelegate(Type delegateType, object target)
		{
			return this._dynamicMethod.CreateDelegate(delegateType, target);
		}

		public object Invoke(object instance, params object[] arguments)
		{
			return this._dynamicMethod.Invoke(instance, arguments);
		}

		#endregion

		#endregion

		#region Properties

		public DynamicMethod DynamicMethod
		{
			[DebuggerStepThrough]
			get { return this._dynamicMethod; }
		}

		public ILGenerator ILGenerator
		{
			[DebuggerStepThrough]
			get { return this._ilGenerator; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
