﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Reflection Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Globalization;
using System.Reflection;
using System.Reflection.Emit;

namespace Nabu.Reflection
{
	public sealed class ILParser
	{
		#region Constants

		private const string UnexpectedEndOfStreamExceptionMessage = "Unexpected end of IL stream at offset {0}.";
		private const string UnexpectedOpCodeExceptionMessage = "Unexpected OpCode {0} at offset {1}. IL: {2:2x} {3:2x} {4:2x} {5:2x} {6:2x} {7:2x}.";

		#endregion

		#region Fields

		private static OpCode[] _1byteOpCodes = new OpCode[0x100];
		private static OpCode[] _2byteOpCodes = new OpCode[0x100];

		private byte[] _il;
		private int _offset;

		#endregion

		#region Methods

		static ILParser()
		{
			foreach (FieldInfo fieldInfo in typeof(OpCodes).GetFields(BindingFlags.Public | BindingFlags.Static))
			{
				OpCode opCode = (OpCode)fieldInfo.GetValue(null);
				UInt16 opCodeValue = (UInt16)opCode.Value;

				if (opCodeValue < 0x100)
				{
					_1byteOpCodes[opCodeValue] = opCode;
				}
				else if ((opCodeValue & 0xFF00) == 0xFE00)
				{
					_2byteOpCodes[opCodeValue & 0xFF] = opCode;
				}
			}
		}

		public ILParser(byte[] il)
		{
			this._il = il;
		}

		private sbyte ReadInt8()
		{
			if (this._offset + 1 > this._il.Length)
			{
				throw new BadImageFormatException(
						string.Format(
							CultureInfo.InvariantCulture,
							UnexpectedEndOfStreamExceptionMessage,
							this._offset));
			}

			return (SByte)this._il[this._offset++];
		}

		private byte ReadUInt8()
		{
			if (this._offset + 1 > this._il.Length)
			{
				throw new BadImageFormatException(
						string.Format(
							CultureInfo.InvariantCulture,
							UnexpectedEndOfStreamExceptionMessage,
							this._offset));
			}

			return this._il[this._offset++];
		}

		private ushort ReadUInt16()
		{
			if (this._offset + 2 > this._il.Length)
			{
				throw new BadImageFormatException(
						string.Format(
							CultureInfo.InvariantCulture,
							UnexpectedEndOfStreamExceptionMessage,
							this._offset));
			}

			ushort value = BitConverter.ToUInt16(this._il, this._offset);

			this._offset += 2;

			return value;
		}

		private int ReadInt32()
		{
			if (this._offset + 4 > this._il.Length)
			{
				throw new BadImageFormatException(
						string.Format(
							CultureInfo.InvariantCulture,
							UnexpectedEndOfStreamExceptionMessage,
							this._offset));
			}

			int value = BitConverter.ToInt32(this._il, this._offset);

			this._offset += 4;

			return value;
		}

		private uint ReadUInt32()
		{
			if (this._offset + 4 > this._il.Length)
			{
				throw new BadImageFormatException(
						string.Format(
							CultureInfo.InvariantCulture,
							UnexpectedEndOfStreamExceptionMessage,
							this._offset));
			}

			uint value = BitConverter.ToUInt32(this._il, this._offset);

			this._offset += 4;

			return value;
		}

		private long ReadInt64()
		{
			if (this._offset + 8 > this._il.Length)
			{
				throw new BadImageFormatException(
						string.Format(
							CultureInfo.InvariantCulture,
							UnexpectedEndOfStreamExceptionMessage,
							this._offset));
			}

			long value = BitConverter.ToInt64(this._il, this._offset);

			this._offset += 8;

			return value;
		}

		private float ReadSingle()
		{
			if (this._offset + 4 > this._il.Length)
			{
				throw new BadImageFormatException(
						string.Format(
							CultureInfo.InvariantCulture,
							UnexpectedEndOfStreamExceptionMessage,
							this._offset));
			}

			float value = BitConverter.ToSingle(this._il, this._offset);

			this._offset += 4;

			return value;
		}

		private double ReadDouble()
		{
			if (this._offset + 8 > this._il.Length)
			{
				throw new BadImageFormatException(
						string.Format(
							CultureInfo.InvariantCulture,
							UnexpectedEndOfStreamExceptionMessage,
							this._offset));
			}

			double value = BitConverter.ToDouble(this._il, this._offset);

			this._offset += 8;

			return value;
		}

		public ILInstruction ReadNextInstruction()
		{
			int offset = this._offset;
			OpCode opCode = OpCodes.Nop;
			byte value = ReadUInt8();

			if (value == 0xFE)
			{
				value = ReadUInt8();
				opCode = _2byteOpCodes[value];
			}
			else
			{
				opCode = _1byteOpCodes[value];
			}

			switch (opCode.OperandType)
			{
				// The operand is a 32-bit integer branch target.
				case OperandType.InlineBrTarget:
					{
						int operand = ReadInt32();

						return new ILInstruction(offset, this._offset - offset, opCode, operand);
					}
				// The operand is a 32-bit metadata token.
				case OperandType.InlineField:
					{
						int operand = ReadInt32();

						return new ILInstruction(offset, this._offset - offset, opCode, operand);
					}
				// The operand is a 32-bit integer.
				case OperandType.InlineI:
					{
						int operand = ReadInt32();

						return new ILInstruction(offset, this._offset - offset, opCode, operand);
					}
				// The operand is a 64-bit integer.
				case OperandType.InlineI8:
					{
						long operand = ReadInt64();

						return new ILInstruction(offset, this._offset - offset, opCode, operand);
					}
				// The operand is a 32-bit metadata token.
				case OperandType.InlineMethod:
					{
						int operand = ReadInt32();

						return new ILInstruction(offset, this._offset - offset, opCode, operand);
					}
				// No operand.
				case OperandType.InlineNone:
					return new ILInstruction(offset, this._offset - offset, opCode);
				// The operand is reserved and should not be used.
				//case OperandType.InlinePhi:
				//break;
				// The operand is a 64-bit IEEE floating point number.
				case OperandType.InlineR:
					{
						double operand = ReadDouble();

						return new ILInstruction(offset, this._offset - offset, opCode, operand);
					}
				// The operand is a 32-bit metadata signature token.
				case OperandType.InlineSig:
					{
						int operand = ReadInt32();

						return new ILInstruction(offset, this._offset - offset, opCode, operand);
					}
				// The operand is a 32-bit metadata string token.
				case OperandType.InlineString:
					{
						int operand = ReadInt32();

						return new ILInstruction(offset, this._offset - offset, opCode, operand);
					}
				// The operand is the 32-bit integer argument to a switch instruction.
				case OperandType.InlineSwitch:
					{
						int[] offsets = new int[ReadUInt32()];

						for (int index = 0; index < offsets.Length; index++)
						{
							offsets[index] = ReadInt32();
						}

						return new ILInstruction(offset, this._offset - offset, opCode, offsets);
					}
				// The operand is a FieldRef, MethodRef, or TypeRef token.
				case OperandType.InlineTok:
					{
						int operand = ReadInt32();

						return new ILInstruction(offset, this._offset - offset, opCode, operand);
					}
				// The operand is a 32-bit metadata token.
				case OperandType.InlineType:
					{
						int operand = ReadInt32();

						return new ILInstruction(offset, this._offset - offset, opCode, operand);
					}
				// The operand is 16-bit integer containing the ordinal of a local variable or an argument.
				case OperandType.InlineVar:
					{
						ushort operand = ReadUInt16();

						return new ILInstruction(offset, this._offset - offset, opCode, operand);
					}
				// The operand is an 8-bit integer branch target.
				case OperandType.ShortInlineBrTarget:
					{
						sbyte operand = ReadInt8();

						return new ILInstruction(offset, this._offset - offset, opCode, operand);
					}
				// The operand is an 8-bit integer.
				case OperandType.ShortInlineI:
					{
						byte operand = ReadUInt8();

						return new ILInstruction(offset, this._offset - offset, opCode, operand);
					}
				// The operand is a 32-bit IEEE floating point number.
				case OperandType.ShortInlineR:
					{
						float operand = ReadSingle();

						return new ILInstruction(offset, this._offset - offset, opCode, operand);
					}
				// The operand is an 8-bit integer containing the ordinal of a local variable or an argument.
				case OperandType.ShortInlineVar:
					{
						byte operand = ReadUInt8();

						return new ILInstruction(offset, this._offset - offset, opCode, operand);
					}
				default:
					throw new BadImageFormatException(
						string.Format(
							CultureInfo.InvariantCulture,
							UnexpectedOpCodeExceptionMessage,
							opCode.Value,
							offset,
							this._il[offset],
							(offset + 1 < this._il.Length) ? this._il[offset + 1] : byte.MinValue,
							(offset + 2 < this._il.Length) ? this._il[offset + 2] : byte.MinValue,
							(offset + 3 < this._il.Length) ? this._il[offset + 3] : byte.MinValue,
							(offset + 4 < this._il.Length) ? this._il[offset + 4] : byte.MinValue,
							(offset + 5 < this._il.Length) ? this._il[offset + 5] : byte.MinValue));
			}
		}

		public IEnumerator<ILInstruction> GetEnumerator()
		{
			while (this._offset < this._il.Length)
			{
				yield return ReadNextInstruction();
			}

			this._offset = 0;
			yield break;
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
