﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Reflection Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;

using Nabu.Text;

namespace Nabu.Reflection
{
	public abstract class ScriptExecutionEngineBase :
		DisposableBase,
		IScriptExecutionEngine
	{
		#region ScriptExecutionEngineBase

		#region Utility

		protected static ScriptEngineCompilerMessage CreateMessage(
			IEnumerable<ISourceCodeUnit> sourceCodeUnits,
			TextPosition position,
			string messageID,
			string messageText,
			bool isWarning,
			bool isError)
		{
			ISourceCodeUnit sourceCodeUnit = null;

			foreach (ISourceCodeUnit sourceCodeUnitCandidate in sourceCodeUnits)
			{
				if (sourceCodeUnitCandidate.Module == position.File)
				{
					sourceCodeUnit = sourceCodeUnitCandidate;
					break;
				}
			}

			IGeneratedSourceCodeUnit generatedSourceCodeUnit = sourceCodeUnit as IGeneratedSourceCodeUnit;

			if (generatedSourceCodeUnit == null)
			{
				return
					new ScriptEngineCompilerMessage(
						position,
						new TextPosition(null, -1, -1),
						messageID,
						messageText,
						isWarning,
						isError);
			}
			else
			{
				long errorPosition = ((long)position.Line << 32) + (position.Column - 1);

				foreach (TextPositionMap map in generatedSourceCodeUnit.PositionMap)
				{
					long startPosition = ((long)map.DestinationStart.Line << 32) + map.DestinationStart.Column;
					long endPosition = ((long)map.DestinationEnd.Line << 32) + map.DestinationEnd.Column;

					if ((errorPosition >= startPosition) && (errorPosition < endPosition))
					{
						if (position.Line == map.DestinationStart.Line)
						{
							return
								new ScriptEngineCompilerMessage(
									position,
									new TextPosition(
										null,
										position.Line - (map.DestinationStart.Line - map.SourceStart.Line),
										position.Column - (map.DestinationStart.Column - map.SourceStart.Column)),
									messageID,
									messageText,
									isWarning,
									isError);
						}
						else
						{
							return
								new ScriptEngineCompilerMessage(
									position,
									new TextPosition(
										null,
										position.Line - (map.DestinationStart.Line - map.SourceStart.Line),
										position.Column),
									messageID,
									messageText,
									isWarning,
									isError);
						}
					}
				}

				return
					new ScriptEngineCompilerMessage(
						position,
						new TextPosition(null, -1, -1),
						messageID,
						messageText,
						isWarning,
						isError);
			}
		}

		#endregion

		#region Constructors

		protected ScriptExecutionEngineBase()
		{
		}

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			base.Dispose(disposing);
		}

		#endregion

		#region IScriptExecutionEngine

		public abstract IntPtr CreateInstance(string scopeName, string typeName);
		public abstract void DestroyInstance(IntPtr cookie);
		public abstract object InvokeMethod(string scopeName, string typeName, string methodName, params object[] arguments);
		public abstract object InvokeMethod(IntPtr cookie, string methodName, params object[] arguments);
		public abstract object GetProperty(string scopeName, string typeName, string propertyName);
		public abstract object GetProperty(IntPtr cookie, string propertyName);
		public abstract void SetProperty(string scopeName, string typeName, string propertyName, object propertyValue);
		public abstract void SetProperty(IntPtr cookie, string propertyName, object propertyValue);

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
