﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Reflection Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Reflection;
using System.Runtime.InteropServices;

namespace Nabu.Reflection
{
	public sealed class AppDomainProxy : MarshalByRefObject
	{
		#region Fields

		private Assembly _assembly;

		#endregion

		#region Methods

		private Type FindType(string scopeName, string typeName)
		{
			foreach (Type type in this._assembly.GetExportedTypes())
			{
				if ((type.Namespace == scopeName) && (type.Name == typeName))
				{
					return type;
				}
			}

			return null;
		}

		public AppDomainProxy()
		{
		}

		public void Load(byte[] assemblyData)
		{
			this._assembly = Assembly.Load(assemblyData);
		}

		public IntPtr CreateInstance(
			string scopeName,
			string typeName)
		{
			Type type = FindType(scopeName, typeName);

			if (type == null)
			{
				return IntPtr.Zero;
			}
			else
			{
				object instance = Activator.CreateInstance(type);
				GCHandle handle = GCHandle.Alloc(instance, GCHandleType.Normal);

				return (IntPtr)handle;
			}
		}

		public void DestroyInstance(
			IntPtr cookie)
		{
			if (cookie != IntPtr.Zero)
			{
				GCHandle handle = (GCHandle)cookie;

				handle.Free();
			}
		}

		public bool InvokeMethod(
			string scopeName,
			string typeName,
			string methodName,
			object[] arguments,
			out object result)
		{
			Type type = FindType(scopeName, typeName);
			MethodInfo methodInfo =
				type.GetMethod(
					methodName,
					BindingFlags.Static | BindingFlags.InvokeMethod | BindingFlags.Public);

			if (methodInfo == null)
			{
				result = null;

				return false;
			}
			else
			{
				result = methodInfo.Invoke(null, arguments);

				return true;
			}
		}

		public bool InvokeMethod(
			IntPtr cookie,
			string methodName,
			object[] arguments,
			out object result)
		{
			GCHandle handle = (GCHandle)cookie;
			object instance = handle.Target;
			Type type = instance.GetType();
			MethodInfo methodInfo =
				type.GetMethod(
					methodName,
					BindingFlags.Instance | BindingFlags.InvokeMethod | BindingFlags.Public);

			if (methodInfo == null)
			{
				result = null;

				return false;
			}
			else
			{
				result = methodInfo.Invoke(instance, arguments);

				return true;
			}
		}

		public bool GetProperty(
			string scopeName,
			string typeName,
			string propertyName,
			out object propertyValue)
		{
			Type type = FindType(scopeName, typeName);
			PropertyInfo propertyInfo =
				type.GetProperty(
					propertyName,
					BindingFlags.Static | BindingFlags.InvokeMethod | BindingFlags.Public);

			if (propertyInfo == null)
			{
				propertyValue = null;

				return false;
			}
			else
			{
				propertyValue = propertyInfo.GetValue(null, null);

				return true;
			}
		}

		public bool GetProperty(
			IntPtr cookie,
			string propertyName,
			out object propertyValue)
		{
			GCHandle handle = (GCHandle)cookie;
			object instance = handle.Target;
			Type type = instance.GetType();
			PropertyInfo propertyInfo =
				type.GetProperty(
					propertyName,
					BindingFlags.Instance | BindingFlags.InvokeMethod | BindingFlags.Public);

			if (propertyInfo == null)
			{
				propertyValue = null;

				return false;
			}
			else
			{
				propertyValue = propertyInfo.GetValue(instance, null);

				return true;
			}
		}

		public bool SetProperty(
			string scopeName,
			string typeName,
			string propertyName,
			object propertyValue)
		{
			Type type = FindType(scopeName, typeName);
			PropertyInfo propertyInfo =
				type.GetProperty(
					propertyName,
					BindingFlags.Static | BindingFlags.InvokeMethod | BindingFlags.Public);

			if (propertyInfo == null)
			{
				return false;
			}
			else
			{
				propertyInfo.SetValue(null, propertyValue, null);

				return true;
			}
		}

		public bool SetProperty(
			IntPtr cookie,
			string propertyName,
			object propertyValue)
		{
			GCHandle handle = (GCHandle)cookie;
			object instance = handle.Target;
			Type type = instance.GetType();
			PropertyInfo propertyInfo =
				type.GetProperty(
					propertyName,
					BindingFlags.Instance | BindingFlags.InvokeMethod | BindingFlags.Public);

			if (propertyInfo == null)
			{
				return false;
			}
			else
			{
				propertyInfo.SetValue(instance, propertyValue, null);

				return true;
			}
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
