﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Reflection Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Globalization;
using System.IO;

using Nabu.Text;

namespace Nabu.Reflection
{
	internal sealed class ManagedScriptExecutionEngine :
		ScriptExecutionEngineBase
	{
		#region ManagedScriptExecutionEngine

		#region Fields

		private AppDomainProxy _appDomainProxy;

		#endregion

		#region Methods

		#region Utility

		private ScriptEngineCompilerMessage[] ProcessCompilerResults(
			IEnumerable<ISourceCodeUnit> sourceCodeUnits,
			CompilerResults results)
		{
			if (results.Errors.HasErrors)
			{
				ScriptEngineCompilerMessage[] messages = new ScriptEngineCompilerMessage[results.Errors.Count];

				for (int index = 0; index < results.Errors.Count; index++)
				{
					CompilerError compilerError = results.Errors[index];

					messages[index] =
						CreateMessage(
							sourceCodeUnits,
							new TextPosition(
								Path.GetFileName(compilerError.FileName),
								compilerError.Line - 1,
								compilerError.Column - 1),
							compilerError.ErrorNumber,
							compilerError.ErrorText,
							compilerError.IsWarning,
							!compilerError.IsWarning);
				}

				return messages;
			}
			else
			{
				return null;
			}
		}

		#endregion

		#region Constructors

		[SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
		internal ManagedScriptExecutionEngine(
			ManagedScriptExecutionHostBase host,
			IEnumerable<ISourceCodeUnit> sourceCodeUnits)
		{
			string basePath =
				string.Format(
					CultureInfo.InvariantCulture,
					"{0}SEE-{1:N}\\",
					Path.GetTempPath(),
					Guid.NewGuid());
			string outputPath = basePath + "assembly.dll";

			Directory.CreateDirectory(basePath);

			try
			{
				using (CodeDomProvider provider = CodeDomProvider.CreateProvider("CSharp"))
				{
					CompilerParameters parameters = new CompilerParameters();

					parameters.GenerateExecutable = false;
					parameters.GenerateInMemory = false;
					parameters.IncludeDebugInformation = true;
					parameters.OutputAssembly = outputPath;

					List<string> inputPaths = new List<string>();

					foreach (ISourceCodeUnit sourceCodeUnit in sourceCodeUnits)
					{
						string inputPath = basePath + sourceCodeUnit.Module;

						inputPaths.Add(inputPath);
						File.WriteAllText(inputPath, sourceCodeUnit.Value);

						foreach (string module in sourceCodeUnit.Modules)
						{
							parameters.ReferencedAssemblies.Add(module);
						}
					}

					CompilerResults results =
						provider.CompileAssemblyFromFile(
							parameters,
							inputPaths.ToArray());
					ScriptEngineCompilerMessage[] messages =
						ProcessCompilerResults(
							sourceCodeUnits,
							results);

					if (messages != null)
					{
						throw new ScriptEngineCompilerException(messages);
					}

					this._appDomainProxy =
						(AppDomainProxy)host.AppDomain.CreateInstanceFromAndUnwrap(
							typeof(AppDomainProxy).Assembly.Location,
							typeof(AppDomainProxy).FullName);
					this._appDomainProxy.Load(File.ReadAllBytes(outputPath));
				}
			}
			finally
			{
				try
				{
					if (Directory.Exists(basePath))
					{
						Directory.Delete(basePath, true);
					}
				}
				catch
				{
				}
			}
		}

		#endregion

		#endregion

		#endregion

		#region IScriptExecutionEngine

		public override IntPtr CreateInstance(
			string scopeName,
			string typeName)
		{
			return
				this._appDomainProxy.CreateInstance(
					scopeName,
					typeName);
		}

		public override void DestroyInstance(
			IntPtr cookie)
		{
			this._appDomainProxy.DestroyInstance(
				cookie);
		}

		public override object InvokeMethod(
			string scopeName,
			string typeName,
			string methodName,
			params object[] arguments)
		{
			object result;

			if (!this._appDomainProxy.InvokeMethod(
				scopeName,
				typeName,
				methodName,
				arguments,
				out result))
			{
				throw new MissingMethodException(typeName, methodName);
			}

			return result;
		}

		public override object InvokeMethod(
			IntPtr cookie,
			string methodName,
			params object[] arguments)
		{
			object result;

			if (!this._appDomainProxy.InvokeMethod(
				cookie,
				methodName,
				arguments,
				out result))
			{
				throw new MissingMethodException(null, methodName);
			}

			return result;
		}

		public override object GetProperty(
			string scopeName,
			string typeName,
			string propertyName)
		{
			object propertyValue;

			if (!this._appDomainProxy.GetProperty(
				scopeName,
				typeName,
				propertyName,
				out propertyValue))
			{
				throw new MissingMemberException(typeName, propertyName);
			}

			return propertyValue;
		}

		public override object GetProperty(
			IntPtr cookie,
			string propertyName)
		{
			object propertyValue;

			if (!this._appDomainProxy.GetProperty(
				cookie,
				propertyName,
				out propertyValue))
			{
				throw new MissingMemberException(null, propertyName);
			}

			return propertyValue;
		}

		public override void SetProperty(
			string scopeName,
			string typeName,
			string propertyName,
			object propertyValue)
		{
			if (!this._appDomainProxy.SetProperty(
				scopeName,
				typeName,
				propertyName,
				propertyValue))
			{
				throw new MissingMemberException(typeName, propertyName);
			}
		}

		public override void SetProperty(
			IntPtr cookie,
			string propertyName,
			object propertyValue)
		{
			if (!this._appDomainProxy.SetProperty(
				cookie,
				propertyName,
				propertyValue))
			{
				throw new MissingMemberException(null, propertyName);
			}
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
