﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Reflection Library                       */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.IO;
using System.Runtime.InteropServices;

using Nabu.Text;

using PaxScript.Net;
using PaxScript.Net.Reflection;

namespace Nabu.Reflection
{
	internal sealed class PaxScriptExecutionEngine :
		ScriptExecutionEngineBase
	{
		#region PaxScriptExecutionEngine

		#region Fields

		private PaxScripter _scripter;

		#endregion

		#region Methods

		#region Event Handlers

		private void Scripter_OnPaxException(PaxScripter sender, Exception e)
		{
			//throw e;
		}

		#endregion

		#region Utility

		private PaxType FindType(string scopeName, string typeName)
		{
			PaxNamespaceInfo namespaceInfo =
				PaxNamespaceInfo.LookupNamespace(
					this._scripter,
					scopeName);

			if (namespaceInfo == null)
			{
				return null;
			}
			else
			{
				return namespaceInfo.LookupMember(typeName) as PaxType;
			}
		}

		private ScriptEngineCompilerMessage[] ProcessScripterErrors(
			IEnumerable<ISourceCodeUnit> sourceCodeUnits)
		{
			if (this._scripter.HasErrors)
			{
				ScriptEngineCompilerMessage[] messages = new ScriptEngineCompilerMessage[this._scripter.Error_List.Count];

				for (int index = 0; index < this._scripter.Error_List.Count; index++)
				{
					ScriptError scriptError = this._scripter.Error_List[index];

					messages[index] =
						CreateMessage(
							sourceCodeUnits,
							new TextPosition(
								scriptError.ModuleName,
								scriptError.LineNumber,
								0),
							string.Empty,
							scriptError.Message,
							false,
							true);
				}

				return messages;
			}
			else
			{
				return null;
			}
		}

		#endregion

		#region Constructors

		internal PaxScriptExecutionEngine(
			PaxScriptExecutionHost host,
			IEnumerable<ISourceCodeUnit> sourceCodeUnits,
			IDictionary<string, object> globalItems)
		{
			this._scripter = new PaxScripter();

			foreach (ISourceCodeUnit sourceCodeUnit in sourceCodeUnits)
			{
				foreach (string module in sourceCodeUnit.Modules)
				{
					string path = Path.GetFullPath(module);

					if (File.Exists(path))
					{
						this._scripter.RegisterAssembly(path);
					}
					else
					{
						if (module.EndsWith(".dll", StringComparison.OrdinalIgnoreCase))
						{
							this._scripter.RegisterAssemblyWithPartialName(module.Substring(0, module.Length - 4));
						}
						else
						{
							this._scripter.RegisterAssemblyWithPartialName(module);
						}
					}
				}
			}

			this._scripter.OnPaxException += new PaxExceptionHandler(Scripter_OnPaxException);

			this._scripter.AddModule(
				"entrypoint");
			this._scripter.AddCode(
				"entrypoint",
@"public sealed class Program
{
	public static void Main()
	{
	}
}");
			foreach (ISourceCodeUnit sourceCodeUnit in sourceCodeUnits)
			{
				this._scripter.AddModule(
					sourceCodeUnit.Module);
				this._scripter.AddCode(
					sourceCodeUnit.Module,
					sourceCodeUnit.Value);
			}

			this._scripter.Compile();
			this._scripter.Link();

			ScriptEngineCompilerMessage[] messages =
				ProcessScripterErrors(
					sourceCodeUnits);

			if (messages != null)
			{
				throw new ScriptEngineCompilerException(messages);
			}

			this._scripter.Run(RunMode.Run);
		}

		#endregion

		#endregion

		#endregion

		#region IScriptExecutionEngine

		public override IntPtr CreateInstance(
			string scopeName,
			string typeName)
		{
			PaxType paxType = FindType(scopeName, typeName);

			if (paxType == null)
			{
				return IntPtr.Zero;
			}
			else
			{
				PaxObject value = paxType.CreateObject();
				GCHandle handle = GCHandle.Alloc(value, GCHandleType.Normal);

				return (IntPtr)handle;
			}
		}

		public override void DestroyInstance(
			IntPtr cookie)
		{
			if (cookie != IntPtr.Zero)
			{
				GCHandle handle = (GCHandle)cookie;

				handle.Free();
			}
		}

		public override object InvokeMethod(
			string scopeName,
			string typeName,
			string methodName,
			params object[] arguments)
		{
			PaxType paxType = FindType(scopeName, typeName);

			if (paxType == null)
			{
				throw new MissingMethodException(methodName);
			}
			else
			{
				PaxMethodInfo paxMethodInfo = paxType.LookupMember(methodName) as PaxMethodInfo;

				if (paxMethodInfo == null)
				{
					throw new MissingMethodException(methodName);
				}
				else
				{
					return
						paxMethodInfo.Invoke(
							null,
							arguments);
				}
			}
		}

		public override object InvokeMethod(
			IntPtr cookie,
			string methodName,
			params object[] arguments)
		{
			try
			{
				GCHandle handle = (GCHandle)cookie;
				PaxObject paxObject = (PaxObject)handle.Target;

				return paxObject.Invoke(methodName, arguments);
			}
			catch (Errors.PaxScriptException e)
			{
				throw new MissingMethodException(methodName, e);
			}
		}

		public override object GetProperty(
			string scopeName,
			string typeName,
			string propertyName)
		{
			PaxType paxType = FindType(scopeName, typeName);

			if (paxType == null)
			{
				throw new MissingMethodException(propertyName);
			}
			else
			{
				PaxPropertyInfo paxPropertyInfo = paxType.LookupMember(propertyName) as PaxPropertyInfo;

				if (paxPropertyInfo == null)
				{
					throw new MissingMethodException(propertyName);
				}
				else
				{
					PaxMethodInfo paxMethodInfo = paxPropertyInfo.GetGetMethod();

					if (paxMethodInfo == null)
					{
						throw new MissingMethodException(propertyName);
					}
					else
					{
						return
							paxMethodInfo.Invoke(
								null);
					}
				}
			}
		}

		public override object GetProperty(
			IntPtr cookie,
			string propertyName)
		{
			try
			{
				GCHandle handle = (GCHandle)cookie;
				PaxObject paxObject = (PaxObject)handle.Target;

				return
					paxObject.GetProperty(
						propertyName);
			}
			catch (Errors.PaxScriptException e)
			{
				throw new MissingMemberException(propertyName, e);
			}
		}

		public override void SetProperty(
			string scopeName,
			string typeName,
			string propertyName,
			object propertyValue)
		{
			PaxType paxType = FindType(scopeName, typeName);

			if (paxType == null)
			{
				throw new MissingMethodException(propertyName);
			}
			else
			{
				PaxPropertyInfo paxPropertyInfo = paxType.LookupMember(propertyName) as PaxPropertyInfo;

				if (paxPropertyInfo == null)
				{
					throw new MissingMethodException(propertyName);
				}
				else
				{
					PaxMethodInfo paxMethodInfo = paxPropertyInfo.GetSetMethod();

					if (paxMethodInfo == null)
					{
						throw new MissingMethodException(propertyName);
					}
					else
					{
						paxMethodInfo.Invoke(
							null,
							propertyValue);
					}
				}
			}
		}

		public override void SetProperty(
			IntPtr cookie,
			string propertyName,
			object propertyValue)
		{
			try
			{
				GCHandle handle = (GCHandle)cookie;
				PaxObject paxObject = (PaxObject)handle.Target;

				paxObject.PutProperty(
					propertyName,
					propertyValue);
			}
			catch (Errors.PaxScriptException e)
			{
				throw new MissingMemberException(propertyName, e);
			}
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
