﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Text Library                             */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Drawing;

using Nabu.Collections;

namespace Nabu.Text
{
	public sealed class CodeText<T>
	{
		#region Types

		private enum LineChangeType
		{
			Add,
			Remove,
			Replace
		}

		private class LineChange
		{
			#region Fields

			private LineChangeType _type;
			private int _index;
			private string _oldValue;
			private string _newValue;

			#endregion

			#region Methods

			[DebuggerStepThrough]
			public LineChange(LineChangeType type, int index, string newValue)
			{
				this._type = type;
				this._index = index;
				this._oldValue = string.Empty;
				this._newValue = newValue;
			}

			[DebuggerStepThrough]
			public LineChange(LineChangeType type, int index, string oldValue, string newValue)
			{
				this._type = type;
				this._index = index;
				this._oldValue = oldValue;
				this._newValue = newValue;
			}

			#endregion

			#region Properties

			public LineChangeType Type
			{
				[DebuggerStepThrough]
				get { return this._type; }
				[DebuggerStepThrough]
				set { this._type = value; }
			}

			public int Index
			{
				[DebuggerStepThrough]
				get { return this._index; }
			}

			public string OldValue
			{
				[DebuggerStepThrough]
				get { return this._oldValue; }
			}

			public string NewValue
			{
				[DebuggerStepThrough]
				get { return this._newValue; }
			}

			#endregion
		}

		private class TextChange
		{
			#region Fields

			private IList<LineChange> _changes;
			private Point _selectionStart;
			private Point _selectionEnd;

			#endregion

			#region Methods

			public TextChange()
			{
				this._changes = new List<LineChange>();
				this._selectionStart = new Point();
				this._selectionEnd = new Point();
			}

			private TextChange(TextChange state)
			{
				this._changes = new List<LineChange>(state.Changes);
				this._selectionStart = state.SelectionStart;
				this._selectionEnd = state.SelectionEnd;
			}

			public TextChange Clone()
			{
				return new TextChange(this);
			}

			#endregion

			#region Properties

			public IList<LineChange> Changes
			{
				[DebuggerStepThrough]
				get { return this._changes; }
			}

			public Point SelectionStart
			{
				[DebuggerStepThrough]
				get { return this._selectionStart; }
				[DebuggerStepThrough]
				set { this._selectionStart = value; }
			}

			public Point SelectionEnd
			{
				[DebuggerStepThrough]
				get { return this._selectionEnd; }
				[DebuggerStepThrough]
				set { this._selectionEnd = value; }
			}

			#endregion
		}

		#endregion

		#region Fields

		private BindingListEx<CodeTextLine<T>> _text = new BindingListEx<CodeTextLine<T>>(256);
		private Point _selectionStart = new Point();
		private Point _selectionEnd = new Point();
		private TextChange _currentChange = new TextChange();
		private Stack<TextChange> _undoStack = new Stack<TextChange>();
		private Stack<TextChange> _redoStack = new Stack<TextChange>();
		private Point _changeSelectionStart = new Point();
		private Point _changeSelectionEnd = new Point();
		private bool _isApplyingChange = true;

		#endregion

		#region Methods

		#region Event Raisers

		private void RaiseSelectionChanged(CodeSelectionChangeType selectionChangeType)
		{
			if (this.SelectionChanged != null)
			{
				this.SelectionChanged(this, new CodeSelectionChangedEventArgs(selectionChangeType));
			}
		}

		#endregion

		#region Event Handlers

		private void Lines_Reseted(object sender, EventArgs e)
		{
		}

		private void Lines_Added(object sender, BindingListExAddedRemovedEventArgs<CodeTextLine<T>> e)
		{
			if (!this._isApplyingChange)
			{
				this._currentChange.Changes.Add(
					new LineChange(
						LineChangeType.Add,
						e.Index,
						e.Value.PlainText));
			}
		}

		private void Lines_Removed(object sender, BindingListExAddedRemovedEventArgs<CodeTextLine<T>> e)
		{
			if (!this._isApplyingChange)
			{
				this._currentChange.Changes.Add(
					new LineChange(
						LineChangeType.Remove,
						e.Index,
						e.Value.PlainText));
			}
		}

		private void Lines_Changed(object sender, BindingListExChangedEventArgs<CodeTextLine<T>> e)
		{
			if (!this._isApplyingChange)
			{
				this._currentChange.Changes.Add(
					new LineChange(
						LineChangeType.Replace,
						e.Index,
						e.OldValue.PlainText,
						e.NewValue.PlainText));
			}
		}

		private void Lines_Moved(object sender, BindingListExMovedEventArgs<CodeTextLine<T>> e)
		{
		}

		#endregion

		#region Position

		public bool IsPositionValid(Point position)
		{
			return
				(position.Y >= 0) &&
				(position.Y < this._text.Count) &&
				(position.X >= 0) &&
				(position.X <= this._text[position.Y].Length);
		}

		public bool IsPositionSelected(Point position)
		{
			ulong lPositionX = (ulong)position.X;
			ulong lPositionY = (ulong)position.Y;
			ulong lPosition = (lPositionY << 32) | lPositionX;

			ulong lSelectionStartX = (ulong)this._selectionStart.X;
			ulong lSelectionStartY = (ulong)this._selectionStart.Y;
			ulong lSelectionStart = (lSelectionStartY << 32) | lSelectionStartX;

			ulong lSelectionEndX = (ulong)this._selectionEnd.X;
			ulong lSelectionEndY = (ulong)this._selectionEnd.Y;
			ulong lSelectionEnd = (lSelectionEndY << 32) | lSelectionEndX;

			return
				((lPosition >= lSelectionStart) && (lPosition < lSelectionEnd)) ||
				((lPosition >= lSelectionEnd) && (lPosition < lSelectionStart));
		}

		#endregion

		#region Selection

		public void SetSelectionRange(Point startPosition, Point endPosition)
		{
			bool selectionStartChanged = false;
			bool selectionEndChanged = false;

			if (this._selectionStart != startPosition)
			{
				this._selectionStart = startPosition;
				selectionStartChanged = true;
			}

			if (this._selectionEnd != endPosition)
			{
				this._selectionEnd = endPosition;
				selectionEndChanged = true;
			}

			if (selectionStartChanged)
			{
				if (selectionEndChanged)
				{
					RaiseSelectionChanged(CodeSelectionChangeType.Both);
				}
				else
				{
					RaiseSelectionChanged(CodeSelectionChangeType.Start);
				}
			}
			else
			{
				if (selectionEndChanged)
				{
					RaiseSelectionChanged(CodeSelectionChangeType.End);
				}
				else
				{
				}
			}
		}

		#endregion

		#region Changes

		public void CommitChange()
		{
			if (this._currentChange.Changes.Count > 0)
			{
				this._currentChange.SelectionStart = this._changeSelectionStart;
				this._currentChange.SelectionEnd = this._changeSelectionEnd;
				this._undoStack.Push(this._currentChange.Clone());
				this._currentChange.Changes.Clear();
			}
			else
			{
				this._changeSelectionStart = this._selectionStart;
				this._changeSelectionEnd = this._selectionEnd;
			}
		}

		public void UndoChange()
		{
			if (this._undoStack.Count > 0)
			{
				this._isApplyingChange = true;

				TextChange changeState = this._undoStack.Pop();

				for (int index = changeState.Changes.Count - 1; index >= 0; --index)
				{
					LineChange change = changeState.Changes[index];

					switch (change.Type)
					{
						case LineChangeType.Add:
							this._text.RemoveAt(change.Index);
							break;
						case LineChangeType.Remove:
							this._text.Insert(change.Index, change.NewValue);
							break;
						case LineChangeType.Replace:
							this._text[change.Index].PlainText = change.OldValue;
							break;
					}
				}

				SetSelectionRange(changeState.SelectionStart, changeState.SelectionEnd);

				this._redoStack.Push(changeState);

				this._isApplyingChange = false;
			}
		}

		public void RedoChange()
		{
			if (this._redoStack.Count > 0)
			{
				this._isApplyingChange = true;

				TextChange changeState = this._redoStack.Pop();

				for (int index = 0; index < changeState.Changes.Count; index++)
				{
					LineChange change = changeState.Changes[index];

					switch (change.Type)
					{
						case LineChangeType.Add:
							this._text.Insert(change.Index, change.NewValue);
							break;
						case LineChangeType.Remove:
							this._text.RemoveAt(change.Index);
							break;
						case LineChangeType.Replace:
							this._text[change.Index].PlainText = change.NewValue;
							break;
					}
				}

				SetSelectionRange(changeState.SelectionStart, changeState.SelectionEnd);

				this._undoStack.Push(changeState);

				this._isApplyingChange = false;
			}
		}

		#endregion

		#region Constructors

		public CodeText()
		{
			this._text.Reseted += new EventHandler(Lines_Reseted);
			this._text.Added += new EventHandler<BindingListExAddedRemovedEventArgs<CodeTextLine<T>>>(Lines_Added);
			this._text.Removed += new EventHandler<BindingListExAddedRemovedEventArgs<CodeTextLine<T>>>(Lines_Removed);
			this._text.Changed += new EventHandler<BindingListExChangedEventArgs<CodeTextLine<T>>>(Lines_Changed);
			this._text.Moved += new EventHandler<BindingListExMovedEventArgs<CodeTextLine<T>>>(Lines_Moved);
			this._text.FromArray(
				new CodeTextLine<T>[]
				{
					new CodeTextLine<T>(string.Empty)
				},
				true);
		}

		#endregion

		#endregion

		#region Properties

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		public BindingListEx<CodeTextLine<T>> Lines
		{
			[DebuggerStepThrough]
			get { return this._text; }
		}

		public Point Start
		{
			get
			{
				return new Point(0, 0);
			}
		}

		public Point End
		{
			get
			{
				return
					new Point(
						this._text.Last.Length,
						this._text.Count - 1);
			}
		}

		public Point SelectionStart
		{
			[DebuggerStepThrough]
			get { return this._selectionStart; }
		}

		public Point SelectionEnd
		{
			[DebuggerStepThrough]
			get { return this._selectionEnd; }
		}

		#endregion

		#region Events

		public event EventHandler<CodeSelectionChangedEventArgs> SelectionChanged;

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
