﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Text Library                             */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Text;

using Nabu.Collections;

namespace Nabu.Text
{
	public sealed class CodeTextLine<T> :
		INotifyPropertyChanging,
		INotifyPropertyChanged,
		ICloneable
	{
		#region CodeTextLine

		#region Fields

		private StringBuilder _text;
		private ListEx<CodeTextLineSegmentInfo<T>> _formatting;

		#endregion

		#region Methods

		#region Utility

		private void Initialize(string value, int capacity)
		{
			this._text = new StringBuilder(value, Math.Max(capacity, value.Length));
			this._formatting = new ListEx<CodeTextLineSegmentInfo<T>>();
			this._formatting.Add(
				new CodeTextLineSegmentInfo<T>(
					default(T),
					value.Length));
		}

		private void InsertFormatting(int offset, int length)
		{
			if (length > 0)
			{
				int currentOffset = 0;
				int startIndex = 0;

				for (int index = 0; index < this._formatting.Count; index++)
				{
					CodeTextLineSegmentInfo<T> formattedRange = this._formatting[index];

					if (
						(currentOffset <= offset) &&
						(currentOffset + formattedRange.Length > offset))
					{
						startIndex = index;
						break;
					}

					currentOffset += formattedRange.Length;
				}

				CodeTextLineSegmentInfo<T> startRange = this._formatting[startIndex];

				this._formatting[startIndex] =
						new CodeTextLineSegmentInfo<T>(
							startRange.Format,
							startRange.Length + length);
			}
		}

		private void RemoveFormatting(int offset, int length)
		{
			if (length > 0)
			{
				int currentOffset = 0;
				int startOffset = 0;
				int endOffset = 0;
				int startIndex = 0;
				int endIndex = this._formatting.Count - 1;

				for (int index = 0; index < this._formatting.Count; index++)
				{
					CodeTextLineSegmentInfo<T> formattedRange = this._formatting[index];

					if (
						(currentOffset <= offset) &&
						(currentOffset + formattedRange.Length > offset))
					{
						startIndex = index;
						startOffset = currentOffset;
					}

					if (
						(currentOffset <= (offset + length)) &&
						(currentOffset + formattedRange.Length > (offset + length)))
					{
						endIndex = index;
						endOffset = currentOffset;
						break;
					}

					currentOffset += formattedRange.Length;
				}

				CodeTextLineSegmentInfo<T> startRange = this._formatting[startIndex];

				if (startIndex == endIndex)
				{
					this._formatting[startIndex] =
						new CodeTextLineSegmentInfo<T>(
							startRange.Format,
							startRange.Length - length);
				}
				else
				{
					CodeTextLineSegmentInfo<T> endRange = this._formatting[endIndex];

					int blockLength = 0;

					for (int index = startIndex + 1; index < endIndex; index++)
					{
						blockLength += this._formatting[index].Length;
					}

					this._formatting.RemoveRange(
						startIndex,
						endIndex - startIndex + 1);

					// int block1Offset = startOffset;
					int block1Length = offset - startOffset;
					int block2Offset = offset + length;
					int block2Length = endOffset + endRange.Length - block2Offset;

					if (block2Length > 0)
					{
						this._formatting.Insert(
							startIndex,
							new CodeTextLineSegmentInfo<T>(
								endRange.Format,
								block2Length));
					}

					if (block1Length > 0)
					{
						this._formatting.Insert(
							startIndex,
							new CodeTextLineSegmentInfo<T>(
								startRange.Format,
								block1Length));
					}
				}
			}
		}

		private void SetFormatting(int offset, int length, T format)
		{
			if (length > 0)
			{
				int currentOffset = 0;
				int startOffset = 0;
				int endOffset = 0;
				int startIndex = 0;
				int endIndex = this._formatting.Count - 1;

				for (int index = 0; index < this._formatting.Count; index++)
				{
					CodeTextLineSegmentInfo<T> formattedRange = this._formatting[index];

					if (
						(currentOffset <= offset) &&
						(currentOffset + formattedRange.Length > offset))
					{
						startIndex = index;
						startOffset = currentOffset;
					}

					if (
						(currentOffset <= (offset + length)) &&
						(currentOffset + formattedRange.Length > (offset + length)))
					{
						endIndex = index;
						endOffset = currentOffset;
						break;
					}

					currentOffset += formattedRange.Length;
				}

				CodeTextLineSegmentInfo<T> startRange = this._formatting[startIndex];

				if (startIndex == endIndex)
				{
					int block1Offset = startOffset;
					int block1Length = offset - block1Offset;
					int block2Offset = offset + length;
					int block2Length = startOffset + startRange.Length - block2Offset;

					this._formatting.RemoveAt(startIndex);

					if (block2Length > 0)
					{
						this._formatting.Insert(
							startIndex,
							new CodeTextLineSegmentInfo<T>(
								startRange.Format,
								block2Length));
					}

					this._formatting.Insert(
						startIndex,
						new CodeTextLineSegmentInfo<T>(
							format,
							length));

					if (block1Length > 0)
					{
						this._formatting.Insert(
							startIndex,
							new CodeTextLineSegmentInfo<T>(
								startRange.Format,
								block1Length));
					}
				}
				else
				{
					CodeTextLineSegmentInfo<T> endRange = this._formatting[endIndex];

					int blockLength = 0;

					for (int index = startIndex + 1; index < endIndex; index++)
					{
						blockLength += this._formatting[index].Length;
					}

					int block1Offset = startOffset;
					int block1Length = offset - block1Offset;
					int block2Offset = offset + length;
					int block2Length = endOffset + endRange.Length - block2Offset;

					this._formatting.RemoveRange(
						startIndex,
						endIndex - startIndex + 1);

					if (block2Length > 0)
					{
						this._formatting.Insert(
							startIndex,
							new CodeTextLineSegmentInfo<T>(
								endRange.Format,
								block2Length));
					}

					this._formatting.Insert(
						startIndex,
						new CodeTextLineSegmentInfo<T>(
							format,
							length));

					if (block1Length > 0)
					{
						this._formatting.Insert(
							startIndex,
							new CodeTextLineSegmentInfo<T>(
								startRange.Format,
								block1Length));
					}
				}
			}
		}

		#endregion

		#region Event Raisers

		private void RaiseChanging(int offset, int length)
		{
			if (this.Changing != null)
			{
				this.Changing(this, new CodeTextLineChangedEventArgs(offset, length));
			}

			if (this.PropertyChanging != null)
			{
				this.PropertyChanging(this, new PropertyChangingEventArgs("FormattedText"));
			}
		}

		private void RaiseChanged(int offset, int length)
		{
			if (this.Changed != null)
			{
				this.Changed(this, new CodeTextLineChangedEventArgs(offset, length));
			}

			if (this.PropertyChanged != null)
			{
				this.PropertyChanged(this, new PropertyChangedEventArgs("FormattedText"));
			}
		}

		#endregion

		#region Constructors

		[DebuggerStepThrough]
		public CodeTextLine()
		{
			Initialize(string.Empty, 0);
		}

		[DebuggerStepThrough]
		public CodeTextLine(int capacity)
		{
			Initialize(string.Empty, capacity);
		}

		[DebuggerStepThrough]
		public CodeTextLine(string value)
		{
			Initialize(value, 0);
		}

		#endregion

		#region Actions

		public void Insert(int offset, string value)
		{
			char[] array = value.ToCharArray();

			Insert(offset, array, 0, array.Length);
		}

		public void Insert(int offset, ArraySegment<char> value)
		{
			Insert(offset, value.Array, value.Offset, value.Count);
		}

		public void Insert(int offset, char[] value)
		{
			Insert(offset, value, 0, value.Length);
		}

		public void Insert(int offset, char[] value, int valueOffset, int valueLength)
		{
			RaiseChanging(offset, valueLength);
			this._text.Insert(offset, value, valueOffset, valueLength);
			InsertFormatting(offset, valueLength);
			RaiseChanged(offset, valueLength);
		}

		public void Insert(int offset, char value)
		{
			RaiseChanging(offset, 1);
			this._text.Insert(offset, value);
			InsertFormatting(offset, 1);
			RaiseChanged(offset, 1);
		}

		public void Remove(int offset, int length)
		{
			RaiseChanging(offset, -length);
			this._text.Remove(offset, length);
			RemoveFormatting(offset, length);
			RaiseChanged(offset, -length);
		}

		public void Format(int offset, int length, T format)
		{
			RaiseChanging(offset, length);
			SetFormatting(offset, length, format);
			RaiseChanged(offset, length);
		}

		#endregion

		#endregion

		#region Properties

		public string PlainText
		{
			[DebuggerStepThrough]
			get
			{
				return this._text.ToString();
			}

			[DebuggerStepThrough]
			set
			{
				Initialize(value, 0);
			}
		}

		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		public IEnumerable<CodeTextLineSegment<T>> FormattedText
		{
			[DebuggerStepThrough]
			get
			{
				int currentOffset = 0;

				foreach (CodeTextLineSegmentInfo<T> range in this._formatting)
				{
					yield return
						new CodeTextLineSegment<T>(
							range.Format,
							this._text.ToString(currentOffset, range.Length));

					currentOffset += range.Length;
				}

				yield break;
			}
		}

		public int Length
		{
			[DebuggerStepThrough]
			get { return this._text.Length; }
		}

		#endregion

		#region Events

		public event EventHandler<CodeTextLineChangedEventArgs> Changing;
		public event EventHandler<CodeTextLineChangedEventArgs> Changed;

		#endregion

		#endregion

		#region INotifyPropertyChanging

		public event PropertyChangingEventHandler PropertyChanging;

		#endregion

		#region INotifyPropertyChanged

		public event PropertyChangedEventHandler PropertyChanged;

		#endregion

		#region ICloneable

		public object Clone()
		{
			return new CodeTextLine<T>(this._text.ToString());
		}

		#endregion

		#region Object

		[DebuggerStepThrough]
		public override string ToString()
		{
			return this._text.ToString();
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
