﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Text Library                             */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Diagnostics;

namespace Nabu.Text
{
	public sealed class XmlLexicalAnalyzer :
		DisposableBase,
		ICodeLexicalAnalyzer
	{
		#region XmlLexicalAnalyzer

		#region Types

		private sealed class XmlTokenizer :
			DisposableBase,
			IEnumerator<CodeLexicalToken>
		{
			#region XmlTokenizer

			#region Types

			private enum XmlTokenizerState
			{
				#region Internal Use

				Unexpected,
				PreEndOfSource,
				EndOfSource,

				#endregion

				#region Whitespace

				// |
				Whitespace,

				// <| tag> <| tag/>
				Whitespace_Before_OpenTagName_Or_EmptyTagName,

				// </| tag>
				Whitespace_Before_CloseTagName,

				// </tag| >
				Whitespace_Before_TagEnd,

				// <tag| >  <tag| key=value>
				Whitespace_Before_AttributeName_Or_TagEnd,

				// <tag key| =value>
				Whitespace_Before_AttributeEqual,

				// <tag key=| value>
				Whitespace_Before_AttributeValue,

				#endregion

				#region Tag

				// <|tag> <|tag/>
				OpenTag_Or_EmptyTag_Start,

				// </|tag>
				CloseTag_Start,

				// <tag|> <tag /|>
				OpenTag_Or_EmptyTag_End,

				// <tag|/>
				EmptyTag_End1,

				// <tag/|>
				EmptyTag_End2,

				// < |tag> < |tag/>
				OpenTag_Or_EmptyTag_Name,

				// </ |tag>
				CloseTag_Name,

				#endregion

				#region Attribute

				// <tag |key=value>
				Attribute_Name,

				// <tag key|=value>
				Attribute_Equal,

				// <tag key=|'value'>
				Attribute_Value_StartSingleQuotes1,

				// <tag key='|value'>
				Attribute_Value_StartSingleQuotes2,

				// <tag key=|"value">
				Attribute_Value_StartDoubleQuotes1,

				// <tag key="|value">
				Attribute_Value_StartDoubleQuotes2,

				// <tag key='value|'>
				Attribute_Value_EndSingleQuotes,

				// <tag key="value|">
				Attribute_Value_EndDoubleQuotes,

				// <tag key=|value>
				Attribute_Value_NoQuotes,

				// <tag key='v|alue'>
				Attribute_Value_SingleQuotes,

				// <tag key="v|alue">
				Attribute_Value_DoubleQuotes,

				#endregion

				#region Comment

				// <!|--
				Comment_Start1,

				// <!-|-
				Comment_Start2,

				// |-->
				Comment_End1,

				// -|->
				Comment_End2,

				// --|>
				Comment_End3,

				// <!-- | -->
				Comment_Body,

				#endregion

				#region Processing Instriction

				// <|?
				ProcessingInstriction_Start,

				// <?|
				ProcessingInstriction_Body,

				// |?>
				ProcessingInstriction_End1,

				// ?|>
				ProcessingInstriction_End2,

				#endregion

				#region Entity

				// |&entity;
				Entity_Ampersand,

				// &|entity;
				Entity_Body,

				// &entity|;
				Entity_Semicolon

				#endregion
			}

			private class XmlLexicalAnalyzerState
			{
				#region Fields

				private long _position;
				private XmlTokenizerState _state;
				private string _tokenValue;

				#endregion

				#region Methods

				[DebuggerStepThrough]
				public XmlLexicalAnalyzerState(
					long position,
					XmlTokenizerState state,
					string tokenValue)
				{
					this._position = position;
					this._state = state;
					this._tokenValue = tokenValue;
				}

				#endregion

				#region Properties

				public long Position
				{
					[DebuggerStepThrough]
					get { return this._position; }
				}

				public XmlTokenizerState State
				{
					[DebuggerStepThrough]
					get { return this._state; }
				}

				public string TokenValue
				{
					[DebuggerStepThrough]
					get { return this._tokenValue; }
				}

				#endregion
			};

			#endregion

			#region Fields

			private ICodeEnumerator _codeEnumerator;
			private CodeLexicalToken _token;
			private XmlTokenizerState _state = XmlTokenizerState.Whitespace;
			private string _tokenValue = string.Empty;
			private long _tokenPositionFrom;

			#endregion

			#region Methods

			private void BuildCurrentToken(
				string type,
				bool includeValue)
			{
				base.CheckDisposed();

				if (includeValue)
				{
					this._token =
						new CodeLexicalToken(
							this._tokenPositionFrom,
							this._codeEnumerator.Position,
							type,
							this._tokenValue);

					this._tokenValue = string.Empty;
				}
				else
				{
					this._token =
						new CodeLexicalToken(
							this._tokenPositionFrom,
							this._codeEnumerator.Position,
							type);
				}

				this._tokenPositionFrom = this._codeEnumerator.Position;
			}

			private void ProcessSymbol(char symbol)
			{
				switch (this._state)
				{
					#region Unexpected

					case XmlTokenizerState.Unexpected:
						switch (symbol)
						{
							case '<':
								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_Start;
								break;
							case '&':
								this._state = XmlTokenizerState.Entity_Ampersand;
								break;
							default:
								break;
						}
						break;

					#endregion

					#region Whitespace

					case XmlTokenizerState.Whitespace:
						switch (symbol)
						{
							case '<':
								if (this._tokenValue.Length > 0)
								{
									BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
								}

								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_Start;
								break;
							case '>':
								this._tokenValue += symbol;
								BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
								this._state = XmlTokenizerState.Unexpected;
								break;
							case '&':
								if (this._tokenValue.Length > 0)
								{
									BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
								}

								this._state = XmlTokenizerState.Entity_Ampersand;
								break;
							default:
								this._tokenValue += symbol;
								break;
						}
						break;

					#endregion

					#region Whitespace_Before_OpenTagName_Or_EmptyTagName

					case XmlTokenizerState.Whitespace_Before_OpenTagName_Or_EmptyTagName:
						if (char.IsWhiteSpace(symbol))
						{
							this._tokenValue += symbol;
						}
						else if (char.IsLetterOrDigit(symbol))
						{
							if (this._tokenValue.Length > 0)
							{
								BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
							}

							this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_Name;
							this._tokenValue += symbol;
						}
						else
						{
							this._tokenValue += symbol;
							BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
							this._state = XmlTokenizerState.Unexpected;
						}
						break;

					#endregion

					#region Whitespace_Before_CloseTagName

					case XmlTokenizerState.Whitespace_Before_CloseTagName:
						if (char.IsWhiteSpace(symbol))
						{
							this._tokenValue += symbol;
						}
						else if (char.IsLetterOrDigit(symbol))
						{
							if (this._tokenValue.Length > 0)
							{
								BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
							}

							this._state = XmlTokenizerState.CloseTag_Name;
							this._tokenValue += symbol;
						}
						else
						{
							this._tokenValue += symbol;
							BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
							this._state = XmlTokenizerState.Unexpected;
						}
						break;

					#endregion

					#region Whitespace_Before_TagEnd

					case XmlTokenizerState.Whitespace_Before_TagEnd:
						switch (symbol)
						{
							case '>':
								if (this._tokenValue.Length > 0)
								{
									BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
								}

								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_End;
								break;
							default:
								if (char.IsWhiteSpace(symbol))
								{
									this._tokenValue += symbol;
								}
								else
								{
									this._tokenValue += symbol;
									BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
									this._state = XmlTokenizerState.Unexpected;
								}
								break;
						}
						break;

					#endregion

					#region Whitespace_Before_AttributeName_Or_TagEnd

					case XmlTokenizerState.Whitespace_Before_AttributeName_Or_TagEnd:
						switch (symbol)
						{
							case '>':
								if (this._tokenValue.Length > 0)
								{
									BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
								}

								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_End;
								break;
							case '/':
								if (this._tokenValue.Length > 0)
								{
									BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
								}

								this._state = XmlTokenizerState.EmptyTag_End1;
								break;
							default:
								if (char.IsWhiteSpace(symbol))
								{
									this._tokenValue += symbol;
								}
								else if (char.IsLetterOrDigit(symbol))
								{
									if (this._tokenValue.Length > 0)
									{
										BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
									}

									this._state = XmlTokenizerState.Attribute_Name;
									this._tokenValue += symbol;
								}
								else
								{
									this._tokenValue += symbol;
									BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
									this._state = XmlTokenizerState.Unexpected;
								}
								break;
						}
						break;

					#endregion

					#region Whitespace_Before_AttributeEqual

					case XmlTokenizerState.Whitespace_Before_AttributeEqual:
						switch (symbol)
						{
							case '=':
								if (this._tokenValue.Length > 0)
								{
									BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
								}

								this._state = XmlTokenizerState.Attribute_Equal;
								break;
							case '>':
								if (this._tokenValue.Length > 0)
								{
									BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
								}

								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_End;
								break;
							case '/':
								if (this._tokenValue.Length > 0)
								{
									BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
								}

								this._state = XmlTokenizerState.EmptyTag_End1;
								break;
							default:
								if (char.IsWhiteSpace(symbol))
								{
									this._tokenValue += symbol;
								}
								else if (char.IsLetterOrDigit(symbol))
								{
									if (this._tokenValue.Length > 0)
									{
										BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
									}

									this._tokenValue += symbol;
									this._state = XmlTokenizerState.Attribute_Name;
								}
								else
								{
									this._tokenValue += symbol;
									BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
									this._state = XmlTokenizerState.Unexpected;
								}
								break;
						}
						break;

					#endregion

					#region Whitespace_Before_AttributeValue

					case XmlTokenizerState.Whitespace_Before_AttributeValue:
						switch (symbol)
						{
							case '\'':
								if (this._tokenValue.Length > 0)
								{
									BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
								}

								this._state = XmlTokenizerState.Attribute_Value_StartSingleQuotes1;
								break;
							case '"':
								if (this._tokenValue.Length > 0)
								{
									BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
								}

								this._state = XmlTokenizerState.Attribute_Value_StartDoubleQuotes1;
								break;
							default:
								if (char.IsWhiteSpace(symbol))
								{
									this._tokenValue += symbol;
								}
								else
								{
									BuildCurrentToken(XmlLexicalTokenType.AttributeEqual, false);
									this._state = XmlTokenizerState.Attribute_Value_NoQuotes;
									this._tokenValue += symbol;
								}
								break;
						}
						break;

					#endregion

					#region OpenTag_Or_EmptyTag_Start

					case XmlTokenizerState.OpenTag_Or_EmptyTag_Start:
						switch (symbol)
						{
							case '/':
								this._state = XmlTokenizerState.CloseTag_Start;
								break;
							case '!':
								this._state = XmlTokenizerState.Comment_Start1;
								break;
							case '?':
								this._state = XmlTokenizerState.ProcessingInstriction_Start;
								break;
							default:
								if (char.IsWhiteSpace(symbol))
								{
									BuildCurrentToken(XmlLexicalTokenType.TagStart, false);
									this._state = XmlTokenizerState.Whitespace_Before_OpenTagName_Or_EmptyTagName;
									this._tokenValue += symbol;
								}
								else if (char.IsLetterOrDigit(symbol))
								{
									BuildCurrentToken(XmlLexicalTokenType.TagStart, false);
									this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_Name;
									this._tokenValue += symbol;
								}
								else
								{
									this._tokenValue += symbol;
									BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
									this._state = XmlTokenizerState.Unexpected;
								}
								break;
						}
						break;

					#endregion

					#region CloseTag_Start

					case XmlTokenizerState.CloseTag_Start:
						BuildCurrentToken(XmlLexicalTokenType.CloseTagStart, false);

						switch (symbol)
						{
							default:
								if (char.IsWhiteSpace(symbol))
								{
									this._state = XmlTokenizerState.Whitespace_Before_CloseTagName;
									this._tokenValue += symbol;
								}
								else if (char.IsLetterOrDigit(symbol))
								{
									if (this._tokenValue.Length > 0)
									{
										BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
									}

									this._state = XmlTokenizerState.CloseTag_Name;
									this._tokenValue += symbol;
								}
								else
								{
									this._tokenValue += symbol;
									BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
									this._state = XmlTokenizerState.Unexpected;
								}
								break;
						}
						break;

					#endregion

					#region OpenTag_Or_EmptyTag_End

					case XmlTokenizerState.OpenTag_Or_EmptyTag_End:
						switch (symbol)
						{
							case '<':
								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_Start;
								break;
							default:
								this._state = XmlTokenizerState.Whitespace;
								this._tokenValue += symbol;
								break;
						}

						BuildCurrentToken(XmlLexicalTokenType.TagEnd, false);
						break;

					#endregion

					#region EmptyTag_End1

					case XmlTokenizerState.EmptyTag_End1:
						switch (symbol)
						{
							case '>':
								this._state = XmlTokenizerState.EmptyTag_End2;
								break;
							default:
								this._tokenValue += symbol;
								BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
								this._state = XmlTokenizerState.Unexpected;
								break;
						}
						break;

					#endregion

					#region EmptyTag_End2

					case XmlTokenizerState.EmptyTag_End2:
						switch (symbol)
						{
							case '<':
								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_Start;
								break;
							default:
								this._state = XmlTokenizerState.Whitespace;
								break;
						}

						BuildCurrentToken(XmlLexicalTokenType.EmptyTagEnd, false);
						break;

					#endregion

					#region OpenTag_Or_EmptyTag_Name

					case XmlTokenizerState.OpenTag_Or_EmptyTag_Name:
						switch (symbol)
						{
							case '>':
								BuildCurrentToken(XmlLexicalTokenType.TagName, true);
								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_End;
								break;
							case '/':
								BuildCurrentToken(XmlLexicalTokenType.TagName, true);
								this._state = XmlTokenizerState.EmptyTag_End1;
								break;
							default:
								if (char.IsWhiteSpace(symbol))
								{
									BuildCurrentToken(XmlLexicalTokenType.TagName, true);
									this._state = XmlTokenizerState.Whitespace_Before_AttributeName_Or_TagEnd;
									this._tokenValue += symbol;
								}
								else if (char.IsLetterOrDigit(symbol))
								{
									this._tokenValue += symbol;
								}
								else
								{
									this._tokenValue += symbol;
									BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
									this._state = XmlTokenizerState.Unexpected;
								}
								break;
						}
						break;

					#endregion

					#region CloseTag_Name

					case XmlTokenizerState.CloseTag_Name:
						switch (symbol)
						{
							case '>':
								BuildCurrentToken(XmlLexicalTokenType.TagName, true);
								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_End;
								break;
							default:
								if (char.IsWhiteSpace(symbol))
								{
									BuildCurrentToken(XmlLexicalTokenType.TagName, true);
									this._state = XmlTokenizerState.Whitespace_Before_TagEnd;
									this._tokenValue += symbol;
								}
								else if (char.IsLetterOrDigit(symbol))
								{
									this._tokenValue += symbol;
								}
								else
								{
									this._tokenValue += symbol;
									BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
									this._state = XmlTokenizerState.Unexpected;
								}
								break;
						}
						break;

					#endregion

					#region Attribute_Name

					case XmlTokenizerState.Attribute_Name:
						switch (symbol)
						{
							case '=':
								BuildCurrentToken(XmlLexicalTokenType.AttributeName, true);
								this._state = XmlTokenizerState.Attribute_Equal;
								break;
							case '>':
								BuildCurrentToken(XmlLexicalTokenType.AttributeName, true);
								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_End;
								break;
							case '/':
								BuildCurrentToken(XmlLexicalTokenType.AttributeName, true);
								this._state = XmlTokenizerState.EmptyTag_End1;
								break;
							default:
								if (char.IsWhiteSpace(symbol))
								{
									BuildCurrentToken(XmlLexicalTokenType.AttributeName, true);
									this._state = XmlTokenizerState.Whitespace_Before_AttributeEqual;
									this._tokenValue += symbol;
								}
								else if (char.IsLetterOrDigit(symbol))
								{
									this._tokenValue += symbol;
								}
								else
								{
									this._tokenValue += symbol;
									BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
									this._state = XmlTokenizerState.Unexpected;
								}
								break;
						}
						break;

					#endregion

					#region Attribute_Equal

					case XmlTokenizerState.Attribute_Equal:
						switch (symbol)
						{
							case '\'':
								BuildCurrentToken(XmlLexicalTokenType.AttributeEqual, false);
								this._state = XmlTokenizerState.Attribute_Value_StartSingleQuotes1;
								break;
							case '"':
								BuildCurrentToken(XmlLexicalTokenType.AttributeEqual, false);
								this._state = XmlTokenizerState.Attribute_Value_StartDoubleQuotes1;
								break;
							default:
								if (char.IsWhiteSpace(symbol))
								{
									BuildCurrentToken(XmlLexicalTokenType.AttributeEqual, false);
									this._state = XmlTokenizerState.Whitespace_Before_AttributeValue;
									this._tokenValue += symbol;
								}
								else
								{
									BuildCurrentToken(XmlLexicalTokenType.AttributeEqual, false);
									this._state = XmlTokenizerState.Attribute_Value_NoQuotes;
									this._tokenValue += symbol;
								}
								break;
						}
						break;

					#endregion

					#region Attribute_Value_StartSingleQuotes1

					case XmlTokenizerState.Attribute_Value_StartSingleQuotes1:
						switch (symbol)
						{
							case '\'':
								BuildCurrentToken(XmlLexicalTokenType.AttributeValueStartSingleQuote, false);
								this._state = XmlTokenizerState.Attribute_Value_EndSingleQuotes;
								break;
							default:
								BuildCurrentToken(XmlLexicalTokenType.AttributeValueStartSingleQuote, false);
								this._state = XmlTokenizerState.Attribute_Value_StartSingleQuotes2;
								this._tokenValue += symbol;
								break;
						}
						break;

					#endregion

					#region Attribute_Value_StartSingleQuotes2

					case XmlTokenizerState.Attribute_Value_StartSingleQuotes2:
						switch (symbol)
						{
							case '\'':
								BuildCurrentToken(XmlLexicalTokenType.AttributeValue, true);
								this._state = XmlTokenizerState.Attribute_Value_EndSingleQuotes;
								break;
							default:
								this._state = XmlTokenizerState.Attribute_Value_SingleQuotes;
								this._tokenValue += symbol;
								break;
						}
						break;

					#endregion

					#region Attribute_Value_StartDoubleQuotes

					case XmlTokenizerState.Attribute_Value_StartDoubleQuotes1:
						switch (symbol)
						{
							case '"':
								BuildCurrentToken(XmlLexicalTokenType.AttributeValueStartDoubleQuote, false);
								this._state = XmlTokenizerState.Attribute_Value_EndDoubleQuotes;
								break;
							default:
								BuildCurrentToken(XmlLexicalTokenType.AttributeValueStartDoubleQuote, false);
								this._state = XmlTokenizerState.Attribute_Value_StartDoubleQuotes2;
								this._tokenValue += symbol;
								break;
						}
						break;

					#endregion

					#region Attribute_Value_StartDoubleQuotes2

					case XmlTokenizerState.Attribute_Value_StartDoubleQuotes2:
						switch (symbol)
						{
							case '"':
								BuildCurrentToken(XmlLexicalTokenType.AttributeValue, true);
								this._state = XmlTokenizerState.Attribute_Value_EndDoubleQuotes;
								break;
							default:
								this._state = XmlTokenizerState.Attribute_Value_DoubleQuotes;
								this._tokenValue += symbol;
								break;
						}
						break;

					#endregion

					#region Attribute_Value_EndSingleQuotes

					case XmlTokenizerState.Attribute_Value_EndSingleQuotes:
						switch (symbol)
						{
							case '>':
								BuildCurrentToken(XmlLexicalTokenType.AttributeValueEndSingleQuote, false);
								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_End;
								break;
							case '/':
								BuildCurrentToken(XmlLexicalTokenType.AttributeValueEndSingleQuote, false);
								this._state = XmlTokenizerState.EmptyTag_End1;
								break;
							default:
								if (char.IsWhiteSpace(symbol))
								{
									BuildCurrentToken(XmlLexicalTokenType.AttributeValueEndSingleQuote, false);
									this._state = XmlTokenizerState.Whitespace_Before_AttributeName_Or_TagEnd;
									this._tokenValue += symbol;
								}
								else if (char.IsLetterOrDigit(symbol))
								{
									BuildCurrentToken(XmlLexicalTokenType.AttributeValueEndSingleQuote, false);
									this._state = XmlTokenizerState.Attribute_Name;
									this._tokenValue += symbol;
								}
								else
								{
									this._tokenValue += symbol;
									BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
									this._state = XmlTokenizerState.Unexpected;
								}
								break;
						}
						break;

					#endregion

					#region Attribute_Value_EndDoubleQuotes

					case XmlTokenizerState.Attribute_Value_EndDoubleQuotes:
						switch (symbol)
						{
							case '>':
								BuildCurrentToken(XmlLexicalTokenType.AttributeValueEndDoubleQuote, false);
								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_End;
								break;
							case '/':
								BuildCurrentToken(XmlLexicalTokenType.AttributeValueEndDoubleQuote, false);
								this._state = XmlTokenizerState.EmptyTag_End1;
								break;
							default:
								if (char.IsWhiteSpace(symbol))
								{
									BuildCurrentToken(XmlLexicalTokenType.AttributeValueEndDoubleQuote, false);
									this._state = XmlTokenizerState.Whitespace_Before_AttributeName_Or_TagEnd;
									this._tokenValue += symbol;
								}
								else if (char.IsLetterOrDigit(symbol))
								{
									BuildCurrentToken(XmlLexicalTokenType.AttributeValueEndDoubleQuote, false);
									this._state = XmlTokenizerState.Attribute_Name;
									this._tokenValue += symbol;
								}
								else
								{
									this._tokenValue += symbol;
									BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
									this._state = XmlTokenizerState.Unexpected;
								}
								break;
						}
						break;

					#endregion

					#region Attribute_Value_NoQuotes

					case XmlTokenizerState.Attribute_Value_NoQuotes:
						switch (symbol)
						{
							case '>':
								BuildCurrentToken(XmlLexicalTokenType.AttributeValue, true);
								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_End;
								break;
							default:
								if (char.IsWhiteSpace(symbol))
								{
									BuildCurrentToken(XmlLexicalTokenType.AttributeValue, true);
									this._state = XmlTokenizerState.Whitespace_Before_AttributeName_Or_TagEnd;
									this._tokenValue += symbol;
								}
								else if (char.IsLetterOrDigit(symbol))
								{
									this._tokenValue += symbol;
								}
								else
								{
									this._tokenValue += symbol;
									BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
									this._state = XmlTokenizerState.Unexpected;
								}
								break;
						}
						break;

					#endregion

					#region Attribute_Value_SingleQuotes

					case XmlTokenizerState.Attribute_Value_SingleQuotes:
						switch (symbol)
						{
							case '\'':
								BuildCurrentToken(XmlLexicalTokenType.AttributeValue, true);
								this._state = XmlTokenizerState.Attribute_Value_EndSingleQuotes;
								break;
							default:
								this._tokenValue += symbol;
								break;
						}
						break;

					#endregion

					#region Attribute_Value_DoubleQuotes

					case XmlTokenizerState.Attribute_Value_DoubleQuotes:
						switch (symbol)
						{
							case '"':
								BuildCurrentToken(XmlLexicalTokenType.AttributeValue, true);
								this._state = XmlTokenizerState.Attribute_Value_EndDoubleQuotes;
								break;
							default:
								this._tokenValue += symbol;
								break;
						}
						break;

					#endregion

					#region Comment_Start1

					case XmlTokenizerState.Comment_Start1:
						switch (symbol)
						{
							case '-':
								this._state = XmlTokenizerState.Comment_Start2;
								break;
							default:
								this._tokenValue += symbol;
								BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
								this._state = XmlTokenizerState.Unexpected;
								break;
						}
						break;

					#endregion

					#region Comment_Start2

					case XmlTokenizerState.Comment_Start2:
						switch (symbol)
						{
							case '-':
								BuildCurrentToken(XmlLexicalTokenType.CommentStart, false);
								this._state = XmlTokenizerState.Comment_Body;
								break;
							default:
								this._tokenValue += symbol;
								BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
								this._state = XmlTokenizerState.Unexpected;
								break;
						}
						break;

					#endregion

					#region Comment_End1

					case XmlTokenizerState.Comment_End1:
						switch (symbol)
						{
							case '-':
								this._state = XmlTokenizerState.Comment_End2;
								break;
							default:
								this._state = XmlTokenizerState.Comment_Body;
								break;
						}
						break;

					#endregion

					#region Comment_End2

					case XmlTokenizerState.Comment_End2:
						switch (symbol)
						{
							case '>':
								BuildCurrentToken(XmlLexicalTokenType.CommentBody, true);
								this._state = XmlTokenizerState.Comment_End3;
								break;
							default:
								this._state = XmlTokenizerState.Comment_Body;
								break;
						}
						break;

					#endregion

					#region Comment_End3

					case XmlTokenizerState.Comment_End3:
						switch (symbol)
						{
							case '<':
								BuildCurrentToken(XmlLexicalTokenType.CommentEnd, false);
								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_Start;
								break;
							case '>':
								this._tokenValue += symbol;
								BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
								this._state = XmlTokenizerState.Unexpected;
								break;
							case '&':
								BuildCurrentToken(XmlLexicalTokenType.CommentEnd, false);
								this._state = XmlTokenizerState.Entity_Ampersand;
								break;
							default:
								BuildCurrentToken(XmlLexicalTokenType.CommentEnd, false);
								this._state = XmlTokenizerState.Whitespace;
								this._tokenValue += symbol;
								break;
						}
						break;

					#endregion

					#region Comment_Body

					case XmlTokenizerState.Comment_Body:
						switch (symbol)
						{
							case '-':
								this._state = XmlTokenizerState.Comment_End1;
								break;
							default:
								this._tokenValue += symbol;
								break;
						}
						break;

					#endregion

					#region ProcessingInstriction_Start

					case XmlTokenizerState.ProcessingInstriction_Start:
						switch (symbol)
						{
							case '?':
								BuildCurrentToken(XmlLexicalTokenType.ProcessingInstructionStart, false);
								this._state = XmlTokenizerState.ProcessingInstriction_End1;
								break;
							default:
								this._tokenValue += symbol;
								BuildCurrentToken(XmlLexicalTokenType.ProcessingInstructionStart, false);
								this._state = XmlTokenizerState.ProcessingInstriction_Body;
								break;
						}
						break;

					#endregion

					#region ProcessingInstriction_Body

					case XmlTokenizerState.ProcessingInstriction_Body:
						switch (symbol)
						{
							case '?':
								BuildCurrentToken(XmlLexicalTokenType.ProcessingInstructionBody, true);
								this._state = XmlTokenizerState.ProcessingInstriction_End1;
								break;
							default:
								this._tokenValue += symbol;
								break;
						}
						break;

					#endregion

					#region ProcessingInstriction_End1

					case XmlTokenizerState.ProcessingInstriction_End1:
						switch (symbol)
						{
							case '>':
								this._state = XmlTokenizerState.ProcessingInstriction_End2;
								break;
							default:
								BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
								this._state = XmlTokenizerState.Unexpected;
								break;
						}
						break;

					#endregion

					#region ProcessingInstriction_End2

					case XmlTokenizerState.ProcessingInstriction_End2:
						switch (symbol)
						{
							case '<':
								BuildCurrentToken(XmlLexicalTokenType.ProcessingInstructionEnd, false);
								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_Start;
								break;
							case '>':
								this._tokenValue += symbol;
								BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
								this._state = XmlTokenizerState.Unexpected;
								break;
							case '&':
								BuildCurrentToken(XmlLexicalTokenType.ProcessingInstructionEnd, false);
								this._state = XmlTokenizerState.Entity_Ampersand;
								break;
							default:
								BuildCurrentToken(XmlLexicalTokenType.ProcessingInstructionEnd, false);
								this._tokenValue += symbol;
								this._state = XmlTokenizerState.Whitespace;
								break;
						}

						break;

					#endregion

					#region Entity_Ampersand

					case XmlTokenizerState.Entity_Ampersand:
						if (char.IsLetterOrDigit(symbol))
						{
							BuildCurrentToken(XmlLexicalTokenType.EntityStart, false);
							this._state = XmlTokenizerState.Entity_Body;
							this._tokenValue += symbol;
						}
						else
						{
							this._tokenValue += symbol;
							BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
							this._state = XmlTokenizerState.Unexpected;
						}
						break;

					#endregion

					#region Entity_Body

					case XmlTokenizerState.Entity_Body:
						switch (symbol)
						{
							case ';':
								BuildCurrentToken(XmlLexicalTokenType.EntityBody, true);
								this._state = XmlTokenizerState.Entity_Semicolon;
								break;
							default:
								if (char.IsLetterOrDigit(symbol))
								{
									this._tokenValue += symbol;
								}
								else
								{
									this._tokenValue += symbol;
									BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
									this._state = XmlTokenizerState.Unexpected;
								}
								break;
						}
						break;

					#endregion

					#region Entity_Semicolon

					case XmlTokenizerState.Entity_Semicolon:
						switch (symbol)
						{
							case '<':
								if (this._tokenValue.Length > 0)
								{
									BuildCurrentToken(XmlLexicalTokenType.Whitespace, true);
								}

								BuildCurrentToken(XmlLexicalTokenType.EntityEnd, false);
								this._state = XmlTokenizerState.OpenTag_Or_EmptyTag_Start;
								break;
							case '>':
								this._tokenValue += symbol;
								BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
								this._state = XmlTokenizerState.Unexpected;
								break;
							case '&':
								BuildCurrentToken(XmlLexicalTokenType.EntityEnd, false);
								this._state = XmlTokenizerState.Entity_Ampersand;
								break;
							default:
								BuildCurrentToken(XmlLexicalTokenType.EntityEnd, false);
								this._state = XmlTokenizerState.Whitespace;
								this._tokenValue += symbol;
								break;
						}
						break;

					#endregion

					default:
						BuildCurrentToken(XmlLexicalTokenType.Unexpected, true);
						this._state = XmlTokenizerState.Unexpected;
						break;
				}
			}

			public XmlTokenizer(ICodeEnumerator codeEnumerator)
			{
				this._codeEnumerator = codeEnumerator;
			}

			public object SaveState()
			{
				base.CheckDisposed();

				return new XmlLexicalAnalyzerState(this._codeEnumerator.Position, this._state, this._tokenValue);
			}

			public void RestoreState(object state)
			{
				base.CheckDisposed();

				XmlLexicalAnalyzerState xmlState = (XmlLexicalAnalyzerState)state;

				this._codeEnumerator.Position = xmlState.Position;
				this._state = xmlState.State;
				this._tokenValue = xmlState.TokenValue;
			}

			#endregion

			#endregion

			#region DisposableBase

			protected override void Dispose(bool disposing)
			{
				base.Dispose(disposing);
			}

			#endregion

			#region IEnumerator<LexicalToken>

			public CodeLexicalToken Current
			{
				[DebuggerStepThrough]
				get
				{
					base.CheckDisposed();

					return this._token;
				}
			}

			#endregion

			#region IEnumerator

			object System.Collections.IEnumerator.Current
			{
				[DebuggerStepThrough]
				get
				{
					base.CheckDisposed();

					return this._token;
				}
			}

			public bool MoveNext()
			{
				base.CheckDisposed();

				if (this._state == XmlTokenizerState.EndOfSource)
				{
					return false;
				}
				else if (this._state == XmlTokenizerState.PreEndOfSource)
				{
					BuildCurrentToken(XmlLexicalTokenType.EndOfSource, false);
					this._state = XmlTokenizerState.EndOfSource;

					return true;
				}
				else
				{
					for (this._token = null; this._token == null; )
					{
						if (this._codeEnumerator.MoveNext())
						{
							ProcessSymbol(this._codeEnumerator.Current);
						}
						else
						{
							ProcessSymbol('\n');
							this._state = XmlTokenizerState.PreEndOfSource;
						}
					}

					return true;
				}
			}

			public void Reset()
			{
				base.CheckDisposed();

				this._codeEnumerator.Reset();
				this._token = null;
				this._state = XmlTokenizerState.Whitespace;
				this._tokenValue = string.Empty;
				this._tokenPositionFrom = 0;
			}

			#endregion
		};

		#endregion

		#region Fields

		private ICodeEnumerator _codeEnumerator;
		private XmlTokenizer _xmlTokenizer;

		#endregion

		#region Methods

		[DebuggerStepThrough]
		public XmlLexicalAnalyzer(ICodeEnumerator codeEnumerator)
		{
			this._codeEnumerator = codeEnumerator;
			this._xmlTokenizer = new XmlTokenizer(this._codeEnumerator);
		}

		#endregion

		#endregion

		#region DisposableBase

		protected override void Dispose(bool disposing)
		{
			if (disposing)
			{
				if (this._codeEnumerator != null)
				{
					this._codeEnumerator.Dispose();
				}

				if (this._xmlTokenizer != null)
				{
					this._xmlTokenizer.Dispose();
				}
			}

			base.Dispose(disposing);
		}

		#endregion

		#region ILexicalAnalyzer

		[DebuggerStepThrough]
		public object SaveState()
		{
			base.CheckDisposed();

			return this._xmlTokenizer.SaveState();
		}

		[DebuggerStepThrough]
		public void RestoreState(object state)
		{
			base.CheckDisposed();

			this._xmlTokenizer.RestoreState(state);
		}

		#endregion

		#region IEnumerable<CodeLexicalToken>

		[DebuggerStepThrough]
		public IEnumerator<CodeLexicalToken> GetEnumerator()
		{
			base.CheckDisposed();

			return this._xmlTokenizer;
		}

		#endregion

		#region IEnumerable

		[DebuggerStepThrough]
		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
		{
			base.CheckDisposed();

			return this._xmlTokenizer;
		}

		#endregion
	};
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
