﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Text Library                             */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System.Collections.Generic;
using System.Diagnostics;

using Nabu.Collections;

namespace Nabu.Text
{
	public sealed class XmlSyntaxAnalyzer : ICodeSyntaxAnalyzer
	{
		#region XmlSyntaxAnalyzer

		#region Types

		private enum XmlSyntaxAnalyzerState
		{
			Before_Document,

			Before_OpenTagName_Or_EmptyTagName,
			Before_CloseTagName,
			Before_OpenTagEnd_Or_EmptyTagEnd_Or_AttributeName,
			Before_CloseTagEnd,

			Before_Attribute_Equal,
			Before_AttributeValue_Start,
			Before_AttributeValue_Body_SingleQuoted,
			Before_AttributeValue_Body_DoubleQuoted,
			Before_AttributeValue_End_SingleQuoted,
			Before_AttributeValue_End_DoubleQuoted,

			Before_Comment_Body,
			Before_Comment_End,

			Before_ProcessingInstruction_Body,
			Before_ProcessingInstruction_End,

			Before_Entity_Body,
			Before_Entity_End
		}

		#endregion

		#region Fields

		private IBindingTree<CodeSyntacticToken> _ast = new BindingTree<CodeSyntacticToken>();
		private ListEx<ITreeNode<CodeSyntacticToken>> _syntaxStack = new ListEx<ITreeNode<CodeSyntacticToken>>();
		private ListEx<CodeLexicalToken> _lexicalStack = new ListEx<CodeLexicalToken>();
		private XmlSyntaxAnalyzerState _state = XmlSyntaxAnalyzerState.Before_Document;
		private int _version;

		#endregion

		#region Methods

		#region Tokens

		private void PushSyntaticToken(string type, string value)
		{
			CodeSyntacticToken token =
				new CodeSyntacticToken(
					this._version,
					type,
					value);

			ITreeNode<CodeSyntacticToken> parent = this._syntaxStack.Last;
			ITreeNode<CodeSyntacticToken> node = parent.AppendChild(token);

			this._syntaxStack.Add(node);
		}

		private bool PopSyntaticToken(string type)
		{
			if (this._syntaxStack.Count <= 1)
			{
				return false;
			}
			else
			{
				if (this._syntaxStack.Last.Data.Type != type)
				{
					return false;
				}
				else
				{
					this._syntaxStack.RemoveAt(this._syntaxStack.Count - 1);

					return true;
				}
			}
		}

		#endregion

		#region Tags

		private void PushOpenOrEmptyTag(string name)
		{
			PushSyntaticToken(XmlSyntacticTokenType.OpenTag, name);
		}

		private void ConvertToEmptyTag()
		{
			this._syntaxStack.Last.Data.Type = XmlSyntacticTokenType.EmptyTag;
		}

		private void PushCloseTag(string name)
		{
			PushSyntaticToken(XmlSyntacticTokenType.CloseTag, name);
		}

		private bool PopOpenTag(string name, out string expectedName)
		{
			if (this._syntaxStack.Count <= 1)
			{
				expectedName = null;

				return false;
			}
			else
			{
				for (int index = this._syntaxStack.Count - 1; index > 0; --index)
				{
					if (this._syntaxStack[index].Data.Type == XmlSyntacticTokenType.OpenTag)
					{
						if (this._syntaxStack[index].Data.Value != name)
						{
							expectedName = this._syntaxStack[index].Data.Value;

							return false;
						}
						else
						{
							this._syntaxStack.RemoveRange(index, this._syntaxStack.Count - index);
							break;
						}
					}
				}

				expectedName = null;

				return true;
			}
		}

		private bool PopEmptyTag()
		{
			return PopSyntaticToken(XmlSyntacticTokenType.EmptyTag);
		}

		private bool PopCloseTag()
		{
			return PopSyntaticToken(XmlSyntacticTokenType.CloseTag);
		}

		private void SetTagLexicalTokens()
		{
			this._syntaxStack.Last.Data.AppendLexicalTokens(this._lexicalStack);
			this._lexicalStack.Clear();
		}

		private CodeSyntacticToken[] GetUnclosedTags()
		{
			List<CodeSyntacticToken> tags = new List<CodeSyntacticToken>();

			for (int index = 1; index < this._syntaxStack.Count; index++)
			{
				if (this._syntaxStack[index].Data.Type == XmlSyntacticTokenType.OpenTag)
				{
					tags.Add(this._syntaxStack[index].Data);
				}
			}

			return tags.ToArray();
		}

		#endregion

		#region Constructors

		[DebuggerStepThrough]
		public XmlSyntaxAnalyzer()
		{
			this._ast.Root.Data = new CodeSyntacticToken(XmlSyntacticTokenType.Root);
			this._syntaxStack.Add(this._ast.Root);
		}

		#endregion

		#endregion

		#endregion

		#region ISyntaxAnalyzer

		#region Methods

		public void ProcessLexicalToken(CodeLexicalToken lexicalToken)
		{
			this._lexicalStack.Add(lexicalToken);

			switch (this._state)
			{
				#region Before_Document

				case XmlSyntaxAnalyzerState.Before_Document:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.Whitespace:
							PushSyntaticToken(XmlSyntacticTokenType.Text, lexicalToken.Value);
							SetTagLexicalTokens();

							if (!PopSyntaticToken(XmlSyntacticTokenType.Text))
							{
								throw
									new CodeSyntaxException(
										lexicalToken,
										new KeyValuePair<string, string>(XmlSyntacticTokenType.Text, null));
							}
							break;
						case XmlLexicalTokenType.TagStart:
							this._state = XmlSyntaxAnalyzerState.Before_OpenTagName_Or_EmptyTagName;
							break;
						case XmlLexicalTokenType.CloseTagStart:
							this._state = XmlSyntaxAnalyzerState.Before_CloseTagName;
							break;
						case XmlLexicalTokenType.CommentStart:
							this._state = XmlSyntaxAnalyzerState.Before_Comment_Body;
							break;
						case XmlLexicalTokenType.ProcessingInstructionStart:
							this._state = XmlSyntaxAnalyzerState.Before_ProcessingInstruction_Body;
							break;
						case XmlLexicalTokenType.EntityStart:
							this._state = XmlSyntaxAnalyzerState.Before_Entity_Body;
							break;
						case XmlLexicalTokenType.EndOfSource:
							if (this._syntaxStack.Count > 1)
							{
								throw
									new CodeSyntaxException(
										lexicalToken,
										GetUnclosedTags(),
										new KeyValuePair<string, string>(XmlLexicalTokenType.TagStart, null));
							}
							break;
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.Whitespace, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.TagStart, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.CloseTagStart, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.CommentStart, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.ProcessingInstructionStart, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.EntityStart, null));
					}
					break;

				#endregion

				#region Before_OpenTagName_Or_EmptyTagName

				case XmlSyntaxAnalyzerState.Before_OpenTagName_Or_EmptyTagName:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.Whitespace:
							break;
						case XmlLexicalTokenType.TagName:
							this._state = XmlSyntaxAnalyzerState.Before_OpenTagEnd_Or_EmptyTagEnd_Or_AttributeName;
							PushOpenOrEmptyTag(lexicalToken.Value);
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.TagName, null));
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.Whitespace, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.TagName, null));
					}
					break;

				#endregion

				#region Before_CloseTagName

				case XmlSyntaxAnalyzerState.Before_CloseTagName:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.Whitespace:
							break;
						case XmlLexicalTokenType.TagName:
							{
								this._state = XmlSyntaxAnalyzerState.Before_CloseTagEnd;

								string expectedName;

								if (!PopOpenTag(lexicalToken.Value, out expectedName))
								{
									throw
										new CodeSyntaxException(
											lexicalToken,
											new KeyValuePair<string, string>(XmlLexicalTokenType.TagName, expectedName));
								}
								else
								{
									PushCloseTag(lexicalToken.Value);
								}
							}
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.TagName, null));
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.Whitespace, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.TagName, null));
					}
					break;

				#endregion

				#region Before_OpenTagEnd_Or_EmptyTagEnd_Or_AttributeName

				case XmlSyntaxAnalyzerState.Before_OpenTagEnd_Or_EmptyTagEnd_Or_AttributeName:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.Whitespace:
							break;
						case XmlLexicalTokenType.TagEnd:
							this._state = XmlSyntaxAnalyzerState.Before_Document;
							SetTagLexicalTokens();
							break;
						case XmlLexicalTokenType.EmptyTagEnd:
							this._state = XmlSyntaxAnalyzerState.Before_Document;
							ConvertToEmptyTag();
							SetTagLexicalTokens();

							if (!PopEmptyTag())
							{
								throw
									new CodeSyntaxException(
										lexicalToken,
										new KeyValuePair<string, string>(XmlSyntacticTokenType.EmptyTag, null));
							}
							break;
						case XmlLexicalTokenType.AttributeName:
							this._state = XmlSyntaxAnalyzerState.Before_Attribute_Equal;
							PushSyntaticToken(XmlSyntacticTokenType.AttributeName, lexicalToken.Value);
							SetTagLexicalTokens();
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.EmptyTagEnd, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeName, null));
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.Whitespace, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.EmptyTagEnd, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeName, null));
					}
					break;

				#endregion

				#region Before_CloseTagEnd

				case XmlSyntaxAnalyzerState.Before_CloseTagEnd:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.Whitespace:
							break;
						case XmlLexicalTokenType.TagEnd:
							this._state = XmlSyntaxAnalyzerState.Before_Document;
							SetTagLexicalTokens();

							if (!PopCloseTag())
							{
								throw
									new CodeSyntaxException(
										lexicalToken,
										new KeyValuePair<string, string>(XmlSyntacticTokenType.CloseTag, null));
							}
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.TagEnd, null));
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.Whitespace, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.TagEnd, null));
					}
					break;

				#endregion

				#region Before_Attribute_Equal

				case XmlSyntaxAnalyzerState.Before_Attribute_Equal:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.Whitespace:
							break;
						case XmlLexicalTokenType.AttributeName:
							this._state = XmlSyntaxAnalyzerState.Before_Attribute_Equal;

							if (!PopSyntaticToken(XmlSyntacticTokenType.AttributeName))
							{
								throw new CodeSyntaxException(lexicalToken);
							}

							PushSyntaticToken(XmlSyntacticTokenType.AttributeName, lexicalToken.Value);
							SetTagLexicalTokens();
							break;
						case XmlLexicalTokenType.AttributeEqual:
							this._state = XmlSyntaxAnalyzerState.Before_AttributeValue_Start;
							break;
						case XmlLexicalTokenType.TagEnd:
							this._state = XmlSyntaxAnalyzerState.Before_Document;
							SetTagLexicalTokens();

							if (!PopSyntaticToken(XmlSyntacticTokenType.AttributeName))
							{
								throw new CodeSyntaxException(lexicalToken);
							}
							break;
						case XmlLexicalTokenType.EmptyTagEnd:
							this._state = XmlSyntaxAnalyzerState.Before_Document;
							ConvertToEmptyTag();
							SetTagLexicalTokens();

							if (!PopSyntaticToken(XmlSyntacticTokenType.AttributeName))
							{
								throw new CodeSyntaxException(lexicalToken);
							}
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.TagEnd, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.EmptyTagEnd, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeEqual, null));
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.Whitespace, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.TagEnd, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.EmptyTagEnd, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeEqual, null));
					}
					break;

				#endregion

				#region Before_AttributeValue_Start

				case XmlSyntaxAnalyzerState.Before_AttributeValue_Start:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.Whitespace:
							break;
						case XmlLexicalTokenType.AttributeValue:
							this._state = XmlSyntaxAnalyzerState.Before_OpenTagEnd_Or_EmptyTagEnd_Or_AttributeName;
							PushSyntaticToken(XmlSyntacticTokenType.AttributeValue, lexicalToken.Value);
							SetTagLexicalTokens();

							if (!PopSyntaticToken(XmlSyntacticTokenType.AttributeValue))
							{
								throw new CodeSyntaxException(lexicalToken);
							}

							if (!PopSyntaticToken(XmlSyntacticTokenType.AttributeName))
							{
								throw new CodeSyntaxException(lexicalToken);
							}
							break;
						case XmlLexicalTokenType.AttributeValueStartSingleQuote:
							this._state = XmlSyntaxAnalyzerState.Before_AttributeValue_Body_SingleQuoted;
							break;
						case XmlLexicalTokenType.AttributeValueStartDoubleQuote:
							this._state = XmlSyntaxAnalyzerState.Before_AttributeValue_Body_DoubleQuoted;
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValue, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValueStartSingleQuote, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValueStartDoubleQuote, null));
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.Whitespace, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValue, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValueStartSingleQuote, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValueStartDoubleQuote, null));
					}
					break;

				#endregion

				#region Before_AttributeValue_Body_SingleQuoted

				case XmlSyntaxAnalyzerState.Before_AttributeValue_Body_SingleQuoted:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.AttributeValue:
							this._state = XmlSyntaxAnalyzerState.Before_AttributeValue_End_SingleQuoted;
							PushSyntaticToken(XmlSyntacticTokenType.AttributeValue, lexicalToken.Value);
							break;
						case XmlLexicalTokenType.AttributeValueEndSingleQuote:
							this._state = XmlSyntaxAnalyzerState.Before_OpenTagEnd_Or_EmptyTagEnd_Or_AttributeName;
							PushSyntaticToken(XmlSyntacticTokenType.AttributeValue, lexicalToken.Value);
							SetTagLexicalTokens();

							if (!PopSyntaticToken(XmlSyntacticTokenType.AttributeValue))
							{
								throw new CodeSyntaxException(lexicalToken);
							}

							if (!PopSyntaticToken(XmlSyntacticTokenType.AttributeName))
							{
								throw new CodeSyntaxException(lexicalToken);
							}
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValue, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValueEndSingleQuote, null));
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValue, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValueEndSingleQuote, null));
					}
					break;

				#endregion

				#region Before_AttributeValue_Body_DoubleQuoted

				case XmlSyntaxAnalyzerState.Before_AttributeValue_Body_DoubleQuoted:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.AttributeValue:
							this._state = XmlSyntaxAnalyzerState.Before_AttributeValue_End_DoubleQuoted;
							PushSyntaticToken(XmlSyntacticTokenType.AttributeValue, lexicalToken.Value);
							break;
						case XmlLexicalTokenType.AttributeValueEndDoubleQuote:
							this._state = XmlSyntaxAnalyzerState.Before_OpenTagEnd_Or_EmptyTagEnd_Or_AttributeName;
							PushSyntaticToken(XmlSyntacticTokenType.AttributeValue, lexicalToken.Value);
							SetTagLexicalTokens();

							if (!PopSyntaticToken(XmlSyntacticTokenType.AttributeValue))
							{
								throw new CodeSyntaxException(lexicalToken);
							}

							if (!PopSyntaticToken(XmlSyntacticTokenType.AttributeName))
							{
								throw new CodeSyntaxException(lexicalToken);
							}
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValue, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValueEndDoubleQuote, null));
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValue, null),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValueEndDoubleQuote, null));
					}
					break;

				#endregion

				#region Before_AttributeValue_End_SingleQuoted

				case XmlSyntaxAnalyzerState.Before_AttributeValue_End_SingleQuoted:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.AttributeValueEndSingleQuote:
							this._state = XmlSyntaxAnalyzerState.Before_OpenTagEnd_Or_EmptyTagEnd_Or_AttributeName;
							SetTagLexicalTokens();

							if (!PopSyntaticToken(XmlSyntacticTokenType.AttributeValue))
							{
								throw new CodeSyntaxException(lexicalToken);
							}

							if (!PopSyntaticToken(XmlSyntacticTokenType.AttributeName))
							{
								throw new CodeSyntaxException(lexicalToken);
							}
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValueEndSingleQuote, null));
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValueEndSingleQuote, null));
					}
					break;

				#endregion

				#region Before_AttributeValue_End_DoubleQuoted

				case XmlSyntaxAnalyzerState.Before_AttributeValue_End_DoubleQuoted:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.AttributeValueEndDoubleQuote:
							this._state = XmlSyntaxAnalyzerState.Before_OpenTagEnd_Or_EmptyTagEnd_Or_AttributeName;
							SetTagLexicalTokens();

							if (!PopSyntaticToken(XmlSyntacticTokenType.AttributeValue))
							{
								throw new CodeSyntaxException(lexicalToken);
							}

							if (!PopSyntaticToken(XmlSyntacticTokenType.AttributeName))
							{
								throw new CodeSyntaxException(lexicalToken);
							}
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValueEndDoubleQuote, null));
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.AttributeValueEndDoubleQuote, null));
					}
					break;

				#endregion

				#region Before_Comment_Body

				case XmlSyntaxAnalyzerState.Before_Comment_Body:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.CommentBody:
							this._state = XmlSyntaxAnalyzerState.Before_Comment_End;
							PushSyntaticToken(XmlSyntacticTokenType.Comment, lexicalToken.Value);
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.CommentBody, null));
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.CommentBody, null));
					}
					break;

				#endregion

				#region Before_Comment_End

				case XmlSyntaxAnalyzerState.Before_Comment_End:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.CommentEnd:
							this._state = XmlSyntaxAnalyzerState.Before_Document;
							SetTagLexicalTokens();

							if (!PopSyntaticToken(XmlSyntacticTokenType.Comment))
							{
								throw new CodeSyntaxException(lexicalToken);
							}
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.CommentEnd, null));
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.CommentEnd, null));
					}
					break;

				#endregion

				#region Before_ProcessingInstruction_Body

				case XmlSyntaxAnalyzerState.Before_ProcessingInstruction_Body:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.ProcessingInstructionBody:
							this._state = XmlSyntaxAnalyzerState.Before_ProcessingInstruction_End;
							PushSyntaticToken(XmlSyntacticTokenType.ProcessingInstruction, lexicalToken.Value);
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.ProcessingInstructionBody, null));
						default:
							throw
								new CodeSyntaxException(
								lexicalToken,
								new KeyValuePair<string, string>(XmlLexicalTokenType.ProcessingInstructionBody, null));
					}
					break;

				#endregion

				#region Before_ProcessingInstruction_End

				case XmlSyntaxAnalyzerState.Before_ProcessingInstruction_End:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.ProcessingInstructionEnd:
							this._state = XmlSyntaxAnalyzerState.Before_Document;
							SetTagLexicalTokens();

							if (!PopSyntaticToken(XmlSyntacticTokenType.ProcessingInstruction))
							{
								throw new CodeSyntaxException(lexicalToken);
							}
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.ProcessingInstructionEnd, null));
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.ProcessingInstructionEnd, null));
					}
					break;

				#endregion

				#region Before_Entity_Body

				case XmlSyntaxAnalyzerState.Before_Entity_Body:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.EntityBody:
							this._state = XmlSyntaxAnalyzerState.Before_Entity_End;
							PushSyntaticToken(XmlSyntacticTokenType.Entity, lexicalToken.Value);
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.EntityBody, null));
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.EntityBody, null));
					}
					break;

				#endregion

				#region Before_Entity_End

				case XmlSyntaxAnalyzerState.Before_Entity_End:
					switch (lexicalToken.Kind)
					{
						case XmlLexicalTokenType.EntityEnd:
							this._state = XmlSyntaxAnalyzerState.Before_Document;
							SetTagLexicalTokens();

							if (!PopSyntaticToken(XmlSyntacticTokenType.Entity))
							{
								throw new CodeSyntaxException(lexicalToken);
							}
							break;
						case XmlLexicalTokenType.EndOfSource:
							throw
								new CodeSyntaxException(
									lexicalToken,
									GetUnclosedTags(),
									new KeyValuePair<string, string>(XmlLexicalTokenType.EntityEnd, null));
						default:
							throw
								new CodeSyntaxException(
									lexicalToken,
									new KeyValuePair<string, string>(XmlLexicalTokenType.EntityEnd, null));
					}
					break;

				#endregion

				#region Default

				default:
					throw new CodeSyntaxException(lexicalToken, null, null);

				#endregion
			}
		}

		[DebuggerStepThrough]
		public int CommitVersion()
		{
			this._version++;

			return this._version;
		}

		[DebuggerStepThrough]
		public void RollbackVersion(int version)
		{
			this._version = version;
		}

		#endregion

		#region Properties

		public IBindingTree<CodeSyntacticToken> Ast
		{
			[DebuggerStepThrough]
			get { return this._ast; }
		}

		#endregion

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
