﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Text Library                             */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Text;

namespace Nabu.Text
{
	public sealed class CSharpSourceCodeGenerator :
		ISourceCodeGenerator,
		IGeneratedSourceCodeUnit
	{
		#region CSharpSourceCodeGenerator

		#region Fields

		private StringBuilder _sourceCode = new StringBuilder();
		private IList<TextPositionMap> _positionMap = new List<TextPositionMap>();
		private IEnumerable<TemplateProperty> _properties;
		private IEnumerable<string> _imports;
		private IEnumerable<string> _modules;
		private string _workerScopeName;
		private string _workerTypeName;
		private string _workerBaseType;
		private int _previousLine;
		private int _previousColumn;
		private int _currentLine;
		private int _currentColumn;

		#endregion

		#region Methods

		[DebuggerStepThrough]
		internal CSharpSourceCodeGenerator(
			string workerScopeName,
			string workerTypeName,
			string workerBaseType,
			IEnumerable<TemplateProperty> properties,
			IEnumerable<string> imports,
			IEnumerable<string> modules)
		{
			this._workerScopeName = workerScopeName;
			this._workerTypeName = workerTypeName;
			this._workerBaseType = workerBaseType;
			this._properties = properties;
			this._imports = imports;
			this._modules = modules;
		}

		private void Append(string value)
		{
			foreach (char symbol in value)
			{
				switch (symbol)
				{
					case '\n':
						this._currentLine++;
						this._currentColumn = 0;
						this._sourceCode.Append(symbol);
						break;
					default:
						this._currentColumn++;
						this._sourceCode.Append(symbol);
						break;
				}
			}
		}

		private static string GetSafeString(string value)
		{
			StringBuilder stringBuilder = new StringBuilder(value.Length);

			foreach (char symbol in value)
			{
				switch (symbol)
				{
					case '\r':
						stringBuilder.Append("\\r");
						break;
					case '\n':
						stringBuilder.Append("\\n");
						break;
					case '\t':
						stringBuilder.Append("\\t");
						break;
					case '\"':
						stringBuilder.Append("\\\"");
						break;
					default:
						stringBuilder.Append(symbol);
						break;
				}
			}

			return stringBuilder.ToString();
		}

		private void MapPositions(TextPosition start, TextPosition end)
		{
			this._positionMap.Add(
				new TextPositionMap(
					start,
					end,
					new TextPosition(null, this._previousLine, this._previousColumn),
					new TextPosition(null, this._currentLine, this._currentColumn)));
			this._previousLine = this._currentLine;
			this._previousColumn = this._currentColumn;
		}

		#endregion

		#endregion

		#region ISourceCodeGenerator

		#region Methods

		public void BeginSource()
		{
			Append("/******************************************************************************/\r\n");
			Append("/*                                                                            */\r\n");
			Append("/* MACHINE GENERATED CODE, DO NOT EDIT MANUALLY                               */\r\n");
			Append("/*                                                                            */\r\n");
			Append(
				"/* UTC DATE AND TIME (yyyy/MM/dd HH:mm:ss): " +
				DateTime.UtcNow.ToString("yyyy/MM/dd HH:mm:ss", CultureInfo.InvariantCulture) +
				"               */\r\n");
			Append("/*                                                                            */\r\n");
			Append("/******************************************************************************/\r\n");
			Append("\r\n");
		}

		public void EndSource()
		{
		}

		public void BeginImport()
		{
		}

		public void EndImport()
		{
			Append("\r\n");
		}

		public void BeginCode()
		{
			Append("namespace " + this._workerScopeName + "\r\n");
			Append("{\r\n");

			if (this._workerBaseType == null)
			{
				Append("\tpublic sealed class " + this._workerTypeName + "\r\n");
			}
			else
			{
				Append("\tpublic sealed class " + this._workerTypeName + " : " + this._workerBaseType + "\r\n");
			}

			Append("\t{\r\n");
		}

		public void EndCode()
		{
			Append("\t\tprivate System.IO.TextWriter _OutputField;\r\n");
			Append("\r\n");
			Append("\t\tpublic System.IO.TextWriter Output\r\n");
			Append("\t\t{\r\n");
			Append("\t\t\tget { return this._OutputField; }\r\n");
			Append("\t\t\tset { this._OutputField = value; }\r\n");
			Append("\t\t}\r\n");
			Append("\r\n");

			foreach (TemplateProperty property in this._properties)
			{
				Append("\t\tprivate global::" + property.Type + " _" + property.Name + "Field;\r\n");
				Append("\r\n");
				Append("\t\tpublic global::" + property.Type + " " + property.Name + "\r\n");
				Append("\t\t{\r\n");
				Append("\t\t\tget { return this._" + property.Name + "Field; }\r\n");
				Append("\t\t\tset { this._" + property.Name + "Field = value; }\r\n");
				Append("\t\t}\r\n");
			}

			Append("\t}\r\n");
			Append("}\r\n");
		}

		public void BeginContentCode()
		{
			Append("\t\tpublic void InProcess()\r\n");
			Append("\t\t{\r\n");
			this._previousLine = this._currentLine;
			this._previousColumn = this._currentColumn;
		}

		public void EndContentCode()
		{
			Append("\t\t}\r\n");
			Append("\r\n");
		}

		public void BeginFeatureCode()
		{
			this._previousLine = this._currentLine;
			this._previousColumn = this._currentColumn;
		}

		public void EndFeatureCode()
		{
			Append("\r\n");
		}

		public void Import(string import)
		{
			Append("using " + import + ";\r\n");
		}

		public void Text(TextPosition startPosition, TextPosition endPosition, string value)
		{
			Append("\t\t\tOutput.Write(\"" + GetSafeString(value) + "\");\r\n");
			MapPositions(startPosition, endPosition);
		}

		public void Code(TextPosition startPosition, TextPosition endPosition, string value)
		{
			Append(value);
			MapPositions(startPosition, endPosition);
		}

		public void Expression(TextPosition startPosition, TextPosition endPosition, string value)
		{
			Append("\t\t\tOutput.Write(" + value + ");\r\n");
			MapPositions(startPosition, endPosition);
		}

		#endregion

		#region Properties

		public IGeneratedSourceCodeUnit SourceCode
		{
			[DebuggerStepThrough]
			get { return this; }
		}

		#endregion

		#endregion

		#region IGeneratedSourceCodeUnit

		public IEnumerable<TextPositionMap> PositionMap
		{
			[DebuggerStepThrough]
			get { return this._positionMap; }
		}

		public string Extension
		{
			[DebuggerStepThrough]
			get { return ".cs"; }
		}

		#endregion

		#region ISourceCodeUnit

		public string Value
		{
			[DebuggerStepThrough]
			get { return this._sourceCode.ToString(); }
		}

		public string Module
		{
			[DebuggerStepThrough]
			get { return this._workerTypeName; }
		}

		public string Language
		{
			[DebuggerStepThrough]
			get { return "C#"; }
		}

		public IEnumerable<string> Imports
		{
			[DebuggerStepThrough]
			get { return this._imports; }
		}

		public IEnumerable<string> Modules
		{
			[DebuggerStepThrough]
			get { return this._modules; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
