﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu.Text Library                             */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Text;
using System.Text.RegularExpressions;

namespace Nabu.Text
{
	public sealed class TemplateProcessor : ITemplateProcessor
	{
		#region TemplateProcessor

		#region Types

		#region ParserState

		private enum ParserState
		{
			//
			TextStart,
			// <|
			TextEnd,
			//
			Text,
			// \|
			TextEscape,
			// <#|
			ContentCodeStart,
			// <#=|
			ContentExpressionStart,
			// <#+|
			FeatureCodeStart,
			// <#@|
			DirectiveStart,
			// <# ... #|
			ContentCodeEnd,
			// <#= ... #|
			ContentExpressionEnd,
			// <#+ ... #|
			FeatureCodeEnd,
			// <#@ ... #|
			DirectiveEnd,
			// <# .... |
			ContentCode,
			// <#= ... |
			ContentExpression,
			// <#+ ... |
			FeatureCode,
			// <#@ ... |
			Directive,
			// <# .... \|
			ContentCodeEscape,
			// <#= ... \|
			ContentExpressionEscape,
			// <#+ ... \|
			FeatureCodeEscape,
			// <#@ ... \|
			DirectiveEscape
		}

		#endregion

		#region FragmentType

		private enum FragmentType
		{
			Text,
			Code,
			Expression
		}

		#endregion

		#region FragmentMode

		private enum FragmentMode
		{
			Text,
			Feature
		}

		#endregion

		#region WhitespaceMode

		private enum WhitespaceMode
		{
			Keep,
			Clear
		}

		#endregion

		#region TemplateContext

		private sealed class TemplateContext
		{
			#region Fields

			private string[] _content;
			private TextPosition _position;
			private FragmentMode _fragmentMode;
			private WhitespaceMode _whitespaceMode;

			#endregion

			#region Methods

			public TemplateContext(string[] content, TextPosition position)
			{
				this._content = content;
				this._position = position;
				this._fragmentMode = FragmentMode.Text;
				this._whitespaceMode = WhitespaceMode.Keep;
			}

			public bool ReadSymbol(out char symbol)
			{
				if (this._position.Line == this._content.Length)
				{
					symbol = '\0';

					return false;
				}
				else
				{
					if (this._position.Column < this._content[this._position.Line].Length)
					{
						symbol = this._content[this._position.Line][this._position.Column];
					}
					else if (this._position.Column == this._content[this._position.Line].Length)
					{
						symbol = '\r';
					}
					else
					{
						symbol = '\n';
					}

					int positionLine = this._position.Line;
					int positionColumn = this._position.Column;

					positionColumn++;

					if (
						(positionLine < this._content.Length) &&
						(positionColumn >= this._content[positionLine].Length + 2))
					{
						positionColumn = 0;
						positionLine++;
					}

					this._position =
						new TextPosition(
							this._position.File,
							positionLine,
							positionColumn);

					return true;
				}
			}

			#endregion

			#region Properties

			public string[] Content
			{
				get { return this._content; }
			}

			public TextPosition Position
			{
				get { return this._position; }
				set { this._position = value; }
			}

			public FragmentMode FragmentMode
			{
				get { return this._fragmentMode; }
				set { this._fragmentMode = value; }
			}

			public WhitespaceMode WhitespaceMode
			{
				get { return this._whitespaceMode; }
				set { this._whitespaceMode = value; }
			}

			#endregion
		}

		#endregion

		#region TemplateFragment

#if !COMPACT_FRAMEWORK
		[System.Diagnostics.DebuggerDisplay("{Start}-{End}, {Type}, {Value})")]
#endif
		private sealed class TemplateFragment
		{
			#region Fields

			private TextPosition _start;
			private TextPosition _end;
			private FragmentType _type;
			private string _value;

			#endregion

			#region Methods

			[DebuggerStepThrough]
			public TemplateFragment(
				TextPosition start,
				TextPosition end,
				FragmentType type,
				string value)
			{
				this._start = start;
				this._end = end;
				this._type = type;
				this._value = value;
			}

			#endregion

			#region Properties

			public TextPosition Start
			{
				[DebuggerStepThrough]
				get { return this._start; }
			}

			public TextPosition End
			{
				[DebuggerStepThrough]
				get { return this._end; }
			}

			public FragmentType Type
			{
				[DebuggerStepThrough]
				get { return this._type; }
			}

			public string Value
			{
				[DebuggerStepThrough]
				get { return this._value; }
			}

			#endregion
		}

		#endregion

		#region CompilationParameters

		private sealed class CompilationParameters
		{
			#region Fields

			private List<TemplateProperty> _properties = new List<TemplateProperty>();
			private List<string> _imports = new List<string>();
			private List<string> _assemblies = new List<string>();
			private string _language = "C#";
			private string _extension = ".cs";
			private string _baseType;

			#endregion

			#region Methods

			[DebuggerStepThrough]
			public CompilationParameters()
			{
			}

			#endregion

			#region Properties

			public IList<TemplateProperty> Properties
			{
				[DebuggerStepThrough]
				get { return this._properties; }
			}

			public IList<string> Imports
			{
				[DebuggerStepThrough]
				get { return this._imports; }
			}

			public IList<string> Assemblies
			{
				[DebuggerStepThrough]
				get { return this._assemblies; }
			}

			public string Language
			{
				[DebuggerStepThrough]
				get { return this._language; }
				[DebuggerStepThrough]
				set { this._language = value; }
			}

			public string Extension
			{
				[DebuggerStepThrough]
				get { return this._extension; }
				[DebuggerStepThrough]
				set { this._extension = value; }
			}

			public string BaseType
			{
				[DebuggerStepThrough]
				get { return this._baseType; }
				[DebuggerStepThrough]
				set { this._baseType = value; }
			}

			#endregion
		}

		#endregion

		#endregion

		#region Fields

		private ITemplateAccessor _templateAccessor;
		private ISourceCodeGenerator _sourceCodeGenerator;
		private List<TemplateFragment> _contentFragments;
		private List<TemplateFragment> _featureFragments;
		private CompilationParameters _compilationParameters;
		private Stack<TemplateContext> _templates;
		private Regex _regExDirective = new Regex(@"([a-z]+)(?:\s+([\S]+)\s*\=\s*\""([^\""]+)\"")+", RegexOptions.IgnoreCase);

		#endregion

		#region Methods

		#region Utility

		private void AppendFragment(
			StringBuilder blockValue,
			TextPosition blockStartPosition,
			TextPosition blockEndPosition,
			FragmentType fragmentType)
		{
			switch (this._templates.Peek().WhitespaceMode)
			{
				case WhitespaceMode.Clear:
					while (
						(blockValue.Length > 0) &&
						(blockValue[0] < ' '))
					{
						blockValue.Remove(0, 1);
					}
					break;
			}

			if (blockValue.Length > 0)
			{
				switch (this._templates.Peek().FragmentMode)
				{
					case FragmentMode.Text:
						this._contentFragments.Add(
							new TemplateFragment(
								blockStartPosition,
								blockEndPosition,
								fragmentType,
								blockValue.ToString()));
						break;
					case FragmentMode.Feature:
						this._featureFragments.Add(
							new TemplateFragment(
								blockStartPosition,
								blockEndPosition,
								fragmentType,
								blockValue.ToString()));
						break;
				}

				blockValue.Length = 0;
			}
		}

		private void ProcessDirective(
			StringBuilder blockValue,
			TextPosition blockStartPosition,
			TextPosition blockEndPosition)
		{
			string directive = blockValue.ToString();
			Match match = this._regExDirective.Match(directive);

			if (match.Success)
			{
				string directiveName = match.Groups[1].Value.ToUpper(CultureInfo.InvariantCulture);
				Dictionary<string, string> directiveParameters = new Dictionary<string, string>();

				for (int index = 0; index < match.Groups[2].Captures.Count; index++)
				{
					directiveParameters[match.Groups[2].Captures[index].Value.ToUpper(CultureInfo.InvariantCulture)] =
						match.Groups[3].Captures[index].Value;
				}

				switch (directiveName)
				{
					case "TEMPLATE":
						{
							string value;

							if (directiveParameters.TryGetValue("LANGUAGE", out value))
							{
								this._compilationParameters.Language = value;
							}

							if (directiveParameters.TryGetValue("INHERITS", out value))
							{
								this._compilationParameters.BaseType = value;
							}
						}
						break;
					case "ASSEMBLY":
						{
							string name;

							if (directiveParameters.TryGetValue("NAME", out name))
							{
								AddAssembly(name);
							}
							else
							{
								throw
									new TemplateParserException(
										TemplateParserErrorType.DirectiveParameterMissing,
										blockStartPosition,
										directive);
							}
						}
						break;
					case "IMPORT":
						{
							string import;

							if (directiveParameters.TryGetValue("NAMESPACE", out import))
							{
								AddImport(import);
							}
							else
							{
								throw
									new TemplateParserException(
										TemplateParserErrorType.DirectiveParameterMissing,
										blockStartPosition,
										directive);
							}
						}
						break;
					case "INCLUDE":
						{
							string file;

							if (directiveParameters.TryGetValue("FILE", out file))
							{
								this._templates.Push(
									new TemplateContext(
										this._templateAccessor.GetContent(file),
										new TextPosition(file, 0, 0)));
							}
							else
							{
								throw
									new TemplateParserException(
										TemplateParserErrorType.DirectiveParameterMissing,
										blockStartPosition,
										directive);
							}
						}
						break;
					case "OUTPUT":
						{
							string extension;

							if (directiveParameters.TryGetValue("EXTENSION", out extension))
							{
								this._compilationParameters.Extension = extension;
							}
							else
							{
								throw
									new TemplateParserException(
										TemplateParserErrorType.DirectiveParameterMissing,
										blockStartPosition,
										directive);
							}
						}
						break;
					case "PROPERTY":
						{
							string name;
							string type;

							if (
								directiveParameters.TryGetValue("NAME", out name) &&
								directiveParameters.TryGetValue("TYPE", out type))
							{
								string converter;
								string editor;

								if (!directiveParameters.TryGetValue("CONVERTER", out converter))
								{
									converter = null;
								}

								if (!directiveParameters.TryGetValue("EDITOR", out editor))
								{
									editor = null;
								}

								this._compilationParameters.Properties.Add(
									new TemplateProperty(
										name,
										type,
										converter,
										editor));
							}
							else
							{
								throw
									new TemplateParserException(
										TemplateParserErrorType.DirectiveParameterMissing,
										blockStartPosition,
										directive);
							}
						}
						break;
					default:
						throw
							new TemplateParserException(
								TemplateParserErrorType.UnknownDirective,
								blockStartPosition,
								directive);
				}
			}
			else
			{
				throw
					new TemplateParserException(
						TemplateParserErrorType.UnknownDirective,
						blockStartPosition,
						directive);
			}

			blockValue.Length = 0;
		}

		private bool ReadSymbol(out char symbol)
		{
			symbol = '\0';

			for (; ; )
			{
				if (this._templates.Count >= 1)
				{
					if (this._templates.Peek().ReadSymbol(out symbol))
					{
						return true;
					}
					else
					{
						if (this._templates.Count == 1)
						{
							return false;
						}
						else
						{
							this._templates.Pop();
						}
					}
				}
				else
				{
					return false;
				}
			}
		}

		#endregion

		#region Constructors

		public TemplateProcessor(
			ITemplateAccessor templateAccessor)
		{
			this._templateAccessor = templateAccessor;
		}

		#endregion

		#endregion

		#endregion

		#region ITemplateProcessor

		#region Methods

		#region Parse

		public void Parse(string templateName)
		{
			this._contentFragments = new List<TemplateFragment>();
			this._featureFragments = new List<TemplateFragment>();
			this._compilationParameters = new CompilationParameters();
			this._templates = new Stack<TemplateContext>();

			this._templates.Push(
				new TemplateContext(
					this._templateAccessor.GetContent(templateName),
					new TextPosition(templateName, 0, 0)));

			ParserState parserState = ParserState.TextStart;
			TextPosition blockStartPosition = this._templates.Peek().Position;
			TextPosition blockEndPosition = null;
			StringBuilder blockValue = new StringBuilder(1024);

			AddAssembly("mscorlib.dll");
			AddAssembly("System.dll");
			AddImport("System");

			for (; ; )
			{
				char symbol;

				if (!ReadSymbol(out symbol))
				{
					break;
				}

				switch (parserState)
				{
					#region TextStart

					//
					case ParserState.TextStart:
						switch (symbol)
						{
							case '<':
								parserState = ParserState.TextEnd;
								break;
							case '\\':
								parserState = ParserState.TextEscape;
								break;
							default:
								parserState = ParserState.Text;
								blockEndPosition = this._templates.Peek().Position;
								blockValue.Append(symbol);
								break;
						}
						break;

					#endregion

					#region TextEnd

					// <|
					case ParserState.TextEnd:
						switch (symbol)
						{
							case '#':
								parserState = ParserState.ContentCodeStart;
								AppendFragment(
									blockValue,
									blockStartPosition,
									blockEndPosition,
									FragmentType.Text);
								blockStartPosition = this._templates.Peek().Position;
								break;
							default:
								parserState = ParserState.Text;
								blockEndPosition = this._templates.Peek().Position;
								blockValue.Append('<');
								blockValue.Append(symbol);
								break;
						}
						break;

					#endregion

					#region Content

					//
					case ParserState.Text:
						switch (symbol)
						{
							case '<':
								parserState = ParserState.TextEnd;
								break;
							case '\\':
								parserState = ParserState.TextEscape;
								break;
							default:
								blockEndPosition = this._templates.Peek().Position;
								blockValue.Append(symbol);
								break;
						}
						break;

					#endregion

					#region TextEscape

					// \|
					case ParserState.TextEscape:
						parserState = ParserState.Text;
						blockEndPosition = this._templates.Peek().Position;
						blockValue.Append(symbol);
						break;

					#endregion

					#region ContentCodeStart

					// <#|
					case ParserState.ContentCodeStart:
						switch (symbol)
						{
							case '=':
								parserState = ParserState.ContentExpressionStart;
								blockStartPosition = this._templates.Peek().Position;
								this._templates.Peek().WhitespaceMode = WhitespaceMode.Keep;
								break;
							case '+':
								parserState = ParserState.FeatureCodeStart;
								blockStartPosition = this._templates.Peek().Position;
								this._templates.Peek().FragmentMode = FragmentMode.Feature;
								this._templates.Peek().WhitespaceMode = WhitespaceMode.Clear;
								break;
							case '@':
								parserState = ParserState.DirectiveStart;
								blockStartPosition = this._templates.Peek().Position;
								this._templates.Peek().WhitespaceMode = WhitespaceMode.Clear;
								break;
							default:
								if (this._templates.Peek().FragmentMode == FragmentMode.Text)
								{
									parserState = ParserState.ContentCode;
									blockEndPosition = this._templates.Peek().Position;
									this._templates.Peek().WhitespaceMode = WhitespaceMode.Keep;
									blockValue.Append(symbol);
								}
								else
								{
									throw
										new TemplateParserException(
											TemplateParserErrorType.UnexpectedSymbol,
											this._templates.Peek().Position,
											string.Format("'{0}'", symbol));
								}
								break;
						}
						break;

					#endregion

					#region ContentExpressionStart

					// <#=|
					case ParserState.ContentExpressionStart:
						switch (symbol)
						{
							case '\\':
								parserState = ParserState.ContentExpressionEscape;
								break;
							case '#':
								parserState = ParserState.ContentExpressionEnd;
								break;
							default:
								blockEndPosition = this._templates.Peek().Position;
								blockValue.Append(symbol);
								break;
						}
						break;

					#endregion

					#region FeatureCodeStart

					// <#+|
					case ParserState.FeatureCodeStart:
						switch (symbol)
						{
							case '\\':
								parserState = ParserState.FeatureCodeEscape;
								break;
							case '#':
								parserState = ParserState.FeatureCodeEnd;
								break;
							default:
								blockEndPosition = this._templates.Peek().Position;
								blockValue.Append(symbol);
								break;
						}
						break;

					#endregion

					#region DirectiveStart

					// <#@|
					case ParserState.DirectiveStart:
						switch (symbol)
						{
							case '\\':
								parserState = ParserState.DirectiveEscape;
								break;
							case '#':
								parserState = ParserState.DirectiveEnd;
								break;
							default:
								blockEndPosition = this._templates.Peek().Position;
								blockValue.Append(symbol);
								break;
						}
						break;

					#endregion

					#region ContentCodeEnd

					// <# ... #|
					case ParserState.ContentCodeEnd:
						switch (symbol)
						{
							case '>':
								parserState = ParserState.TextStart;
								AppendFragment(
									blockValue,
									blockStartPosition,
									blockEndPosition,
									FragmentType.Code);
								blockStartPosition = this._templates.Peek().Position;
								break;
							default:
								parserState = ParserState.ContentCode;
								blockEndPosition = this._templates.Peek().Position;
								blockValue.Append('#');
								blockValue.Append(symbol);
								break;
						}
						break;

					#endregion

					#region ContentExpressionEnd

					// <#= ... #|
					case ParserState.ContentExpressionEnd:
						switch (symbol)
						{
							case '>':
								parserState = ParserState.TextStart;
								AppendFragment(
									blockValue,
									blockStartPosition,
									blockEndPosition,
									FragmentType.Expression);
								blockStartPosition = this._templates.Peek().Position;
								break;
							default:
								parserState = ParserState.ContentExpression;
								blockEndPosition = this._templates.Peek().Position;
								blockValue.Append('#');
								blockValue.Append(symbol);
								break;
						}
						break;

					#endregion

					#region FeatureCodeEnd

					// <#+ ... #|
					case ParserState.FeatureCodeEnd:
						switch (symbol)
						{
							case '>':
								parserState = ParserState.TextStart;
								AppendFragment(
									blockValue,
									blockStartPosition,
									blockEndPosition,
									FragmentType.Code);
								blockStartPosition = this._templates.Peek().Position;
								break;
							default:
								parserState = ParserState.FeatureCode;
								blockEndPosition = this._templates.Peek().Position;
								blockValue.Append('#');
								blockValue.Append(symbol);
								break;
						}
						break;

					#endregion

					#region DirectiveEnd

					// <#@ ... #|
					case ParserState.DirectiveEnd:
						switch (symbol)
						{
							case '>':
								parserState = ParserState.TextStart;
								ProcessDirective(
									blockValue,
									blockStartPosition,
									blockEndPosition);
								blockStartPosition = this._templates.Peek().Position;
								break;
							default:
								parserState = ParserState.Directive;
								blockEndPosition = this._templates.Peek().Position;
								blockValue.Append('#');
								blockValue.Append(symbol);
								break;
						}
						break;

					#endregion

					#region ContentCode

					// <# .... |
					case ParserState.ContentCode:
						switch (symbol)
						{
							case '#':
								parserState = ParserState.ContentCodeEnd;
								break;
							case '\\':
								parserState = ParserState.ContentCodeEscape;
								break;
							default:
								blockEndPosition = this._templates.Peek().Position;
								blockValue.Append(symbol);
								break;
						}
						break;

					#endregion

					#region ContentExpression

					// <#= ... |
					case ParserState.ContentExpression:
						switch (symbol)
						{
							case '#':
								parserState = ParserState.ContentExpressionEnd;
								break;
							case '\\':
								parserState = ParserState.ContentExpressionEscape;
								break;
							default:
								blockEndPosition = this._templates.Peek().Position;
								blockValue.Append(symbol);
								break;
						}
						break;

					#endregion

					#region FeatureCode

					// <#+ ... |
					case ParserState.FeatureCode:
						switch (symbol)
						{
							case '#':
								parserState = ParserState.FeatureCodeEnd;
								break;
							case '\\':
								parserState = ParserState.FeatureCodeEscape;
								break;
							default:
								blockEndPosition = this._templates.Peek().Position;
								blockValue.Append(symbol);
								break;
						}
						break;

					#endregion

					#region Directive

					// <#@ ... |
					case ParserState.Directive:
						switch (symbol)
						{
							case '#':
								parserState = ParserState.DirectiveEnd;
								break;
							case '\\':
								parserState = ParserState.DirectiveEscape;
								break;
							default:
								blockEndPosition = this._templates.Peek().Position;
								blockValue.Append(symbol);
								break;
						}
						break;

					#endregion

					#region ContentCodeEscape

					// <# .... \|
					case ParserState.ContentCodeEscape:
						parserState = ParserState.ContentCode;
						blockEndPosition = this._templates.Peek().Position;
						blockValue.Append(symbol);
						break;

					#endregion

					#region ContentExpressionEscape

					// <#= ... \|
					case ParserState.ContentExpressionEscape:
						parserState = ParserState.ContentExpression;
						blockEndPosition = this._templates.Peek().Position;
						blockValue.Append(symbol);
						break;

					#endregion

					#region FeatureCodeEscape

					// <#+ ... \|
					case ParserState.FeatureCodeEscape:
						parserState = ParserState.FeatureCode;
						blockEndPosition = this._templates.Peek().Position;
						blockValue.Append(symbol);
						break;

					#endregion

					#region DirectiveEscape

					// <#@ ... \|
					case ParserState.DirectiveEscape:
						parserState = ParserState.Directive;
						blockEndPosition = this._templates.Peek().Position;
						blockValue.Append(symbol);
						break;

					#endregion
				}
			}

			AppendFragment(
				blockValue,
				blockStartPosition,
				blockEndPosition,
				FragmentType.Text);
		}

		#endregion

		#region Configure

		public void AddAssembly(string assembly)
		{
			if (!this._compilationParameters.Assemblies.Contains(assembly))
			{
				this._compilationParameters.Assemblies.Add(assembly);
			}
		}

		public void AddImport(string import)
		{
			if (!this._compilationParameters.Imports.Contains(import))
			{
				this._compilationParameters.Imports.Add(import);
			}
		}

		#endregion

		#region Generate

		public void Generate(string workerScopeName, string workerTypeName)
		{
			switch (this._compilationParameters.Language)
			{
				case "C#":
				case "C#v2.0":
				case "C#v3.0":
				case "C#v3.5":
					this._sourceCodeGenerator =
						new CSharpSourceCodeGenerator(
							workerScopeName,
							workerTypeName,
							this._compilationParameters.BaseType,
							this._compilationParameters.Properties,
							this._compilationParameters.Imports,
							this._compilationParameters.Assemblies);
					break;
				default:
					throw new KeyNotFoundException(this._compilationParameters.Language);
			}

			this._sourceCodeGenerator.BeginSource();

			this._sourceCodeGenerator.BeginImport();

			foreach (string import in this._compilationParameters.Imports)
			{
				this._sourceCodeGenerator.Import(import);
			}

			this._sourceCodeGenerator.EndImport();

			this._sourceCodeGenerator.BeginCode();

			this._sourceCodeGenerator.BeginFeatureCode();

			foreach (TemplateFragment fragment in this._featureFragments)
			{
				switch (fragment.Type)
				{
					case FragmentType.Text:
						this._sourceCodeGenerator.Text(fragment.Start, fragment.End, fragment.Value);
						break;
					case FragmentType.Code:
						this._sourceCodeGenerator.Code(fragment.Start, fragment.End, fragment.Value);
						break;
					case FragmentType.Expression:
						this._sourceCodeGenerator.Expression(fragment.Start, fragment.End, fragment.Value);
						break;
				}
			}

			this._sourceCodeGenerator.EndFeatureCode();

			this._sourceCodeGenerator.BeginContentCode();

			foreach (TemplateFragment fragment in this._contentFragments)
			{
				switch (fragment.Type)
				{
					case FragmentType.Text:
						this._sourceCodeGenerator.Text(fragment.Start, fragment.End, fragment.Value);
						break;
					case FragmentType.Code:
						this._sourceCodeGenerator.Code(fragment.Start, fragment.End, fragment.Value);
						break;
					case FragmentType.Expression:
						this._sourceCodeGenerator.Expression(fragment.Start, fragment.End, fragment.Value);
						break;
				}
			}

			this._sourceCodeGenerator.EndContentCode();

			this._sourceCodeGenerator.EndCode();

			this._sourceCodeGenerator.EndSource();
		}

		#endregion

		#endregion

		#region Properties

		public IGeneratedSourceCodeUnit SourceCode
		{
			[DebuggerStepThrough]
			get { return this._sourceCodeGenerator.SourceCode; }
		}

		public IList<TemplateProperty> Properties
		{
			[DebuggerStepThrough]
			get { return this._compilationParameters.Properties; }
		}

		public IList<string> Imports
		{
			[DebuggerStepThrough]
			get { return this._compilationParameters.Imports; }
		}

		public IList<string> Assemblies
		{
			[DebuggerStepThrough]
			get { return this._compilationParameters.Assemblies; }
		}

		public string Language
		{
			[DebuggerStepThrough]
			get { return this._compilationParameters.Language; }
		}

		public string Extension
		{
			[DebuggerStepThrough]
			get { return this._compilationParameters.Extension; }
		}

		public string BaseType
		{
			[DebuggerStepThrough]
			get { return this._compilationParameters.BaseType; }
			[DebuggerStepThrough]
			set { this._compilationParameters.BaseType = value; }
		}

		#endregion

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
