﻿using System;
using System.ComponentModel;

using Nabu.Collections;

using NUnit.Framework;

namespace Nabu.UnitTest.nUnit.Nabu.Collections
{
	public abstract class BindingListExTestBase
	{
		protected static void TestNoAdd<T>(BindingListExEventSink<T> sink)
		{
			Assert.IsTrue(
				(sink.ListChanged == null) ||
				(sink.ListChanged.ListChangedType != ListChangedType.ItemAdded));
			Assert.IsNull(sink.BeforeAdd);
			Assert.IsNull(sink.AfterAdd);
		}

		protected static void TestNoRemove<T>(BindingListExEventSink<T> sink)
		{
			Assert.IsTrue(
				(sink.ListChanged == null) ||
				(sink.ListChanged.ListChangedType != ListChangedType.ItemDeleted));
			Assert.IsNull(sink.BeforeRemove);
			Assert.IsNull(sink.AfterRemove);
		}

		protected static void TestNoChange<T>(BindingListExEventSink<T> sink)
		{
			Assert.IsTrue(
				(sink.ListChanged == null) ||
				(sink.ListChanged.ListChangedType != ListChangedType.ItemChanged));
			Assert.IsNull(sink.BeforeChange);
			Assert.IsNull(sink.AfterChange);
		}

		protected static void TestNoMove<T>(BindingListExEventSink<T> sink)
		{
			Assert.IsTrue(
				(sink.ListChanged == null) ||
				(sink.ListChanged.ListChangedType != ListChangedType.ItemMoved));
			Assert.IsNull(sink.BeforeMove);
			Assert.IsNull(sink.AfterMove);
		}

		protected static void TestNoReset<T>(BindingListExEventSink<T> sink)
		{
			Assert.IsTrue(
				(sink.ListChanged == null) ||
				(sink.ListChanged.ListChangedType != ListChangedType.Reset));
			Assert.IsNull(sink.BeforeReset);
			Assert.IsNull(sink.AfterReset);
		}

		protected static void TestAdd<T>(BindingListExEventSink<T> sink, int index)
		{
			Assert.IsNotNull(sink.ListChanged);
			Assert.IsNotNull(sink.BeforeAdd);
			Assert.IsNotNull(sink.AfterAdd);

			Assert.AreEqual(ListChangedType.ItemAdded, sink.ListChanged.ListChangedType);
			Assert.AreEqual(index, sink.ListChanged.NewIndex);
			Assert.AreEqual(index, sink.BeforeAdd.Index);
			Assert.AreEqual(index, sink.AfterAdd.Index);
		}

		protected static void TestRemove<T>(BindingListExEventSink<T> sink, int index)
		{
			Assert.IsNotNull(sink.ListChanged);
			Assert.IsNotNull(sink.BeforeRemove);
			Assert.IsNotNull(sink.AfterRemove);

			Assert.AreEqual(ListChangedType.ItemDeleted, sink.ListChanged.ListChangedType);
			Assert.AreEqual(index, sink.ListChanged.NewIndex);
			Assert.AreEqual(index, sink.BeforeRemove.Index);
			Assert.AreEqual(index, sink.AfterRemove.Index);
		}

		protected static void TestChange<T>(BindingListExEventSink<T> sink, int index)
		{
			Assert.IsNotNull(sink.ListChanged);
			Assert.IsNotNull(sink.BeforeChange);
			Assert.IsNotNull(sink.AfterChange);

			Assert.AreEqual(ListChangedType.ItemChanged, sink.ListChanged.ListChangedType);
			Assert.AreEqual(index, sink.ListChanged.NewIndex);
			Assert.AreEqual(index, sink.BeforeChange.Index);
			Assert.AreEqual(index, sink.AfterChange.Index);
		}

		protected static void TestMove<T>(BindingListExEventSink<T> sink, int index1, int index2)
		{
			Assert.IsNotNull(sink.ListChanged);
			Assert.IsNotNull(sink.BeforeMove);
			Assert.IsNotNull(sink.AfterMove);

			Assert.AreEqual(ListChangedType.ItemMoved, sink.ListChanged.ListChangedType);
			Assert.AreEqual(index1, sink.ListChanged.OldIndex);
			Assert.AreEqual(index2, sink.ListChanged.NewIndex);
			Assert.AreEqual(index1, sink.BeforeMove.Index1);
			Assert.AreEqual(index2, sink.BeforeMove.Index2);
			Assert.AreEqual(index1, sink.AfterMove.Index1);
			Assert.AreEqual(index2, sink.AfterMove.Index2);
		}

		protected static void TestReset<T>(BindingListExEventSink<T> sink)
		{
			Assert.IsNotNull(sink.ListChanged);
			Assert.IsNotNull(sink.BeforeReset);
			Assert.IsNotNull(sink.AfterReset);

			Assert.AreEqual(ListChangedType.Reset, sink.ListChanged.ListChangedType);
		}
	}

	public class BindingListExEventSink<T>
	{
		#region Fields

		public ListChangedEventArgs ListChanged;
		public BindingListExAddingRemovingEventArgs<T> BeforeAdd;
		public BindingListExAddedRemovedEventArgs<T> AfterAdd;
		public BindingListExAddingRemovingEventArgs<T> BeforeRemove;
		public BindingListExAddedRemovedEventArgs<T> AfterRemove;
		public BindingListExMovingEventArgs<T> BeforeMove;
		public BindingListExMovedEventArgs<T> AfterMove;
		public BindingListExChangingEventArgs<T> BeforeChange;
		public BindingListExChangedEventArgs<T> AfterChange;
		public CancelEventArgs BeforeReset;
		public EventArgs AfterReset;

		#endregion

		#region Methods

		#region Constructor

		public BindingListExEventSink(IBindingListEx<T> list)
		{
			((IBindingList)list).ListChanged += new ListChangedEventHandler(List_ListChanged);
			list.Adding += new EventHandler<BindingListExAddingRemovingEventArgs<T>>(List_BeforeAdd);
			list.Added += new EventHandler<BindingListExAddedRemovedEventArgs<T>>(List_AfterAdd);
			list.Removing += new EventHandler<BindingListExAddingRemovingEventArgs<T>>(List_BeforeRemove);
			list.Removed += new EventHandler<BindingListExAddedRemovedEventArgs<T>>(List_AfterRemove);
			list.Moving += new EventHandler<BindingListExMovingEventArgs<T>>(List_BeforeMove);
			list.Moved += new EventHandler<BindingListExMovedEventArgs<T>>(List_AfterMove);
			list.Changing += new EventHandler<BindingListExChangingEventArgs<T>>(List_BeforeChange);
			list.Changed += new EventHandler<BindingListExChangedEventArgs<T>>(List_AfterChange);
			list.Reseting += new CancelEventHandler(List_BeforeReset);
			list.Reseted += new EventHandler(List_AfterReset);
		}

		#endregion

		#region ListChanged

		private void List_ListChanged(object sender, ListChangedEventArgs e)
		{
			ListChanged = e;
		}

		#endregion

		#region Add

		private void List_BeforeAdd(object sender, BindingListExAddingRemovingEventArgs<T> e)
		{
			BeforeAdd = e;
		}

		private void List_AfterAdd(object sender, BindingListExAddedRemovedEventArgs<T> e)
		{
			AfterAdd = e;
		}

		#endregion

		#region Remove

		private void List_BeforeRemove(object sender, BindingListExAddingRemovingEventArgs<T> e)
		{
			BeforeRemove = e;
		}

		private void List_AfterRemove(object sender, BindingListExAddedRemovedEventArgs<T> e)
		{
			AfterRemove = e;
		}

		#endregion

		#region Move

		private void List_BeforeMove(object sender, BindingListExMovingEventArgs<T> e)
		{
			BeforeMove = e;
		}

		private void List_AfterMove(object sender, BindingListExMovedEventArgs<T> e)
		{
			AfterMove = e;
		}

		#endregion

		#region Change

		private void List_BeforeChange(object sender, BindingListExChangingEventArgs<T> e)
		{
			BeforeChange = e;
		}

		private void List_AfterChange(object sender, BindingListExChangedEventArgs<T> e)
		{
			AfterChange = e;
		}

		#endregion

		#region Reset

		private void List_BeforeReset(object sender, CancelEventArgs e)
		{
			BeforeReset = e;
		}

		private void List_AfterReset(object sender, EventArgs e)
		{
			AfterReset = e;
		}

		#endregion

		#endregion
	}
}
