﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu Library                                  */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Globalization;
using System.IO;
using System.Text;

namespace Nabu
{
	public class CommandLine
	{
		#region Constants

		internal static CommandLineParameterDescriptor[] NoParameters = new CommandLineParameterDescriptor[0];

		#endregion

		#region Fields

		private readonly CommandLineLocalization _localization;
		private readonly IDictionary<string, CommandLineNamedArgumentDescriptor> _namedArgumentDescriptions = new Dictionary<string, CommandLineNamedArgumentDescriptor>();
		private readonly IList<CommandLineUnnamedArgumentDescriptor> _unnamedArgumentDescriptions = new List<CommandLineUnnamedArgumentDescriptor>();
		[SuppressMessage("Microsoft.Design", "CA1006:DoNotNestGenericTypesInMemberSignatures")]
		private IDictionary<string, IList<CommandLineNamedArgument>> _namedArgumentValues = new Dictionary<string, IList<CommandLineNamedArgument>>();
		private readonly IList<CommandLineUnnamedArgument> _unnamedArgumentValues = new List<CommandLineUnnamedArgument>();
		private readonly string[] _arguments;
		private readonly int _minUnnamedArgumentCount;

		#endregion

		#region Methods

		#region Constructors

		public CommandLine(
			string[] arguments)
		{
			this._localization = new CommandLineLocalization_enUS();
			this._arguments = arguments;
		}

		public CommandLine(
			string[] arguments,
			int minUnnamedArgumentCount)
		{
			this._localization = new CommandLineLocalization_enUS();
			this._arguments = arguments;
			this._minUnnamedArgumentCount = minUnnamedArgumentCount;
		}

		public CommandLine(
			int localeID,
			string[] arguments)
		{
			this._localization = CommandLineLocalization.Create(localeID);
			this._arguments = arguments;
		}

		public CommandLine(
			int localeID,
			string[] arguments,
			int minUnnamedArgumentCount)
		{
			this._localization = CommandLineLocalization.Create(localeID);
			this._arguments = arguments;
			this._minUnnamedArgumentCount = minUnnamedArgumentCount;
		}

		#endregion

		#region Print

		private static void SetInfoColor()
		{
#if !COMPACT_FRAMEWORK
			Console.ForegroundColor = ConsoleColor.Gray;
			Console.BackgroundColor = ConsoleColor.Black;
#endif
		}

		private static void SetErrorColor()
		{
#if !COMPACT_FRAMEWORK
			Console.ForegroundColor = ConsoleColor.Red;
			Console.BackgroundColor = ConsoleColor.Black;
#endif
		}

		protected virtual void PrintHeadingUsage()
		{
			Console.Error.Write(Localization.HeadingUsage());
		}

		protected virtual void PrintHeadingParameters()
		{
			Console.Error.WriteLine(Localization.HeadingParameters());
		}

		protected virtual void PrintMessageNamedArgumentIsUnknown(
			string argumentName)
		{
			SetErrorColor();
			Console.Error.WriteLine();
			Console.Error.WriteLine(Localization.MessageNamedArgumentIsUnknown(argumentName));
			SetInfoColor();
		}

		protected virtual void PrintMessageNamedArgumentInstanceCountIsInvalid(
			CommandLineNamedArgumentDescriptor argumentDescription,
			int argumentInstanceCount)
		{
			SetErrorColor();
			Console.Error.WriteLine();

			if (argumentInstanceCount < argumentDescription.MinInstanceCount)
			{
				Console.Error.WriteLine(
					Localization.MessageNamedArgumentShortage(
						argumentDescription.Name,
						argumentInstanceCount,
						argumentDescription.MinInstanceCount));
			}
			else if (argumentInstanceCount > argumentDescription.MaxInstanceCount)
			{
				Console.Error.WriteLine(
					Localization.MessageNamedArgumentExcess(
						argumentDescription.Name,
						argumentInstanceCount,
						argumentDescription.MaxInstanceCount));
			}

			SetInfoColor();
		}

		protected virtual void PrintMessageUnnamedArgumentInstanceCountIsInvalid(
			int argumentInstanceCount)
		{
			SetErrorColor();
			Console.Error.WriteLine();

			if (argumentInstanceCount < MinUnnamedArgumentCount)
			{
				string[] arguments = new string[UnnamedArgumentDescriptions.Count];

				for (int index = 0; index < arguments.Length; index++)
				{
					arguments[index] = UnnamedArgumentDescriptions[index].Name;
				}

				Console.Error.WriteLine(
					Localization.MessageUnnamedArgumentShortage(
						argumentInstanceCount,
						MinUnnamedArgumentCount,
						arguments));
			}

			if (argumentInstanceCount > UnnamedArgumentDescriptions.Count)
			{
				Console.Error.WriteLine(
					Localization.MessageUnnamedArgumentExcess(
						argumentInstanceCount,
						UnnamedArgumentDescriptions.Count));
			}

			SetInfoColor();
		}

		protected virtual void PrintMessageArgumentParameterIsMissing(
			string argumentName,
			IList<CommandLineParameterDescriptor> parameterDescriptions,
			int parameterOffset)
		{
			SetErrorColor();

			string[] parameterNames = new string[parameterDescriptions.Count - parameterOffset];

			for (int index = 0; index < parameterNames.Length; index++)
			{
				parameterNames[index] = parameterDescriptions[index + parameterOffset].Name;
			}

			Console.Error.WriteLine();
			Console.Error.WriteLine(
				Localization.MessageParameterShortage(
					argumentName,
					parameterNames));

			SetInfoColor();
		}

		protected virtual void PrintMessageArgumentParameterTypeIsInvalid(
			string argumentName,
			CommandLineParameterDescriptor parameterDescription)
		{
			SetErrorColor();

			Console.Error.WriteLine();
			Console.Error.WriteLine(
				Localization.MessageParameterTypeIsInvalid(
					argumentName,
					parameterDescription.Name,
					parameterDescription.Type));

			SetInfoColor();
		}

		public virtual void PrintHelpInfo()
		{
			SetInfoColor();

			Console.Error.WriteLine();
			PrintHeadingUsage();
			Console.Error.Write(
				string.Format(
					this._localization.FormatProvider,
					" {0}",
					Path.GetFileName(SpecialPath.ApplicationFile)));

			foreach (CommandLineNamedArgumentDescriptor argumentDescription in NamedArgumentDescriptions.Values)
			{
				StringBuilder consoleOutput = new StringBuilder();

				if (argumentDescription.MinInstanceCount == 0)
				{
					consoleOutput.AppendFormat(
						this._localization.FormatProvider,
						" [-{0}",
						argumentDescription.Name);

					foreach (CommandLineParameterDescriptor parameterDescription in argumentDescription.Parameters)
					{
						consoleOutput.AppendFormat(
							this._localization.FormatProvider,
							" {0}",
							parameterDescription.Name);
					}

					consoleOutput.Append("]");
				}
				else
				{
					for (int index = 0; index < argumentDescription.MinInstanceCount; index++)
					{
						consoleOutput.AppendFormat(
							this._localization.FormatProvider,
							" -{0}",
							argumentDescription.Name);

						foreach (CommandLineParameterDescriptor parameterDescription in argumentDescription.Parameters)
						{
							consoleOutput.AppendFormat(
								this._localization.FormatProvider,
								" {0}",
								parameterDescription.Name);
						}
					}
				}

#if !COMPACT_FRAMEWORK
				if (Console.CursorLeft + consoleOutput.Length > Console.BufferWidth)
				{
					Console.Error.WriteLine();
					Console.Error.Write(new string(' ', Localization.HeadingUsage().Length));
				}
#endif

				Console.Error.Write(consoleOutput.ToString());
			}

			for (int index = 0; index < UnnamedArgumentDescriptions.Count; index++)
			{
				CommandLineUnnamedArgumentDescriptor argumentDescription = UnnamedArgumentDescriptions[index];
				StringBuilder consoleOutput = new StringBuilder();

				if (index < MinUnnamedArgumentCount)
				{
					consoleOutput.AppendFormat(
						this._localization.FormatProvider,
						" {0}",
						argumentDescription.Name);
				}
				else
				{
					consoleOutput.AppendFormat(
						this._localization.FormatProvider,
						" [{0}]",
						argumentDescription.Name);
				}

#if !COMPACT_FRAMEWORK
				if (Console.CursorLeft + consoleOutput.Length > Console.BufferWidth)
				{
					Console.Error.WriteLine();
					Console.Error.Write(new string(' ', Localization.HeadingUsage().Length));
				}
#endif

				Console.Error.Write(consoleOutput.ToString());
			}

			Console.Error.WriteLine();
			Console.Error.WriteLine();

			PrintHeadingParameters();

			Dictionary<CommandLineNamedArgumentDescriptor, string> argumentFullNames = new Dictionary<CommandLineNamedArgumentDescriptor, string>();
			int argumentFullNameMaxLength = 15;

			foreach (CommandLineNamedArgumentDescriptor argumentDescription in NamedArgumentDescriptions.Values)
			{
				if (argumentDescription.Description != null)
				{
					StringBuilder argumentFullNameBuilder = new StringBuilder(argumentDescription.Name);

					foreach (CommandLineParameterDescriptor parameterDescription in argumentDescription.Parameters)
					{
						argumentFullNameBuilder.AppendFormat(
							this._localization.FormatProvider,
							" {0}",
							parameterDescription.Name);
					}

					string argumentFullName = argumentFullNameBuilder.ToString();

					argumentFullNames.Add(argumentDescription, argumentFullName);
					argumentFullNameMaxLength = Math.Max(argumentFullNameMaxLength, argumentFullName.Length);
				}
			}

			foreach (CommandLineNamedArgumentDescriptor argumentDescription in NamedArgumentDescriptions.Values)
			{
				if (argumentDescription.Description != null)
				{
					Console.Error.WriteLine(
						string.Format(
							this._localization.FormatProvider,
							"    -{0} {1}",
							argumentFullNames[argumentDescription].PadRight(argumentFullNameMaxLength),
							argumentDescription.Description));
				}
			}

			Console.Error.WriteLine();

			foreach (CommandLineUnnamedArgumentDescriptor argumentDescription in UnnamedArgumentDescriptions)
			{
				if (argumentDescription.Description != null)
				{
					Console.Error.WriteLine(
						string.Format(
							this._localization.FormatProvider,
							"    {0} {1}",
							argumentDescription.Name.PadRight(argumentFullNameMaxLength + 1),
							argumentDescription.Description));
				}
			}
		}

		#endregion

		#region Configuration

		public void AddNamed(
			string name,
			string description)
		{
			NamedArgumentDescriptions.Add(
				name,
				new CommandLineNamedArgumentDescriptor(
					name,
					description));
		}

		public void AddNamed(
			string name,
			string description,
			bool isMandatory)
		{
			NamedArgumentDescriptions.Add(
				name,
				new CommandLineNamedArgumentDescriptor(
					name,
					description,
					isMandatory));
		}

		public void AddNamed(
			string name,
			string description,
			bool isMandatory,
			params CommandLineParameterDescriptor[] parameters)
		{
			NamedArgumentDescriptions.Add(
				name,
				new CommandLineNamedArgumentDescriptor(
					name,
					description,
					isMandatory,
					parameters));
		}

		public void AddNamed(
			string name,
			string description,
			int minInstanceCount,
			int maxInstanceCount,
			params CommandLineParameterDescriptor[] parameters)
		{
			NamedArgumentDescriptions.Add(
				name,
				new CommandLineNamedArgumentDescriptor(
					name,
					description,
					minInstanceCount,
					maxInstanceCount,
					parameters));
		}

		public void AddUnnamed(
			string name,
			string description)
		{
			UnnamedArgumentDescriptions.Add(
				new CommandLineUnnamedArgumentDescriptor(
					name,
					description));
		}

		#endregion

		#region Parsing

		public bool Parse()
		{
			if (Arguments.Count == 0)
			{
				bool isMandatoryFound = MinUnnamedArgumentCount > 0;

				if (!isMandatoryFound)
				{
					foreach (CommandLineNamedArgumentDescriptor argumentDecription in NamedArgumentDescriptions.Values)
					{
						isMandatoryFound |= (argumentDecription.MinInstanceCount > 0);

						if (isMandatoryFound)
						{
							break;
						}
					}
				}

				if (isMandatoryFound)
				{
					PrintHelpInfo();

					return false;
				}
			}

			bool unnamedFound = false;
			CommandLineNamedArgumentDescriptor argumentDescription = null;
			IList<CommandLineNamedArgument> argumentValues = null;
			CommandLineNamedArgument argumentValue = null;
			int argumentParametersLeft = 0;

			foreach (string argument in Arguments)
			{
				if (argumentParametersLeft == 0)
				{
					if (argument.Length < 2)
					{
						unnamedFound = true;
						UnnamedArgumentValues.Add(new CommandLineUnnamedArgument(argument));
					}
					else
					{
						if ((argument[0] == '-') || (argument[0] == '/'))
						{
							if (unnamedFound)
							{
								UnnamedArgumentValues.Add(new CommandLineUnnamedArgument(argument));
							}
							else
							{
								string argumentName;
								bool argumentAction;

								if ((argument[1] == '-') || (argument[1] == '+'))
								{
									argumentName = argument.Substring(2);
									argumentAction = argument[1] == '+';
								}
								else
								{
									argumentName = argument.Substring(1);
									argumentAction = true;
								}

								if (NamedArgumentDescriptions.TryGetValue(
									argumentName,
									out argumentDescription))
								{
									if (!NamedArgumentValues.TryGetValue(argumentName, out argumentValues))
									{
										argumentValues = new List<CommandLineNamedArgument>();
										NamedArgumentValues.Add(argumentName, argumentValues);
									}

									argumentValue = new CommandLineNamedArgument(argumentAction);
									argumentParametersLeft = argumentDescription.Parameters.Count;

									if (argumentParametersLeft == 0)
									{
										argumentValues.Add(argumentValue);
									}
								}
								else
								{
									PrintMessageNamedArgumentIsUnknown(argumentName);
									PrintHelpInfo();

									return false;
								}
							}
						}
						else
						{
							unnamedFound = true;
							UnnamedArgumentValues.Add(new CommandLineUnnamedArgument(argument));
						}
					}
				}
				else
				{
					CommandLineParameterDescriptor parameterDescription = argumentDescription.Parameters[argumentDescription.Parameters.Count - argumentParametersLeft];
					object parameterValue;

					try
					{
						parameterValue = Convert.ChangeType(
							argument,
							parameterDescription.Type,
							CultureInfo.CurrentCulture.NumberFormat);
					}
					catch (InvalidCastException)
					{
						PrintMessageArgumentParameterTypeIsInvalid(
							argumentDescription.Name,
							parameterDescription);
						PrintHelpInfo();

						return false;
					}

					argumentValue.ParameterValues.Add(parameterValue);
					argumentParametersLeft--;

					if (argumentParametersLeft == 0)
					{
						argumentValues.Add(argumentValue);
					}
				}
			}

			if (argumentParametersLeft > 0)
			{
				PrintMessageArgumentParameterIsMissing(
					argumentDescription.Name,
					argumentDescription.Parameters,
					argumentDescription.Parameters.Count - argumentParametersLeft);
				PrintHelpInfo();

				return false;
			}

			foreach (CommandLineNamedArgumentDescriptor argumentDecription in NamedArgumentDescriptions.Values)
			{
				if (NamedArgumentValues.TryGetValue(argumentDecription.Name, out argumentValues))
				{
					if (
						(argumentValues.Count < argumentDecription.MinInstanceCount) ||
						(argumentValues.Count > argumentDecription.MaxInstanceCount))
					{
						PrintMessageNamedArgumentInstanceCountIsInvalid(
							argumentDecription,
							argumentValues.Count);
						PrintHelpInfo();

						return false;
					}
				}
				else
				{
					if (argumentDecription.MinInstanceCount > 0)
					{
						PrintMessageNamedArgumentInstanceCountIsInvalid(
							argumentDecription,
							0);
						PrintHelpInfo();

						return false;
					}
				}
			}

			if (UnnamedArgumentValues.Count < MinUnnamedArgumentCount)
			{
				PrintMessageUnnamedArgumentInstanceCountIsInvalid(UnnamedArgumentValues.Count);
				PrintHelpInfo();

				return false;
			}

			if (UnnamedArgumentValues.Count > UnnamedArgumentDescriptions.Count)
			{
				PrintMessageUnnamedArgumentInstanceCountIsInvalid(UnnamedArgumentValues.Count);
				PrintHelpInfo();

				return false;
			}

			return true;
		}

		#endregion

		#region Retrival

		public IList<CommandLineNamedArgument> GetNamedValues(string argumentName)
		{
			IList<CommandLineNamedArgument> argumentValues;

			if (NamedArgumentValues.TryGetValue(argumentName, out argumentValues))
			{
				if (argumentValues.Count > 0)
				{
					return argumentValues;
				}
			}

			return null;
		}

		public CommandLineNamedArgument GetNamedValue(string argumentName)
		{
			IList<CommandLineNamedArgument> argumentValues;

			if (NamedArgumentValues.TryGetValue(argumentName, out argumentValues))
			{
				if (argumentValues.Count > 0)
				{
					return argumentValues[0];
				}
			}

			return null;
		}

		public bool GetNamedValueAction(string argumentName, bool defaultAction)
		{
			CommandLineNamedArgument argumentValue = GetNamedValue(argumentName);

			return (argumentValue == null) ? defaultAction : argumentValue.Action;
		}

		public object GetNamedValueParameter(string argumentName)
		{
			CommandLineNamedArgument argumentValue = GetNamedValue(argumentName);

			return (argumentValue == null) ? null : argumentValue.ParameterValues[0];
		}

		[SuppressMessage("Microsoft.Design", "CA1024")]
		public IList<CommandLineUnnamedArgument> GetUnnamedValues()
		{
			if (UnnamedArgumentValues.Count > 0)
			{
				return UnnamedArgumentValues;
			}

			return null;
		}

		#endregion

		#endregion

		#region Properties

		protected CommandLineLocalization Localization
		{
			[DebuggerStepThrough]
			get { return this._localization; }
		}

		protected IDictionary<string, CommandLineNamedArgumentDescriptor> NamedArgumentDescriptions
		{
			[DebuggerStepThrough]
			get { return this._namedArgumentDescriptions; }
		}

		protected IList<CommandLineUnnamedArgumentDescriptor> UnnamedArgumentDescriptions
		{
			[DebuggerStepThrough]
			get { return this._unnamedArgumentDescriptions; }
		}

		[SuppressMessage("Microsoft.Design", "CA1006")]
		protected IDictionary<string, IList<CommandLineNamedArgument>> NamedArgumentValues
		{
			[DebuggerStepThrough]
			get { return this._namedArgumentValues; }
		}

		protected IList<CommandLineUnnamedArgument> UnnamedArgumentValues
		{
			[DebuggerStepThrough]
			get { return this._unnamedArgumentValues; }
		}

		protected IList<string> Arguments
		{
			[DebuggerStepThrough]
			get { return this._arguments; }
		}

		protected int MinUnnamedArgumentCount
		{
			[DebuggerStepThrough]
			get { return this._minUnnamedArgumentCount; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
