﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu Library                                  */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics;
using System.Globalization;
using System.Text;

using Nabu.Text;

namespace Nabu
{
	internal abstract class CommandLineLocalization_en : CommandLineLocalization
	{
		#region CommandLineLocalization_en

		#region Constants

		private static readonly string[] ArgumentForms = new string[]
		{
			"argument was",
			"arguments were"
		};

		private static readonly string[] InstanceForms = new string[]
		{
			"instance",
			"instances"
		};

		private static readonly string[] ParameterForms = new string[]
		{
			"parameter",
			"parameters"
		};

		private static readonly string[] TimeForms = new string[]
		{
			"time",
			"times"
		};

		#endregion

		#region Fields

		private readonly NumeralFormatter _numeralFormatter;
		private readonly CultureInfo _cultureInfo;

		#endregion

		#region Methods

		protected CommandLineLocalization_en(int localeID)
		{
			this._numeralFormatter = NumeralFormatter.Create(localeID);
			this._cultureInfo = CultureInfo.GetCultureInfo(localeID);
		}

		private int WordFormIndexFromNumeral(long value)
		{
			switch (this._numeralFormatter.NounFormFromNumeral(
				value,
				NumeralType.Cardinal,
				WordGender.Neuter,
				WordForms.EN_Default))
			{
				case WordForms.EN_Singular:
					return 0;
				case WordForms.EN_Plural:
					return 1;
			}

			return 0;
		}

		private string NumeralNameFromValue(long value)
		{
			return this._numeralFormatter.NumeralNameFromValue(
				value,
				NumeralType.Cardinal,
				WordGender.Neuter,
				WordForms.EN_Default,
				false);
		}

		#endregion

		#endregion

		#region CommandLineLocalization

		#region Methods

		public override string HeadingUsage()
		{
			return "Usage:";
		}

		public override string HeadingParameters()
		{
			return "Parameters:";
		}

		public override string MessageNamedArgumentIsUnknown(string argumentName)
		{
			return string.Format(
				this._cultureInfo,
				"Argument '{0}' is unknown.",
				argumentName);
		}

		public override string MessageNamedArgumentShortage(string argumentName, int count, int limit)
		{
			if (count == 0)
			{
				return string.Format(
					this._cultureInfo,
					"Argument '{0}' was not specified. At least {1} argument {2} expected.",
					argumentName,
					NumeralNameFromValue(limit),
					InstanceForms[WordFormIndexFromNumeral(limit)]);
			}
			else
			{
				return string.Format(
					this._cultureInfo,
					"Argument '{0}' was specified {1} {2}. At least {3} argument {4} expected.",
					argumentName,
					NumeralNameFromValue(count),
					TimeForms[WordFormIndexFromNumeral(count)],
					NumeralNameFromValue(limit),
					InstanceForms[WordFormIndexFromNumeral(limit)]);
			}
		}

		public override string MessageNamedArgumentExcess(string argumentName, int count, int limit)
		{
			return string.Format(
				this._cultureInfo,
				"Argument '{0}' was specified {1} {2}. At most {3} argument {4} expected.",
				argumentName,
				NumeralNameFromValue(count),
				TimeForms[WordFormIndexFromNumeral(count)],
				NumeralNameFromValue(limit),
				InstanceForms[WordFormIndexFromNumeral(limit)]);
		}

		public override string MessageUnnamedArgumentShortage(int count, int limit, string[] arguments)
		{
			StringBuilder result = new StringBuilder();

			if (count == 0)
			{
				result.AppendFormat(
					this._cultureInfo,
					"No unnamed argument was specified. At least {0} unnamed {1} expected ({2}",
					NumeralNameFromValue(limit),
					ArgumentForms[WordFormIndexFromNumeral(limit)],
					arguments[0]);
			}
			else
			{
				result.AppendFormat(
					this._cultureInfo,
					"{0} unnamed {1} specified. At least {2} unnamed {3} expected ({4}",
					WordCase.ToTitle(NumeralNameFromValue(count), _cultureInfo),
					ArgumentForms[WordFormIndexFromNumeral(count)],
					NumeralNameFromValue(limit),
					ArgumentForms[WordFormIndexFromNumeral(limit)],
					arguments[0]);
			}

			for (int index = 1; index < arguments.Length; index++)
			{
				result.AppendFormat(
					this._cultureInfo,
					", {0}",
					arguments[index]);
			}

			result.Append(").");

			return result.ToString();
		}

		public override string MessageUnnamedArgumentExcess(int count, int limit)
		{
			if (limit == 0)
			{
				return string.Format(
					this._cultureInfo,
					"{0} unnamed {1} specified. No unnamed arguments expected.",
					WordCase.ToTitle(NumeralNameFromValue(count), _cultureInfo),
					ArgumentForms[WordFormIndexFromNumeral(count)]);
			}
			else
			{
				return string.Format(
					this._cultureInfo,
					"{0} unnamed {1} specified. At most {2} unnamed {3} expected.",
					WordCase.ToTitle(NumeralNameFromValue(count), _cultureInfo),
					ArgumentForms[WordFormIndexFromNumeral(count)],
					NumeralNameFromValue(limit),
					ArgumentForms[WordFormIndexFromNumeral(limit)]);
			}
		}

		public override string MessageParameterShortage(string argumentName, string[] parameterNames)
		{
			StringBuilder result = new StringBuilder();

			result.AppendFormat(
				this._cultureInfo,
				"Argument '{0}' is missing {1} '{2}'",
				argumentName,
				ParameterForms[WordFormIndexFromNumeral(parameterNames.Length)],
				parameterNames[0]);

			for (int index = 1; index < parameterNames.Length; index++)
			{
				result.AppendFormat(
					this._cultureInfo,
					", '{0}'",
					parameterNames[index]);
			}

			result.Append(".");

			return result.ToString();
		}

		public override string MessageParameterTypeIsInvalid(string argumentName, string parameterName, Type parameterType)
		{
			if (parameterType == null)
			{
				throw new ArgumentNullException("parameterType");
			}

			string parameterTypeName;

			if (parameterType == typeof(sbyte))
			{
				parameterTypeName = string.Format(
					this._cultureInfo,
					"integer from {0} to {1}",
					sbyte.MinValue,
					sbyte.MaxValue);
			}
			else if (parameterType == typeof(byte))
			{
				parameterTypeName = string.Format(
					this._cultureInfo,
					"integer from {0} to {1}",
					byte.MinValue,
					byte.MaxValue);
			}
			else if (parameterType == typeof(short))
			{
				parameterTypeName = string.Format(
					this._cultureInfo,
					"integer from {0} to {1}",
					short.MinValue,
					short.MaxValue);
			}
			else if (parameterType == typeof(ushort))
			{
				parameterTypeName = string.Format(
					this._cultureInfo,
					"integer from {0} to {1}",
					ushort.MinValue,
					ushort.MaxValue);
			}
			else if (parameterType == typeof(int))
			{
				parameterTypeName = string.Format(
					this._cultureInfo,
					"integer from {0} to {1}",
					int.MinValue,
					int.MaxValue);
			}
			else if (parameterType == typeof(uint))
			{
				parameterTypeName = string.Format(
					this._cultureInfo,
					"integer from {0} to {1}",
					uint.MinValue,
					uint.MaxValue);
			}
			else if (parameterType == typeof(long))
			{
				parameterTypeName = string.Format(
					this._cultureInfo,
					"integer from {0} to {1}",
					long.MinValue,
					long.MaxValue);
			}
			else if (parameterType == typeof(ulong))
			{
				parameterTypeName = string.Format(
					this._cultureInfo,
					"integer from {0} to {1}",
					ulong.MinValue,
					ulong.MaxValue);
			}
			else
			{
				parameterTypeName = string.Format(
					this._cultureInfo,
					"convertible to type '{0}'",
					parameterType.FullName);
			}

			return string.Format(
				this._cultureInfo,
				"Parameter '{1}' of argument '{0}' must be {2}",
				argumentName,
				parameterName,
				parameterTypeName);
		}

		#endregion

		#region Properties

		public override IFormatProvider FormatProvider
		{
			[DebuggerStepThrough]
			get { return this._cultureInfo; }
		}

		#endregion

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
