﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu Library                                  */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics;
using System.IO;
using System.Reflection;

namespace Nabu
{
	/// <summary>
	/// Provides information about assembly version.
	/// </summary>
	public sealed class AssemblyVersionInfo
	{
		#region Fields

		private readonly string _companyName;
		private readonly string _productName;
		private readonly string _productDescription;
		private readonly string _productVersion;
		private readonly string _productCopyright;

		#endregion

		#region Methods

		private AssemblyVersionInfo(string path, int fieldCount)
		{
			try
			{
				this._companyName = string.Empty;
				this._productName = string.Empty;
				this._productDescription = string.Empty;
				this._productCopyright = string.Empty;

#if !COMPACT_FRAMEWORK
				Assembly assembly = Assembly.ReflectionOnlyLoadFrom(path);

				foreach (CustomAttributeData customAttributeData in CustomAttributeData.GetCustomAttributes(assembly))
				{
					if (customAttributeData.Constructor.DeclaringType == typeof(AssemblyCompanyAttribute))
					{
						this._companyName = customAttributeData.ConstructorArguments[0].Value as string;
					}
					else if (customAttributeData.Constructor.DeclaringType == typeof(AssemblyProductAttribute))
					{
						this._productName = customAttributeData.ConstructorArguments[0].Value as string;
					}
					else if (customAttributeData.Constructor.DeclaringType == typeof(AssemblyDescriptionAttribute))
					{
						this._productDescription = customAttributeData.ConstructorArguments[0].Value as string;
					}
					else if (customAttributeData.Constructor.DeclaringType == typeof(AssemblyCopyrightAttribute))
					{
						this._productCopyright = customAttributeData.ConstructorArguments[0].Value as string;
					}
				}

#else

				Assembly assembly = Assembly.LoadFrom(path);

#endif

				this._productVersion = assembly.GetName().Version.ToString(fieldCount);
			}
			catch (FileNotFoundException)
			{
				this._companyName = string.Empty;
				this._productName = string.Empty;
				this._productDescription = string.Empty;
				this._productCopyright = string.Empty;
				this._productVersion = string.Empty;
			}

#if !COMPACT_FRAMEWORK
			catch (FileLoadException)
			{
				this._companyName = string.Empty;
				this._productName = string.Empty;
				this._productDescription = string.Empty;
				this._productCopyright = string.Empty;
				this._productVersion = string.Empty;
			}
#endif

			catch (BadImageFormatException)
			{
				this._companyName = string.Empty;
				this._productName = string.Empty;
				this._productDescription = string.Empty;
				this._productCopyright = string.Empty;
				this._productVersion = string.Empty;
			}
		}

		/// <summary>
		/// Returns assembly version information for an assembly with given path.
		/// </summary>
		/// <param name="path">Path to assembly file.</param>
		/// <param name="fieldCount">Number of fields in version string. Must be from 1 to 4.</param>
		/// <returns>AssemblyVersionInfo</returns>
		[DebuggerStepThrough]
		public static AssemblyVersionInfo GetVersionInfo(string path, int fieldCount)
		{
			return new AssemblyVersionInfo(path, fieldCount);
		}

		/// <summary>
		///  Returns assembly version information for an assembly with given path.
		/// </summary>
		/// <param name="path">Path to assembly file.</param>
		/// <returns></returns>
		[DebuggerStepThrough]
		public static AssemblyVersionInfo GetVersionInfo(string path)
		{
			return new AssemblyVersionInfo(path, 4);
		}

		#endregion

		#region Properties

		/// <summary>
		/// Company name.
		/// </summary>
		public string CompanyName
		{
			[DebuggerStepThrough]
			get { return this._companyName; }
		}

		/// <summary>
		/// Product name.
		/// </summary>
		public string ProductName
		{
			[DebuggerStepThrough]
			get { return this._productName; }
		}

		/// <summary>
		/// Product description.
		/// </summary>
		public string ProductDescription
		{
			[DebuggerStepThrough]
			get { return this._productDescription; }
		}

		/// <summary>
		/// Product version.
		/// </summary>
		public string ProductVersion
		{
			[DebuggerStepThrough]
			get { return this._productVersion; }
		}

		/// <summary>
		/// Product copyright.
		/// </summary>
		public string ProductCopyright
		{
			[DebuggerStepThrough]
			get { return this._productCopyright; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
