﻿/******************************************************************************/
/*                                                                            */
/*    _                ___        _..-._                                      */
/*    \`.|\..----...-'`   `-._.-'' _.-..'                                     */
/*    /  ' `         ,       __.-''                                           */
/*    )/` _/     \   `-_,   /                                                 */
/*    `-'" `"\_  ,_.-;_.-\_ ',  NABU Project                                  */
/*        _.-'_./   {_.'   ; /  Nabu Library                                  */
/*       {_.-``-'         {_/   Copyright © 2005-2009, TrifleSoft             */
/*                                                                            */
/******************************************************************************/
using System;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Text;

namespace Nabu
{
	/// <summary>
	/// Provides information about predefined paths.
	/// </summary>
	public sealed class SpecialPath
	{
		#region Fields

		private static readonly SpecialPath _specialPath = new SpecialPath();
		private readonly AssemblyVersionInfo _applicationVersion;
		private readonly string _applicationFile;
		private readonly string _applicationPath;
		private readonly string _currentUserAllMachinesSettingsPath;
		private readonly string _allUsersCurrentMachineSettingsPath;
		private readonly string _currentUserCurrentMachineSettingsPath;

		#endregion

		#region Methods

		private SpecialPath()
		{
			StringBuilder pathApplicationExecutable = new StringBuilder(260);

			if (NativeMethods.GetModuleFileNameW(IntPtr.Zero, pathApplicationExecutable, pathApplicationExecutable.Capacity) == 0)
			{
				throw NativeException.Create();
			}

			this._applicationFile = pathApplicationExecutable.ToString();
			this._applicationPath = Path.GetDirectoryName(this._applicationFile);

			this._applicationVersion = AssemblyVersionInfo.GetVersionInfo(this._applicationFile, 3);
			string format;

			if (this._applicationVersion.CompanyName == null)
			{
				format = "{1}";
			}
			else
			{
				if (this._applicationVersion.ProductName == null)
				{
					format = "{1}{0}{2}";
				}
				else
				{
					if (this._applicationVersion.ProductVersion == null)
					{
						format = "{1}{0}{2}{0}{3}";
					}
					else
					{
						format = "{1}{0}{2}{0}{3}{0}{4}";
					}
				}
			}

			this._currentUserAllMachinesSettingsPath = string.Format(
				CultureInfo.CurrentCulture.NumberFormat,
				format,
				Path.DirectorySeparatorChar,
				Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData),
				this._applicationVersion.CompanyName,
				this._applicationVersion.ProductName,
				this._applicationVersion.ProductVersion);
			this._allUsersCurrentMachineSettingsPath = string.Format(
				CultureInfo.CurrentCulture.NumberFormat,
				format,
				Path.DirectorySeparatorChar,
				Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData),
				this._applicationVersion.CompanyName,
				this._applicationVersion.ProductName,
				this._applicationVersion.ProductVersion);
			this._currentUserCurrentMachineSettingsPath = string.Format(
				CultureInfo.CurrentCulture.NumberFormat,
				format,
				Path.DirectorySeparatorChar,
				Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData),
				this._applicationVersion.CompanyName,
				this._applicationVersion.ProductName,
				this._applicationVersion.ProductVersion);
		}

		#endregion

		#region Properties

#if !COMPACT_FRAMEWORK
		/// <summary>
		/// Application version of currently executing executable.
		/// </summary>
		public static AssemblyVersionInfo ApplicationVersion
		{
			[DebuggerStepThrough]
			get { return _specialPath._applicationVersion; }
		}
#endif

		/// <summary>
		/// Path of file of currently executing executable.
		/// </summary>
		public static string ApplicationFile
		{
			[DebuggerStepThrough]
			get { return _specialPath._applicationFile; }
		}

		/// <summary>
		/// Path of directory of currently executing executable.
		/// </summary>
		public static string ApplicationPath
		{
			[DebuggerStepThrough]
			get { return _specialPath._applicationPath; }
		}

		/// <summary>
		/// Path folder where roaming settings for current user must be placed.
		/// </summary>
		public static string CurrentUserAllMachinesSettingsPath
		{
			[DebuggerStepThrough]
			get { return _specialPath._currentUserAllMachinesSettingsPath; }
		}

		/// <summary>
		/// Path folder where non-roaming settings for all users must be placed.
		/// </summary>
		public static string AllUsersCurrentMachineSettingsPath
		{
			[DebuggerStepThrough]
			get { return _specialPath._allUsersCurrentMachineSettingsPath; }
		}

		/// <summary>
		/// Path folder where non-roaming settings for current user must be placed.
		/// </summary>
		public static string CurrentUserCurrentMachineSettingsPath
		{
			[DebuggerStepThrough]
			get { return _specialPath._currentUserCurrentMachineSettingsPath; }
		}

		#endregion
	}
}
/******************************************************************************/
/*                                END OF FILE                                 */
/******************************************************************************/
