﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;

namespace Nabu.SvnToVersion
{
	public sealed class FileProcessor
	{
		#region Fields

		private Encoding _encoding;
		private List<KeyValuePair<string, string>> _regexes = new List<KeyValuePair<string, string>>();

		#endregion

		#region Methods

		public FileProcessor(string encodingName)
		{
			if (encodingName == null)
			{
				this._encoding = Encoding.UTF8;
			}
			else
			{
				this._encoding = Encoding.GetEncoding(encodingName);
			}
		}

		public void LoadConfiguration(
			string filePath)
		{
			switch (filePath)
			{
				case "?.net":
					this._regexes.Add(new KeyValuePair<string, string>(
						@"\[assembly:\s*AssemblyVersion\(\""(\d+).(\d+).(\d+).(\d+|\*)""\)\]",
						@"[assembly: AssemblyVersion(""$1.$2.$mrh.$mrl"")]"));
					this._regexes.Add(new KeyValuePair<string, string>(
						@"\[assembly:\s*AssemblyFileVersion\(\""(\d+).(\d+).(\d+).(\d+|\*)""\)\]",
						@"[assembly: AssemblyFileVersion(""$1.$2.$mrh.$mrl"")]"));

					this._regexes.Add(new KeyValuePair<string, string>(
						@"\[assembly:\s*AssemblyVersionAttribute\(\""(\d+).(\d+).(\d+).(\d+|\*)""\)\]",
						@"[assembly: AssemblyVersionAttribute(""$1.$2.$mrh.$mrl"")]"));
					this._regexes.Add(new KeyValuePair<string, string>(
						@"\[assembly:\s*AssemblyFileVersionAttribute\(\""(\d+).(\d+).(\d+).(\d+|\*)""\)\]",
						@"[assembly: AssemblyFileVersionAttribute(""$1.$2.$mrh.$mrl"")]"));

					this._regexes.Add(new KeyValuePair<string, string>(
						@"\<Assembly:\s*AssemblyVersion\(\""(\d+).(\d+).(\d+).(\d+|\*)""\)\>",
						@"<Assembly: AssemblyVersion(""$1.$2.$mrh.$mrl"")>"));
					this._regexes.Add(new KeyValuePair<string, string>(
						@"\<Assembly:\s*AssemblyFileVersion\(\""(\d+).(\d+).(\d+).(\d+|\*)""\)\>",
						@"<Assembly: AssemblyFileVersion(""$1.$2.$mrh.$mrl"")>"));

					this._regexes.Add(new KeyValuePair<string, string>(
						@"\<Assembly:\s*AssemblyVersionAttribute\(\""(\d+).(\d+).(\d+).(\d+|\*)""\)\>",
						@"<Assembly: AssemblyVersionAttribute(""$1.$2.$mrh.$mrl"")>"));
					this._regexes.Add(new KeyValuePair<string, string>(
						@"\<Assembly:\s*AssemblyFileVersionAttribute\(\""(\d+).(\d+).(\d+).(\d+|\*)""\)\>",
						@"<Assembly: AssemblyFileVersionAttribute(""$1.$2.$mrh.$mrl"")>"));
					break;
				case "?rc":
					this._regexes.Add(new KeyValuePair<string, string>(
						@"PRODUCTVERSION (\d+)[,\.]\s*(\d+)[,\.]\s*(\d+)[,\.]\s*(\d+)",
						@"PRODUCTVERSION $1,$2,$prh,$prl"));
					this._regexes.Add(new KeyValuePair<string, string>(
						@"FILEVERSION (\d+)[,\.]\s*(\d+)[,\.]\s*(\d+)[,\.]\s*(\d+)",
						@"FILEVERSION $1,$2,$mrh,$mrl"));

					this._regexes.Add(new KeyValuePair<string, string>(
						@"VALUE ""ProductVersion"", ""(\d+)[,\.]\s*(\d+)[,\.]\s*(\d+)[,\.]\s*(\d+)""",
						@"VALUE ""ProductVersion"", ""$1, $2, $prh, $prl"""));
					this._regexes.Add(new KeyValuePair<string, string>(
						@"VALUE ""FileVersion"", ""(\d+)[,\.]\s*(\d+)[,\.]\s*(\d+)[,\.]\s*(\d+)""",
						@"VALUE ""FileVersion"", ""$1, $2, $mrh, $mrl"""));
					break;
				default:
					{
						string[] lines;

						using (TextReader fileReader = new StreamReader(filePath, this._encoding))
						{
							lines = fileReader.ReadToEnd().Split(new char[] { '\r', '\n' }, StringSplitOptions.RemoveEmptyEntries);
						}

						for (int index = 0; index < lines.Length; index += 2)
						{
							this._regexes.Add(new KeyValuePair<string, string>(lines[index], lines[index + 1]));
						}
					}
					break;
			}
		}

		public string ProcessContent(
			string fileContent,
			int productRevision,
			int moduleRevision)
		{
			string productRevisionLow = (productRevision & 0xFFFF).ToString();
			string productRevisionHigh = (productRevision >> 16).ToString();
			string productRevisionFull = productRevision.ToString();
			string moduleRevisionLow = (moduleRevision & 0xFFFF).ToString();
			string moduleRevisionHigh = (moduleRevision >> 16).ToString();
			string moduleRevisionFull = moduleRevision.ToString();

			foreach (KeyValuePair<string, string> pair in this._regexes)
			{
				fileContent = Regex.Replace(
					fileContent,
					pair.Key,
					pair.Value
						.Replace("$prl", productRevisionLow)
						.Replace("$prh", productRevisionHigh)
						.Replace("$prf", productRevisionFull)
						.Replace("$mrl", moduleRevisionLow)
						.Replace("$mrh", moduleRevisionHigh)
						.Replace("$mrf", moduleRevisionFull),
					RegexOptions.Multiline);
			}

			return fileContent.Trim();
		}

		public void ProcessFile(
			string filePath,
			int productRevision,
			int moduleRevision)
		{
			Console.WriteLine("Updating '{0}'.", filePath);

			string fileContent;

			using (TextReader fileReader = new StreamReader(filePath, this._encoding))
			{
				fileContent = fileReader.ReadToEnd();
			}

			fileContent = ProcessContent(
				fileContent,
				productRevision,
				moduleRevision);

			using (TextWriter textWriter = new StreamWriter(filePath, false, this._encoding))
			{
				textWriter.WriteLine(fileContent);
			}
		}

		public void ProcessDirectory(
			string rootDirectoryPath,
			string fileMask,
			int productRevision,
			int moduleRevision)
		{
			foreach (string directoryPath in Directory.GetDirectories(rootDirectoryPath))
			{
				ProcessDirectory(
					directoryPath,
					fileMask,
					productRevision,
					moduleRevision);
			}

			foreach (string pathFile in Directory.GetFiles(rootDirectoryPath, fileMask))
			{
				ProcessFile(
					pathFile,
					productRevision,
					moduleRevision);
			}
		}

		#endregion
	}
}
